/* Borders.java

	Purpose:
		
	Description:
		
	History:
		11:44 AM 2019/9/24, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import io.keikai.model.SBorder;

/**
 * @author jumperchen
 */
public class Borders {
	Map<Border, Border> _bordersMap = new HashMap();
	List<Border> _borders = new ArrayList<>();
	public Borders() {}

	public void addBorder(Border border) {
		// keep the same instance to reduce memory footprint
		Border exist = _bordersMap.get(border);
		if (exist != null) {
			_borders.add(exist);
		} else {
			_bordersMap.put(border, border);
			_borders.add(border);
		}
	}
	public Border getBordreAt(int index) {
		return _borders.size() > index ? _borders.get(index) : null;
	}
	public static class Border {
		private boolean diagonalUp;
		private boolean diagonalDown;
		private boolean outline;
		private BorderPr left;
		private BorderPr top;
		private BorderPr right;
		private BorderPr bottom;
		private BorderPr diagonal;
		private BorderPr vertical;
		private BorderPr horizontal;

		public Border(boolean diagonalUp, boolean diagonalDown, boolean outline,
				BorderPr left, BorderPr top, BorderPr right, BorderPr bottom,
				BorderPr diagonal, BorderPr vertical, BorderPr horizontal) {
			this.diagonalUp = diagonalUp;
			this.diagonalDown = diagonalDown;
			this.outline = outline;
			this.left = left;
			this.top = top;
			this.right = right;
			this.bottom = bottom;
			this.diagonal = diagonal;
			this.vertical = vertical;
			this.horizontal = horizontal;
		}

		public boolean isDiagonalUp() {
			return diagonalUp;
		}

		public boolean isDiagonalDown() {
			return diagonalDown;
		}

		public boolean isOutline() {
			return outline;
		}

		public BorderPr getLeft() {
			return left;
		}

		public BorderPr getTop() {
			return top;
		}

		public BorderPr getRight() {
			return right;
		}

		public BorderPr getBottom() {
			return bottom;
		}

		public BorderPr getDiagonal() {
			return diagonal;
		}

		public BorderPr getVertical() {
			return vertical;
		}

		public BorderPr getHorizontal() {
			return horizontal;
		}

		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;
			Border border = (Border) o;
			return diagonalUp == border.diagonalUp
					&& diagonalDown == border.diagonalDown
					&& outline == border.outline && Objects
					.equals(left, border.left) && Objects
					.equals(top, border.top) && Objects
					.equals(right, border.right) && Objects
					.equals(bottom, border.bottom) && Objects
					.equals(diagonal, border.diagonal) && Objects
					.equals(vertical, border.vertical) && Objects
					.equals(horizontal, border.horizontal);
		}

		public int hashCode() {
			return Objects
					.hash(diagonalUp, diagonalDown, outline, left, top, right,
							bottom, diagonal, vertical, horizontal);
		}
	}
	public static class BorderPr {
		private SBorder.BorderType style;
		private String color;
		public BorderPr(String style, String color) {
			style = style == null ? "none" : style;
			switch (style) {
			case "none":
				this.style = SBorder.BorderType.NONE;
				break;
			case "thin":
				this.style = SBorder.BorderType.THIN;
				break;
			case "medium":
				this.style = SBorder.BorderType.MEDIUM;
				break;
			case "dashed":
				this.style = SBorder.BorderType.DASHED;
				break;
			case "dotted":
				this.style = SBorder.BorderType.DOTTED;
				break;
			case "thick":
				this.style = SBorder.BorderType.THICK;
				break;
			case "double":
				this.style = SBorder.BorderType.DOUBLE;
				break;
			case "hair":
				this.style = SBorder.BorderType.HAIR;
				break;
			case "mediumDashed":
				this.style = SBorder.BorderType.MEDIUM_DASHED;
				break;
			case "dashDot":
				this.style = SBorder.BorderType.DASH_DOT;
				break;
			case "mediumDashDot":
				this.style = SBorder.BorderType.MEDIUM_DASH_DOT;
				break;
			case "dashDotDot":
				this.style = SBorder.BorderType.DASH_DOT_DOT;
				break;
			case "mediumDashDotDot":
				this.style = SBorder.BorderType.MEDIUM_DASH_DOT_DOT;
				break;
			case "slantDashDot":
				this.style = SBorder.BorderType.SLANTED_DASH_DOT;
				break;
			}
			this.color = color;
		}

		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;
			BorderPr borderPr = (BorderPr) o;
			return style == borderPr.style && Objects
					.equals(color, borderPr.color);
		}

		public int hashCode() {
			return Objects.hash(style, color);
		}

		public SBorder.BorderType getStyle() {
			return style;
		}

		public String getColor() {
			return color;
		}
	}
}
