/* Colors.java

	Purpose:
		
	Description:
		
	History:
		10:18 AM 2019/9/24, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import static io.keikai.util.Converter.asList;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Colors responsible for manage colors in a workbook.
 * see http://www.datypic.com/sc/ooxml/t-ssml_CT_Colors.html
 * @author jumperchen
 */
public class Colors {

	private final static java.util.List<String> BUILTIN_INDEXED_COLORS = asList("ff000000", "ffffffff", "ffff0000", "ff00ff00", "ff0000ff", "ffffff00", "ffff00ff", "ff00ffff", "ff000000", "ffffffff", "ffff0000", "ff00ff00", "ff0000ff", "ffffff00", "ffff00ff", "ff00ffff", "ff800000", "ff008000", "ff000080", "ff808000", "ff800080", "ff008080", "ffc0c0c0", "ff808080", "ff9999ff", "ff993366", "ffffffcc", "ffccffff", "ff660066", "ffff8080", "ff0066cc", "ffccccff", "ff000080", "ffff00ff", "ffffff00", "ff00ffff", "ff800080", "ff800000", "ff008080", "ff0000ff", "ff00ccff", "ffccffff", "ffccffcc", "ffffff99", "ff99ccff", "ffff99cc", "ffcc99ff", "ffffcc99", "ff3366ff", "ff33cccc", "ff99cc00", "ffffcc00", "ffff9900", "ffff6600", "ff666699", "ff969696", "ff003366", "ff339966", "ff003300", "ff333300", "ff993300", "ff993366", "ff333399", "ff333333", "ff000000", "ffffffff");
	private List<String> _indexColors = new ArrayList<>();
	private Map<String, String> _colorsMap = new HashMap<>();

	private boolean ignoreBuiltin;

	public Colors() {
		this(false);
	}

	public Colors(boolean ignoreBuiltin) {
		this.ignoreBuiltin = ignoreBuiltin;
		if (!this.ignoreBuiltin) {
			for (final String argb : BUILTIN_INDEXED_COLORS) {
				this.addIndexedColor(argb);
			}
		}
	}

	public String getIndexColorByIndex(int index) {
		// TODO: KEIKAI-48 ignore index 64, 65 when supporting theme, e.g. SchemeColorVal.tx1, SchemeColorVal.bg1
		return index >= this._indexColors.size()? null : this._indexColors.get(index);
	}

	/**
	 * Add color if not present.
	 * @param argb
	 */
	public void addColor(String argb) {
		if (!_colorsMap.containsKey(argb)) {
			addIndexedColor(argb);
		}
	}

	public void addIndexedColor(String argb) {
		// keep the same instance to reduce memory footprint
		String exist = _colorsMap.get(argb);
		if (exist != null) {
			this._indexColors.add(exist);
		} else {
			this._indexColors.add(argb);
			this._colorsMap.put(argb, argb);
		}
	}
}