/* Fills.java

	Purpose:
		
	Description:
		
	History:
		11:18 AM 2019/9/24, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import static io.keikai.model.SFill.FillPattern;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * @author jumperchen
 */
public class Fills {
	private Map<Fill, Fill> _fillsMap = new HashMap();
	private List<Fill> _fills = new ArrayList<>();

	public <T extends Fill> T getFillAt(int index) {
		return _fills.size() > index ? (T) _fills.get(index) : null;
	}

	public void addFill(Fill fill) {
		// keep the same instance to reduce memory footprint
		Fill exist = _fillsMap.get(fill);
		if (exist != null) {
			_fills.add(exist);
		} else {
			_fillsMap.put(fill, fill);
			_fills.add(fill);
		}
	}

	public abstract static class Fill {}
	public static class PatternFill extends Fill {
		private FillPattern type;
		private String fgColor;
		private String bgColor;
		public PatternFill(String type, String fgColor, String bgColor) {
			switch (type == null ? "none" : type) {
			case "none":
				this.type = FillPattern.NONE;
				break;
			case "solid":
				this.type = FillPattern.SOLID;
				break;
			case "mediumGray":
				this.type = FillPattern.MEDIUM_GRAY;
				break;
			case "darkGray":
				this.type = FillPattern.DARK_GRAY;
				break;
			case "lightGray":
				this.type = FillPattern.LIGHT_GRAY;
				break;
			case "darkHorizontal":
				this.type = FillPattern.DARK_HORIZONTAL;
				break;
			case "darkVertical":
				this.type = FillPattern.DARK_VERTICAL;
				break;
			case "darkDown":
				this.type = FillPattern.DARK_DOWN;
				break;
			case "darkUp":
				this.type = FillPattern.DARK_UP;
				break;
			case "darkGrid":
				this.type = FillPattern.DARK_GRID;
				break;
			case "darkTrellis":
				this.type = FillPattern.DARK_TRELLIS;
				break;
			case "lightHorizontal":
				this.type = FillPattern.LIGHT_HORIZONTAL;
				break;
			case "lightVertical":
				this.type = FillPattern.LIGHT_VERTICAL;
				break;
			case "lightDown":
				this.type = FillPattern.LIGHT_DOWN;
				break;
			case "lightUp":
				this.type = FillPattern.LIGHT_UP;
				break;
			case "lightGrid":
				this.type = FillPattern.LIGHT_GRID;
				break;
			case "lightTrellis":
				this.type = FillPattern.LIGHT_TRELLIS;
				break;
			case "gray125":
				this.type = FillPattern.GRAY125;
				break;
			case "gray0625":
				this.type = FillPattern.GRAY0625;
				break;
			}
			this.fgColor = fgColor;
			this.bgColor = bgColor;
		}
		public FillPattern getType() {
			return type;
		}

		public String getFgColor() {
			return fgColor;
		}

		public String getBgColor() {
			return bgColor;
		}

		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;
			PatternFill that = (PatternFill) o;
			return type == that.type && Objects.equals(fgColor, that.fgColor)
					&& Objects.equals(bgColor, that.bgColor);
		}

		public int hashCode() {
			return Objects.hash(type, fgColor, bgColor);
		}
	}
}
