package io.keikai.importer;

import org.apache.poi.ss.util.CellReference;

/**
 * Store the impose phase information for exception.
 */
// KEIKAI-164
class ImportPhase {
    // In the book
    final static String XLSX = "xlsx";
    final static String STYLE_SHEET = "stylesheet";
    final static String EXTERNAL_BOOK_REFERENCE = "external book reference";
    final static String WORKBOOK_PROTECTION = "workbook protection";
    final static String NAMED_RANGE = "named range";
    final static String SHEETS = "sheets";

    // In the sheet
    final static String SHEET = "sheet";
    final static String SHEET_FORMAT = "sheet format";
    final static String SHEET_VIEW = "sheet view";
    final static String TABLE = "table named %s";
    final static String TABLE_STYLE = "style of table named %s";
    final static String TABLE_AUTO_FILTER = "auto filter of table named %s";
    final static String TABLE_COLUMN = "table column %s[%s]";
    final static String HYPERLINK = "hyperlink";
    final static String VML_DRAWING = "VML drawing";
    final static String COMMENT = "comment";
    final static String ROW = "row";
    final static String COL = "column";
    final static String CELL = "cell";
    final static String MERGED_CELL = "merged cell";
    final static String DRAWING = "drawing";
    final static String CHART = "chart";
    final static String PICTURE = "picture";
    final static String DATA_VALIDATION = "data validation";
    final static String AUTO_FILTER = "auto filter";
    final static String SHEET_PROTECTION = "sheet protection";
    final static String CONDITIONAL_FORMATTING = "conditional formatting";
    final static String PRINT_AREA = "print area";
    final static String EXT_LST = "extLst";
    final private String _bookName;
    private String _phase;
    private String _sheetName;
    private String _tableName;
    private String _tableColumnName;
    private int _colIndex = -1;
    private int _rowIndex = -1;

    public ImportPhase(String bookName) {
        _bookName = bookName;
    }

    void setPhase(String phase) {
        _phase = phase;
    }

    void setSheetName(String sheetName) {
        _sheetName = sheetName;
    }

    void setRowIndex(int rowIndex) {
        _rowIndex = rowIndex;
    }

    void setColIndex(int colIndex) {
        _colIndex = colIndex;
    }

    void setTableName(String tableName) {
        _tableName = tableName;
    }

    void setTableColumnName(String tableColumnName) {
        _tableColumnName = tableColumnName;
    }

    @Override
    public String toString() {
        return appendSheetAndBookName(toPhaseMessage()).toString();
    }

    void clear() {
        _phase = null;
        _sheetName = null;
        _tableName = null;
        _tableColumnName = null;
        _rowIndex = -1;
        _colIndex = -1;
    }

    private StringBuilder toPhaseMessage() {
        final StringBuilder sb = new StringBuilder();
        if (_phase == null) {
            return sb.append(XLSX);
        }
        switch (_phase) {
            case TABLE:
            case TABLE_AUTO_FILTER:
            case TABLE_STYLE: {
                sb.append(String.format(_phase, _tableName));
                break;
            }
            case TABLE_COLUMN: {
                sb.append(String.format(_phase, _tableName, _tableColumnName));
                break;
            }
            case COL: {
                sb.append(_phase);
                if (_colIndex != -1) {
                    sb.append(" ").append(CellReference.convertNumToColString(_colIndex));
                }
                break;
            }
            case ROW: {
                sb.append(_phase);
                if (_rowIndex != -1) {
                    sb.append(" ").append(_rowIndex + 1);
                }
                break;
            }
            case CELL: {
                sb.append(_phase);
                if (_colIndex != -1) {
                    sb.append(" ").append(CellReference.convertNumToColString(_colIndex));
                }
                if (_rowIndex != -1) {
                    sb.append(_rowIndex + 1);
                }
                break;
            }
            default: {
                sb.append(_phase);
            }
        }
        return sb;
    }

    private StringBuilder appendSheetAndBookName(StringBuilder sb) {
        if (_sheetName != null) {
            sb.append(" in sheet ");
            sb.append(_sheetName);
        }
        if (_bookName != null) {
            sb.append(" from book ");
            sb.append(_bookName);
        }
        return sb;
    }
}
