/* NumFmts.java

	Purpose:
		
	Description:
		
	History:
		10:45 AM 2019/9/24, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import static io.keikai.util.Converter.asList;

import java.util.HashMap;
import java.util.Map;

/**
 * @author jumperchen
 */
public class NumFmts {
	private final static java.util.List<String> BUILTIN_FMTS = asList(
			"General", // 0
			"0", // 1
			"0.00", // 2
			"#,##0", // 3
			"#,##0.00", // 4
			"\"$\"#,##0_);\\(\"$\"#,##0\\)", // 5
			"\"$\"#,##0_);[Red]\\(\"$\"#,##0\\)", // 6
			"\"$\"#,##0.00_);\\(\"$\"#,##0.00\\)", // 7
			"\"$\"#,##0.00_);[Red]\\(\"$\"#,##0.00\\)", // 8
			"0%", // 9
			"0.00%", // 10
			"0.00E+00", // 11
			"# ?/?", // 12
			"# ??/??", // 13
			"m/d/yyyy", // 14; == short os system date format
			"d-mmm-yy", // 15
			"d-mmm", // 16
			"mmm-yy", // 17
			"h:mm AM/PM", // 18
			"h:mm:ss AM/PM", // 19
			"h:mm", // 20
			"h:mm:ss", // 21
			"m/d/yyyy h:mm", // 22
			// 0x17(23) - 0x24(36) reserved for international and undocumented
			"reserved-0x17", // 23
			"reserved-0x18", // 24
			"reserved-0x19", // 25
			"reserved-0x1a", // 26
			"reserved-0x1b", // 27
			"reserved-0x1c", // 28
			"reserved-0x1d", // 29
			"reserved-0x1e", // 30; == short os system date format (yyyy/m/d) (see #2011)
			"reserved-0x1f", // 31
			"reserved-0x20", // 32
			"reserved-0x21", // 33
			"reserved-0x22", // 34
			"reserved-0x23", // 35
			"reserved-0x24", // 36
			"#,##0_);\\(#,##0\\)", // 37
			"#,##0_);[Red]\\(#,##0\\)", // 38
			"#,##0.00_);\\(#,##0.00\\)", // 39
			"#,##0.00_);[Red]\\(#,##0.00\\)", // 40
			"_(* #,##0_);_(* \\(#,##0\\);_(* \"-\"_);_(@_)", // 41
			"_(\"$\"* #,##0_);_(\"$\"* \\(#,##0\\);_(\"$\"* \"-\"_);_(@_)", // 42
			"_(* #,##0.00_);_(* \\(#,##0.00\\);_(* \"-\"??_);_(@_)", // 43
			"_(\"$\"* #,##0.00_);_(\"$\"* \\(#,##0.00\\);_(\"$\"* \"-\"??_);_(@_)", //  44
			"mm:ss", // 45
			"[h]:mm:ss", // 46
			"mm:ss.0", // 47
			"##0.0E+0", // 48
			"@" // 49
			);
	private final static int NUM_FMT_INDEX_BASE = 164;
	private Map<String , Integer> numFmtMap = new HashMap<>();
	private Map<Integer , String> customNumFmts = new HashMap<>();
	private int indexBase;
	private int maxIndex;

	public NumFmts() {
		this(NUM_FMT_INDEX_BASE);
	}
	public NumFmts(int indexBase) {
		this.indexBase = indexBase;
		this.maxIndex = indexBase - 1;
		int j = 0;
		for (String fmt : BUILTIN_FMTS) {
			this.numFmtMap.put(fmt, j);
			j = j + 1;
		}
	}

	public static boolean isBuiltIn(int id) {
		return id < BUILTIN_FMTS.size();
	}

	public boolean isCustomize(int id) {
		return id >= this.indexBase;
	}

	public int getCount() {
		return this.customNumFmts.size();
	}

	public String getNumFmtAt(int id) {
		String fmt0 = customNumFmts.get(id); // look in customized map
		//#2011
		if (fmt0 == null && isReservedBuiltin(id)) {
			id = 14; // convert it into OS System Short Date id; i.e., 14
		}
		final String fmt = fmt0 != null ? fmt0 :
		(id < BUILTIN_FMTS.size()
				? BUILTIN_FMTS.get(id)
				: null); // then built-in if not exists
		return fmt;
	}

	//#2011
	// per the observation for the special file t6660.xlsx, we assume all
	// reserved numFmtId is interpreted by Excel as "OS System Short Date"
	// (in the file, both numFmtId = "30" and numFmtId = "58"); thus avoid
	// parsing complain by Keikai format parser.
	private boolean isReservedBuiltin(int id) {
		return (23 <= id && id <= 36)
				|| (BUILTIN_FMTS.size() <= id && id < NUM_FMT_INDEX_BASE);
	}

	public void addNumFmt(String fmt, int index) {
		if (fmt != null) {
			if (index > maxIndex) {
				maxIndex = index;
			}
			this.customNumFmts.put(index, fmt);
			this.numFmtMap.put(fmt, index);
		}
	}
}
