/* VmlDrawing.java

	Purpose:
		
	Description:
		
	History:
		10:51 AM 2019/9/26, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import static io.keikai.util.Converter.asList;
import static io.keikai.util.Converter.getFirst;
import static io.keikai.util.Converter.ptToPx;
import static io.keikai.util.Converter.rangeToA1;
import static io.keikai.util.XlsxJsonHelper.$;
import static io.keikai.util.XlsxJsonHelper.$x;
import static io.keikai.util.XlsxJsonHelper.toInt;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * @author jumperchen
 */
public class VmlDrawing {
	final private Map json;
	final private Map _mapEntry;
	final private byte[] imageData; // #1105
	Map <String, VShapeJson> _refIdToShape = new HashMap<>(3);
	private List<VShapeJson> _shapes;
	Map _idmap;
	public VmlDrawing(Map jsonMap, byte[] imageData) {
		this.json = jsonMap;
		this.imageData = imageData;
		this._mapEntry = $(jsonMap, "xml");
		_idmap = $x(_mapEntry, "o:shapelayout/o:idmap");
		Object shapeJson = $(_mapEntry, "v:shape");
		if (shapeJson != null) {
			List<Map> shapeList = shapeJson instanceof List ?
					((List<Map>) shapeJson) :
					asList((Map) shapeJson);
			_shapes = shapeList.stream().map((entry) -> new VShapeJson(this, entry)).collect(
					Collectors.toList());
			for (VShapeJson shape : _shapes) {
				_addToMap(shape);
			}
		} else {
			_shapes = Collections.EMPTY_LIST;
		}
	}

	public VShapeJson getShape(String refOrId) {return _refIdToShape.get(refOrId);}

	private void _addToMap(VShapeJson shape) {
		// only ObjectType="Note", TODO: other types (like form controls)
		if (shape.isNote()) {
			_refIdToShape.put(rangeToA1(shape.getRow(), shape.getColumn()), shape);
		}

		// form-control's id
		final String spid = (String) shape.shape.get("o:spid");
		if (spid != null) {
			_refIdToShape.put(spid, shape);
		}
		_refIdToShape.put(shape.getId(), shape);
	}


	public static class VShapeJson {
		final private VmlDrawing parent;
		final private Map shape;
		final private Map _clientData;
		final private Map<String, String> _style;
		public VShapeJson(VmlDrawing parent, Map shape) {
			this.parent = parent;
			this.shape = shape;
			this._clientData = $(shape, "x:ClientData");
			_style = Stream.of(shape.get("style").toString().split(";")).map((v) -> v.split(":")).collect(Collectors
					.toMap((v) -> v[0].trim(), (v) -> v[v.length - 1].trim()));
		}

		public boolean isFormCtrlType() {
			switch(getObjectType()) {
			case "Button":
			case "Label":
			case "Radio":
			case "Checkbox":
			case "GBox":
			case "Spin":
			case "Scroll":
			case "List":
			case "Drop":
				return true;
			default: return false;
			}
		}

		public String getObjectType() {
			return $(_clientData, "ObjectType");
		}

		public boolean isNote() {
			return "Note".equals(getObjectType());
		}

		public int getColumn() {
			return toInt(_clientData, "x:Column");
		}
		public int getRow() {
			return toInt(_clientData, "x:Row");
		}

		public String getId() {
			return $(shape, "id");
		}
		public String getSpid() {
			return $(shape, "o:spid");
		}
		public String getFillColor() {
			String fill = Optional.ofNullable($(shape, "fillcolor")).orElse($(shape, "v:fill", "color2")).toString();
			return getFirst(fill.split(" "));
		}
		private int toSize(String size) {
			if (size != null && size.length() > 2) {
				return ptToPx(Double.parseDouble(size.substring(0, size.length() - 2)));
			}
			return 0;
		}
		public int getWidth() {
			return toSize(_style.get("width"));
		}
		public int getHeight() {
			return toSize(_style.get("height"));
		}

		public int getLeft() {
			return toSize(_style.get("margin-left"));
		}

		public int getTop() {
			return toSize(_style.get("margin-top"));
		}
		public boolean isVisibility() {
			return !"hidden".equals(_style.get("visibility"));
		}
		public String getFmlaRange() {
			String val = $(this._clientData, "x:FmlaRange");
			return val != null ? val.trim() : val;
		}
		public String getFmlaLink() {
			String val = $(this._clientData, "x:FmlaLink");
			return val != null ? val.trim() : val;
		}
		public int getVal() {
			return toInt(this._clientData, "x:Val");
		}
		public int getMin() {
			return toInt(this._clientData, "x:Min");
		}

		public int getMax() {
			return toInt(this._clientData, "x:Max");
		}

		public int getIncrement() {
			return toInt(this._clientData, "x:Inc");
		}

		public boolean isHorizontal() {
			return this._clientData.containsKey("x:Horiz");
		}

		public int getPage() {
			return toInt(this._clientData, "x:Page");
		}
	}
}
