/* XlsxExtractor.java

	Purpose:
		
	Description:
		
	History:
		10:25 AM 2019/9/19, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import static io.keikai.model.SDataValidation.AlertStyle;
import static io.keikai.model.SDataValidation.OperatorType;
import static io.keikai.model.SDataValidation.ValidationType;
import static io.keikai.util.Converter.asList;
import static io.keikai.util.Converter.asMap;
import static io.keikai.util.Converter.getLast;
import static io.keikai.util.XlsxJsonHelper.$;
import static io.keikai.util.XlsxJsonHelper.$1;
import static io.keikai.util.XlsxJsonHelper.$x;
import static io.keikai.util.XlsxJsonHelper.toBool;
import static io.keikai.util.XlsxJsonHelper.toBoolean;
import static io.keikai.util.XlsxJsonHelper.toDouble;
import static io.keikai.util.XlsxJsonHelper.toFloat;
import static io.keikai.util.XlsxJsonHelper.toInt;
import static io.keikai.util.XlsxJsonHelper.toList;
import static io.keikai.util.XlsxJsonHelper.toNum;
import static io.keikai.util.XlsxJsonHelper.toStringValue;
import static io.keikai.util.XlsxJsonHelper.toValue;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import io.keikai.importer.xml.Cell;
import io.keikai.importer.xml.Row;
import io.keikai.importer.xml.SheetData;
import io.keikai.json.JSONArray;
import io.keikai.json.JSONAware;
import io.keikai.json.JSONObject;
import io.keikai.model.SAutoFilter;
import io.keikai.model.SCFValueObject;
import io.keikai.model.SConditionalFormattingRule;
import io.keikai.model.SHyperlink;
import io.keikai.model.SIconSet;
import io.keikai.model.SPrintSetup;
import io.keikai.model.STableColumn;
import io.keikai.util.Converter;
import io.keikai.util.XlsxJsonHelper;

import org.zkoss.util.Pair;

/**
 * @author jumperchen
 */
public class XlsxExtractor {

	final static String PrinterRelType = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/printerSettings";
	final static String TableRelType = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/table";
	final static String CommentsRelType = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/comments";
	final static String VmlDrawingRelType = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/vmlDrawing";
	final static String DrawingRelType = "http://schemas.openxmlformats.org/officeDocument/2006/relationships/drawing";

	private static Map<String, String> _getSheetPrinterRels(String sheetPath,
			Map bookJson) {
		return _getSheetRels(sheetPath, PrinterRelType, bookJson);
	}

	private static Map<String, String> _getSheetTableRels(String sheetPath,
			Map bookJson) {
		return _getSheetRels(sheetPath, TableRelType, bookJson);
	}

	private static Map<String, String> _getSheetCommentsRels(String sheetPath,
			Map bookJson) {
		return _getSheetRels(sheetPath, CommentsRelType, bookJson);
	}

	private static Map<String, String> _getSheetVmlDrawingRels(String sheetPath,
			Map bookJson) {
		return _getSheetRels(sheetPath, VmlDrawingRelType, bookJson);
	}

	private static Map<String, String> _getSheetDrawingRels(String sheetPath,
			Map bookJson) {
		return _getSheetRels(sheetPath, DrawingRelType, bookJson);
	}

	// {rid: target path} for the specified sheet path
	private static Map<String, String> _getSheetRels(String sheetPath,
			String relType, Map bookJson) {
		String[] strings = sheetPath.split("/");
		Map rels = $(bookJson,
				"xl/worksheets/_rels/" + strings[strings.length - 1] + ".rels");
		Map<String, String> sheetRels = null;
		if (rels != null) {
			Object relations = $(rels, "Relationships", "Relationship");
			if (relations instanceof List) {
				sheetRels = ((List<Map>) relations).stream()
						.filter((r) -> relType.equals(r.get("Type"))).collect(
								Collectors.toMap((r) -> (String) r.get("Id"),
										(r) -> (String) r.get("Target")));
			} else if (relations instanceof Map) {
				Map relMap = (Map) relations;
				if (relType.equals(relMap.get("Type"))) {
					sheetRels = asMap(relMap.get("Id"), relMap.get("Target"));
				}
			}
		}
		return sheetRels;
	}

	private static String _getPath(String path, String relatePath) {
		if (path.startsWith("../")) {
			return relatePath + path.substring(2);
		} else if (path.startsWith("/")) {
			return path.substring(1);
		}
		return path;
	}

	public static XlsxWorkbookExtractor getWorkbook(Map jsonMap) {
		return new XlsxWorkbookExtractor(jsonMap);
	}

	public static class XlsxWorkbookExtractor {
		final private Map jsonMap;
		private static String _DEFAULT_CONTENT_TYPE = "[Content_Types].xml";
		private Map<String, Object> _filePaths;
		private static Map<String, String> _CONTENT_TYPE_MAPPING = asMap(
				"application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml",
				"workbooks", "application/vnd.ms-excel.binIndexWs", "TODO",
				"application/vnd.ms-excel.chartsheet", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.chartsheet+xml",
				"TODO", "application/vnd.ms-excel.dialogsheet", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.dialogsheet+xml",
				"TODO", "application/vnd.ms-excel.macrosheet", "TODO",
				"application/vnd.ms-excel.macrosheet+xml", "TODO",
				"application/vnd.ms-excel.intlmacrosheet", "TODO",
				"application/vnd.ms-excel.binIndexMs", "TODO",
				"application/vnd.openxmlformats-package.core-properties+xml",
				"coreprops",
				"application/vnd.openxmlformats-officedocument.custom-properties+xml",
				"custprops",
				"application/vnd.openxmlformats-officedocument.extended-properties+xml",
				"extprops",
				"application/vnd.openxmlformats-officedocument.customXmlProperties+xml",
				"TODO", "application/vnd.ms-excel.comments", "comments",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.comments+xml",
				"comments", "application/vnd.ms-excel.pivotTable", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.pivotTable+xml",
				"TODO", "application/vnd.ms-excel.calcChain", "calcchains",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.calcChain+xml",
				"calcchains",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.printerSettings",
				"TODO", "application/vnd.ms-office.activeX", "TODO",
				"application/vnd.ms-office.activeX+xml", "TODO",
				"application/vnd.ms-excel.attachedToolbars", "TODO",
				"application/vnd.ms-excel.connections", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.connections+xml",
				"TODO", "application/vnd.ms-excel.externalLink",
				"externalReferences",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.externalLink+xml",
				"externalReferences", "application/vnd.ms-excel.sheetMetadata",
				"TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.sheetMetadata+xml",
				"TODO", "application/vnd.ms-excel.pivotCacheDefinition", "TODO",
				"application/vnd.ms-excel.pivotCacheRecords", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.pivotCacheDefinition+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.pivotCacheRecords+xml",
				"TODO", "application/vnd.ms-excel.queryTable", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.queryTable+xml",
				"TODO", "application/vnd.ms-excel.userNames", "TODO",
				"application/vnd.ms-excel.revisionHeaders", "TODO",
				"application/vnd.ms-excel.revisionLog", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.revisionHeaders+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.revisionLog+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.userNames+xml",
				"TODO", "application/vnd.ms-excel.tableSingleCells", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.tableSingleCells+xml",
				"TODO", "application/vnd.ms-excel.slicer", "TODO",
				"application/vnd.ms-excel.slicerCache", "TODO",
				"application/vnd.ms-excel.slicer+xml", "TODO",
				"application/vnd.ms-excel.slicerCache+xml", "TODO",
				"application/vnd.ms-excel.wsSortMap", "TODO",
				"application/vnd.ms-excel.table", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.table+xml",
				"tables",
				"application/vnd.openxmlformats-officedocument.theme+xml",
				"themes", "application/vnd.ms-excel.Timeline+xml", "TODO",
				"application/vnd.ms-excel.TimelineCache+xml", "TODO",
				"application/vnd.ms-office.vbaProject", "vba",
				"application/vnd.ms-office.vbaProjectSignature", "vba",
				"application/vnd.ms-office.volatileDependencies", "TODO",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.volatileDependencies+xml",
				"TODO", "application/vnd.ms-excel.controlproperties+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.model+data",
				"TODO", "application/vnd.ms-excel.Survey+xml", "TODO",
				"application/vnd.openxmlformats-officedocument.drawing+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.drawingml.chart+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.drawingml.chartshapes+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.drawingml.diagramColors+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.drawingml.diagramData+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.drawingml.diagramLayout+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.drawingml.diagramStyle+xml",
				"TODO",
				"application/vnd.openxmlformats-officedocument.vmlDrawing",
				"TODO",
				"application/vnd.openxmlformats-package.relationships+xml",
				"rels",
				"application/vnd.openxmlformats-officedocument.oleObject",
				"TODO", "sheet", "js",
				"application/vnd.ms-excel.sheet.macroEnabled.main+xml",
				"workbooks",
				"application/vnd.ms-excel.sheet.binary.macroEnabled.main",
				"workbooks",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.template.main+xml",
				"workbooks",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.sharedStrings+xml",
				"strs", "application/vnd.ms-excel.sharedStrings", "strs",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml",
				"sheets", "application/vnd.ms-excel.worksheet", "sheets",
				"application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml",
				"styles", "application/vnd.ms-excel.styles", "styles");

		public XlsxWorkbookExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			_filePaths = new HashMap();
			XlsxJsonHelper.<List<Map>>$(this.jsonMap, _DEFAULT_CONTENT_TYPE,
					"Types", "Override").forEach((Map k) -> {
				String key = _CONTENT_TYPE_MAPPING.get(k.get("ContentType"));
				Object value = _filePaths.get(key);
				if (value instanceof List) {
					((List) value)
							.add(((String) k.get("PartName")).substring(1));
				} else if (value != null) {
					List val = asList(value);
					_filePaths.put(key, val);
					val.add(((String) k.get("PartName")).substring(1));
				} else {
					_filePaths.put(key,
							((String) k.get("PartName")).substring(1));
				}
			});
		}

		Map _workbookJson;

		private Map getWorkbookJson() {
			if (_workbookJson == null) {
				_workbookJson = $(jsonMap, $(_filePaths, "workbooks"),
						"workbook");
			}
			return _workbookJson;
		}

		//		public ThemesExtractor getThemes() {
		//			/*final*/	Object file = _filePaths["themes"];
		//			if (file != null) {
		//				return new XlsxThemesExtractor("Win10',unsafeCast(jsonMap[_filePaths['themes")]['a:theme")))
		//						;
		//			} else {
		//				return null;
		//			}
		//		}

		public XlsxStyleSheetExtractor getStyleSheet() {
			return new XlsxStyleSheetExtractor(
					$(jsonMap, $(_filePaths, "styles"), "styleSheet"));
		}

		public XlsxThemesExtractor getThemes() {
			String filePath = $(_filePaths, "themes");

			if (filePath != null) {
				return new XlsxThemesExtractor("Win10",
						$(jsonMap, filePath, "a:theme"));
			} else {
				return null;
			}
		}

		public Map getSst() {
			Map r = $1(jsonMap, $(_filePaths, "strs"));
			if (r != null) {
				return $(r, "sst");
			} else {
				return Collections.EMPTY_MAP;
			}
		}

		public Map getMcAlternateContent() {
			return $(getWorkbookJson(), "mc:AlternateContent");
		}

		public Map getXrRevisionPtr() {
			return $(getWorkbookJson(), "xr:revisionPtr");
		}

		public List<XlsxExternalLinkExtractor> getExternalReferences() {
			Map workbookJson = getWorkbookJson();
			Map extRefsJson = (Map) workbookJson.get("externalReferences");
			if (extRefsJson == null) {
				return Collections.EMPTY_LIST;
			}
			Object extRefs = extRefsJson.get("externalReference");
			if (extRefs == null) {
				return Collections.EMPTY_LIST;
			}
			List<Map> extRefsList = null;
			if (extRefs instanceof Map) {
				extRefsList = asList((Map) extRefs);
			}
			if (extRefsList == null || extRefsList.isEmpty()) {
				return Collections.EMPTY_LIST;
			}
			List<Map> workbookRels = $(jsonMap, "xl/_rels/workbook.xml.rels",
					"Relationships", "Relationship");
			Map idMap = new LinkedHashMap();
			workbookRels.forEach((rel) -> {
				idMap.put(rel.get("Id"), rel.get("Target"));
			});
			List<Map> result = new ArrayList<>();
			Object externalReferences = _filePaths.get("externalReferences");
			if (externalReferences instanceof List) {
				Map resultMap = new LinkedHashMap(); // externalLinkPath -> externalLink contents
				((List<String>) externalReferences)
						.forEach((String extRefPath) -> {
							Object extLink = $(jsonMap, extRefPath); // contents
							String extRefPath0 = extRefPath
									.substring(3); // trim 'xl/'
							resultMap.put(extRefPath0, extLink);
						});
				// externalReference order is determined by workbook.externalReferences, not by file path order
				extRefsList.forEach((extRef) -> {
					// KEIKAI-671: support absolute relationship paths
					String extRefPath = $1(idMap, $(extRef, "r:id"));
					String extRefPath0 = extRefPath.replaceFirst("^/?xl/", "");
					Map extLink = $1(resultMap, extRefPath0);
					if (extLink != null) {
						extLink.putAll(extRef);
						result.add(extLink);
					} else {
						result.add(extRef);
					}
				});
			} else {
				String extRefPath = $(_filePaths, "externalReferences");
				Map extLink = $(jsonMap, extRefPath);
				Map extRef = extRefsList.get(0);
				if (extLink != null) {
					extLink.putAll(extRef);
					result.add(extLink);
				} else {
					result.add(extRef);
				}
			}
			return result.stream().map((extLink) -> {
				return extLink.get("externalLink") != null ?
						new XlsxExternalLinkExtractor(jsonMap, extLink) :
						null;
			}).filter((d) -> d != null).collect(Collectors.toList());
		}

		public XlsxWorkbookProtectionExtractor getWorkbookProtection() {
			final Map sheetProtection = $(getWorkbookJson(), "workbookProtection");
			if (sheetProtection != null) {
				return new XlsxWorkbookProtectionExtractor(sheetProtection);
			}
			return null;
		}

		public List<XlsxSheetExtractor> getSheets() {
			List<Map> workbookRels = $(jsonMap, "xl/_rels/workbook.xml.rels",
					"Relationships", "Relationship");
			Map workbook = getWorkbookJson();
			Map sharedStrings = getSst();
			Map<String, String> idMap = new LinkedHashMap<>(); // rid -> sheetPath
			workbookRels.forEach((rel) -> {
				idMap.put($(rel, "Id"), $(rel, "Target"));
			});
			Object sheets = $(workbook, "sheets", "sheet");
			List<Map> sheetsList = null;
			if (sheets instanceof Map) {
				sheetsList = asList((Map) sheets);
			} else if (sheets instanceof List) {
				sheetsList = (List) sheets;
			} else {
				throw new IllegalStateException("Sheets cannot be empty!");
			}

			List<Map> result = new ArrayList<>();
			Object sheetsData = _filePaths.get("sheets");
			if (sheetsData instanceof List) {
				Map<String, Map> resultMap = new LinkedHashMap<>(); // sheetPath -> ws
				((List<String>) sheetsData).forEach((sheetPath) -> {
					Map ws = (Map) jsonMap.get(sheetPath);
					String sheetPath0 = sheetPath.substring(3);
					resultMap.put(sheetPath0, ws);
				});
				// #762: sheet order is determined by workbook.sheets, not by file path order
				sheetsList.forEach((sheet) -> {
					// KEIKAI-671: support absolute relationship paths
					String sheetPath = idMap.get(sheet.get("r:id"));
					String sheetPath0 = sheetPath.replaceFirst("^/?xl/", "");
					Map ws = $(resultMap, sheetPath0);
					if (ws != null) {
						ws.putAll(sheet);
						result.add(ws);
					} else {
						result.add(sheet);
					}
				});
			} else {
				String sheetPath = (String) sheetsData;
				Map ws = (Map) jsonMap.get(sheetPath);
				Map sheet = sheetsList.get(0);
				if (ws != null) {
					ws.putAll(sheet);
					result.add(ws);
				} else {
					result.add(sheet);
				}
			}
			return result.stream().map((d) -> {
				return new XlsxSheetExtractor(jsonMap, d, sharedStrings,
						new HashMap());
			}).collect(Collectors.toList());
		}

		public List<XlsxDefinedNameExtractor> getDefinedNames() {
			return toList($(_workbookJson, "definedNames"), "definedName", XlsxDefinedNameExtractor.class);
		}

		//		public Iterable<WorkbookViewExtractor> getBookViews() {
		//			/*final*/	Object workbook = jsonMap[_filePaths["workbooks"]]["workbook"] as Map;
		//			if (workbook.containsKey("bookViews")) {
		//				/*final*/	Object bookViews = workbook["bookViews"] as Map;
		//				if (bookViews.containsKey("workbookView")) {
		//					/*final*/	Object workbookView = bookViews["workbookView"];
		//					if (workbookView instanceOf List) {
		//						return workbookView.map((e)=>new XlsxWorkbookViewExtractor(unsafeCast(e)));
		//					} else {
		//						return asList(new XlsxWorkbookViewExtractor(unsafeCast(workbookView)));
		//					}
		//				}
		//			}
		//			return asList();
		//		}

		//		public FileVersionExtractor getFileVersion() {
		//			/*final*/	Object workbook = jsonMap[_filePaths["workbooks"]]["workbook"] as Map;
		//			if (workbook.containsKey("fileVersion")) {
		//				return new XlsxFileVersionExtractor(workbook['fileVersion") as Map)
		//						;
		//			}
		//			return null;
		//		}
		//
		//		public List<JsonRelationEntry> getCustomXml() {
		//			/*final*/	Object workbookRels = jsonMap["xl/_rels/workbook.xml.rels"]["Relationships"]["Relationship"];
		//			/*var*/	Object relList = asList();
		//			for (/*var*/ Object rel : workbookRels) {
		//				if (rel["Type"] == "http://schemas.openxmlformats.org/officeDocument/2006/relationships/customXml") {
		//					relList.add(rel);
		//				}
		//			}
		//			if (relList.isNotEmpty) {
		//				return relList.map((rel0){final entry={};var target=rel0['Target");var index=target.substring("../customXml/item'.length);index=index.substring(0,index.length-'.xml'.length);for(String key in jsonMap.keys){if(key.startsWith("customXml/item${index}.xml")||key.startsWith("customXml/itemProps${index}.xml")||key.startsWith("customXml/_rels/item${index}.xml.rels")){entry[key]=jsonMap[key];}}return new JsonRelationEntry(unsafeCast(rel0['Target")),unsafeCast(rel0['Type")),entry);}).toList;
		//			}
		//			return null;
		//		}
		//
		//		public Map getConnections() {
		//			return $(jsonMap, "xl/connections.xml");
		//		}
		//
		//		public Map getCtrlProps() {
		//			/*final*/	Object data = asMap();
		//			int idx = 1;
		//			//while(true){final key='xl/ctrlProps/ctrlProp${idx}.xml';var v=jsonMap[key];if(v!=null){data[key]=v;idx++ ;}else{break;}}
		//			return data;
		//		}
	}

	public static class XlsxStyleSheetExtractor {
		final private Map jsonMap;

		public XlsxStyleSheetExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxBorderExtractor> getBorders() {
			return jsonMap.containsKey("borders") ?
					(List<XlsxBorderExtractor>) toList((jsonMap.get("borders")),
							"border").stream()
							.map((data) -> new XlsxBorderExtractor((Map) data))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public List<XlsxXfExtractor> getCellStyleXfs() {
			return jsonMap.containsKey("cellStyleXfs") ?
					(List<XlsxXfExtractor>) toList(
							(jsonMap.get("cellStyleXfs")), "xf").stream()
							.map((data) -> new XlsxXfExtractor((Map) (data)))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public List<XlsxCellStyleExtractor> getCellStyles() {
			return jsonMap.containsKey("cellStyles") ?
					(List<XlsxCellStyleExtractor>) toList(
							jsonMap.get("cellStyles"), "cellStyle").stream()
							.map((data) -> new XlsxCellStyleExtractor(
									(Map) (data)))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public List<XlsxXfExtractor> getCellXfs() {
			return jsonMap.containsKey("cellXfs") ?
					(List<XlsxXfExtractor>) toList(jsonMap.get("cellXfs"), "xf")
							.stream()
							.map((data) -> new XlsxXfExtractor((Map) (data)))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public XlsxRGBColorsExtractor getColors() {
			return jsonMap.containsKey("colors") ?
					new XlsxRGBColorsExtractor((Map) jsonMap.get("colors")) :
					null;
		}

		public List<XlsxDxfExtractor> getDxfs() {
			return jsonMap.containsKey("dxfs") ?
					(List<XlsxDxfExtractor>) toList(jsonMap.get("dxfs"), "dxf")
							.stream()
							.map((data) -> new XlsxDxfExtractor((Map) (data)))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public List<XlsxFillExtractor> getFills() {
			return jsonMap.containsKey("fills") ?
					(List<XlsxFillExtractor>) toList(jsonMap.get("fills"),
							"fill").stream().map((data) -> XlsxFillExtractor
							.getInstance((Map) (data)))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public List<XlsxFontExtractor> getFonts() {
			return jsonMap.containsKey("fonts") ?
					(List<XlsxFontExtractor>) toList(jsonMap.get("fonts"),
							"font").stream()
							.map((data) -> new XlsxFontExtractor((data == null ?
									Collections.EMPTY_MAP :
									(Map) data))).collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public List<XlsxNumFmtExtractor> getNumFmts() {
			return jsonMap.containsKey("numFmts") ?
					(List<XlsxNumFmtExtractor>) toList(jsonMap.get("numFmts"),
							"numFmt").stream()
							.map((data) -> new XlsxNumFmtExtractor(
									(Map) (data)))
							.collect(Collectors.toList()) :
					Collections.EMPTY_LIST;
		}

		public XlsxTableStylesExtractor getTableStyles() {
			return jsonMap.containsKey("tableStyles") ?
					new XlsxTableStylesExtractor(
							(Map) (jsonMap.get("tableStyles"))) :
					null;
		}
	}

	public static class XlsxTableStylesExtractor {
		final private Map jsonMap;

		public XlsxTableStylesExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getDefaultPivotStyle() {
			return toStringValue(jsonMap, "defaultPivotStyle");
		}

		public String getDefaultTableStyle() {
			return toStringValue(jsonMap, "defaultTableStyle");
		}

		public List<XlsxTableStyleExtractor> getTableStyles() {
			return toList(jsonMap, "tableStyle", XlsxTableStyleExtractor.class);
		}
	}

	public static class XlsxTableStyleExtractor {
		final private Map jsonMap;

		public XlsxTableStyleExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public Boolean isPivot() {
			return toBool(jsonMap, "pivot");
		}

		public Boolean isTable() {
			return toBool(jsonMap, "table");
		}

		public List<XlsxTableStyleElementExtractor> getTableStyleElements() {
			return toList(jsonMap, "tableStyleElement", XlsxTableStyleElementExtractor.class);
		}
	}

	public static class XlsxTableStyleElementExtractor {
		final private Map jsonMap;

		public XlsxTableStyleElementExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getDxfId() {
			return toNum(jsonMap, "dxfId");
		}

		public Number getSize() {
			return toNum(jsonMap, "size");
		}

		public String getType() {
			return toStringValue(jsonMap, "type");
		}
	}

	public static class XlsxFillExtractor {
		public static XlsxFillExtractor getInstance(Map jsonMap) {
			if (jsonMap.containsKey("patternFill")) {
				return new XlsxPatternExtractor(
						(Map) (jsonMap.get("patternFill")));
			} else {
				if (jsonMap.containsKey("gradientFill"))
					return new XlsxGradientExtractor(
							(Map) jsonMap.get("gradientFill"));
			}
			return null;
		}
	}

	public static class XlsxRGBColorsExtractor {
		Map jsonMap;

		public XlsxRGBColorsExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<String> getIndexedColors() {
			Object indexColor = this.jsonMap.get("indexedColors");
			if (indexColor != null) {
				return (List<String>) toList(indexColor, "rgbColor").stream()
						.map((data) -> ((Map) data).get("rgb"))
						.collect(Collectors.toList());
			}
			return Collections.EMPTY_LIST;
		}

		public List<String> getMruColors() {
			/*final*/
			Object indexColor = this.jsonMap.get("mruColors");
			if (indexColor != null) {
				return (List<String>) toList(indexColor, "color").stream()
						.map((data) -> ((Map) data).get("rgb"))
						.collect(Collectors.toList());
			}
			return Collections.EMPTY_LIST;
		}
	}

	public static class XlsxGradientExtractor extends XlsxFillExtractor {
		final private Map jsonMap;

		public XlsxGradientExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getBottom() {
			return toNum(jsonMap, "bottom");
		}

		public Number getDegree() {
			return toNum(jsonMap, "degree");
		}

		public Number getLeft() {
			return toNum(jsonMap, "left");
		}

		public Number getRight() {
			return toNum(jsonMap, "right");
		}

		public List<XlsxStopExtractor> getStop() {
			return toList(jsonMap, "stop", XlsxStopExtractor.class);
		}

		public Number getTop() {
			return toNum(jsonMap, "top");
		}

		public String getType() {
			return toStringValue(jsonMap, "type");
		}
	}

	public static class XlsxStopExtractor {
		final private Map jsonMap;

		public XlsxStopExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxColorExtractor getColor() {
			return jsonMap.containsKey("color") ?
					new XlsxColorExtractor((Map) jsonMap.get("color")) :
					null;
		}

		public Number getPosition() {
			return toNum(jsonMap, "position");
		}
	}

	public static class XlsxPatternExtractor extends XlsxFillExtractor {
		final private Map jsonMap;

		public XlsxPatternExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxColorExtractor getBgColor() {
			return jsonMap.containsKey("bgColor") ?
					new XlsxColorExtractor((Map) jsonMap.get("bgColor")) :
					null;
		}

		public XlsxColorExtractor getFgColor() {
			return jsonMap.containsKey("fgColor") ?
					new XlsxColorExtractor((Map) jsonMap.get("fgColor")) :
					null;
		}

		public String getPatternType() {
			return toStringValue(jsonMap, "patternType");
		}
	}

	public static class XlsxFontExtractor {
		final private Map jsonMap;

		public XlsxFontExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isOverrideBold() {
			return jsonMap.containsKey("b");
		}

		public Boolean isBold() {
			return toBoolean(toValue(jsonMap, "b"), jsonMap.containsKey("b"));
		}

		public Number getCharset() {
			return toNum(jsonMap, "charset");
		}

		public boolean isOverrideColor() {
			return jsonMap.containsKey("color");
		}

		public XlsxColorExtractor getColor() {
			return jsonMap.containsKey("color") ?
					new XlsxColorExtractor((Map) jsonMap.get("color")) :
					null;
		}

		public Boolean isCondense() {
			return toBool(jsonMap, "condense", jsonMap.containsKey("condense"));
		}

		public Boolean isExtend() {
			return jsonMap.containsKey("extend");
		}

		public Number getFamily() {
			return toNum(jsonMap, "family");
		}

		public boolean isOverrideItalic() {
			return jsonMap.containsKey("i");
		}

		public Boolean isItalic() {
			return toBoolean(toValue(jsonMap, "i"), jsonMap.containsKey("i"));
		}

		public boolean isOverrideName() {
			return jsonMap.containsKey("name");
		}

		public String getName() {
			return Optional.ofNullable(toStringValue(jsonMap, "name",
					toStringValue(jsonMap, "rFont"))).orElse("Calibri");
		}

		public Boolean isOutline() {
			return toBool(jsonMap, "outline", jsonMap.containsKey("outline"));
		}

		public String getScheme() {
			return toStringValue(jsonMap, "scheme");
		}

		public Boolean isShadow() {
			return toBool(jsonMap, "shadow", jsonMap.containsKey("shadow"));
		}

		public boolean isOverrideStrike() {
			return jsonMap.containsKey("strike");
		}

		public Boolean isStrike() {
			return toBoolean(toValue(jsonMap, "strike"), jsonMap.containsKey("strike"));
		}

		public boolean isOverrideSize() {
			return jsonMap.containsKey("sz");
		}

		public Number getSize() {
			return toNum(jsonMap, "sz");
		}

		public boolean isOverrideUnderline() {
			return jsonMap.containsKey("u");
		}

		public String getUnderline() {
			if (jsonMap.containsKey("u")) {
				String v = toStringValue(jsonMap, "u");
				if (v == null)
					return "single";
				return v;
			} else {
				return null;
			}
		}

		public boolean isOverrideVertAlign() {
			return jsonMap.containsKey("vertAlign");
		}

		public String getVertAlign() {
			return toStringValue(jsonMap, "vertAlign");
		}
	}

	public static class XlsxNumFmtExtractor {
		final private Map jsonMap;

		public XlsxNumFmtExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getFormatCode() {
			return (String) jsonMap.get("formatCode");
		}

		public Number getNumFmtId() {
			return toNum(jsonMap, "numFmtId");
		}
	}

	public static class XlsxCellStyleExtractor {
		final private Map jsonMap;

		public XlsxCellStyleExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getBuiltinId() {
			return Optional.ofNullable(toInt(jsonMap, "builtinId")).orElse(0);
		}

		public boolean isCustomBuiltin() {
			return toBool(jsonMap, "customBuiltin", false);
		}

		public boolean isHidden() {
			return toBool(jsonMap, "hidden", false);
		}

		public Number getILevel() {
			return toNum(jsonMap, "iLevel");
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public int getXfId() {
			return toInt(jsonMap, "xfId");
		}
	}

	public static class XlsxCellXfExtractor extends XlsxXfExtractor {
		private XlsxXfExtractor self;
		private XlsxXfExtractor parentXf;

		public XlsxCellXfExtractor(XlsxXfExtractor self,
				XlsxXfExtractor parentXf) {
			super(parentXf.jsonMap);
			this.parentXf = parentXf;
			this.self = self;
		}

		public Number getParentXfId() {
			return super.getXfId();
		}

		public Number getXfId() {
			return self.getXfId();
		}

		public XlsxAlignmentExtractor getAlignment() {
			if (self.isSetAlignment() != null || parentXf.isSetAlignment()) {
				return new XlsxAlignmentExtractor((Map) parentXf.jsonMap.get("alignment"), self);
			} else {
				return super.getAlignment();
			}
		}

		public XlsxBorderExtractor getBorder() {
			return Optional.ofNullable(self.getBorder())
					.orElse(super.getBorder());
		}

		public Number getBorderId() {
			return Optional.ofNullable(self.getBorderId())
					.orElse(super.getBorderId());
		}

		public XlsxFillExtractor getFill() {
			return Optional.ofNullable(self.getFill()).orElse(super.getFill());
		}

		public Number getFillId() {
			return Optional.ofNullable(self.getFillId())
					.orElse(super.getFillId());
		}

		public XlsxFontExtractor getFont() {
			return Optional.ofNullable(self.getFont()).orElse(super.getFont());
		}

		public Number getFontId() {
			return Optional.ofNullable(self.getFontId())
					.orElse(super.getFontId());
		}

		public XlsxNumFmtExtractor getNumFmt() {
			return Optional.ofNullable(self.getNumFmt())
					.orElse(super.getNumFmt());
		}

		public Number getNumFmtId() {
			return Optional.ofNullable(self.getNumFmtId())
					.orElse(super.getNumFmtId());
		}

		public XlsxProtectionExtractor getProtection() {
			return Optional.ofNullable(self.getProtection())
					.orElse(super.getProtection());
		}

		public Boolean isPivotButton() {
			return Optional.ofNullable(self.isPivotButton())
					.orElse(super.isPivotButton());
		}

		public Boolean isQuotePrefix() {
			return Optional.ofNullable(self.isQuotePrefix())
					.orElse(super.isQuotePrefix());
		}

		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;
			if (!super.equals(o))
				return false;
			XlsxCellXfExtractor that = (XlsxCellXfExtractor) o;
			return Objects.equals(self, that.self);
		}

		public int hashCode() {
			return Objects.hash(super.hashCode(), self);
		}
	}

	public static class XlsxXfExtractor {
		final private Map jsonMap;

		public XlsxXfExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getXfId() {
			return toNum(jsonMap, "xfId");
		}

		public Boolean isSetApplyAlignment() {
			return (Boolean) jsonMap.containsKey("applyAlignment");
		}

		public Boolean getApplyAlignment() {
			return toBool(jsonMap, "applyAlignment", false);
		}

		public Boolean isSetAlignment() {
			return jsonMap.containsKey("alignment");
		}

		public XlsxAlignmentExtractor getAlignment() {
			return jsonMap.containsKey("alignment") ?
					new XlsxAlignmentExtractor((Map) jsonMap.get("alignment"), null) :
					null;
		}

		public XlsxBorderExtractor getBorder() {
			return jsonMap.containsKey("border") ?
					new XlsxBorderExtractor((Map) jsonMap.get("border")) :
					null;
		}

		public Number getBorderId() {
			return jsonMap.containsKey("borderId") ?
					toNum(jsonMap, "borderId") :
					null;
		}

		public XlsxFillExtractor getFill() {
			return jsonMap.containsKey("fill") ?
					XlsxFillExtractor.getInstance((Map) jsonMap.get("fill")) :
					null;
		}

		public Number getFillId() {
			return jsonMap.containsKey("fillId") ?
					toNum(jsonMap, "fillId") :
					null;
		}

		public XlsxFontExtractor getFont() {
			return jsonMap.containsKey("font") ?
					new XlsxFontExtractor((Map) jsonMap.get("font")) :
					null;
		}

		public Number getFontId() {
			return jsonMap.containsKey("fontId") ?
					toNum(jsonMap, "fontId") :
					null;
		}

		public XlsxNumFmtExtractor getNumFmt() {
			return jsonMap.containsKey("numFmt") ?
					new XlsxNumFmtExtractor((Map) (jsonMap.get("numFmt"))) :
					null;
		}

		public Number getNumFmtId() {
			return jsonMap.containsKey("numFmtId") ?
					toNum(jsonMap, "numFmtId") :
					null;
		}

		public XlsxProtectionExtractor getProtection() {
			return jsonMap.containsKey("protection") ?
					new XlsxProtectionExtractor(
							(Map) (jsonMap.get("protection"))) :
					null;
		}

		public Boolean isPivotButton() {
			return toBool(jsonMap, "pivotButton");
		}

		public Boolean isQuotePrefix() {
			return toBool(jsonMap, "quotePrefix");
		}

		public int hashCode() {
			return jsonMap.hashCode();
		}

		public boolean equals(Object obj) {
			if (obj == this)
				return true;
			if (obj instanceof XlsxXfExtractor) {
				return jsonMap.equals(((XlsxXfExtractor) obj).jsonMap);
			}
			return false;
		}
	}

	public static class XlsxProtectionExtractor {
		final private Map jsonMap;
		final private static boolean defaultHidden = false;
		final private static boolean defaultLocked = true;

		public XlsxProtectionExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Boolean isHidden() {
			// KEIKAI-214
			return jsonMap != null ? toBool(jsonMap, "hidden", defaultHidden) : defaultHidden;
		}

		public Boolean isLocked() {
			// KEIKAI-214
			return jsonMap != null ? toBool(jsonMap, "locked", defaultLocked) : defaultLocked;
		}
	}

	public static class XlsxColorExtractor {
		final private Map jsonMap;

		public XlsxColorExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Boolean isAuto() {
			return toBool(jsonMap, "auto");
		}

		public Number getIndexed() {
			return toNum(jsonMap, "indexed");
		}

		public Number getTheme() {
			return toNum(jsonMap, "theme");
		}

		public Number getTint() {
			return toNum(jsonMap, "tint");
		}

		public String getRgb() {
			return toStringValue(jsonMap, "rgb");
		}
	}

	public static class XlsxSchemeClrExtractor {
		final private Map jsonMap;
		final private Map _schemeClr;
		final private Map _srgbClr;

		public XlsxSchemeClrExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			_schemeClr = $(jsonMap, "a:schemeClr");
			_srgbClr = $(jsonMap, "a:srgbClr");
		}

		public boolean isSchemeClr() {
			return _schemeClr != null;
		}

		public boolean isSrgbClr() {
			return _srgbClr != null;
		}

		public String getVal() {
			return toStringValue(jsonMap, "val");
		}

		public String getColorVal() {
			if (isSchemeClr()) {
				return toStringValue(jsonMap, "a:schemeClr");
			} else if (isSrgbClr()) {
				return toStringValue(jsonMap, "a:srgbClr");
			}
			return null;
		}

		public Map getColor() {
			return _schemeClr != null ? _schemeClr : _srgbClr;
		}

		public boolean isModified() {
			return _schemeClr != null ? _schemeClr.size() > 1 : (_srgbClr != null ? _srgbClr.size() > 1 : false);
		}

		public Number getTint() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:tint");
		}
		public Number getShade() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:shade");
		}
		public Number getAlpha() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:alpha");
		}
		public Number getHueMod() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:hueMod");
		}
		public Number getSat() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:sat");
		}
		public Number getSatOff() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:satOff");
		}
		public Number getSatMod() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:satMod");
		}
		public Number getLum() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:lum");
		}
		public Number getLumOff() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:lumOff");
		}
		public Number getLumMod() {
			return toNum(isSchemeClr() ? _schemeClr : _srgbClr, "a:lumMod");
		}

	}

	public static class XlsxAlignmentExtractor {
		final private Map jsonMap;
		private XlsxXfExtractor self;

		public XlsxAlignmentExtractor(Map jsonMap, XlsxXfExtractor self) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			this.self = self;
		}

		public String getHorizontal() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("horizontal")) {
					return (String) (align.getHorizontal());
				}
				return null;
			} else {
				return (String) (jsonMap.get("horizontal"));
			}
		}

		public short getIndent() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("indent")) {
					return Short.valueOf(align.getIndent());
				}
				return 0;
			} else {
				return Optional.ofNullable(toNum(jsonMap, "indent")).orElse(0).shortValue();
			}
		}

		public Boolean isJustifyLastLine() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("justifyLastLine")) {
					return align.isJustifyLastLine();
				}
				return false;
			} else {
				return toBool(jsonMap, "justifyLastLine", false);
			}
		}

		public short getReadingOrder() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("readingOrder")) {
					return Short.valueOf(align.getReadingOrder());
				}
				return 0;
			} else {
				return Optional.ofNullable(toNum(jsonMap, "readingOrder")).orElse(0).shortValue();
			}
		}

		public short getRelativeIndent() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("relativeIndent")) {
					return Short.valueOf(align.getRelativeIndent());
				}
				return 0;
			} else {
				return Optional.ofNullable(toNum(jsonMap, "relativeIndent")).orElse(0).shortValue();
			}
		}

		public Boolean isShrinkToFit() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("shrinkToFit")) {
					return align.isShrinkToFit();
				}
				return false;
			} else {
				return toBool(jsonMap, "shrinkToFit", false);
			}
		}

		public short getTextRotation() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("textRotation")) {
					return Short.valueOf(align.getTextRotation());
				}
				return 0;
			} else {
				return Optional.ofNullable(toNum(jsonMap, "textRotation")).orElse(0).shortValue();
			}
		}

		public String getVertical() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("vertical")) {
					return (String) (align.getVertical());
				}
				if (self.getApplyAlignment()) {
					return null;
				}
			}
			if (jsonMap != null) {
				return (String) (jsonMap.get("vertical"));
			}
			return null;
		}

		public Boolean isWrapText() {
			if (self != null) {
				XlsxAlignmentExtractor align = self.getAlignment();
				if (align != null && align.jsonMap.containsKey("wrapText")) {
					return align.isWrapText();
				}
				return false;
			} else {
				return toBool(jsonMap, "wrapText", false);
			}
		}
	}

	// KEIKAI-395, <x14:dxf> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting><x14:cfRule>
	public static class XlsxExtDxfExtractor extends XlsxDxfExtractor {
		public XlsxExtDxfExtractor(Map jsonMap) {
			super(jsonMap);
		}
	}

	public static class XlsxDxfExtractor {
		final private Map jsonMap;

		public XlsxDxfExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxAlignmentExtractor getAlignment() {
			return jsonMap != null && jsonMap.get("alignment") != null ?
					new XlsxAlignmentExtractor((Map) jsonMap.get("alignment"), null) :
					null;
		}

		public XlsxBorderExtractor getBorder() {
			return jsonMap != null && jsonMap.get("border") != null ?
					new XlsxBorderExtractor((Map) jsonMap.get("border")) :
					null;
		}

		public XlsxFontExtractor getFont() {
			return jsonMap != null && jsonMap.get("font") != null ?
					new XlsxFontExtractor((Map) jsonMap.get("font")) :
					null;
		}

		public XlsxNumFmtExtractor getNumFmt() {
			return jsonMap != null && jsonMap.get("numFmt") != null ?
					new XlsxNumFmtExtractor((Map) (jsonMap.get("numFmt"))) :
					null;
		}

		public XlsxFillExtractor getFill() {
			return jsonMap != null && jsonMap.get("fill") != null ?
					XlsxFillExtractor.getInstance((Map) jsonMap.get("fill")) :
					null;
		}

		public XlsxProtectionExtractor getProtection() {
			return jsonMap != null && jsonMap.get("protection") != null ?
					new XlsxProtectionExtractor(
							(Map) (jsonMap.get("protection"))) :
					null;
		}
	}

	public static class XlsxBorderExtractor {
		final private Map jsonMap;

		public XlsxBorderExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxBorderPrExtractor getBottom() {
			if (jsonMap.containsKey("bottom")) {
				Object data = jsonMap.get("bottom");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}

		public boolean isDiagonalUp() {
			return toBool(jsonMap, "diagonalUp", false);
		}

		public XlsxBorderPrExtractor getDiagonal() {
			if (jsonMap.containsKey("diagonal")) {
				Object data = jsonMap.get("diagonal");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}

		public boolean isDiagonalDown() {
			return toBool(jsonMap, "diagonalDown", false);
		}

		public XlsxBorderPrExtractor getHorizontal() {
			if (jsonMap.containsKey("horizontal")) {
				Object data = jsonMap.get("horizontal");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}

		public XlsxBorderPrExtractor getLeft() {
			if (jsonMap.containsKey("left")) {
				Object data = jsonMap.get("left");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}

		public boolean isOutline() {
			return toBool(jsonMap, "outline", false);
		}

		public XlsxBorderPrExtractor getRight() {
			if (jsonMap.containsKey("right")) {
				Object data = jsonMap.get("right");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}

		public XlsxBorderPrExtractor getTop() {
			if (jsonMap.containsKey("top")) {
				Object data = jsonMap.get("top");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}

		public XlsxBorderPrExtractor getVertical() {
			if (jsonMap.containsKey("vertical")) {
				Object data = jsonMap.get("vertical");
				return data == null ?
						XlsxBorderPrExtractor.EMPTY :
						new XlsxBorderPrExtractor((Map) data);
			}
			return null;
		}
	}

	public static class XlsxBorderPrExtractor {
		public static XlsxBorderPrExtractor EMPTY = new XlsxBorderPrExtractor(
				Collections.EMPTY_MAP);
		final private Map jsonMap;

		public XlsxBorderPrExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxColorExtractor getColor() {
			return jsonMap.containsKey("color") ?
					new XlsxColorExtractor((Map) jsonMap.get("color")) :
					null;
		}

		public String getStyle() {
			return (String) jsonMap.get("style");
		}
	}

	public static class XlsxWorkbookProtectionExtractor {
		final private Map jsonMap;
		public XlsxWorkbookProtectionExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}
		public Boolean isLockStructure() {
			return toBool(jsonMap, "lockStructure");
		}
		public Boolean isLockWindows() {
			return toBool(jsonMap, "lockWindows");
		}
		public String getAlgName() {
			return toStringValue(jsonMap, "workbookAlgorithmName");
		}
		public String getHashValue() {
			return toStringValue(jsonMap, "workbookHashValue");
		}
		public String getSaltValue() {
			return toStringValue(jsonMap, "workbookSaltValue");
		}
		public String getSpinCount() {
			return toStringValue(jsonMap, "workbookSpinCount");
		}
	}

	public static class XlsxSheetExtractor {
		final Map workbookJson;
		final private Map jsonMap;
		final Map sharedStrings;
		final Map worksheetJson;
		final Map calcChain;
		private String _sheetPath;

		public XlsxSheetExtractor(Map workbookJson, Map jsonMap,
				Map sharedStrings, Map calcChain) {
			this.workbookJson = workbookJson;
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			this.sharedStrings = sharedStrings;
			this.calcChain = calcChain;
			Map worksheetJson = (Map) jsonMap.get("worksheet");
			this.worksheetJson = worksheetJson == null ?
					Collections.EMPTY_MAP :
					worksheetJson;
		}

		public String getSheetPath() {
			if (_sheetPath == null) {
				List<Map> workbookRels = $(workbookJson,
						"xl/_rels/workbook.xml.rels", "Relationships",
						"Relationship");
				Map<String, String> idMap = new LinkedHashMap<>();// rid -> sheetPath
				workbookRels.forEach((rel) -> {
					idMap.put((String) rel.get("Id"),
							(String) rel.get("Target"));
				});

				String rid = (String) this.jsonMap.get("r:id");
				String[] array = idMap.get(rid).split("/");
				_sheetPath = array[array.length - 1];
			}
			return _sheetPath;
		}

		public List<XlsxColExtractor> getCols() {
			return toList($(worksheetJson, "cols"), "col", XlsxColExtractor.class);
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public List<XlsxRowExtractor> getRows() {
			SheetData sheetData = $(worksheetJson, "sheetData");
			if (sheetData != null) {
				Object rows = sheetData.get("row");
				if (rows instanceof List) {
					return (List<XlsxRowExtractor>) ((List) rows).stream()
							.map((r) -> new XlsxRowExtractor((Row) r,
									sharedStrings, calcChain))
							.collect(Collectors.toList());
				} else if (rows != null) {
					return asList(
							new XlsxRowExtractor((Row) rows, sharedStrings,
									calcChain));
				}
			}
			return Collections.EMPTY_LIST;
		}

		public List<XlsxTableExtractor> getTables() {
			String sheetPath = getSheetPath();
			Map sheetRels = getSheetRefs();
			if (sheetRels != null) {
				Map tableRels = _getSheetTableRels(sheetPath, workbookJson);
				final Object tableParts = $(worksheetJson, "tableParts",
						"tablePart");
				if (tableParts != null) {
					if (tableParts instanceof List) {
						return ((List<Map>) tableParts).stream().map((part) -> {
							final String rid = (String) part.get("r:id");
							final String tableRel = tableRels.get(rid).toString();
							final String tablePath = _getPath(tableRel, "xl");
							final String tableName = getLast(tablePath.split("/"));
							final Map rels = $(workbookJson,
									"xl/tables/_rels/" + tableName + ".rels");
							String queryTablePath = null;
							if (rels != null) {
								final Map rel = $(rels, "Relationships",
										"Relationship");
								if ("http://schemas.openxmlformats.org/officeDocument/2006/relationships/queryTable"
										.equals(rel.get("Type"))) {
									queryTablePath = (String) rel.get("Target");
								}
							}
							return new XlsxTableExtractor(
									$(this.workbookJson, tablePath, "table"),
									queryTablePath != null ?
											$(this.workbookJson, _getPath(queryTablePath, "xl")) : null);
						}).collect(Collectors.toList());
					} else {
						final String tableRel = tableRels.values().iterator().next().toString();
						final String tablePath = _getPath(tableRel, "xl");
						final String tableName = getLast(tablePath.split("/"));
						final Map rels = $(this.workbookJson,
								"xl/tables/_rels/" + tableName + ".rels");
						String queryTablePath = null;
						if (rels != null) {
							final Map rel = $(rels, "Relationships",
									"Relationship");
							if ("http://schemas.openxmlformats.org/officeDocument/2006/relationships/queryTable"
									.equals(rel.get("Type"))) {
								queryTablePath = (String) rel.get("Target");
							}
						}
						return asList(new XlsxTableExtractor(
								$(this.workbookJson, tablePath, "table"),
								queryTablePath != null ?
										(Map) $(this.workbookJson, _getPath(queryTablePath, "xl")) :
										null));
					}
				}
			}
			return Collections.EMPTY_LIST;
		}

		public int getId() {
			String rId = toValue(jsonMap, "r:id");
			if (rId != null) {
				return Integer.parseInt(rId.substring(3));
			}
			return -1;
		}

		public String getSheetId() {
			return toStringValue(jsonMap, "sheetId");
		}

		public String getState() {
			return toStringValue(jsonMap, "state");
		}

		public String getDimension() {
			return toStringValue($(worksheetJson, "dimension"), "ref");
		}

		public XlsxDataValidationsExtractor getDataValidations() {
			Map dataValidations = (Map) worksheetJson.get("dataValidations");
			if (dataValidations != null) {
				return new XlsxDataValidationsExtractor((Map) dataValidations);
			}
			return null;
		}

		public XlsxSheetProtectionExtractor getSheetProtection() {
			Map sheetProtection = $(worksheetJson, "sheetProtection");
			if (sheetProtection != null) {
				return new XlsxSheetProtectionExtractor(sheetProtection);
			}
			return null;
		}

		public XlsxAutoFilterExtractor getAutoFilter() {
			Map autoFilter = (Map) worksheetJson.get("autoFilter");
			if (autoFilter != null) {
				return new XlsxAutoFilterExtractor(autoFilter);
			}
			return null;
		}

		public List<XlsxConditionalFormattingExtractor> getConditionalFormattings() {
			return toList(worksheetJson, "conditionalFormatting", XlsxConditionalFormattingExtractor.class);
		}

		public List<XlsxHyperlinkExtractor> getHyperlinks() {
			String sheetPath = getSheetPath();
			Map sheetRels = $(this.workbookJson,
					"xl/worksheets/_rels/" + sheetPath + ".rels");
			Map hyperlinkRels = null;
			if (sheetRels != null) {
				Object relations = $(sheetRels, "Relationships",
						"Relationship");
				if (relations instanceof List) {
					hyperlinkRels = ((List<Map>) relations).stream()
							.filter((r) -> "http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink"
									.equals(r.get("Type"))).collect(Collectors
									.toMap((r) -> r.get("Id"),
											(r) -> r.get("Target")));
				} else if (relations instanceof Map) {
					Map rels = (Map) relations;
					if ("http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink"
							.equals(rels.get("Type"))) {
						hyperlinkRels = asMap(rels.get("Id"),
								rels.get("Target"));
					}
				}
			}
			final Object hyperlink = $(worksheetJson, "hyperlinks",
					"hyperlink");
			if (hyperlink != null) {
				if (hyperlink instanceof List) {
					final Map finalHyperlinkRels = hyperlinkRels;
					return ((List<Map>) hyperlink).stream().map((e) -> {
						if (e.containsKey("r:id")) {
							e.put("r:id", $(finalHyperlinkRels,
									(String) e.get("r:id")));
						}
						return new XlsxHyperlinkExtractor(e);
					}).collect(Collectors.toList());
				} else if (hyperlink instanceof Map) {
					Map hMap = (Map) hyperlink;
					final String rId = (String) hMap.get("r:id");
					if (rId != null) {
						hMap.put("r:id", hyperlinkRels.get(rId));
					}
					return asList(new XlsxHyperlinkExtractor(hMap));
				}
			}
			return Collections.EMPTY_LIST;
		}

		//
		//		public SortStateExtractor getSortState() {
		//			return //worksheetJson.containsKey("sortState")?new XlsxSortStateExtractor((Map)(worksheetJson.get("sortState"))):null
		//					;
		//		}
		//
		public List<XlsxSheetViewExtractor> getSheetViews() {
			Object sheetView = $(jsonMap, "worksheet", "sheetViews",
					"sheetView");
			if (sheetView != null) {
				if (sheetView instanceof List) {
					return (List<XlsxSheetViewExtractor>) ((List) sheetView)
							.stream()
							.map((e) -> new XlsxSheetViewExtractor((Map) (e)))
							.collect(Collectors.toList());
				} else {
					return asList(new XlsxSheetViewExtractor((Map) sheetView));
				}
			}
			return Collections.EMPTY_LIST;
		}

		public List<String> getMergeCells() {
			Object mergeCell = $(worksheetJson, "mergeCells", "mergeCell");
			if (mergeCell instanceof List) {
				return ((List<Map>) mergeCell).stream()
						.map((e) -> (String) e.get("ref"))
						.collect(Collectors.toList());
			} else if (mergeCell instanceof Map) {
				return asList((String) ((Map) mergeCell).get("ref"));
			}
			return Collections.EMPTY_LIST;
		}

		public XlsxSheetFormatPrExtractor getSheetFormatPr() {
			Map sheetFormatPr = $(worksheetJson, "sheetFormatPr");
			if (sheetFormatPr != null) {
				return new XlsxSheetFormatPrExtractor(sheetFormatPr);
			}
			return null;
		}

		public XlsxPageMarginsExtractor getPageMargins() {
			if (worksheetJson.containsKey("pageMargins")) {
				return new XlsxPageMarginsExtractor((Map)(worksheetJson.get("pageMargins")))
						;
			}
			return null;
		}

		public XlsxPageSetupExtractor getPageSetup() {
			if (worksheetJson.containsKey("pageSetup")) {
				return new XlsxPageSetupExtractor((Map)(worksheetJson.get("pageSetup")))
						;
			}
			return null;
		}

		public XlsxPrintOptionsExtractor getPrintOptions() {
			if (worksheetJson.containsKey("printOptions")) {
				return new XlsxPrintOptionsExtractor((Map)(worksheetJson.get("printOptions")))
						;
			}
			return null;
		}

		public XlsxHeaderFootersExtractor getHeaderFooter() {
			if (worksheetJson.containsKey("headerFooter")) {
				return new XlsxHeaderFootersExtractor((Map)(worksheetJson.get("headerFooter")))
						;
			}
			return null;
		}

		public int[] getColBreaks() {
			Object brk = $(worksheetJson, "colBreaks", "brk");
			if (brk == null) {
				return new int[0];
			}
			if (brk instanceof List) {
				return ((List<Map>)brk).stream().mapToInt((b) -> toInt(b, "id", 0) - 1).toArray();
			} else {
				Map data =(Map) brk;
				return new int[] {toInt(data, "id", 0) - 1};
			}
		}

		public int[] getRowBreaks() {
			Object brk = $(worksheetJson, "rowBreaks", "brk");
			if (brk == null) {
				return new int[0];
			}
			if (brk instanceof List) {
				return ((List<Map>)brk).stream().mapToInt((b) -> toInt(b, "id", 0) - 1).toArray();
			} else {
				Map data =(Map) brk;
				return new int[] {toInt(data, "id", 0) - 1};
			}
		}

		//
		//		public List<ShapeJson> getDrawingsJson() {
		//			/*final*/	Object sheetPath = getSheetPath;
		//			/*final*/	Object sheetRels = this.workbookJson['xl/worksheets/_rels/${sheetPath}.rels");
		//			if (sheetRels != null) {
		//				/*final*/	Object drawingsRels = _getSheetDrawingRels(sheetPath, workbookJson);
		//				if (drawingsRels.isNotEmpty == true) {
		//					/*final*/	Object wsDr = this.workbookJson['xl/${drawingsRels.values.first.substring(3)}")["xdr:wsDr"];
		//					/*var*/	Object anchors = JsonHelper.toJsonEntry(wsDr["xdr:twoCellAnchor"]);
		//					/*final*/	Object wsDrEntry = JsonHelper.toJsonEntry(wsDr);
		//					/*var*/	Object oneAnchor = wsDrEntry.into("xdr:oneCellAnchor");
		//					/*var*/	Object altAnchors = wsDrEntry.jump("mc:AlternateContent");
		//					/*final*/	Object fileName = drawingsRels.values.first.split("/").last;
		//					/*final*/	Object drawingRels = this.workbookJson['xl/drawings/_rels/${fileName}.rels");
		//					/*final*/	Object result = anchors.asList.where((anchor)=>anchor.isNotNull).map((viewAnchor){final cChart=viewAnchor.jump("xdr:graphicFrame/a:graphic/a:graphicData/c:chart");final pic=viewAnchor.jump("xdr:pic");final grpSpPic=viewAnchor.jump("xdr:grpSp/xdr:pic");if(cChart.isNotNull){final relation=JsonHelper.toJsonEntry(drawingRels['Relationships")['Relationship")).asList.firstWhere((entry)=>entry.get("Id")==cChart.get("r:id"));if(relation.isNotNull){final chartPath=relation.get("Target").asString.substring(3);final chartFile=chartPath.split("/").last;final chartData=JsonHelper.toJsonEntry(this.workbookJson['xl/${chartPath}"));final chartRef=JsonHelper.toJsonEntry(this.workbookJson['xl/charts/_rels/${chartFile}.rels")).jump("Relationships/Relationship").asList;final chartColorStyleTarget=chartRef.firstWhere((rel)=>rel.get("Type").asString=='http://schemas.microsoft.com/office/2011/relationships/chartColorStyle',orElse:()=>JsonHelper.toJsonEntry(null)).get("Target").asString;final chartStyleTarget=chartRef.firstWhere((rel)=>rel.get("Type").asString=='http://schemas.microsoft.com/office/2011/relationships/chartStyle',orElse:()=>JsonHelper.toJsonEntry(null)).get("Target").asString;final chartColorStyle=JsonHelper.toJsonEntry(this.workbookJson['xl/charts/${chartColorStyleTarget}"));final chartStyle=JsonHelper.toJsonEntry(this.workbookJson['xl/charts/${chartStyleTarget}"));return new ChartSpaceJson(chartSpace:chartData,viewAnchor:viewAnchor,chartColorStyle:chartColorStyle,chartStyle:chartStyle,uuid:viewAnchor.jump("xdr:graphicFrame/xdr:nvGraphicFramePr/xdr:cNvPr").get("id").asString);}throw new UnsupportedError("The chart relation cannot be null!");}else if(pic.isNotNull||grpSpPic.isNotNull){createPic(JsonEntry pic0){final relation=JsonHelper.toJsonEntry(drawingRels['Relationships")['Relationship")).asList.firstWhere((entry){final blip=pic0.jump("xdr:blipFill/a:blip");return blip.get("r:embed").isNotNull==true?entry.get("Id")==blip.get("r:embed"):entry.get("Id")==blip.get("r:link");});if(relation.isNotNull){if(relation.get("TargetMode").asString=='External"){final imagePath=relation.get("Target").asString;return new PictureJson(viewAnchor,pic0,filePath:imagePath,uuid:pic0.jump("xdr:nvPicPr/xdr:cNvPr").get("id").asString);}else{final imagePath=relation.get("Target").asString.substring(3);final imageFile=imagePath.split("/").last;return new PictureJson(viewAnchor,pic0,fileContent:unsafeCast(this.workbookJson['xl/media/${imageFile}")),fileType:imageFile.split(".").last,uuid:pic0.jump("xdr:nvPicPr/xdr:cNvPr").get("id").asString);}}return null;}if(pic.isNotNull){return createPic(pic);}else{final hlinkClick=viewAnchor.jump("xdr:grpSp/xdr:nvGrpSpPr/xdr:cNvPr/a:hlinkClick");Tuple2<String,String> hlink;if(hlinkClick.isNotNull){final rId=hlinkClick.get("r:id").asString;final relation=JsonHelper.toJsonEntry(drawingRels['Relationships")['Relationship")).asList.firstWhere((entry){return entry.get("Id").asString==rId;});hlink=new Tuple2(relation.get("Target").asString,hlinkClick.get("tooltip").asString);}final gpic=new GroupPictureJson(viewAnchor,hlink,uuid:viewAnchor.jump("xdr:grpSp/xdr:nvGrpSpPr/xdr:cNvPr").get("id").asString);grpSpPic.asList.forEach((pic0){gpic.addPicture(createPic(pic0));});return gpic;}}else{final hlinkClick=viewAnchor.jump("xdr:sp/xdr:nvSpPr/xdr:cNvPr/a:hlinkClick");Tuple2<String,String> hlink;if(hlinkClick.isNotNull){final rId=hlinkClick.get("r:id").asString;final relation=JsonHelper.toJsonEntry(drawingRels['Relationships")['Relationship")).asList.firstWhere((entry){return entry.get("Id").asString==rId;});hlink=new Tuple2(relation.get("Target").asString,hlinkClick.get("tooltip").asString);}return new DrawingShapeJson(viewAnchor,hlink);}}).toList;
		//					if (oneAnchor.isNotNull) {
		//						result.addAll(oneAnchor.asList.map((viewAnchor){final hlinkClick=viewAnchor.jump("xdr:sp/xdr:nvSpPr/xdr:cNvPr/a:hlinkClick");Tuple2<String,String> hlink;if(hlinkClick.isNotNull){final rId=hlinkClick.get("r:id").asString;final relation=JsonHelper.toJsonEntry(drawingRels['Relationships")['Relationship")).asList.firstWhere((entry){return entry.get("Id").asString==rId;});hlink=new Tuple2(relation.get("Target").asString,hlinkClick.get("tooltip").asString);}return new UnimplementedShapeJson(viewAnchor,hlink);}));
		//					}
		//					if (altAnchors.isNotNull) {
		//						/*final*/	Object altContentList = JsonHelper.toJsonEntry(this.worksheetJson["mc:AlternateContent"]).jump("mc:Choice/controls/mc:AlternateContent").asList;
		//						if (altContentList.isNotNull) {
		//							result.addAll(altAnchors.asList.map((alternateContent){final twoCellAnchor=alternateContent.jump("mc:Choice/xdr:twoCellAnchor");final shapeId=twoCellAnchor.jump("xdr:sp/xdr:nvSpPr/xdr:cNvPr").get("id").asString;final altContent=altContentList.firstWhere((entry)=>entry.jump("mc:Choice/control").get("shapeId").asString==shapeId,orElse:()=>null);final rid=altContent.jump("mc:Choice/control").get("r:id").asString;final relation=JsonHelper.toJsonEntry(sheetRels['Relationships")['Relationship")).asList.firstWhere((entry){return entry.get("Id").asString==rid;});final targetPath=relation.get("Target").asString.replaceAll("../','");return new CtrlShapeJson(twoCellAnchor,JsonHelper.toJsonEntry(this.workbookJson['xl/${targetPath}")['formControlPr")),uuid:shapeId);}));
		//						}
		//					}
		//					return result;
		//				}
		//			}
		//			return asList();
		//		}
		//
		public XlsxCommentsExtractor getComments() {
			Map sheetRels = getSheetRefs();
			if (sheetRels != null) {
				Map<String, String> commentsRels = _getSheetCommentsRels(
						getSheetPath(), workbookJson);
				if (commentsRels != null && !commentsRels.isEmpty()) {
					final String commentRel = commentsRels.values().iterator().next();
					return new XlsxCommentsExtractor($(workbookJson, _getPath(commentRel, "xl"), "comments"));
				}
			}
			return null;
		}

		private Map sheetRefs;
		public Map getSheetRefs() {
			if (sheetRefs == null) {
				sheetRefs = $(workbookJson,
						"xl/worksheets/_rels/" + getSheetPath() + ".rels");
			}
			return sheetRefs;
		}
//		private Map drawingRefs;
//		public Map getDrawingRefs() {
//			if (drawingRefs == null) {
//				final String sheetPath = getSheetPath();
//				final Map sheetRels = getSheetRefs();
//				if (sheetRels != null) {
//					final Map<String, String> drawingRels = _getSheetDrawingRels(sheetPath, workbookJson);
//					if (drawingRels != null && !drawingRels.isEmpty()) {
//						final String filePath = drawingRels.values().iterator().next().substring(3);
//						final String fileName = getLast(filePath.split("/"));
//						drawingRefs = $(workbookJson, "xl/drawings/_rels/" + fileName + ".rels");
//					}
//				}
//			}
//			return drawingRefs;
//		}

		public XlsxDrawingsExtractor getDrawings() {
			final String sheetPath = getSheetPath();
			final Map sheetRels = getSheetRefs();
			if (sheetRels != null) {
				final Map<String, String> drawingRels = _getSheetDrawingRels(sheetPath, workbookJson);
				if (drawingRels != null && !drawingRels.isEmpty()) {
					final String drv = drawingRels.values().iterator().next();
					final String filePath = _getPath(drv, "xl");
					final String fileName = getLast(drv.split("/"));
					final Map drawingRefs = $(workbookJson, "xl/drawings/_rels/" + fileName + ".rels");
					final Map wsDr = $(workbookJson, filePath, "xdr:wsDr");
					return new XlsxDrawingsExtractor(workbookJson, drawingRefs, wsDr);
				}
			}
			return null;
		}

		public XlsxVmlDrawingExtractor getLegacyDrawing() {
			String sheetPath = getSheetPath();
			Map sheetRels = getSheetRefs();
			if (sheetRels != null) {
				Map<String, String> vmlDrawingRels = _getSheetVmlDrawingRels(sheetPath, workbookJson);
				if (vmlDrawingRels != null && !vmlDrawingRels.isEmpty()) {
					// todo: make it importable/exportable only
					final String vmlDrawingRel = vmlDrawingRels.values().iterator().next();
					String filePath = _getPath(vmlDrawingRel, "xl");
					Map vmlJson = $(this.workbookJson, filePath);
					Map vImageData = $x(vmlJson, "xml/v:shape/v:imagedata");
					if (vImageData != null) {
						List<Map> relations;
						Object relationsObj = $(this
										.workbookJson, "xl/drawings/_rels/" + getLast(filePath.split("/")) + ".rels", "Relationships", "Relationship");
						if (relationsObj instanceof JSONObject) {
							relations = new LinkedList<>();
							relations.add((Map) relationsObj);
						} else {
							relations = (List<Map>) relationsObj;
						}
						Map relation = null;
						if (relations != null) {
							relation = relations.stream().filter((entry) -> Objects.equals(entry.get("Id"), vImageData.get("o:relid"))).findFirst().get();
						}
						if (relation == null)
							throw new IllegalStateException("vmlDrawing relation should not be null! [xl/drawings/_rels/" + filePath + ".rels]");

						final String relationTarget = relation.get("Target").toString();
						String imagePath = _getPath(relationTarget, "xl");
						return new XlsxVmlDrawingExtractor(vmlJson, (byte[]) this.workbookJson.get(imagePath));
					} else {
						return new XlsxVmlDrawingExtractor(vmlJson, null);
					}
				}
			}
			return null;
		}
		//
		//		public String getPrinterSettings() {
		//			/*final*/	Object sheetPath = getSheetPath;
		//			/*final*/	Object sheetRels = this.workbookJson['xl/worksheets/_rels/${sheetPath}.rels");
		//			if (sheetRels != null) {
		//				/*final*/	Object printerRels = _getSheetPrinterRels(sheetPath, workbookJson);
		//				if (printerRels.isNotEmpty == true) {
		//					return unsafeCast(this.workbookJson['xl/${printerRels.values.first.substring(3)}"));
		//				}
		//			}
		//			return null;
		//		}
		//

		public XlsxSheetPropertiesExtractor getSheetPr() {
			if (worksheetJson.containsKey("sheetPr")) {
				return new XlsxSheetPropertiesExtractor((Map)(worksheetJson.get("sheetPr")));
			}
			return null;
		}

		public Map getAlternateContent() {
			return $(worksheetJson, "mc:AlternateContent");
		}

		public XlsxExtLstExtractor getExtLst() {
			if (worksheetJson.containsKey("extLst")) {
				return new XlsxExtLstExtractor((Map)(worksheetJson.get("extLst")));
			}
			return null;
		}
	}

	public static class XlsxExtLstExtractor {
		final private Map jsonMap;

		public XlsxExtLstExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxExtExtractor> getExt() {
			Object ext = jsonMap.get("ext");
			if (ext instanceof List) {
				return ((List<Map>)ext).stream().map((s) -> new XlsxExtExtractor(s)).collect(Collectors.toList());
			} else if (ext instanceof Map) {
				return asList(new XlsxExtExtractor((Map)ext));
			}
			return Collections.EMPTY_LIST;
		}
	}

	public static class XlsxExtExtractor {
		final private Map jsonMap;

		public XlsxExtExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxSparklineGroupsExtractor getSparklineGroups() {
			if (jsonMap.containsKey("x14:sparklineGroups")) {
				return new XlsxSparklineGroupsExtractor((Map)(jsonMap.get("x14:sparklineGroups")));
			}
			return null;
		}

		// KEIKAI-395
		public List<XlsxExtConditionalFormattingExtractor> getConditionalFormattings() {
			return toList((Map)jsonMap.get("x14:conditionalFormattings"), "x14:conditionalFormatting", XlsxExtConditionalFormattingExtractor.class);
		}

		// KEIKAI-464
		public List<XlsxExtDataValidationExtractor> getDataValidations() {
			return toList((Map)jsonMap.get("x14:dataValidations"), "x14:dataValidation", XlsxExtDataValidationExtractor.class);
		}
	}

	public static class XlsxSparklineGroupsExtractor {
		final private Map jsonMap;

		public XlsxSparklineGroupsExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxSparklineGroupExtractor> getSparklineGroups() {
			Object sparklineGroup = jsonMap.get("x14:sparklineGroup");
			if (sparklineGroup instanceof List) {
				return ((List<Map>)sparklineGroup).stream().map((s) -> new XlsxSparklineGroupExtractor(s)).collect(Collectors.toList());
			} else if (sparklineGroup instanceof Map) {
				return asList(new XlsxSparklineGroupExtractor((Map)sparklineGroup));
			}
			return Collections.EMPTY_LIST;
		}
	}

	public static class XlsxSparklineGroupExtractor {
		final private Map jsonMap;

		public XlsxSparklineGroupExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getType() {
			return toStringValue(jsonMap, "type");
		}

		public XlsxSparklinesExtractor getSparklines() {
			if (jsonMap.containsKey("x14:sparklines")) {
				return new XlsxSparklinesExtractor((Map) jsonMap.get("x14:sparklines"));
			}
			return null;
		}

		public Boolean isMarkersVisible() {
			return toBool(jsonMap, "markers");
		}

		public Boolean isNegativeVisible() {
			return toBool(jsonMap, "negative");
		}

		public Boolean isFirstVisible() {
			return toBool(jsonMap, "first");
		}

		public Boolean isLastVisible() {
			return toBool(jsonMap, "last");
		}

		public Boolean isHighVisible() {
			return toBool(jsonMap, "high");
		}

		public Boolean isLowVisible() {
			return toBool(jsonMap, "low");
		}

		public XlsxColorExtractor getColorSeries() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorSeries"));
		}

		public XlsxColorExtractor getColorMarkers() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorMarkers"));
		}

		public XlsxColorExtractor getColorNegative() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorNegative"));
		}

		public XlsxColorExtractor getColorFirst() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorFirst"));
		}

		public XlsxColorExtractor getColorLast() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorLast"));
		}

		public XlsxColorExtractor getColorHigh() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorHigh"));
		}

		public XlsxColorExtractor getColorLow() {
			return new XlsxColorExtractor($(jsonMap, "x14:colorLow"));
		}
	}

	public static class XlsxSparklinesExtractor {
		final private Map jsonMap;

		public XlsxSparklinesExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxSparklineExtractor> getSparklines() {
			Object sparklines = jsonMap.get("x14:sparkline");
			if (sparklines instanceof List) {
				return ((List<Map>)sparklines).stream().map(XlsxSparklineExtractor::new).collect(Collectors.toList());
			} else if (sparklines instanceof Map) {
				return asList(new XlsxSparklineExtractor((Map)sparklines));
			}
			return Collections.EMPTY_LIST;
		}
	}

	public static class XlsxSparklineExtractor {
		final private Map jsonMap;

		public XlsxSparklineExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getF() {
			return toStringValue(jsonMap, "xm:f");
		}

		public String getSqref() {
			return toStringValue(jsonMap, "xm:sqref");
		}
	}

	public static class XlsxSheetViewExtractor {
		final private Map jsonMap;

		public XlsxSheetViewExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Boolean isWindowProtection() {
			return toBool(jsonMap, "windowProtection");
		}

		public Boolean isShowFormulas() {
			return toBool(jsonMap, "showFormulas");
		}

		public Boolean isShowGridLines() {
			return toBool(jsonMap, "showGridLines");
		}

		public Boolean isShowRowColHeaders() {
			return toBool(jsonMap, "showRowColHeaders");
		}

		public Boolean isShowZeros() {
			return toBool(jsonMap, "showZeros");
		}

		public Boolean isRightToLeft() {
			return toBool(jsonMap, "rightToLeft");
		}

		public Boolean isTabSelected() {
			Boolean tabSelected = toBool(jsonMap, "tabSelected");
			if (tabSelected == null) return Boolean.FALSE;
			return tabSelected;
		}

		public Boolean isShowRuler() {
			return toBool(jsonMap, "showRuler");
		}

		public Boolean isShowOutlineSymbols() {
			return toBool(jsonMap, "showOutlineSymbols");
		}

		public Boolean isDefaultGridColor() {
			return toBool(jsonMap, "defaultGridColor");
		}

		public Boolean isShowWhiteSpace() {
			return toBool(jsonMap, "showWhiteSpace");
		}

		public String getView() {
			return toStringValue(jsonMap, "view");
		}

		public String getTopLeftCell() {
			return toStringValue(jsonMap, "topLeftCell");
		}

		public Number getColorId() {
			return toNum(jsonMap, "colorId");
		}

		public Number getZoomScale() {
			return toNum(jsonMap, "zoomScale");
		}

		public Number getZoomScaleNormal() {
			return toNum(jsonMap, "zoomScaleNormal");
		}

		public Number getZoomScaleSheetLayoutView() {
			return toNum(jsonMap, "zoomScaleSheetLayoutView");
		}

		public Number getZoomScalePageLayoutView() {
			return toNum(jsonMap, "zoomScalePageLayoutView");
		}

		public Number getWorkbookViewId() {
			return toNum(jsonMap, "workbookViewId");
		}

		public XlsxPaneExtractor getPane() {
			return jsonMap.containsKey("pane") ?
					new XlsxPaneExtractor($(jsonMap, "pane")) :
					null;
		}

		public List<XlsxSelectionExtractor> getSelections() {
			return  toList(jsonMap, "selection", XlsxSelectionExtractor.class);
		}
	}

	public static class XlsxPaneExtractor {
		final private Map jsonMap;

		public XlsxPaneExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getXSplit() {
			return Optional.ofNullable(toNum(jsonMap, "xSplit")).orElse(0)
					.intValue();
		}

		public int getYSplit() {
			return Optional.ofNullable(toNum(jsonMap, "ySplit")).orElse(0)
					.intValue();
		}

		public String getTopLeftCell() {
			return toStringValue(jsonMap, "topLeftCell");
		}

		public String getActivePane() {
			return toStringValue(jsonMap, "activePane");
		}

		public String getState() {
			return toStringValue(jsonMap, "state");
		}
	}

	public static class XlsxSelectionExtractor {
		final private Map jsonMap;

		public XlsxSelectionExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getPane() {
			return toStringValue(jsonMap, "pane");
		}

		public String getActiveCell() {
			return toStringValue(jsonMap, "activeCell");
		}

		public Number getActiveCellId() {
			return toNum(jsonMap, "activeCellId");
		}

		public String getSqref() {
			return toStringValue(jsonMap, "sqref");
		}
	}

	public static class XlsxSheetPropertiesExtractor {
		final private Map jsonMap;

		public XlsxSheetPropertiesExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxColorExtractor getTabColor() {
			return jsonMap.containsKey("tabColor") ?
					new XlsxColorExtractor((Map) jsonMap.get("tabColor")) : null;
		}

		public XlsxOutlinePrExtractor getOutlinePr() {
			return jsonMap.containsKey("outlinePr") ?
					new XlsxOutlinePrExtractor((Map) jsonMap.get("outlinePr")) : null;
		}
	}

	public static class XlsxOutlinePrExtractor {
		final private Map jsonMap;

		public XlsxOutlinePrExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}
		public boolean isApplyStyles() {
			return toBool(jsonMap, "applyStyles", false);
		}
		public boolean isSummaryBelow() {
			return toBool(jsonMap, "summaryBelow", true);
		}
		public boolean isSummaryRight() {
			return toBool(jsonMap, "summaryRight", true);
		}
		public boolean isShowOutlineSymbols() {
			return toBool(jsonMap, "showOutlineSymbols", true);
		}
	}

	public static class XlsxSheetFormatPrExtractor {
		final private Map jsonMap;

		public XlsxSheetFormatPrExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getBaseColWidth() {
			return Optional.ofNullable(toInt(jsonMap, "baseColWidth"))
					.orElse(8);
		}

		public boolean isSetBaseColWidth() {
			return jsonMap.containsKey("baseColWidth");
		}

		public Boolean isCustomHeight() {
			return toBool(jsonMap, "customHeight");
		}

		public Number getDefaultColWidth() {
			return toNum(jsonMap, "defaultColWidth");
		}

		public boolean isSetDefaultColWidth() {
			return jsonMap.containsKey("defaultColWidth");
		}

		public float getDefaultRowHeight() {
			return Optional.ofNullable(toFloat(jsonMap, "defaultRowHeight"))
					.orElse(20f).floatValue();
		}

		public Number getOutlineLevelCol() {
			return toNum(jsonMap, "outlineLevelCol");
		}

		public Number getOutlineLevelRow() {
			return toNum(jsonMap, "outlineLevelRow");
		}

		public Boolean isThickBottom() {
			return toBool(jsonMap, "thickBottom");
		}

		public Boolean isThickTop() {
			return toBool(jsonMap, "thickTop");
		}

		public Boolean isZeroHeight() {
			return toBool(jsonMap, "zeroHeight");
		}
	}

	public static class XlsxSheetProtectionExtractor {
		final private Map jsonMap;

		public XlsxSheetProtectionExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getPassword() {
			return toStringValue(jsonMap, "password");
		}
		public String getHashValue() {
			return toStringValue(jsonMap, "hashValue");
		}
		public String getSpinCount() {
			return toStringValue(jsonMap, "spinCount");
		}
		public String getSaltValue() {
			return toStringValue(jsonMap, "saltValue");
		}
		public String getAlgName() {
			return toStringValue(jsonMap, "algorithmName");
		}

		public boolean isSheet() {
			return toBool(jsonMap, "sheet", false);
		}

		public boolean isObjects() {
			return isSheet() && !(toBool(jsonMap, "objects", false));
		}

		public boolean isScenarios() {
			return isSheet() && !(toBool(jsonMap, "scenarios", false));
		}

		public boolean isFormatCells() {
			return !(toBool(jsonMap, "formatCells", true));
		}

		public boolean isFormatColumns() {
			return !(toBool(jsonMap, "formatColumns", true));
		}

		public boolean isFormatRows() {
			return !(toBool(jsonMap, "formatRows", true));
		}

		public boolean isInsertColumns() {
			return !(toBool(jsonMap, "insertColumns", true));
		}

		public boolean isInsertRows() {
			return !(toBool(jsonMap, "insertRows", true));
		}

		public boolean isInsertHyperlinks() {
			return !(toBool(jsonMap, "insertHyperlinks", true));
		}

		public boolean isDeleteColumns() {
			return !(toBool(jsonMap, "deleteColumns", true));
		}

		public boolean isDeleteRows() {
			return !(toBool(jsonMap, "deleteRows", true));
		}

		public boolean isSelectLockedCells() {
			return !(toBool(jsonMap, "selectLockedCells", false));
		}

		public boolean isSort() {
			return !(toBool(jsonMap, "sort", true));
		}

		public boolean isAutoFilter() {
			return !(toBool(jsonMap, "autoFilter", true));
		}

		public boolean isPivotTables() {
			return !(toBool(jsonMap, "pivotTables", true));
		}

		public boolean isSelectUnlockedCells() {
			return !(toBool(jsonMap, "selectUnlockedCells", false));
		}
	}

	public static class XlsxRowExtractor {
		final Row jsonMap;
		final Map sharedStrings;
		final Map calcChain;

		public XlsxRowExtractor(Row jsonMap, Map sharedStrings, Map calcChain) {
			this.jsonMap = jsonMap;
			this.sharedStrings = sharedStrings;
			this.calcChain = calcChain;
		}

		public Boolean isCollapsed() {
			return toBoolean(jsonMap.get("collapsed"), null);
		}

		public Boolean isCustomFormat() {
			return toBoolean(jsonMap.get("customFormat"), null);
		}

		public Boolean isCustomHeight() {
			return toBoolean(jsonMap.get("customHeight"), null);
		}

		public Boolean isHidden() {
			return toBoolean(jsonMap.get("hidden"), null);
		}

		public int getIndex() {
			return Integer.parseInt((String) jsonMap.get("r"))
					- 1; // 1-based to 0-based
		}

		public Number getOutlineLevel() {
			Object outlineLevel = jsonMap.get("outlineLevel");
			return outlineLevel != null ?
					Double.parseDouble((String) outlineLevel) :
					null;
		}

		public Boolean isPhonetic() {
			return toBoolean(jsonMap.get("ph"), null);
		}

		public float getRowHeight() {
			Object ht = jsonMap.get("ht");
			return ht != null ? (float) Float.parseFloat((String) ht) : -1f;
		}

		public XlsxSpanExtractor getSpans() {
			Object spans = jsonMap.get("spans");
			return spans instanceof String ?
					new XlsxSpanExtractor((String) spans) :
					null;
		}

		public int getStyle() {
			Object s = jsonMap.get("s");
			return s != null ? Integer.parseInt((String) s) : -1;
		}

		public Boolean isThickBot() {
			return toBoolean(jsonMap.get("thickBot"), null);
		}

		public Boolean isThickTop() {
			return toBoolean(jsonMap.get("thickTop"), null);
		}

		public List<XlsxCellExtractor> getCells() {
			JSONArray<JSONAware> cells = jsonMap.getCells();
			if (cells != null) {
				return (List<XlsxCellExtractor>) (cells.stream()
						.map((d) -> new XlsxCellExtractor((Cell) d,
								sharedStrings, calcChain))
						.collect(Collectors.toList()));
			}
			return Collections.EMPTY_LIST;
		}
	}

	public static class XlsxSpanExtractor {
		private int _first;
		private int _last;

		public XlsxSpanExtractor(String jsonMap) {
			String[] args = jsonMap.split(":");
			_first = Integer.parseInt(args[0]);
			_last = Integer.parseInt(args[args.length - 1]);
		}

		public int getFirst() {
			return _first;
		}

		public int getLast() {
			return _last;
		}
	}

	public static class XlsxCellExtractor {
		final Cell jsonMap;
		final Map sharedStrings;
		final int index;
		final Map calcChain;

		public XlsxCellExtractor(Cell jsonMap, Map sharedStrings,
				Map calcChain) {
			this.jsonMap = jsonMap;
			this.sharedStrings = sharedStrings;
			this.calcChain = calcChain;
			this.index = Converter.a1ToRef((String) jsonMap.get("r")).getLeft();
		}

		public Number getCm() {
			Object cm = jsonMap.get("cm");
			return cm != null ? (int) Double.parseDouble((String) cm) : null;
		}

		public Boolean isPhonetic() {
			return toBoolean(jsonMap.get("ph"), null);
		}

		public String getRef() {
			return (String) jsonMap.get("r");
		}

		public int getStyle() {
			Object s = jsonMap.get("s");
			return s != null ? Integer.parseInt((String) s) : -1;
		}

		public String getType() {
			return (String) jsonMap.get("t");
		}

		public Number getVm() {
			Object cm = jsonMap.get("vm");
			return cm != null ? (int) Double.parseDouble((String) cm) : null;
		}

		public XlsxFormulaExtractor getFormula() {
			Object f = jsonMap.get("f");
			return f != null ?
					new XlsxFormulaExtractor((String) (jsonMap.get("r")),
							(Map) f, calcChain) :
					null;
		}

		public XlsxRichTextExtractor getRichTextInline() {
			Object is = jsonMap.get("is");
			return is != null ? new XlsxRichTextExtractor((Map) (is)) : null;
		}

		private List _sharedStrings;

		public Object getValue() {
			Object v = jsonMap.get("v");
			if (v != null) {
				String t = getType();
				if (t != null) {
					if ("s".equals(t)) {
						if (_sharedStrings == null) {
							_sharedStrings = toList(sharedStrings, "si");
						}
						try {
							int n = Integer.parseInt((String) v);
							Map si = (Map) _sharedStrings.get(n);
							if (si.containsKey("r") || si.containsKey("rPh")) {
								return new XlsxRichTextExtractor(si);
							} else {
								return toValue(si, "t");
							}
						} catch (NumberFormatException e) {
							return v;
						}
					} else {
						//Todo
						return v;
					}
				} else {
					return v;
				}
			}
			return null;
		}
	}

	public static class XlsxFormulaExtractor {
		final String cref;
		final private Map jsonMap;
		final Map calcChain;

		public XlsxFormulaExtractor(String cref, Map jsonMap, Map calcChain) {
			this.cref = cref;
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			this.calcChain = calcChain;
		}

		public Boolean isAca() {
			return toBool(jsonMap, "aca");
		}

		public Boolean isBx() {
			return toBool(jsonMap, "bx");
		}

		public Boolean isCa() {
			return toBool(jsonMap, "ca");
		}

		public Boolean isDel1() {
			return toBool(jsonMap, "del1");
		}

		public Boolean isDel2() {
			return toBool(jsonMap, "del2");
		}

		public Boolean isDt2D() {
			return toBool(jsonMap, "dt2D");
		}

		public Boolean isDtr() {
			return toBool(jsonMap, "dtr");
		}

		public String getR1() {
			return toStringValue(jsonMap, "r1");
		}

		public String getR2() {
			return toStringValue(jsonMap, "r2");
		}

		public String getRef() {
			return toStringValue(jsonMap, "ref");
		}

		public Number getSi() {
			return toNum(jsonMap, "si");
		}

		public String getType() {
			return toStringValue(jsonMap, "t");
		}

		public Object getValue() {
			Object v = toValue(jsonMap, "_");
			String type = getType();
			if ("shared".equals(type)) {
				String ref0 = getRef();
				Number si = getSi();
				if (ref0 != null)
					calcChain.put(si, new Pair<>(cref, v));
				else
					return calcChain.get(si);
			}
			return (String) v;
		}

		public String getSharedRef() {
			if ("shared".equals(getType()) && getRef() == null) {
				return ((Pair<String, String>) calcChain.get(cref)).y;
			}
			return null;
		}
	}

	public static class XlsxRichTextExtractor {
		final private Map jsonMap;

		public XlsxRichTextExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getBaseText() {
			return toStringValue(jsonMap, "t", "");
		}

		public List<XlsxREItExtractor> getRuns() {
			Object rs = jsonMap.get("r");
			if (rs != null) {
				if (rs instanceof List) {
					return (List<XlsxREItExtractor>) ((List) rs).stream()
							.map((r) -> new XlsxREItExtractor((Map) (r)))
							.collect(Collectors.toList());
				}
				return asList(new XlsxREItExtractor((Map) (rs)));
			}
			return Collections.EMPTY_LIST;
		}

		public List<XlsxPhoneticRunExtractor> getPhoneticRuns() {
			Object rs = jsonMap.get("rPh");
			if (rs != null) {
				if (rs instanceof List) {
					return (List<XlsxPhoneticRunExtractor>) ((List) rs).stream()
							.map((r) -> new XlsxPhoneticRunExtractor((Map) (r)))
							.collect(Collectors.toList());
				}
				return asList(new XlsxPhoneticRunExtractor((Map) (rs)));
			}
			return Collections.EMPTY_LIST;
		}

		public XlsxPhoneticPrExtractor getPhoneticPr() {
			return jsonMap.containsKey("phoneticPr") ?
					new XlsxPhoneticPrExtractor(
							(Map) (jsonMap.get("phoneticPr"))) :
					null;
		}
	}

	public static class XlsxREItExtractor {
		final private Map jsonMap;

		public XlsxREItExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxFontExtractor getFont() {
			// Issue1188Test: rPr with null value
			Map rPr = (Map) jsonMap.get("rPr");
			return rPr != null ?
					new XlsxFontExtractor(rPr) :
					null;
		}

		public String getText() {
			return toStringValue(jsonMap, "t", "");
		}
	}

	public static class XlsxPhoneticRunExtractor {
		final private Map jsonMap;

		public XlsxPhoneticRunExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getStartIndex() {
			return toInt(jsonMap, "sb");
		}

		public String getText() {
			return toStringValue(jsonMap, "t");
		}

		public int getEndIndex() {
			return toInt(jsonMap, "eb");
		}
	}

	public static class XlsxPhoneticPrExtractor {
		final private Map jsonMap;

		public XlsxPhoneticPrExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getFontId() {
			return toInt(jsonMap, "fontId");
		}

		public String getAlignment() {
			return toStringValue(jsonMap, "alignment");
		}

		public String getPhoneticType() {
			return toStringValue(jsonMap, "type");
		}
	}

	public static class XlsxThemesExtractor {

		final String name;

		final String os;

		final private Map jsonMap;

		XlsxThemesExtractor(String os, Map jsonMap) {
			this.os = os;
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			name = toStringValue(jsonMap, "name");
		}

		public XlsxClrSchemeExtractor getClrScheme() {
			return new XlsxClrSchemeExtractor(
					$(jsonMap, "a:themeElements", "a:clrScheme"));
		}
	}

	public static class XlsxClrSchemeExtractor {
		final private Map jsonMap;

		public XlsxClrSchemeExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		private static String getColorScheme(Map jsonMap, String key) {
			final Map color = $(jsonMap, key);
			if (color.containsKey("a:srgbClr")) {
				return $(color, "a:srgbClr", "val");
			} else if (color.containsKey("a:sysClr")) {
				return $(color, "a:sysClr", "lastClr");
			} else if (color.containsKey("a:schemeClr")) {
				return $(color, "a:schemeClr", "val");
			} else if (color.containsKey("a:prstClr")) {
				return $(color, "a:prstClr", "val");
			} else {
				// TODO: for a:scrgbClr, a:hslClr
				throw new UnsupportedOperationException("Not implemented yet.");
			}
		}

		public String getName() {
			return $(jsonMap, "name");
		}

		public String getAccent1() {
			return getColorScheme(jsonMap, "a:accent1");
		}

		public String getAccent2() {
			return getColorScheme(jsonMap, "a:accent2");
		}

		public String getAccent3() {
			return getColorScheme(jsonMap, "a:accent3");
		}

		public String getAccent4() {
			return getColorScheme(jsonMap, "a:accent4");
		}

		public String getAccent5() {
			return getColorScheme(jsonMap, "a:accent5");
		}

		public String getAccent6() {
			return getColorScheme(jsonMap, "a:accent6");
		}

		public String getDk1() {
			return getColorScheme(jsonMap, "a:dk1");
		}

		public String getTx1() {
			return getColorScheme(jsonMap, "a:dk1");
		}

		public String getDk2() {
			return getColorScheme(jsonMap, "a:dk2");
		}

		public String getTx2() {
			return getColorScheme(jsonMap, "a:dk2");
		}

		public String getFolHlink() {
			return getColorScheme(jsonMap, "a:folHlink");
		}

		public String getHlink() {
			return getColorScheme(jsonMap, "a:hlink");
		}

		public String getLt1() {
			return getColorScheme(jsonMap, "a:lt1");
		}

		public String getBg1() {
			return getColorScheme(jsonMap, "a:lt1");
		}

		public String getLt2() {
			return getColorScheme(jsonMap, "a:lt2");
		}

		public String getBg2() {
			return getColorScheme(jsonMap, "a:lt2");
		}

	}

	public static class XlsxExternalLinkExtractor {
		final private Map workbookJson;
		final private Map jsonMap;
		final private Map extBookJson;
		private String _extLinkPath;

		public XlsxExternalLinkExtractor(Map workbookJson, Map jsonMap) {
			this.workbookJson = workbookJson;
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			Map extBookJson = $(jsonMap, "externalLink", "externalBook");
			this.extBookJson = (extBookJson == null) ?
					Collections.EMPTY_MAP :
					extBookJson;
		}

		public String getExternalLinkPath() {
			if (_extLinkPath == null) {
				List<Map> workbookRels = $(workbookJson,
						"xl/_rels/workbook.xml.rels", "Relationships",
						"Relationship");
				Map<String, String> idMap = new LinkedHashMap<>(); // rid -> sheetPath
				workbookRels.forEach((rel) -> {
					idMap.put((String) rel.get("Id"),
							(String) rel.get("Target"));
				});

				String rid = (String) this.jsonMap.get("r:id");
				String[] array = idMap.get(rid).split("/");
				_extLinkPath = array[array.length - 1];
			}
			return _extLinkPath;
		}

		public XlsxExternalBookExtractor getExternalBook() {
			final String extLinkPath = getExternalLinkPath();
			final Map extLinkRels = $(this.workbookJson,
					"xl/externalLinks/_rels/" + extLinkPath + ".rels");
			Map extBookNames;
			if (extLinkRels != null) {
				final Object relations = $(extLinkRels, "Relationships",
						"Relationship");
				if (relations instanceof List) {
					extBookNames = ((List<Map>) relations).stream()
							.filter((r) -> {
								Object type = r.get("Type");
								return "http://schemas.openxmlformats.org/officeDocument/2006/relationships/externalLinkPath"
										.equals(type)
										|| "http://schemas.microsoft.com/office/2006/relationships/xlExternalLinkPath/xlPathMissing"
										.equals(type);
							}).collect(Collectors.toMap((Map r) -> r.get("Id"),
									(Map r) -> r.get("Target")));
				} else {
					final Map relationsMap = (Map) relations;
					final Object type = relationsMap.get("Type");
					if ("http://schemas.openxmlformats.org/officeDocument/2006/relationships/externalLinkPath"
							.equals(type)
							|| "http://schemas.microsoft.com/office/2006/relationships/xlExternalLinkPath/xlPathMissing"
							.equals(type)) {
						extBookNames = asMap(relationsMap.get("Id"),
								relationsMap.get("Target"));
					} else {
						extBookNames = null;
					}
				}
				String extBookName = (String) extBookNames
						.get(extBookJson.get("r:id"));
				return new XlsxExternalBookExtractor(extBookName, extBookJson);
			}
			return null;
		}
	}

	public static class XlsxExternalBookExtractor {
		final private String name;
		final private Map jsonMap;

		public XlsxExternalBookExtractor(String name, Map jsonMap) {
			this.name = name;
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getName() {
			return name;
		}
	}

	public static class XlsxColExtractor {
		final private Map jsonMap;

		public XlsxColExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isBestFit() {
			return toBool(jsonMap, "bestFit", false);
		}

		public boolean isCollapsed() {
			return toBool(jsonMap, "collapsed", false);
		}

		public boolean isCustomWidth() {
			return toBool(jsonMap, "customWidth", false);
		}

		public boolean isHidden() {
			return toBool(jsonMap, "hidden", false);
		}

		public int getMax() {
			return Optional.ofNullable(toInt(jsonMap, "max")).orElse(0);
		}

		public int getMin() {
			return Optional.ofNullable(toInt(jsonMap, "min")).orElse(0);
		}

		public Number getOutlineLevel() {
			return toNum(jsonMap, "outlineLevel");
		}

		public boolean isPhonetic() {
			return toBool(jsonMap, "phonetic", false);
		}

		public int getStyle() {
			return Optional.ofNullable(toInt(jsonMap, "style")).orElse(-1);
		}

		public Number getWidth() {
			return toNum(jsonMap, "width");
		}
	}

	public static class XlsxHyperlinkExtractor {
		final private Map jsonMap;

		public XlsxHyperlinkExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getRef() {
			return $(jsonMap, "ref");
		}

		public String getTarget() {
			return $(jsonMap,
					"r:id"); // already convert the value of r:id to the value of target.
		}

		public String getLocation() {
			return Optional.ofNullable(Optional.ofNullable($(jsonMap, "location"))
					.orElse(getTarget())).orElse("").toString();
		}

		public String getTooltip() {
			return $(jsonMap, "tooltip");
		}

		public String getDisplay() {
			return $(jsonMap, "display");
		}

		public SHyperlink.HyperlinkType getType() {
			String link = getTarget();
			if (link == null || link.isEmpty()) {
				return SHyperlink.HyperlinkType.DOCUMENT;
			} else {
				link = link.toLowerCase();
				if (link.startsWith("http://") || link.startsWith("https://")
						|| link.startsWith("ftp://")) {
					return SHyperlink.HyperlinkType.URL;
				} else if (link.startsWith("mailto:")) {
					return SHyperlink.HyperlinkType.EMAIL;
				} else {
					return SHyperlink.HyperlinkType.FILE;
				}
			}
		}
	}

	public static class XlsxCommentsExtractor {
		final private Map jsonMap;

		public XlsxCommentsExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<String> getAuthors() {
			Object author = $(jsonMap, "authors", "author");
			if (author instanceof List) {
				return ((List<Map>) author).stream()
						.map((r) -> toStringValue(r, "_", ""))
						.collect(Collectors.toList());
			} else if (author instanceof Map) {
				return asList((String) ((Map) author).get("_"));
			}
			return Collections.EMPTY_LIST;
		}

		public List<XlsxCommentExtractor> getCommentList() {
			return  toList($(jsonMap, "commentList"), "comment", XlsxCommentExtractor.class);
		}
	}

	public static class XlsxCommentExtractor {
		final private Map jsonMap;
		private String authorName;
		private VmlDrawing.VShapeJson vShape;

		public XlsxCommentExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getAuthorId() {
			return toNum(this.jsonMap, "authorId");
		}

		public XlsxRichTextExtractor getText() {
			return new XlsxRichTextExtractor($(jsonMap, "text"));
		}

		public Number getShapeId() {
			return toNum(this.jsonMap, "shapeId");
		}

		public String getRef() {
			return $(jsonMap, "ref");
		}

		public void setAuthorName(String authorName) {
			this.authorName = authorName;
		}
		public String getAuthorName() {
			return authorName;
		}

		public void setShape(VmlDrawing.VShapeJson shape) {
			this.vShape = shape;
		}
		public boolean isVisible() {
			return vShape != null ? vShape.isVisibility() : false;
		}
	}

	public static class XlsxDrawingsExtractor {
		final private Map workbookJson;
		final private Map drawingRefs;
		final private Map jsonMap;

		public XlsxDrawingsExtractor(Map workbookJson, Map drawingRefs, Map jsonMap) {
			this.workbookJson = workbookJson;
			this.drawingRefs = drawingRefs;
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<ShapeJson> getDrawingsJson() {
			final List<ShapeJson> result =
					((List<Map>) toList(jsonMap, "xdr:twoCellAnchor")).stream()
					.filter((data) -> !(data).isEmpty())
					.map((data) -> XlsxDrawingFactory.create(workbookJson, drawingRefs, data))
					.collect(Collectors.toList());
			return result;
		}
	}

	public static abstract class XlsxDrawingFactory {
		private static ChartSpaceJson createChart(Map workbookJson, Map rels, Map viewAnchor, Map data) {
			final List<Map> relList = ((List<Map>) toList($(rels, "Relationships"), "Relationship"));
			final String cId = data.get("r:id").toString();
			final Map relation = relList.stream().filter((entry) -> cId.equals(entry.get("Id").toString()))
					.findAny().orElse(null);
			if (relation != null) {
				final String relationTarget = relation.get("Target").toString();
				final String chartPath = _getPath(relationTarget, "xl");
				final String chartFile = chartPath.substring(chartPath.lastIndexOf("/") + 1);
				final Map chartData = $(workbookJson, chartPath);
				Map chartColorStyle = null;
				Map chartStyle = null;
				final List<Map> chartRef = $x($(workbookJson, "xl/charts/_rels/" + chartFile + ".rels"),
						"Relationships/Relationship");
				if (chartRef != null) {
					final String chartColorStyleTarget = (String) chartRef.stream()
							.filter((rel) -> "http://schemas.microsoft.com/office/2011/relationships/chartColorStyle"
									.equals(rel.get("Type").toString())).findAny().get().get("Target");
					final String chartStyleTarget = (String) chartRef.stream()
							.filter((rel) -> "http://schemas.microsoft.com/office/2011/relationships/chartStyle"
									.equals(rel.get("Type").toString())).findAny().get().get("Target");
					chartColorStyle = $(workbookJson, "xl", "charts",
							chartColorStyleTarget);
					chartStyle = $(workbookJson, "xl", "charts",
							chartStyleTarget);
				}
				final String uuid = $x(viewAnchor,
						"xdr:graphicFrame/xdr:nvGraphicFramePr/xdr:cNvPr/id")
						.toString();
				return new ChartSpaceJson(uuid, viewAnchor, chartData, chartColorStyle, chartStyle);
			}
			return null;
		}

		private static PictureJson createPicture(Map workbookJson, Map rels, Map viewAnchor, Map data) {
			final List<Map> relList = ((List<Map>) toList($(rels, "Relationships"), "Relationship"));
			final Map hlinkClick = $x(viewAnchor, "xdr:grpSp/xdr:nvGrpSpPr/xdr:cNvPr/a:hlinkClick");
			String hlinkTarget = "";
			String hlinkTooltip = "";
			if (hlinkClick != null) {
				final String rId = hlinkClick.get("r:id").toString();
				final Map relation = relList.stream().filter((entry) -> rId.equals(entry.get("Id").toString())).findAny().get();
				if (relation.containsKey("Target")) {
					hlinkTarget = relation.get("Target").toString();
				}
				if (relation.containsKey("tooltip")) {
					hlinkTooltip = hlinkClick.get("tooltip").toString();
				}
			}
			final Map relation = relList.stream().filter((entry) -> {
				final String id = (String) entry.get("Id");
				if (id == null) {
					return false;
				}
				final Map blip = $x(data, "xdr:blipFill/a:blip");
				final String embed = (String) blip.get("r:embed");
				return id.equals(embed != null ? embed : (String) blip.get("r:link"));
			}).findAny().orElse(null);
			if (relation != null) {
				final String uuid = $(data, "xdr:nvPicPr", "xdr:cNvPr", "id").toString();
				final String target = relation.get("Target").toString();
				final String targetMode = relation.containsKey("TargetMode") ?
						relation.get("TargetMode").toString() : "";
				if ("External".equals(targetMode)) {
					return new PictureJson(uuid, viewAnchor, data, target, hlinkTarget, hlinkTooltip, true);
				} else {
					final String imagePath = target.substring(3);
					final String imageFile = imagePath.substring(imagePath.lastIndexOf("/") + 1);
					final byte[] fileContent = (byte[]) workbookJson.get("xl/media/" + imageFile);
					final String fileType = imagePath.substring(imagePath.lastIndexOf(".") + 1);
					return new PictureJson(uuid, viewAnchor, data, fileContent, fileType, hlinkTarget, hlinkTooltip);
				}
			}
			return null;
		}

		private static GroupPictureJson createGroupPicture(Map workbookJson, Map rels, Map viewAnchor, List<Map> data) {
			final Map hlinkClick = $x(viewAnchor, "xdr:grpSp/xdr:nvGrpSpPr/xdr:cNvPr/a:hlinkClick");
			String hlinkTarget = "";
			String hlinkTooltip = "";
			if (hlinkClick != null) {
				final String rId = hlinkClick.get("r:id").toString();
				final Map relation = ((List<Map>) toList($(rels, "Relationships"), "Relationship"))
						.stream().filter((entry) -> rId.equals(entry.get("Id").toString())).findAny().get();
				if (relation.containsKey("Target")) {
					hlinkTarget = relation.get("Target").toString();
				}
				if (relation.containsKey("tooltip")) {
					hlinkTooltip = hlinkClick.get("tooltip").toString();
				}
			}
			final String uuid = $x(viewAnchor, "xdr:grpSp/xdr:nvGrpSpPr/xdr:cNvPr/id").toString();
			final GroupPictureJson gpicJson = new GroupPictureJson(viewAnchor, uuid, hlinkTarget, hlinkTooltip);
			data.forEach((pic) ->
				gpicJson.addPicture(createPicture(workbookJson, rels, viewAnchor, pic)));
			return gpicJson;
		}

		public static ShapeJson create(Map workbookJson, Map rels, Map data) {
			final Map cchartData = $x(data, "xdr:graphicFrame/a:graphic/a:graphicData/c:chart");
			if (cchartData != null) {
				return createChart(workbookJson, rels, data, cchartData);
			}
			final Map chartData = $x(data, "xdr:graphicFrame/a:graphic/a:graphicData/chart");
			if (chartData != null) {
				return createChart(workbookJson, rels, data, chartData);
			}
			final Map picData = $x(data, "xdr:pic");
			if (picData != null) {
				return createPicture(workbookJson, rels, data, picData);
			}
			final List<Map> grpSpPicData = toList($x(data, "xdr:grpSp"), "xdr:pic");
			if (grpSpPicData != null) {
				return createGroupPicture(workbookJson, rels, data, grpSpPicData);
			}
			return null;
		}
	}

	public static class XlsxVmlDrawingExtractor {
		final private Map jsonMap;
		final private byte[] imageData;

		public XlsxVmlDrawingExtractor(Map jsonMap, byte[] imageData) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			this.imageData = imageData;
		}

		public Map getJsonMap() {
			return jsonMap;
		}

		public byte[] getImageData() {
			return imageData;
		}
	}

	public static  class XlsxDefinedNameExtractor {

		final private Map jsonMap;

		public XlsxDefinedNameExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getComment() {
			return toStringValue(jsonMap, "comment");
		}

		public String getCustomMenu() {
			return toStringValue(jsonMap, "comment");
		}

		public String getDescription() {
			return toStringValue(jsonMap, "description");
		}

		public String getFormula() {
			return toStringValue(jsonMap, "_");
		}

		public boolean isFunction() {
			return toBool(jsonMap, "function", false);
		}

		public Number getFunctionGroupId() {
			return toNum(jsonMap, "functionGroupId");
		}

		public String getHelp() {
			return toStringValue(jsonMap, "help");
		}

		public Boolean isHidden() {
			return toBool(jsonMap, "hidden");
		}

		public String getLocalSheetId() {
			return toStringValue(jsonMap, "localSheetId");
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public boolean isPublishToServer() {
			return toBool(jsonMap, "publishToServer", false);
		}

		public String getShortcutKey() {
			return toStringValue(jsonMap, "shortcutKey");
		}

		public String getStatusBar() {
			return toStringValue(jsonMap, "statusBar");
		}

		public boolean isVbProcedure() {
			return toBool(jsonMap, "vbProcedure", false);
		}

		public boolean isWorkbookParameter() {
			return toBool(jsonMap, "workbookParameter", false);
		}

		public boolean isXlm() {
			return toBool(jsonMap, "xlm");
		}
	}
	public static class XlsxDataValidationsExtractor {

		final private Map jsonMap;

		public XlsxDataValidationsExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getCount() {
			return Optional.ofNullable(toInt(jsonMap, "count")).orElse(0);
		}

		public boolean isDisablePrompts() {
			return toBool(jsonMap, "disablePrompts", false);
		}

		public List<XlsxDataValidationExtractor> getValidations() {
			Object validations = jsonMap.get("dataValidation");
			if (validations instanceof List) {
				return ((List<Map>)validations).stream().map((r) -> new XlsxDataValidationExtractor((r))).collect(
						Collectors.toList());
			} else if (validations instanceof Map) {
				return asList(new XlsxDataValidationExtractor((Map) validations));
			} else { // empty rows
				return Collections.EMPTY_LIST;
			}
		}

		public int getXWindow() {
			return Optional.ofNullable(toInt(jsonMap, "xWindow")).orElse(0);
		}

		public int getYWindow() {
			return Optional.ofNullable(toInt(jsonMap, "yWindow")).orElse(0);
		}
	}

	//KEIKAI-464
	public static class XlsxExtDataValidationExtractor extends XlsxDataValidationExtractor {

		public XlsxExtDataValidationExtractor(Map jsonMap) {
			super(jsonMap);
		}

		public String getSqref() {
			return toStringValue(jsonMap, "xm:sqref");
		}

		public String getFormula1() {
			if (jsonMap.containsKey("x14:formula1")) {
				return toStringValue(((Map)jsonMap.get("x14:formula1")), "xm:f");
			}
			return null;
		}

		public String getFormula2() {
			if (jsonMap.containsKey("x14:formula2")) {
				return toStringValue(((Map)jsonMap.get("x14:formula2")), "xm:f");
			}
			return null;
		}
	}
		public static class XlsxDataValidationExtractor {
			final protected Map jsonMap;

			public XlsxDataValidationExtractor(Map jsonMap) {
				this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			}

			public boolean isAllowBlank() {
				return toBool(jsonMap, "allowBlank", false);
			}

			public String getError() {
				return toStringValue(jsonMap, "error");
			}

			public AlertStyle getErrorStyle() {
				final String style = toStringValue(jsonMap, "errorStyle", "stop");
				switch (style) {
				case "information":
					return AlertStyle.INFO;
				case "warning":
					return AlertStyle.WARNING;
				case "stop":
				default:
					return AlertStyle.STOP;
				}
			}

			public String getErrorTitle() {
				return toStringValue(jsonMap, "errorTitle");
			}

			public String getImeMode() {
				return toStringValue(jsonMap, "imeMode");
			}

			public OperatorType getOperator() {
				final String operator = toStringValue(jsonMap, "operator", "between");
				switch(operator) {
				case "equal":
					return OperatorType.EQUAL;
				case "greaterThanOrEqual":
					return OperatorType.GREATER_OR_EQUAL;
				case "greaterThan":
					return OperatorType.GREATER_THAN;
				case "lessThanOrEqual":
					return OperatorType.LESS_OR_EQUAL;
				case "lessThan":
					return OperatorType.LESS_THAN;
				case "notBetween":
					return OperatorType.NOT_BETWEEN;
				case "notEqual":
					return OperatorType.NOT_EQUAL;
				case "between":
				default:
					return OperatorType.BETWEEN;
				}
			}

			public String getPrompt() {
				return toStringValue(jsonMap, "prompt");
			}

			public String getPromptTitle() {
				return toStringValue(jsonMap, "promptTitle");
			}

			public boolean isShowDropDown() {
				// Excel will use 0 as show and 1 as hidden for showDropDown (same logic in workbook_exporter).
				return jsonMap.containsKey("showDropDown") ? (!toBool(jsonMap,"showDropDown")) : true;
			}

			public boolean isShowErrorMessage() {
				return toBool(jsonMap, "showErrorMessage", false);
			}

			public boolean isShowInputMessage() {
				return toBool(jsonMap, "showInputMessage", false);
			}

			public String getSqref() {
				return toStringValue(jsonMap, "sqref");
			}

			public ValidationType getType() {
				final String type = toStringValue(jsonMap, "type", "none");
				switch(type) {
				case "time":
					return ValidationType.TIME;
				case "textLength":
					return ValidationType.TEXT_LENGTH;
				case "list":
					return ValidationType.LIST;
				case "whole":
					return ValidationType.INTEGER;
				case "custom":
					return ValidationType.CUSTOM;
				case "decimal":
					return ValidationType.DECIMAL;
				case "date":
					return ValidationType.DATE;
				case "none":
				default:
					return ValidationType.ANY;
				}
			}

			public String getFormula1() {
				return toStringValue(jsonMap, "formula1");
			}

			public String getFormula2() {
				return toStringValue(jsonMap, "formula2");
			}

			public String getValidator() {
				return toStringValue(jsonMap, "validator");
			}
		}


	public static class XlsxAutoFilterExtractor {

		final private Map jsonMap;
		public XlsxAutoFilterExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxFilterColumnExtractor> getFilterColumns() {
			Object filterColumns = $(jsonMap, "filterColumn");
			if (filterColumns instanceof List) {
				return ((List<Map>)filterColumns).stream().map((r) -> new XlsxFilterColumnExtractor(
						r)).collect(Collectors.toList());
			} else if (filterColumns instanceof Map) {
				return asList(new XlsxFilterColumnExtractor((Map)(filterColumns)));
			}
			return Collections.EMPTY_LIST;
		}


		public String getRef() { return toStringValue(jsonMap, "ref");}

		public XlsxSortStateExtractor getSortState() {
			return jsonMap.containsKey("sortState") ?
					new XlsxSortStateExtractor((Map)(jsonMap.get("sortState"))) : null;
		}
	}

	public static class XlsxFiltersExtractor {
		final private Map jsonMap;

		public XlsxFiltersExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isBlank() {
			return toBool(jsonMap, "blank", false);
		}

		public String getCalendarType() {
			return toStringValue(jsonMap, "calendarType");
		}

		public List<XlsxDateGroupItemExtractor> getDateGroupItems() {
			Object dateGroupItems = jsonMap.get("dateGroupItem");
			if (dateGroupItems instanceof List) {
				return ((List<Map>)dateGroupItems).stream().map((r) -> new XlsxDateGroupItemExtractor(r)).collect(
						Collectors.toList());
			} else if (dateGroupItems instanceof Map) {
				return asList(new XlsxDateGroupItemExtractor((Map)dateGroupItems));
			} else { // empty rows
				return Collections.EMPTY_LIST;
			}
		}

		public List<String> getFilters() {
			final Object filters = jsonMap.get("filter");
			if (filters instanceof List) {
				return ((List<Map>)filters).stream().map((r) -> toStringValue(r, "val")).collect(
						Collectors.toList());
			} else if (filters instanceof Map) {
				return asList(toStringValue((Map)filters, "val"));
			} else { // empty rows
				return Collections.EMPTY_LIST;
			}
		}
	}
	public static class XlsxDateGroupItemExtractor {
		final private Map jsonMap;

		public XlsxDateGroupItemExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getDateTimeGrouping() {
			return toStringValue(jsonMap, "dateTimeGrouping");
		}

		public Number getDay() {
			return toNum(jsonMap, "day");
		}

		public Number getHour() {
			return toNum(jsonMap, "hour");
		}

		public Number getMinute() {
			return toNum(jsonMap, "minute");
		}

		public Number getMonth() {
			return toNum(jsonMap, "month");
		}

		public Number getSecond() {
			return toNum(jsonMap, "second");
		}

		public Number getYear() {
			return toNum(jsonMap, "year");
		}
	}
	public static class XlsxTop10Extractor {
		final private Map jsonMap;

		public XlsxTop10Extractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public double getFilterVal() {
			return Optional.ofNullable(toNum(jsonMap, "filterVal")).orElse(0).doubleValue();
		}

		public boolean isPercent() {
			return toBool(jsonMap, "percent", false);
		}

		public boolean isTop() {
			return toBool(jsonMap, "top", false);
		}

		public double getVal() {
			return Optional.ofNullable(toNum(jsonMap, "val")).orElse(0).doubleValue();
		}
	}
	public static class XlsxIconFilterExtractor {
		final private Map jsonMap;

		public XlsxIconFilterExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getIconId() {
			return toNum(jsonMap, "iconId");
		}

		public String getIconSet() {
			return toStringValue(jsonMap, "iconSet");
		}
	}

	public static class XlsxFilterColumnExtractor {
		final private Map jsonMap;

		public XlsxFilterColumnExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getColId() {
			return toNum(jsonMap, "colId");
		}

		public XlsxColorFilterExtractor getColorFilter() {
			if (jsonMap.containsKey("colorFilter")) {
				return new XlsxColorFilterExtractor((Map)(jsonMap.get("colorFilter")))
						;
			}
			return null;
		}

		public XlsxCustomFiltersExtractor getCustomFilters() {
			if (jsonMap.containsKey("customFilters")) {
				return new XlsxCustomFiltersExtractor((Map)(jsonMap.get("customFilters")))
						;
			}
			return null;
		}

		public XlsxDynamicFilterExtractor getDynamicFilter() {
			if (jsonMap.containsKey("dynamicFilter")) {
				return new XlsxDynamicFilterExtractor((Map)(jsonMap.get("dynamicFilter")))
						;
			}
			return null;
		}

		public XlsxFiltersExtractor getFilters() {
			if (jsonMap.containsKey("filters")) {
				return new XlsxFiltersExtractor((Map)(jsonMap.get("filters")))
						;
			}
			return null;
		}

		public Boolean isHiddenButton() {
			return toBool(jsonMap, "hiddenButton");
		}

		public boolean isSetHiddenButton() {
			return jsonMap.containsKey("hiddenButton");
		}

		public XlsxIconFilterExtractor getIconFilter() {
			if (jsonMap.containsKey("iconFilter")) {
				return new XlsxIconFilterExtractor((Map)(jsonMap.get("iconFilter")))
						;
			}
			return null;
		}

		public Boolean isShowButton() {
			return toBool(jsonMap, "showButton");
		}

		public boolean isSetShowButton() {
			return jsonMap.containsKey("showButton");
		}

		public XlsxTop10Extractor getTop10() {
			if (jsonMap.containsKey("top10")) {
				return new XlsxTop10Extractor((Map)(jsonMap.get("top10")));
			}
			return null;
		}
	}

	public static class XlsxColorFilterExtractor {
		final private Map jsonMap;

		public XlsxColorFilterExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Boolean isCellColor() {
			return toBool(jsonMap, "cellColor");
		}

		public Number getDxfId() {
			return toNum(jsonMap, "dxfId");
		}
	}

	public static class XlsxCustomFiltersExtractor {
		final private Map jsonMap;

		public XlsxCustomFiltersExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isAnd() {
			return toBool(jsonMap, "and", false);
		}

		public XlsxCustomFilterExtractor getCustomFilter1() {
			final Object jsonValue = jsonMap.get("customFilter");
			if (jsonValue instanceof List) {
				return new XlsxCustomFilterExtractor((Map)(((List) jsonValue).get(0)));
			} else if (jsonValue instanceof Map) {
				return new XlsxCustomFilterExtractor((Map)(jsonValue));
			} else {
				return null;
			}
		}

		public XlsxCustomFilterExtractor getCustomFilter2() {
			final Object jsonValue = jsonMap.get("customFilter");
			if (jsonValue instanceof List) {
				List jsonList = (List) jsonValue;
				return jsonList.size() == 2 ? new XlsxCustomFilterExtractor((Map)(((List) jsonValue).get(1))) : null;
			}
			return null;
		}
	}

	public static class XlsxCustomFilterExtractor {
		final private Map jsonMap;

		public XlsxCustomFilterExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public SAutoFilter.FilterOp getOperator() {
			final String op = toStringValue(jsonMap, "operator");
			if (op == null) {
				boolean endWith = false;
				boolean beginWith = false;
				final String val = toStringValue(jsonMap, "val");
				if (val.startsWith("*")) {
					endWith = true;
				}
				if (val.endsWith("*")) {
					beginWith = true;
				}
				boolean contains = beginWith && endWith;
				if (contains) {
					return SAutoFilter.FilterOp.contains;
				} else if (beginWith) {
					return SAutoFilter.FilterOp.beginWith;
				} else if (endWith) {
					return SAutoFilter.FilterOp.endWith;
				} else { //equal
					return SAutoFilter.FilterOp.equal;
				}
			}
			return SAutoFilter.FilterOp.valueOf("null0".equals(op) ? "null" : op);
		}

		public String getVal() {
			final String val = toStringValue(jsonMap, "val");
			boolean endWith = false;
			boolean beginWith = false;
			if (val.startsWith("*")) {
				endWith = true;
			}
			if (val.endsWith("*")) {
				beginWith = true;
			}
			boolean contains = beginWith && endWith;

			if (contains) {
				return val.substring(1, val.length()-1);
			} else if (beginWith) {
				return val.substring(0, val.length()-1);
			} else if (endWith) {
				return val.substring(1);
			} else {
				return val;
			}
		}
	}

	public static class XlsxDynamicFilterExtractor {
		final private Map jsonMap;

		public XlsxDynamicFilterExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Number getMaxVal() {
			return toNum(jsonMap, "maxVal");
		}

		public String getType() {
			return toStringValue(jsonMap, "type");
		}

		public Number getVal() {
			return toNum(jsonMap, "val");
		}
	}

	public static class XlsxSortStateExtractor {
		final private Map jsonMap;

		public XlsxSortStateExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isCaseSensitive() {
			return toBool(jsonMap, "caseSensitive");
		}

		public boolean isColumnSort() {
			return toBool(jsonMap, "columnSort");
		}

		public String getRef() {
			return toStringValue(jsonMap, "ref");
		}

		public List<XlsxSortConditionExtractor> getSortConditions() {
			final Object sortConditions = jsonMap.get("sortCondition");
			if (sortConditions instanceof List) {
				return ((List<Map>)sortConditions).stream().map((r) -> new XlsxSortConditionExtractor(r)).collect(
						Collectors.toList());
			} else if (sortConditions instanceof Map) {
				return asList(new XlsxSortConditionExtractor((Map)sortConditions));
			} else { // empty rows
				return Collections.EMPTY_LIST;
			}
		}

		public String getSortMethod() {
			return toStringValue(jsonMap, "sortMethod");
		}
	}

	public static class XlsxSortConditionExtractor {
		final private Map jsonMap;

		public XlsxSortConditionExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getCustomList() {
			return toStringValue(jsonMap, "customList");
		}

		public boolean isDescending() {
			return toBool(jsonMap, "descending");
		}

		public Number getDxfId() {
			return toNum(jsonMap, "dxfId");
		}

		public Number getIconId() {
			return toNum(jsonMap, "iconId");
		}

		public String getIconSet() {
			return toStringValue(jsonMap, "iconSet");
		}

		public String getRef() {
			return toStringValue(jsonMap, "ref");
		}

		public String getSortBy() {
			return toStringValue(jsonMap, "sortBy");
		}
	}
	// KEIKAI-395, <x14:conditionalFormatting> in <extLst><ext><x14:conditionalFormattings>
	public static class XlsxExtConditionalFormattingExtractor extends XlsxConditionalFormattingExtractor {

		public XlsxExtConditionalFormattingExtractor(Map jsonMap) {
			super(jsonMap);
		}

		@Override
		public List<XlsxCfRuleExtractor> getCfRules() {
			return toList(jsonMap, "x14:cfRule", XlsxExtCfRuleExtractor.class);
		}

		@Override
		public String getSqref() {
			return toStringValue(jsonMap, "xm:sqref");
		}
	}
	public static class XlsxConditionalFormattingExtractor {
		final protected Map jsonMap;

		public XlsxConditionalFormattingExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxCfRuleExtractor> getCfRules() {
			return toList(jsonMap, "cfRule", XlsxCfRuleExtractor.class);
		}

		public List<String> getDuplicateList() {
			return jsonMap.containsKey("duplicateList")?(List<String>) jsonMap.get("duplicateList"): Collections.EMPTY_LIST;
		}

		public String getSqref() {
			return toStringValue(jsonMap, "sqref");
		}

		public List<Number> getSortedList() {
			return jsonMap.containsKey("sortedList")?(List<Number>) jsonMap.get("sortedList"): Collections.EMPTY_LIST;
		}
	}
	// KEIKAI-395, <x14:cfRule> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting>
	public static class XlsxExtCfRuleExtractor extends XlsxCfRuleExtractor {
		public XlsxExtCfRuleExtractor(Map jsonMap) {
			super(jsonMap);
		}
		public List<String> getFormulas() {
			return getFormulas0("xm:f");
		}
		public XlsxDxfExtractor getDxf() {
			return jsonMap.containsKey("x14:dxf")
					? new XlsxExtDxfExtractor((Map) jsonMap.get("x14:dxf"))
					: null;
		}

		@Override
		public XlsxColorScaleExtractor getColorScale() {
			return jsonMap.containsKey("x14:colorScale")?new XlsxExtColorScaleExtractor((Map)(jsonMap.get("x14:colorScale"))):null
					;
		}

		@Override
		public XlsxDataBarExtractor getDataBar() {
			return jsonMap.containsKey("x14:dataBar")
					?new XlsxExtDataBarExtractor((Map)(jsonMap.get("x14:dataBar"))):null;
		}

		@Override
		public XlsxIconSetExtractor getIconSet() {
			return jsonMap.containsKey("x14:iconSet")
					? new XlsxExtIconSetExtractor((Map)jsonMap.get("x14:iconSet")):null;
		}

		@Override
		public List<Map> getInfo() {
			return jsonMap.containsKey("x14:info")?(List<Map>)jsonMap.get("x14:info") :Collections.EMPTY_LIST;
		}
		public String getId() {
			return jsonMap.containsKey("id") ? (String) jsonMap.get("id") : null;
		}
	}
	public static class XlsxCfRuleExtractor {
		final protected Map jsonMap;

		public XlsxCfRuleExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<String> getFormulas() {
			return getFormulas0("formula");
		}
		protected List<String> getFormulas0(String key) {
			final Object formula = jsonMap.get(key);
			if (formula instanceof List) {
				return ((List<Map>)formula).stream().map((f)-> toStringValue(f, "_")).collect(
						Collectors.toList());
			} else if (formula instanceof Map) {
				return asList(toStringValue(jsonMap, key));
			}
			return Collections.EMPTY_LIST;
		}

		public SConditionalFormattingRule.RuleType getType() {
			String type = toStringValue(jsonMap, "type");
			switch (type) {
			case "expression": return SConditionalFormattingRule.RuleType.EXPRESSION;
			case "cellIs": return SConditionalFormattingRule.RuleType.CELL_IS;
			case "colorScale": return SConditionalFormattingRule.RuleType.COLOR_SCALE;
			case "dataBar": return SConditionalFormattingRule.RuleType.DATA_BAR;
			case "iconSet": return SConditionalFormattingRule.RuleType.ICON_SET;
			case "top10": return SConditionalFormattingRule.RuleType.TOP_10;
			case "uniqueValues": return SConditionalFormattingRule.RuleType.UNIQUE_VALUES;
			case "duplicateValues": return SConditionalFormattingRule.RuleType.DUPLICATE_VALUES;
			case "containsText": return SConditionalFormattingRule.RuleType.CONTAINS_TEXT;
			case "notContainsText": return SConditionalFormattingRule.RuleType.NOT_CONTAINS_TEXT;
			case "beginsWith": return SConditionalFormattingRule.RuleType.BEGINS_WITH;
			case "endsWith": return SConditionalFormattingRule.RuleType.ENDS_WITH;
			case "containsBlanks": return SConditionalFormattingRule.RuleType.CONTAINS_BLANKS;
			case "notContainsBlanks": return SConditionalFormattingRule.RuleType.NOT_CONTAINS_BLANKS;
			case "containsErrors": return SConditionalFormattingRule.RuleType.CONTAINS_ERRORS;
			case "notContainsErrors": return SConditionalFormattingRule.RuleType.NOT_CONTAINS_ERRORS;
			case "timePeriod": return SConditionalFormattingRule.RuleType.TIME_PERIOD;
			case "aboveAverage": return SConditionalFormattingRule.RuleType.ABOVE_AVERAGE;
			}
			return null;
		}

		public int getDxfId() {
			return toInt(jsonMap, "dxfId");
		}

		public boolean isSetDxfId() {
			return jsonMap.containsKey("dxfId");
		}

		public int getPriority() {
			return toInt(jsonMap, "priority", 0);
		}

		public boolean isStopIfTrue() {
			return toBool(jsonMap, "stopIfTrue");
		}

		public boolean isSetStopIfTrue() {
			return jsonMap.containsKey("stopIfTrue");
		}

		public boolean isAboveAverage() {
			return toBool(jsonMap, "aboveAverage", false);
		}
		public boolean isSetAboveAverage() {
			return jsonMap.containsKey("aboveAverage");
		}

		public boolean isPercent() {
			return toBool(jsonMap, "percent", false);
		}
		public boolean isSetPercent() {
			return jsonMap.containsKey("percent");
		}

		public boolean isBottom() {
			return toBool(jsonMap, "bottom", false);
		}
		public boolean isSetBottom() {
			return jsonMap.containsKey("bottom");
		}

		public SConditionalFormattingRule.RuleOperator getOperator() {
			String operator = toStringValue(jsonMap, "operator");
			switch (operator) {
			case "lessThan": return SConditionalFormattingRule.RuleOperator.LESS_THAN;
			case "lessThanOrEqual": return SConditionalFormattingRule.RuleOperator.LESS_THAN_OR_EQUAL;
			case "equal": return SConditionalFormattingRule.RuleOperator.EQUAL;
			case "notEqual": return SConditionalFormattingRule.RuleOperator.NOT_EQUAL;
			case "greaterThanOrEqual": return SConditionalFormattingRule.RuleOperator.GREATER_THAN_OR_EQUAL;
			case "greaterThan": return SConditionalFormattingRule.RuleOperator.GREATER_THAN;
			case "between": return SConditionalFormattingRule.RuleOperator.BETWEEN;
			case "notBetween": return SConditionalFormattingRule.RuleOperator.NOT_BETWEEN;
			case "containsText": return SConditionalFormattingRule.RuleOperator.CONTAINS_TEXT;
			case "notContains": return SConditionalFormattingRule.RuleOperator.NOT_CONTAINS;
			case "beginsWith": return SConditionalFormattingRule.RuleOperator.BEGINS_WITH;
			case "endsWith": return SConditionalFormattingRule.RuleOperator.ENDS_WITH;
			}
			return null;
		}
		public boolean isSetOperator() {
			return jsonMap.containsKey("operator");
		}

		public String getText() {
			return toStringValue(jsonMap, "text");
		}

		public SConditionalFormattingRule.RuleTimePeriod getTimePeriod() {
			String timePeriod = toStringValue(jsonMap, "timePeriod");
			switch (timePeriod) {
			case "today": return SConditionalFormattingRule.RuleTimePeriod.TODAY;
			case "yesterday": return SConditionalFormattingRule.RuleTimePeriod.YESTERDAY;
			case "Tomorrow": return SConditionalFormattingRule.RuleTimePeriod.TOMORROW;
			case "last7Days": return SConditionalFormattingRule.RuleTimePeriod.LAST_7_DAYS;
			case "thisMonth": return SConditionalFormattingRule.RuleTimePeriod.THIS_MONTH;
			case "lastMonth": return SConditionalFormattingRule.RuleTimePeriod.LAST_MONTH;
			case "nextMonth": return SConditionalFormattingRule.RuleTimePeriod.NEXT_MONTH;
			case "thisWeek": return SConditionalFormattingRule.RuleTimePeriod.THIS_WEEK;
			case "lastWeek": return SConditionalFormattingRule.RuleTimePeriod.LAST_WEEK;
			case "nextWeek": return SConditionalFormattingRule.RuleTimePeriod.NEXT_WEEK;
			}
			return null;
		}

		public int getRank() {
			return toInt(jsonMap, "rank");
		}

		public boolean isSetRank() {
			return jsonMap.containsKey("rank");
		}

		public int getStdDev() {
			return toInt(jsonMap, "stdDev");
		}

		public boolean isSetStdDev() {
			return jsonMap.containsKey("stdDev");
		}

		public boolean isEqualAverage() {
			return toBool(jsonMap, "equalAverage", false);
		}

		public boolean isSetEqualAverage() {
			return jsonMap.containsKey("equalAverage");
		}

		public XlsxColorScaleExtractor getColorScale() {
			return jsonMap.containsKey("colorScale")?new XlsxColorScaleExtractor((Map)(jsonMap.get("colorScale"))):null
					;
		}

		public XlsxDataBarExtractor getDataBar() {
			return jsonMap.containsKey("dataBar")?new XlsxDataBarExtractor((Map)(jsonMap.get("dataBar"))):null
					;
		}

		public XlsxIconSetExtractor getIconSet() {
			return jsonMap.containsKey("iconSet") ? new XlsxIconSetExtractor((Map)jsonMap.get("iconSet")):null
					;
		}

		public List<Map> getInfo() {
			return jsonMap.containsKey("info")?(List<Map>)jsonMap.get("info") :Collections.EMPTY_LIST
					;
		}

		//KEIKAI-469
		public String getExtId() {
			Object value = null;
			if ((value = jsonMap.get("extLst")) != null) {
				if ((value = ((Map)value).get("ext")) != null) {
					return toStringValue(((Map)value), "x14:id");
				}
			}
			return null;
		}
	}

	// KEIKAI-395, <x14:colorScale> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting><x14:cfRule>
	public static class XlsxExtColorScaleExtractor extends XlsxColorScaleExtractor {
		public XlsxExtColorScaleExtractor(Map jsonMap) {
			super(jsonMap);
		}

		public List<XlsxCfvoExtractor> getCfvos() {
			return toList(jsonMap, "x14:cfvo", XlsxExtCfvoExtractor.class);
		}

		public List<XlsxColorExtractor> getColors() {
			return toList(jsonMap, "x14:color", XlsxColorExtractor.class);
		}
	}

	public static class XlsxColorScaleExtractor {
		final protected Map jsonMap;

		public XlsxColorScaleExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public List<XlsxCfvoExtractor> getCfvos() {
			return toList(jsonMap, "cfvo", XlsxCfvoExtractor.class);
		}

		public List<XlsxColorExtractor> getColors() {
			return toList(jsonMap, "color", XlsxColorExtractor.class);
		}
	}

	// KEIKAI-395, <x14:cfvo> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting><x14:cfRule>
	public static class XlsxExtCfvoExtractor extends XlsxCfvoExtractor {
		public XlsxExtCfvoExtractor(Map jsonMap) {
			super(jsonMap);
		}

		@Override
		public boolean isSetVal() {
			return jsonMap.containsKey("xm:f");
		}

		@Override
		public String getVal() {
			return toStringValue(jsonMap, "xm:f");
		}
	}

	public static class XlsxCfvoExtractor {
		final protected Map jsonMap;

		public XlsxCfvoExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public SCFValueObject.CFValueObjectType getType() {
			String type = toStringValue(jsonMap, "type");
			switch (type) {
			case "num": return SCFValueObject.CFValueObjectType.NUM;
			case "percent": return SCFValueObject.CFValueObjectType.PERCENT;
			case "max": return SCFValueObject.CFValueObjectType.MAX;
			case "min": return SCFValueObject.CFValueObjectType.MIN;
			case "formula": return SCFValueObject.CFValueObjectType.FORMULA;
			case "percentile": return SCFValueObject.CFValueObjectType.PERCENTILE;
			case "autoMin": return SCFValueObject.CFValueObjectType.AUTOMIN;
			case "autoMax": return SCFValueObject.CFValueObjectType.AUTOMAX;
			}
			return null;
		}

		public String getVal() {
			return toStringValue(jsonMap, "val");
		}

		public boolean isSetVal() {
			return jsonMap.containsKey("val");
		}

		public boolean isGte() {
			return toBool(jsonMap, "gte", false);
		}
		public boolean isSetGte() {
			return jsonMap.containsKey("gte");
		}

	}

	// KEIKAI-395, <x14:dataBar> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting><x14:cfRule>
	public static class XlsxExtDataBarExtractor extends XlsxDataBarExtractor {
		public XlsxExtDataBarExtractor(Map jsonMap) {
			super(jsonMap);
		}

		public List<XlsxCfvoExtractor> getCfvos() {
			return toList(jsonMap, "x14:cfvo", XlsxExtCfvoExtractor.class);
		}

		public XlsxColorExtractor getBorderColor() {
			return jsonMap.containsKey("x14:borderColor") ?
					new XlsxColorExtractor(
							(Map) (jsonMap.get("x14:borderColor"))) :
					null;
		}

		public XlsxColorExtractor getNegativeBorderColor() {
			return jsonMap.containsKey("x14:negativeBorderColor") ?
					new XlsxColorExtractor(
							(Map) (jsonMap.get("x14:negativeBorderColor"))) :
					null;
		}

		public XlsxColorExtractor getNegativeFillColor() {
			return jsonMap.containsKey("x14:negativeFillColor") ?
					new XlsxColorExtractor(
							(Map) (jsonMap.get("x14:negativeFillColor"))) :
					null;
		}

		public XlsxColorExtractor getAxisColor() {
			return jsonMap.containsKey("x14:axisColor") ?
					new XlsxColorExtractor(
							(Map) (jsonMap.get("x14:axisColor"))) :
					null;
		}

		public boolean isBorder() {
			return toBool(jsonMap, "border", false);
		}

		public boolean isGradient() {
			return toBool(jsonMap, "gradient", true);
		}

		public boolean isNegativeBarColorSameAsPositive() {
			return toBool(jsonMap, "negativeBarColorSameAsPositive", true);
		}

		public boolean isNegativeBarBorderColorSameAsPositive() {
			return toBool(jsonMap, "negativeBarBorderColorSameAsPositive", true);
		}

		public String getAxisPosition() {
			return toStringValue(jsonMap, "axisPosition", "auto");
		}

		public String getDirection() {
			return toStringValue(jsonMap, "direction", "leftToRight");
		}
	}

	public static class XlsxDataBarExtractor {
		final protected Map jsonMap;

		public XlsxDataBarExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public int getMinLength() {
			return toInt(jsonMap, "minLength");
		}

		public boolean isSetMinLength() {
			return jsonMap.containsKey("minLength");
		}

		public int getMaxLength() {
			return toInt(jsonMap, "maxLength");
		}

		public boolean isSetMaxLength() {
			return jsonMap.containsKey("maxLength");
		}

		public boolean isSetShowValue() {
			return jsonMap.containsKey("showValue");
		}

		public boolean isShowValue() {
			return toBool(jsonMap, "showValue", false);
		}

		public List<XlsxCfvoExtractor> getCfvos() {
			return toList(jsonMap, "cfvo", XlsxCfvoExtractor.class);
		}

		public XlsxColorExtractor getColor() {
			return jsonMap.containsKey("color")?new XlsxColorExtractor((Map)(jsonMap.get("color"))):null
					;
		}
	}

	// KEIKAI-395, <x14:iconSet> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting><x14:cfRule>
	public static class XlsxExtIconSetExtractor extends XlsxIconSetExtractor {
		public XlsxExtIconSetExtractor(Map jsonMap) {
			super(jsonMap);
		}

		public List<XlsxCfvoExtractor> getCfvos() {
			return toList(jsonMap, "x14:cfvo", XlsxExtCfvoExtractor.class);
		}

		public List<XlsxCfIconExtractor> getCfIcons() {
			return toList(jsonMap, "x14:cfIcon", XlsxCfIconExtractor.class);
		}
	}

	// KEIKAI-395, <x14:cfIcon> in <extLst><ext><x14:conditionalFormattings><x14:conditionalFormatting><x14:cfRule><x14:iconSet>
	public static class XlsxCfIconExtractor {
		final protected Map jsonMap;

		public XlsxCfIconExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public Integer getIconId() {
			return toInt(jsonMap, "iconId");
		}

		public SIconSet.IconSetType getIconSet() {
			String iconSet = toStringValue(jsonMap, "iconSet", "NoIcons");
			switch (iconSet) {
				case "3Arrows": return SIconSet.IconSetType.X_3_ARROWS;
				case "3ArrowsGray": return SIconSet.IconSetType.X_3_ARROWS_GRAY;
				case "3Flags": return SIconSet.IconSetType.X_3_FLAGS;
				case "3TrafficLights1": return SIconSet.IconSetType.X_3_TRAFFIC_LIGHTS_1;
				case "3TrafficLights2": return SIconSet.IconSetType.X_3_TRAFFIC_LIGHTS_2;
				case "3Signs": return SIconSet.IconSetType.X_3_SIGNS;
				case "3Symbols": return SIconSet.IconSetType.X_3_SYMBOLS;
				case "3Symbols2": return SIconSet.IconSetType.X_3_SYMBOLS_2;
				case "4Arrows": return SIconSet.IconSetType.X_4_ARROWS;
				case "4ArrowsGray": return SIconSet.IconSetType.X_4_ARROWS_GRAY;
				case "4RedToBlack": return SIconSet.IconSetType.X_4_RED_TO_BLACK;
				case "4Rating": return SIconSet.IconSetType.X_4_RATING;
				case "4TrafficLights": return SIconSet.IconSetType.X_4_TRAFFIC_LIGHTS;
				case "5Arrows": return SIconSet.IconSetType.X_5_ARROWS;
				case "5ArrowsGray": return SIconSet.IconSetType.X_5_ARROWS_GRAY;
				case "5Rating": return SIconSet.IconSetType.X_5_RATING;
				case "5Quarters": return SIconSet.IconSetType.X_5_QUARTERS;
			}
			return null;
		}
	}

	public static class XlsxIconSetExtractor {
		final protected Map jsonMap;

		public XlsxIconSetExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public SIconSet.IconSetType getIconSet() {
			String iconSet = toStringValue(jsonMap, "iconSet", "3TrafficLights1");
			switch (iconSet) {
			case "3Arrows": return SIconSet.IconSetType.X_3_ARROWS;
			case "3ArrowsGray": return SIconSet.IconSetType.X_3_ARROWS_GRAY;
			case "3Flags": return SIconSet.IconSetType.X_3_FLAGS;
			case "3TrafficLights1": return SIconSet.IconSetType.X_3_TRAFFIC_LIGHTS_1;
			case "3TrafficLights2": return SIconSet.IconSetType.X_3_TRAFFIC_LIGHTS_2;
			case "3Signs": return SIconSet.IconSetType.X_3_SIGNS;
			case "3Symbols": return SIconSet.IconSetType.X_3_SYMBOLS;
			case "3Symbols2": return SIconSet.IconSetType.X_3_SYMBOLS_2;
			case "4Arrows": return SIconSet.IconSetType.X_4_ARROWS;
			case "4ArrowsGray": return SIconSet.IconSetType.X_4_ARROWS_GRAY;
			case "4RedToBlack": return SIconSet.IconSetType.X_4_RED_TO_BLACK;
			case "4Rating": return SIconSet.IconSetType.X_4_RATING;
			case "4TrafficLights": return SIconSet.IconSetType.X_4_TRAFFIC_LIGHTS;
			case "5Arrows": return SIconSet.IconSetType.X_5_ARROWS;
			case "5ArrowsGray": return SIconSet.IconSetType.X_5_ARROWS_GRAY;
			case "5Rating": return SIconSet.IconSetType.X_5_RATING;
			case "5Quarters": return SIconSet.IconSetType.X_5_QUARTERS;
			}
			return null;
		}
		public boolean isSetPercent() {
			return jsonMap.containsKey("percent");
		}
		public boolean isPercent() {
			return toBool(jsonMap, "percent", false);
		}
		public boolean isSetReverse() {
			return jsonMap.containsKey("reverse");
		}

		public boolean isReverse() {
			return toBool(jsonMap, "reverse", false);
		}
		public boolean isSetShowValue() {
			return jsonMap.containsKey("showValue");
		}

		public boolean isShowValue() {
			return toBool(jsonMap, "showValue", false);
		}

		public List<XlsxCfvoExtractor> getCfvos() {
			return toList(jsonMap, "cfvo", XlsxCfvoExtractor.class);
		}

		// KEIKAI-448
		public List<XlsxCfIconExtractor> getCfIcons() {
			return null;
		}
	}

	public static class XlsxHeaderFootersExtractor {
		final private Map jsonMap;

		public XlsxHeaderFootersExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isAlignWithMargins() {
			return toBool(jsonMap, "alignWithMargins", false);
		}

		public boolean isDifferentFirst() {
			return toBool(jsonMap, "differentFirst", false);
		}

		public boolean isDifferentOddEven() {
			return toBool(jsonMap, "differentOddEven", false);
		}

		public String getEvenFooter() {
			return toStringValue(jsonMap, "evenFooter");
		}

		public String getEvenHeader() {
			return toStringValue(jsonMap, "evenHeader");
		}

		public String getFirstFooter() {
			return toStringValue(jsonMap, "firstFooter");
		}

		public String getFirstHeader() {
			return toStringValue(jsonMap, "firstHeader");
		}

		public String getOddFooter() {
			return toStringValue(jsonMap, "oddFooter");
		}

		public String getOddHeader() {
			return toStringValue(jsonMap, "oddHeader");
		}

		public boolean isScaleWithDoc() {
			return toBool(jsonMap, "scaleWithDoc", false);
		}
	}

	public static class XlsxPageSetupExtractor {
		final private Map jsonMap;
		/** US Letter 8 1/2 x 11 in */
		public static final short LETTER_PAPERSIZE = 1;
		/** US Letter Small 8 1/2 x 11 in */
		public static final short LETTER_SMALL_PAGESIZE = 2;
		/** US Tabloid 11 x 17 in */
		public static final short TABLOID_PAPERSIZE = 3;
		/** US Ledger 17 x 11 in */
		public static final short LEDGER_PAPERSIZE = 4;
		/** US Legal 8 1/2 x 14 in */
		public static final short LEGAL_PAPERSIZE = 5;
		/** US Statement 5 1/2 x 8 1/2 in */
		public static final short STATEMENT_PAPERSIZE = 6;
		/** US Executive 7 1/4 x 10 1/2 in */
		public static final short EXECUTIVE_PAPERSIZE = 7;
		/** A3 - 297x420 mm */
		public static final short A3_PAPERSIZE = 8;
		/** A4 - 210x297 mm */
		public static final short A4_PAPERSIZE = 9;
		/** A4 Small - 210x297 mm */
		public static final short A4_SMALL_PAPERSIZE = 10;
		/** A5 - 148x210 mm */
		public static final short A5_PAPERSIZE = 11;
		/** B4 (JIS) 250x354 mm */
		public static final short B4_PAPERSIZE = 12;
		/** B5 (JIS) 182x257 mm */
		public static final short B5_PAPERSIZE = 13;
		/** Folio 8 1/2 x 13 in */
		public static final short FOLIO8_PAPERSIZE = 14;
		/** Quarto 215x275 mm */
		public static final short QUARTO_PAPERSIZE = 15;
		/** 10 x 14 in */
		public static final short TEN_BY_FOURTEEN_PAPERSIZE = 16;
		/** 11 x 17 in */
		public static final short ELEVEN_BY_SEVENTEEN_PAPERSIZE = 17;
		/** US Note 8 1/2 x 11 in */
		public static final short NOTE8_PAPERSIZE = 18;
		/** US Envelope #9 3 7/8 x 8 7/8 */
		public static final short ENVELOPE_9_PAPERSIZE = 19;
		/** US Envelope #10 4 1/8 x 9 1/2 */
		public static final short ENVELOPE_10_PAPERSIZE = 20;
		/** Envelope DL 110x220 mm */
		public static final short ENVELOPE_DL_PAPERSIZE = 27;
		/** Envelope C5 162x229 mm */
		public static final short ENVELOPE_CS_PAPERSIZE = 28;
		public static final short ENVELOPE_C5_PAPERSIZE = 28;
		/** Envelope C3 324x458 mm */
		public static final short ENVELOPE_C3_PAPERSIZE = 29;
		/** Envelope C4 229x324 mm */
		public static final short ENVELOPE_C4_PAPERSIZE = 30;
		/** Envelope C6 114x162 mm */
		public static final short ENVELOPE_C6_PAPERSIZE = 31;

		public static final short ENVELOPE_MONARCH_PAPERSIZE = 37;
		/** A4 Extra - 9.27 x 12.69 in */
		public static final short A4_EXTRA_PAPERSIZE = 53;
		/** A4 Transverse - 210x297 mm */
		public static final short A4_TRANSVERSE_PAPERSIZE = 55;
		/** A4 Plus - 210x330 mm */
		public static final short A4_PLUS_PAPERSIZE = 60;
		/** US Letter Rotated 11 x 8 1/2 in */
		public static final short LETTER_ROTATED_PAPERSIZE = 75;
		/** A4 Rotated - 297x210 mm */
		public static final short A4_ROTATED_PAPERSIZE = 77;

		public XlsxPageSetupExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isBlackAndWhite() {
			return toBool(jsonMap, "blackAndWhite");
		}

		public int getCellComments() {
			switch (toStringValue(jsonMap, "cellComments", "none")) {
			case "none":
				return 0;
			case "asDisplayed":
				return 1;
			case "atEnd":
				return 2;
			}
			return 0;
		}

		public Number getCopies() {
			return toNum(jsonMap, "copies");
		}

		public boolean isDraft() {
			return toBool(jsonMap, "draft");
		}

		public int getErrors() {
			String errors = toStringValue(jsonMap, "errors", "displayed");
			switch (errors) {
			case "blank": return 1;
			case "dash": return 2;
			case "NA": return 3;
			default:
				return 0;
			}
		}

		public int getFirstPageNumber() {
			return toInt(jsonMap, "firstPageNumber", 0);
		}

		public int getFitToHeight() {
			return toInt(jsonMap, "fitToHeight", 0);
		}

		public int getFitToWidth() {
			return toInt(jsonMap, "fitToWidth", 0);
		}

		public int getHorizontalDpi() {
			return toInt(jsonMap, "horizontalDpi", 0);
		}

		public String getOrientation() {
			return toStringValue(jsonMap, "orientation");
		}

		public String getPageOrder() {
			return toStringValue(jsonMap, "pageOrder");
		}

		public SPrintSetup.PaperSize getPaperSize() {
			int paperSize = toInt(jsonMap, "paperSize", (int) A4_PAPERSIZE);
			switch (paperSize) {
			case A3_PAPERSIZE:
				return SPrintSetup.PaperSize.A3;
			case A4_EXTRA_PAPERSIZE:
				return SPrintSetup.PaperSize.A4_EXTRA;
			case A4_PAPERSIZE:
				return SPrintSetup.PaperSize.A4;
			case A4_PLUS_PAPERSIZE:
				return SPrintSetup.PaperSize.A4_PLUS;
			case A4_ROTATED_PAPERSIZE:
				return SPrintSetup.PaperSize.A4_ROTATED;
			case A4_SMALL_PAPERSIZE:
				return SPrintSetup.PaperSize.A4_SMALL;
			case A4_TRANSVERSE_PAPERSIZE:
				return SPrintSetup.PaperSize.A4_TRANSVERSE;
			case A5_PAPERSIZE:
				return SPrintSetup.PaperSize.A5;
			case B4_PAPERSIZE:
				return SPrintSetup.PaperSize.B4;
			case B5_PAPERSIZE:
				return SPrintSetup.PaperSize.B5;
			case ELEVEN_BY_SEVENTEEN_PAPERSIZE:
				return SPrintSetup.PaperSize.ELEVEN_BY_SEVENTEEN;
			case ENVELOPE_10_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_10;
			case ENVELOPE_9_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_9;
			case ENVELOPE_C3_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_C3;
			case ENVELOPE_C4_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_C4;
			case ENVELOPE_C5_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_C5;
			case ENVELOPE_C6_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_C6;
			case ENVELOPE_DL_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_DL;
			case ENVELOPE_MONARCH_PAPERSIZE:
				return SPrintSetup.PaperSize.ENVELOPE_MONARCH;
			case EXECUTIVE_PAPERSIZE:
				return SPrintSetup.PaperSize.EXECUTIVE;
			case FOLIO8_PAPERSIZE:
				return SPrintSetup.PaperSize.FOLIO8;
			case LEDGER_PAPERSIZE:
				return SPrintSetup.PaperSize.LEDGER;
			case LETTER_PAPERSIZE:
				return SPrintSetup.PaperSize.LETTER;
			case LETTER_ROTATED_PAPERSIZE:
				return SPrintSetup.PaperSize.LETTER_ROTATED;
			case LETTER_SMALL_PAGESIZE:
				return SPrintSetup.PaperSize.LETTER_SMALL;
			case NOTE8_PAPERSIZE:
				return SPrintSetup.PaperSize.NOTE8;
			case QUARTO_PAPERSIZE:
				return SPrintSetup.PaperSize.QUARTO;
			case STATEMENT_PAPERSIZE:
				return SPrintSetup.PaperSize.STATEMENT;
			case TABLOID_PAPERSIZE:
				return SPrintSetup.PaperSize.TABLOID;
			case TEN_BY_FOURTEEN_PAPERSIZE:
				return SPrintSetup.PaperSize.TEN_BY_FOURTEEN;
			default:
				return SPrintSetup.PaperSize.A4;
			}
		}

		public int getScale() {
			return toInt(jsonMap, "scale", 0);
		}

		public boolean isUseFirstPageNumber() {
			return toBool(jsonMap, "useFirstPageNumber", false);
		}

		public boolean isUsePrinterDefaults() {
			return toBool(jsonMap, "usePrinterDefaults", false);
		}

		public Number getVerticalDpi() {
			return toNum(jsonMap, "verticalDpi");
		}
	}

	public static class XlsxPageMarginsExtractor {
		final private Map jsonMap;

		public XlsxPageMarginsExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public double getBottom() {
			return toDouble(jsonMap, "bottom", 0d);
		}

		public double getFooter() {
			return toDouble(jsonMap, "footer", 0d);
		}

		public double getHeader() {
			return toDouble(jsonMap, "header", 0d);
		}

		public double getLeft() {
			return toDouble(jsonMap, "left", 0d);
		}

		public double getRight() {
			return toDouble(jsonMap, "right", 0d);
		}

		public double getTop() {
			return toDouble(jsonMap, "top", 0d);
		}
	}

	public static class XlsxPrintOptionsExtractor {
		final private Map jsonMap;

		public XlsxPrintOptionsExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isGridLines() {
			return toBool(jsonMap, "gridLines", false);
		}

		public boolean isGridLinesSet() {
			return toBool(jsonMap, "gridLinesSet", false);
		}

		public boolean isHeadings() {
			return toBool(jsonMap, "headings", false);
		}

		public boolean isHorizontalCentered() {
			return toBool(jsonMap, "horizontalCentered", false);
		}

		public boolean isVerticalCentered() {
			return toBool(jsonMap, "verticalCentered", false);
		}

		// KEIKAI-403
		public boolean isNotEmpty() { return jsonMap != null && !jsonMap.isEmpty(); }
	}

	public static class XlsxTableExtractor {
		final private Map jsonMap;
		final private Map _queryTable;

		public XlsxTableExtractor(Map jsonMap, Map queryTable) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
			this._queryTable =  queryTable;
		}

		public XlsxAutoFilterExtractor getAutoFilter() {
			if (jsonMap.containsKey("autoFilter")) {
				return new XlsxAutoFilterExtractor((Map)(jsonMap.get("autoFilter")));
			}
			return null;
		}

		public String getComment() {
			return toStringValue(jsonMap, "comment");
		}

		public Number getConnectionId() {
			return toNum(jsonMap, "connectionId");
		}

		public String getDataCellStyle() {
			return toStringValue(jsonMap, "dataCellStyle");
		}

		public Number getDataDxfId() {
			return toNum(jsonMap, "dataDxfId");
		}

		public String getDisplayName() {
			return toStringValue(jsonMap, "displayName");
		}

		public Number getHeaderRowBorderDxfId() {
			return toNum(jsonMap, "headerRowBorderDxfId");
		}

		public String getHeaderRowCellStyle() {
			return toStringValue(jsonMap, "headerRowCellStyle");
		}

		public int getHeaderRowCount() {
			return toInt(jsonMap, "headerRowCount", 1); // default to 1;
		}

		public Number getHeaderRowDxfId() {
			return toNum(jsonMap, "headerRowDxfId");
		}

		public Number getId() {
			return toNum(jsonMap, "id");
		}

		public boolean isInsertRow() {
			return toBool(jsonMap, "insertRow");
		}

		public boolean isInsertRowShift() {
			return toBool(jsonMap, "insertRowShift");
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public boolean isPublished() {
			return toBool(jsonMap, "published");
		}

		public String getRef() {
			return toStringValue(jsonMap, "ref");
		}

		public XlsxSortStateExtractor getSortState() {
			return jsonMap.containsKey("sortState")?new XlsxSortStateExtractor((Map)jsonMap.get("sortState")):null
					;
		}

		public Number getTableBorderDxfId() {
			return toNum(jsonMap, "tableBorderDxfId");
		}

		public List<XlsxTableColumnExtractor> getTableColumns() {
			return toList($(jsonMap, "tableColumns"), "tableColumn", XlsxTableColumnExtractor.class);
		}

		public XlsxTableStyleInfoExtractor getTableStyleInfo() {
			return jsonMap.containsKey("tableStyleInfo")?new XlsxTableStyleInfoExtractor((Map)jsonMap.get("tableStyleInfo")):null
					;
		}

		public String getTableType() {
			return toStringValue(jsonMap, "tableType");
		}

		public Number getTotalsRowBorderDxfId() {
			return toNum(jsonMap, "totalsRowBorderDxfId");
		}

		public String getTotalsRowCellStyle() {
			return toStringValue(jsonMap, "totalsRowCellStyle");
		}

		public int getTotalsRowCount() {
			return toInt(jsonMap, "totalsRowCount", 0); // default to 0
		}

		public Number getTotalsRowDxfId() {
			return toNum(jsonMap, "totalsRowDxfId");
		}

		public boolean isTotalsRowShown() {
			return toBool(jsonMap, "totalsRowShown", true);
		}

		public Map getQueryTable() {
			return _queryTable;
		}
	}

	public static class XlsxTableStyleInfoExtractor {
		final private Map jsonMap;

		public XlsxTableStyleInfoExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public boolean isShowColumnStripes() {
			return toBool(jsonMap, "showColumnStripes", false);
		}

		public boolean isShowFirstColumn() {
			return toBool(jsonMap, "showFirstColumn", false);
		}

		public boolean isShowLastColumn() {
			return toBool(jsonMap, "showLastColumn", false);
		}

		public boolean isShowRowStripes() {
			return toBool(jsonMap, "showRowStripes", false);
		}
	}

	public static class XlsxTableColumnExtractor {
		final private Map jsonMap;

		public XlsxTableColumnExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public XlsxTableFormulaExtractor getCalculatedColumnFormula() {
			return null;
		}

		public String getDataCellStyle() {
			return toStringValue(jsonMap, "dataCellStyle");
		}

		public Number getDataDxfId() {
			return toNum(jsonMap, "dataDxfId");
		}

		public String getHeaderRowCellStyle() {
			return toStringValue(jsonMap, "headerRowCellStyle");
		}

		public Number getHeaderRowDxfId() {
			return toNum(jsonMap, "headerRowDxfId");
		}

		public Number getId() {
			return toNum(jsonMap, "id");
		}

		public String getName() {
			return toStringValue(jsonMap, "name");
		}

		public Number getQueryTableFieldId() {
			return toNum(jsonMap, "queryTableFieldId");
		}

		public String getTotalsRowCellStyle() {
			return toStringValue(jsonMap, "totalsRowCellStyle");
		}

		public Number getTotalsRowDxfId() {
			return toNum(jsonMap, "totalsRowDxfId");
		}

		public XlsxTableFormulaExtractor getTotalsRowFormula() {
			return jsonMap.containsKey("totalsRowFormula")?new XlsxTableFormulaExtractor((Map)(jsonMap.get("totalsRowFormula"))):null
					;
		}

		public STableColumn.STotalsRowFunction getTotalsRowFunction() {
			String rfun = toStringValue(jsonMap, "totalsRowFunction", "none");
			switch (rfun) {
			case "none": return STableColumn.STotalsRowFunction.none;
			case "sum": return STableColumn.STotalsRowFunction.sum;
			case "min": return STableColumn.STotalsRowFunction.min;
			case "max": return STableColumn.STotalsRowFunction.max;
			case "average": return STableColumn.STotalsRowFunction.average;
			case "count": return STableColumn.STotalsRowFunction.count;
			case "countNums": return STableColumn.STotalsRowFunction.countNums;
			case "stdDev": return STableColumn.STotalsRowFunction.stdDev;
			case "var": return STableColumn.STotalsRowFunction.var;
			case "custom": return STableColumn.STotalsRowFunction.custom;
			}
			return null;
		}

		public String getTotalsRowLabel() {
			return toStringValue(jsonMap, "totalsRowLabel");
		}

		public String getUniqueName() {
			return toStringValue(jsonMap, "uniqueName");
		}

		public XlsxXmlColumnPrExtractor getXmlColumnPr() {
			return jsonMap.containsKey("xmlColumnPr")?new XlsxXmlColumnPrExtractor((Map)jsonMap.get("xmlColumnPr")) : null;
		}
	}

	public static class XlsxTableFormulaExtractor {
		final private Map jsonMap;

		public XlsxTableFormulaExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isArray() {
			return toBool(jsonMap, "array");
		}

		public String getExpression() {
			return toStringValue(jsonMap, "_");
		}
	}

	public static class XlsxXmlColumnPrExtractor {
		final private Map jsonMap;

		public XlsxXmlColumnPrExtractor(Map jsonMap) {
			this.jsonMap = Optional.ofNullable(jsonMap).orElse(Collections.EMPTY_MAP);
		}

		public boolean isDenormalized() {
			return toBool(jsonMap, "denormalized");
		}

		public Number getMapId() {
			return toNum(jsonMap, "mapId");
		}

		public String getXmlDataType() {
			return toStringValue(jsonMap, "xmlDataType");
		}

		public String getXpath() {
			return toStringValue(jsonMap, "xpath");
		}
	}

}