/* XlsxImporter.java

	Purpose:
		
	Description:
		
	History:
		4:04 PM 2019/9/18, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer;

import static io.keikai.model.SChart.ChartType;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeHueMod;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeLum;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeLumMod;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeLumOff;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeSat;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeSatMod;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeSatOff;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeShade;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithSchemeTint;
import static io.keikai.range.impl.imexp.BookHelper.getRgbWithTint;
import static io.keikai.range.impl.imexp.BookHelper.toHex;
import static io.keikai.util.XlsxJsonHelper.getParts;
import static io.keikai.util.XlsxJsonHelper.toBoolean;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import io.keikai.json.JSONObject;
import io.keikai.model.CFIconImpl;
import io.keikai.model.CellRegion;
import io.keikai.model.SAutoFilter;
import io.keikai.model.SBook;
import io.keikai.model.SBookSeries;
import io.keikai.model.SBooks;
import io.keikai.model.SBorder;
import io.keikai.model.SBorderLine;
import io.keikai.model.SCFIcon;
import io.keikai.model.SCFValueObject;
import io.keikai.model.SCell;
import io.keikai.model.SCellStyle;
import io.keikai.model.SCellStyleBuilder;
import io.keikai.model.SChart;
import io.keikai.model.SChart.ChartGrouping;
import io.keikai.model.SChartAxis;
import io.keikai.model.SColor;
import io.keikai.model.SColorFilter;
import io.keikai.model.SColorScale;
import io.keikai.model.SColumn;
import io.keikai.model.SColumnArray;
import io.keikai.model.SComment;
import io.keikai.model.SConditionalFormatting;
import io.keikai.model.SCustomFilter;
import io.keikai.model.SCustomFilters;
import io.keikai.model.SDataBar;
import io.keikai.model.SDataValidation;
import io.keikai.model.SDynamicFilter;
import io.keikai.model.SExtraStyle;
import io.keikai.model.SFill;
import io.keikai.model.SFont;
import io.keikai.model.SFooter;
import io.keikai.model.SHeader;
import io.keikai.model.SHyperlink;
import io.keikai.model.SIconSet;
import io.keikai.model.SName;
import io.keikai.model.SNamedStyle;
import io.keikai.model.SPicture;
import io.keikai.model.SPrintSetup;
import io.keikai.model.SRichText;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.model.SSheetProtection;
import io.keikai.model.SSheetViewInfo;
import io.keikai.model.STable;
import io.keikai.model.STableColumn;
import io.keikai.model.STableStyle;
import io.keikai.model.STableStyleElem;
import io.keikai.model.STableStyleInfo;
import io.keikai.model.STop10Filter;
import io.keikai.model.SheetRegion;
import io.keikai.model.ViewAnchor;
import io.keikai.model.chart.SChartData;
import io.keikai.model.chart.SGeneralChartData;
import io.keikai.model.chart.SSeries;
import io.keikai.model.impl.AbstractBookAdv;
import io.keikai.model.impl.AbstractCellAdv;
import io.keikai.model.impl.AbstractChartAdv;
import io.keikai.model.impl.AbstractChartAxisAdv;
import io.keikai.model.impl.AbstractDataValidationAdv;
import io.keikai.model.impl.AbstractFontBuilderAdv;
import io.keikai.model.impl.AbstractSeriesAdv;
import io.keikai.model.impl.AbstractSheetAdv;
import io.keikai.model.impl.BorderImpl;
import io.keikai.model.impl.BorderLineImpl;
import io.keikai.model.impl.CFValueObjectImpl;
import io.keikai.model.impl.ColorFilterImpl;
import io.keikai.model.impl.ColorImpl;
import io.keikai.model.impl.ColorScaleImpl;
import io.keikai.model.impl.ConditionalFormattingImpl;
import io.keikai.model.impl.ConditionalFormattingRuleImpl;
import io.keikai.model.impl.CustomFilterImpl;
import io.keikai.model.impl.CustomFiltersImpl;
import io.keikai.model.impl.DataBarImpl;
import io.keikai.model.impl.DynamicFilterImpl;
import io.keikai.model.impl.ExtraFillImpl;
import io.keikai.model.impl.HeaderFooterImpl;
import io.keikai.model.impl.IconSetImpl;
import io.keikai.model.impl.ImmutableExtraStyleImpl;
import io.keikai.model.impl.NamedStyleImpl;
import io.keikai.model.impl.RichTextImpl;
import io.keikai.model.impl.SheetImpl;
import io.keikai.model.impl.TableColumnImpl;
import io.keikai.model.impl.TableImpl;
import io.keikai.model.impl.TableStyleElemImpl;
import io.keikai.model.impl.TableStyleImpl;
import io.keikai.model.impl.TableStyleInfoImpl;
import io.keikai.model.impl.Top10FilterImpl;
import io.keikai.model.impl.WorkbookProtectionImpl;
import io.keikai.model.impl.chart.AreaChartDataImpl;
import io.keikai.model.impl.chart.BarChartDataImpl;
import io.keikai.model.impl.chart.CategoryAxisImpl;
import io.keikai.model.impl.chart.DateAxisImpl;
import io.keikai.model.impl.chart.GeneralChartDataImpl;
import io.keikai.model.impl.chart.LineChartDataImpl;
import io.keikai.model.impl.chart.ValueAxisImpl;
import io.keikai.model.impl.sys.formula.FormulaEngineImpl;
import io.keikai.model.impl.sys.formula.ParsingBook;
import io.keikai.model.sys.EngineFactory;
import io.keikai.model.sys.formula.FormulaEngine;
import io.keikai.model.sys.formula.FormulaExpression;
import io.keikai.model.sys.formula.FormulaParseContext;
import io.keikai.model.sys.formula.FormulaType;
import io.keikai.model.util.Strings;
import io.keikai.range.impl.imexp.AbstractImporter;
import io.keikai.range.impl.imexp.PoiEnumConversion;
import io.keikai.range.impl.imexp.StyleCache;
import io.keikai.range.impl.imexp.UnitUtil;
import io.keikai.util.Converter;
import io.keikai.util.Ref;
import org.apache.commons.io.IOUtils;
import org.apache.poi.ss.SpreadsheetVersion;
import org.apache.poi.ss.formula.FormulaRenderer;
import org.apache.poi.ss.formula.SharedFormula;
import org.apache.poi.ss.formula.eval.BoolEval;
import org.apache.poi.ss.formula.eval.ErrorEval;
import org.apache.poi.ss.formula.eval.NumberEval;
import org.apache.poi.ss.formula.eval.StringEval;
import org.apache.poi.ss.formula.eval.ValueEval;
import org.apache.poi.ss.formula.ptg.FuncVarPtg;
import org.apache.poi.ss.formula.ptg.Ptg;
import org.apache.poi.ss.util.CellRangeAddress;
import org.jsoup.nodes.Document;
import org.jsoup.parser.ParseSettings;
import org.jsoup.parser.Parser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.zkoss.util.Locales;
import org.zkoss.util.Pair;

/**
 * An excel xlsx importer
 *
 * @author jumperchen
 * @since 5.0.0
 */
public class XlsxImporter extends AbstractImporter implements Serializable {

	private static final long serialVersionUID = 20190918160712L;
	private static final Logger _logger = LoggerFactory.getLogger(XlsxImporter.class);

	protected Map<XlsxExtractor.XlsxXfExtractor, SCellStyle> importedXfs = new HashMap<XlsxExtractor.XlsxXfExtractor, SCellStyle>();    //ZSS-685

	protected StyleCache styleCaches = new StyleCache();

	// KEIKAI-164
	protected ImportPhase _importPhase;

	/**
	 * <poi Font index, {@link SFont} object>
	 **/
//	protected Map<Short, SFont> importedFont = new HashMap<Short, SFont>();
	//ZSS-735, poiPictureData -> SPictureData index
	protected Map<PictureJson, Integer> importedPictureJson = new HashMap<>();

	protected Map<String, Ptg[]> _sharedFormulaPtgArray = new HashMap<>();

	/**
	 * target book model
	 */
	protected SBook book;
	XlsxExtractor.XlsxThemesExtractor themes;
	// xml data map
	XlsxExtractor.XlsxWorkbookExtractor workbookExtractor;
	public SBook imports(InputStream inputStream, String bookName)
			throws IOException {
		if (_logger.isDebugEnabled()) {
			_logger.debug("start to import {}", bookName);
		}
		// clear cache for reuse
		importedXfs.clear();
		styleCaches.clear();

		book = SBooks.createBook(bookName);

		workbookExtractor = newXlsxExtractor(
				inputStream);

		//ZSS-715: Enforce internal Locale.US Locale so formula is in consistent internal format
		Locale old = Locales.setThreadLocal(Locale.US);
		SBookSeries bookSeries = book.getBookSeries();
		boolean isCacheClean = bookSeries.isAutoFormulaCacheClean();
		try {
			_importPhase = new ImportPhase(bookName);
			((AbstractBookAdv) book).setPostProcessing(true); //ZSS-1283
			_importPhase.setPhase(ImportPhase.STYLE_SHEET);
			XlsxExtractor.XlsxStyleSheetExtractor styleSheet = workbookExtractor
					.getStyleSheet();

			// refer to KK importer
			this.themes = workbookExtractor
					.getThemes();

			importDefaultChartSeriesColors();
			importStyleSheet(styleSheet);
			_importPhase.clear();

			setBookType(book);

			bookSeries.setAutoFormulaCacheClean(false);// disable it to avoid
			// unnecessary clean up
			// during importing

			_importPhase.setPhase(ImportPhase.EXTERNAL_BOOK_REFERENCE);
			importExternalBookLinks(workbookExtractor.getExternalReferences());

			_importPhase.setPhase(ImportPhase.WORKBOOK_PROTECTION);
			importWorkbookProtection(workbookExtractor, book);

			_importPhase.setPhase(ImportPhase.SHEETS);
			List<XlsxExtractor.XlsxSheetExtractor> sheets = workbookExtractor
					.getSheets();
			importSheets(sheets);
			_importPhase.setPhase(ImportPhase.NAMED_RANGE);
			importNamedRange();
			_importPhase.clear();

			for (int i = 0; i < sheets.size(); i++) {
				_importPhase.setPhase(ImportPhase.SHEET);
				SSheet sheet = book.getSheet(i);
				_importPhase.setSheetName(sheet.getSheetName());
				XlsxExtractor.XlsxSheetExtractor xSheet = sheets.get(i);

				_importPhase.setPhase(ImportPhase.HYPERLINK);
				importHyperlinks(xSheet);

				_importPhase.setPhase(ImportPhase.VML_DRAWING);
				importVmlDrawing(xSheet);

				_importPhase.setPhase(ImportPhase.COMMENT);
				importComments(xSheet);

				_importPhase.setPhase(ImportPhase.ROW);
				for (XlsxExtractor.XlsxRowExtractor xRow : xSheet.getRows()) {
					importRow(xRow, sheet);
				}
				_importPhase.setRowIndex(-1);
				_importPhase.setColIndex(-1);

				importColumn(xSheet, sheet);
				_importPhase.setColIndex(-1);

				_importPhase.setPhase(ImportPhase.MERGED_CELL);
				importMergedRegions(xSheet, sheet);

				// KEIKAI-395, retrieve <extLst> early and to be used by importSparklines() and importConditionalFormatting()
				_importPhase.setPhase(ImportPhase.EXT_LST);
				final XlsxExtractor.XlsxExtLstExtractor extLst = xSheet.getExtLst();
				List<XlsxExtractor.XlsxExtExtractor> extList = extLst != null
						? extLst.getExt()
						: Collections.EMPTY_LIST;

				_importPhase.setPhase(ImportPhase.DRAWING);
				importDrawings(xSheet, sheet);
				importSparklines(xSheet, sheet, extList); // KEIKAI-395, modify to take Ext list argumet

				_importPhase.setPhase(ImportPhase.DATA_VALIDATION);
				importValidation(xSheet, sheet, extList);

				_importPhase.setPhase(ImportPhase.AUTO_FILTER);
				importAutoFilter(xSheet, sheet);

				_importPhase.setPhase(ImportPhase.SHEET_PROTECTION);
				importSheetProtection(xSheet, sheet); //ZSS-576

				_importPhase.setPhase(ImportPhase.CONDITIONAL_FORMATTING);
				importConditionalFormatting(xSheet, sheet, extList); //ZSS-1130, //KEIKAI-395

				// do this after importNamedRange();
				_importPhase.setPhase(ImportPhase.PRINT_AREA);
				importPrintArea(xSheet, sheet);
			}
		} catch (RuntimeException e) { // KEIKAI-164
			throw new ImportException("Fail to import " +  _importPhase.toString(), e);
		} finally {
			book.getBookSeries().setAutoFormulaCacheClean(isCacheClean);
			Locales.setThreadLocal(old);
			//ZSS-1283
			((AbstractBookAdv) book).setPostProcessing(false);
			_importPhase.clear();
			if (_logger.isDebugEnabled()) {
				_logger.debug("complete to import {}", bookName);
			}
		}
		return book;
	}


	/**
	 * A built-in defined name that specifies the workbook's print area
	 */
	public final static String BUILTIN_PRINT_AREA = "_xlnm.Print_Area";
	/**
	 * A built-in defined name that specifies the row(s) or column(s) to repeat
	 * at the top of each printed page.
	 */
	public final static String BUILTIN_PRINT_TITLE = "_xlnm.Print_Titles";

	protected void importPrintArea(XlsxExtractor.XlsxSheetExtractor xSheet,
			SSheet sheet) {
		SName printArea = book
				.getNameByName(BUILTIN_PRINT_AREA, sheet.getSheetName());
		SPrintSetup sps= sheet.getPrintSetup();

		XlsxExtractor.XlsxHeaderFootersExtractor headerFooter = xSheet
				.getHeaderFooter();

		if (headerFooter != null) {
			SSheetViewInfo viewInfo = sheet.getViewInfo();

			sps.setAlignWithMargins(headerFooter.isAlignWithMargins());
			String[] oddHeaders = getParts(headerFooter.getOddHeader());
			String[] oddFooters = getParts(headerFooter.getOddFooter());

			SHeader header = viewInfo.getHeader();
			header.setCenterText(oddHeaders[1]);
			header.setLeftText(oddHeaders[0]);
			header.setRightText(oddHeaders[2]);
			sps.setHeader(header);

			SFooter footer = viewInfo.getFooter();
			footer.setCenterText(oddFooters[1]);
			footer.setLeftText(oddFooters[0]);
			footer.setRightText(oddFooters[2]);
			sps.setFooter(footer);


			if (headerFooter.isDifferentOddEven()) {
				String evenHeader1 = headerFooter.getEvenHeader();
				if (evenHeader1 != null) {
					SHeader evenHeader = new HeaderFooterImpl();
					String[] parts = getParts(evenHeader1);
					evenHeader.setCenterText(parts[1]);
					evenHeader.setLeftText(parts[0]);
					evenHeader.setRightText(parts[2]);
					sps.setEvenHeader(evenHeader);
				}
				String evenFooter1 = headerFooter.getEvenFooter();
				if (evenFooter1 != null) {
					SFooter evenFooter = new HeaderFooterImpl();
					String[] parts = getParts(evenFooter1);
					evenFooter.setCenterText(parts[1]);
					evenFooter.setLeftText(parts[0]);
					evenFooter.setRightText(parts[2]);
					sps.setEvenFooter(evenFooter);
				}
			}

			if (headerFooter.isDifferentFirst()) {
				String firstHeader1 = headerFooter.getFirstHeader();
				if (firstHeader1 != null) {
					SHeader firstHeader = new HeaderFooterImpl();
					String[] parts = getParts(firstHeader1);
					firstHeader.setCenterText(parts[1]);
					firstHeader.setLeftText(parts[0]);
					firstHeader.setRightText(parts[2]);
					sps.setFirstHeader(firstHeader);
				}
				String firstFooter1 = headerFooter.getFirstFooter();
				if (firstFooter1 != null) {
					SFooter firstFooter = new HeaderFooterImpl();
					String[] parts = getParts(firstHeader1);
					firstFooter.setCenterText(parts[1]);
					firstFooter.setLeftText(parts[0]);
					firstFooter.setRightText(parts[2]);
					sps.setFirstFooter(firstFooter);
				}
			}

			sps.setScaleWithDoc(headerFooter.isScaleWithDoc());
			sps.setDifferentOddEvenPage(headerFooter.isDifferentOddEven());
			sps.setDifferentFirstPage(headerFooter.isDifferentFirst());
		}

		XlsxExtractor.XlsxPageSetupExtractor pageSetup = xSheet.getPageSetup();
		XlsxExtractor.XlsxPageMarginsExtractor pageMargins = xSheet
				.getPageMargins();
		XlsxExtractor.XlsxPrintOptionsExtractor printOptions = xSheet
				.getPrintOptions();

		if (pageMargins != null) {
			sps.setBottomMargin(pageMargins.getBottom());
			sps.setTopMargin(pageMargins.getTop());
			sps.setLeftMargin(pageMargins.getLeft());
			sps.setRightMargin(pageMargins.getRight());
			sps.setHeaderMargin(pageMargins.getHeader());
			sps.setFooterMargin(pageMargins.getFooter());
		}
		if (printOptions != null && printOptions.isNotEmpty()) { // KEIKAI-403
			sps.setHCenter(printOptions.isHorizontalCentered());
			sps.setVCenter(printOptions.isVerticalCentered());
			sps.setPrintGridlines(printOptions.isGridLines());
			sps.setPrintHeadings(printOptions.isHeadings());
		}
		if (pageSetup != null) {
			sps.setErrorPrintMode(pageSetup.getErrors());
			sps.setFitHeight(pageSetup.getFitToHeight());
			sps.setFitWidth(pageSetup.getFitToWidth());
			sps.setLandscape("landscape".equals(pageSetup.getOrientation()));
			sps.setLeftToRight("overThenDown".equals(pageSetup.getPageOrder()));
			sps.setPaperSize(pageSetup.getPaperSize());
			sps.setCommentsMode(pageSetup.getCellComments());
			sps.setScale(pageSetup.getScale());
			sps.setPageStart(pageSetup.isUseFirstPageNumber() ? pageSetup.getFirstPageNumber() : 0);
		}

		if (printArea != null) {
			sps.setPrintArea(printArea.getRefersToFormula());
		}

		SName printTitle = book
				.getNameByName(BUILTIN_PRINT_TITLE, sheet.getSheetName());
		if (printTitle != null) {
			String refersToFormula = printTitle.getRefersToFormula();
			if (refersToFormula != null) {
				CellRangeAddress rowrng = getRepeatingRowsOrColumns(
						refersToFormula, true);
				if (rowrng != null) {
					sps.setRepeatingRowsTitle(rowrng.getFirstRow(), rowrng.getLastRow());
				}
				CellRangeAddress colrng = getRepeatingRowsOrColumns(
						refersToFormula, false);
				if (colrng != null) {
					sps.setRepeatingColumnsTitle(colrng.getFirstColumn(),
							colrng.getLastColumn());
				}
			}
		}
	}

	protected static CellRangeAddress getRepeatingRowsOrColumns(String refs, boolean rows) {
		String[] parts = refs.split(",");
		int maxRowIndex = Ref.MAX_ROW - 1;
		int maxColIndex = Ref.MAX_COL - 1;
		for (String part : parts) {
			CellRangeAddress range = CellRangeAddress.valueOf(part);
			if ((range.getFirstColumn() == 0
					&& range.getLastColumn() == maxColIndex)
					|| (range.getFirstColumn() == -1
					&& range.getLastColumn() == -1)) {
				if (rows) {
					return range;
				}
			} else if (range.getFirstRow() == 0
					&& range.getLastRow() == maxRowIndex
					|| (range.getFirstRow() == -1
					&& range.getLastRow() == -1)) {
				if (!rows) {
					return range;
				}
			}
		}
		return null;
	}

	// Sheet scope
	Map<String, XlsxExtractor.XlsxHyperlinkExtractor> hyperlinksCache;
	protected void importHyperlinks(XlsxExtractor.XlsxSheetExtractor xSheet) {
		List<XlsxExtractor.XlsxHyperlinkExtractor> hyperlinks = xSheet
				.getHyperlinks();
		if (hyperlinks.isEmpty()) {
			hyperlinksCache = null;
		} else {
			hyperlinksCache = new HashMap<>();
			for (XlsxExtractor.XlsxHyperlinkExtractor extractor : hyperlinks) {
				hyperlinksCache.put(extractor.getRef(), extractor);
			}
		}
	}
	// Sheet scope
	VmlDrawing vmlDrawingCache;
	protected void importVmlDrawing(XlsxExtractor.XlsxSheetExtractor xSheet) {
		XlsxExtractor.XlsxVmlDrawingExtractor legacyDrawing = xSheet
				.getLegacyDrawing();
		if (legacyDrawing != null) {
			vmlDrawingCache = new VmlDrawing(legacyDrawing.getJsonMap(), legacyDrawing.getImageData());
		} else {
			vmlDrawingCache = null; //reset
		}
	}

	// Sheet scope
	Map<String, XlsxExtractor.XlsxCommentExtractor> commentsCache;
	protected void importComments(XlsxExtractor.XlsxSheetExtractor xSheet) {
		XlsxExtractor.XlsxCommentsExtractor comments = xSheet
				.getComments();
		if (comments != null) {
			List<String> authors = comments.getAuthors();
			List<XlsxExtractor.XlsxCommentExtractor> commentList = comments
					.getCommentList();
			if (commentList.isEmpty()) {
				commentsCache = null; //reset
			} else {
				commentsCache = new HashMap<>();
				for (XlsxExtractor.XlsxCommentExtractor extractor : commentList) {
					Number authorId = extractor.getAuthorId();
					if (authorId != null && authorId.intValue() < authors.size()) {
						extractor.setAuthorName(authors.get(authorId.intValue()));
					}
					Number shapeId = extractor.getShapeId();
					if (shapeId != null) {
						extractor.setShape(vmlDrawingCache.getShape(extractor.getRef()));
					}
					commentsCache.put(extractor.getRef(), extractor);
				}
			}
		} else {
			commentsCache = null; //reset
		}
	}

	protected void importDefaultChartSeriesColors() {
		if (this.themes == null) {
			return;
		}
		final XlsxExtractor.XlsxClrSchemeExtractor clrScheme = themes.getClrScheme();
		if (clrScheme == null) {
			return;
		}
		final List<SColor> list = new ArrayList<>();
		list.add(new ColorImpl(clrScheme.getAccent1()));
		list.add(new ColorImpl(clrScheme.getAccent2()));
		list.add(new ColorImpl(clrScheme.getAccent3()));
		list.add(new ColorImpl(clrScheme.getAccent4()));
		list.add(new ColorImpl(clrScheme.getAccent5()));
		list.add(new ColorImpl(clrScheme.getAccent6()));
		((AbstractBookAdv) book).setDefaultChartSeriesColors(list);
	}


	protected void importStyleSheet(
			XlsxExtractor.XlsxStyleSheetExtractor extractor) {
		importColors(extractor.getColors());
		importNumFmts(extractor.getNumFmts());
		importFonts(extractor.getFonts());
		importFills(extractor.getFills());
		importBorders(extractor.getBorders());

		//ZSS-854: same as importCellStyleXfs() in KK
		importDefaultCellStyles(extractor);

		importCellXfs(extractor.getCellXfs());

		// same as importCellStyles() in KK
		importNamedStyles(extractor);

		//ZSS-1140: same sa importDxfs() in KK
		importExtraStyles(extractor);

		//ZSS-992
		importTableStyles(extractor);
	}
	CellXfs cellXfsCache;
	void importCellXfs(List<XlsxExtractor.XlsxXfExtractor> extractors) {
		cellXfsCache = new CellXfs();
		if (!extractors.isEmpty()) {
			for (XlsxExtractor.XlsxXfExtractor extractor : extractors) {
				XlsxExtractor.XlsxXfExtractor parentXf = cellStyleXfsCache
						.get(extractor.getXfId().intValue());
				cellXfsCache.addCellXf(new XlsxExtractor.XlsxCellXfExtractor(extractor, parentXf));
			}
		}
	}

	Colors colorsCache;
	protected void importColors(XlsxExtractor.XlsxRGBColorsExtractor extractor) {
		if (extractor != null) {
			List<String> indexedColorsList = extractor.getIndexedColors();
			boolean ignoreBuiltin = indexedColorsList != null && indexedColorsList.isEmpty() == false;
			colorsCache = new Colors(ignoreBuiltin);
			for (String color: indexedColorsList) {
				colorsCache.addIndexedColor(color);
			}
		} else {
			colorsCache = new Colors();
		}
	}
	NumFmts numFmtsCache;
	protected void importNumFmts(List<XlsxExtractor.XlsxNumFmtExtractor> extractors) {
		numFmtsCache = new NumFmts();

		// add default one
		if (extractors.isEmpty()) {
			numFmtsCache.addNumFmt("General", 0);
		} else {
			for (XlsxExtractor.XlsxNumFmtExtractor extractor: extractors) {
				numFmtsCache.addNumFmt(extractor.getFormatCode(), extractor.getNumFmtId().intValue());
			}
		}
	}

	Fonts fontsCache;
	protected void importFonts(List<XlsxExtractor.XlsxFontExtractor> extractors) {
		 fontsCache = new Fonts();
		for (XlsxExtractor.XlsxFontExtractor font : extractors) {
			fontsCache.addFont(font, createZssFont(font));
		}
	}
	Fills fillsCache;
	protected void importFills(List<XlsxExtractor.XlsxFillExtractor> extractors) {
		fillsCache = new Fills();
		if (!extractors.isEmpty()) {
			for (XlsxExtractor.XlsxFillExtractor extractor : extractors) {
				if (extractor instanceof XlsxExtractor.XlsxPatternExtractor) {
					XlsxExtractor.XlsxPatternExtractor fill = ((XlsxExtractor.XlsxPatternExtractor) extractor);
					String fgColor = getForegroundHTMLColor(fill.getFgColor(), themes, this);
					String bgColor = getBackgroundHTMLColor(fill.getBgColor(), themes, this);
					fillsCache.addFill(new Fills.PatternFill(fill.getPatternType(), fgColor, bgColor));
					colorsCache.addColor(fgColor);
					colorsCache.addColor(bgColor);
				}/** TODO: else if (fill is GradientExtractor) {
					final f = createGradientFill(fill);
					if (f != null) {
						styleCaches.fillsCache.add(f);
					}
				}*/
			}
		} else {
			if (themes != null) {
				XlsxExtractor.XlsxClrSchemeExtractor clrScheme = themes
						.getClrScheme();
				if (clrScheme != null) {
					fillsCache.addFill(new Fills.PatternFill("none", clrScheme.getLt1(), clrScheme.getDk1()));
					fillsCache.addFill(new Fills.PatternFill("gray125", clrScheme.getLt1(), clrScheme.getDk1()));
				} else {
					fillsCache.addFill(new Fills.PatternFill("none", "#ffffff", "#000000"));
					fillsCache.addFill(new Fills.PatternFill("gray125", "#ffffff", "#000000"));
				}
			} else {
				fillsCache.addFill(new Fills.PatternFill("none", "#ffffff", "#000000"));
				fillsCache.addFill(new Fills.PatternFill("gray125", "#ffffff", "#000000"));
			}
		}
	}
	Borders bordersCache;
	protected void importBorders(List<XlsxExtractor.XlsxBorderExtractor> extractors) {
		bordersCache = new Borders();
		if (!extractors.isEmpty()) {
			for (XlsxExtractor.XlsxBorderExtractor extractor : extractors) {
				bordersCache.addBorder(createBorder(extractor));
			}
		} else {
			Borders.BorderPr defaultBorderPr = new Borders.BorderPr("none",
					getHTMLColorFromHex(colorsCache.getIndexColorByIndex(0)));
			bordersCache.addBorder(new Borders.Border(false, false, false,
					defaultBorderPr, defaultBorderPr, defaultBorderPr, defaultBorderPr, defaultBorderPr, null, null));
		}
	}
	protected Borders.Border createBorder(
			XlsxExtractor.XlsxBorderExtractor extractor) {
		boolean diagonalUp = extractor.isDiagonalUp();
		boolean diagonalDown = extractor.isDiagonalDown();
		boolean outline = extractor.isOutline();
		Borders.BorderPr left;
		Borders.BorderPr top;
		Borders.BorderPr right;
		Borders.BorderPr bottom;
		Borders.BorderPr diagonal;
		Borders.BorderPr vertical;
		Borders.BorderPr horizontal;

		left = createBorderPr(extractor.getLeft());
		top = createBorderPr(extractor.getTop());
		right = createBorderPr(extractor.getRight());
		bottom = createBorderPr(extractor.getBottom());
		diagonal = createBorderPr(extractor.getDiagonal());
		vertical = createBorderPr(extractor.getVertical());
		horizontal = createBorderPr(extractor.getHorizontal());
		return new Borders.Border(diagonalUp, diagonalDown, outline, left, top, right, bottom, diagonal, vertical, horizontal);
	}
	protected Borders.BorderPr createBorderPr(XlsxExtractor.XlsxBorderPrExtractor extractor) {
		if (extractor == null) return null;
		if (extractor.getColor() == null && extractor.getStyle() == null) {
			return new Borders.BorderPr("none", getHTMLColorFromHex(colorsCache.getIndexColorByIndex(0)));
		}
		String fontHTMLColor = getBorderHTMLColor(extractor.getColor(), themes, this);
		colorsCache.addColor(fontHTMLColor);
		return new Borders.BorderPr(extractor.getStyle(), fontHTMLColor);
	}
	protected void importAutoFilter(
			XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet) {
		XlsxExtractor.XlsxAutoFilterExtractor xAutoFilter = xSheet
				.getAutoFilter();

		if (xAutoFilter != null) {
			String ref = xAutoFilter.getRef();

			CellRegion cellRegion = new CellRegion(ref);
			SAutoFilter autoFilter = sheet.createAutoFilter(cellRegion);
			importAutoFilterColumns(xAutoFilter, autoFilter); //ZSS-1019
		}
	}

	protected void importSheetProtection(
			XlsxExtractor.XlsxSheetExtractor xSheet,
			SSheet sheet) { //ZSS-576
		XlsxExtractor.XlsxSheetProtectionExtractor sheetProtection = xSheet
				.getSheetProtection();
		if (sheetProtection != null) {
			sheet.setPassword(sheetProtection.isSheet() ? "" : null);

			//import hashed password directly
			importPassword(sheetProtection, sheet);

			SSheetProtection ssp = sheet.getSheetProtection();

			ssp.setAutoFilter(sheetProtection.isAutoFilter());
			ssp.setDeleteColumns(sheetProtection.isDeleteColumns());
			ssp.setDeleteRows(sheetProtection.isDeleteRows());
			ssp.setFormatCells(sheetProtection.isFormatCells());
			ssp.setFormatColumns(sheetProtection.isFormatColumns());
			ssp.setFormatRows(sheetProtection.isFormatRows());
			ssp.setInsertColumns(sheetProtection.isInsertColumns());
			ssp.setInsertHyperlinks(sheetProtection.isInsertHyperlinks());
			ssp.setInsertRows(sheetProtection.isInsertRows());
			ssp.setPivotTables(sheetProtection.isPivotTables());
			ssp.setSort(sheetProtection.isSort());
			ssp.setObjects(sheetProtection.isObjects());
			ssp.setScenarios(sheetProtection.isScenarios());
			ssp.setSelectLockedCells(sheetProtection.isSelectLockedCells());
			ssp.setSelectUnlockedCells(sheetProtection.isSelectUnlockedCells());
		}
	}

	/**
	 * Reference BookHelper.validate()
	 */
	protected void importValidation(XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet, List<XlsxExtractor.XlsxExtExtractor> extList) {
		XlsxExtractor.XlsxDataValidationsExtractor dataValidations = xSheet
				.getDataValidations();
		Consumer<XlsxExtractor.XlsxDataValidationExtractor> func = (extractor) -> {
			SDataValidation dataValidation = sheet
					.addDataValidation(null, null);
			// getExplicitListValues() will be represented as formula1
			dataValidation.setOperatorType(extractor.getOperator());
			dataValidation.setValidationType(extractor.getType());
			((AbstractDataValidationAdv) dataValidation).setEscapedFormulas(extractor.getFormula1(),
					extractor.getFormula2());

			dataValidation.setIgnoreBlank(extractor.isAllowBlank());
			dataValidation.setErrorTitle(extractor.getErrorTitle());
			dataValidation.setErrorMessage(extractor.getError());
			dataValidation.setAlertStyle(extractor.getErrorStyle());
			dataValidation.setInputTitle(extractor.getPromptTitle());
			dataValidation.setInputMessage(extractor.getPrompt());
			if (dataValidation.getValidationType() == SDataValidation.ValidationType.LIST) {
				/*
				 * According to ISO/IEC 29500-1 \ 18.3.1.32  dataValidation (Data Validation)
				 * Excel file contains reversed value against format specification . If not showing drop down, file contains 1 (true). If showing,
				 * attribute "showDropDown" won't exist and POI get false. But POI's API reverse its value again.
				 * So we just directly accept the POI returned value.
				 */
				dataValidation.setInCellDropdown(extractor.isShowDropDown());
			}
			dataValidation.setShowError(extractor.isShowErrorMessage());
			dataValidation.setShowInput(extractor.isShowInputMessage());

			/*
			 * According to ISO/IEC 29500-1 \ 18.18.76  ST_Sqref (Reference Sequence) and A.2
			 * Its XML Schema indicates it's a required attribute, so CellRangeAddresses must have at least one address.
			 */
			LinkedHashSet<CellRegion> regions = new LinkedHashSet<CellRegion>();
			for (String ref :  extractor.getSqref().split(" ")) {
				regions.add(new CellRegion(ref));
			}
			dataValidation.setRegions(regions);
		};

		if (dataValidations != null) {
			for (XlsxExtractor.XlsxDataValidationExtractor extractor : dataValidations.getValidations()) {
				func.accept(extractor);
			}
		}

		if (!extList.isEmpty()) {
			extList.forEach(xlsxExtExtractor -> {
				xlsxExtExtractor.getDataValidations().forEach(func);
			});
		}
	}

	protected void importMergedRegions(XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet) {
		final SheetImpl sheetImpl = (SheetImpl) sheet;
		List<String> mergeCells = xSheet.getMergeCells();

		// consistent with AbstractExcelImporter, read from last merged region for backward compatibility
		for (int i = mergeCells.size(); --i >= 0;) {
			Ref ref1 = Converter.a1ToRef(mergeCells.get(i));
			//ZSS-1114: any new merged region that overlapped with previous merged region is thrown away
			final CellRegion r = new CellRegion(ref1.getTop(),
					ref1.getLeft(), ref1.getBottom(),
					ref1.getRight());
			final CellRegion overlapped = sheetImpl.checkMergedRegion(r);
			if (overlapped != null) {
				_logger.warn("Drop the region " + r
						+ " which is overlapped with existing merged area "
						+ overlapped + ".");
				continue;
			}
			sheetImpl.addDirectlyMergedRegion(r);
		}
	}

	protected void importSparklines(XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet, List<XlsxExtractor.XlsxExtExtractor> extList) {
		if (extList.isEmpty()) {
			return;
		}
		for (XlsxExtractor.XlsxExtExtractor ext : extList) {
			XlsxExtractor.XlsxSparklineGroupsExtractor sparklineGroups = ext.getSparklineGroups();
			if (sparklineGroups == null) {
				return;
			}
			final List<XlsxExtractor.XlsxSparklineGroupExtractor> sparklineGroupList = sparklineGroups.getSparklineGroups();
			if (sparklineGroupList.isEmpty()) {
				return;
			}
			for (XlsxExtractor.XlsxSparklineGroupExtractor sparklineGroup : sparklineGroupList) {
				final Boolean isMarkersVisible = sparklineGroup.isMarkersVisible();
				final Boolean isNegativeVisible = sparklineGroup.isNegativeVisible();
				final Boolean isFirstVisible = sparklineGroup.isFirstVisible();
				final Boolean isLastVisible = sparklineGroup.isLastVisible();
				final Boolean isHighVisible = sparklineGroup.isHighVisible();
				final Boolean isLowVisible = sparklineGroup.isLowVisible();
				final SColor seriesColor = toColor(sparklineGroup.getColorSeries());
				final SColor markerColor = toColor(sparklineGroup.getColorMarkers());
				final SColor negativeColor = toColor(sparklineGroup.getColorNegative());
				final SColor firstColor = toColor(sparklineGroup.getColorFirst());
				final SColor lastColor = toColor(sparklineGroup.getColorLast());
				final SColor highColor = toColor(sparklineGroup.getColorHigh());
				final SColor lowColor = toColor(sparklineGroup.getColorLow());
				List<XlsxExtractor.XlsxSparklineExtractor> sparklineList = sparklineGroup.getSparklines().getSparklines();
				for (XlsxExtractor.XlsxSparklineExtractor sparkline : sparklineList) {
					String anchorRef = sparkline.getSqref();
					ViewAnchor viewAnchor = toSparklineViewAnchor(anchorRef, sheet);
					final String type = sparklineGroup.getType();
					final boolean winlose = "stacked".equalsIgnoreCase(type);
					ChartType chartType = ChartType.LINE;
					if (ChartType.AREA.name().equalsIgnoreCase(type)) {
						chartType = ChartType.AREA;
					} else if (ChartType.COLUMN.name().equalsIgnoreCase(type) || winlose) {
						chartType = ChartType.COLUMN;
					}
					SChart chart = sheet.addChart(chartType, viewAnchor);
					final AbstractChartAdv chartAdv = (AbstractChartAdv) chart;
					SGeneralChartData chartData = (SGeneralChartData) chart.getData();
					chart.setSparkline(true);
					if (isMarkersVisible != null) {
						chartAdv.setSparklineMarkerVisible(isMarkersVisible);
					}

					if (isNegativeVisible != null) {
						chartAdv.setSparklineNegativePointVisible(isNegativeVisible);
					}

					if (isFirstVisible != null) {
						chartAdv.setSparklineFirstPointVisible(isFirstVisible);
					}

					if (isLastVisible != null) {
						chartAdv.setSparklineLastPointVisible(isLastVisible);
					}

					if (isHighVisible != null) {
						chartAdv.setSparklineHighPointVisible(isHighVisible);
					}

					if (isLowVisible != null) {
						chartAdv.setSparklineLowPointVisible(isLowVisible);
					}

					if (seriesColor != null) {
						chartAdv.setSparklineSeriesColor(seriesColor);
					}

					if (markerColor != null) {
						chartAdv.setSparklineMarkerColor(markerColor);
					}

					if (negativeColor != null) {
						chartAdv.setSparklineNegativePointColor(negativeColor);
					}

					if (firstColor != null) {
						chartAdv.setSparklineFirstPointColor(firstColor);
					}

					if (lastColor != null) {
						chartAdv.setSparklineLastPointColor(lastColor);
					}

					if (highColor != null) {
						chartAdv.setSparklineHighPointColor(highColor);
					}

					if (lowColor != null) {
						chartAdv.setSparklineLowPointColor(lowColor);
					}
					if (winlose) {
						chart.setWinloseSpark(true);
					}
					final SSeries series = chartData.addSeries();
					series.setFormula(null, sparkline.getF());
				}
			}
		}
	}

	protected SColor toColor(XlsxExtractor.XlsxColorExtractor extractor) {
		final String hex = getHTMLColor(extractor, themes, this, null);
		if (hex == null) {
			return null;
		}
		return new ColorImpl(hex);
	}

	protected ViewAnchor toSparklineViewAnchor(String anchorRef, SSheet sheet) {
		final int rowIndex = sheet.getCell(anchorRef).getRowIndex();
		final int colIndex = sheet.getCell(anchorRef).getColumnIndex();
		CellRegion cellRegion = sheet.getMergedRegion(rowIndex, colIndex);
		int height = 0;
		int width = 0;
		if (cellRegion != null) {
			for (int i = 0; i < cellRegion.getColumnCount(); i++) {
				width += sheet.getColumn(colIndex + i).getWidth();
			}
			for (int i = 0; i < cellRegion.getRowCount(); i++) {
				height += sheet.getRow(rowIndex + i).getHeight();
			}
			width -= 4;
			height -= 4;
		} else {
			SRow row = sheet.getRow(rowIndex);
			height = row.isHidden() ? 0 : (row.getHeight() - 4);
			SColumn column = sheet.getColumn(colIndex);
			width = column.isHidden() ? 0 : column.getWidth() - 4;
		}
		ViewAnchor viewAnchor = new ViewAnchor(rowIndex, colIndex, width, height);
		viewAnchor.setXOffset(2);
		viewAnchor.setYOffset(2);
		return viewAnchor;
	}

	/*
	 * import drawings which include charts and pictures.
	 * reference DrawingManagerImpl.initXSSFDrawings()
	 */
	Drawings drawingsCache;
	protected void importDrawings(XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet) {
		final XlsxExtractor.XlsxDrawingsExtractor drawings = xSheet.getDrawings();
		if (drawings != null) {
			for (final ShapeJson shapeJson : drawings.getDrawingsJson()) {
				if (shapeJson instanceof ChartSpaceJson) {
					_importPhase.setPhase(ImportPhase.CHART);
					importChart((ChartSpaceJson) shapeJson, xSheet, sheet);
				} else if (shapeJson instanceof PictureJson) {
					_importPhase.setPhase(ImportPhase.PICTURE);
					importPicture((PictureJson) shapeJson, xSheet, sheet);
				}
			}
		}
	}
//		List<ZssChartX> poiCharts = new ArrayList<ZssChartX>();
//		List<Picture> poiPictures = new ArrayList<Picture>();
//
//		XSSFDrawing patriarch = null;
//		for (POIXMLDocumentPart dr : ((XSSFSheet) poiSheet).getRelations()) {
//			if (dr instanceof XSSFDrawing) {
//				patriarch = (XSSFDrawing) dr;
//				break;
//			}
//		}
//		if (patriarch != null) {
//			final CTDrawing ctdrawing = patriarch.getCTDrawing();
//			for (CTTwoCellAnchor anchor : ctdrawing.getTwoCellAnchorArray()) {
//				final CTMarker from = anchor.getFrom();
//				final CTMarker to = anchor.getTo();
//				XSSFClientAnchor clientAnchor = null;
//				if (from != null && to != null) {
//					clientAnchor = new XSSFClientAnchor((int) from.getColOff(),
//							(int) from.getRowOff(), (int) to.getColOff(),
//							(int) to.getRowOff(), from.getCol(), from.getRow(),
//							to.getCol(), to.getRow());
//				}
//				CTPicture ctPicture = anchor.getPic();
//				if (ctPicture == null) {
//					final CTGraphicalObjectFrame gfrm = anchor
//							.getGraphicFrame();
//					if (gfrm != null) {
//						final XSSFChartX chartX = createXSSFChartX(patriarch,
//								gfrm, clientAnchor);
//						poiCharts.add(chartX);
//					}
//				} else {
//					poiPictures.add(XSSFPictureHelper
//							.newXSSFPicture(patriarch, clientAnchor,
//									ctPicture));
//				}
//			}
//		}
//		importChart(poiCharts, poiSheet, sheet);
//		importPicture(poiPictures, poiSheet, sheet);
//	}
//
//	protected XSSFChartX createXSSFChartX(XSSFDrawing patriarch,
//			CTGraphicalObjectFrame gfrm, XSSFClientAnchor xanchor) {
//		//chart
//		final String name = gfrm.getNvGraphicFramePr().getCNvPr().getName();
//		final CTGraphicalObject gobj = gfrm.getGraphic();
//		final CTGraphicalObjectData gdata = gobj.getGraphicData();
//		String chartId = null;
//		// ZSS-358: must skip other XML nodes such as TEXT_NODE
//		Node child = gdata.getDomNode().getFirstChild();
//		while (child != null) {
//			if ("chart".equals(child.getLocalName())) {
//				chartId = child.getAttributes().getNamedItemNS(
//						"http://schemas.openxmlformats.org/officeDocument/2006/relationships",
//						"id").getNodeValue();
//				break;
//			}
//			child = child.getNextSibling();
//		}
//		final XSSFChartX chartX = new XSSFChartX(patriarch, xanchor, name,
//				chartId);
//		return chartX;
//	}

	// TODO handle group picture
	protected void importPicture(PictureJson pictureJson, XlsxExtractor.XlsxSheetExtractor xSheet,
			SSheet sheet) {
		final Integer picDataIx = importedPictureJson.get(pictureJson); //ZSS-735
		ViewAnchor viewAnchor = toViewAnchor(pictureJson, sheet);
		if (picDataIx != null) {
			sheet.addPicture(picDataIx, viewAnchor);
		} else {
			final SPicture.Format format = SPicture.Format.valueOfFileExtension(pictureJson.fileType);
			if (format != null) {
				SPicture pic = sheet
						.addPicture(format, pictureJson.getPictureData(), viewAnchor);
				importedPictureJson
						.put(pictureJson, pic.getPictureData().getIndex());
			} else {
				if (pictureJson.isExternal()) { //KEIKAI-636
					_logger.warn("ignore an external image at ['" + sheet.getSheetName()
						+ "'!" + new CellRegion(viewAnchor.getRowIndex(), viewAnchor.getColumnIndex()).getReferenceString() + "]");
				}
				// TODO log we ignore a picture with unsupported format
			}
		}
	}

	protected ViewAnchor toViewAnchor(ShapeJson shapeJson, SSheet sheet) {
		int width = getAnchorWidthInPx(shapeJson, sheet);
		int height = getAnchorHeightInPx(shapeJson, sheet);
		ViewAnchor viewAnchor = new ViewAnchor(shapeJson.getFromRow(),
				shapeJson.getFromCol(), width, height);
		viewAnchor.setXOffset(getXoffsetInPixel(shapeJson));
		viewAnchor.setYOffset(getYoffsetInPixel(shapeJson));
		viewAnchor.setAnchorType(shapeJson.getAnchorType());
		return viewAnchor;
	}

	/**
	 * Reference DefaultBookWidgetLoader.getXSSFWidthInPx()
	 */
	protected int getAnchorWidthInPx(ShapeJson shapeJson, SSheet sheet) {
		// first column
		final int firstColumn = shapeJson.getFromCol();
		final SColumn firstSColumn = sheet.getColumn(firstColumn);
		// KEIKAI-75
		boolean firstSColumnHidden = firstSColumn.isHidden();
		final int firstColumnWidth = firstSColumnHidden ? 0 : firstSColumn.getWidth();
		final int offsetInFirstColumn = firstSColumnHidden ? 0 : UnitUtil.emuToPx(shapeJson.getFromColOff());
		final int anchorWidthInFirstColumn = firstColumnWidth - offsetInFirstColumn;

		// last column
		final int lastColumn = shapeJson.getToCol();
		final int anchorWidthInLastColumn = sheet.getColumn(lastColumn).isHidden() ? 0 : UnitUtil.emuToPx(shapeJson.getToColOff());
		//in between
		int width = firstColumn == lastColumn ? anchorWidthInLastColumn - offsetInFirstColumn :
				anchorWidthInFirstColumn + anchorWidthInLastColumn;
		width = Math.abs(width); // just in case

		for (int j = firstColumn+1; j < lastColumn; ++j) {
			final SColumn scolumn = sheet.getColumn(j);
			// KEIKAI-75
			if (!scolumn.isHidden()) {
				width += scolumn.getWidth();
			}
		}

		return width;
	}

	/**
	 * DefaultBookWidgetLoader.getXSSFHeightInPx()
	 */
	protected int getAnchorHeightInPx(ShapeJson shapeJson, SSheet sheet) {
		//first row
		final int firstRow = shapeJson.getFromRow();
		final SRow firstSRow = sheet.getRow(firstRow);
		// KEIKAI-75
		boolean firstSRowHidden = firstSRow.isHidden();
		final int firstRowHeight = firstSRowHidden ? 0 : firstSRow.getHeight();
		int offsetInFirstRow = firstSRowHidden ? 0 : UnitUtil.emuToPx(shapeJson.getFromRowOff());
		final int anchorHeightInFirstRow = firstRowHeight - offsetInFirstRow;

		//last row
		final int lastRow = shapeJson.getToRow();
		int anchorHeightInLastRow = sheet.getRow(lastRow).isHidden() ? 0 : UnitUtil.emuToPx(shapeJson.getToRowOff());
		int height = lastRow == firstRow ? anchorHeightInLastRow - offsetInFirstRow :
				anchorHeightInFirstRow + anchorHeightInLastRow ;
		height = Math.abs(height); // just in case

		//add inter-row height
		for (int row = firstRow+1; row < lastRow; ++row) {
			final SRow srow = sheet.getRow(row);
			// KEIKAI-75
			if (!srow.isHidden()) {
				height += sheet.getRow(row).getHeight();
			}
		}

		return height;
	}

	protected int getXoffsetInPixel(ShapeJson shapeJson) {
		return UnitUtil.emuToPx(shapeJson.getFromColOff());
	}

	protected int getYoffsetInPixel(ShapeJson shapeJson) {
		return UnitUtil.emuToPx(shapeJson.getFromRowOff());
	}

	protected void importChart(ChartSpaceJson chartSpaceJson, XlsxExtractor.XlsxSheetExtractor xSheet,
							   SSheet sheet) {
		final ViewAnchor viewAnchor = toViewAnchor(chartSpaceJson, sheet);
		AbstractChartAdv chart = null;

		for (final ChartTypeJson json: chartSpaceJson.getChartTypes()) {
			final ChartType chartType = toChartType(json.getChartType());
			SChartData data;
			if (chartType != null) {
				if (chart == null) {
					chart = (AbstractChartAdv) sheet.addChart(chartType, viewAnchor);
					//ZSS-822
					importAxis(chartSpaceJson, chart);
					data = chart.getData();
				} else {
					data = chart.createChartData(chartType);
					chart.addChartData(data);
				}
				if (data instanceof SGeneralChartData) {
					final GeneralChartDataImpl generalChartData = (GeneralChartDataImpl) data;
					if (data instanceof AreaChartDataImpl) {
						final AreaChartDataImpl areaChartData = (AreaChartDataImpl) data;
						areaChartData.setGrouping(toChartGrouping(json.getGrouping()));
					} else if (data instanceof BarChartDataImpl) {
						final ChartGrouping chartGrouping = toChartGrouping(json.getGrouping());
						final BarChartDataImpl barChartData = (BarChartDataImpl) data;
						barChartData.setBarOverlap(toBarOverlap(json, chartGrouping)); //ZSS-830
						barChartData.setGrouping(chartGrouping);
//					} else if (data instanceof BubbleChartDataImpl) {
//					} else if (data instanceof DoughnutChartDataImpl) {
					} else if (data instanceof LineChartDataImpl) {
						final LineChartDataImpl lineChartData = (LineChartDataImpl) data;
						lineChartData.setSmooth(json.isSmooth());
						lineChartData.setMarker(json.hasMarker());
						lineChartData.setGrouping(toChartGrouping(json.getGrouping()));
//					} else if (data instanceof PieChartDataImpl) {
//					} else if (data instanceof ScatterChartDataImpl) {
					}
					final List<Long> axids = json.getAxIds();
					if (!axids.isEmpty()) {
						final long id = axids.get(0);
						chart.getCategoryAxises().stream().filter((v) -> v.getId() == id)
								.findFirst().ifPresent(generalChartData::setCategoryAxis);
						chart.getDateAxises().stream().filter((v) -> v.getId() == id)
								.findFirst().ifPresent(generalChartData::setDateAxis);
						final List<SChartAxis> valAxes = new ArrayList<>();
						for (long valId: axids) {
							chart.getValueAxises().stream()
									.filter((v) -> v.getId() == valId).findAny().ifPresent(valAxes::add);
						}
						generalChartData.setValueAxes(valAxes);
					}
					final List<? extends SeriesJson> seriesList = json.getSeriesList();
					if (seriesList != null) {
						for (final SeriesJson seriesJson: seriesList) {
							importSeries(seriesJson, generalChartData);
						}
					}
				}
			}
		}
		if (chart == null) {
			return;
		}

		chart.setSeriesColors(((AbstractBookAdv) book).getDefaultChartSeriesColors());

		// KEIKAI-359
		if (chartSpaceJson.hasAutoTitleDeleted()){
			chart.setAutoTitleDeleted(chartSpaceJson.isAutoTitleDeleted());
		}
		if (chartSpaceJson.hasTitle()) {
			chart.setShowTitle(true);
			chart.setTitle(chartSpaceJson.getTitle());
		} else {
			chart.setShowTitle(false);
		}
		chart.setThreeD(chartSpaceJson.isSetView3D());
		//ZSS-830
		if (chart.isThreeD()) {
			final View3DJson view3DJson = chartSpaceJson.getView3D();
			chart.setRotX(view3DJson.getRotX());
			chart.setRotY(view3DJson.getRotY());
			chart.setPerspective(view3DJson.getPerspective());
			chart.setHPercent(view3DJson.getHPercent());
			chart.setDepthPercent(view3DJson.getDepthPercent());
			chart.setRightAngleAxes(view3DJson.isRightAngleAxes());
		}
		final LegendStyle legendStyle = chartSpaceJson.getLegendStyle();
		if (legendStyle != null) {
			chart.setLegendPosition(toChartLegendPosition(legendStyle.getLegendPosition()));
		}

		chart.setEmptyAsGaps("gap".equals(chartSpaceJson.getDispBlanksAs()));
		chart.setShowDataLabelsOverMax(chartSpaceJson.isShowDLblsOverMax());
		chart.setPlotOnlyVisibleCells(chartSpaceJson.isPlotOnlyVisibleCells());
	}

	public ChartGrouping toChartGrouping(String grouping) {
		if (grouping == null) {
			return ChartGrouping.STANDARD;
		}
		switch (grouping) {
			case "stacked":
				return ChartGrouping.STACKED;
			case "percentStacked":
				return ChartGrouping.PERCENT_STACKED;
			case "clustered":
				return ChartGrouping.CLUSTERED;
			case "standard":
			default:
				return ChartGrouping.STANDARD;
		}
	}

	protected int toBarOverlap(ChartTypeJson json, ChartGrouping chartGrouping) {
		final Integer barOverlap = json.getBarOverlap();
		if (barOverlap != null) {
			return barOverlap;
		}
		switch (chartGrouping) {
			case STACKED:
			case PERCENT_STACKED:
				return 100;
			default:
				return 0;
		}
	}

	protected SChart.ChartLegendPosition toChartLegendPosition(String legendPosition) {
		if (legendPosition == null) {
			return null;
		}
		switch (legendPosition) {
			case "l":
				return SChart.ChartLegendPosition.LEFT;
			case "t":
				return SChart.ChartLegendPosition.TOP;
			case "tr":
				return SChart.ChartLegendPosition.TOP_RIGHT;
			case "b":
				return SChart.ChartLegendPosition.BOTTOM;
			case "r":
				return SChart.ChartLegendPosition.RIGHT;
			default:
				return null;
		}
	}

	protected AbstractChartAxisAdv.Position toChartAxisPosition(String legendPosition) {
		if (legendPosition == null) {
			return null;
		}
		switch (legendPosition) {
			case "l":
				return AbstractChartAxisAdv.Position.LEFT;
			case "t":
				return AbstractChartAxisAdv.Position.TOP;
			case "r":
				return AbstractChartAxisAdv.Position.RIGHT;
			case "b":
				return AbstractChartAxisAdv.Position.BOTTOM;
			default:
				return null;
		}
	}

	protected AbstractChartAxisAdv.Orientation toChartAxisOrientation(String orientation) {
		if (orientation == null) {
			return null;
		}
		switch (orientation) {
			case "maxMin":
				return AbstractChartAxisAdv.Orientation.MAX_MIN;
			case "minMax":
				return AbstractChartAxisAdv.Orientation.MIN_MAX;
			default:
				return null;
		}
	}

	protected AbstractChartAxisAdv.TickLabelPosition toChartAxisTickLabelPosition(String tickLabelPosition) {
		if (tickLabelPosition == null) {
			return null;
		}
		switch (tickLabelPosition) {
			case "low":
				return AbstractChartAxisAdv.TickLabelPosition.LOW;
			case "none":
				return AbstractChartAxisAdv.TickLabelPosition.NONE;
			case "high":
				return AbstractChartAxisAdv.TickLabelPosition.HIGH;
			case "nextTo":
				return AbstractChartAxisAdv.TickLabelPosition.NEXT_TO;
			default:
				return null;
		}
	}

	protected AbstractChartAxisAdv.Crosses toChartAxisCrosses(String crosses) {
		if (crosses == null) {
			return null;
		}
		switch (crosses) {
			case "max":
				return AbstractChartAxisAdv.Crosses.MAX;
			case "min":
				return AbstractChartAxisAdv.Crosses.MIN;
			case "autoZero":
				return AbstractChartAxisAdv.Crosses.AUTO_ZERO;
			default:
				return null;
		}
	}

	protected CategoryAxisImpl.LabelAlign toChartAxisLabelAlign(String labelAlign) {
		if (labelAlign == null) {
			return null;
		}
		switch (labelAlign) {
			case "l":
				return CategoryAxisImpl.LabelAlign.LEFT;
			case "r":
				return CategoryAxisImpl.LabelAlign.RIGHT;
			case "ctr":
				return CategoryAxisImpl.LabelAlign.CENTER;
			default:
				return null;
		}
	}

	protected DateAxisImpl.TimeUnit toChartAxisTimeUnit(String timeUnit) {
		if (timeUnit == null) {
			return null;
		}
		switch (timeUnit) {
			case "days":
				return DateAxisImpl.TimeUnit.DAYS;
			case "months":
				return DateAxisImpl.TimeUnit.MONTHS;
			case "years":
				return DateAxisImpl.TimeUnit.YEARS;
			default:
				return null;
		}
	}

	protected ValueAxisImpl.CrossBetween toChartAxisCrossBetween(String crossBetween) {
		if (crossBetween == null) {
			return null;
		}
		switch (crossBetween) {
			case "between":
				return ValueAxisImpl.CrossBetween.BETWEEN;
			case "midCat":
				return ValueAxisImpl.CrossBetween.MIDPOINT_CATEGORY;
			default:
				return null;
		}
	}

	protected void importCategory(CategorySeriesJson categorySeriesJson, SGeneralChartData chartData) {
		chartData.setCategoriesFormula(categorySeriesJson.getCategoryFormula());
	}

	/**
	 * reference ChartHelper.prepareCategoryModel()
	 * Category normally indicates the values show on X axis.
	 * Tile indicates the name for each series.
	 */
	protected void importSeries(SeriesJson seriesJson, SGeneralChartData chartData) {
		switch (seriesJson.getChartType()) {
			case AREA:
			case AREA_3D:
			case BAR:
			case BAR_3D:
			case COLUMN:
			case COLUMN_3D:
			case DOUGHNUT:
			case PIE:
			case PIE_3D:
//		    case RADAR:
			{
				importCategorySeries((CategorySeriesJson) seriesJson, chartData, false, false);
				break;
			}
			case LINE:
			{
				importCategorySeries((CategorySeriesJson) seriesJson, chartData, true, true);
				break;
			}
			case LINE_3D:
			{
				importCategorySeries((CategorySeriesJson) seriesJson, chartData, true, false);
				break;
			}
			case BUBBLE: {
				importXyzSeries((XYZSeriesJson) seriesJson, chartData);
				break;
			}
			case SCATTER: {
				importXySeries((XYSeriesJson) seriesJson, chartData);
				break;
			}
//        case Stock:
//          categoryData = new XSSFStockChartData(xssfChart); TODO XSSFStockChartData contains wrong implementation
//          break;
			default:
				// TODO ignore unsupported charts
		}
	}

	protected AbstractSeriesAdv createSeries(SeriesJson json, SGeneralChartData chartData) {
		final AbstractSeriesAdv series = (AbstractSeriesAdv) chartData.addSeries();
		series.setOrder(json.getOrder());
		return series;
	}

	protected void importCategorySeries(CategorySeriesJson categorySeriesJson, SGeneralChartData chartData,
									  boolean isLine, boolean checkSmooth) {
		final AbstractSeriesAdv series = createSeries(categorySeriesJson, chartData);
		series.setFormula(getTitleFormula(categorySeriesJson),
				categorySeriesJson.getValueFormula());
		final DataPointJson dataPointJson = categorySeriesJson.getDataPoint();
		final XlsxExtractor.XlsxSchemeClrExtractor colorExtractor = isLine ?
				dataPointJson.getLineColor() : dataPointJson.getFillColor();
		series.setDefaultColor(getSchemeHTMLColor(colorExtractor, themes));
		if (checkSmooth) {
			series.setSmooth(categorySeriesJson.isSmooth());
		}
		// KEIKAI-579
		series.setDataLabelsVisible(categorySeriesJson.isDataLabelsVisible());
		final Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> map = isLine ?
				dataPointJson.getDataPointLineColors() : dataPointJson.getDataPointFillColors();
		map.forEach((key, val) -> {
			final SColor color = getSchemeHTMLColor(val, themes);
			if (color != null) {
				series.addDataPointColors(key, color);
			}
		});
		if (chartData.getCategoriesFormula() == null) {
			importCategory(categorySeriesJson, chartData);
		}
	}

	protected void importXySeries(XYSeriesJson xySeriesJson, SGeneralChartData chartData) {
		final AbstractSeriesAdv series = createSeries(xySeriesJson, chartData);
		final DataPointJson dataPointJson = xySeriesJson.getDataPoint();
		final XlsxExtractor.XlsxSchemeClrExtractor colorExtractor = dataPointJson.getMarkerColor();
		series.setDefaultColor(getSchemeHTMLColor(colorExtractor, themes));
		final Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> map = dataPointJson.getDataPointMarkerColors();
		map.forEach((key, val) -> {
			final SColor color = getSchemeHTMLColor(val, themes);
			if (color != null) {
				series.addDataPointColors(key, color);
			}
		});
		series.setXYFormula(getTitleFormula(xySeriesJson), xySeriesJson.getXFormula(), xySeriesJson.getYFormula());
	}

	/**
	 * reference ChartHelper.prepareXYZModel()
	 *
	 * @param xyzSeriesJson
	 * @param chartData
	 */
	protected void importXyzSeries(XYZSeriesJson xyzSeriesJson, SGeneralChartData chartData) {
		final AbstractSeriesAdv series = createSeries(xyzSeriesJson, chartData);
		// reference to ChartHelper.prepareTitle()
		final String nameExpression = getTitleFormula(xyzSeriesJson);
		final String xValueExpression = xyzSeriesJson.getXFormula();
		final String yValueExpression = xyzSeriesJson.getYFormula();
		final String zValueExpression = xyzSeriesJson.getZFormula();
		// KEIKAI-234: ZK Charts bug
		final DataPointJson dataPointJson = xyzSeriesJson.getDataPoint();
		final XlsxExtractor.XlsxSchemeClrExtractor colorExtractor = dataPointJson.getFillColor();
		series.setDefaultColor(getSchemeHTMLColor(colorExtractor, themes));
		final Map<Integer, XlsxExtractor.XlsxSchemeClrExtractor> map = dataPointJson.getDataPointFillColors();
		map.forEach((key, val) -> series.addDataPointColors(key, getSchemeHTMLColor(val, themes)));
		series.setXYZFormula(nameExpression, xValueExpression, yValueExpression, zValueExpression);
	}

	//ZSS-822
	protected void importAxis(ChartSpaceJson chartSpaceJson, SChart chart) {
		final List<CategoryAxisJson> categoryAxises = chartSpaceJson.getCategoryAxises();
		final List<DateAxisJson> dateAxisJsons = chartSpaceJson.getDateAxises();
		final List<ValueAxisJson> valueAxises = chartSpaceJson.getValueAxises();
		if (categoryAxises != null) {
			for (CategoryAxisJson categoryAxis : categoryAxises) {
				final CategoryAxisImpl saxis = new CategoryAxisImpl(categoryAxis.getId(),
						toChartAxisPosition(categoryAxis.getPosition()),
						categoryAxis.getMinimum(),
						categoryAxis.getMaximum(),
						categoryAxis.getFormatCode(),
						toChartAxisOrientation(categoryAxis.getOrientation()),
						toChartAxisTickLabelPosition(categoryAxis.getTickLabelPosition()),
						categoryAxis.getCrossAx(),
						toChartAxisCrosses(categoryAxis.getCrosses()),
						toChartAxisLabelAlign(categoryAxis.getLabelAlign()),
						categoryAxis.getLabelOffset(),
						categoryAxis.hasMajorGridlines(),
						!categoryAxis.isDelete());
				chart.addCategoryAxis(saxis);
			}
		}
		if (dateAxisJsons != null) {
			for (DateAxisJson dateAxis : dateAxisJsons) {
				final SChartAxis saxis = new DateAxisImpl(dateAxis.getId(),
						toChartAxisPosition(dateAxis.getPosition()),
						dateAxis.getMinimum(),
						dateAxis.getMaximum(),
						dateAxis.getFormatCode(),
						toChartAxisOrientation(dateAxis.getOrientation()),
						toChartAxisTickLabelPosition(dateAxis.getTickLabelPosition()),
						dateAxis.getCrossAx(),
						toChartAxisCrosses(dateAxis.getCrosses()),
						dateAxis.hasMajorGridlines(),
						!dateAxis.isDelete(),
						dateAxis.getAuto(),
						dateAxis.getLabelOffset(),
						toChartAxisTimeUnit(dateAxis.getBaseTimeUnit()),
						dateAxis.getMajorUnit(),
						toChartAxisTimeUnit(dateAxis.getMajorTimeUnit()),
						dateAxis.getMinorUnit(),
						toChartAxisTimeUnit(dateAxis.getMinorTimeUnit())
				);
				chart.addDateAxis(saxis);
			}
		}
		if (valueAxises != null) {
			for (ValueAxisJson valueAxis : valueAxises) {
				final SChartAxis saxis = new ValueAxisImpl(valueAxis.getId(),
						toChartAxisPosition(valueAxis.getPosition()),
						valueAxis.getMinimum(),
						valueAxis.getMaximum(),
						valueAxis.getFormatCode(),
						toChartAxisOrientation(valueAxis.getOrientation()),
						toChartAxisTickLabelPosition(valueAxis.getTickLabelPosition()),
						valueAxis.getCrossAx(),
						toChartAxisCrosses(valueAxis.getCrosses()),
						toChartAxisCrossBetween(valueAxis.getCrossBetween()),
						valueAxis.hasMajorGridlines(),
						!valueAxis.isDelete());
				chart.addValueAxis(saxis);
			}
		}
	}

	/**
	 * return a formula or generate a default title ("Series[N]")if title doesn't exist.
	 * reference ChartHelper.prepareTitle()
	 *
	 * @param seriesJson
	 * @return
	 */
	protected String getTitleFormula(SeriesJson seriesJson) {
		if (seriesJson.hasTitle()) {
			if (seriesJson.isReference()) {
				return seriesJson.getTitleFormula();
			} else {
				return "\"" + seriesJson.getTitleText() + "\"";
			}
		}
		return "\"\"";
	}

	protected ChartType toChartType(io.keikai.importer.ChartType chartType) {
		switch (chartType) {
			case AREA:
			case AREA_3D:
				return ChartType.AREA;
			case BAR:
			case BAR_3D:
				return ChartType.BAR;
			case COLUMN:
			case COLUMN_3D:
				return ChartType.COLUMN;
			case DOUGHNUT:
				return ChartType.DOUGHNUT;
			case LINE:
			case LINE_3D:
				return ChartType.LINE;
			case PIE:
			case PIE_3D:
				return ChartType.PIE;
			case BUBBLE:
				return ChartType.BUBBLE;
			case SCATTER:
				return ChartType.SCATTER;
			default:
				// TODO ignore unsupported charts
				return null;
		}
	}

	/**
	 * [ISO/IEC 29500-1 1st Edition] 18.3.1.13 col (Column Width & Formatting)
	 * By experiments, CT_Col is always created in ascending order by min and the range specified by min & max doesn't overlap each other.<br/>
	 * For example:
	 * <x:cols xmlns:x="http://schemas.openxmlformats.org/spreadsheetml/2006/main">
	 * <x:col min="1" max="1" width="11.28515625" customWidth="1" />
	 * <x:col min="2" max="4" width="11.28515625" hidden="1" customWidth="1" />
	 * <x:col min="5" max="5" width="11.28515625" customWidth="1" />
	 * </x:cols>
	 */
	protected void importColumn(XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet) {
		List<XlsxExtractor.XlsxColExtractor> cols = xSheet.getCols();
		if (cols.isEmpty()) {
			return;
		}

		//ZSS-1132
		final int charWidth = ((AbstractBookAdv) book).getCharWidth();

		//ZSS-952: xssf's default column width includes the 5 pixels padding and is in charwidth unit.
		//		int defaultWidth = UnitUtil.defaultColumnWidthToPx(poiSheet.getDefaultColumnWidth(), CHRACTER_WIDTH);
		double defaultColWidth = getXlsxDefaultColumnWidth(xSheet, charWidth);
		int defaultWidth = UnitUtil.xssfDefaultColumnWidthToPx(defaultColWidth, charWidth);

		_importPhase.setPhase(ImportPhase.COL);

		//KEIKAI-424, setupColumnArray() requires to set column in ascending order of "min"
		Collections.sort(cols, Comparator.comparingInt(XlsxExtractor.XlsxColExtractor::getMin));
		for (XlsxExtractor.XlsxColExtractor col : cols) {
			//max is 16384
			int min = col.getMin(), max = col.getMax();

			int columnIndex = min - 1;
			_importPhase.setColIndex(columnIndex);

			SColumnArray columnArray = sheet
					.setupColumnArray(columnIndex, max - 1);

			boolean hidden = col.isHidden();
			columnArray.setHidden(hidden);
			Number colWdh = col.getWidth();
			if (colWdh == null) {
				colWdh = defaultColWidth * 256;
			} else {
				colWdh = colWdh.doubleValue() * 256;
			}

			int width = colWdh.doubleValue() == defaultColWidth * 256 ?
					UnitUtil.xssfDefaultColumnWidthToPx(colWdh.doubleValue() / 256.0, charWidth):
					UnitUtil.fileChar256ToPx(colWdh.intValue(), charWidth);
			if (width > 0 && width != defaultWidth) {
				//when CT_Col is hidden with default width, We don't import the width for it's 0.
				columnArray.setWidth(width);
			}
			columnArray.setCustomWidth(
					col.isCustomWidth() || width != defaultWidth); //ZSS-896
			int style = col.getStyle();
			if (style >= 0) {
				columnArray.setCellStyle(importCellXf(cellXfsCache.getCellXfAt(style), false));
			}

			// KEIKAI-555
			if (col.isBestFit()) {
				columnArray.setBestFit(true);
			}

			Boolean collapsed = col.isCollapsed();
			if (collapsed != null) {
				columnArray.setCollapsed(collapsed);
			}

			Number outlineLevel = col.getOutlineLevel();
			if (outlineLevel != null) {
				columnArray.setOutlineLevel(outlineLevel.intValue());
			}
		}
	}

	protected double getXlsxDefaultColumnWidth(
			XlsxExtractor.XlsxSheetExtractor extractor, int charWidth) {
		XlsxExtractor.XlsxSheetFormatPrExtractor sheetFormatPr = extractor
				.getSheetFormatPr();
		final double XSSF_DEFAULT_COL_WIDTH = 9.142857; // base on font Calibri 11
		if (sheetFormatPr == null) {
			return XSSF_DEFAULT_COL_WIDTH;
		} else {
			Number defaultColWidth = sheetFormatPr.getDefaultColWidth();
			if (defaultColWidth != null) {
				return defaultColWidth.doubleValue();
			} else {
				return sheetFormatPr.isSetBaseColWidth() ? geDefaultColumnWidthPerBaseColWidth(sheetFormatPr.getBaseColWidth(), charWidth) : XSSF_DEFAULT_COL_WIDTH;
			}
		}
	}

	//ZSS-1132: default font size will affect the CHAR_WIDTH
	//@since 3.9.9
	//20151015, henrichen:
	public double geDefaultColumnWidthPerBaseColWidth(int baseWidth, int charWidth) {
		final int defaultPx = UnitUtil.defaultColumnWidthToPx(baseWidth, charWidth);
		return UnitUtil.fileChar256ToPx(defaultPx, charWidth);
	}

	protected SRow importRow(XlsxExtractor.XlsxRowExtractor xRow, SSheet sheet) {
		final int rowIndex = xRow.getIndex();
		_importPhase.setRowIndex(rowIndex);
		SRow row = sheet.getRow(rowIndex);
		float rh = xRow.getRowHeight();

		// row height measured in twips (1/20th of a point), see XSSFRow.java#getHight()
		if (rh != -1) { // KEIKAI-658: only set when exist
			row.setHeight(UnitUtil.twipToPx((short)(rh * 20)));
		}
		Boolean isCustomHeight = xRow.isCustomHeight();
		if (isCustomHeight != null) {
			row.setCustomHeight(isCustomHeight);
		}
		Boolean isHidden = xRow.isHidden();
		if (isHidden != null) {
			row.setHidden(isHidden);
		}
		Number outlineLevel = xRow.getOutlineLevel();
		if (outlineLevel != null) {
			row.setOutlineLevel(outlineLevel.intValue());
		}
		Boolean collapsed = xRow.isCollapsed();
		if (collapsed != null) {
			row.setCollapsed(collapsed);
		}

		int sIndex = xRow.getStyle();
		if (sIndex > -1) {
			row.setCellStyle(importCellXf(cellXfsCache.getCellXfAt(sIndex), false));
		}

		_importPhase.setPhase(ImportPhase.CELL);
		for (XlsxExtractor.XlsxCellExtractor xCell : xRow.getCells()) {
			importCell(xCell, xRow.getIndex(), sheet);
		}
		_importPhase.setPhase(ImportPhase.ROW);

		return row;
	}

	protected SCell importCell(XlsxExtractor.XlsxCellExtractor xCell, int row, SSheet sheet) {
		int cellIndex = xCell.index;
		_importPhase.setColIndex(cellIndex);
		SCell cell = sheet.getCell(row, cellIndex);
		int sIndex = xCell.getStyle();
		if (sIndex > -1) {
			cell.setCellStyle(importCellXf(cellXfsCache.getCellXfAt(sIndex), true));
		}

		XlsxExtractor.XlsxFormulaExtractor xCellFormula = xCell.getFormula();

		// convert to ZSS type
		final String xCellType = xCell.getType();
		final String cellType = xCellFormula != null ? "f" : xCellType;

		// KEIKAI-217
		if (cellType != null) {
			switch (cellType) {
				case "n":
					cell.setNumberValue(Double.parseDouble((String) xCell.getValue()));
					break;
				case "inlineStr":
					XlsxExtractor.XlsxRichTextExtractor xRichText = xCell
							.getRichTextInline();
					if (xRichText != null) {
						SRichText richText = cell.setupRichTextValue();
						importRichText(xRichText, richText);
					}
					break;
				case "s":
				case "str":
					Object xCellValue = xCell.getValue();
					if (xCellValue instanceof XlsxExtractor.XlsxRichTextExtractor) {
						SRichText richText = cell.setupRichTextValue();
						importRichText((XlsxExtractor.XlsxRichTextExtractor) xCellValue, richText);
					} else {
						cell.setStringValue((String) xCellValue);
					}
					break;
				case "b":
					cell.setBooleanValue(toBoolean(xCell.getValue(), false));
					break;
				case "f":
					Object formulaValue = xCellFormula.getValue();
					final String type = xCellFormula.getType();
					final FormulaType formulaType = "array".equals(type) ? FormulaType.ARRAY : FormulaType.CELL; 
					if (formulaValue instanceof String) {
						if (formulaType != FormulaType.ARRAY) {
							cell.setFormulaValue((String) formulaValue);
						} else {
							// array formula
							sheet.setArrayFormula((String) formulaValue, new CellRegion(xCellFormula.getRef()));
						}
						if (_logger.isDebugEnabled()) {
							_logger.debug("imported a formula: {} at {}",
									formulaValue,
									new SheetRegion(sheet, cell.getRowIndex(),
											cell.getColumnIndex()).getReferenceString());
						}
					} else if (formulaValue != null) {
						String ref0 = ((Pair<String, String>) formulaValue).x;
						String formula = ((Pair<String, String>) formulaValue).y;
						if (formulaType != FormulaType.ARRAY) {
							Ref ref1 = Converter.a1ToRef(ref0);
							ParsingBook fpb = new ParsingBook(book);
							SharedFormula sf = new SharedFormula(
									SpreadsheetVersion.EXCEL2007);
							FormulaParseContext formulaCtx = new FormulaParseContext(
									book, sheet, cell, sheet.getSheetName(),
									null, Locale.US, formulaType);
							FormulaEngine fe = EngineFactory.getInstance().createFormulaEngine();
							Ptg[] ptgs = _sharedFormulaPtgArray.get(formula);
							if (ptgs == null) {
								ptgs = ((FormulaEngineImpl) fe).parse(formula,
										fpb, book.getSheetIndex(sheet),
										formulaCtx);
								_sharedFormulaPtgArray.put(formula, ptgs);
							}
							Ptg[] fmla = sf.convertSharedFormulas(ptgs, row - ref1.getTop(),
									cellIndex - ref1.getLeft());
							formula = FormulaRenderer.toFormulaString(fpb, fmla);
							cell.setFormulaValue(formula);
							if (_logger.isDebugEnabled()) {
								_logger.debug(
										"imported a shared formula: {} at {}",
										formula, new SheetRegion(sheet, cell.getRowIndex(),
												cell.getColumnIndex()).getReferenceString());
							}
						} else {
							// array formula
							sheet.setArrayFormula(formula, new CellRegion(ref0));
						}
					} else {
						// KEIKAI-398:
						// It could be array formula without any shared info.
						// And ZSS (KEIKAI 5.0) doesn't support array formula.
						// So, we use Excel's cache value to display as cell value.
						cell.setNumberValue(Double.parseDouble((String) xCell.getValue()));
						if (_logger.isDebugEnabled()) {
							_logger.debug(
									"unsupported formula, imported cell cache instead at {}",
									new SheetRegion(sheet, cell.getRowIndex(),
											cell.getColumnIndex()).getReferenceString());
						}
					}
					//ZSS-873
					if (isImportCache() && /*
			TODO: CT_Worksheet>ml:sheetCalcPr>fullCalcOnLoad
			!xCell.isCalcOnLoad() &&
			*/ !mustCalc(cell) && xCellType != null) {
						ValueEval val = null;
						switch (xCellType) {
							case "n":
								val = new NumberEval(Double.parseDouble((String) xCell.getValue()));
								break;
							case "inlineStr":
								XlsxExtractor.XlsxRichTextExtractor richTextInline0 = xCell
										.getRichTextInline();
								if (richTextInline0 != null) {
									SRichText richText = new RichTextImpl();
									importRichText(richTextInline0, richText);
									val = new StringEval(richText.getText());
								}
								break;
							case "s":
							case "str":
								Object xCellValue0 = xCell.getValue();
								if (xCellValue0 instanceof XlsxExtractor.XlsxRichTextExtractor) {
									SRichText richText = new RichTextImpl();
									importRichText((XlsxExtractor.XlsxRichTextExtractor) xCellValue0, richText);
									val = new StringEval(richText.getText());
								} else {
									// KEIKAI-181: if xCell contains formula, we should use "" instead of null,
									// which is the same as XSSFCell#getRichStringCellValue()
									if (xCellValue0 == null && xCell.getFormula() != null)
										xCellValue0 = "";

									val = new StringEval((String) xCellValue0);
								}
								break;
							case "b":
								val = BoolEval.valueOf(Boolean.parseBoolean((String) xCell.getValue()));
								break;
							case "e":
								val = ErrorEval.valueOf(getErrorCode((String) xCell.getValue()));
								break;
							default:
								// do nothing
						}
						if (val != null) {
							((AbstractCellAdv) cell).setFormulaResultValue(val);
						}
					}
					break;
				case "e":
					cell.setErrorValue(
							PoiEnumConversion.toErrorCode(getErrorCode((String) xCell.getValue())));
					break;
				default:
					// treat as number type.
					Object val = xCell.getValue();
					if (val != null) {
						cell.setNumberValue(Double.parseDouble((String) val));
					}
					if (_logger.isDebugEnabled()) {
						_logger.debug(
								"treat a cell type: {} as a number value: {}",
								cellType, val);
					}
					break;
			}
		}
		if (_logger.isDebugEnabled()) {
			_logger.debug("imported a cell type: {} at {}", cellType,
					new SheetRegion(sheet, cell.getRowIndex(), cell.getColumnIndex()).getReferenceString());
		}

		if (hyperlinksCache != null && !hyperlinksCache.isEmpty()) {
			XlsxExtractor.XlsxHyperlinkExtractor hyperlinkExtractor = hyperlinksCache
					.get(Converter.rangeToA1(row, cellIndex));
			if (hyperlinkExtractor != null) {
				String addr = hyperlinkExtractor.getLocation();
				String label = hyperlinkExtractor.getDisplay();
				SHyperlink hyperlink = cell.setupHyperlink(hyperlinkExtractor.getType(),
						addr == null ? "" : addr, label == null ? "" : label);
				cell.setHyperlink(hyperlink);
			}
		}

		if (commentsCache != null && !commentsCache.isEmpty()) {
			XlsxExtractor.XlsxCommentExtractor xComment = commentsCache
					.get(Converter.rangeToA1(row, cellIndex));
			if (xComment != null) {
				SComment comment = cell.setupComment();
				comment.setAuthor(xComment.getAuthorName());
				comment.setVisible(xComment.isVisible());

				XlsxExtractor.XlsxRichTextExtractor richTextExtractor = xComment.getText();
				if (richTextExtractor != null) {

					List<XlsxExtractor.XlsxREItExtractor> runs = richTextExtractor
							.getRuns();
					if (!runs.isEmpty()) {
						importRichText(richTextExtractor, comment.setupRichText());
					} else {
						comment.setText(richTextExtractor.getBaseText());
					}
				} else {
					comment.setText(richTextExtractor.getBaseText());
				}
			}
		}

		return cell;
	}

	/** <b>#NULL!</b>  - Intersection of two cell ranges is empty */
	public static final int ERROR_NULL = 0x00;
	/** <b>#DIV/0!</b> - Division by zero */
	public static final int ERROR_DIV_0 = 0x07;
	/** <b>#VALUE!</b> - Wrong type of operand */
	public static final int ERROR_VALUE = 0x0F;
	/** <b>#REF!</b> - Illegal or deleted cell reference */
	public static final int ERROR_REF = 0x17;
	/** <b>#NAME?</b> - Wrong function or range name */
	public static final int ERROR_NAME = 0x1D;
	/** <b>#NUM!</b> - Value range overflow */
	public static final int ERROR_NUM = 0x24;
	/** <b>#N/A</b> - Argument or function not available */
	public static final int ERROR_NA = 0x2A;

	//ZSS-810
	public static final int ERROR_FORMULA = 0x7f;

	/**
	 * @return Standard Excel error literal for the specified error code.
	 * @throws IllegalArgumentException if the specified error code is not one of the 7
	 * standard error codes
	 */
	public static final String getErrorText(int errorCode) {
		switch(errorCode) {
		case ERROR_NULL:  return "#NULL!";
		case ERROR_DIV_0: return "#DIV/0!";
		case ERROR_VALUE: return "#VALUE!";
		case ERROR_REF:   return "#REF!";
		case ERROR_NAME:  return "#NAME?";
		case ERROR_NUM:   return "#NUM!";
		case ERROR_NA:    return "#N/A";
		case ERROR_FORMULA: return "#ERROR!";
		}
		throw new IllegalArgumentException("Bad error code (" + errorCode + ")");
	}
	public static final byte getErrorCode(String errorText) {
		switch(errorText) {
		case "#NULL!": return ERROR_NULL;
		case "#DIV/0!": return ERROR_DIV_0;
		case "#VALUE!": return ERROR_VALUE;
		case "#REF!": return ERROR_REF;
		case "#NAME?": return ERROR_NAME;
		case "#NUM!": return ERROR_NUM;
		case "#N/A": return ERROR_NA;
		case "#ERROR!": return ERROR_FORMULA;
		}
		throw new IllegalArgumentException("Bad error text (" + errorText + ")");

	}
	//ZSS-873
	//Must evaluate INDIRECT() function to make the dependency table)
	//Issue845Test-checkIndirectNameRange
	protected boolean mustCalc(SCell cell) {
		FormulaExpression val = ((AbstractCellAdv) cell).getFormulaExpression();
		for (Ptg ptg : val.getPtgs()) {
			if (ptg instanceof FuncVarPtg
					&& ((FuncVarPtg) ptg).getFunctionIndex()
					== 148) { //148 is INDIRECT
				return true;
			}
		}
		return false;
	}

	protected void importRichText(XlsxExtractor.XlsxRichTextExtractor xRichText, SRichText richText) {
		//ZSS-1138, ZSS-1189
		List<XlsxExtractor.XlsxREItExtractor> runs = xRichText.getRuns();

		String cellValue = xRichText.getBaseText();
		richText.addSegment(cellValue != null ? cellValue : "",
				null); // ZSS-1138: null font means use cell's font

		for (XlsxExtractor.XlsxREItExtractor run : runs) {
			XlsxExtractor.XlsxFontExtractor runFont = run.getFont();
			richText.addSegment(run.getText(), runFont == null ? null :
					Optional.ofNullable(fontsCache.getFont(runFont)).orElse(createZssFont(runFont)));
		}
	}

	//ZSS-1140
	protected SFont createDxfZssFont(XlsxExtractor.XlsxFontExtractor xFont) {
		final AbstractFontBuilderAdv fb = (AbstractFontBuilderAdv) book.createFont();
		// font
		if (xFont.isOverrideName()) {
			fb.overrideName(true).name(xFont.getName());
		}
		if (xFont.isOverrideBold()) {
			fb.overrideBold(true).boldweight(PoiEnumConversion.toBoldweight((short) (xFont.isBold() ? 700 : 400)));
		}
		if (xFont.isOverrideItalic()) {
			fb.overrideItalic(true).italic(xFont.isItalic());
		}
		if (xFont.isOverrideStrike()) {
			fb.overrideStrikeout(true).strikeout(xFont.isStrike());
		}
		if (xFont.isOverrideUnderline()) {
			fb.overrideUnderline(true).underline(toUnderline(xFont.getUnderline()));
		}

		Number size = xFont.getSize();
		if (size != null) {
			fb.overrideHeightPoints(true).heightPoints(size.intValue());
		} else {
			fb.heightPoints(220 / 20); // same as XSSFont
		}

		if (xFont.isOverrideVertAlign()) {
			fb.overrideTypeOffset(true).typeOffset(toTypeOffset(xFont.getVertAlign()));
		}
		if (xFont.isOverrideColor()) {
			fb.overrideColor(true).color(getFontHTMLColor(xFont.getColor(), themes, this));
		}
		return fb.build();
	}

	protected SFont createZssFont(XlsxExtractor.XlsxFontExtractor xFont) {
		final Number size = xFont.getSize();
		return book.addFont()
				.name(xFont.getName())
				.boldweight(PoiEnumConversion.toBoldweight((short) (xFont.isBold() ? 700 : 400)))
				.italic(xFont.isItalic())
				.strikeout(xFont.isStrike())
				.underline(toUnderline(xFont.getUnderline()))
				.heightPoints(size != null ? size.intValue() : 220 / 20) // same as XSSFont
				.typeOffset(toTypeOffset(xFont.getVertAlign()))
				.color(getFontHTMLColor(xFont.getColor(), themes, this)).build();
	}
	protected static String getHTMLColorFromHex(String colorHax) {
		byte[] argb = UnitUtil.hexToArgbColor(colorHax);
		return argb.length > 3 ? "#" + toHex(argb[1]) + toHex(argb[2]) + toHex(argb[3]) : "#" + toHex(argb[0]) + toHex(argb[1]) + toHex(argb[2]);
	}

	public static String getBorderHTMLColor(XlsxExtractor.XlsxColorExtractor color,
										  XlsxExtractor.XlsxThemesExtractor themes, XlsxImporter xlsxImporter) {
		//reference BookHelper.colorToBorderHTML()
		return getHTMLColor(color, themes, xlsxImporter, "#000000");
	}

	public static String getFontHTMLColor(XlsxExtractor.XlsxColorExtractor color,
												XlsxExtractor.XlsxThemesExtractor themes, XlsxImporter xlsxImporter) {
		//reference BookHelper.colorToFontHTML()
		return getHTMLColor(color, themes, xlsxImporter, "#000000");
	}

	public static String getForegroundHTMLColor(XlsxExtractor.XlsxColorExtractor color,
			XlsxExtractor.XlsxThemesExtractor themes, XlsxImporter xlsxImporter) {
		//reference BookHelper.colorToForegroundHTML()
		return getHTMLColor(color, themes, xlsxImporter, "#000000");
	}

	public static String getBackgroundHTMLColor(XlsxExtractor.XlsxColorExtractor color,
										  XlsxExtractor.XlsxThemesExtractor themes, XlsxImporter xlsxImporter) {
		//reference BookHelper.colorToBackgroundHTML()
		return getHTMLColor(color, themes, xlsxImporter, "#ffffff");
	}

	protected static String getHTMLColor(XlsxExtractor.XlsxColorExtractor color,
										  XlsxExtractor.XlsxThemesExtractor themes, XlsxImporter xlsxImporter,
									   String nullAs) {
		if (color != null) {
			byte[] argb;
			Number indexed = color.getIndexed();
			if (indexed != null) {
				argb = UnitUtil.hexToArgbColor(xlsxImporter.colorsCache.getIndexColorByIndex(indexed.intValue()));
				if (argb != null) {
					return argb.length > 3 ? "#" + toHex(argb[1]) + toHex(argb[2]) + toHex(argb[3]) : "#" + toHex(argb[0]) + toHex(argb[1]) + toHex(argb[2]);
				}
			}

			String colorRgb = color.getRgb();
			Number colorTint = color.getTint();
			if (colorRgb != null) {
				colorRgb = colorRgb.toLowerCase();
				if (colorRgb.length() == 6) {
					colorRgb = "ff" + colorRgb;
				}
				argb = colorTint != null ? getRgbWithTint(UnitUtil.hexToArgbColor(colorRgb), colorTint.doubleValue()) : UnitUtil.hexToArgbColor(colorRgb);
				return argb.length > 3 ? "#" + toHex(argb[1]) + toHex(argb[2]) + toHex(argb[3]) : "#" + toHex(argb[0]) + toHex(argb[1]) + toHex(argb[2]);
			}

			Number colorTheme = color.getTheme();
			if (colorTheme != null) {
				if (themes != null) {
					XlsxExtractor.XlsxClrSchemeExtractor clrScheme = themes
							.getClrScheme();
					if (clrScheme != null) {
						String rgb = null;
						switch (colorTheme.intValue()) {
							case 0:
								rgb = clrScheme.getLt1();
								break;
							case 1:
								rgb = clrScheme.getDk1();
								break;
							case 2:
								rgb = clrScheme.getLt2();
								break;
							case 3:
								rgb = clrScheme.getDk2();
								break;
							case 4:
								rgb = clrScheme.getAccent1();
								break;
							case 5:
								rgb = clrScheme.getAccent2();
								break;
							case 6:
								rgb = clrScheme.getAccent3();
								break;
							case 7:
								rgb = clrScheme.getAccent4();
								break;
							case 8:
								rgb = clrScheme.getAccent5();
								break;
							case 9:
								rgb = clrScheme.getAccent6();
								break;
							case 10:
								rgb = clrScheme.getHlink();
								break;
							case 11:
								rgb = clrScheme.getFolHlink();
								break;
						}

						if (rgb != null) {
							argb = colorTint != null ? getRgbWithTint(UnitUtil.hexToArgbColor(rgb), colorTint.doubleValue()) : UnitUtil.hexToArgbColor(rgb);
							return argb.length > 3 ? "#" + toHex(argb[1]) + toHex(argb[2]) + toHex(argb[3]) : "#" + toHex(argb[0]) + toHex(argb[1]) + toHex(argb[2]);
						}
					}
				}
			}
		}
		return nullAs;
	}

	protected static SColor getSchemeHTMLColor(XlsxExtractor.XlsxSchemeClrExtractor color,
				XlsxExtractor.XlsxThemesExtractor themes) {
		String colorVal = color.getColorVal();
		String rgb = null;
		if (color.isSchemeClr()) {
			if (themes != null) {
				XlsxExtractor.XlsxClrSchemeExtractor clrScheme = themes
						.getClrScheme();
				if (clrScheme != null) {
					switch (colorVal) {
						case "lt1":
						case "bg1":
							rgb = clrScheme.getLt1();
							break;
						case "dk1":
						case "tx1":
							rgb = clrScheme.getDk1();
							break;
						case "lt2":
						case "bg2":
							rgb = clrScheme.getLt2();
							break;
						case "dk2":
						case "tx2":
							rgb = clrScheme.getDk2();
							break;
						case "accent1":
							rgb = clrScheme.getAccent1();
							break;
						case "accent2":
							rgb = clrScheme.getAccent2();
							break;
						case "accent3":
							rgb = clrScheme.getAccent3();
							break;
						case "accent4":
							rgb = clrScheme.getAccent4();
							break;
						case "accent5":
							rgb = clrScheme.getAccent5();
							break;
						case "accent6":
							rgb = clrScheme.getAccent6();
							break;
						case "hlink":
							rgb = clrScheme.getHlink();
							break;
						case "folhlink":
							rgb = clrScheme.getFolHlink();
							break;
					}
				}
			}
		} else if (color.isSrgbClr()) {
			rgb = colorVal;
		}
		if (rgb != null) {
			byte[] argb = UnitUtil.hexToArgbColor(rgb);
			if (color.isModified()) {
				Map modifies = color.getColor();
				boolean _hasAlpha = false;
				for (Object tag : modifies.keySet()) {
					if (tag instanceof String) {
						switch ((String) tag) {
							case "a:tint":
								argb = getRgbWithSchemeTint(argb, color.getTint().intValue() / 100000D);
								break;
							case "a:shade":
								argb = getRgbWithSchemeShade(argb, color.getShade().intValue() / 100000D);
								break;
							case "a:alpha":
								_hasAlpha = true;
								break;
							case "a:hueMod":
								argb = getRgbWithSchemeHueMod(argb, color.getHueMod().intValue() / 100000D);
								break;
							case "a:sat":
								argb = getRgbWithSchemeSat(argb, color.getSat().intValue() / 100000D);
								break;
							case "a:satOff":
								argb = getRgbWithSchemeSatOff(argb, color.getSatOff().intValue() / 100000D);
								break;
							case "a:satMod":
								argb = getRgbWithSchemeSatMod(argb, color.getSatMod().intValue() / 100000D);
								break;
							case "a:lum":
								argb = getRgbWithSchemeLum(argb, color.getLum().intValue() / 100000D);
								break;
							case "a:lumOff":
								argb = getRgbWithSchemeLumOff(argb, color.getLumOff().intValue() / 100000D);
								break;
							case "a:lumMod":
								argb = getRgbWithSchemeLumMod(argb, color.getLumMod().intValue() / 100000D);
								break;
						}
					}
				}
				// KEIKAI-299
				// TODO: support rgba
//				if (_hasAlpha) {
//					final double alpha = color.getAlpha().intValue() / 100000D;
//					final String hex = argb.length > 3 ?
//							rgbToRgbaString(argb[1] & 0xff, argb[2] & 0xff, argb[3] & 0xff, alpha) :
//							rgbToRgbaString(argb[0] & 0xff, argb[1] & 0xff, argb[2] & 0xff, alpha);
//					return new ColorImpl(hex);
//				}
			}
			final String hex = argb.length > 3 ? toHex(argb[1]) + toHex(argb[2]) + toHex(argb[3]) :
					toHex(argb[0]) + toHex(argb[1]) + toHex(argb[2]);
			return new ColorImpl(hex);
		}
		return null;
	}

	public static SFont.Underline toUnderline(String underline) {
		underline = underline == null ? "none" : underline;
		switch (underline) {
		case "single":
			return SFont.Underline.SINGLE;
		case "double":
			return SFont.Underline.DOUBLE;
		case "singleAccounting":
			return SFont.Underline.SINGLE_ACCOUNTING;
		case "doubleAccounting":
			return SFont.Underline.DOUBLE_ACCOUNTING;
		case "none":
		default:
			return SFont.Underline.NONE;
		}
	}
	public static SFont.TypeOffset toTypeOffset(String typeOffset) {
		typeOffset = typeOffset == null ? "baseline" : typeOffset;
		switch (typeOffset) {
		case "subscript":
			return SFont.TypeOffset.SUB;
		case "superscript":
			return SFont.TypeOffset.SUPER;
		case "baseline":
		default:
			return SFont.TypeOffset.NONE;
		}
	}
	protected boolean skipName(XlsxExtractor.XlsxDefinedNameExtractor definedName) {
		String namename = definedName.getName();
		if(namename == null) {
			return true;
		}
		// ignore defined name of functions, they are macro functions that we don't support
		if (definedName.isFunction()){
			return true;
		}

		if(definedName.getFormula() == null) { // ignore defined name with null formula, don't know when will have this case
			return true;
		}

		// because POI export doesn't support to hide name range
		Boolean hidden = definedName.isHidden();
		if (hidden == Boolean.TRUE) {
			return true; // ignore hidden definedName
		}

		return false;
	}
	/**
	 * Name should be created after sheets created. A special defined name,
	 * _xlnm._FilterDatabase (xlsx) or _FilterDatabase (xls), stores the
	 * selected cells for auto-filter
	 */
	protected void importNamedRange() {
		Iterable<XlsxExtractor.XlsxDefinedNameExtractor> definedNames = workbookExtractor
				.getDefinedNames();
		for (XlsxExtractor.XlsxDefinedNameExtractor definedName : definedNames) {
			if(skipName(definedName)){
				continue;
			}
			SName namedRange = null;
			if (definedName.getLocalSheetId() == null) {// workbook scope
				namedRange = book.createName(definedName.getName());
			} else {
				namedRange = book.createName(definedName.getName(), book.getSheet(Integer.parseInt(definedName.getLocalSheetId())).getSheetName());
			}
			namedRange.setRefersToFormula(definedName.getFormula());
		}
	}

	protected static InputStream convertZipInputStreamToInputStream(
			final ZipInputStream in) throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		IOUtils.copy(in, out);
		InputStream is = new ByteArrayInputStream(out.toByteArray());
		return is;
	}

	protected XlsxExtractor.XlsxWorkbookExtractor newXlsxExtractor(InputStream inputStream) throws IOException {
		ZipInputStream zipFile = inputStream instanceof ZipInputStream ?
				(ZipInputStream) inputStream :
				new ZipInputStream(inputStream);
		Map<String, InputStream> mapping = new HashMap(12);
		ZipEntry entry;
		while ((entry = zipFile.getNextEntry()) != null) {
			mapping.put(entry.getName(),
					convertZipInputStreamToInputStream(zipFile));
		}
		JSONObject xlsxMap = new JSONObject();

		XMLInputFactory factory = XMLInputFactory.newInstance();
		try {
			for (Map.Entry<String, InputStream> me : mapping.entrySet()) {
				final String key = me.getKey();
				final InputStream is = me.getValue();
//				System.out.println("name: " + key);
				if (key.endsWith(".rels") || key.endsWith(".xml")) {
					// ignore unsupported file type.
					if (key.endsWith("calcChain.xml") || key
							.startsWith("docProps/"))
						continue;

					XMLStreamReader reader = factory.createXMLStreamReader(is);
					xlsxMap.put(key, XlsxParser.parseXML(reader));

				} else if (key.endsWith(".vml")) {

					// issue #1547
					BufferedInputStream bufferedInputStream = null;
					try {
						bufferedInputStream = new BufferedInputStream(is);
						bufferedInputStream
								.mark(bufferedInputStream.available());
						XMLStreamReader reader = factory
								.createXMLStreamReader(bufferedInputStream);
						xlsxMap.put(key, XlsxParser.parseXML(reader));
					} catch (XMLStreamException e) {
						// for #1547, format the invalid content.
						Parser parser = Parser.htmlParser();

						// preserve-case
						parser.settings(new ParseSettings(true, true));

						inputStream.reset();
						Document document = parser.parseInput(
								IOUtils.toString(inputStream, "UTF-8"), "");
						document.outputSettings()
								.syntax(Document.OutputSettings.Syntax.xml);
						StringReader sreader = new StringReader(
								document.body().html());
						XMLStreamReader reader = factory
								.createXMLStreamReader(sreader);
						xlsxMap.put(key, XlsxParser.parseXML(reader));
					}
				} else { // image or bin
					//				Todo: Dart server is not supported yet.
					//					String s = Base64.getEncoder().encodeToString(IOUtils.toByteArray(is));
					xlsxMap.put(key, IOUtils.toByteArray(is));
				}
				is.close();
			}
		} catch (XMLStreamException e) {
			throw new IOException(e);
		}
		return XlsxExtractor.getWorkbook(xlsxMap);
	}

	//ZSS-854
	protected void importDefaultCellStyles(
			XlsxExtractor.XlsxStyleSheetExtractor styleSheet) {
		((AbstractBookAdv) book).clearDefaultCellStyles();
		for (XlsxExtractor.XlsxXfExtractor cellStyleXf : styleSheet
				.getCellStyleXfs()) {
			book.addDefaultCellStyle(importCellStyleXf(cellStyleXf, false));
		}

		// in case of XLS files which we have not support defaultCellStyles
		if (book.getDefaultCellStyles().isEmpty()) {
			((AbstractBookAdv) book).initDefaultCellStyles();
		}
		// ZSS-1132: setup default font
		((AbstractBookAdv) book).initDefaultFont();
	}

	protected String getNumberFormat(XlsxExtractor.XlsxNumFmtExtractor extractor) {
		if (extractor == null) {
			return numFmtsCache.getNumFmtAt(0);
		}
		String formatCode = extractor.getFormatCode();
		if (formatCode != null) return formatCode;
		return numFmtsCache.getNumFmtAt(extractor.getNumFmtId().intValue());
	}
	protected SCellStyle.Alignment toHorizontalAlignment(String align) {
		align = align == null ? "general" : align;
		switch (align) {
		case "left":
			return SCellStyle.Alignment.LEFT;
		case "right":
			return SCellStyle.Alignment.RIGHT;
		case "center":
			return SCellStyle.Alignment.CENTER;
		case "centerContinuous": //ZSS-779
			return SCellStyle.Alignment.CENTER_SELECTION;
		case "fill":
			return SCellStyle.Alignment.FILL;
		case "justify":
			return SCellStyle.Alignment.JUSTIFY;
		case "general":
		default:
			return SCellStyle.Alignment.GENERAL;
		}
	}
	protected SCellStyle.VerticalAlignment toVerticalAlignment(String align) {
		align = align == null ? "bottom" : align;
		switch (align) {
		case "top":
			return SCellStyle.VerticalAlignment.TOP;
		case "center":
			return SCellStyle.VerticalAlignment.CENTER;
		case "justify":
			return SCellStyle.VerticalAlignment.JUSTIFY;
		case "bottom":
		default:
			return SCellStyle.VerticalAlignment.BOTTOM;
		}
	}

	protected short convertToUIRotation(short rotation) {
		if (90 < rotation && rotation <= 180) {
			rotation = (short)(90 - rotation);
		}

		return rotation;
	}
	List<XlsxExtractor.XlsxXfExtractor> cellStyleXfsCache = new ArrayList<>();
	protected SCellStyle importCellStyleXf(XlsxExtractor.XlsxXfExtractor xCellXf,
			boolean inStyleTable) {

		// keep the file indexing here, so we don't need to check whether it's duplicated or not
		cellStyleXfsCache.add(xCellXf);

		return importCellXf(xCellXf, inStyleTable);
	}

	// without cache the CellXf
	protected SCellStyle importCellXf(XlsxExtractor.XlsxXfExtractor xCellXf, boolean inStyleTable) {
		SCellStyle cellStyle = importedXfs.get(xCellXf);
		//				short idx = cellXf.getIndex(); // ZSS-685
		if (cellStyle == null) { // ZSS-685
			final SCellStyleBuilder csb = inStyleTable ? book.addCellStyle() : book.createCellStyle();
			Number numFmtId = xCellXf.getNumFmtId();
			if (numFmtId != null) {
				String dataFormat = numFmtsCache.getNumFmtAt(numFmtId.intValue());
				if (dataFormat == null) {//just in case
					dataFormat = SCellStyle.FORMAT_GENERAL;
				}
				csb.dataFormat(dataFormat).directFormat(!NumFmts.isBuiltIn(numFmtId.intValue()));
			}
			XlsxExtractor.XlsxAlignmentExtractor alignment = xCellXf
					.getAlignment();
			if (alignment != null) {
				csb.wrapText(alignment.isWrapText())
					.alignment(toHorizontalAlignment(alignment.getHorizontal()))
					.verticalAlignment(toVerticalAlignment(alignment.getVertical()))
					.rotation(convertToUIRotation(alignment.getTextRotation())) //ZSS-918
					.indention(alignment.getIndent()); //ZSS-915
			}

			XlsxExtractor.XlsxProtectionExtractor protection = xCellXf
					.getProtection();
			if (protection != null) {
				csb.locked(protection.isLocked()).hidden(protection.isHidden());
			}
			Number fillId = xCellXf.getFillId();
			if (fillId != null) {
				Fills.Fill fill = fillsCache.getFillAt(fillId.intValue());
				if (fill instanceof Fills.PatternFill) {
					Fills.PatternFill patternFill = (Fills.PatternFill) fill;
					//ZSS-857: SOLID pattern: switch fillColor and backColor
					String fgColor = patternFill.getFgColor();
					String bgColor = patternFill.getBgColor();
					if (patternFill.getType() == SFill.FillPattern.SOLID) {
						String tmp = fgColor;
						fgColor = bgColor;
						bgColor = tmp;
					}
					csb.fillPattern(patternFill.getType()).fillColor(fgColor).backColor(bgColor); //ZSS-780
				}
			}
			Number borderId = xCellXf.getBorderId();
			if (borderId != null) {
				Borders.Border border = bordersCache
						.getBordreAt(borderId.intValue());
				if (border != null) {
					Borders.BorderPr borderPr = border.getLeft();
					if (borderPr != null) {
						csb.borderLeft(borderPr.getStyle()).borderLeftColor(borderPr.getColor());
					}
					borderPr = border.getRight();
					if (borderPr != null) {
						csb.borderRight(borderPr.getStyle()).borderRightColor(borderPr.getColor());
					}
					borderPr = border.getTop();
					if (borderPr != null) {
						csb.borderTop(borderPr.getStyle()).borderTopColor(borderPr.getColor());
					}
					borderPr = border.getBottom();
					if (borderPr != null) {
						csb.borderBottom(borderPr.getStyle()).borderBottomColor(borderPr.getColor());
					}
				}
			}
			Number fontId = xCellXf.getFontId();
			if (fontId != null) {
				csb.font(fontsCache.getFontAt(fontId.intValue()));
			}
			cellStyle = csb.build();
			importedXfs.put(xCellXf, cellStyle); // ZSS-685
		}
		return cellStyle;
	}

	//ZSS-854
	protected void importNamedStyles(
			XlsxExtractor.XlsxStyleSheetExtractor styleSheet) {
		((AbstractBookAdv) book).clearNamedStyles();
		for (XlsxExtractor.XlsxCellStyleExtractor extractor : styleSheet.getCellStyles()) {
			SNamedStyle namedStyle =
					new NamedStyleImpl(extractor.getName(), extractor.isCustomBuiltin(),
							extractor.getBuiltinId(), book, extractor.getXfId());
			book.addNamedCellstyle(namedStyle);
		}
	}//ZSS-855

	protected void importTables(XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet) {
		for (XlsxExtractor.XlsxTableExtractor xTable : xSheet.getTables()) {
			final String displayName = xTable.getDisplayName();
			_importPhase.setPhase(ImportPhase.TABLE);
			_importPhase.setTableName(displayName);
			final CellRegion region0 = new CellRegion(xTable.getRef());
			final SheetRegion region = new SheetRegion(sheet, region0);
			_importPhase.setPhase(ImportPhase.TABLE_STYLE);
			XlsxExtractor.XlsxTableStyleInfoExtractor tableStyleInfo = xTable
					.getTableStyleInfo();
			final STableStyleInfo info = new TableStyleInfoImpl(
					tableStyleInfo.getName(), tableStyleInfo.isShowColumnStripes(),
					tableStyleInfo.isShowRowStripes(), tableStyleInfo.isShowFirstColumn(),
					tableStyleInfo.isShowLastColumn());

			_importPhase.setPhase(ImportPhase.TABLE);
			final STable table = new TableImpl((AbstractBookAdv) book,
					xTable.getName(), displayName, region,
					xTable.getHeaderRowCount(), xTable.getTotalsRowCount(),
					info, xTable.isTotalsRowShown());

			for (XlsxExtractor.XlsxTableColumnExtractor xTableColumn : xTable.getTableColumns()) {
				_importPhase.setPhase(ImportPhase.TABLE_COLUMN);
				final String name = xTableColumn.getName();
				_importPhase.setTableColumnName(name);
				final STableColumn tbCol = new TableColumnImpl(
						xTableColumn.getName());
				tbCol.setTotalsRowFunction(xTableColumn
								.getTotalsRowFunction());
				if (tbCol.getTotalsRowFunction()
						== STableColumn.STotalsRowFunction.none) {
					tbCol.setTotalsRowLabel(xTableColumn.getTotalsRowLabel());
				} else if (tbCol.getTotalsRowFunction()
						== STableColumn.STotalsRowFunction.custom) {
					XlsxExtractor.XlsxTableFormulaExtractor totalsRowFormula = xTableColumn
							.getTotalsRowFormula();
					if (totalsRowFormula != null) {
						tbCol.setTotalsRowFormula(totalsRowFormula.getExpression()); //ZSS-977
					}
				}
				table.addColumn(tbCol);
			}

			_importPhase.setPhase(ImportPhase.TABLE);
			XlsxExtractor.XlsxAutoFilterExtractor autoFilter = xTable
					.getAutoFilter();
			table.enableAutoFilter(autoFilter != null);

			sheet.addTable(table);

			//ZSS-1019
			if (autoFilter != null) {
				_importPhase.setPhase(ImportPhase.TABLE_AUTO_FILTER);
				final SAutoFilter zssFilter = table.getAutoFilter();
				importAutoFilterColumns(autoFilter, zssFilter);
				_importPhase.setPhase(ImportPhase.TABLE);
			}
			importTableName(table);
			((AbstractBookAdv) book).addTable(table);
			_importPhase.clear();
		}
	}

	//ZSS-855
	protected void importTableName(STable table) {
		SName namedRange = ((AbstractBookAdv) book).createTableName(table);
		SheetRegion rgn = table.getDataRegion();
		namedRange.setRefersToFormula(rgn.getReferenceString());
	}

	//ZSS-1191
	//@since 3.9.0
	protected SColorFilter importColorFilter(
			XlsxExtractor.XlsxColorFilterExtractor colorFilter) {
		if (colorFilter == null) return null;

		final SExtraStyle src = styleCaches.getExtraStyleCache().get(colorFilter.getDxfId().intValue());
		if (src == null) return null;

		final SExtraStyle style = book.getOrAddExtraStyle(src);
		final Boolean cellColor = colorFilter.isCellColor();
		return new ColorFilterImpl(style, cellColor != null && !cellColor);
	}

	//ZSS-1019
	protected void importAutoFilterColumns(
			XlsxExtractor.XlsxAutoFilterExtractor xFilter,
			SAutoFilter zssFilter) {
		List<XlsxExtractor.XlsxFilterColumnExtractor> filterColumns = xFilter
				.getFilterColumns();
		Map<String, Object> extra = new HashMap<String, Object>();
		for (XlsxExtractor.XlsxFilterColumnExtractor extractor : filterColumns) {
			SAutoFilter.NFilterColumn destColumn = zssFilter
					.getFilterColumn(extractor.getColId().intValue(), true);

			//ZSS-1191
			final SColorFilter destColorFilter = importColorFilter(
					extractor.getColorFilter());
			extra.put("colorFilter", destColorFilter);

			//ZSS-1224
			final SCustomFilters destCustomFilters = importCustomFilters(
					extractor.getCustomFilters());
			extra.put("customFilters", destCustomFilters);

			//ZSS-1226
			final SDynamicFilter destDynamicFilter = importDynamicFilter(
					extractor.getDynamicFilter());
			extra.put("dynamicFilter", destDynamicFilter);

			//ZSS-1227
			final STop10Filter destTop10Filter = importTop10Filter(
					extractor.getTop10());
			extra.put("top10Filter", destTop10Filter);

			SAutoFilter.FilterOp operator = SAutoFilter.FilterOp.values;
			Set criteria1 = null;
			Set criteria2 = null;

			XlsxExtractor.XlsxFiltersExtractor filters = extractor.getFilters();
			if (filters != null) {
				List<String> filters1 = filters.getFilters();
				if (filters1 != null) {
					criteria1 = filters1.stream().collect(Collectors.toSet());
					if (filters.isBlank()) {
						criteria1.add("=");
					}
					operator =  SAutoFilter.FilterOp.values;
				}
			}

			if (destColorFilter != null) {
				operator = destColorFilter.isByFontColor() ?
						SAutoFilter.FilterOp.fontColor: SAutoFilter.FilterOp.cellColor;
			}

			//ZSS-1226
			if (destDynamicFilter != null) {
				String dynamicFilterType = destDynamicFilter.getType();
				if ("aboveAverage".equals(dynamicFilterType)) {
					operator = SAutoFilter.FilterOp.aboveAverage;
				} else if ("belowAverage".equals(dynamicFilterType)) {
					operator = SAutoFilter.FilterOp.belowAverage;
				}
			}

			//ZSS-1227
			if (destTop10Filter != null) {
				operator = SAutoFilter.FilterOp.top10;
			}
			destColumn.setProperties(operator,
					criteria1, criteria2,
					extractor.isSetHiddenButton() ? !extractor.isHiddenButton() :
							(extractor.isSetShowButton() ? extractor.isShowButton() : true), extra);
		}
	}

	//ZSS-1224
	//@since 3.9.0
	protected SCustomFilters importCustomFilters(
			XlsxExtractor.XlsxCustomFiltersExtractor xCustomFilters) {
		if (xCustomFilters == null)
			return null;
		final boolean and = xCustomFilters.isAnd();
		final XlsxExtractor.XlsxCustomFilterExtractor poiFilter1 = xCustomFilters.getCustomFilter1();
		final XlsxExtractor.XlsxCustomFilterExtractor poiFilter2 = xCustomFilters.getCustomFilter2();
		final SCustomFilter filter1 = poiFilter1 != null ? new CustomFilterImpl(
				poiFilter1.getVal(), poiFilter1.getOperator()) : null;
		final SCustomFilter filter2 = poiFilter2 == null ?
				null :
				new CustomFilterImpl(poiFilter2.getVal(), poiFilter2.getOperator());
		return new CustomFiltersImpl(filter1, filter2, and);
	}

	//ZSS-1226
	//@since 3.9.0
	protected SDynamicFilter importDynamicFilter(
			XlsxExtractor.XlsxDynamicFilterExtractor xDynamicFilter) {
		if (xDynamicFilter == null)
			return null;
		Number maxval = xDynamicFilter.getMaxVal();
		Number val = xDynamicFilter.getVal();
		String type = xDynamicFilter.getType(); //ZSS-1234
		return new DynamicFilterImpl(maxval != null ? maxval.doubleValue() : null, val != null ? val.doubleValue() : null, type); //ZSS-1234
	}

	//ZSS-1227
	//@since 3.9.0
	protected STop10Filter importTop10Filter(
			XlsxExtractor.XlsxTop10Extractor xTop10Filter) {
		if (xTop10Filter == null)
			return null;
		return new Top10FilterImpl(xTop10Filter.isTop(),
				xTop10Filter.getVal(), xTop10Filter.isPercent(),
				xTop10Filter.getFilterVal());
	}

	protected void importExtraStyles(
			XlsxExtractor.XlsxStyleSheetExtractor styleSheet) {
		((AbstractBookAdv) book).clearExtraStyles();
		for (XlsxExtractor.XlsxDxfExtractor extractor: styleSheet.getDxfs()) {
			styleCaches.getExtraStyleCache().add(importExtraStyle(extractor));
		}
	}

	//ZSS-1140
	protected SExtraStyle importExtraStyle(XlsxExtractor.XlsxDxfExtractor xDxf) {
		final XlsxExtractor.XlsxNumFmtExtractor xNumFmt = xDxf.getNumFmt();
		String dataFormat = xNumFmt != null ? getNumberFormat(xDxf.getNumFmt()) : null;

		XlsxExtractor.XlsxFillExtractor xDxfFill = xDxf.getFill();
		SFill fill = null;
		if (xDxfFill instanceof XlsxExtractor.XlsxPatternExtractor) {
			XlsxExtractor.XlsxPatternExtractor dxfFill = (XlsxExtractor.XlsxPatternExtractor) xDxfFill;
			final XlsxExtractor.XlsxColorExtractor xFgColor = dxfFill.getFgColor();
			final XlsxExtractor.XlsxColorExtractor xBgColor = dxfFill.getBgColor();
			String fgColor = xFgColor != null ? getForegroundHTMLColor(dxfFill.getFgColor(), themes, this) : null;
			String bgColor = xBgColor != null ? getBackgroundHTMLColor(dxfFill.getBgColor(), themes, this) : null;
			final String patternType = dxfFill.getPatternType();
			SFill.FillPattern fillPattern = patternType != null ?
					new Fills.PatternFill(patternType, fgColor, bgColor).getType() : null;
			SColor fgSColor = fgColor == null ? null : book.createColor(fgColor);
			SColor bgSColor = bgColor == null ? null : book.createColor(bgColor);
			fill = new ExtraFillImpl(fillPattern, fgSColor, bgSColor); //ZSS-1162
		}
		XlsxExtractor.XlsxBorderExtractor xBorder = xDxf.getBorder();

		SBorder border = null;
		if (xBorder != null) {
			Borders.Border border1 = createBorder(xBorder);
			border = createSBorder(border1);
		}
		SFont font = null;
		XlsxExtractor.XlsxFontExtractor xDxfFont = xDxf.getFont();
		if (xDxfFont != null) {
			font = createDxfZssFont(xDxfFont);
		}

		return new ImmutableExtraStyleImpl(font, fill, border, dataFormat);
	}

	protected SBorder createSBorder(Borders.Border border) {
		Borders.BorderPr borderLeft = border.getLeft();
		Borders.BorderPr borderRight = border.getRight();
		Borders.BorderPr borderTop = border.getTop();
		Borders.BorderPr borderBottom = border.getBottom();
		Borders.BorderPr borderDiagonal = border.getDiagonal();
		Borders.BorderPr borderVertical = border.getVertical();
		Borders.BorderPr borderHorizontal = border.getHorizontal();

		SBorderLine left = borderLeft == null ? null : new BorderLineImpl(borderLeft.getStyle(), borderLeft.getColor() == null ? null : this.book.createColor(borderLeft.getColor()));
		SBorderLine top = borderTop == null ? null : new BorderLineImpl(borderTop.getStyle(), borderTop.getColor() == null ? null : this.book.createColor(borderTop.getColor()));
		SBorderLine right = borderRight == null ? null : new BorderLineImpl(borderRight.getStyle(), borderRight.getColor() == null ? null : this.book.createColor(borderRight.getColor()));
		SBorderLine bottom = borderBottom == null ? null : new BorderLineImpl(borderBottom.getStyle(), borderBottom.getColor() == null ? null : this.book.createColor(borderBottom.getColor()));
		SBorderLine diagonal = borderDiagonal == null ? null : new BorderLineImpl(borderDiagonal.getStyle(), borderDiagonal.getColor() == null ? null : this.book.createColor(borderDiagonal.getColor()));
		SBorderLine vertical = borderVertical == null ? null : new BorderLineImpl(borderVertical.getStyle(), borderVertical.getColor() == null ? null : this.book.createColor(borderVertical.getColor()));
		SBorderLine horizontal = borderHorizontal == null ? null : new BorderLineImpl(borderHorizontal.getStyle(), borderHorizontal.getColor() == null ? null : this.book.createColor(borderHorizontal.getColor()));
		return new BorderImpl(left, top, right, bottom, diagonal, vertical, horizontal);
	}

	//ZSS-992
	protected void importTableStyles(
			XlsxExtractor.XlsxStyleSheetExtractor extractor) {
		((AbstractBookAdv) book).clearTableStyles();
		XlsxExtractor.XlsxTableStylesExtractor tableStyles = extractor
				.getTableStyles();
		if (tableStyles != null) {
			for (XlsxExtractor.XlsxTableStyleExtractor tableStyle : tableStyles.getTableStyles()) {
				book.addTableStyle(importTableStyle(tableStyle));
			}
			book.setDefaultPivotStyleName(tableStyles.getDefaultPivotStyle());
			book.setDefaultTableStyleName(tableStyles.getDefaultTableStyle());
		}
	}

	//ZSS-992
	protected STableStyleElem importTableStyleElem(XlsxExtractor.XlsxTableStyleElementExtractor styleElement) {
		final SExtraStyle extraStyle = styleCaches.getExtraStyleCache().get(styleElement.getDxfId().intValue());
		if (extraStyle == null) return null;
		final SExtraStyle style = book.getOrAddExtraStyle(extraStyle);
		return new TableStyleElemImpl(style.getFont(), style.getFill(), style.getBorder());
	}

	//ZSS-992
	protected STableStyle importTableStyle(XlsxExtractor.XlsxTableStyleExtractor tableStyle) {
		final String name = tableStyle.getName();
		final Map<String, STableStyleElem> map = new HashMap(13);
		for (XlsxExtractor.XlsxTableStyleElementExtractor styleElement: tableStyle.getTableStyleElements()) {
			map.put(styleElement.getType(), importTableStyleElem(
					styleElement));
		}
		return new TableStyleImpl(
				name,
				map.get("wholeTable"),
				map.get("firstColumnStripe"),
				1, 		//colStripe1Size,

				map.get("secondColumnStripe"),
				1, 		//colStripe2Size,

				map.get("firstRowStripe"),
				1, 		//rowStripe1Size,

				map.get("secondRowStripe"),
				1, 		//rowStripe2Size,
				map.get("lastColumn"),
				map.get("firstColumn"),
				map.get("headerRow"),
				map.get("totalRow"),
				map.get("firstHeaderCell"),
				map.get("lastHeaderCell"),
				map.get("firstTotalCell"),
				map.get("lastTotalCell")
		);
	}

	protected void setBookType(SBook book) {
		book.setAttribute(BOOK_TYPE_KEY, "xlsx");
	}

	protected void importExternalBookLinks(
			List<XlsxExtractor.XlsxExternalLinkExtractor> externalReferences) {
		if (!externalReferences.isEmpty()) {
			// directly use XML to retrieve external book references and fetch external book names
			// if retrieving ExternalLink from relations list of workbook, the order might be incorrect.
			// according to spec. ISO 29500-1, 18.17.2.3, p. 2048, the order should be natural order of XML tags.
			List<String> bookNames = new ArrayList<String>();
				for (XlsxExtractor.XlsxExternalLinkExtractor exRef : externalReferences) {
					XlsxExtractor.XlsxExternalBookExtractor externalBook = exRef
							.getExternalBook();
					if (externalBook != null) {
						bookNames.add(externalBook.getName());
					}
				}
			if (bookNames.size() > 0) {
				this.book.setAttribute(FormulaEngine.KEY_EXTERNAL_BOOK_NAMES,
						bookNames);
			}
		}
	}

	protected void importWorkbookProtection(XlsxExtractor.XlsxWorkbookExtractor xBook, SBook book) {
		final XlsxExtractor.XlsxWorkbookProtectionExtractor xBookProtection = xBook.getWorkbookProtection();
		if (xBookProtection != null) {
			final WorkbookProtectionImpl bookProtection = (WorkbookProtectionImpl) book.getWorkbookProtection();
			bookProtection.setLockStructure(xBookProtection.isLockStructure());
			bookProtection.setLockWindows(xBookProtection.isLockWindows());
			bookProtection.setAlgName(xBookProtection.getAlgName());
			bookProtection.setHashValue(xBookProtection.getHashValue());
			bookProtection.setSaltValue(xBookProtection.getSaltValue());
			bookProtection.setSpinCount(xBookProtection.getSpinCount());
		}
	}

	protected void importSheets(List<XlsxExtractor.XlsxSheetExtractor> sheets) {
		sheets.forEach((xSheet) -> {
			SSheet sheet = importSheet(xSheet);

			// importTables before importRow and importNamedRange
			importTables(xSheet, sheet); //ZSS-855, ZSS-1011
		});
	}

	/*
	 * import sheet scope content
	 */
	protected SSheet importSheet(XlsxExtractor.XlsxSheetExtractor xSheet) {
		final String sheetName = xSheet.getName();
		_importPhase.setPhase(ImportPhase.SHEET);
		_importPhase.setSheetName(sheetName);

		SSheet sheet = book.createSheet(sheetName);
		List<XlsxExtractor.XlsxSheetViewExtractor> sheetViews = xSheet
				.getSheetViews();
		XlsxExtractor.XlsxSheetFormatPrExtractor sheetFormatPr = xSheet
				.getSheetFormatPr();
		_importPhase.setPhase(ImportPhase.SHEET_FORMAT);
		if (sheetFormatPr != null) {
			sheet.setDefaultRowHeight(
				UnitUtil.twipToPx((short)(sheetFormatPr.getDefaultRowHeight() * 20))); // in twips (1/20 of  a point)

			// refer to XSSFSheet#getXssfDefaultColumnWidth()
			int charWidth = ((AbstractBookAdv) book).getCharWidth();
			double defaultColWidth = 9.142857D;
			if (sheetFormatPr.isSetDefaultColWidth()) {
				defaultColWidth = sheetFormatPr.getDefaultColWidth().doubleValue();
			} else if (sheetFormatPr.isSetBaseColWidth()) {
				defaultColWidth = UnitUtil.pxToCTChar(UnitUtil.defaultColumnWidthToPx(sheetFormatPr.getBaseColWidth(), charWidth), charWidth);
			}
			Number outlineLevelCol = sheetFormatPr.getOutlineLevelCol();
			Number outlineLevelRow = sheetFormatPr.getOutlineLevelRow();
			if (outlineLevelCol != null) {
				sheet.setOutlineLevelCol(outlineLevelCol.intValue());
			}
			if (outlineLevelRow != null) {
				sheet.setOutlineLevelRow(outlineLevelRow.intValue());
			}
			//ZSS-952
			sheet.setDefaultColumnWidth(UnitUtil
					.xssfDefaultColumnWidthToPx(defaultColWidth,
							charWidth));//ZSS-1132
			_importPhase.setPhase(ImportPhase.SHEET);
		} else {
			// refer to XSSFSheet#getXssfDefaultColumnWidth()
			sheet.setDefaultColumnWidth(UnitUtil
					.xssfDefaultColumnWidthToPx(9.142857D,
							((AbstractBookAdv) book).getCharWidth()));//ZSS-1132
		}
		SSheetViewInfo viewInfo = sheet.getViewInfo();
		if (sheetViews != null && !sheetViews.isEmpty()) {
			_importPhase.setPhase(ImportPhase.SHEET_VIEW);
			XlsxExtractor.XlsxSheetViewExtractor xlsxSheetViewExtractor = sheetViews
					.get(0);
			XlsxExtractor.XlsxPaneExtractor pane = xlsxSheetViewExtractor
					.getPane();
			if (pane != null) {
				viewInfo.setNumOfRowFreeze(pane.getYSplit());
				viewInfo.setNumOfColumnFreeze(pane.getXSplit());
			}
			Boolean showGridLines = xlsxSheetViewExtractor.isShowGridLines();
			if (showGridLines != null) {
				viewInfo.setDisplayGridlines(showGridLines); // Note isDisplayGridlines() and isPrintGridlines() are different
			}

			viewInfo.setColumnBreaks(xSheet.getColBreaks());
			viewInfo.setRowBreaks(xSheet.getRowBreaks());

			viewInfo.setTabSelected(xlsxSheetViewExtractor.isTabSelected());
			List<XlsxExtractor.XlsxSelectionExtractor> selections = xlsxSheetViewExtractor.getSelections();
			if (selections != null && !selections.isEmpty()) {
				XlsxExtractor.XlsxSelectionExtractor xlsxSelectionExtractor = selections.get(0);
				if (pane != null && pane.getActivePane() != null) {
					String activePane = pane.getActivePane();
					for (XlsxExtractor.XlsxSelectionExtractor selection : selections) {
						if (activePane.equals(selection.getPane())) {
							xlsxSelectionExtractor = selection;
							break;
						}
					}
				}
				String activeCell = xlsxSelectionExtractor.getActiveCell();
				String selectionAreas = xlsxSelectionExtractor.getSqref();
				viewInfo.setActiveCell(Strings.isBlank(activeCell) ? "A1" : activeCell); // the default is "A1"
				viewInfo.setSelectionAreas(Strings.isBlank(selectionAreas) ? "A1" : selectionAreas); // the default is "A1"
			}
		}

		final XlsxExtractor.XlsxSheetPropertiesExtractor sheetPr = xSheet.getSheetPr();
		if (sheetPr != null) {
			final XlsxExtractor.XlsxColorExtractor xTabColor = sheetPr.getTabColor();
			String tabColor = xTabColor != null ? getHTMLColor(xTabColor, themes, this, null) : null;
			sheet.setTabColor(tabColor);

			final XlsxExtractor.XlsxOutlinePrExtractor xOutlinePrExtractor = sheetPr.getOutlinePr();
			if (xOutlinePrExtractor != null) {
				if (xOutlinePrExtractor.isApplyStyles()) {
					sheet.setApplyStyles(true);
				}
				if (!xOutlinePrExtractor.isSummaryBelow()) {
					sheet.setSummaryBelow(false);
				}
				if (!xOutlinePrExtractor.isSummaryRight()) {
					sheet.setSummaryRight(false);
				}
				if (!xOutlinePrExtractor.isShowOutlineSymbols()) {
					sheet.setShowOutlineSymbols(false);
				}
			}
		}
		_importPhase.setPhase(ImportPhase.SHEET);
		//ZSS-832
		//import sheet visible
		final String state = xSheet.getState();
		if ("hidden".equals(state)) {
			sheet.setSheetVisible(SSheet.SheetVisible.HIDDEN);
		} else if ("veryHidden".equals(state)) {
			sheet.setSheetVisible(SSheet.SheetVisible.VERY_HIDDEN);
		} else {
			sheet.setSheetVisible(SSheet.SheetVisible.VISIBLE);
		}

		return sheet;
	}

	protected void importPassword(
			XlsxExtractor.XlsxSheetProtectionExtractor protectionExtractor, SSheet sheet) {
		String password = protectionExtractor.getPassword();
		if (password != null) {
			short hashpass = getPasswordHash(password);
			sheet.setHashedPassword(hashpass);
		}
		//ZSS-1063
		String hashValue = protectionExtractor.getHashValue();
		sheet.setHashValue(hashValue);
		String spinCount = protectionExtractor.getSpinCount();
		sheet.setSpinCount(spinCount);
		String saltValue = protectionExtractor.getSaltValue();
		sheet.setSaltValue(saltValue);
		String algName = protectionExtractor.getAlgName();
		sheet.setAlgName(algName);
	}
	protected short getPasswordHash(String password) {
		// POI5 uses XOR password (Excel 2010 and earlier)
		return (short) Integer.parseInt(password, 16);
	}

	//ZSS-1130
	protected void importConditionalFormatting(
			XlsxExtractor.XlsxSheetExtractor xSheet, SSheet sheet, List<XlsxExtractor.XlsxExtExtractor> extList) {
		List<XlsxExtractor.XlsxConditionalFormattingExtractor> conditionalFormattings = xSheet
				.getConditionalFormattings();

		if (!conditionalFormattings.isEmpty()) {
			for (XlsxExtractor.XlsxConditionalFormattingExtractor cf : conditionalFormattings) {
				SConditionalFormatting scf = prepareConditionalFormattingImpl(sheet, cf);
				((AbstractSheetAdv) sheet).addConditionalFormatting(scf);
			}
		}

		// KEIKAI-395
		if (!extList.isEmpty()) {
			for (XlsxExtractor.XlsxExtExtractor ext : extList) {
				List<XlsxExtractor.XlsxExtConditionalFormattingExtractor> extConditionalFormattings =
						ext.getConditionalFormattings();
				if (!extConditionalFormattings.isEmpty()) {
					for (XlsxExtractor.XlsxExtConditionalFormattingExtractor cf : extConditionalFormattings) {
						SConditionalFormatting scf = prepareConditionalFormattingImpl(sheet, cf);
						((AbstractSheetAdv) sheet).addConditionalFormatting(scf);
					}
				}
			}
		}
	}

	//ZSS-1130
	protected ConditionalFormattingImpl prepareConditionalFormattingImpl(
			SSheet sheet, XlsxExtractor.XlsxConditionalFormattingExtractor cf) {
		final ConditionalFormattingImpl cfi = new ConditionalFormattingImpl(
				sheet);
		for (Ref ref : Converter.toRefs(cf.getSqref())) {
			final int r1 = ref.getTop();
			final int c1 = ref.getLeft();
			final int r2 = ref.getBottom();
			final int c2 = ref.getRight();
			cfi.addRegion(new CellRegion(r1, c1, r2, c2));
		}

		for (XlsxExtractor.XlsxCfRuleExtractor cfRule: cf.getCfRules()) {
			cfi.addRule(prepareConditonalFormattingRuleImpl(cfi, cfRule)); //ZSS-1142
		}

		// KEIKAI-395 remember it is from ExtLst so we can export correctly?
		if (cf instanceof XlsxExtractor.XlsxExtConditionalFormattingExtractor) {
			cfi.setUnderExt(true);
		}
		return cfi;
	}

	//ZSS-1130
	protected ConditionalFormattingRuleImpl prepareConditonalFormattingRuleImpl(
			SConditionalFormatting cfi, XlsxExtractor.XlsxCfRuleExtractor xRule) {
		final ConditionalFormattingRuleImpl cfri = new ConditionalFormattingRuleImpl(
				cfi); //ZSS-1142
		cfri.setPriority(xRule.getPriority());
		cfri.setType(xRule.getType());
		cfri.setExtId(xRule.getExtId()); // KEIKAI-469
		if (xRule.isSetStopIfTrue())
			cfri.setStopIfTrue(xRule.isStopIfTrue());
		if (xRule.isSetDxfId()) {
			final SExtraStyle extraStyle = styleCaches.getExtraStyleCache().get(xRule.getDxfId());
			if (extraStyle != null) {
				cfri.setExtraStyle(book.getOrAddExtraStyle(extraStyle));
			}
		} else {
			// KEIKAI-395
			if (xRule instanceof XlsxExtractor.XlsxExtCfRuleExtractor) {
				XlsxExtractor.XlsxExtCfRuleExtractor xExtRule = (XlsxExtractor.XlsxExtCfRuleExtractor) xRule;
				final XlsxExtractor.XlsxDxfExtractor dxfExtractor = xExtRule.getDxf();
				if (dxfExtractor != null) {
					final SExtraStyle extraStyle = importExtraStyle(dxfExtractor);
					cfri.setExtraStyle(extraStyle);
					cfri.setUnderExt(true); // indicate this CfriRule is under <extLst><ext> ....
				}
				// KEIKAI-469 setId as well
				cfri.setId(xExtRule.getId());
			}
		}
		switch (cfri.getType()) {
		case ABOVE_AVERAGE:
			if (xRule.isSetAboveAverage())
				cfri.setAboveAverage(xRule.isAboveAverage());
			if (xRule.isSetEqualAverage())
				cfri.setEqualAverage(xRule.isEqualAverage());
			if (xRule.isSetStdDev())
				cfri.setStandardDeviation(xRule.getStdDev());
			break;
		case CELL_IS:
			if (xRule.isSetOperator())
				cfri.setOperator(xRule.getOperator());
			prepareFormulas(cfri, xRule);
			break;
		case COLOR_SCALE:
			cfri.setColorScale(prepareColorScale(xRule));
			break;
		case CONTAINS_BLANKS:
		case NOT_CONTAINS_BLANKS:
		case EXPRESSION:
		case CONTAINS_ERRORS:
		case NOT_CONTAINS_ERRORS:
			prepareFormulas(cfri, xRule);
			break;
		case CONTAINS_TEXT:
		case NOT_CONTAINS_TEXT:
		case BEGINS_WITH:
		case ENDS_WITH:
			cfri.setText(xRule.getText());
			cfri.setOperator(xRule.getOperator());
			prepareFormulas(cfri, xRule);
			break;
		case DATA_BAR:
			cfri.setDataBar(prepareDataBar(xRule));
			break;
		case ICON_SET:
			cfri.setIconSet(prepareIconSet(xRule));
			break;
		case TIME_PERIOD:
			cfri.setTimePeriod(xRule.getTimePeriod());
			prepareFormulas(cfri, xRule);
			break;
		case TOP_10:
			if (xRule.isSetRank())
				cfri.setRank(xRule.getRank());
			if (xRule.isSetPercent())
				cfri.setPercent(xRule.isPercent());
			if (xRule.isSetBottom())
				cfri.setBottom(xRule.isBottom());
			break;
		case DUPLICATE_VALUES:
		case UNIQUE_VALUES:
			// No extra setting
			break;
		}

		return cfri;
	}

	//ZSS-1130
	protected SColorScale prepareColorScale(
			XlsxExtractor.XlsxCfRuleExtractor ctRule) {
		XlsxExtractor.XlsxColorScaleExtractor colorScale1 = ctRule
				.getColorScale();
		final ColorScaleImpl colorScale = new ColorScaleImpl();
		for (XlsxExtractor.XlsxCfvoExtractor ctvo : colorScale1.getCfvos()) {
			colorScale.addValueObject(prepareValueObject(ctvo));
		}
		for (XlsxExtractor.XlsxColorExtractor xColor : colorScale1.getColors()) {
			final SColor color = new ColorImpl(getFontHTMLColor(xColor, themes, this));
			colorScale.addColor(color);
		}
		return colorScale;
	}

	//ZSS-1130
	protected SIconSet prepareIconSet(XlsxExtractor.XlsxCfRuleExtractor ctRule) {
		final IconSetImpl iconSet = new IconSetImpl();
		XlsxExtractor.XlsxIconSetExtractor iconSet1 = ctRule.getIconSet();
		for (XlsxExtractor.XlsxCfvoExtractor ctvo : iconSet1.getCfvos()) {
			iconSet.addValueObject(prepareValueObject(ctvo));
		}
		iconSet.setType(iconSet1.getIconSet());
		if (iconSet1.isSetPercent())
			iconSet.setPercent(iconSet1.isPercent());
		if (iconSet1.isSetReverse())
			iconSet.setReverse(iconSet1.isReverse());
		if (iconSet1.isSetShowValue())
			iconSet.setShowValue(iconSet1.isShowValue());

		// KEIKAI-448
		if (iconSet1.getCfIcons() != null) {
			for (XlsxExtractor.XlsxCfIconExtractor icon : iconSet1.getCfIcons()) {
				iconSet.addCfIcon(prepareCFIcon(icon));
			}
		}
		return iconSet;
	}

	//ZSS-1130
	protected void prepareFormulas(ConditionalFormattingRuleImpl cfri,
			XlsxExtractor.XlsxCfRuleExtractor xRule) {
		//ZSS-1142
		Iterator<String> it = xRule.getFormulas().iterator();
		if (it.hasNext()) {
			cfri.setFormula1(it.next());
		} else {
			return;
		}
		if (it.hasNext()) {
			cfri.setFormula2(it.next());
		} else {
			return;
		}
		if (it.hasNext()) {
			cfri.setFormula3(it.next());
		} else {
			return;
		}
	}

	//ZSS-1130
	protected SDataBar prepareDataBar(XlsxExtractor.XlsxCfRuleExtractor ctRule) {
		final DataBarImpl dataBar = new DataBarImpl();
		XlsxExtractor.XlsxDataBarExtractor xDataBar = ctRule.getDataBar();
		for (XlsxExtractor.XlsxCfvoExtractor ctvo : xDataBar.getCfvos()) {
			dataBar.addValueObject(prepareValueObject(ctvo));
		}
		final SColor color = new ColorImpl(getFontHTMLColor(xDataBar.getColor(), themes, this));
		dataBar.setColor(color);
		if (xDataBar.isSetMaxLength())
			dataBar.setMaxLength((int) xDataBar.getMaxLength());
		if (xDataBar.isSetMinLength())
			dataBar.setMinLength((int) xDataBar.getMinLength());
		if (xDataBar.isSetShowValue())
			dataBar.setShowValue(xDataBar.isShowValue());

		if (xDataBar instanceof XlsxExtractor.XlsxExtDataBarExtractor) {
			XlsxExtractor.XlsxExtDataBarExtractor extDataBar = (XlsxExtractor.XlsxExtDataBarExtractor) xDataBar;
			XlsxExtractor.XlsxColorExtractor colorValue = extDataBar.getBorderColor();
			if (colorValue != null) {
				dataBar.setBorderColor(new ColorImpl(getFontHTMLColor(colorValue, themes, this)));
			}
			colorValue = extDataBar.getNegativeBorderColor();
			if (colorValue != null) {
				dataBar.setNegativeBorderColor(new ColorImpl(getFontHTMLColor(colorValue, themes, this)));
			}
			colorValue = extDataBar.getNegativeFillColor();
			if (colorValue != null) {
				dataBar.setNegativeFillColor(new ColorImpl(getFontHTMLColor(colorValue, themes, this)));
			}
			colorValue = extDataBar.getAxisColor();
			if (colorValue != null) {
				dataBar.setAxisColor(new ColorImpl(getFontHTMLColor(colorValue, themes, this)));
			}
			dataBar.setBorder(extDataBar.isBorder());
			dataBar.setGradient(extDataBar.isGradient());
			dataBar.setNegativeBarBorderColorSameAsPositive(
					extDataBar.isNegativeBarBorderColorSameAsPositive());
			dataBar.setNegativeBarColorSameAsPositive(extDataBar.isNegativeBarColorSameAsPositive());
			dataBar.setAxisPosition(extDataBar.getAxisPosition());
			dataBar.setDirection(extDataBar.getDirection());
		}
		return dataBar;
	}

	//ZSS-1130
	protected SCFValueObject prepareValueObject(
			XlsxExtractor.XlsxCfvoExtractor ctvo) {
		CFValueObjectImpl vo = new CFValueObjectImpl();
		if (ctvo.isSetGte())
			vo.setGreaterOrEqual(ctvo.isGte());
		vo.setType(ctvo.getType());
		if (ctvo.isSetVal())
			vo.setValue(ctvo.getVal());

		return vo;
	}

	//KIEKAI-448
	protected SCFIcon prepareCFIcon(XlsxExtractor.XlsxCfIconExtractor cfIcon) {
		final CFIconImpl icon = new CFIconImpl();
		icon.setType(cfIcon.getIconSet());
		icon.setIconId(cfIcon.getIconId());
		return icon;
	}
}
