/* Cell.java

	Purpose:
		
	Description:
		
	History:
		3:54 PM 03/07/2017, Created by jumperchen

Copyright (C) 2017 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer.xml;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import io.keikai.json.JSONAware;
import io.keikai.json.JSONValue;
import io.keikai.util.Converter;

/**
 * see http://www.datypic.com/sc/ooxml/t-ssml_CT_Cell.html
 * @author jumperchen
 */
public class Cell implements JSONNode {
		// attributes
		private int r = -1;
		private int c = -1;
		private int s = -1;
		private CellType t;
		private int cm = -1;
		private int vm = -1;
		private Boolean ph;

		// child elements
		private JSONAware f;
		private Object v;
		private JSONAware is;
		private JSONAware extLst;

	public void addAttributes(Map<String, String> attributes) {
		String r = attributes.remove("r");
		if (r != null) {
			int[] colRow = getIndexes(r);
			this.c = colRow[0];
			this.r = colRow[1];
		}
		String v = attributes.remove("s");
		if (v != null) {
			this.s = Integer.parseInt(v);
		}
		if ((v = attributes.remove("t")) != null) {
			this.t = CellType.valueOf(v);
		}
		if ((v = attributes.remove("cm")) != null) {
			this.cm = Integer.parseInt(v);
		}
		if ((v = attributes.remove("vm")) != null) {
			this.vm = Integer.parseInt(v);
		}
		if ((v = attributes.remove("ph")) != null) {
			this.ph = toBoolean(v);
		}

		if (!attributes.isEmpty()) {
			throw new IllegalArgumentException("Unknown attributes [" + attributes + "]");
		}
	}
	private static int[] getIndexes(String a1Notation) {
		int colIndex = 0;
		int index = 0;
		final int len = a1Notation.length();
		for (; index < len; index++) {
			char code = a1Notation.charAt(index);
			if (code >= 'A' && code <= 'Z') {
				final int digit = code - 64;
				colIndex = colIndex * 26 + digit;
			} else {
				break;
			}
		}

		if (colIndex > 0) {
			colIndex--;
		}

		return new int[]{colIndex, Integer.parseInt(a1Notation.substring(index))};
	}
	public void addElement(String name, JSONAware element) {
		switch(name) {
		case "v":
			if (element instanceof Map) {
				v = (String) ((Map) element).get("_");
				if (t == CellType.s) {
					this.v = Integer.parseInt((String) v);
				} else if (t == null || t == CellType.n) {
					String str = (String) this.v;
					boolean parsed = false;
					if (str.toLowerCase().indexOf('e') == -1) {
						if (str.indexOf('.') == -1) {
							try {
								this.v = Integer.parseInt(str);
								parsed = true;
							} catch (NumberFormatException e) {
								try {
									this.v = Long.parseLong(str);
									parsed = true;
								} catch (NumberFormatException ee) {
								}
							}
						} else {
							try {
								this.v = Double.parseDouble(str);
								parsed = true;
							} catch (NumberFormatException e) {
							}
						}
					}
					if (!parsed) {
						this.v = new BigDecimal(str);
					}
				}
			} else if (v == null) {
				v = NULL_STRING_VALUE;
			}
			break;
		case "f":
			f = element;
			break;
		case "is":
			is = element;
			break;
		case "extLst":
			extLst = element;
			break;
		default:
			throw new IllegalArgumentException("Unknown element [" + name + "]");
		}
	}

	public Object get(String key) {
		switch(key) {
		case "r":
			return Converter.numToAbc(c) + r;
		case "s":
			return s == -1 ? null : String.valueOf(s);
		case "t":
			// KEIKAI-217
			return t == null ?
					((v == NULL_STRING_VALUE || v == null) ? null : CellType.n.toString())
					: String.valueOf(t);
		case "cm":
			return cm == -1 ? null : String.valueOf(cm);
		case "vm":
			return vm == -1 ? null : String.valueOf(vm);
		case "ph":
			return ph == null ? null : (ph ? '1' : '0');
		case "v":
			return v == NULL_STRING_VALUE || v == null ? null : String.valueOf(v);
		case "f":
			return f;
		case "is":
			return is;
		case "extLst":
			return extLst;
		default:
			throw new IllegalArgumentException("Unknown key [" + key + "]");
		}
	}

	public void addContent(String text) {
		if (!text.trim().isEmpty()) // KEIKAI-402
			throw new UnsupportedOperationException("Unsupported content value: [" + text + "]");
	}

	private List getAttributeKeys() {
		List<String> list = new ArrayList<>();
		if (r != -1) {
			list.add("r");
		}
		if (s != -1) {
			list.add("s");
		}
		if (t != null) {
			list.add("t");
		}
		if (cm != -1) {
			list.add("cm");
		}
		if (vm != -1) {
			list.add("vm");
		}
		if (ph != null) {
			list.add("ph");
		}
		return list;
	}
	public String toJSONString() {
		return JSONValue.toJSONString(new HashMap(6){{
			List keys = getAttributeKeys();
			if (!keys.isEmpty()) {
				put("", keys);
			}
			if (r != -1) {
				put("r", Converter.numToAbc(c) + r);
			}
			if (s != -1) {
				put("s", String.valueOf(s));
			}
			if (t != null) {
				put("t", String.valueOf(t));
			}
			if (ph != null) {
				put("ph", ph ? '1' : '0');
			}
			if (cm != -1) {
				put("cm", String.valueOf(cm));
			}
			if (vm != -1) {
				put("vm", String.valueOf(vm));
			}
			if (f != null) {
				put("f", f);
			}
			if (v != null) {
				if (v == NULL_STRING_VALUE) {
					put("v", null);
				} else {
					put("v", new HashMap(1) {{
						put("_", String.valueOf(v));
					}});
				}
			}
			if (is != null) {
				put("is", is);
			}
		}});
	}
	private enum CellType {
		b, n, e, s, str, inlineStr;
	}
	//KEIKAI-196
	private static boolean toBoolean(String s) {
		return "1".equals(s) || Boolean.parseBoolean(s);
	}
}
