/* Row.java

	Purpose:
		
	Description:
		
	History:
		3:49 PM 03/07/2017, Created by jumperchen

Copyright (C) 2017 Potix Corporation. All Rights Reserved.
*/
package io.keikai.importer.xml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import io.keikai.json.JSONArray;
import io.keikai.json.JSONAware;
import io.keikai.json.JSONValue;

/**
 * see http://www.datypic.com/sc/ooxml/t-ssml_CT_Row.html
 * @author jumperchen
 */
public class Row implements JSONNode {
	// attributes
	private int r = -1;
	private String spans;
	private int s = -1;
	private Boolean customFormat;
	private String ht;
	private Boolean hidden;
	private Boolean customHeight;
	private String outlineLevel;
	private Boolean collapsed;
	private Boolean thickTop;
	private Boolean thickBot;
	private Boolean ph;
	private String x14ac;

	private JSONArray<JSONAware> cells = new JSONArray<>();

	public void addAttributes(Map<String, String> attributes) {
		String r = attributes.remove("r");
		if (r != null) {
			this.r = Integer.parseInt(r);
		}
		this.spans = attributes.remove("spans");
		String v = attributes.remove("s");
		if (v != null) {
			this.s = Integer.parseInt(v);
		}
		if ((v = attributes.remove("customFormat"))!= null) {
			this.customFormat = toBoolean(v);
		}
		this.ht = attributes.remove("ht");
		if ((v = attributes.remove("hidden"))!= null) {
			this.hidden = toBoolean(v);
		}
		if ((v = attributes.remove("customHeight"))!= null) {
			this.customHeight = toBoolean(v);
		}
		this.outlineLevel = attributes.remove("outlineLevel");
		if ((v = attributes.remove("collapsed"))!= null) {
			this.collapsed = toBoolean(v);
		}
		if ((v = attributes.remove("thickTop"))!= null) {
			this.thickTop = toBoolean(v);
		}
		if ((v = attributes.remove("thickBot"))!= null) {
			this.thickBot = toBoolean(v);
		}
		if ((v = attributes.remove("ph"))!= null) {
			this.ph = toBoolean(v);
		}
		this.x14ac = attributes.remove("x14ac:dyDescent");
		if (!attributes.isEmpty()) {
			throw new IllegalArgumentException("Unknown attributes [" + attributes + "]");
		}
	}

	public void addElement(String name, JSONAware element) {
		if ("c".equals(name)) {
			cells.add(element);
		} else {
			throw new IllegalArgumentException("Unknown element [" + name + "]");
		}
	}

	public void addContent(String text) {
		if (!text.trim().isEmpty()) // KEIKAI-402
			throw new UnsupportedOperationException("Unsupported content value: [" + text + "]");
	}

	private List getAttributeKeys() {
		List<String> list = new ArrayList<>();
		if (r != -1)
			list.add("r");
		if (spans != null)
			list.add("spans");
		if (s != -1)
			list.add("s");
		if (customFormat != null)
			list.add("customFormat");
		if (ht != null)
			list.add("ht");
		if (hidden != null)
			list.add("hidden");
		if (customHeight != null)
			list.add("customHeight");
		if (outlineLevel != null)
			list.add("outlineLevel");
		if (collapsed != null)
			list.add("collapsed");
		if (thickTop != null)
			list.add("thickTop");
		if (thickBot != null)
			list.add("thickBot");
		if (ph != null)
			list.add("ph");
		if (x14ac != null) {
			list.add("x14ac:dyDescent");
		}
		return list;
	}

	public String toJSONString() {
		return JSONValue.toJSONString(new HashMap(14) {
			{
				if (!cells.isEmpty()) {
					List<String> keys = getAttributeKeys();
					if (!keys.isEmpty()) {
						put("", keys);
					}
					put("c", cells);
				}
				if (r != -1)
					put("r", String.valueOf(r));
				if (spans != null)
					put("spans", spans);
				if (s != -1)
					put("s", String.valueOf(s));
				if (customFormat != null)
					put("customFormat", customFormat ? "1" : "0");
				if (ht != null)
					put("ht", ht);
				if (hidden != null)
					put("hidden", hidden ? "1" : "0");
				if (customHeight != null)
					put("customHeight", customHeight ? "1" : "0");
				if (outlineLevel != null)
					put("outlineLevel", outlineLevel);
				if (collapsed != null)
					put("collapsed", collapsed ? "1" : "0");
				if (thickTop != null)
					put("thickTop", thickTop ? "1" : "0");
				if (thickBot != null)
					put("thickBot", thickBot ? "1" : "0");
				if (ph != null)
					put("ph", ph ? "1" : "0");
				if (x14ac != null)
					put("x14ac", x14ac);
			}
		});
	}

	public JSONArray<JSONAware> getCells() {
		return cells;
	}

	public Object get(String key) {
		switch(key) {
		case "r":
			return r == -1 ? null : String.valueOf(r);
		case "spans":
			return spans;
		case "s":
			return s == -1 ? null : String.valueOf(s);
		case "customFormat":
			return customFormat;
		case "ht":
			return ht;
		case "hidden":
			return hidden;
		case "customHeight":
			return customHeight;
		case "outlineLevel":
			return outlineLevel;
		case "collapsed":
			return collapsed;
		case "thickTop":
			return thickTop;
		case "thickBot":
			return thickBot;
		case "ph":
			return ph;
		case "x14ac":
			return x14ac;
		default:
			throw new IllegalArgumentException("Unknown key [" + key + "]");
		}
	}

	//KEIKAI-196
	private static boolean toBoolean(String s) {
		return "1".equals(s) || Boolean.parseBoolean(s);
	}
}
