/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/01 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.model;
import io.keikai.model.SFill.FillPattern;
import io.keikai.model.SBorder.BorderType;

/**
 * Represent style information e.g. alignment, border, format pattern, font, color, fill pattern, wrap text, and hidden status. It may associate with a book or cells.
 * @author dennis
 * @since 3.5.0
 */
public interface SCellStyle {

	public static final String FORMAT_GENERAL = "General";

	/**
	 * The horizontal alignment
	 * @since 3.5.0
	 */
	public enum Alignment {
		GENERAL, LEFT, CENTER, RIGHT, FILL, JUSTIFY, CENTER_SELECTION
	}

	/**
	 * The vertical alignment
	 * @since 3.5.0
	 */
	public enum VerticalAlignment {
		TOP, CENTER, BOTTOM, JUSTIFY
	}
	
	/**
	 * @return fill foreground-color
	 */
	public SColor getFillColor();

	/**
	 * @return fill background-color
	 * @since 3.6.0
	 */
	public SColor getBackColor();

	/**
	 * Gets the fill/background pattern <br/>
	 * @return the fill pattern
	 */
	public FillPattern getFillPattern();

	/**
	 * Gets the horizontal alignment <br/>
	 * @return the horizontal alignment
	 */
	public Alignment getAlignment();

	/**
	 * Gets vertical alignment <br/>
	 * @return
	 */
	public VerticalAlignment getVerticalAlignment();

	/**
	 * @return true if wrap-text
	 */
	public boolean isWrapText();

	/**
	 * @return left border
	 */
	public BorderType getBorderLeft();

	/**
	 * @return top border
	 */
	public BorderType getBorderTop();

	/**
	 * @return right border
	 */
	public BorderType getBorderRight();

	/**
	 * @return bottom border
	 */
	public BorderType getBorderBottom();

	/**
	 * @return vertical border
	 * @since 3.8.0
	 */
	public BorderType getBorderVertical();

	/**
	 * @return horizontal border
	 * @since 3.8.0
	 */
	public BorderType getBorderHorizontal();

	/**
	 * @return diagonal border
	 * @since 3.8.0
	 */
	public BorderType getBorderDiagonal();

	/**
	 * @return top border color
	 */
	public SColor getBorderTopColor();

	/** 
	 * @return left border color
	 */
	public SColor getBorderLeftColor();

	/**
	 * 
	 * @return bottom border color
	 */
	public SColor getBorderBottomColor();

	/**
	 * @return right border color
	 */
	public SColor getBorderRightColor();
	
	/**
	 * @return vertical border color
	 * @since 3.8.0
	 */
	public SColor getBorderVerticalColor();

	/**
	 * @return horizontal border color
	 * @since 3.8.0
	 */
	public SColor getBorderHorizontalColor();

	/**
	 * @return diagonal border color
	 * @since 3.8.0
	 */
	public SColor getBorderDiagonalColor();

	/**
	 * @return data format
	 */
	public String getDataFormat();
	
	/**
	 * 
	 * @return true if the data format is direct data format, which mean it will not care Locale when formatting
	 */
	public boolean isDirectDataFormat();
	
	/**
	 * 
	 * @return true if locked
	 */
	public boolean isLocked();
	
	/**
	 * 
	 * @return true if hidden
	 */
	public boolean isHidden();


	@Deprecated public void setFillColor(SColor fillColor);

	@Deprecated public void setBackColor(SColor backColor);

	@Deprecated public void setFillPattern(FillPattern fillPattern);

	@Deprecated public void setAlignment(Alignment alignment);

	@Deprecated public void setVerticalAlignment(VerticalAlignment verticalAlignment);

	@Deprecated public void setWrapText(boolean wrapText);

	@Deprecated public void setBorderLeft(BorderType borderLeft);
	@Deprecated public void setBorderLeft(BorderType borderLeft,SColor color);

	@Deprecated public void setBorderTop(BorderType borderTop);
	@Deprecated public void setBorderTop(BorderType borderTop,SColor color);

	@Deprecated public void setBorderRight(BorderType borderRight);
	@Deprecated public void setBorderRight(BorderType borderRight,SColor color);

	@Deprecated public void setBorderBottom(BorderType borderBottom);
	@Deprecated public void setBorderBottom(BorderType borderBottom,SColor color);

	@Deprecated public void setBorderVertical(BorderType borderVertical);
	@Deprecated public void setBorderVertical(BorderType borderVertical,SColor color);

	@Deprecated public void setBorderHorizontal(BorderType borderHorizontal);
	@Deprecated public void setBorderHorizontal(BorderType borderHorizontal,SColor color);

	@Deprecated public void setBorderDiagonal(BorderType borderDiagonal);
	@Deprecated public void setBorderDiagonal(BorderType borderDiagonal,SColor color);

	@Deprecated public void setBorderTopColor(SColor borderTopColor);

	@Deprecated public void setBorderLeftColor(SColor borderLeftColor);

	@Deprecated public void setBorderBottomColor(SColor borderBottomColor);

	@Deprecated public void setBorderRightColor(SColor borderRightColor);

	@Deprecated public void setBorderVerticalColor(SColor color);

	@Deprecated public void setBorderHorizontalColor(SColor color);

	@Deprecated public void setBorderDiagonalColor(SColor color);

	@Deprecated public void setDataFormat(String dataFormat);

	@Deprecated public void setDirectDataFormat(String dataFormat);

	@Deprecated public void setLocked(boolean locked);

	@Deprecated public void setHidden(boolean hidden);
	
	public SFont getFont();

	@Deprecated public void setFont(SFont font);

	@Deprecated public void copyFrom(SCellStyle src);
		
    /**
     * get the degree of rotation for the text in the cell
     * @return rotation degrees (between -90 and 90 degrees); 255 means "vertical text"
     * @since 3.8.0
     */
	public int getRotation();
	@Deprecated public void setRotation(int rotation);
	
	public int getIndention();
	@Deprecated public void setIndention(int indent);

	//ZSS-977
	/**
	 * Returns the border
	 * @return
	 * @since 3.8.0
	 */
	public SBorder getBorder();
	
	//ZSS-977
	/**
	 * Returns the fill
	 * @return
	 * @since 3.8.0
	 */
	public SFill getFill();
	

	//ZSS-1145
	/**
	 * 
	 * @return
	 * @since 3.8.2
	 */
    public boolean isShowDiagonalUpBorder();

	//ZSS-1145
	/**
	 * 
	 * @return
	 * @since 3.8.2
	 */
    public boolean isShowDiagonalDownBorder();
}
