/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/01 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.model;

/**
 * A column of a sheet. But you cannot get cells from it. You should get cell via {@link SSheet#getCell(int, int)}. 
 * @author dennis
 * @since 3.5.0
 */
public interface SColumn extends CellStyleHolder{

	public int getIndex();
	public SSheet getSheet();

	/**
	 * @return TRUE if this is a blank column whose cells has no data, otherwise returns FALSE.
	 */
	public boolean isNull();
	
	public int getWidth();
	public boolean isHidden();
	public boolean isCustomWidth();
	
	public void setWidth(int width);
	public void setHidden(boolean hidden);
	public void setCustomWidth(boolean custom);
	/**
	 * the bestfit attribute
	 * @return is bestfit or not
	 * @since 5.11.0
	 */
	public boolean isBestFit();

	/**
	 * set the bestfit attribute
	 * @since 5.11.0
	 */
	public void setBestFit(boolean bestFit);


	/**
	 * Set column groupings (like groupColumn) in a stream-friendly manner
	 *
	 * <p>
	 *    groupColumns requires all rows in the group to be in the current window.
	 *    This is not always practical.  Instead use setOutlineLevel to
	 *    explicitly set the group level.  Level 1 is the top level group,
	 *    followed by 2, etc.  It is up to the user to ensure that level 2
	 *    groups are correctly nested under level 1, etc.
	 * </p>
	 * <p>Note: The maximum outlines are up to eight levels.
	 * @param outlineLevel     outline level (greater than 0)
	 * @since 6.0.0
	 */
	public void setOutlineLevel(int outlineLevel);

	/**
	 * Returns the column groupings
	 * @since 6.0.0
	 */
	public int getOutlineLevel();

	/**
	 * Returns true if it has "collapsed" attribute
	 * @since 6.0.0
	 */
	public boolean isCollapsed();

	/**
	 * Sets the "collapsed" attribute
	 * @since 6.0.0
	 */
	public void setCollapsed(boolean collapsed);
}
