/* SDataBar.java

	Purpose:
		
	Description:
		
	History:
		Oct 23, 2015 4:32:28 PM, Created by henrichen

	Copyright (C) 2015 Potix Corporation. All Rights Reserved.
*/

package io.keikai.model;

import java.util.List;

/**
 * Condiditional Formatting of "dataBar" type.
 * @author henri
 * @since 3.8.2
 */
public interface SDataBar {
	/**
	 * Returns the value objects.
	 * @return
	 */
	List<SCFValueObject> getCFValueObjects();
	
	/** Returns the color of the bar */
	SColor getColor();
	
	/** Returns the minimum length in percentage of this data bar; default to 10. */
	int getMinLength();
	
	/** Returns the maximum length in percentage of this data bar; default to 90 */
	int getMaxLength();
	
	/** Returns whether show the value in the data bar */
	boolean isShowValue();

	/**Return the border color of the bar in extLst.
	 * since 5.9.0
	 * @return
	 */
	SColor getBorderColor();

	/**Return the negative border color of the bar in extLst.
	 * since 5.9.0
	 * @return
	 */
	SColor getNegativeBorderColor();

	/**Return the negative fill color of the bar in extLst.
	 * since 5.9.0
	 * @return
	 */
	SColor getNegativeFillColor();

	/**Return the axis color of the bar in extLst.
	 * since 5.9.0
	 * @return
	 */
	SColor getAxisColor();

	/**Return whether enables gradient or not in extLst.
	 * Default: true
	 * since 5.9.0
	 * @return
	 */
	boolean isGradient();

	/**Return whether enables border or not in extLst.
	 * Default: false
	 * since 5.9.0
	 * @return
	 */
	boolean isBorder();

	/**Return whether enables negativeBarColorSameAsPositive in extLst.
	 * Default: true
	 * since 5.9.0
	 * @return
	 */
	boolean isNegativeBarColorSameAsPositive();

	/**Return whether enables negativeBarBorderColorSameAsPositive in extLst.
	 * Default: true
	 * since 5.9.0
	 * @return
	 */
	boolean isNegativeBarBorderColorSameAsPositive();

	/**Return the axisPosition of this bar in extLst.
	 * Default: "auto"
	 * since 5.9.0
	 * @return
	 */
	String getAxisPosition();

	/**Return the direction of this bar in extLst.
	 * Default: "leftToRight"
	 * since 5.9.0
	 * @return
	 */
	String getDirection();
}
