package io.keikai.model.impl;

import io.keikai.model.SChartAxis;

import java.io.Serializable;

/**
 * @since 5.3.0
 */
public abstract class AbstractChartAxisAdv implements SChartAxis, Serializable {

    public enum Crosses {
        AUTO_ZERO,
        MIN,
        MAX
    }

    public enum Position {
        BOTTOM,
        LEFT,
        RIGHT,
        TOP
    }

    public enum Orientation {
        MAX_MIN,
        MIN_MAX
    }

    // KEIKAI-405
    public enum TickMark {
        CROSS,
        IN,
        NONE,
        OUT
    }

    public enum TickLabelPosition {
        LOW,
        NONE,
        HIGH,
        NEXT_TO
    }

    /**
     * Returns if this axis is visible
     * @return visible if the axis is visible
     */
    abstract public boolean isVisible();

    /**
     * Returns the position of this axis.
     * @return position the position of this axis
     */
    abstract public Position getPosition();

    /**
     * Returns if this axis has major gridline
     * @return majorGridline if the axis has major gridline
     */
    abstract public boolean hasMajorGridline();

    /**
     * Returns the orientation of the values.
     * @return orientation the orientation of the values
     */
    abstract public Orientation getOrientation();

    /**
     * Returns the position of the tick label.
     * @return tickLabelPosition the position of the tick label.
     */
    abstract public TickLabelPosition getTickLabelPosition();

    /**
     * Returns the id of the axis cross another axis
     */
    abstract public long getCrossAxisId();

    /**
     * Sets the id of the axis cross another axis
     */
    abstract public void setCrossAxisId(long id);

    /**
     * Returns the possible crossing points for the axis.
     * @return axisCrosses the possible crossing points for the axis
     */
    abstract public Crosses getAxisCrosses();
}
