package io.keikai.model.impl;

import io.keikai.model.SBorder;
import io.keikai.model.SBorderBuilder;
import io.keikai.model.SBorderLine;


public class BorderBuilderImpl implements SBorderBuilder {
    final private BookImpl bookImpl;
    final private boolean inStyleTable;

    private SBorder.BorderType borderLeft;
    private String borderLeftColor;

    private SBorder.BorderType borderTop;
    private String borderTopColor;

    private SBorder.BorderType borderRight;
    private String borderRightColor;

    private SBorder.BorderType borderBottom;
    private String borderBottomColor;

    private SBorder.BorderType borderVertical;
    private String borderVerticalColor;

    private SBorder.BorderType borderHorizontal;
    private String borderHorizontalColor;
    private SBorder.BorderType borderDiagonal;
    private String borderDiagonalColor;

    private boolean showDiagonalDown;
    private boolean showDiagonalUp;

    public BorderBuilderImpl(BookImpl bookImpl, SBorder src) {
        this.bookImpl = bookImpl;
        inStyleTable = bookImpl != null;
        if (src != null) {
            borderLeft = src.getBorderLeft();
            borderLeftColor = src.getBorderLeftColor().getHtmlColor();
            borderTop = src.getBorderTop();
            borderTopColor = src.getBorderTopColor().getHtmlColor();
            borderRight = src.getBorderRight();
            borderRightColor = src.getBorderRightColor().getHtmlColor();
            borderBottom = src.getBorderBottom();
            borderBottomColor = src.getBorderBottomColor().getHtmlColor();
            borderVertical = src.getBorderVertical();
            borderVerticalColor = src.getBorderVerticalColor().getHtmlColor();
            borderHorizontal = src.getBorderHorizontal();
            borderHorizontalColor = src.getBorderHorizontalColor().getHtmlColor();
            borderDiagonal = src.getBorderDiagonal();
            borderDiagonalColor = src.getBorderDiagonalColor().getHtmlColor();
            showDiagonalUp = src.isShowDiagonalUpBorder();
            showDiagonalDown = src.isShowDiagonalDownBorder();
        } else {
            borderLeft = DefaultStyle.BORDER_TYPE;
            borderLeftColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            borderTop = DefaultStyle.BORDER_TYPE;
            borderTopColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            borderRight = DefaultStyle.BORDER_TYPE;
            borderRightColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            borderBottom = DefaultStyle.BORDER_TYPE;
            borderBottomColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            borderVertical = DefaultStyle.BORDER_TYPE;
            borderVerticalColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            borderHorizontal = DefaultStyle.BORDER_TYPE;
            borderHorizontalColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            borderDiagonal = DefaultStyle.BORDER_TYPE;
            borderDiagonalColor = DefaultStyle.BORDER_COLOR.getHtmlColor();
            showDiagonalUp = DefaultStyle.BORDER_SHOW_UP;
            showDiagonalDown = DefaultStyle.BORDER_SHOW_DOWN;
        }
    }

    static private SBorderLine toBorderLine(SBorder.BorderType type, String htmlColor) {
        if (type == SBorder.BorderType.NONE) {
            return ImmutableBorderLineImpl.BLANK_BORDER_LINE;
        }
        return new ImmutableBorderLineImpl(type, new ColorImpl(htmlColor));
    }

    static private SBorderLine toBorderLine(BookImpl book, SBorder.BorderType type, String htmlColor) {
        if (type == SBorder.BorderType.NONE) {
            return ImmutableBorderLineImpl.BLANK_BORDER_LINE;
        }
        return new ImmutableBorderLineImpl(type, book.createColor(htmlColor));
    }

    @Override
    public SBorder build() {
        final boolean showDiagonal = (showDiagonalDown || showDiagonalUp) &&
                borderDiagonal != SBorder.BorderType.NONE;
        if (inStyleTable) {
            final SBorderLine diagonalLine = showDiagonal ?
                    new ImmutableBorderLineImpl(borderDiagonal, bookImpl.createColor(borderDiagonalColor),
                            showDiagonalUp, showDiagonalDown) : ImmutableBorderLineImpl.BLANK_BORDER_LINE;
            return new ImmutableBorderImpl(toBorderLine(bookImpl, borderLeft, borderLeftColor),
                    toBorderLine(bookImpl, borderTop, borderTopColor),
                    toBorderLine(bookImpl, borderRight, borderRightColor),
                    toBorderLine(bookImpl, borderBottom, borderBottomColor),
                    diagonalLine,
                    toBorderLine(bookImpl, borderVertical, borderVerticalColor),
                    toBorderLine(bookImpl, borderHorizontal, borderHorizontalColor));
        } else {
            final SBorderLine diagonalLine = showDiagonal ?
                    new ImmutableBorderLineImpl(borderDiagonal, new ColorImpl(borderDiagonalColor),
                            showDiagonalUp, showDiagonalDown) : ImmutableBorderLineImpl.BLANK_BORDER_LINE;
            return new ImmutableBorderImpl(toBorderLine(borderLeft, borderLeftColor),
                    toBorderLine(borderTop, borderTopColor),
                    toBorderLine(borderRight, borderRightColor),
                    toBorderLine(borderBottom, borderBottomColor),
                    diagonalLine,
                    toBorderLine(borderVertical, borderVerticalColor),
                    toBorderLine(borderHorizontal, borderHorizontalColor)
            );
        }
    }

    @Override
    public SBorderBuilder borderLeft(SBorder.BorderType type) {
        borderLeft = type;
        return this;
    }

    @Override
    public SBorderBuilder borderLeftColor(String color) {
        borderLeftColor = color;
        return this;
    }

    @Override
    public SBorderBuilder borderTop(SBorder.BorderType type) {
        borderTop = type;
        return this;
    }

    @Override
    public SBorderBuilder borderTopColor(String color) {
        borderTopColor = color;
        return this;
    }

    @Override
    public SBorderBuilder borderRight(SBorder.BorderType type) {
        borderRight = type;
        return this;
    }

    @Override
    public SBorderBuilder borderRightColor(String color) {
        borderRightColor = color;
        return this;
    }

    @Override
    public SBorderBuilder borderBottom(SBorder.BorderType type) {
        borderBottom = type;
        return this;
    }

    @Override
    public SBorderBuilder borderBottomColor(String color) {
        borderBottomColor = color;
        return this;
    }

    @Override
    public SBorderBuilder borderVertical(SBorder.BorderType type) {
        borderVertical = type;
        return this;
    }

    @Override
    public SBorderBuilder borderVerticalColor(String color) {
        borderVerticalColor = color;
        return this;
    }

    @Override
    public SBorderBuilder borderHorizontal(SBorder.BorderType type) {
        borderHorizontal = type;
        return this;
    }

    @Override
    public SBorderBuilder borderHorizontalColor(String color) {
        borderHorizontalColor = color;
        return this;
    }

    @Override
    public SBorderBuilder borderDiagonal(SBorder.BorderType type) {
        borderDiagonal = type;
        return this;
    }

    @Override
    public SBorderBuilder borderDiagonalColor(String color) {
        borderDiagonalColor = color;
        return this;
    }

    @Override
    public SBorderBuilder showDiagonalUpBorder(boolean show) {
        showDiagonalUp = show;
        return this;
    }

    @Override
    public SBorderBuilder showDiagonalDownBorder(boolean show) {
        showDiagonalDown = show;
        return this;
    }
}
