/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/01 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.model.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;

import io.keikai.model.SChartAxis;
import io.keikai.model.SColor;
import io.keikai.model.SSheet;
import io.keikai.model.ViewAnchor;
import io.keikai.model.chart.*;
import io.keikai.model.impl.chart.*;
import io.keikai.model.util.Validations;

/**
 * @author dennis
 * @since 3.5.0
 */
public class ChartImpl extends AbstractChartAdv {
    private static final long serialVersionUID = 1L;
    private final String _id;
    private final List<SChartAxis> _valueAxises = new ArrayList<>();
    private final List<SChartAxis> _categoryAxises = new ArrayList<>();
    private final List<SChartAxis> _dateAxises = new ArrayList<>();
    private final List<SChartData> _chartSeries = new ArrayList<>();
    private ViewAnchor _anchor;
    private String _title;
    private String _xAxisTitle;
    private String _yAxisTitle;
    private AbstractSheetAdv _sheet;
    private ChartLegendPosition _legendPosition;
    private boolean _threeD;
    private boolean _emptyAsGaps;
    private boolean _sparkline;
    private boolean _winlose;
    //ZSS-830
    private boolean _rAngAx = true; //default true
    private int _rotX = 0;
    private int _rotY = 0;
    private int _hPercent = 100;
    private int _depthPercent = 100;
    private int _perspective = 30;
    private int _dataCounter;
    private boolean _showTitle;
    private Boolean _autoTitleDeleted = null;
    private boolean _plotOnlyVisibleCells = true; //default true
    private boolean _showDataLabelsOverMax = true; //default true
    private List<SColor> colors = new ArrayList<>();
    private boolean _sparklineNegativePointVisible;
    private boolean _sparklineFirstPointVisible;
    private boolean _sparklineLastPointVisible;
    private boolean _sparklineHighPointVisible;
    private boolean _sparklineLowPointVisible;
    private boolean _sparklineMarkPointVisible;
    private SColor _sparklineSeriesColor;
    private SColor _sparklineNegativePointColor;
    private SColor _sparklineFirstPointColor;
    private SColor _sparklineLastPointColor;
    private SColor _sparklineHighPointColor;
    private SColor _sparklineLowPointColor;
    private SColor _sparklineMarkPointColor;

    private static final AtomicLong axisIdCount = new AtomicLong(0);

    private static long generateAxisId() {
        return axisIdCount.getAndIncrement();
    }

    private static void compareAxisId(long id) {
        if (axisIdCount.get() < id) {
            axisIdCount.set(id + 1);
        }
    }

    public ChartImpl(AbstractSheetAdv sheet, String id, ChartType type, ViewAnchor anchor) {
        Validations.argNotNull(anchor);
        this._sheet = sheet;
        this._id = id;
        this._anchor = anchor;
        addChartData(createChartData(type));
    }

    private ChartImpl(AbstractSheetAdv sheet, String id, ViewAnchor anchor) {
        Validations.argNotNull(anchor);
        this._sheet = sheet;
        this._id = id;
        this._anchor = anchor;
    }

    @Override
    public SSheet getSheet() {
        checkOrphan();
        return _sheet;
    }

    @Override
    public String getId() {
        return _id;
    }

    @Override
    public ViewAnchor getAnchor() {
        return _anchor;
    }

    @Override
    public void setAnchor(ViewAnchor anchor) {
        Validations.argNotNull(anchor);
        this._anchor = anchor;
    }

    @Override
    public ChartType getType() {
        final SChartData data = getData();
        if (data instanceof SGeneralChartData) {
            return ((SGeneralChartData) data).getType();
        }
        return null;
    }

    @Override
    public SChartData getData() {
        if (_chartSeries.isEmpty()) {
            return null;
        }
        return _chartSeries.get(0);
    }

    @Override
    public String getTitle() {
        return _title;
    }

    @Override
    public void setTitle(String title) {
        this._title = title;
    }

    @Override
    public String getXAxisTitle() {
        return _xAxisTitle;
    }

    @Override
    public void setXAxisTitle(String xAxisTitle) {
        this._xAxisTitle = xAxisTitle;
    }

    @Override
    public String getYAxisTitle() {
        return _yAxisTitle;
    }

    @Override
    public void setYAxisTitle(String yAxisTitle) {
        this._yAxisTitle = yAxisTitle;
    }

    @Override
    public void destroy() {
        checkOrphan();

        ((ChartDataAdv) getData()).destroy();

        _sheet = null;
    }

    @Override
    public void checkOrphan() {
        if (_sheet == null) {
            throw new IllegalStateException("doesn't connect to parent");
        }
    }

    @Override
    public ChartLegendPosition getLegendPosition() {
        return _legendPosition;
    }

    @Override
    public void setLegendPosition(ChartLegendPosition pos) {
        this._legendPosition = pos;
    }

    @Override
    public ChartGrouping getGrouping() {
        if (!_chartSeries.isEmpty()) {
            final SChartData data = _chartSeries.get(0);
            if (data instanceof GroupingChartDataImpl) {
                return ((GroupingChartDataImpl) data).getGrouping();
            }
        }
        return null;
    }

    @Override
    public void setGrouping(ChartGrouping grouping) {
        if (!_chartSeries.isEmpty()) {
            final SChartData data = _chartSeries.get(0);
            if (data instanceof GroupingChartDataImpl) {
                ((GroupingChartDataImpl) data).setGrouping(grouping);
            }
        }
    }

    @Override
    public BarDirection getBarDirection() {
        if (!_chartSeries.isEmpty()) {
            final SChartData data = _chartSeries.get(0);
            if (data instanceof BarChartDataImpl) {
                return ((BarChartDataImpl) data).getBarDirection();
            }
        }
        return null;
    }

    @Override
    public boolean isThreeD() {
        return _threeD;
    }

    @Override
    public void setThreeD(boolean threeD) {
        this._threeD = threeD;
    }

    @Override
    public boolean isEmptyAsGaps() {
        return _emptyAsGaps;
    }

    @Override
    public void setEmptyAsGaps(boolean emptyAsGaps) {
        _emptyAsGaps = emptyAsGaps;
    }

    @Override
    public boolean isSparkline() {
        return _sparkline;
    }

    @Override
    public void setSparkline(boolean sparkline) {
        _sparkline = sparkline;
    }

    @Override
    public boolean isWinloseSpark() {
        return _winlose;
    }

    @Override
    public void setWinloseSpark(boolean winlose) {
        _winlose = winlose;
    }

    //ZSS-688
    //@since 3.6.0
    /*package*/ ChartImpl cloneChartImpl(AbstractSheetAdv sheet) {
        final ChartImpl tgt = new ChartImpl(sheet, this._id, this._anchor.cloneViewAnchor());
        final Map<Long, Long> catAxMap = new HashMap<>();
        final Map<Long, Long> dateAxMap = new HashMap<>();
        final Map<Long, Long> valAxMap = new HashMap<>();
        for (SChartAxis axis : _valueAxises) {
            final long newId = generateAxisId();
            tgt.addValueAxis(((ValueAxisImpl) axis).cloneValueAxisImpl(newId));
            catAxMap.put(axis.getId(), newId);
        }
        for (SChartAxis axis : _categoryAxises) {
            final long newId = generateAxisId();
            tgt.addCategoryAxis(((CategoryAxisImpl) axis).cloneCategoryAxisImpl(newId));
            valAxMap.put(axis.getId(), newId);
        }
        for (SChartAxis axis : _dateAxises) {
            final long newId = generateAxisId();
            tgt.addDateAxis(((DateAxisImpl) axis).cloneDateAxisImpl(newId));
            dateAxMap.put(axis.getId(), newId);
        }
        for (SChartAxis axis : tgt._categoryAxises) {
            if (axis instanceof CategoryAxisImpl) {
                final CategoryAxisImpl catAx = (CategoryAxisImpl) axis;
                final long crossCatAxId = catAx.getCrossAxisId();
                if (catAxMap.containsKey(crossCatAxId)) {
                    catAx.setCrossAxisId(catAxMap.get(crossCatAxId));
                }
            }
        }
        for (SChartAxis axis : tgt._dateAxises) {
            if (axis instanceof DateAxisImpl) {
                final DateAxisImpl dateAx = (DateAxisImpl) axis;
                final long crossDateAxId = dateAx.getCrossAxisId();
                if (dateAxMap.containsKey(crossDateAxId)) {
                    dateAx.setCrossAxisId(dateAxMap.get(crossDateAxId));
                }
            }
        }
        for (SChartAxis axis : tgt._valueAxises) {
            if (axis instanceof ValueAxisImpl) {
                final ValueAxisImpl valAx = (ValueAxisImpl) axis;
                final long crossValAxId = valAx.getCrossAxisId();
                if (valAxMap.containsKey(crossValAxId)) {
                    valAx.setCrossAxisId(valAxMap.get(crossValAxId));
                }
            }
        }
        for (SChartData chartData : _chartSeries) {
            if (chartData instanceof GeneralChartDataImpl) {
                final GeneralChartDataImpl dataImpl = (GeneralChartDataImpl) chartData;
                final GeneralChartDataImpl cloned = (GeneralChartDataImpl) tgt.createChartData(dataImpl.getType());
                cloned.copyFrom(dataImpl);
                tgt.addChartData(cloned);
            }
        }
        tgt._anchor = this._anchor.cloneViewAnchor();
        tgt._title = this._title;
        tgt._xAxisTitle = this._xAxisTitle;
        tgt._yAxisTitle = this._yAxisTitle;
        tgt._legendPosition = this._legendPosition;
        tgt._threeD = this._threeD;
        tgt._emptyAsGaps = this._emptyAsGaps;
        tgt._sparkline = this._sparkline;
        tgt._winlose = this._winlose;
        tgt._rAngAx = this._rAngAx;
        tgt._rotX = this._rotX;
        tgt._rotY = this._rotY;
        tgt._hPercent = this._hPercent;
        tgt._perspective = this._perspective;
        tgt._autoTitleDeleted = this._autoTitleDeleted;
        tgt._plotOnlyVisibleCells = this._plotOnlyVisibleCells;
        tgt._showDataLabelsOverMax = this._showDataLabelsOverMax;
        if (this.colors != null) {
            tgt.colors = new ArrayList<>(this.colors);
        }
        return tgt;
    }

    @Override
    public void addValueAxis(SChartAxis axis) {
        compareAxisId(axis.getId());
        _valueAxises.add(axis);
    }

    @Override
    public void addCategoryAxis(SChartAxis axis) {
        compareAxisId(axis.getId());
        _categoryAxises.add(axis);
    }

    @Override
    public void addDateAxis(SChartAxis axis) {
        compareAxisId(axis.getId());
        _dateAxises.add(axis);
    }

    @Override
    public List<SChartAxis> getValueAxises() {
        return _valueAxises;
    }

    @Override
    public List<SChartAxis> getCategoryAxises() {
        return _categoryAxises;
    }

    @Override
    public List<SChartAxis> getDateAxises() {
        return _dateAxises;
    }

    //ZSS-830
    @Override
    public int getRotX() {
        return _rotX;
    }

    //ZSS-830
    @Override
    public void setRotX(int rotX) {
        _rotX = rotX;
    }

    //ZSS-830
    @Override
    public int getRotY() {
        return _rotY;
    }

    //ZSS-830
    @Override
    public void setRotY(int rotY) {
        _rotY = rotY;
    }

    //ZSS-830
    @Override
    public int getPerspective() {
        return _perspective;
    }

    //ZSS-830
    @Override
    public void setPerspective(int perspective) {
        _perspective = perspective;
    }

    //ZSS-830
    @Override
    public int getHPercent() {
        return _hPercent;
    }

    //ZSS-830
    @Override
    public void setHPercent(int percent) {
        _hPercent = percent;
    }

    //ZSS-830
    @Override
    public int getDepthPercent() {
        return _depthPercent;
    }

    //ZSS-830
    @Override
    public void setDepthPercent(int percent) {
        _depthPercent = percent;
    }

    //ZSS-830
    @Override
    public boolean isRightAngleAxes() {
        return _rAngAx;
    }

    //ZSS-830
    @Override
    public void setRightAngleAxes(boolean b) {
        _rAngAx = b;
    }

    @Override
    public int getBarOverlap() {
        if (!_chartSeries.isEmpty()) {
            final SChartData data = _chartSeries.get(0);
            if (data instanceof BarChartDataImpl) {
                return ((BarChartDataImpl) data).getBarOverlap();
            }
        }
        return 0;
    }

    @Override
    public void setBarOverlap(int overlap) {
        if (_chartSeries.isEmpty()) {
            return;
        }
        final SChartData data = _chartSeries.get(0);
        if (data instanceof BarChartDataImpl) {
            ((BarChartDataImpl) data).setBarOverlap(overlap);
        }
    }

    @Override
    public boolean isPlotOnlyVisibleCells() {
        return _plotOnlyVisibleCells;
    }

    @Override
    public void setPlotOnlyVisibleCells(boolean plotOnlyVisibleCells) {
        _plotOnlyVisibleCells = plotOnlyVisibleCells;

    }

    @Override
    public List<SColor> getSeriesColors() {
        return colors;
    }

    @Override
    public void setSeriesColors(List<SColor> colors) {
        this.colors = colors;
    }

    @Override
    public List<SChartData> getChartData() {
        return _chartSeries;
    }


    @Override
    public SChartData createChartData(ChartType type) {
        switch (type) {
            case AREA:
                return new AreaChartDataImpl(this, toId());
            case BAR:
                return new BarChartDataImpl(this, toId(), BarDirection.HORIZONTAL);
            case BUBBLE:
                return new BubbleChartDataImpl(this, toId());
            case COLUMN:
                return new BarChartDataImpl(this, toId(), BarDirection.VERTICAL);
            case DOUGHNUT:
                return new DoughnutChartDataImpl(this, toId());
            case LINE:
                return new LineChartDataImpl(this, toId());
            case PIE:
                return new PieChartDataImpl(this, toId());
            case SCATTER:
                return new ScatterChartDataImpl(this, toId());
            // not supported
            case OF_PIE:
            case RADAR:
            case STOCK:
            case SURFACE:
                return new UnsupportedChartDataImpl(this);
        }
        throw new UnsupportedOperationException("Unsupported chart type " + type);
    }

    @Override
    public void addChartData(SChartData chartData) {
        _chartSeries.add(chartData);
    }

    @Override
    public Boolean isAutoTitleDeleted() {
        return _autoTitleDeleted;
    }

    @Override
    public void setAutoTitleDeleted(Boolean autoTitleDeleted) {
        _autoTitleDeleted = autoTitleDeleted;
    }

    @Override
    public boolean isShowTitle() {
        return _showTitle;
    }

    @Override
    public void setShowTitle(boolean showTitle) {
        _showTitle = showTitle;
    }

    @Override
    public boolean isShowDataLabelsOverMax() {
        return _showDataLabelsOverMax;
    }

    @Override
    public void setShowDataLabelsOverMax(boolean showDataLabelsOverMax) {
        _showDataLabelsOverMax = showDataLabelsOverMax;
    }

    @Override
    public void setSparklineNegativePointVisible(boolean visible) {
        _sparklineNegativePointVisible = visible;
    }

    @Override
    public boolean isSparklineNegativePointVisible() {
        return _sparklineNegativePointVisible;
    }

    @Override
    public void setSparklineFirstPointVisible(boolean visible) {
        _sparklineFirstPointVisible = visible;
    }

    @Override
    public boolean isSparklineFirstPointVisible() {
        return _sparklineFirstPointVisible;
    }

    @Override
    public void setSparklineLastPointVisible(boolean visible) {
        _sparklineLastPointVisible = visible;
    }

    @Override
    public boolean isSparklineLastPointVisible() {
        return _sparklineLastPointVisible;
    }

    @Override
    public void setSparklineHighPointVisible(boolean visible) {
        _sparklineHighPointVisible = visible;
    }

    @Override
    public boolean isSparklineHighPointVisible() {
        return _sparklineHighPointVisible;
    }

    @Override
    public void setSparklineLowPointVisible(boolean visible) {
        _sparklineLowPointVisible = visible;
    }

    @Override
    public boolean isSparklineLowPointVisible() {
        return _sparklineLowPointVisible;
    }

    @Override
    public void setSparklineMarkerVisible(boolean visible) {
        _sparklineMarkPointVisible = visible;
    }

    @Override
    public boolean isSparklineMarkerVisible() {
        return _sparklineMarkPointVisible;
    }

    @Override
    public void setSparklineSeriesColor(SColor color) {
        _sparklineSeriesColor = color;
    }

    @Override
    public SColor getSparklineSeriesColor() {
        return _sparklineSeriesColor;
    }

    @Override
    public void setSparklineNegativePointColor(SColor color) {
        _sparklineNegativePointColor = color;
    }

    @Override
    public SColor getSparklineNegativePointColor() {
        return _sparklineNegativePointColor;
    }

    @Override
    public void setSparklineFirstPointColor(SColor color) {
        _sparklineFirstPointColor = color;
    }

    @Override
    public SColor getSparklineFirstPointColor() {
        return _sparklineFirstPointColor;
    }

    @Override
    public void setSparklineLastPointColor(SColor color) {
        _sparklineLastPointColor = color;
    }

    @Override
    public SColor getSparklineLastPointColor() {
        return _sparklineLastPointColor;
    }

    @Override
    public void setSparklineHighPointColor(SColor color) {
        _sparklineHighPointColor = color;
    }

    @Override
    public SColor getSparklineHighPointColor() {
        return _sparklineHighPointColor;
    }

    @Override
    public void setSparklineLowPointColor(SColor color) {
        _sparklineLowPointColor = color;
    }

    @Override
    public SColor getSparklineLowPointColor() {
        return _sparklineLowPointColor;
    }

    @Override
    public void setSparklineMarkerColor(SColor color) {
        _sparklineMarkPointColor = color;
    }

    @Override
    public SColor getSparklineMarkerColor() {
        return _sparklineMarkPointColor;
    }

    private String toId() {
        return String.format("%s-%s", _id, _dataCounter++);
    }
}
