package io.keikai.model.impl;

import io.keikai.model.SFill;
import io.keikai.model.SFillBuilder;


public class FillBuilderImpl implements SFillBuilder {
    final private BookImpl bookImpl;
    final private boolean inStyleTable;

    private String fillColor;
    private String backColor;
    private SFill.FillPattern fillPattern;

    public FillBuilderImpl(BookImpl bookImpl) {
        this(bookImpl, null);
    }

    public FillBuilderImpl(BookImpl bookImpl, SFill src) {
        this.bookImpl = bookImpl;
        this.inStyleTable = bookImpl != null;
        if (src != null) {
            this.fillPattern = src.getFillPattern();
            this.fillColor = src.getFillColor().getHtmlColor();
            this.backColor = src.getBackColor().getHtmlColor();
        } else {
            this.fillPattern = DefaultStyle.FILL_PATTERN;
            this.fillColor = DefaultStyle.FONT_COLOR.getHtmlColor();
            this.backColor = DefaultStyle.BACK_COLOR.getHtmlColor();
        }
    }

    @Override
    public SFill build() {
        if (inStyleTable) {
            return new ImmutableFillImpl(this.fillPattern, bookImpl.createColor(fillColor), bookImpl.createColor(backColor));
        } else {
            return new ImmutableFillImpl(this.fillPattern, fillColor, backColor);
        }
    }

    @Override
    public SFillBuilder fillColor(String color) {
        fillColor = color;
        return this;
    }

    @Override
    public SFillBuilder backColor(String color) {
        backColor = color;
        return this;
    }

    @Override
    public SFillBuilder fillPattern(SFill.FillPattern pattern) {
        fillPattern = pattern;
        return this;
    }
}
