package io.keikai.model.impl;

import io.keikai.model.SFont;

public class FontBuilderImpl implements AbstractFontBuilderAdv {
    final private BookImpl bookImpl;
    final private boolean inStyleTable;

    private String name;
    private String color;
    private SFont.Boldweight boldweight;
    private boolean italic;
    private boolean strikeout;
    private SFont.Underline underline;
    private SFont.TypeOffset typeOffset;
    private int heightPoints;

    private boolean overrideName;
    private boolean overrideColor;
    private boolean overrideBold;
    private boolean overrideItalic;
    private boolean overrideStrikeout;
    private boolean overrideUnderline;
    private boolean overrideHeightPoints;
    private boolean overrideTypeOffset;

    FontBuilderImpl(BookImpl bookImpl, boolean inStyleTable) {
        this(bookImpl, null, inStyleTable);
    }

    FontBuilderImpl(BookImpl bookImpl, SFont src, boolean inStyleTable) {
        this.bookImpl = bookImpl;
        this.inStyleTable = inStyleTable;
        if (src != null) {
            name = src.getName();
            color = src.getColor().getHtmlColor();
            boldweight = src.getBoldweight();
            italic = src.isItalic();
            strikeout = src.isStrikeout();
            underline = src.getUnderline();
            heightPoints = src.getHeightPoints();
            typeOffset = src.getTypeOffset();
        } else {
            name = DefaultStyle.FONT_NAME;
            color = DefaultStyle.FONT_COLOR.getHtmlColor();
            boldweight = DefaultStyle.FONT_BOLDWEIGHT;
            italic = DefaultStyle.FONT_ITALIC;
            strikeout = DefaultStyle.FONT_STRIKEOUT;
            underline = DefaultStyle.FONT_UNDERLINE;
            heightPoints = DefaultStyle.FONT_HEIGHT_POINTS;
            typeOffset = DefaultStyle.FONT_TYPE_OFFSET;
        }
    }


    @Override
    public SFont build() {
        if (inStyleTable) {
            final ImmutableFontImpl font = new ImmutableFontImpl(name, bookImpl.createColor(color),
                    boldweight, italic, strikeout, underline, heightPoints, typeOffset, overrideName,
                    overrideColor, overrideBold, overrideItalic, overrideStrikeout, overrideUnderline,
                    overrideHeightPoints, overrideTypeOffset);
            return bookImpl.getOrAddFont(font);
        } else {
            return new ImmutableFontImpl(name, new ColorImpl(color),
                    boldweight, italic, strikeout, underline, heightPoints, typeOffset, overrideName,
                    overrideColor, overrideBold, overrideItalic, overrideStrikeout, overrideUnderline,
                    overrideHeightPoints, overrideTypeOffset);
        }
    }


    @Override
    public AbstractFontBuilderAdv color(String color) {
        this.color = color;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv name(String name) {
        this.name = name;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv boldweight(SFont.Boldweight boldweight) {
        this.boldweight = boldweight;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv heightPoints(int heightPoints) {
        this.heightPoints = heightPoints < 0 ? DefaultStyle.FONT_HEIGHT_POINTS : heightPoints;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv italic(boolean italic) {
        this.italic = italic;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv strikeout(boolean strikeout) {
        this.strikeout = strikeout;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv typeOffset(SFont.TypeOffset typeOffset) {
        this.typeOffset = typeOffset;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv underline(SFont.Underline underline) {
        this.underline = underline;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideName(boolean b) {
        overrideName = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideColor(boolean b) {
        overrideColor = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideBold(boolean b) {
        overrideBold = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideItalic(boolean b) {
        overrideItalic = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideStrikeout(boolean b) {
        overrideStrikeout = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideUnderline(boolean b) {
        overrideUnderline = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideHeightPoints(boolean b) {
        overrideHeightPoints = b;
        return this;
    }

    @Override
    public AbstractFontBuilderAdv overrideTypeOffset(boolean b) {
        overrideTypeOffset = b;
        return this;
    }
}
