package io.keikai.model.impl;

import io.keikai.model.SBook;
import io.keikai.model.SBorder;
import io.keikai.model.SBorderLine;
import io.keikai.model.SColor;
import org.zkoss.lang.Objects;

public class ImmutableBorderImpl extends AbstractBorderAdv {
    private static final long serialVersionUID = 8668618009034300532L;
    final private SBorderLine leftLine;
    final private SBorderLine topLine;
    final private SBorderLine rightLine;
    final private SBorderLine bottomLine;
    final private SBorderLine diagonalLine;
    final private SBorderLine verticalLine;
    final private SBorderLine horizontalLine;

    final private int hashCode;

    // ZSS-977, used for extra styles
    public ImmutableBorderImpl(SBorderLine left, SBorderLine top,
                               SBorderLine right, SBorderLine bottom, SBorderLine diagonal,
                               SBorderLine vertical, SBorderLine horizontal) {
        leftLine = left;
        topLine = top;
        rightLine = right;
        bottomLine = bottom;
        diagonalLine = diagonal;
        verticalLine = vertical;
        horizontalLine = horizontal;
        hashCode = calcHashCode(this);
    }

    static private BorderType getBorderType(SBorderLine borderline) {
        if (borderline == null) {
            return DefaultStyle.BORDER_TYPE;
        }
        return borderline.getBorderType();
    }

    static private SColor getBorderColor(SBorderLine borderline) {
        if (borderline == null) {
            return DefaultStyle.BORDER_COLOR;
        }
        return borderline.getColor();
    }

    static private int calcHashCode(ImmutableBorderImpl impl) {
        int hash = impl.leftLine == null ? 0 : impl.leftLine.hashCode();
        hash = hash * 31 + (impl.topLine == null ? 0 : impl.topLine.hashCode());
        hash = hash * 31 + (impl.rightLine == null ? 0 : impl.rightLine.hashCode());
        hash = hash * 31 + (impl.bottomLine == null ? 0 : impl.bottomLine.hashCode());
        hash = hash * 31 + (impl.diagonalLine == null ? 0 : impl.diagonalLine.hashCode());
        hash = hash * 31 + (impl.verticalLine == null ? 0 : impl.verticalLine.hashCode());
        hash = hash * 31 + (impl.horizontalLine == null ? 0 : impl.horizontalLine.hashCode());
        return hash;
    }

    @Override
    public BorderType getBorderLeft() {
        return getBorderType(leftLine);
    }

    @Override
    public void setBorderLeft(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border left");
    }

    @Override
    public BorderType getBorderTop() {
        return getBorderType(topLine);
    }

    @Override
    public void setBorderTop(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border top");
    }

    @Override
    public BorderType getBorderRight() {
        return getBorderType(rightLine);
    }

    @Override
    public void setBorderRight(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border right");
    }

    @Override
    public BorderType getBorderBottom() {
        return getBorderType(bottomLine);
    }

    @Override
    public void setBorderBottom(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border bottom");
    }

    @Override
    public BorderType getBorderVertical() {
        return getBorderType(verticalLine);
    }

    @Override
    public void setBorderVertical(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border vertical");
    }

    @Override
    public BorderType getBorderHorizontal() {
        return getBorderType(horizontalLine);
    }

    @Override
    public void setBorderHorizontal(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border horizontal");
    }

    @Override
    public BorderType getBorderDiagonal() {
        return getBorderType(diagonalLine);
    }

    @Override
    public void setBorderDiagonal(BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border diagonal");
    }

    @Override
    public SColor getBorderTopColor() {
        return getBorderColor(topLine);
    }

    @Override
    public void setBorderTopColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border top color");
    }

    @Override
    public SColor getBorderLeftColor() {
        return getBorderColor(leftLine);
    }

    @Override
    public void setBorderLeftColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border left color");
    }

    @Override
    public SColor getBorderBottomColor() {
        return getBorderColor(bottomLine);
    }

    @Override
    public void setBorderBottomColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border bottom color");
    }

    @Override
    public SColor getBorderRightColor() {
        return getBorderColor(rightLine);
    }

    @Override
    public void setBorderRightColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border right color");
    }

    @Override
    public SColor getBorderVerticalColor() {
        return getBorderColor(verticalLine);
    }

    @Override
    public void setBorderVerticalColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border vertical color");
    }

    @Override
    public SColor getBorderHorizontalColor() {
        return getBorderColor(horizontalLine);
    }

    @Override
    public void setBorderHorizontalColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border horizontal color");
    }

    @Override
    public SColor getBorderDiagonalColor() {
        return getBorderColor(diagonalLine);
    }

    @Override
    public void setBorderDiagonalColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set border diagonal color");
    }

    @Override
    public boolean isShowDiagonalUpBorder() {
        return diagonalLine == null ? DefaultStyle.BORDER_SHOW_UP : diagonalLine.isShowDiagonalUpBorder();
    }

    @Override
    public void setShowDiagonalUpBorder(boolean show) {
        throw new UnsupportedOperationException("doesn't support set show diagonal up border");
    }

    @Override
    public boolean isShowDiagonalDownBorder() {
        return diagonalLine == null ? DefaultStyle.BORDER_SHOW_DOWN : diagonalLine.isShowDiagonalDownBorder();
    }

    @Override
    public void setShowDiagonalDownBorder(boolean show) {
        throw new UnsupportedOperationException("doesn't support set show diagonal down border");
    }

    @Override
    String getStyleKey() {
        return (leftLine == null ? "" : ((AbstractBorderLineAdv) leftLine).getStyleKey()) +
                (topLine == null ? "" : ((AbstractBorderLineAdv) topLine).getStyleKey()) +
                (rightLine == null ? "" : ((AbstractBorderLineAdv) rightLine).getStyleKey()) +
                (bottomLine == null ? "" : ((AbstractBorderLineAdv) bottomLine).getStyleKey()) +
                (diagonalLine == null ? "" : ((AbstractBorderLineAdv) diagonalLine).getStyleKey()) +
                (verticalLine == null ? "" : ((AbstractBorderLineAdv) verticalLine).getStyleKey()) +
                (horizontalLine == null ? "" : ((AbstractBorderLineAdv) horizontalLine).getStyleKey());
    }

    public SBorderLine getLeftLine() {
        return leftLine;
    }

    public SBorderLine getTopLine() {
        return topLine;
    }

    public SBorderLine getRightLine() {
        return rightLine;
    }

    public SBorderLine getBottomLine() {
        return bottomLine;
    }

    public SBorderLine getDiagonalLine() {
        return diagonalLine;
    }

    public SBorderLine getVerticalLine() {
        return verticalLine;
    }

    public SBorderLine getHorizontalLine() {
        return horizontalLine;
    }

    @Override
    public boolean equals(Object other) {
        if (other == this) {
            return true;
        }
        if (!(other instanceof ImmutableBorderImpl)) {
            return false;
        }
        final ImmutableBorderImpl o = (ImmutableBorderImpl) other;
        return Objects.equals(this.leftLine, o.leftLine)
                && Objects.equals(this.topLine, o.topLine)
                && Objects.equals(this.rightLine, o.rightLine)
                && Objects.equals(this.bottomLine, o.bottomLine)
                && Objects.equals(this.diagonalLine, o.diagonalLine)
                && Objects.equals(this.verticalLine, o.verticalLine)
                && Objects.equals(this.horizontalLine, o.horizontalLine);
    }

    @Override
    public int hashCode() {
        return this.hashCode;
    }

    // ZSS-1183
    SBorder cloneBorder(SBook book) {
        final SBorderLine left = this.leftLine == null ? null :
                ((AbstractBorderLineAdv) this.leftLine).cloneBorderLine(book);
        final SBorderLine top = this.topLine == null ? null :
                ((AbstractBorderLineAdv) this.topLine).cloneBorderLine(book);
        final SBorderLine right = this.rightLine == null ? null :
                ((AbstractBorderLineAdv) this.rightLine).cloneBorderLine(book);
        final SBorderLine bottom = this.bottomLine == null ? null :
                ((AbstractBorderLineAdv) this.bottomLine).cloneBorderLine(book);
        final SBorderLine diagonal = this.diagonalLine == null ? null :
                ((AbstractBorderLineAdv) this.diagonalLine).cloneBorderLine(book);
        final SBorderLine vertical = this.verticalLine == null ? null :
                ((AbstractBorderLineAdv) this.verticalLine).cloneBorderLine(book);
        final SBorderLine horizontal = this.horizontalLine == null ? null :
                ((AbstractBorderLineAdv) this.horizontalLine).cloneBorderLine(book);

        return new ImmutableBorderImpl(left, top, right, bottom, diagonal,
                vertical, horizontal);
    }
}
