package io.keikai.model.impl;

import io.keikai.model.SBook;
import io.keikai.model.SBorder;
import io.keikai.model.SBorderLine;
import io.keikai.model.SColor;

import java.util.Objects;
import java.util.Optional;

public class ImmutableBorderLineImpl extends AbstractBorderLineAdv {
    public final static SBorderLine BLANK_BORDER_LINE = new ImmutableBorderLineImpl(DefaultStyle.BORDER_TYPE,
            DefaultStyle.BORDER_COLOR);
    private static final long serialVersionUID = 2236585361952473608L;
    final private SBorder.BorderType type;
    final private SColor color;
    final private boolean showUp;
    final private boolean showDown;
    final private int hashCode;

    // ZSS-977
    public ImmutableBorderLineImpl(SBorder.BorderType type, SColor color) {
        this(type, color, DefaultStyle.BORDER_SHOW_UP, DefaultStyle.BORDER_SHOW_DOWN);
    }

    public ImmutableBorderLineImpl(SBorder.BorderType type, SColor color, boolean showUp, boolean showDown) {
        this.type = Optional.of(type).orElse(DefaultStyle.BORDER_TYPE);
        this.color = color;
        this.showUp = showUp;
        this.showDown = showDown;
        hashCode = calcHashCode(this);
    }

    static private int calcHashCode(ImmutableBorderLineImpl impl) {
        int result = impl.type.hashCode();
        result = 31 * result + impl.color.hashCode();
        result = 31 * result + (impl.showUp ? 1 : 0);
        result = 31 * result + (impl.showDown ? 1 : 0);
        return result;
    }

    public SBorder.BorderType getBorderType() {
        return type;
    }

    @Override
    public void setBorderType(SBorder.BorderType type) {
        throw new UnsupportedOperationException("doesn't support set border type");
    }

    public SColor getColor() {
        return color;
    }

    @Override
    public void setColor(SColor color) {
        throw new UnsupportedOperationException("doesn't support set color");
    }

    public boolean isShowDiagonalUpBorder() {
        return showUp;
    }

    @Override
    public void setShowDiagonalUpBorder(boolean show) {
        throw new UnsupportedOperationException("doesn't support set show diagonal up border");
    }

    public boolean isShowDiagonalDownBorder() {
        return showDown;
    }

    @Override
    public void setShowDiagonalDownBorder(boolean show) {
        throw new UnsupportedOperationException("doesn't support set show diagonal down border");
    }

    String getStyleKey() {
        return type.ordinal() +
                "." + color.getHtmlColor() +
                "." + (showUp ? "1" : "0") +
                "." + (showDown ? "1" : "0");
    }

    // ZSS-1183
    @Override
    SBorderLine cloneBorderLine(SBook book) {
        final AbstractColorAdv srcColor = (AbstractColorAdv) this.getColor();
        final SColor color = srcColor == null ? null : srcColor.cloneColor(book);
        return new ImmutableBorderLineImpl(this.type, color, this.showUp, this.showDown);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        final ImmutableBorderLineImpl that = (ImmutableBorderLineImpl) o;

        if (showUp != that.showUp) {
            return false;
        }
        if (showDown != that.showDown) {
            return false;
        }
        if (type != that.type) {
            return false;
        }
        return Objects.equals(color, that.color);
    }

    @Override
    public int hashCode() {
        return this.hashCode;
    }
}
