package io.keikai.model.impl;

import io.keikai.model.*;
import io.keikai.model.SBorder.BorderType;
import io.keikai.model.SFill.FillPattern;
import org.zkoss.lang.Objects;

/**
 * @since 5.3.0
 */
public class ImmutableCellStyleImpl extends AbstractCellStyleAdv {
    private static final long serialVersionUID = 5810442678993127830L;
    private final int hashCode;
    protected AbstractFontAdv font;
    //SFill
    protected AbstractFillAdv fill;
    //SBorder
    protected AbstractBorderAdv border;
    protected Alignment alignment = DefaultStyle.CELL_ALIGNMENT;
    protected VerticalAlignment verticalAlignment = DefaultStyle.CELL_VERTICAL_ALIGNMENT;
    protected boolean wrapText = DefaultStyle.CELL_WRAP_TEXT;
    protected String dataFormat; //ZSS-1140
    protected boolean directFormat = DefaultStyle.CELL_DIRECT_FORMAT;
    protected boolean locked = DefaultStyle.CELL_LOCKED;// default locked as excel.
    protected boolean hidden = DefaultStyle.CELL_HIDDEN;
    protected int rotation; // ZSS-918
    protected int indention; // ZSS-915

    // ZSS-1183
    ImmutableCellStyleImpl(ImmutableCellStyleImpl src, BookImpl book) {
        this.font = (AbstractFontAdv)
                (src.font == null ? null : src.font.cloneFont(book));
        this.fill = (AbstractFillAdv)
                (src.fill == null ? null : src.fill.cloneFill(book));
        this.border = (AbstractBorderAdv)
                (src.border == null ? null : src.border.cloneBorder(book));
        this.alignment = src.alignment;
        this.verticalAlignment = src.verticalAlignment;
        this.wrapText = src.wrapText;
        this.dataFormat = src.dataFormat;
        this.directFormat = src.directFormat;
        this.locked = src.locked;
        this.hidden = src.hidden;
        this.rotation = src.rotation;
        this.indention = src.indention;
        hashCode = calcHashCode(this);
    }

    ImmutableCellStyleImpl(SFont font, SFill fill, SBorder border, Alignment alignment,
                           VerticalAlignment verticalAlignment, boolean wrapText, String dataFormat,
                           boolean directFormat, boolean locked, boolean hidden, int rotation, int indention) {
        if (font != null) {
            this.font = (AbstractFontAdv) font;
        }
        if (fill != null) {
            this.fill = (AbstractFillAdv) fill;
        }
        if (border != null) {
            this.border = (AbstractBorderAdv) border;
        }
        if (alignment != null) {
            this.alignment = alignment;
        }
        if (verticalAlignment != null) {
            this.verticalAlignment = verticalAlignment;
        }
        this.wrapText = wrapText;
        if (dataFormat != null) {
            this.dataFormat = dataFormat;
        }
        this.directFormat = directFormat;
        this.locked = locked;
        this.hidden = hidden;
        this.rotation = rotation;
        this.indention = indention;
        hashCode = calcHashCode(this);
    }

    public ImmutableCellStyleImpl(AbstractFontAdv font, AbstractFillAdv fill, AbstractBorderAdv border) {
        this.font = font;
        this.fill = fill;
        this.border = border;
        hashCode = calcHashCode(this);
    }

    private static int calcHashCode(ImmutableCellStyleImpl impl) {
        int hash = impl.font == null ? 0 : impl.font.hashCode();
        hash = hash * 31 + (impl.fill == null ? 0 : impl.fill.hashCode());
        hash = hash * 31 + (impl.alignment == null ? 0 : impl.alignment.hashCode());
        hash = hash * 31 + (impl.verticalAlignment == null ? 0 : impl.verticalAlignment.hashCode());
        hash = hash * 31 + (impl.wrapText ? 1 : 0);
        hash = hash * 31 + (impl.border == null ? 0 : impl.border.hashCode());
        hash = hash * 31 + (impl.getDataFormat() == null ? 0 : impl.getDataFormat().hashCode()); //ZSS-1140
        hash = hash * 31 + (impl.directFormat ? 1 : 0);
        hash = hash * 31 + (impl.locked ? 1 : 0);
        hash = hash * 31 + (impl.hidden ? 1 : 0);
        hash = hash * 31 + impl.rotation;
        hash = hash * 31 + impl.indention;

        return hash;
    }

    public SFont getFont() {
        return font;
    }

    public void setFont(SFont font) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void copyFrom(SCellStyle src) {
        throw new UnsupportedOperationException();
    }

    @Override
    public SColor getFillColor() {
        return fill == null ? DefaultStyle.FILL_COLOR : fill.getFillColor();
    }

    @Override
    public void setFillColor(SColor fillColor) {
        throw new UnsupportedOperationException();
    }

    @Override
    public FillPattern getFillPattern() {
        return fill == null ? DefaultStyle.FILL_PATTERN : fill.getFillPattern();
    }

    @Override
    public void setFillPattern(FillPattern fillPattern) {
        throw new UnsupportedOperationException();
    }

    @Override
    public Alignment getAlignment() {
        return alignment;
    }

    @Override
    public void setAlignment(Alignment alignment) {
        throw new UnsupportedOperationException();
    }

    @Override
    public VerticalAlignment getVerticalAlignment() {
        return verticalAlignment;
    }

    @Override
    public void setVerticalAlignment(VerticalAlignment verticalAlignment) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isWrapText() {
        return wrapText;
    }

    @Override
    public void setWrapText(boolean wrapText) {
        throw new UnsupportedOperationException();
    }

    @Override
    public BorderType getBorderLeft() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderLeft();
    }

    @Override
    public void setBorderLeft(BorderType borderLeft) {
        throw new UnsupportedOperationException();
    }

    @Override
    public BorderType getBorderTop() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderTop();
    }

    @Override
    public void setBorderTop(BorderType borderTop) {
        throw new UnsupportedOperationException();
    }

    @Override
    public BorderType getBorderRight() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderRight();
    }

    @Override
    public void setBorderRight(BorderType borderRight) {
        throw new UnsupportedOperationException();
    }

    @Override
    public BorderType getBorderBottom() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderBottom();
    }

    @Override
    public void setBorderBottom(BorderType borderBottom) {
        throw new UnsupportedOperationException();
    }

    @Override
    public SColor getBorderTopColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderTopColor();
    }

    @Override
    public void setBorderTopColor(SColor borderTopColor) {
        throw new UnsupportedOperationException();
    }

    @Override
    public SColor getBorderLeftColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderLeftColor();
    }

    @Override
    public void setBorderLeftColor(SColor borderLeftColor) {
        throw new UnsupportedOperationException();
    }

    @Override
    public SColor getBorderBottomColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderBottomColor();
    }

    @Override
    public void setBorderBottomColor(SColor borderBottomColor) {
        throw new UnsupportedOperationException();
    }

    @Override
    public SColor getBorderRightColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderRightColor();
    }

    @Override
    public void setBorderRightColor(SColor borderRightColor) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getDataFormat() {
        return dataFormat == null ? DefaultStyle.CELL_DATA_FORMAT : dataFormat; //ZSS-1140
    }

    @Override
    public void setDataFormat(String dataFormat) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isDirectDataFormat() {
        return directFormat;
    }

    @Override
    public void setDirectDataFormat(String dataFormat) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isLocked() {
        return locked;
    }

    @Override
    public void setLocked(boolean locked) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isHidden() {
        return hidden;
    }

    @Override
    public void setHidden(boolean hidden) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderLeft(BorderType borderLeft, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderTop(BorderType borderTop, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderRight(BorderType borderRight, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderBottom(BorderType borderBottom, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderVertical(BorderType borderVertical, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderHorizontal(BorderType borderHorizontal, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setBorderDiagonal(BorderType borderDiagonal, SColor color) {
        throw new UnsupportedOperationException();
    }

    @Override
    String getStyleKey() {
        return (font == null ? "" : font.getStyleKey()) +
                "." + (fill == null ? "" : fill.getStyleKey()) +
                "." + (border == null ? "" : border.getStyleKey()) +
                "." + alignment.ordinal() +
                "." + verticalAlignment.ordinal() +
                "." + (wrapText ? "T" : "F") +
                "." + getDataFormat() + // ZSS-1140
                "." + (locked ? "T" : "F") +
                "." + (hidden ? "T" : "F") +
                "." + rotation;
    }

    // ZSS-780
    @Override
    public SColor getBackColor() {
        return fill == null ? DefaultStyle.BACK_COLOR : fill.getBackColor();
    }

    @Override
    public void setBackColor(SColor backColor) {
        throw new UnsupportedOperationException();
    }

    // ZSS-841
    @Override
    public String getFillPatternHtml() {
        return fill == null ? "" : ((ImmutableFillImpl) fill).getFillPatternHtml();
    }

    @Override
    public boolean equals(Object other) {
        if (other == this) {
            return true;
        }
        if (!(other instanceof ImmutableCellStyleImpl)) {
            return false;
        }
        final ImmutableCellStyleImpl o = (ImmutableCellStyleImpl) other;
        return Objects.equals(this.font, o.font)
                && Objects.equals(this.fill, o.fill)
                && Objects.equals(this.alignment, o.alignment)
                && Objects.equals(this.verticalAlignment, o.verticalAlignment)
                && Objects.equals(this.wrapText, o.wrapText)
                && Objects.equals(this.border, o.border)
                && Objects.equals(this.getDataFormat(), o.getDataFormat()) //ZSS-1140
                && Objects.equals(this.directFormat, o.directFormat)
                && Objects.equals(this.locked, o.locked)
                && Objects.equals(this.hidden, o.hidden)
                && Objects.equals(this.rotation, o.rotation)
                && Objects.equals(this.indention, o.indention);
    }

    @Override
    public int hashCode() {
        return hashCode;
    }

    // ZSS-918
    @Override
    public int getRotation() {
        return rotation;
    }

    @Override
    public void setRotation(int rotation) {
        throw new UnsupportedOperationException();
    }

    // ZSS-915
    @Override
    public int getIndention() {
        return indention;
    }

    @Override
    public void setIndention(int indent) {
        throw new UnsupportedOperationException();
    }

    // ZSS-977
    @Override
    public SBorder getBorder() {
        return border;
    }

    // ZSS-977
    @Override
    public SFill getFill() {
        return fill;
    }

    // ZSS-977
    @Override
    public BorderType getBorderVertical() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderVertical();
    }

    @Override
    public void setBorderVertical(BorderType borderVertical) {
        throw new UnsupportedOperationException();
    }

    // ZSS-977
    @Override
    public BorderType getBorderHorizontal() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderHorizontal();
    }

    @Override
    public void setBorderHorizontal(BorderType borderHorizontal) {
        throw new UnsupportedOperationException();
    }

    // ZSS-977
    @Override
    public BorderType getBorderDiagonal() {
        return border == null ? DefaultStyle.BORDER_TYPE : border.getBorderDiagonal();
    }

    @Override
    public void setBorderDiagonal(BorderType borderDiagonal) {
        throw new UnsupportedOperationException();
    }

    // ZSS-977
    @Override
    public SColor getBorderVerticalColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderVerticalColor();
    }

    @Override
    public void setBorderVerticalColor(SColor color) {
        throw new UnsupportedOperationException();
    }

    // ZSS-977
    @Override
    public SColor getBorderHorizontalColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderHorizontalColor();
    }

    @Override
    public void setBorderHorizontalColor(SColor color) {
        throw new UnsupportedOperationException();
    }

    // ZSS-977
    @Override
    public SColor getBorderDiagonalColor() {
        return border == null ? DefaultStyle.BORDER_COLOR : border.getBorderDiagonalColor();
    }

    @Override
    public void setBorderDiagonalColor(SColor color) {
        throw new UnsupportedOperationException();
    }

    // ZSS-1145
    @Override
    public boolean isShowDiagonalUpBorder() {
        if (border == null) {
            return DefaultStyle.BORDER_SHOW_UP;
        }
        return border.isShowDiagonalUpBorder();
    }

    // ZSS-1145
    @Override
    public boolean isShowDiagonalDownBorder() {
        if (border == null) {
            return DefaultStyle.BORDER_SHOW_DOWN;
        }
        return border.isShowDiagonalDownBorder();
    }

    // ZSS-1183
    @Override
    SCellStyle cloneCellStyle(SBook book) {
        return book == null ? this : book.getOrAddCellStyle(this);
    }

    // ZSS-1183
    @Override
    SCellStyle createCellStyle(SBook book) {
        return new ImmutableCellStyleImpl(this, (BookImpl) book);
    }
}
