package io.keikai.model.impl;

import io.keikai.model.*;
import io.keikai.model.SBorder.BorderType;
import io.keikai.model.util.Strings;
import io.keikai.model.util.Validations;

/**
 * @since 5.3.0
 */
public class ImmutableExtraStyleImpl extends ImmutableCellStyleImpl implements SExtraStyle {

    private static final long serialVersionUID = -7358673570667872723L;

    // ZSS-1183
    ImmutableExtraStyleImpl(ImmutableExtraStyleImpl src, BookImpl book) {
        super(src, book);
    }

    public ImmutableExtraStyleImpl(SFont font, SFill fill, SBorder border, String dataFormat) {
        super((AbstractFontAdv) font, (AbstractFillAdv) fill, (AbstractBorderAdv) border);
        // ZSS-1183
        if (fill != null) {
            Validations.argInstance(fill, ExtraFillImpl.class);
        }

        if (dataFormat != null && !Strings.isBlank(dataFormat)) {
            this.dataFormat = dataFormat;
        }
    }

    // ZSS-1145
    @Override
    public BorderType getBorderLeft() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getLeftLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public BorderType getBorderTop() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getTopLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public BorderType getBorderRight() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getRightLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public BorderType getBorderBottom() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getBottomLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public BorderType getBorderDiagonal() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getDiagonalLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public BorderType getBorderVertical() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getVerticalLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public BorderType getBorderHorizontal() {
        if (border == null) {
            return null;
        }
        final SBorderLine bline = border.getHorizontalLine();
        return bline == null ? null : bline.getBorderType();
    }

    // ZSS-1145
    @Override
    public String getDataFormat() {
        return dataFormat;
    }

    // ZSS-1183
    @Override
    SCellStyle createCellStyle(SBook book) {
        return new ImmutableExtraStyleImpl(this, (BookImpl) book);
    }

    // ZSS-1183
    @Override
    SCellStyle cloneCellStyle(SBook book) {
        return book == null ? this : book.getOrAddExtraStyle(this);
    }
}
