package io.keikai.model.impl;

import io.keikai.model.SBook;
import io.keikai.model.SColor;
import io.keikai.model.SFont;

/**
 * @since 5.3.0
 */
public class ImmutableFontImpl extends AbstractFontAdv {
    //ZSS-1191
    public final static SFont BLANK_FONT =
            new ImmutableFontImpl(DefaultStyle.FONT_COLOR.getHtmlColor(), false, DefaultStyle.FONT_ITALIC,
                    DefaultStyle.FONT_STRIKEOUT, DefaultStyle.FONT_UNDERLINE);
    final private String fontName;
    final private SColor fontColor;
    final private Boldweight fontBoldweight;
    final private boolean fontItalic;
    final private boolean fontStrikeout;
    final private Underline fontUnderline;
    final private int fontHeightPoint;
    final private TypeOffset fontTypeOffset;
    //ZSS-1140, ZSS-1145
    final private boolean overrideName;
    final private boolean overrideColor;
    final private boolean overrideBold;
    final private boolean overrideItalic;
    final private boolean overrideStrikeout;
    final private boolean overrideUnderline;
    final private boolean overrideHeightPoints;
    final private boolean overrideTypeOffset;
    final private int hashCode;

    public ImmutableFontImpl() {
        fontName = DefaultStyle.FONT_NAME;
        fontColor = DefaultStyle.FONT_COLOR;
        fontBoldweight = DefaultStyle.FONT_BOLDWEIGHT;
        fontItalic = DefaultStyle.FONT_ITALIC;
        fontStrikeout = DefaultStyle.FONT_STRIKEOUT;
        fontUnderline = DefaultStyle.FONT_UNDERLINE;
        fontHeightPoint = DefaultStyle.FONT_HEIGHT_POINTS;
        fontTypeOffset = DefaultStyle.FONT_TYPE_OFFSET;
        overrideName = false;
        overrideColor = false;
        overrideBold = false;
        overrideItalic = false;
        overrideStrikeout = false;
        overrideUnderline = false;
        overrideHeightPoints = false;
        overrideTypeOffset = false;
        hashCode = calcHashCode(this);
    }

    //ZSS-977
    public ImmutableFontImpl(String fontColor, boolean bold, boolean fontItalic, boolean fontStrikeout, Underline fontUnderline) {
        fontName = DefaultStyle.FONT_NAME;
        this.fontColor = fontColor != null ? new ColorImpl(fontColor) : DefaultStyle.FONT_COLOR;
        fontBoldweight = bold ? Boldweight.BOLD : Boldweight.NORMAL;
        this.fontStrikeout = fontStrikeout;
        this.fontItalic = fontItalic;
        this.fontUnderline = fontUnderline;
        fontHeightPoint = DefaultStyle.FONT_HEIGHT_POINTS;
        fontTypeOffset = DefaultStyle.FONT_TYPE_OFFSET;
        overrideName = false;
        overrideColor = false;
        overrideBold = false;
        overrideItalic = false;
        overrideStrikeout = false;
        overrideUnderline = false;
        overrideHeightPoints = false;
        overrideTypeOffset = false;
        hashCode = calcHashCode(this);
    }

    ImmutableFontImpl(String fontName, SColor fontColor, Boldweight boldweight, boolean fontItalic,
                      boolean fontStrikeout, Underline fontUnderline, int fontHeightPoint, TypeOffset fontTypeOffset,
                      boolean overrideName, boolean overrideColor, boolean overrideBold, boolean overrideItalic,
                      boolean overrideStrikeout, boolean overrideUnderline, boolean overrideHeightPoints,
                      boolean overrideTypeOffset) {
        this.fontName = fontName;
        this.fontColor = fontColor;
        fontBoldweight = boldweight;
        this.fontItalic = fontItalic;
        this.fontStrikeout = fontStrikeout;
        this.fontUnderline = fontUnderline;
        this.fontTypeOffset = fontTypeOffset;
        this.fontHeightPoint = fontHeightPoint;
        this.overrideName = overrideName;
        this.overrideColor = overrideColor;
        this.overrideBold = overrideBold;
        this.overrideItalic = overrideItalic;
        this.overrideStrikeout = overrideStrikeout;
        this.overrideUnderline = overrideUnderline;
        this.overrideHeightPoints = overrideHeightPoints;
        this.overrideTypeOffset = overrideTypeOffset;
        hashCode = calcHashCode(this);
    }

    private int calcHashCode(ImmutableFontImpl impl) {
       return impl.getStyleKey().hashCode();
    }

    @Override
    public String getName() {
        return fontName;
    }

    @Override
    public void setName(String fontName) {
        throw new UnsupportedOperationException("doesn't support set name");
    }

    @Override
    public SColor getColor() {
        return fontColor;
    }

    @Override
    public void setColor(SColor fontColor) {
        throw new UnsupportedOperationException("doesn't support set color");
    }

    @Override
    public Boldweight getBoldweight() {
        return fontBoldweight;
    }

    @Override
    public void setBoldweight(Boldweight fontBoldweight) {
        throw new UnsupportedOperationException("doesn't support set bold weight");
    }

    @Override
    public int getHeightPoints() {
//		return _fontHeightPoint < 0 ? 11 : _fontHeightPoint;
        return fontHeightPoint;
    }

    @Override
    public void setHeightPoints(int fontHeightPoint) {
        throw new UnsupportedOperationException("doesn't support set height points");
    }

    @Override
    public boolean isItalic() {
        return fontItalic;
    }

    @Override
    public void setItalic(boolean fontItalic) {
        throw new UnsupportedOperationException("doesn't support set italic");
    }

    @Override
    public boolean isStrikeout() {
        return fontStrikeout;
    }

    @Override
    public void setStrikeout(boolean fontStrikeout) {
        throw new UnsupportedOperationException("doesn't support set strikeout");
    }

    @Override
    public TypeOffset getTypeOffset() {
        return fontTypeOffset;
    }

    @Override
    public void setTypeOffset(TypeOffset fontTypeOffset) {
        throw new UnsupportedOperationException("doesn't support set typeoffset");
    }

    @Override
    public Underline getUnderline() {
        return fontUnderline;
    }

    @Override
    public void setUnderline(Underline fontUnderline) {
        throw new UnsupportedOperationException("doesn't support set underline");
    }

    @Override
    public void copyFrom(SFont src) {
        throw new UnsupportedOperationException("doesn't support copy from");
    }

    @Override
    String getStyleKey() {
        return fontName +
                "." + fontColor.getHtmlColor() +
                "." + fontBoldweight.ordinal() +
                "." + fontHeightPoint +
                "." + (fontItalic ? "T" : "F") +
                "." + (fontStrikeout ? "T" : "F") +
                "." + fontTypeOffset.ordinal() +
                "." + fontUnderline.ordinal();
    }

    //ZSS-977
    @Override
    public boolean equals(Object other) {
        if (other == this)
            return true;
        if (!(other instanceof ImmutableFontImpl))
            return false;
        final ImmutableFontImpl o = (ImmutableFontImpl) other;
        return this.getStyleKey().equals(o.getStyleKey());
    }

    //ZSS-977
    @Override
    public int hashCode() {
        return hashCode;
    }

    //ZSS-1145
    public boolean isOverrideName() {
        return overrideName;
    }

    //ZSS-1140
    public void setOverrideName(boolean overrideName) {
        throw new UnsupportedOperationException("doesn't support set override name");
    }

    public boolean isOverrideColor() {
        return overrideColor;
    }

    public void setOverrideColor(boolean overrideColor) {
        throw new UnsupportedOperationException("doesn't support set override color");
    }

    public boolean isOverrideBold() {
        return overrideBold;
    }

    public void setOverrideBold(boolean overrideBold) {
        throw new UnsupportedOperationException("doesn't support set override bold");
    }

    public boolean isOverrideItalic() {
        return overrideItalic;
    }

    public void setOverrideItalic(boolean overrideItalic) {
        throw new UnsupportedOperationException("doesn't support set override italic");
    }

    public boolean isOverrideStrikeout() {
        return overrideStrikeout;
    }

    public void setOverrideStrikeout(boolean overrideStrikeout) {
        throw new UnsupportedOperationException("doesn't support set override strikeout");
    }

    public boolean isOverrideUnderline() {
        return overrideUnderline;
    }

    public void setOverrideUnderline(boolean overrideUnderline) {
        throw new UnsupportedOperationException("doesn't support set override underline");
    }

    public boolean isOverrideHeightPoints() {
        return overrideHeightPoints;
    }

    public void setOverrideHeightPoints(boolean overrideHeightPoints) {
        throw new UnsupportedOperationException("doesn't support set override height points");
    }

    public boolean isOverrideTypeOffset() {
        return overrideTypeOffset;
    }

    public void setOverrideTypeOffset(boolean overrideTypeOffset) {
        throw new UnsupportedOperationException("doesn't support set override type offset");
    }

    //ZSS-1183
    @Override
    SFont cloneFont(SBook book) {
        return book == null ? this : ((AbstractBookAdv) book).getOrAddFont(this);
    }
}
