/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/01 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.model.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentSkipListMap;

/**
 * 
 * @author dennis
 * @since 3.5.0
 */
/*package*/ abstract class IndexPool<T> implements Serializable{
	private static final long serialVersionUID = 1L;
	private Integer _firstKey;
	private Integer _lastKey;
	private ConcurrentSkipListMap<Integer,T> _objs = new ConcurrentSkipListMap<>();
	
	public T get(int idx) {
		return _objs.get(idx);
	}
	
	public T put(int idx, T obj) {
		T old = _objs.put(idx, obj);
		resetCacheKey();
		return old;
	}
	
	public int firstKey() {
		if (_firstKey == null) {
			_firstKey = size() == 0 ? -1 : _objs.firstKey();
		}
		return _firstKey;
	}
	
	public int lastKey() {
		if (_lastKey == null) {
			_lastKey = size() == 0 ? -1 : _objs.lastKey();
		}
		return _lastKey;
	}
	
	public Collection<T> clear(int start,int end) {
		
		NavigableMap<Integer, T> effected = _objs.subMap(start,true,end,true);
		LinkedList<T> remove = new LinkedList<>();
		Iterator<Entry<Integer,T>> iter = effected.entrySet().iterator();
		while(iter.hasNext()){
			Entry<Integer,T> entry = iter.next();
			T obj = entry.getValue();
			remove.add(obj);
			iter.remove();
		}
		resetCacheKey();
		return remove;
	}
	
	public void insert(int start, int size) {
		if(size<=0) return;

		//get last, reversed cell
		SortedMap<Integer,T> effected = _objs.descendingMap().headMap(start,true);
		
		//shift from the end
		for(Entry<Integer,T> entry:new ArrayList<>(effected.entrySet())){
			int idx = entry.getKey();
			int newidx = idx+size;
			T obj = entry.getValue();
			
			_objs.remove(idx);
			
			resetIndex(newidx,obj);
			
			_objs.put(newidx, obj);
		}
		resetCacheKey();
	}
	
	abstract void resetIndex(int newidx, T obj);

	public Collection<T> delete(int start, int size) {
		if(size<=0) return Collections.EMPTY_LIST;
		//get last,
		SortedMap<Integer,T> effected = _objs.tailMap(start,true);
		LinkedList<T> remove = new LinkedList<>();
		//shift
		for(Entry<Integer,T> entry:new ArrayList<>(effected.entrySet())){
			int idx = entry.getKey();
			int newidx = idx-size;
			T obj = entry.getValue();
			_objs.remove(idx);
			if(newidx>=start){
				resetIndex(newidx,obj);
				_objs.put(newidx, obj);
			}else{
				remove.add(obj);
			}
		}
		resetCacheKey();
		return remove;
	}
	
	public Collection<T> trim(int start) {
		//get last,
		SortedMap<Integer,T> effected = _objs.tailMap(start,true);
		LinkedList<T> remove = new LinkedList<>();
		//shift
		for(Entry<Integer,T> entry:new ArrayList<>(effected.entrySet())){
			int idx = entry.getKey();
			_objs.remove(idx);
			remove.add(entry.getValue());
		}
		resetCacheKey();
		return remove;
	}
	
	public Set<Integer> keySet(){
		return _objs.keySet();
	}

	public Collection<T> values() {
		return _objs.values();
	}
	public Collection<T> descendingValues() {
		return _objs.descendingMap().values();
	}
	
	public Collection<T> subValues(int start,int end){
		return _objs.subMap(start, true, end, true).values();
	}
	public Collection<T> descendingSubValues(int start,int end){
		return _objs.subMap(start, true, end, true).descendingMap().values();
	}
	public int size() {
		return _objs.size();
	}
	public void clear() {
		_objs.clear();
		resetCacheKey();
	}

	private void resetCacheKey() {
		_firstKey = null;
		_lastKey = null;
	}
}
