/* OutlineLevelTree.java

	Purpose:
		
	Description:
		
	History:
		11:29 AM 2023/11/15, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
package io.keikai.model.impl;

import java.io.Serializable;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.List;

import io.keikai.util.Ref;

/**
 * An outline level tree structure.
 * @author jumperchen
 * @since 6.0.0
 */
public class OutlineLevelTree implements Serializable {
	private OutlineLevelNode[] indexToNode = new OutlineLevelNode[8];
	public OutlineLevelNode root = new OutlineLevelNode(0);

	private static boolean overlaps(OutlineLevelNode node, Ref ref) {
		return node.end >= ref.getTop() && node.start <= ref.getBottom();
	}
	public List<OutlineLevelNode> search(Ref ref) {
		Deque<OutlineLevelNode> stack = new ArrayDeque<>();
		stack.push(root);
		List<OutlineLevelNode> result = new ArrayList<>();
		while (!stack.isEmpty()) {
			OutlineLevelNode node = stack.pop();
			if (overlaps(node, ref)) {
				if (node.level < ref.getRight()) {// exclusive
					if (node.children != null) {
						stack.addAll(node.children);
					}
					if (node.level >= ref.getLeft()) {
						result.add(node);
					}
				}
			}
		}
		return result;
	}

	public OutlineLevelNode getOrCreateNode(int level) {
		return getOrCreateNode(level, false);
	}

	private OutlineLevelNode getOrCreateNode(int level, boolean skipClear) {
		if (level == 0) {
			if (!skipClear) {
				indexToNode = new OutlineLevelNode[8];
			}
			return root;
		}
		OutlineLevelNode node = indexToNode[level];
		if (node != null) {
			// clear the rest.
			if (!skipClear) {
				for (int i = level + 1; i < 8; i++) {
					indexToNode[i] = null;
				}
			}
			return node;
		}
		node = new OutlineLevelNode(level);
		OutlineLevelNode parent = getOrCreateNode(level - 1, true);
		parent.appendChild(node);
		indexToNode[level] = node;
		return node;
	}

	public static class OutlineLevelNode implements Serializable {
		private int start = -1;
		private int end = -1;
		private int level = 0;
		private OutlineLevelNode parent;
		private List<OutlineLevelNode> children;
		private int maxLevel = 0;

		public OutlineLevelNode(int level) {
			this.level = level;
		}

		public void appendChild(OutlineLevelNode node) {
			if (children == null) {
				children = new ArrayList<>();
			}
			children.add(node);
			node.parent = this;
		}
		public void setIndex(int index, int level) {
			if (start < 0 || start > index) {
				start = index;
			}
			if (end < index) {
				end = index;
			}
			if (maxLevel < level) {
				maxLevel = level;
			}
			if (parent != null) {
				parent.setIndex(index, level);
			}
		}

		public int getStart() {
			return start;
		}

		public int getEnd() {
			return end;
		}

		public int getLevel() {
			return level;
		}

		public int getMaxLevel() {
			return maxLevel;
		}
	}
}
