package io.keikai.model.impl.chart;

import io.keikai.model.SChartAxis;
import io.keikai.model.impl.AbstractChartAxisAdv;
import io.keikai.model.impl.ChartAxisImpl;

/**
 * @since 5.7.0
 */
public class DateAxisImpl extends ChartAxisImpl {
    private static final long serialVersionUID = 7392848813857669086L;

    public enum TimeUnit {
        DAYS,
        MONTHS,
        YEARS
    }
    private final Boolean auto;
    private final Integer lblOffset;
    private final TimeUnit baseTimeUnit;
    private final Double majorUnit;
    private final TimeUnit majorTimeUnit;
    private final Double minorUnit;
    private final TimeUnit minorTimeUnit;

    public DateAxisImpl(long id, AbstractChartAxisAdv.Position position, Double min, Double max,
                        String format, AbstractChartAxisAdv.Orientation orientation,
                        AbstractChartAxisAdv.TickLabelPosition tickLabelPosition, long crossAxisId,
                        Crosses axisCrosses, boolean majorGridline, boolean visible,
                        Boolean auto, Integer lblOffset, TimeUnit baseTimeUnit, Double majorUnit,
                        TimeUnit majorTimeUnit, Double minorUnit, TimeUnit minorTimeUnit) {
        super(id, position, min, max, format, orientation, tickLabelPosition, crossAxisId, axisCrosses, majorGridline,
                visible);
        this.auto = auto;
        this.lblOffset = lblOffset;
        this.baseTimeUnit = baseTimeUnit;
        this.majorUnit = majorUnit;
        this.majorTimeUnit = majorTimeUnit;
        this.minorUnit = minorUnit;
        this.minorTimeUnit = minorTimeUnit;
    }

    @Override
    public SChartAxis.SChartAxisType getType() {
        return SChartAxis.SChartAxisType.DATE;
    }

    public Boolean getAuto() {
        return auto;
    }

    public Integer getLblOffset() {
        return lblOffset;
    }

    public TimeUnit getBaseTimeUnit() {
        return baseTimeUnit;
    }

    public Double getMajorUnit() {
        return majorUnit;
    }

    public TimeUnit getMajorTimeUnit() {
        return majorTimeUnit;
    }

    public Double getMinorUnit() {
        return minorUnit;
    }

    public TimeUnit getMinorTimeUnit() {
        return minorTimeUnit;
    }

    public DateAxisImpl cloneDateAxisImpl(long id) {
        return new DateAxisImpl(id, getPosition(), getMin(), getMax(), getFormat(), getOrientation(),
                getTickLabelPosition(), getCrossAxisId(),getAxisCrosses(), hasMajorGridline(),
                isVisible(), auto, lblOffset, baseTimeUnit, majorUnit, majorTimeUnit, minorUnit, minorTimeUnit);
    }
}
