/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/01 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.model.impl.chart;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import io.keikai.model.ErrorValue;
import io.keikai.model.SChart;
import io.keikai.model.SChartAxis;
import io.keikai.model.chart.SGeneralChartData;
import io.keikai.model.chart.SSeries;
import io.keikai.model.impl.AbstractBookSeriesAdv;
import io.keikai.model.impl.AbstractChartAdv;
import io.keikai.model.impl.EvaluationUtil;
import io.keikai.model.impl.ObjectRefImpl;
import io.keikai.model.sys.EngineFactory;
import io.keikai.model.sys.dependency.Ref;
import io.keikai.model.sys.formula.EvaluationResult;
import io.keikai.model.sys.formula.FormulaEngine;
import io.keikai.model.sys.formula.FormulaEvaluationContext;
import io.keikai.model.sys.formula.FormulaExpression;
import io.keikai.model.sys.formula.FormulaParseContext;
import io.keikai.model.sys.formula.EvaluationResult.ResultType;
/**
 * 
 * @author dennis
 * @since 3.5.0
 */
public abstract class GeneralChartDataImpl extends AbstractGeneralChartDataAdv implements SGeneralChartData {

	private static final long serialVersionUID = 1L;

	private FormulaExpression _catFormulaExpr; //horizontal axis labels
	
	final private List<SeriesImpl> _serieses = new LinkedList<>();
	private AbstractChartAdv _chart;
	final private String _id;
	
	private Object _evalResult;
	
	private boolean _evaluated = false;
	private boolean _visibleEvaluated = false;
	
	private int _seriesCount = 0;

	private boolean[] _hiddenCategoriesInfo;

	private SChartAxis categoryAxis;
    private SChartAxis dateAxis;
	private List<SChartAxis> valueAxes;
	
	public GeneralChartDataImpl(AbstractChartAdv chart,String id){
		this._chart = chart;
		this._id = id;
	}
	
	public SChart getChart(){
		return _chart;
	}
	
	/*package*/ void evalFormula(){
		//ZSS-739
		//20140730, henrichen: when share the same book, many users might 
		//populate GeneralChartDataImpl simultaneously; must synchronize it.
		if(_evaluated) return;
		synchronized (this) {
			if(!_evaluated){
				if(_catFormulaExpr!=null){
					FormulaEngine fe = EngineFactory.getInstance().createFormulaEngine();
					EvaluationResult result = fe.evaluate(_catFormulaExpr,new FormulaEvaluationContext(_chart.getSheet(),getRef()));
					Object val = result.getValue();
					if(result.getType() == ResultType.SUCCESS){
						_evalResult = val;
					}else if(result.getType() == ResultType.ERROR){
						_evalResult = (val instanceof ErrorValue)?val: ErrorValue.valueOf(ErrorValue.INVALID_NAME); //ZSS-672
					}
				}
				_evaluated = true;
			}
		}
	}
	
	public int getNumOfSeries() {
		return _serieses.size();
	}
	public SSeries getSeries(int i) {
		return _serieses.get(i);
	}
	public int getNumOfCategory() {
		evalFormula();
		return EvaluationUtil.sizeOf(_evalResult);
	}
	public Object getCategory(int i) {
		evalFormula();
		if(i>=EvaluationUtil.sizeOf(_evalResult)){
			//ZSS-862
			return String.valueOf(i+1);
		}
		//ZSS-882
		Object cat = EvaluationUtil.valueOf(_evalResult,i);
		boolean showCat = false;
		for (SSeries ser : _serieses) {
			final Object val = ser.getValue(i);
			if ((val instanceof Integer && ((Integer) val) != 0) && !"0".equals(val)) {
				showCat = true;
				break;
			}
		}
		return cat == null || "".equals(cat) ? (showCat ? String.valueOf(i+1) : "") : cat;
	}
	public SSeries addSeries() {
		SeriesImpl series = new SeriesImpl(_chart,_chart.getId() + "-series" + (_seriesCount++));
		_serieses.add(series);
		return series;
	}
	
	protected void checkOwnership(SSeries series){
		if(!_serieses.contains(series)){
			throw new IllegalStateException("doesn't has ownership "+ series);
		}
	}
	
	public void removeSeries(SSeries series) {
		checkOwnership(series);
		((SeriesImpl)series).destroy();
		_serieses.remove(series);
	}
	public void setCategoriesFormula(String expr) {
		checkOrphan();
		_evaluated = false;
		
		clearFormulaDependency();
		
		if(expr!=null && expr != ""){ //ZSS-1129
			FormulaEngine fe = EngineFactory.getInstance().createFormulaEngine();
			_catFormulaExpr = fe.parse(expr, new FormulaParseContext(_chart.getSheet(),getRef()));
		}else{
			_catFormulaExpr = null;
		}
	}

	@Override
	public String getCategoriesFormula() {
		return _catFormulaExpr==null?null:_catFormulaExpr.getFormulaString();
	}

	@Override
	public void clearFormulaResultCache() {
		_evalResult = null;
		_evaluated = false;
		_visibleEvaluated = false;
		for(SeriesImpl series:_serieses){
			series.clearFormulaResultCache();
		}
	}
	
	@Override
	public boolean isFormulaParsingError() {
		return _catFormulaExpr==null?false:_catFormulaExpr.hasError();
	}
	
	private void clearFormulaDependency(){
		if(_catFormulaExpr!=null){
			((AbstractBookSeriesAdv) _chart.getSheet().getBook().getBookSeries())
					.getDependencyTable().clearDependents(getRef());
		}
	}
	
	private Ref getRef(){
		return new ObjectRefImpl(_chart,new String[]{_chart.getId(),_id});
	}

	@Override
	public void destroy() {
		checkOrphan();
		clearFormulaDependency();
		clearFormulaResultCache();
		for(SeriesImpl series:_serieses){
			series.destroy();
		}
		_chart = null;
	}

	@Override
	public void checkOrphan() {
		if(_chart==null){
			throw new IllegalStateException("doesn't connect to parent");
		}
	}

	//ZSS-688
	//@since 3.6.0
	public void copyFrom(GeneralChartDataImpl src) {
		checkOrphan();
		final String expr = src.getCategoriesFormula(); //ZSS-1183
		if (expr != null) {
			this.setCategoriesFormula(expr);
		}
		
		for (SeriesImpl s : src._serieses) {
			_serieses.add(s.cloneSeriesImpl(_chart));
		}
		_seriesCount = src._seriesCount;

		if (src.categoryAxis != null) {
			final List<SChartAxis> srcCatAxes = src._chart.getCategoryAxises();
			for (int i = 0; i < srcCatAxes.size(); ++i) {
				if (srcCatAxes.get(i).getId() == src.categoryAxis.getId()) {
					final List<SChartAxis> catAxes = _chart.getCategoryAxises();
					categoryAxis = catAxes.get(i);
				}
			}
		}

		if (src.dateAxis != null) {
			final List<SChartAxis> srcDateAxes = src._chart.getDateAxises();
			for (int i = 0; i < srcDateAxes.size(); ++i) {
				if (srcDateAxes.get(i).getId() == src.dateAxis.getId()) {
					final List<SChartAxis> dateAxes = _chart.getDateAxises();
					dateAxis = dateAxes.get(i);
				}
			}
		}

		final List<SChartAxis> valueAxes = new ArrayList<>();
		final List<SChartAxis> srcValueAxes = src.valueAxes;
		if (srcValueAxes != null) {
			final List<SChartAxis> srcChartValAxes = src._chart.getValueAxises();
			for (SChartAxis srcValAx : srcValueAxes) {
				final long id = srcValAx.getId();
				for (int j = 0; j < srcChartValAxes.size(); ++j) {
					if (srcChartValAxes.get(j).getId() == id) {
						valueAxes.add(_chart.getValueAxises().get(j));
					}
				}
			}
			this.valueAxes = valueAxes;
		}
		// do not copy _evalResult, _evaluated.
		//private Object _evalResult;
		//private boolean _evaluated = false;
	}
	
	//ZSS-747
	/**
	 * 
	 * @return
	 * @since 3.6.0
	 */
	public FormulaExpression getCategoriesFormulaExpression() {
		return _catFormulaExpr;  
	}
	//ZSS-747
	/**
	 * 
	 * @param fexpr
	 * @since 3.6.0
	 */
	public void setCategoriesFormula(FormulaExpression fexpr) {
		checkOrphan();
		_evaluated = false;
		clearFormulaDependency();
		
		_catFormulaExpr = fexpr;
		
		//update dependency table
		if(fexpr != null) {
			FormulaEngine fe = EngineFactory.getInstance().createFormulaEngine();
			fe.updateDependencyTable(fexpr, new FormulaParseContext(_chart.getSheet(),getRef()));
		}
	}

	@Override
	public boolean isCategoryHidden(int index) {
		evalVisibleInfo();
		// select multiple empty rows may cause exceeding array index exception
		return _hiddenCategoriesInfo == null ? false : _hiddenCategoriesInfo.length < index + 1 ? false : _hiddenCategoriesInfo[index];
	}

	public SChartAxis getCategoryAxis() {
		return categoryAxis;
	}

	public void setCategoryAxis(SChartAxis category) {
		categoryAxis = category;
	}

	public SChartAxis getDateAxis() {
		return dateAxis;
	}

	public void setDateAxis(SChartAxis dateAxis) {
		this.dateAxis = dateAxis;
	}

	public List<SChartAxis> getValueAxes() {
		return valueAxes;
	}

	public void setValueAxes(List<SChartAxis> values) {
		valueAxes = values;
	}

	private void evalVisibleInfo() {
		if(_visibleEvaluated)
			return;
		
		synchronized (this) {
			if(!_visibleEvaluated) {
				Map<Integer, Boolean> cachedRowValues = new HashMap<Integer, Boolean>(16);
				Map<Integer, Boolean> cachedColumnValues = new HashMap<Integer, Boolean>(16);
				_hiddenCategoriesInfo = new boolean[getNumOfCategory()];
				
				ChartUtil.evalVisibleInfo(_chart, _catFormulaExpr, _hiddenCategoriesInfo, cachedRowValues, cachedColumnValues);
				_visibleEvaluated = true;
			}
		}
		
	}
}
