/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/01 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.model.sys;

import java.util.Optional;

import io.keikai.model.impl.sys.CalendarUtilImpl;
import io.keikai.model.impl.sys.DependencyTableImpl;
import io.keikai.model.impl.sys.FormatEngineImpl;
import io.keikai.model.impl.sys.InputEngineImpl;
import io.keikai.model.impl.sys.formula.FormulaEngineImpl;
import io.keikai.model.sys.dependency.DependencyTable;
import io.keikai.model.sys.format.FormatEngine;
import io.keikai.model.sys.formula.FormulaEngine;
import io.keikai.model.sys.input.InputEngine;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.zkoss.lang.Library;

/**
 * 
 * @author dennis
 * @since 3.5.0
 */
public class EngineFactory {

	private static final Logger _logger = LoggerFactory.getLogger(EngineFactory.class);

	static private EngineFactory _instance;
	
	static private CalendarUtil _calendarUtil = new CalendarUtilImpl();

	//KEIKAI-453
	// maximum iterations when use a iterative calculation;
	// 0 means do NOT use iterative calculation.
	private int _maxIters;

	// KEIKAI-453
	// iterations stop if change between iterations less than this value [maxChange]
	private double _maxChange;

	private EngineFactory() {
	}

	public static EngineFactory getInstance() {
		if (_instance == null) {
			synchronized (EngineFactory.class) {
				if (_instance == null) {
					// TODO from config
					_instance = new EngineFactory();
				}
			}
		}
		return _instance;
	}

	static InputEngine _inputEngine;
	static Class<?> inputEngineClazz;
	static {
		String clz = Optional.ofNullable(Library.getProperty("io.keikai.model.InputEngine.class")).orElse(Library.getProperty("org.zkoss.zss.model.InputEngine.class"));
		if(clz!=null){
			try {
				inputEngineClazz = Class.forName(clz);
			} catch(Exception e) {
				_logger.error(e.getMessage(), e);
			}			
		}
		
	}
	public InputEngine createInputEngine() {
		if (_inputEngine == null) {
			try {
				if(inputEngineClazz != null) {
					return (_inputEngine = (InputEngine)inputEngineClazz.newInstance());
				}
			} catch(Exception e) {
				_logger.error(e.getMessage(), e);
				inputEngineClazz = null;
			}
			_inputEngine = new InputEngineImpl();
		}
		return _inputEngine;
	}

	static Class<?> formulaEngineClazz;
	static {
		String clz = Optional.ofNullable(Library.getProperty("io.keikai.model.FormulaEngine.class")).orElse(Library.getProperty("org.zkoss.zss.model.FormulaEngine.class"));
		if(clz!=null){
			try {
				formulaEngineClazz = Class.forName(clz);
			} catch(Exception e) {
				_logger.error(e.getMessage(), e);
			}			
		}
		
	}
	
	public FormulaEngine createFormulaEngine() {
		try {
			if(formulaEngineClazz != null) {
				return (FormulaEngine)formulaEngineClazz.newInstance();
			}
		} catch(Exception e) {
			_logger.error(e.getMessage(), e);
			formulaEngineClazz = null;
		}
		return new FormulaEngineImpl();
	}

	//ZSS-815
	static Class<?> dependencyTableClazz;
	static {
		String clz = Optional.ofNullable(Library.getProperty("io.keikai.model.DependencyTable.class")).orElse(Library.getProperty("org.zkoss.zss.model.DependencyTable.class"));
		if(clz!=null){
			try {
				dependencyTableClazz = Class.forName(clz);
			} catch(Exception e) {
				_logger.error(e.getMessage(), e);
			}			
		}
		
	}
	//ZSS-815
	public DependencyTable createDependencyTable() {
		try {
			if(dependencyTableClazz != null) {
				return (DependencyTable)dependencyTableClazz.newInstance();
			}
		} catch(Exception e) {
			_logger.error(e.getMessage(), e);
			dependencyTableClazz = null;
		}
		return new DependencyTableImpl();
	}
	
	static FormatEngine _formatEngine;
	static Class<?> formatEngineClazz;
	static {
		String clz = Optional.ofNullable(Library.getProperty("io.keikai.model.FormatEngine.class")).orElse(Library.getProperty("org.zkoss.zss.model.FormatEngine.class"));
		if(clz!=null){
			try {
				formatEngineClazz = Class.forName(clz);
			} catch(Exception e) {
				_logger.error(e.getMessage(), e);
			}			
		}
		
	}
	public FormatEngine createFormatEngine() {
		if (_formatEngine == null) {
			try {
				if(formatEngineClazz != null) {
					return (_formatEngine = (FormatEngine)formatEngineClazz.newInstance());
				}
			} catch(Exception e) {
				_logger.error(e.getMessage(), e);
				formatEngineClazz = null;
			}
			_formatEngine = new FormatEngineImpl();
		}
		return _formatEngine;
	}
	
	public CalendarUtil getCalendarUtil(){
		return _calendarUtil;
	}

}
