/* FormulaType.java

	Purpose:
		
	Description:
		
	History:
		3:18 PM 2023/9/7, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
package io.keikai.model.sys.formula;

/**
 * Enumeration of various formula types.
 *
 * See Sections 3 and 4.8 of https://www.openoffice.org/sc/excelfileformat.pdf
 * @author jumperchen
 * @since 6.0.0
 */
public enum FormulaType {

	/** Regular cell formula */
	CELL,

	/**
	 * A Shared Formula ("{=SUM(A1:E1*{1,2,3,4,5}}")
	 * Similar to an array formula, but stored in a SHAREDFMLA instead of ARRAY record as a file size optimization.
	 * See Section 4.8 of https://www.openoffice.org/sc/excelfileformat.pdf
	 */
	SHARED,

	/**
	 * An Array formula ("{=SUM(A1:E1*{1,2,3,4,5}}")
	 * https://support.office.com/en-us/article/Guidelines-and-examples-of-array-formulas-7D94A64E-3FF3-4686-9372-ECFD5CAA57C7
	 */
	ARRAY,

	/** Conditional formatting */
	CONDFORMAT,

	/** Named range */
	NAMEDRANGE,

	/**
	 * This constant is currently very specific.  The exact differences from general data
	 * validation formulas or conditional format formulas is not known yet
	 */
	DATAVALIDATION_LIST;
}
