/* FunctionResolverFactory.java

	Purpose:
		
	Description:
		
	History:
		Dec 26, 2013 Created by Pao Wang

Copyright (C) 2013 Potix Corporation. All Rights Reserved.
 */
package io.keikai.model.sys.formula;


import java.util.Optional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zkoss.lang.Library;
import io.keikai.model.impl.sys.formula.FunctionResolverImpl;

/**
 * A factory of formula function resolver.
 * @author Pao
 */
public class FunctionResolverFactory {
	private static final Logger logger = LoggerFactory.getLogger(FunctionResolverFactory.class);

	private static Class<?> functionResolverClazz;
	static {
		String clz = Optional.ofNullable(Library.getProperty("io.keikai.model.FunctionResolver.class")).orElse(Library.getProperty("org.zkoss.zss.model.FunctionResolver.class"));
		if(clz!=null){
			try {
				functionResolverClazz = Class.forName(clz);
			} catch(Exception e) {
				logger.error(e.getMessage(), e);
			}			
		}
	}

	public static FunctionResolver createFunctionResolver() {
		try {
			if(functionResolverClazz != null) {
				return (FunctionResolver)functionResolverClazz.newInstance();
			}
		} catch(Exception e) {
			logger.error(e.getMessage(), e);
			functionResolverClazz = null;
		}
		return new FunctionResolverImpl();
	}
}
