/* DataRegionHelper.java
 * 
{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.range.impl;

import io.keikai.model.CellRegion;
import io.keikai.model.InvalidModelOpException;
import io.keikai.model.SCell;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.model.STable;
import io.keikai.model.impl.AbstractRowAdv;
import io.keikai.model.impl.AbstractSheetAdv;
import io.keikai.range.SRange;

import java.util.Iterator;
import java.util.List;

/**
 * To help search the data region
 *
 * @author dennis
 * @since 3.5.0
 */
//these code if from XRangeImpl and migrate to new model
public class DataRegionHelper extends RangeHelperBase { //ZSS-1261
    private static final long serialVersionUID = 7502892116280609724L;

    public DataRegionHelper(SRange range) {
        super(range);
    }

    public static boolean isOneCell(SSheet sheet, CellRegion rng) {
        if (rng.isSingle()) {
            return true;
        }
        final int l = rng.getColumn();
        final int t = rng.getRow();
        final int r = rng.getLastColumn();
        final int b = rng.getLastRow();

        int s = sheet.getNumOfMergedRegion();
        for (int i = 0; i < s; i++) {
            final CellRegion ref = sheet.getMergedRegion(i);
            final int top = ref.getRow();
            final int bottom = ref.getLastRow();
            final int left = ref.getColumn();
            final int right = ref.getLastColumn();

            if (l == left && t == top && r == right && b == bottom) {
                return true;
            }
        }
        return false;
    }

    // ZSS-1261: give an API for user to select all range from selecting a cell
    public CellRegion findCustomSortRegion() {
        final CellRegion currentArea = new CellRegion(getRow(), getColumn(), getLastRow(), getLastColumn());
        //If it's a single cell range, it has to be extend to a continuous range by looking up the near 8 cells of the single cell.
        if (isOneCell(sheet, currentArea)) {
            final FilterRegionHelper frHelper = new FilterRegionHelper();
            final CellRegion cra = frHelper.findOutlineRegion(sheet, getRow(), getColumn(), getRow(), getColumn());
            if (cra == null) { // cannot extend
                throw new InvalidModelOpException("This cannot be applied to the selected range");
            }
            return cra;
        } else {
            return currentArea;
        }
    }

    // ZSS-246: give an API for user checking the auto-filtering range before applying it.
    public CellRegion findAutoFilterDataRegion() {
        //The logic to decide the actual affected range to implement autofilter:
        //If it's a multiple cell range, it's the range intersect with largest range of the sheet.
        //If it's a single cell range, it has to be extend to a continuous range by looking up the near 8 cells of the single cell.
        final CellRegion currentArea = new CellRegion(getRow(), getColumn(), getLastRow(), getLastColumn());
        final FilterRegionHelper frHelper = new FilterRegionHelper();
        //ZSS-199
        if (isWholeRow()) {
            //extend to a continuous range from the top row
            if (isPartialOverlapTables()) {
                throw new InvalidModelOpException("Range overlaps the table");
            }
            return frHelper.getRowCurrentRegion(sheet, getRow(), getLastRow());

        } else if (isOneCell(sheet, currentArea)) {
            //only one cell selected(include merged one), try to look the max range surround by blank cells
            return frHelper.findOutlineRegion(sheet, getRow(), getColumn(), getRow(), getColumn());
        } else {
            if (isPartialOverlapTables()) {
                throw new InvalidModelOpException(
                        "Range " + currentArea.getReferenceString() + " overlaps the table");
            }
            final CellRegion region = frHelper.findOutlineRegion(sheet,
                    getRow(), getColumn(), getLastRow(), getLastColumn());
            if (region == null) {
                return null;
            } else {
                final int row = Math.max(getRow(), region.getRow());
                final int col = Math.max(getColumn(), region.getColumn());
                final int lastRow = region.getLastRow();
                final int lastCol = Math.min(getLastColumn(), region.getLastColumn());
                if (row > lastRow || col > lastCol) {
                    return null;
                }
                return new CellRegion(row, col, lastRow, lastCol);
            }
        }
    }

    private boolean isPartialOverlapTables() {
        final List<STable> tables = sheet.getTables();
        final int col = getColumn();
        final int lastCol = getLastColumn();
        final int row = getRow();
        final int lastRow = getLastRow();
        for (final STable table : tables) {
            final CellRegion rgn = table.getAllRegion().getRegion();
            if (rgn.getColumn() <= lastCol && rgn.getLastColumn() >= col &&
                    rgn.getRow() <= lastRow && rgn.getLastRow() >= row) {
                return true; // found
            }
        }
        return false;
    }

    //ZSS-988: consider auto filter surrounded by Table
    public static class FilterRegionHelper {
        static private int LTRB_L = 0;
        static private int LTRB_T = 1;
        static private int LTRB_R = 2;
        static private int LTRB_B = 3;
        static private int LR_L = 0;
        static private int LR_R = 1;
        //outline region surrounded by Tables
        int llm, tlm, rlm, blm;

        //[0]: left, [1]: top, [2]: right, [3]: bottom; null mean blank cell
        private int[] getRowMinMax(SSheet sheet, SRow row) {
            final int rowIndex = row.getIndex();
            final int sc = sheet.getStartCellIndex(rowIndex);
            final int es = sheet.getEndCellIndex(rowIndex);
            if (sc == -1) {
                return null;
            }
            return getRowMinMax(sheet, row, sc, es);
        }

        private int[] getRowMinMax(SSheet sheet, SRow row, int startCol, int endCol) {
            final int rowIdx = row.getIndex();
            int minc = startCol;
            int maxc = endCol;
            int minr = rowIdx;
            int maxr = rowIdx;
            boolean found = false;
            final Iterator<SCell> iter = row.getCellIterator(startCol, endCol);
            while (iter.hasNext()) {
                final int[] ltrb = getCellMinMax(sheet, rowIdx, iter);
                if (ltrb != null) {
                    found = true;
                    minc = Math.min(minc, ltrb[LTRB_L]);
                    minr = Math.min(minr, ltrb[LTRB_T]);
                    maxc = Math.max(maxc, ltrb[LTRB_R]);
                    minr = Math.max(minr, ltrb[LTRB_B]);
                }
            }
            return found ? new int[]{minc, minr, maxc, maxr} : null;
        }

//        /**
//         * Search adjacent cells and return a range with non-blank cells based on a given cell.
//         * It searches row by row to find minimal and maximal non-blank columns.
//         *
//         * @param sheet target searching sheet
//         * @param row   starting cell's row index
//         * @param col   starting cell's column index
//         * @return
//         */
//        CellRegion findCurrentRegion(SSheet sheet, int row, int col) {
//            // TODO: search should be able to across the table
//            return findOutlineRegion(sheet, row, col);
//        }

        //[0]: left, [1]: top, [2]: right, [3]: bottom; null mean blank cell
        int[] getCellMinMax(SSheet sheet, int row, Iterator<SCell> iter) {
            SCell ltCell = iter.next();
            final int col = ltCell.getColumnIndex();
            //ZSS-988: as long as out of region, deemed as blank
            if (tlm > row || blm < row || llm > col || rlm < col) {
                return null;
            }

            final CellRegion rng = sheet.getMergedRegion(row, col);
            int l;
            int t;
            int r;
            int b;
            if (rng != null) {
                l = rng.getColumn();
                t = rng.getRow();
                r = rng.getLastColumn();
                b = rng.getLastRow();
                if (t != row || l != col) {
                    ltCell = sheet.getCell(t, l);
                }
            } else {
                l = col;
                t = row;
                r = col;
                b = row;
            }
            return isBlank(ltCell) ? null : new int[]{l, t, r, b};
        }

        /**
         * Search adjacent cells and return a range with non-blank cells based on a given cell.
         * The region should not across the table if the cell is outside the table, or return
         * the data region in the table which the cell located
         *
         * @param sheet   target searching sheet
         * @param row     starting cell's row index
         * @param col     starting cell's column index
         * @param lastRow ending cell's row index
         * @param lastCol ending cell's column index
         * @return
         */
        CellRegion findOutlineRegion(SSheet sheet, int row, int col, int lastRow, int lastCol) {
            return findOutlineRegion0(sheet, row, col, lastRow, lastCol);
        }

        /// get the outline region surrounded by Tables
        private void calcBoundaries(SSheet sheet, int row, int col, int lastRow, int lastCol) {
            // ZSS-988: get the outline region surrounded by Tables
            llm = 0;
            rlm = Integer.MAX_VALUE;
            tlm = 0;
            blm = Integer.MAX_VALUE;

            // locate the row limits first
            for (STable tb : sheet.getTables()) {
                final CellRegion rgn = tb.getAllRegion().getRegion();

                final int l = rgn.getColumn();
                final int r = rgn.getLastColumn();
                final int t = rgn.getRow();
                final int b = rgn.getLastRow();

                //Don't consider row limits unless the cell is between the table's first column and last column
                if (l <= col && lastCol <= r) {
                    if (row < t && blm >= t) {
                        blm = t - 1;
                    }
                    if (lastRow > b && tlm <= b) {
                        tlm = b + 1;
                    }
                }
            }

            // then locate the column limits
            for (STable tb : sheet.getTables()) {
                final CellRegion rgn = tb.getAllRegion().getRegion();

                final int l = rgn.getColumn();
                final int r = rgn.getLastColumn();
                final int t = rgn.getRow();
                final int b = rgn.getLastRow();

                //Don't consider column limits unless the table overlapped with found row limits
                if (t <= blm && b >= tlm) {
                    if (col < l && rlm >= l) {
                        rlm = l - 1;
                    }
                    if (lastCol > r && llm <= r) {
                        llm = r + 1;
                    }
                }
            }

            // then locates the data limits
            final int sr = sheet.getStartRowIndex();
            final int er = sheet.getEndRowIndex();
            if (sr != -1 && sr > tlm) {
                tlm = sr;
            }
            if (er != -1 && er < blm) {
                blm = er;
            }
        }

        private void calcBoundaries(SSheet sheet, int row, int lastRow) {
            llm = 0;
            tlm = row;
            rlm = Integer.MAX_VALUE;
            blm = lastRow;
        }

        private CellRegion findOutlineRegion0(SSheet sheet, int row, int col, int lastRow, int lastCol) {
            final AbstractSheetAdv absSheetAdv = (AbstractSheetAdv) sheet;
            // select table data if selection contained by a table
            final STable table = absSheetAdv.getTableByRowCol(row, col);
            if (table != null) {
                return table.getDataRegion().getRegion();
            }
            calcBoundaries(sheet, row, col, lastRow, lastCol);
            // bounding original 9 cells centered at this cell Range
            final int t = row > tlm ? row - 1 : row;
            final int b = lastRow < blm ? lastRow + 1 : lastRow;
            final int l = col > llm ? col - 1 : col;
            final int r = lastCol < rlm ? lastCol + 1 : lastCol;
            int tt = Integer.MAX_VALUE;
            int bb = -1;
            int ll = Integer.MAX_VALUE;
            int rr = -1;
            final Iterator<SRow> iter = sheet.getRowIterator(t, b);
            while (iter.hasNext()) {
                final SRow srow = iter.next();
                final Iterator<SCell> cellIter = srow.getCellIterator(l, r);
                while (cellIter.hasNext()) {
                    final int[] ltrb = getCellMinMax(sheet, srow.getIndex(), cellIter);
                    if (ltrb != null) {
                        ll = Math.min(ll, ltrb[LTRB_L]);
                        tt = Math.min(tt, ltrb[LTRB_T]);
                        rr = Math.max(rr, ltrb[LTRB_R]);
                        bb = Math.max(bb, ltrb[LTRB_B]);
                    }
                }
            }
            if (tt > bb || ll > rr) {
                return null;
            }
            return findCurrentRegion0(absSheetAdv, tt, ll, bb, rr);
        }

        private boolean isRowBlank(AbstractRowAdv row, int start, int end) {
            final int index = row.getIndex();
            final Iterator<SCell> iter = row.getCellIterator(start, end);
            while (iter.hasNext()) {
                final int[] ltrb = getCellMinMax(row.getSheet(), index, iter);
                if (ltrb != null) {
                    return false;
                }
            }
            return true;
        }

        private int[] getNonBlankCell(AbstractSheetAdv sheet, AbstractRowAdv row, int left, int right) {
            if (isRowBlank(row, left, right)) {
                return null;
            }
            final int rowIdx = row.getIndex();
            int minc = left;
            if (llm <= left - 1) {
                final Iterator<SCell> leftIter = row.getCellIterator(true, llm, left - 1);
                while (leftIter.hasNext()) {
                    final int[] ltrb = getCellMinMax(sheet, rowIdx, leftIter);
                    if (ltrb != null) {
                        if (minc - 1 > ltrb[LTRB_R]) {
                            break;
                        } else {
                            minc = ltrb[LTRB_L];
                        }
                    }
                }
            }
            int maxc = right;
            if (rlm >= right + 1) {
                final Iterator<SCell> rightIter = row.getCellIterator(right + 1, rlm);
                while (rightIter.hasNext()) {
                    final int[] ltrb = getCellMinMax(sheet, rowIdx, rightIter);
                    if (ltrb != null) {
                        if (maxc + 1 < ltrb[LTRB_L]) {
                            break;
                        } else {
                            maxc = ltrb[LTRB_R];
                        }
                    }
                }
            }
            return new int[]{minc, maxc};
        }

        private CellRegion findCurrentRegion0(AbstractSheetAdv sheet, int row, int col, int lastRow, int lastCol) {
            int l = col;
            int t = row;
            int r = lastCol;
            int b = lastRow;
            int t0 = t - 1;
            int b0 = b + 1;
            if (t0 >= tlm) {
                final Iterator<AbstractRowAdv> topIter = sheet.getRowIterator(true, tlm, t0);
                // find if there contain non blank contiguous cells at top of the range
                while (topIter.hasNext()) {
                    final AbstractRowAdv topRow = topIter.next();
                    if (topRow.getIndex() < t0) {
                        // empty row in between
                        break;
                    } else {
                        final int[] lr = getNonBlankCell(sheet, topRow, l, r);
                        if (lr == null) {
                            break;
                        } else {
                            t = t0;
                            l = Math.min(l, lr[LR_L]);
                            r = Math.max(r, lr[LR_R]);
                            t0 -= 1;
                        }
                    }
                }

            }
            // find if there contain non blank contiguous cells at bottom of the range
            if (b0 <= blm) {
                final Iterator<SRow> bottomIter = sheet.getRowIterator(b0, blm);
                while (bottomIter.hasNext()) {
                    final SRow next = bottomIter.next();
                    if (next.getIndex() > b0) {
                        // empty row in between
                        break;
                    } else {
                        final int[] lr = getNonBlankCell(sheet, (AbstractRowAdv) next, l, r);
                        if (lr == null) {
                            break;
                        } else {
                            b = b0;
                            l = Math.min(l, lr[LR_L]);
                            r = Math.max(r, lr[LR_R]);
                            b0 += 1;
                        }
                    }
                }
            }
            final int t1 = Math.max(tlm, t - 1);
            final int b1 = Math.min(b + 1, blm);
            if (t1 <= b1) {
                final Iterator<SRow> iter = sheet.getRowIterator(t1, b1);
                while (iter.hasNext()) {
                    final SRow srow = iter.next();
                    final int rowIndex = srow.getIndex();
                    // find if there contain non blank contiguous cells at row of the range
                    int l0 = l - 1;
                    if (l0 >= llm) {
                        final Iterator<SCell> leftIter = ((AbstractRowAdv) srow).getCellIterator(
                                true, llm, l0);
                        while (leftIter.hasNext()) {
                            final int[] ltrb = getCellMinMax(sheet, rowIndex, leftIter);
                            if (ltrb != null) {
                                if (l0 > ltrb[LTRB_R]) {
                                    break;
                                } else {
                                    l = l0;
                                    l0 = ltrb[LTRB_L];
                                    t = Math.min(t, ltrb[LTRB_T]);
                                    b = Math.max(b, ltrb[LTRB_B]);
                                }
                            }
                        }
                    }

                    // find if there contain non blank contiguous cells at right of the range
                    int r0 = r + 1;
                    if (r0 <= rlm) {
                        boolean isExtended = false;
                        final Iterator<SCell> rightIter = srow.getCellIterator(r0, rlm);
                        while (rightIter.hasNext()) {
                            final int[] ltrb = getCellMinMax(sheet, rowIndex, rightIter);
                            if (ltrb != null) {
                                if (r0 < ltrb[LTRB_L]) {
                                    break;
                                } else {
                                    r = r0;
                                    r0 = ltrb[LTRB_R];
                                    t = Math.min(t, ltrb[LTRB_T]);
                                    b = Math.max(b, ltrb[LTRB_B]);
                                }
                            }
                        }
                    }
                }
            }

            // the neighboring cells are all blank, return current range
            if (l == col && t == row && r == lastCol && b == lastRow) {
                if (l == r && t == b && isBlank(sheet.getCell(t, l))) {
                    return null;
                } else {
                    return new CellRegion(t, l, b, r);
                }
            }
            // recursive for next round
            return findCurrentRegion0(sheet, t, l, b, r);
        }

        private CellRegion getRowCurrentRegion(SSheet sheet, int topRow, int btmRow) {
            calcBoundaries(sheet, topRow, btmRow);
            int minc = -1;
            int maxc = -1;
            int minr = -1;
            int maxr = -1;
            boolean found = false;
            final Iterator<SRow> iter = sheet.getRowIterator(topRow, btmRow);
            while (iter.hasNext()) {
                final SRow row = iter.next();
                final int[] ltrb = getRowMinMax(sheet, row);
                if (ltrb != null) {
                    if (found) {
                        minc = Math.min(minc, ltrb[LTRB_L]);
                        minr = Math.min(minr, ltrb[LTRB_T]);
                        maxc = Math.max(maxc, ltrb[LTRB_R]);
                        maxr = Math.max(maxr, ltrb[LTRB_B]);
                    } else {
                        found = true;
                        minc = ltrb[LTRB_L];
                        minr = ltrb[LTRB_T];
                        maxc = ltrb[LTRB_R];
                        maxr = ltrb[LTRB_B];
                    }
                }
            }
            return found ? new CellRegion(minr, minc, maxr, maxc) : null;
        }
    }
}
