/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/12/18 , Created by Hawk
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
 */
package io.keikai.range.impl.imexp;

import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.stream.Collectors;

import io.keikai.model.CellRegion;
import io.keikai.model.ErrorValue;
import io.keikai.model.SBook;
import io.keikai.model.SCell;
import io.keikai.model.SCell.CellType;
import io.keikai.model.SCellStyle;
import io.keikai.model.SColor;
import io.keikai.model.SColumnArray;
import io.keikai.model.SComment;
import io.keikai.model.SExtraStyle;
import io.keikai.model.SFill.FillPattern;
import io.keikai.model.SFont;
import io.keikai.model.SFooter;
import io.keikai.model.SHeader;
import io.keikai.model.SHyperlink;
import io.keikai.model.SName;
import io.keikai.model.SNamedStyle;
import io.keikai.model.SPictureData;
import io.keikai.model.SPrintSetup;
import io.keikai.model.SRichText;
import io.keikai.model.SRichText.Segment;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.model.SSheetProtection;
import io.keikai.model.SSheetViewInfo;
import io.keikai.model.STableStyle;
import io.keikai.model.SWorkbookProtection;
import io.keikai.model.SheetRegion;
import io.keikai.model.impl.AbstractBookAdv;
import io.keikai.model.impl.ColumnArrayImpl;
import io.keikai.model.impl.TableNameImpl;
import io.keikai.model.sys.EngineFactory;
import io.keikai.model.sys.format.FormatContext;
import io.keikai.model.sys.format.FormatEngine;
import io.keikai.model.util.Strings;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFDataFormat;
import org.apache.poi.hssf.usermodel.HSSFPrintSetup;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.util.HSSFColor;
import org.apache.poi.ooxml.POIXMLDocumentPart;
import org.apache.poi.ooxml.POIXMLDocumentPartHelper;
import org.apache.poi.ss.SpreadsheetVersion;
import org.apache.poi.ss.usermodel.BorderStyle;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.ClientAnchor;
import org.apache.poi.ss.usermodel.Color;
import org.apache.poi.ss.usermodel.Comment;
import org.apache.poi.ss.usermodel.CreationHelper;
import org.apache.poi.ss.usermodel.DataFormat;
import org.apache.poi.ss.usermodel.Drawing;
import org.apache.poi.ss.usermodel.FillPatternType;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.Footer;
import org.apache.poi.ss.usermodel.Header;
import org.apache.poi.ss.usermodel.HorizontalAlignment;
import org.apache.poi.ss.usermodel.Hyperlink;
import org.apache.poi.ss.usermodel.Name;
import org.apache.poi.ss.usermodel.NamedStyle;
import org.apache.poi.ss.usermodel.PrintSetup;
import org.apache.poi.ss.usermodel.RichTextString;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.SheetProtection;
import org.apache.poi.ss.usermodel.SheetVisibility;
import org.apache.poi.ss.usermodel.TableStyle;
import org.apache.poi.ss.usermodel.VerticalAlignment;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.ZssContext;
import org.apache.poi.ss.util.AreaReference;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.xssf.model.ExternalLinksTable;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFPrintSetup;
import org.apache.poi.xssf.usermodel.XSSFPrintSetupHelper;
import org.apache.poi.xssf.usermodel.XSSFRow;
import org.apache.poi.xssf.usermodel.XSSFSheet;
import org.apache.poi.xssf.usermodel.XSSFSheetHelper;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbookHelper;
import org.openxmlformats.schemas.spreadsheetml.x2006.main.CTExternalReference;
import org.openxmlformats.schemas.spreadsheetml.x2006.main.CTExternalReferences;
import org.openxmlformats.schemas.spreadsheetml.x2006.main.CTRow;
import org.openxmlformats.schemas.spreadsheetml.x2006.main.CTSheetFormatPr;
import org.openxmlformats.schemas.spreadsheetml.x2006.main.CTWorksheet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Common exporting behavior for both XLSX and XLS.
 * 
 * @author kuro, Hawk
 * @since 3.5.0
 */
abstract public class AbstractExcelExporter extends AbstractExporter {
	private static final long serialVersionUID = -453628222967064565L;

	//Though there is customHeight defined in <sheetFormatPr>. However, it is 
	//  only for empty rows; not rows with cells. For rows with cells, still 
	//  have to specifies its height or it will use application's default row 
	//  height instead.
	protected static final int DEFAULT_ROW_HEIGHT = 20; //ZSS-1012

	/**
	 * Exporting destination, POI book model
	 */
	protected Workbook workbook;

	protected SBook sbook;

	/**
	 * The map stores the exported {@link CellStyle} during exporting, so that
	 * we can reuse them for exporting other cells.
	 */
	protected Map<SCellStyle, CellStyle> styleTable = new HashMap<SCellStyle, CellStyle>();
	protected Map<SFont, Font> fontTable = new HashMap<SFont, Font>();
	protected Map<SColor, Color> colorTable = new HashMap<SColor, Color>();
	//ZSS-688: SPictureData index -> poi PictureData index
	protected Map<Integer, Integer> exportedPicDataMap = new HashMap<Integer, Integer>();
	//ZSS-1188
	protected Map<STableStyle, TableStyle> tbStyleTable = new HashMap<STableStyle, TableStyle>();
	
	private static final Logger _logger = LoggerFactory.getLogger(AbstractExcelExporter.class);

	abstract protected void exportColumnArray(SSheet sheet, Sheet poiSheet, SColumnArray columnArr);

	abstract protected Workbook createPoiBook();

	abstract protected void exportWorkbookProtection(SWorkbookProtection workbookProtection);

	abstract protected void exportChart(SSheet sheet, Sheet poiSheet);

	abstract protected void exportPicture(SSheet sheet, Sheet poiSheet);

	abstract protected void exportValidation(SSheet sheet, Sheet poiSheet);

	abstract protected void exportAutoFilter(SSheet sheet, Sheet poiSheet);

	protected abstract void exportSheetPr(SSheet sheet, Sheet poiSheet);

	protected abstract void exportExtLst(SSheet sheet, Sheet poiSheet);

	abstract protected void exportPassword(SSheet sheet, Sheet poiSheet);
	
	abstract protected int exportTables(SSheet sheet, Sheet poiSheet, int tbId); //ZSS-855
	
	abstract protected void exportConditionalFormatting(SSheet sheet, Sheet poiSheet); //ZSS-1141
	// KEIKAI-454
	protected ExportPhase _exportPhase;
	
	/**
	 * Export the model according to reversed depended order: book, sheet,
	 * defined name, cells, chart, pictures, validation. Because named ranges
	 * (defined names) require sheet index, they should be imported after sheets
	 * created. Besides, cells, charts, and validations may have formulas
	 * referring to named ranges, they must be imported after named ranged.
	 * Pictures depend on cells.
	 */
	@Override
	public void export(SBook book, OutputStream fos) throws IOException {
		ReadWriteLock lock = book.getBookSeries().getLock();
		lock.readLock().lock();

		try {
			// clear cache for reuse
			clearCache();
			_exportPhase = new ExportPhase(book.getBookName());
			workbook = createPoiBook();
			sbook = book;

			// fix for IssueK201Test for xlsx only to add external book for Ref3dPxg to use
			List<SBook> books = book.getBookSeries().getBooks();
			if (!books.isEmpty() && workbook instanceof XSSFWorkbook) {
				books.forEach(sbook -> workbook.linkExternalWorkbook(sbook.getBookName().replace(" ", "+"), createPoiBook()));
			}

			//ZSS-854: export default cell styles
			_exportPhase.setPhase(ExportPhase.CELL_STYLE);

			// Xls doesn't support these exporting functions.
			if (workbook instanceof XSSFWorkbook) {
				XSSFWorkbook xssfWorkbook = (XSSFWorkbook) workbook;
				XSSFWorkbookHelper.clearDefaultCellStyles(xssfWorkbook);
				for (SCellStyle style : book.getDefaultCellStyles()) {
					CellStyle cellStyle = toPOIDefaultCellStyle(style); // put the default cellStyle
					XSSFWorkbookHelper.addDefaultCellStyle(xssfWorkbook, cellStyle);
					toPOICellStyle(style); //ZSS-1027: put the default cellStyle into cellXfs
				}
				//ZSS-854: export named cell styles
				_exportPhase.setPhase(ExportPhase.NAMED_CELL_STYLE);
				XSSFWorkbookHelper.clearNamedStyles(xssfWorkbook);
				XSSFWorkbookHelper.addNamedStyles(xssfWorkbook, book.getNamedStyles().stream().map(this::toPOINamedStyle).collect(
						Collectors.toList()));

				//ZSS-1145: export Dxf cell styles
				_exportPhase.setPhase(ExportPhase.DXF_CELL_STYLE);
				XSSFWorkbookHelper.clearDxfCellStyles(xssfWorkbook);
				for (SExtraStyle style : book.getExtraStyles()) {
					addPOIDxfCellStyle(style); // put the Dxf cellStyle
				}

				//ZSS-1188: export TableStyles
				_exportPhase.setPhase(ExportPhase.WORKBOOK_TABLE_STYLE);
				XSSFWorkbookHelper.setDefaultPivotStyle(xssfWorkbook, book.getDefaultPivotStyleName());
				XSSFWorkbookHelper.setDefaultTableStyle(xssfWorkbook, book.getDefaultTableStyleName());
				XSSFWorkbookHelper.clearTableStyles(xssfWorkbook);
				for (STableStyle style : book.getTableStyles()) {
					addPOITableStyle(style); // put the POI TableStyle
				}
			}
			_exportPhase.setPhase(ExportPhase.WORKBOOK_PROTECTION);
			exportWorkbookProtection(book.getWorkbookProtection());
			
			int tbId = 0;
			_exportPhase.setPhase(ExportPhase.SHEETS);
			int firstVisibleSheetIndex = -1;
			int selectedSheetIndex = -1;
			for (int n = 0; n < book.getSheets().size(); n++) {
				SSheet sheet = book.getSheet(n);
				exportSheet(sheet);
				if (SSheet.SheetVisible.VISIBLE.equals(sheet.getSheetVisible())) {
					if (firstVisibleSheetIndex == -1) {
						firstVisibleSheetIndex = n; // first visible sheet
					}
					if (selectedSheetIndex == -1 && sheet.getViewInfo().isTabSelected()) {
						selectedSheetIndex = n; // first visible and selected sheet
					}
				}
				Sheet poiSheet = workbook.getSheetAt(n);
				tbId = exportTables(sheet, poiSheet, tbId); //ZSS-855, ZSS-1011
				
				exportConditionalFormatting(sheet, poiSheet); //ZSS-1141
				_exportPhase.setSheetName(null);
			}
			boolean selectedSheetExisted = selectedSheetIndex != -1;
			if (selectedSheetExisted || firstVisibleSheetIndex != -1) {
				_exportPhase.setPhase(ExportPhase.SHEET_ACTIVE);
				workbook.setActiveSheet(selectedSheetExisted ? selectedSheetIndex : firstVisibleSheetIndex);
			}
			_exportPhase.setPhase(ExportPhase.NAMED_RANGE);
			exportNamedRange(book);
			_exportPhase.setPhase(ExportPhase.WORKBOOK_PICTURE_DATA);
			exportPictureData(book); //ZSS-735
			for (int n = 0; n < book.getSheets().size(); n++) {
				_exportPhase.setPhase(ExportPhase.SHEET);
				SSheet sheet = book.getSheet(n);
				_exportPhase.setSheetName(sheet.getSheetName());
				Sheet poiSheet = workbook.getSheetAt(n);

				if (poiSheet instanceof XSSFSheet) {
					XSSFSheet xssfSheet = (XSSFSheet) poiSheet;
					CTWorksheet ctSheet = xssfSheet.getCTWorksheet();
					CTSheetFormatPr ctSheetFormatPr = ctSheet.getSheetFormatPr();
					if (ctSheetFormatPr == null) {
						ctSheetFormatPr = ctSheet.addNewSheetFormatPr();
					}
					if (sheet.getOutlineLevelRow() > 0) {
						ctSheetFormatPr.setOutlineLevelRow((short) sheet.getOutlineLevelRow());
					}
					if (sheet.getOutlineLevelCol() > 0) {
						ctSheetFormatPr.setOutlineLevelCol((short) sheet.getOutlineLevelCol());
					}
				}
				exportRowColumn(sheet, poiSheet);
				exportMergedRegions(sheet, poiSheet);
				exportChart(sheet, poiSheet);
				exportPicture(sheet, poiSheet);
				exportValidation(sheet, poiSheet);
				exportAutoFilter(sheet, poiSheet);
				exportExtLst(sheet, poiSheet);
				_exportPhase.setSheetName(null);
			}

			_exportPhase.setPhase(ExportPhase.WRITE_WORKBOOK);

			// fix for IssueK201Test for xlsx only
			if (!books.isEmpty() && workbook instanceof XSSFWorkbook) {
				XSSFWorkbook xssfWorkbook = (XSSFWorkbook) workbook;
				// remove all temp external books.
				List<POIXMLDocumentPart.RelationPart> relationParts = xssfWorkbook.getRelationParts();
				CTExternalReferences externalReferences = xssfWorkbook.getCTWorkbook()
						.getExternalReferences();
				for (SBook sbook : books) {
					CTExternalReference[] externalReferenceArray = externalReferences.getExternalReferenceArray();
					for (POIXMLDocumentPart.RelationPart relationPart : relationParts) {
						if (relationPart.getDocumentPart() instanceof ExternalLinksTable) {
							ExternalLinksTable linksTable = relationPart.getDocumentPart();
							String linkedFileName = linksTable.getLinkedFileName();
							if (linkedFileName.equals(sbook.getBookName())) {
								for (int i = 0; i < externalReferenceArray.length; i++) {
									if (externalReferenceArray[i].getId().equals(relationPart.getRelationship().getId())) {
										externalReferences.removeExternalReference(i);
										break;
									}
								}
								POIXMLDocumentPartHelper.removeRelation(xssfWorkbook, relationPart.getDocumentPart());
								break;
							}
						}
					}
				}
				if (externalReferences.sizeOfExternalReferenceArray() == 0) {
					xssfWorkbook.getCTWorkbook().unsetExternalReferences();
				}
			}

			workbook.write(fos);
		} catch (RuntimeException e) { // KEIKAI-454
			throw new ExportException("Fail to export " +  _exportPhase.toString(), e);
		} finally {
			lock.readLock().unlock();
			clearCache();
		}
	}

	private void clearCache() {
		styleTable.clear();
		fontTable.clear();
		colorTable.clear();
		tbStyleTable.clear(); //ZSS-1188
		_exportPhase = null; //reset
		workbook = null; // clear workbook
		sbook = null;
	}
	
	//ZSS-735
	protected void exportPictureData(SBook book) {
		for (SPictureData picData : book.getPicturesDatas()) {
			int poiIndex = workbook.addPicture(picData.getData(), PoiEnumConversion.toPoiPictureFormat(picData.getFormat()));
			exportedPicDataMap.put(picData.getIndex(), poiIndex);
		}
	}

	protected void exportNamedRange(SBook book) {
		for (SName name : book.getNames()) {
			if (name instanceof TableNameImpl) continue; //ZSS-855: skip Name associated with Table
			if (name.getName().startsWith("_xlnm.")) continue; // KEIKAI-135: skip built-in named range
			Name poiName = workbook.createName();
			try{
				String sheetName = name.getApplyToSheetName();//ZSS-699
				if(sheetName!=null){
					poiName.setSheetIndex(workbook.getSheetIndex(sheetName));
				}
				poiName.setNameName(name.getName()); //ZSS-699: execute after setSheetIndex() to avoid check duplicate name issue
				//zss-214, to tolerate the name refers to formula error (#REF!!$A$1:$I$18)
				if(!name.isFormulaParsingError()){
					poiName.setRefersToFormula(name.getRefersToFormula());
				}
			// KEIKAI-135
//			} catch (IllegalArgumentException e) {
//				String nn = name.getName();
//				if (nn != null && nn.startsWith("_xlnm.")) {
//					//ignore name with _xlnm.*
//				} else {
//					throw e;
//				}
			} catch (Exception e) {
				//ZSS-645 catch the exception happens when a book has a named range referring to an external book in XLS
				_logger.warn("Cannot export a name range: "+name.getName(),e);
				if (poiName.getNameName()!=null){
					workbook.removeName(poiName);
				}
			}
			
		}
	}

	protected void exportSheet(SSheet sheet) {
		_exportPhase.setPhase(ExportPhase.SHEET);
		String sheetName = sheet.getSheetName();
		_exportPhase.setSheetName(sheetName);
		Sheet poiSheet = workbook.createSheet(sheetName);

		exportSheetPr(sheet, poiSheet);

		_exportPhase.setPhase(ExportPhase.SHEET_VIEW);
		SSheetViewInfo viewInfo = sheet.getViewInfo();
		// refer to BookHelper#setFreezePanel
		int freezeRow = viewInfo.getNumOfRowFreeze();
		int freezeCol = viewInfo.getNumOfColumnFreeze();
		poiSheet.createFreezePane(freezeCol <= 0 ? 0 : freezeCol, freezeRow <= 0 ? 0 : freezeRow);

		poiSheet.setDisplayGridlines(viewInfo.isDisplayGridlines());
		for (Integer rowIndex : viewInfo.getRowBreaks()){
			poiSheet.setRowBreak(rowIndex);
		}
		for (Integer colIndex : viewInfo.getColumnBreaks()){
			poiSheet.setColumnBreak(colIndex);
		}

		poiSheet.setSelected(viewInfo.isTabSelected());

		_exportPhase.setPhase(ExportPhase.SHEET_PROTECTION);
		exportSheetProtection(sheet, poiSheet); //ZSS-576
		if (sheet.isProtected()) {
			if (workbook.getSpreadsheetVersion() == SpreadsheetVersion.EXCEL97) {
				poiSheet.protectSheet(""); // without password; set hashed password directly later
			} else {
				XSSFSheetHelper.protectSheet((XSSFSheet) poiSheet, "");
			}
			exportPassword(sheet, poiSheet);
		} else {
			//ZSS-679
			if (poiSheet instanceof XSSFSheet) {
				XSSFSheetHelper.unsetSheetProtection((XSSFSheet) poiSheet);
			} else {
				poiSheet.protectSheet(null);
			}
		}

		_exportPhase.setPhase(ExportPhase.SHEET_FORMAT);
		poiSheet.setDefaultRowHeight((short) UnitUtil.pxToTwip(sheet.getDefaultRowHeight()));
		//ZSS-1132
		final AbstractBookAdv book = (AbstractBookAdv) sheet.getBook();
		poiSheet.setDefaultColumnWidth((int) UnitUtil.pxToDefaultColumnWidth(sheet.getDefaultColumnWidth(), book.getCharWidth()));

		// Header
		_exportPhase.setPhase(ExportPhase.SHEET_HEADER);
		Header header = poiSheet.getHeader();
		header.setLeft(viewInfo.getHeader().getLeftText());
		header.setCenter(viewInfo.getHeader().getCenterText());
		header.setRight(viewInfo.getHeader().getRightText());

		// Footer
		_exportPhase.setPhase(ExportPhase.SHEET_FOOTER);
		Footer footer = poiSheet.getFooter();
		footer.setLeft(viewInfo.getFooter().getLeftText());
		footer.setCenter(viewInfo.getFooter().getCenterText());
		footer.setRight(viewInfo.getFooter().getRightText());

		SPrintSetup sps = sheet.getPrintSetup();
		
		// Margin
		_exportPhase.setPhase(ExportPhase.SHEET_PRINT_SETUP);
		poiSheet.setMargin(Sheet.LeftMargin, sps.getLeftMargin());
		poiSheet.setMargin(Sheet.RightMargin, sps.getRightMargin());
		poiSheet.setMargin(Sheet.TopMargin, sps.getTopMargin());
		poiSheet.setMargin(Sheet.BottomMargin, sps.getBottomMargin());
		poiSheet.setMargin(Sheet.HeaderMargin, sps.getHeaderMargin());
		poiSheet.setMargin(Sheet.FooterMargin, sps.getFooterMargin());

		// Print Setup Information
		PrintSetup poips = poiSheet.getPrintSetup();
		if (sps.isDifferentOddEvenPage()) {
			SHeader evenHeader = sps.getEvenHeader();
			if (evenHeader != null) {
				Header poiEvenHeader = poiSheet instanceof HSSFSheet ? ((HSSFSheet) poiSheet)
						.getEvenHeader() : ((XSSFSheet) poiSheet).getEvenHeader();
				poiEvenHeader.setCenter(evenHeader.getCenterText());
				poiEvenHeader.setLeft(evenHeader.getLeftText());
				poiEvenHeader.setRight(evenHeader.getRightText());
			}
			SFooter evenFooter = sps.getEvenFooter();
			if (evenFooter != null) {
				Footer poiEvenFooter = poiSheet instanceof HSSFSheet ? ((HSSFSheet) poiSheet)
						.getEvenFooter() : ((XSSFSheet) poiSheet).getEvenFooter();
				poiEvenFooter.setCenter(evenFooter.getCenterText());
				poiEvenFooter.setLeft(evenFooter.getLeftText());
				poiEvenFooter.setRight(evenFooter.getRightText());
			}
		}
		if (sps.isDifferentFirstPage()) {
			SHeader firstHeader = sps.getFirstHeader();
			if (firstHeader != null) {
				Header poiFirstHeader = poiSheet instanceof HSSFSheet ? ((HSSFSheet) poiSheet)
						.getFirstHeader() : ((XSSFSheet) poiSheet).getFirstHeader();
				poiFirstHeader.setCenter(firstHeader.getCenterText());
				poiFirstHeader.setLeft(firstHeader.getLeftText());
				poiFirstHeader.setRight(firstHeader.getRightText());
			}
			SFooter firstFooter = sps.getFirstFooter();
			if (firstFooter != null) {
				Footer poiFirstFooter = poiSheet instanceof HSSFSheet ? ((HSSFSheet) poiSheet)
						.getFirstFooter() : ((XSSFSheet) poiSheet).getFirstFooter();
				poiFirstFooter.setCenter(firstFooter.getCenterText());
				poiFirstFooter.setLeft(firstFooter.getLeftText());
				poiFirstFooter.setRight(firstFooter.getRightText());
			}
		}

		if (poiSheet instanceof HSSFSheet) {
			((HSSFSheet) poiSheet).setAlignMargins(sps.isAlignWithMargins());
		} else {
			((XSSFSheet) poiSheet).getHeaderFooterProperties().setAlignWithMargins(sps.isAlignWithMargins());
		}

		if (poips instanceof HSSFPrintSetup) {
			((HSSFPrintSetup) poips).setErrorsMode(sps.getErrorPrintMode());
		} else {
			XSSFPrintSetupHelper.setErrorsMode((XSSFPrintSetup) poips, sps.getErrorPrintMode());
		}

		poips.setFitHeight((short) sps.getFitHeight());
		poips.setFitWidth((short) sps.getFitWidth());
		
		poiSheet.setHorizontallyCenter(sps.isHCenter());
		poiSheet.setVerticallyCenter(sps.isVCenter());
		poips.setLandscape(sps.isLandscape());
		poips.setLeftToRight(sps.isLeftToRight());

		int pageStart = sps.getPageStart();
		poips.setUsePage(pageStart > 0);
		poips.setPageStart((short) (pageStart > 0 ? pageStart : 0));
		poips.setPaperSize(PoiEnumConversion.toPoiPaperSize(sps.getPaperSize()));
		if (poips instanceof HSSFPrintSetup) {
			((HSSFPrintSetup) poips).setCommentsMode(sps.getCommentsMode());
		} else {
			XSSFPrintSetupHelper.setCommentsMode((XSSFPrintSetup) poips, sps.getCommentsMode());
		}
		poiSheet.setPrintGridlines(sps.isPrintGridlines());
		if (poiSheet instanceof HSSFSheet) {
			((HSSFSheet) poiSheet).setPrintHeadings(sps.isPrintHeadings());
		} else {
			XSSFSheetHelper.setPrintHeadings((XSSFSheet) poiSheet, sps.isPrintHeadings());
		}
		poips.setScale((short)sps.getScale());
		if (poiSheet instanceof HSSFSheet) {
			((HSSFSheet) poiSheet).setScalWithDoc(sps.isScaleWithDoc());
		} else {
			((XSSFSheet) poiSheet).getHeaderFooterProperties().setScaleWithDoc(sps.isScaleWithDoc());
		}
		if (poiSheet instanceof HSSFSheet) {
			((HSSFSheet) poiSheet).setDiffOddEven(sps.isDifferentOddEvenPage());
		} else {
			((XSSFSheet) poiSheet).getHeaderFooterProperties().setDifferentOddEven(sps.isDifferentOddEvenPage());
		}
		if (poiSheet instanceof HSSFSheet) {
			((HSSFSheet) poiSheet).setDiffFirst(sps.isDifferentOddEvenPage());
		} else {
			((XSSFSheet) poiSheet).getHeaderFooterProperties().setDifferentFirst(sps.isDifferentFirstPage());
		}
		
		int sheetIndex = workbook.getNumberOfSheets() - 1;
		String area = sps.getPrintArea();
		if (!Strings.isEmpty(area)) {
			_exportPhase.setPhase(ExportPhase.SHEET_PRINT_AREA);
			AreaReference areaReference = new AreaReference(area,
					workbook.getSpreadsheetVersion());
			workbook.setPrintArea(sheetIndex,
					areaReference.getFirstCell().formatAsString(false) + ':'
							+ areaReference.getLastCell().formatAsString(false));
			_exportPhase.setPhase(ExportPhase.SHEET_PRINT_SETUP);
		}

		final CellRegion rgn = sps.getRepeatingRowsTitle();
		if (rgn != null) {
			CellRangeAddress rowrng = new CellRangeAddress(rgn.getRow(), rgn.getLastRow(), -1, -1);
			poiSheet.setRepeatingRows(rowrng);
		}
		
		final CellRegion crgn= sps.getRepeatingColumnsTitle();
		if (crgn != null) {
			CellRangeAddress colrng = new CellRangeAddress(-1, -1, crgn.getColumn(), crgn.getLastColumn());
			poiSheet.setRepeatingColumns(colrng);
		}
		
		//ZSS-832
		// export sheet visible
		_exportPhase.setPhase(ExportPhase.SHEET_VISIBLE);
		SheetVisibility option = SheetVisibility.VISIBLE;
		switch(sheet.getSheetVisible()) {
		default:
		case VISIBLE:
			option = SheetVisibility.VISIBLE;
			break;
		case HIDDEN:
			option = SheetVisibility.HIDDEN;
			break;
		case VERY_HIDDEN:
			option = SheetVisibility.VERY_HIDDEN;
			break;
		}
	    workbook.setSheetVisibility(sheetIndex, option);
	}

	protected void exportMergedRegions(SSheet sheet, Sheet poiSheet) {
		_exportPhase.setPhase(ExportPhase.MERGED_CELL);
		// consistent with importer, read from last merged region
		for (int i = sheet.getNumOfMergedRegion() - 1; i >= 0; i--) {
			CellRegion region = sheet.getMergedRegion(i);
			poiSheet.addMergedRegion(new CellRangeAddress(region.row, region.lastRow, region.column, region.lastColumn));
		}
	}

	protected void exportRowColumn(SSheet sheet, Sheet poiSheet) {
		// export rows
		Iterator<SRow> rowIterator = sheet.getRowIterator();
		while (rowIterator.hasNext()) {
			SRow row = rowIterator.next();
			exportRow(sheet, poiSheet, row);
		}
		_exportPhase.setRowIndex(-1);
		_exportPhase.setColIndex(-1);

		// export columns
		Iterator<SColumnArray> columnArrayIterator = sheet.getColumnArrayIterator();
		while (columnArrayIterator.hasNext()) {
			SColumnArray columnArr = columnArrayIterator.next();
			if (((ColumnArrayImpl)columnArr).shouldSkip()) continue; //ZSS-1132, ZSS-1190
			exportColumnArray(sheet, poiSheet, columnArr);
		}
		_exportPhase.setColIndex(-1);
	}

	protected void exportRow(SSheet sheet, Sheet poiSheet, SRow row) {
		_exportPhase.setPhase(ExportPhase.ROW);
		int rowIndex = row.getIndex();
		_exportPhase.setRowIndex(rowIndex);
		Row poiRow = poiSheet.getRow(rowIndex);
		if (poiRow == null) {
			poiRow = poiSheet.createRow(rowIndex);
		}

		if (row.isHidden()) {
			// hidden, set height as 0
			poiRow.setZeroHeight(true);
		} else {
			// not hidden, calculate height
			//ZSS-1012
			// We export both customHeight and Height when customHeight is set.
			// otherwise, we may export only height value if height is not equal to DEFAULT_ROW_HEIGHT
			if (row.isCustomHeight()) {
//				poiRow.setCustomHeight(true);
				poiRow.setHeight((short) UnitUtil.pxToTwip(row.getHeight()));
			} else if (row.getHeight() != DEFAULT_ROW_HEIGHT) {
				poiRow.setHeight((short) UnitUtil.pxToTwip(row.getHeight()));
			}
		}
		if (poiRow instanceof XSSFRow) {
			CTRow ctRow = ((XSSFRow) poiRow).getCTRow();
			if (row.isCollapsed()) {
				ctRow.setCollapsed(true);
			}
			if (row.getOutlineLevel() > 0) {
				ctRow.setOutlineLevel((short) row.getOutlineLevel());
			}
		}

		SCellStyle rowStyle = row.getCellStyle();
		CellStyle poiRowStyle = toPOICellStyle(rowStyle);
		poiRow.setRowStyle(poiRowStyle);

		// Export Cell
		_exportPhase.setPhase(ExportPhase.CELL);
		Iterator<SCell> cellIterator = sheet.getCellIterator(row.getIndex());
		while (cellIterator.hasNext()) {
			SCell cell = cellIterator.next();
			exportCell(poiRow, cell);
		}
		_exportPhase.setPhase(ExportPhase.ROW);
	}

	protected void exportCell(Row poiRow, SCell cell) {
		int columnIndex = cell.getColumnIndex();
		_exportPhase.setColIndex(columnIndex);
		Cell poiCell = poiRow.createCell(columnIndex);

		SCellStyle cellStyle = cell.getCellStyle();
		poiCell.setCellStyle(toPOICellStyle(cellStyle));

		switch (cell.getType()) {
		case BLANK:
			poiCell.setBlank();
			break;
		case ERROR:
			if(cell.getErrorValue().getCode() == ErrorValue.INVALID_FORMULA){
				//ignore the value of this cell, excel doesn't allow it invalid formula (parsing error).
				_logger.warn("Don't export an invalid formula: {} at {}",
						"="+cell.getFormulaValue(),
						new SheetRegion(cell.getSheet(), cell.getReferenceString()).getReferenceString());
			}else{
				poiCell.setCellErrorValue(cell.getErrorValue().getCode());
			}
			break;
		case BOOLEAN:
			poiCell.setCellValue(cell.getBooleanValue());
			break;
		case FORMULA:
			if(cell.getFormulaResultType()==CellType.ERROR && cell.getErrorValue().getCode() == ErrorValue.INVALID_FORMULA){//ZSS-891
				//ignore the value of this cell, excel doesn't allow invalid formulas (parsing error).
				_logger.warn("Don't export an invalid formula: {} at {}",
						"="+cell.getFormulaValue(),
						new SheetRegion(cell.getSheet(), cell.getReferenceString()).getReferenceString());
			}else{
				if (cell.isPartOfArrayFormulaGroup()) {
					CellRegion arrayFormulaRange = cell.getArrayFormulaRegion();

					// KEIKAI-667, only top-left cell has array formula
					if (arrayFormulaRange.getRow() == cell.getRowIndex() && arrayFormulaRange.getColumn() == cell.getColumnIndex()) {
						poiCell.getSheet().setArrayFormula(getFormulaString(cell), CellRangeAddress.valueOf(arrayFormulaRange.getReferenceString()));
					}
				} else {
					poiCell.setCellFormula(getFormulaString(cell));
					if (_logger.isDebugEnabled()) {
						_logger.debug("export a formula at {}",
								new SheetRegion(cell.getSheet(), cell.getReferenceString()).getReferenceString());
					}
					//ZSS-873
					if (isExportCache()) {
						switch (cell.getFormulaResultType()) {
						default:
						case BLANK:
							break;
						case BOOLEAN:
							poiCell.setCellValue(cell.getBooleanValue());
							break;
						case ERROR:
							poiCell.setCellErrorValue(cell.getErrorValue().getCode());
							break;
						case NUMBER:
							poiCell.setCellValue((Double) cell.getNumberValue());
							break;
						case STRING:
							if (cell.isRichTextValue()) {
								poiCell.setCellValue(toPOIRichText(cell.getRichTextValue()));
							} else {
								poiCell.setCellValue(cell.getStringValue());
							}
						}
					}
				}
			}
			break;
		case NUMBER:
			poiCell.setCellValue(cell.getNumberValue());
			break;
		case STRING:
			if(cell.isRichTextValue()) {
				poiCell.setCellValue(toPOIRichText(cell.getRichTextValue()));
			} else {
				poiCell.setCellValue(cell.getStringValue());
			}
			break;
		default:
		}
		if (_logger.isDebugEnabled()) {
			_logger.debug("exported a cell type {} at {}", cell.getType(),
					new SheetRegion(cell.getSheet(), cell.getReferenceString()).getReferenceString());
		}
		_exportPhase.setPhase(ExportPhase.HYPERLINK);
		SHyperlink hyperlink = cell.getHyperlink();
		if (hyperlink != null) {
			CreationHelper helper = workbook.getCreationHelper();
			try{
				Hyperlink poiHyperlink = helper.createHyperlink(PoiEnumConversion.toPoiHyperlinkType(hyperlink.getType()));
				poiHyperlink.setAddress(hyperlink.getAddress());
				poiHyperlink.setLabel(hyperlink.getLabel());
				poiCell.setHyperlink(poiHyperlink);
			}catch (Exception e) {
				//ZSS-644 catch the exception happens when a hyperlink has an invalid URI in XLSX
				_logger.warn("Cannot export a hyperlink: "+hyperlink.getAddress(),e);
			}
			
		}
		
		_exportPhase.setPhase(ExportPhase.COMMENT);
		SComment comment = cell.getComment();
		if (comment != null) {
			// Refer to the POI Official Tutorial
			// http://poi.apache.org/spreadsheet/quick-guide.html#CellComments
			CreationHelper helper = workbook.getCreationHelper();
			Drawing drawing = poiCell.getSheet().createDrawingPatriarch();
			ClientAnchor anchor = helper.createClientAnchor();
			anchor.setCol1(poiCell.getColumnIndex());
			anchor.setCol2(poiCell.getColumnIndex() + 1);
			anchor.setRow1(poiRow.getRowNum());
			anchor.setRow2(poiRow.getRowNum() + 3);
			Comment poiComment = drawing.createCellComment(anchor);
			SRichText richText = comment.getRichText();
			if (richText != null) {
				poiComment.setString(toPOIRichText(richText));
			} else {
				poiComment.setString(helper.createRichTextString(comment.getText()));
			}
			poiComment.setAuthor(comment.getAuthor());
			poiComment.setVisible(comment.isVisible());
			poiCell.setCellComment(poiComment);
		}	
		_exportPhase.setPhase(ExportPhase.CELL);
	}

	//CellImpl.getFormulaValue() doesn't contains external book name
	protected String getFormulaString(SCell cell){
		FormatEngine fe = EngineFactory.getInstance().createFormatEngine();
		// KEIKAI-734: handle "_xlfn."
		String originalFormulaValue = cell.getFormulaValue();
		String prefix = "_xlfn.";
		boolean prefixFound = originalFormulaValue.startsWith(prefix);
		String formulaString = fe.getEditText(cell, new FormatContext(ZssContext.getCurrent().getLocale())).substring(1);
		return prefixFound && !formulaString.startsWith(prefix) ? prefix + formulaString: formulaString;
	}

	protected RichTextString toPOIRichText(SRichText richText) {

		CreationHelper helper = workbook.getCreationHelper();

		RichTextString poiRichTextString = helper.createRichTextString(richText.getText());

		int start = 0;
		int end = 0;
		for (Segment sg : richText.getSegments()) {
			SFont font = sg.getFont();
			int len = sg.getText().length();
			end += len;
			poiRichTextString.applyFont(start, end, toPOIFont(font));
			start += len;
		}

		return poiRichTextString;
	}
	//ZSS-854
	protected NamedStyle toPOINamedStyle(SNamedStyle cellStyle) {
		// only work for xlsx
		if (workbook instanceof XSSFWorkbook) {
			return XSSFWorkbookHelper.createNamedStyle((XSSFWorkbook) workbook, cellStyle.getName(), cellStyle.isCustomBuiltin(), cellStyle.getBuiltinId(),
					cellStyle.getIndex());
		}
		return null;
	}
	
	//ZSS-854
	protected CellStyle toPOIDefaultCellStyle(SCellStyle cellStyle) {
		return null;
	}

	protected CellStyle toPOICellStyle(SCellStyle cellStyle) {
		// instead of creating a new style, use old one if exist
		CellStyle poiCellStyle = styleTable.get(cellStyle);
		if (poiCellStyle != null) {
			return poiCellStyle;
		}
		poiCellStyle = workbook.createCellStyle();

		//set Border
		BorderStyle bottom = PoiEnumConversion.toPoiBorderType(cellStyle.getBorderBottom());
		BorderStyle left = PoiEnumConversion.toPoiBorderType(cellStyle.getBorderLeft());
		BorderStyle right = PoiEnumConversion.toPoiBorderType(cellStyle.getBorderRight());
		BorderStyle top = PoiEnumConversion.toPoiBorderType(cellStyle.getBorderTop());
		Color bottomColor = toPOIColor(cellStyle.getBorderBottomColor());
		Color leftColor = toPOIColor(cellStyle.getBorderLeftColor());
		Color rightColor = toPOIColor(cellStyle.getBorderRightColor());
		Color topColor = toPOIColor(cellStyle.getBorderTopColor());
		poiCellStyle.setBorderRight(right);
		poiCellStyle.setBorderLeft(left);
		poiCellStyle.setBorderTop(top);
		poiCellStyle.setBorderBottom(bottom);

		if (poiCellStyle instanceof HSSFCellStyle) {
			((HSSFCellStyle) poiCellStyle).setBottomBorderColor((HSSFColor) bottomColor);
			((HSSFCellStyle) poiCellStyle).setTopBorderColor((HSSFColor) topColor);
			((HSSFCellStyle) poiCellStyle).setRightBorderColor((HSSFColor) rightColor);
			((HSSFCellStyle) poiCellStyle).setLeftBorderColor((HSSFColor) leftColor);
		} else {
			((XSSFCellStyle) poiCellStyle).setBottomBorderColor((XSSFColor) bottomColor);
			((XSSFCellStyle) poiCellStyle).setTopBorderColor((XSSFColor) topColor);
			((XSSFCellStyle) poiCellStyle).setRightBorderColor((XSSFColor) rightColor);
			((XSSFCellStyle) poiCellStyle).setLeftBorderColor((XSSFColor) leftColor);
		}

		// fill
		//ZSS-857: SOLID pattern; switch fgColor and bgColor 
		SColor fgColor = cellStyle.getFillColor();
		SColor bgColor = cellStyle.getBackColor();
		if (cellStyle.getFillPattern() == FillPattern.SOLID) {
			SColor tmp = fgColor;
			fgColor = bgColor;
			bgColor = tmp;
		}
		Color fillColor = toPOIColor(fgColor);
		Color backColor = toPOIColor(bgColor);
		FillPatternType pattern = PoiEnumConversion.toPoiFillPattern(cellStyle.getFillPattern());
		poiCellStyle.setFillForegroundColor(fillColor);
		poiCellStyle.setFillBackgroundColor(backColor);
		poiCellStyle.setFillPattern(pattern);
		
		//cell Alignment
		HorizontalAlignment hAlign = PoiEnumConversion.toPoiHorizontalAlignment(cellStyle.getAlignment());
		VerticalAlignment vAlign = PoiEnumConversion.toPoiVerticalAlignment(cellStyle.getVerticalAlignment());
		boolean wrapText = cellStyle.isWrapText();

		//ZSS-1020
		poiCellStyle.setAlignment(hAlign);
		poiCellStyle.setVerticalAlignment(vAlign);
		poiCellStyle.setWrapText(wrapText);
		poiCellStyle.setRotation((short) cellStyle.getRotation());

		//protect
		boolean locked = cellStyle.isLocked();
		boolean hidden = cellStyle.isHidden();
		poiCellStyle.setLocked(locked);
		poiCellStyle.setHidden(hidden);

		// number format, refer from BookHelper#setDataFormat
		DataFormat dataFormat = workbook.createDataFormat();
		if(cellStyle.isDirectDataFormat()){ // explicit number format
			if (dataFormat instanceof HSSFDataFormat) {
				short index = ((HSSFDataFormat) dataFormat).newFormat(cellStyle.getDataFormat());
				poiCellStyle.setDataFormat(index);
			} else {
				poiCellStyle.setDataFormat((short) ((XSSFWorkbook) workbook).getStylesSource().putNumberFormat(cellStyle.getDataFormat()));
			}
		}else { // implied number format
			short index = dataFormat.getFormat(cellStyle.getDataFormat());
			poiCellStyle.setDataFormat(index);
		}

		// font
		poiCellStyle.setFont(toPOIFont(cellStyle.getFont()));

		// put into table
		styleTable.put(cellStyle, poiCellStyle);
		
		int indention = cellStyle.getIndention();
		if (indention > 0) 
			poiCellStyle.setIndention((short) indention);

		return poiCellStyle;

	}

	protected Color toPOIColor(SColor color) {
		if (color == null) return null; //ZSS-1145
		
		Color poiColor = colorTable.get(color);

		if (poiColor != null) {
			return poiColor;
		}
		poiColor = BookHelper.HTMLToColor(workbook, color.getHtmlColor());
		colorTable.put(color, poiColor);
		return poiColor;
	}

	/**
	 * Convert ZSS Font into POI Font. Cache font in the fontTable. If font
	 * exist, don't create a new one.
	 * 
	 * @param font
	 * @return
	 */
	protected Font toPOIFont(SFont font) {
		if (font == null) return null; //ZSS-1138
			
		Font poiFont = fontTable.get(font);
		if (poiFont != null) {
			return poiFont;
		}

		poiFont = workbook.createFont();
		poiFont.setBold(PoiEnumConversion.toPoiBoldweight(font.getBoldweight()) == PoiEnumConversion.BOLDWEIGHT_BOLD);
		poiFont.setStrikeout(font.isStrikeout());
		poiFont.setItalic(font.isItalic());
		BookHelper.setFontColor(workbook, poiFont, toPOIColor(font.getColor()));
		poiFont.setFontHeightInPoints((short) font.getHeightPoints());
		poiFont.setFontName(font.getName());
		poiFont.setTypeOffset(PoiEnumConversion.toPoiTypeOffset(font.getTypeOffset()));
		poiFont.setUnderline(PoiEnumConversion.toPoiUnderline(font.getUnderline()));

		// put into table
		fontTable.put(font, poiFont);

		return poiFont;
	}

	/**
	 * POI SheetProtection.
	 * @param sheet destination sheet
	 * @param poiSheet source POI sheet
	 */
	protected void exportSheetProtection(SSheet sheet, Sheet poiSheet) { //ZSS-576
		SSheetProtection ssp = sheet.getSheetProtection();
		SheetProtection sp = poiSheet instanceof HSSFSheet ? ((HSSFSheet) poiSheet).getOrCreateSheetProtection()
				: XSSFSheetHelper.getOrCreateSheetProtection((XSSFSheet) poiSheet);

	    sp.setAutoFilter(ssp.isAutoFilter());
	    sp.setDeleteColumns(ssp.isDeleteColumns());
	    sp.setDeleteRows(ssp.isDeleteRows());
	    sp.setFormatCells(ssp.isFormatCells());
	    sp.setFormatColumns(ssp.isFormatColumns());
	    sp.setFormatRows(ssp.isFormatRows());
	    sp.setInsertColumns(ssp.isInsertColumns());
	    sp.setInsertHyperlinks(ssp.isInsertHyperlinks());
	    sp.setInsertRows(ssp.isInsertRows());
	    sp.setPivotTables(ssp.isPivotTables());
	    sp.setSort(ssp.isSort());
	    sp.setObjects(ssp.isObjects());
	    sp.setScenarios(ssp.isScenarios());
	    sp.setSelectLockedCells(ssp.isSelectLockedCells());
	    sp.setSelectUnlockedCells(ssp.isSelectUnlockedCells());
	}
	
	//ZSS-873: Import formula cache result from an Excel file
	protected boolean _exportCache = false;
	/**
	 * Set whether export cached value into excel file(must be called before
	 * export() is called.
	 * @param b
	 * @since 3.7.0
	 */
	public void setExportCache(boolean b) {
		_exportCache = b;
	}
	/**
	 * Returns whether export cached value into excel file.
	 * @return
	 * @since 3.7.0
	 */
	protected boolean isExportCache() {
		return _exportCache;
	}
	
	/**
	 * Add DxfCellStyle as poi Dxf. 
	 * @param extraStyle
	 * @since 3.8.2
	 */
	protected void addPOIDxfCellStyle(SExtraStyle extraStyle) {
		//Should be override in ExcelXlsxExporter
	}
	
	/**
	 * Add TableStyle as poi TableStyle
	 * @param tableStyle
	 * @since 3.8.3
	 */
	protected void addPOITableStyle(STableStyle tableStyle) {
		//Should be override in ExcelXlsxExporter
	}
}
