/*

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2014/1/24 , Created by Kuro
}}IS_NOTE

Copyright (C) 2014 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/

package io.keikai.range.impl.imexp;

import static org.apache.poi.xddf.usermodel.chart.XDDFDateAxisHelper.TimeUnit;

import io.keikai.model.ErrorValue;
import io.keikai.model.SAutoFilter.FilterOp;
import io.keikai.model.SBorder.BorderType;
import io.keikai.model.SCellStyle.Alignment;
import io.keikai.model.SCellStyle.VerticalAlignment;
import io.keikai.model.SChart;
import io.keikai.model.SDataValidation.AlertStyle;
import io.keikai.model.SDataValidation.OperatorType;
import io.keikai.model.SDataValidation.ValidationType;
import io.keikai.model.SFill.FillPattern;
import io.keikai.model.SFont;
import io.keikai.model.SFont.Boldweight;
import io.keikai.model.SFont.TypeOffset;
import io.keikai.model.SFont.Underline;
import io.keikai.model.SHyperlink;
import io.keikai.model.SHyperlink.HyperlinkType;
import io.keikai.model.SPicture.Format;
import io.keikai.model.SPrintSetup;
import io.keikai.model.impl.AbstractChartAxisAdv;
import io.keikai.model.impl.chart.CategoryAxisImpl;
import io.keikai.model.impl.chart.DateAxisImpl;
import io.keikai.model.impl.chart.ValueAxisImpl;
import org.apache.poi.ss.usermodel.AutoFilter;
import org.apache.poi.ss.usermodel.BorderStyle;
import org.apache.poi.ss.usermodel.DataValidation;
import org.apache.poi.ss.usermodel.DataValidationConstraint;
import org.apache.poi.ss.usermodel.FillPatternType;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.HorizontalAlignment;
import org.apache.poi.ss.usermodel.PrintSetup;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xddf.usermodel.chart.AxisCrossBetween;
import org.apache.poi.xddf.usermodel.chart.AxisCrosses;
import org.apache.poi.xddf.usermodel.chart.AxisLabelAlignment;
import org.apache.poi.xddf.usermodel.chart.AxisOrientation;
import org.apache.poi.xddf.usermodel.chart.AxisPosition;
import org.apache.poi.xddf.usermodel.chart.AxisTickLabelPosition;
import org.apache.poi.xddf.usermodel.chart.BarDirection;
import org.apache.poi.xddf.usermodel.chart.LegendPosition;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

/**
 * An utility to convert enumeration constant between ZSS model and POI model.
 * This utility helps developers who want to write their own importer or exporter based on POI.
 * 
 * @author kuro, Hawk
 * @since 3.5.0
 */
public class PoiEnumConversion {
	
	public static org.apache.poi.common.usermodel.HyperlinkType toPoiHyperlinkType(SHyperlink.HyperlinkType type) {
		switch(type) {
		case DOCUMENT:
			return org.apache.poi.common.usermodel.HyperlinkType.DOCUMENT;
		case EMAIL:
			return org.apache.poi.common.usermodel.HyperlinkType.EMAIL;
		case FILE:
			return org.apache.poi.common.usermodel.HyperlinkType.FILE;
		case URL:
			default:
			return org.apache.poi.common.usermodel.HyperlinkType.URL;
		}
	}
	
	public static SHyperlink.HyperlinkType toHyperlinkType(
			org.apache.poi.common.usermodel.HyperlinkType type) {
		switch (type) {
		case DOCUMENT:
			return HyperlinkType.DOCUMENT;
		case EMAIL:
			return HyperlinkType.EMAIL;
		case FILE:
			return HyperlinkType.FILE;
		case URL:
		default:
			return HyperlinkType.URL;
		}
	}

	public static int toPoiFilterOperator(FilterOp operator){
		switch(operator){
			case and:
				return AutoFilter.FILTEROP_AND;
			case or:
				return AutoFilter.FILTEROP_OR;
			case top10:
				return AutoFilter.FILTEROP_TOP10;
			case cellColor: //ZSS-1191
				return AutoFilter.FILTEROP_CELL_COLOR;
			case fontColor: //ZSS-1191
				return AutoFilter.FILTEROP_FONT_COLOR;
			case values:
			default:
				return AutoFilter.FILTEROP_VALUES;
		}
	}

	public static short toPoiPaperSize(SPrintSetup.PaperSize paperSize) {
		switch(paperSize) {
		case A3:
			return PrintSetup.A3_PAPERSIZE;
		case A4_EXTRA:
			return PrintSetup.A4_EXTRA_PAPERSIZE;
		case A4_PLUS:
			return PrintSetup.A4_PLUS_PAPERSIZE;
		case A4_ROTATED:
			return PrintSetup.A4_ROTATED_PAPERSIZE;
		case A4_SMALL:
			return PrintSetup.A4_SMALL_PAPERSIZE;
		case A4:
			return PrintSetup.A4_PAPERSIZE;
		case A4_TRANSVERSE:
			return PrintSetup.A4_TRANSVERSE_PAPERSIZE;
		case A5:
			return PrintSetup.A5_PAPERSIZE;
		case B4:
			return PrintSetup.B4_PAPERSIZE;
		case B5:
			return PrintSetup.B5_PAPERSIZE;
		case ELEVEN_BY_SEVENTEEN:
			return PrintSetup.ELEVEN_BY_SEVENTEEN_PAPERSIZE;
		case ENVELOPE_10:
			return PrintSetup.ENVELOPE_10_PAPERSIZE;
		case ENVELOPE_9:
			return PrintSetup.ENVELOPE_9_PAPERSIZE;
		case ENVELOPE_C3:
			return PrintSetup.ENVELOPE_C3_PAPERSIZE;
		case ENVELOPE_C4:
			return PrintSetup.ENVELOPE_C4_PAPERSIZE;
		case ENVELOPE_C5:
			return PrintSetup.ENVELOPE_C5_PAPERSIZE;
		case ENVELOPE_C6:
			return PrintSetup.ENVELOPE_C6_PAPERSIZE;
		case ENVELOPE_CS:
			return PrintSetup.ENVELOPE_CS_PAPERSIZE;
		case ENVELOPE_DL:
			return PrintSetup.ENVELOPE_DL_PAPERSIZE;
		case ENVELOPE_MONARCH:
			return PrintSetup.ENVELOPE_MONARCH_PAPERSIZE;
		case EXECUTIVE:
			return PrintSetup.EXECUTIVE_PAPERSIZE;
		case FOLIO8:
			return PrintSetup.FOLIO8_PAPERSIZE;
		case LEDGER:
			return PrintSetup.LEDGER_PAPERSIZE;
		case LEGAL:
			return PrintSetup.LEGAL_PAPERSIZE;
		case LETTER:
			return PrintSetup.LETTER_PAPERSIZE;
		case LETTER_ROTATED:
			return PrintSetup.LETTER_ROTATED_PAPERSIZE;
		case LETTER_SMALL:
			return PrintSetup.LETTER_SMALL_PAGESIZE;
		case NOTE8:
			return PrintSetup.NOTE8_PAPERSIZE;
		case QUARTO:
			return PrintSetup.QUARTO_PAPERSIZE;
		case STATEMENT:
			return PrintSetup.STATEMENT_PAPERSIZE;
		case TABLOID:
			return PrintSetup.TABLOID_PAPERSIZE;
		case TEN_BY_FOURTEEN:
			return PrintSetup.TEN_BY_FOURTEEN_PAPERSIZE;
			default:
				return PrintSetup.A4_PAPERSIZE;
		}
	}
	
	public static SPrintSetup.PaperSize toPaperSize(short paperSize) {
		switch (paperSize) {
		case PrintSetup.A3_PAPERSIZE:
			return SPrintSetup.PaperSize.A3;
		case PrintSetup.A4_EXTRA_PAPERSIZE:
			return SPrintSetup.PaperSize.A4_EXTRA;
		case PrintSetup.A4_PAPERSIZE:
			return SPrintSetup.PaperSize.A4;
		case PrintSetup.A4_PLUS_PAPERSIZE:
			return SPrintSetup.PaperSize.A4_PLUS;
		case PrintSetup.A4_ROTATED_PAPERSIZE:
			return SPrintSetup.PaperSize.A4_ROTATED;
		case PrintSetup.A4_SMALL_PAPERSIZE:
			return SPrintSetup.PaperSize.A4_SMALL;
		case PrintSetup.A4_TRANSVERSE_PAPERSIZE:
			return SPrintSetup.PaperSize.A4_TRANSVERSE;
		case PrintSetup.A5_PAPERSIZE:
			return SPrintSetup.PaperSize.A5;
		case PrintSetup.B4_PAPERSIZE:
			return SPrintSetup.PaperSize.B4;
		case PrintSetup.B5_PAPERSIZE:
			return SPrintSetup.PaperSize.B5;
		case PrintSetup.ELEVEN_BY_SEVENTEEN_PAPERSIZE:
			return SPrintSetup.PaperSize.ELEVEN_BY_SEVENTEEN;
		case PrintSetup.ENVELOPE_10_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_10;
		case PrintSetup.ENVELOPE_9_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_9;
		case PrintSetup.ENVELOPE_C3_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_C3;
		case PrintSetup.ENVELOPE_C4_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_C4;
		case PrintSetup.ENVELOPE_C5_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_C5;
		case PrintSetup.ENVELOPE_C6_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_C6;
		case PrintSetup.ENVELOPE_DL_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_DL;
		case PrintSetup.ENVELOPE_MONARCH_PAPERSIZE:
			return SPrintSetup.PaperSize.ENVELOPE_MONARCH;
		case PrintSetup.EXECUTIVE_PAPERSIZE:
			return SPrintSetup.PaperSize.EXECUTIVE;
		case PrintSetup.FOLIO8_PAPERSIZE:
			return SPrintSetup.PaperSize.FOLIO8;
		case PrintSetup.LEDGER_PAPERSIZE:
			return SPrintSetup.PaperSize.LEDGER;
		case PrintSetup.LETTER_PAPERSIZE:
			return SPrintSetup.PaperSize.LETTER;
		case PrintSetup.LETTER_ROTATED_PAPERSIZE:
			return SPrintSetup.PaperSize.LETTER_ROTATED;
		case PrintSetup.LETTER_SMALL_PAGESIZE:
			return SPrintSetup.PaperSize.LETTER_SMALL;
		case PrintSetup.NOTE8_PAPERSIZE:
			return SPrintSetup.PaperSize.NOTE8;
		case PrintSetup.QUARTO_PAPERSIZE:
			return SPrintSetup.PaperSize.QUARTO;
		case PrintSetup.STATEMENT_PAPERSIZE:
			return SPrintSetup.PaperSize.STATEMENT;
		case PrintSetup.TABLOID_PAPERSIZE:
			return SPrintSetup.PaperSize.TABLOID;
		case PrintSetup.TEN_BY_FOURTEEN_PAPERSIZE:
			return SPrintSetup.PaperSize.TEN_BY_FOURTEEN;
		default:
			return SPrintSetup.PaperSize.A4;
		}
	}

	public static org.apache.poi.ss.usermodel.VerticalAlignment toPoiVerticalAlignment(VerticalAlignment vAlignment) {
		switch(vAlignment) {
			case BOTTOM:
				return org.apache.poi.ss.usermodel.VerticalAlignment.BOTTOM;
			case CENTER:
				return org.apache.poi.ss.usermodel.VerticalAlignment.CENTER;
			case JUSTIFY:
				return org.apache.poi.ss.usermodel.VerticalAlignment.JUSTIFY;
			case TOP:
			default:
				return org.apache.poi.ss.usermodel.VerticalAlignment.TOP;
		}
	}
	
	public static VerticalAlignment toVerticalAlignment(
			org.apache.poi.ss.usermodel.VerticalAlignment poiVerticalAlignment) {
		switch (poiVerticalAlignment) {
		case TOP:
			return VerticalAlignment.TOP;
		case CENTER:
			return VerticalAlignment.CENTER;
		case JUSTIFY:
			return VerticalAlignment.JUSTIFY;
		case BOTTOM:
		default:
			return VerticalAlignment.BOTTOM;
		}
	}

	public static FillPatternType toPoiFillPattern(FillPattern fillPattern) {
		switch(fillPattern) {
		case DARK_GRAY:
			return FillPatternType.ALT_BARS;
		case DARK_GRID:
			return FillPatternType.BIG_SPOTS;
		case DARK_TRELLIS:
			return FillPatternType.BRICKS;
		case LIGHT_TRELLIS:
			return FillPatternType.DIAMONDS;
		case MEDIUM_GRAY:
			return FillPatternType.FINE_DOTS;
		case GRAY0625:
			return FillPatternType.LEAST_DOTS;
		case GRAY125:
			return FillPatternType.LESS_DOTS;
		case SOLID:
			return FillPatternType.SOLID_FOREGROUND;
		case LIGHT_GRAY:
			return FillPatternType.SPARSE_DOTS;
		case LIGHT_GRID:
			return FillPatternType.SQUARES;
		case DARK_DOWN:
			return FillPatternType.THICK_BACKWARD_DIAG;
		case DARK_UP:
			return FillPatternType.THICK_FORWARD_DIAG;
		case DARK_HORIZONTAL:
			return FillPatternType.THICK_HORZ_BANDS;
		case DARK_VERTICAL:
			return FillPatternType.THICK_VERT_BANDS;
		case LIGHT_DOWN:
			return FillPatternType.THIN_BACKWARD_DIAG;
		case LIGHT_UP:
			return FillPatternType.THIN_FORWARD_DIAG;
		case LIGHT_HORIZONTAL:
			return FillPatternType.THIN_HORZ_BANDS;
		case LIGHT_VERTICAL:
			return FillPatternType.THIN_VERT_BANDS;
		case NONE:
		default:
			return FillPatternType.NO_FILL;
		}
	}
	
	public static FillPattern toFillPattern(FillPatternType poiFillPattern) {
		if (poiFillPattern == null) return null; //ZSS-1140
		
		switch (poiFillPattern) {
		case SOLID_FOREGROUND:
			return FillPattern.SOLID;
		case FINE_DOTS:
			return FillPattern.MEDIUM_GRAY;
		case ALT_BARS:
			return FillPattern.DARK_GRAY;
		case SPARSE_DOTS:
			return FillPattern.LIGHT_GRAY;
		case THICK_HORZ_BANDS:
			return FillPattern.DARK_HORIZONTAL;
		case THICK_VERT_BANDS:
			return FillPattern.DARK_VERTICAL;
		case THICK_BACKWARD_DIAG:
			return FillPattern.DARK_DOWN;
		case THICK_FORWARD_DIAG:
			return FillPattern.DARK_UP;
		case BIG_SPOTS:
			return FillPattern.DARK_GRID;
		case BRICKS:
			return FillPattern.DARK_TRELLIS;
		case THIN_HORZ_BANDS:
			return FillPattern.LIGHT_HORIZONTAL;
		case THIN_VERT_BANDS:
			return FillPattern.LIGHT_VERTICAL;
		case THIN_BACKWARD_DIAG:
			return FillPattern.LIGHT_DOWN;
		case THIN_FORWARD_DIAG:
			return FillPattern.LIGHT_UP;
		case SQUARES:
			return FillPattern.LIGHT_GRID;
		case DIAMONDS:
			return FillPattern.LIGHT_TRELLIS;
		case LESS_DOTS:
			return FillPattern.GRAY125;
		case LEAST_DOTS:
			return FillPattern.GRAY0625;
		case NO_FILL:
		default:
			return FillPattern.NONE;
		}
	}

	public static org.apache.poi.ss.usermodel.BorderStyle toPoiBorderType(BorderType borderType) {
		switch(borderType) {
		case DASH_DOT:
			return BorderStyle.DASH_DOT;
		case DASHED:
			return BorderStyle.DASHED;
		case DOTTED:
			return BorderStyle.DOTTED;
		case DOUBLE:
			return BorderStyle.DOUBLE;
		case HAIR:
			return BorderStyle.HAIR;
		case MEDIUM:
			return BorderStyle.MEDIUM;
		case MEDIUM_DASH_DOT:
			return BorderStyle.DASH_DOT;
		case MEDIUM_DASH_DOT_DOT:
			return BorderStyle.DASH_DOT_DOT;
		case MEDIUM_DASHED:
			return BorderStyle.MEDIUM_DASHED;
		case SLANTED_DASH_DOT:
			return BorderStyle.SLANTED_DASH_DOT;
		case THICK:
			return BorderStyle.THICK;
		case THIN:
			return BorderStyle.THIN;
		case DASH_DOT_DOT:
			return BorderStyle.DASH_DOT_DOT;
		case NONE:
		default:
			return BorderStyle.NONE;
		}
	}
	
	public static BorderType toBorderType(BorderStyle poiBorder) {
		if (poiBorder == null) return null; //ZSS-1140
		
		switch (poiBorder) {
		case THIN:
			return BorderType.THIN;
		case MEDIUM:
			return BorderType.MEDIUM;
		case DASHED:
			return BorderType.DASHED;
		case HAIR:
			return BorderType.HAIR;
		case THICK:
			return BorderType.THICK;
		case DOUBLE:
			return BorderType.DOUBLE;
		case DOTTED:
			return BorderType.DOTTED;
		case MEDIUM_DASHED:
			return BorderType.MEDIUM_DASHED;
		case DASH_DOT:
			return BorderType.DASH_DOT;
		case MEDIUM_DASH_DOT:
			return BorderType.MEDIUM_DASH_DOT;
		case DASH_DOT_DOT:
			return BorderType.DASH_DOT_DOT;
		case MEDIUM_DASH_DOT_DOT:
			return BorderType.MEDIUM_DASH_DOT_DOT;
		case SLANTED_DASH_DOT:
			return BorderType.SLANTED_DASH_DOT;
		case NONE:
		default:
			return BorderType.NONE;
		}
	}

	public static HorizontalAlignment toPoiHorizontalAlignment(Alignment alignment) {
		switch(alignment) {
		case CENTER:
			return HorizontalAlignment.CENTER;
		case FILL:
			return HorizontalAlignment.FILL;
		case JUSTIFY:
			return HorizontalAlignment.JUSTIFY;
		case RIGHT:
			return HorizontalAlignment.RIGHT;
		case LEFT:
			return HorizontalAlignment.LEFT;
		case CENTER_SELECTION:
			return HorizontalAlignment.CENTER_SELECTION;
		case GENERAL:
			default:
			return HorizontalAlignment.GENERAL;
		}
	}
	
	public static Alignment toHorizontalAlignment(HorizontalAlignment poiHorizontalAlignment) {
		switch (poiHorizontalAlignment) {
		case LEFT:
			return Alignment.LEFT;
		case RIGHT:
			return Alignment.RIGHT;
		case CENTER:
			return Alignment.CENTER;
		case CENTER_SELECTION: //ZSS-779
			return Alignment.CENTER_SELECTION;
		case FILL:
			return Alignment.FILL;
		case JUSTIFY:
			return Alignment.JUSTIFY;
		case GENERAL:
		default:
			return Alignment.GENERAL;
		}
	}

	public static short toPoiBoldweight(Boldweight bold) {
		switch(bold) {
			case BOLD:
				return BOLDWEIGHT_BOLD;
			case NORMAL:
			default:
				return BOLDWEIGHT_NORMAL;
		}
	}


	/**
	 * Normal boldness (not bold) for xls use only
	 */
	private static final short BOLDWEIGHT_NORMAL = 0x190;

	/**
	 * Bold boldness (bold) for xls use only
	 */
	public static final short BOLDWEIGHT_BOLD = 0x2bc;

	public static Boldweight toBoldweight(boolean bold) {
		return bold ? Boldweight.BOLD : Boldweight.NORMAL;
	}
	
	public static Boldweight toBoldweight(short bold) {
		switch(bold) {
			case BOLDWEIGHT_BOLD:
				return Boldweight.BOLD;
			case BOLDWEIGHT_NORMAL:
			default:
				return Boldweight.NORMAL;
		}
	}
	
	public static short toPoiTypeOffset(TypeOffset typeOffset) {
		switch(typeOffset) {
			case SUPER:
				return Font.SS_SUPER;
			case SUB:
				return Font.SS_SUB;
			case NONE:
			default:
				return Font.SS_NONE;
		}
	}
	
	public static TypeOffset toTypeOffset(short typeOffset) {
		switch (typeOffset) {
		case Font.SS_SUB:
			return TypeOffset.SUB;
		case Font.SS_SUPER:
			return TypeOffset.SUPER;
		case Font.SS_NONE:
		default:
			return TypeOffset.NONE;
		}
	}

	public static byte toPoiUnderline(Underline underline) {
		switch(underline) {
			case SINGLE:
				return Font.U_SINGLE;
			case DOUBLE:
				return Font.U_DOUBLE;
			case DOUBLE_ACCOUNTING:
				return Font.U_DOUBLE_ACCOUNTING;
			case SINGLE_ACCOUNTING:
				return Font.U_SINGLE_ACCOUNTING;
			case NONE:
			default:
				return Font.U_NONE;
		}
	}

	/*
	 * reference BookHelper.getFontCSSStyle()
	 */
	public static Underline toUnderline(byte underline) {
		switch (underline) {
		case Font.U_SINGLE:
			return SFont.Underline.SINGLE;
		case Font.U_DOUBLE:
			return SFont.Underline.DOUBLE;
		case Font.U_SINGLE_ACCOUNTING:
			return SFont.Underline.SINGLE_ACCOUNTING;
		case Font.U_DOUBLE_ACCOUNTING:
			return SFont.Underline.DOUBLE_ACCOUNTING;
		case Font.U_NONE:
		default:
			return SFont.Underline.NONE;
		}
	}

	public static int toPoiPictureFormat(Format format){
		switch(format){
		case GIF:
			return XSSFWorkbook.PICTURE_TYPE_GIF;
		case JPG:
			return Workbook.PICTURE_TYPE_JPEG;
		case PNG:
		default:
			return Workbook.PICTURE_TYPE_PNG;
		}
	}
	//There is no format information on POI's Picture class

	public static org.apache.poi.xddf.usermodel.chart.Grouping toPoiGrouping(
			SChart.ChartGrouping grouping){
		if (grouping == null){ //no API to read grouping from XLS, so chart might have null grouping to export
			return org.apache.poi.xddf.usermodel.chart.Grouping.STANDARD;
		}
		switch(grouping) {
		case PERCENT_STACKED:
			return org.apache.poi.xddf.usermodel.chart.Grouping.PERCENT_STACKED;
		case STACKED:
			return org.apache.poi.xddf.usermodel.chart.Grouping.STACKED;
		case STANDARD:
		default:
			return org.apache.poi.xddf.usermodel.chart.Grouping.STANDARD;
		}
	}
	public static org.apache.poi.xddf.usermodel.chart.BarGrouping toPoiBarGrouping(
			SChart.ChartGrouping grouping){
		if (grouping == null){ //no API to read grouping from XLS, so chart might have null grouping to export
			return org.apache.poi.xddf.usermodel.chart.BarGrouping.STANDARD;
		}
		switch(grouping){
		case CLUSTERED:
			return org.apache.poi.xddf.usermodel.chart.BarGrouping.CLUSTERED;
		case PERCENT_STACKED:
			return org.apache.poi.xddf.usermodel.chart.BarGrouping.PERCENT_STACKED;
		case STACKED:
			return org.apache.poi.xddf.usermodel.chart.BarGrouping.STACKED;
		case STANDARD:
		default:
			return org.apache.poi.xddf.usermodel.chart.BarGrouping.STANDARD;
		}
	}

//	public static io.keikai.model.SChart.ChartGrouping
//		toChartGrouping(org.apache.poi.xddf.usermodel.chart.Grouping grouping){
//		if (grouping == null) {
//			return io.keikai.model.SChart.ChartGrouping.STANDARD;
//		}
//		switch(grouping){
//		case STACKED:
//			return io.keikai.model.SChart.ChartGrouping.STACKED;
//		case PERCENT_STACKED:
//			return io.keikai.model.SChart.ChartGrouping.PERCENT_STACKED;
//		case CLUSTERED:
//			return io.keikai.model.SChart.ChartGrouping.CLUSTERED;
//		case STANDARD:
//		default:
//			return io.keikai.model.SChart.ChartGrouping.STANDARD;
//		}
//	}

	public static BarDirection toPoiBarDirection(SChart.BarDirection direction){
		switch(direction){
			case VERTICAL:
				return BarDirection.COL;
			case HORIZONTAL:
			default:
				return BarDirection.BAR;
		}

	}

	public static SChart.BarDirection toBarDirection(BarDirection direction){
		switch(direction){
		case COL:
			return SChart.BarDirection.VERTICAL;
		case BAR:
		default:
			return SChart.BarDirection.HORIZONTAL;
		}
	}

	public static LegendPosition toPoiLegendPosition(
			SChart.ChartLegendPosition position){
		switch(position){
			case BOTTOM:
				return LegendPosition.BOTTOM;
			case TOP:
				return LegendPosition.TOP;
			case TOP_RIGHT:
				return LegendPosition.TOP_RIGHT;
			case LEFT:
				return LegendPosition.LEFT;
			case RIGHT:
			default:
				return LegendPosition.RIGHT;

		}
	}

	public static SChart.ChartLegendPosition toLengendPosition(LegendPosition position){
		switch(position){
		case BOTTOM:
			return SChart.ChartLegendPosition.BOTTOM;
		case LEFT:
			return SChart.ChartLegendPosition.LEFT;
		case TOP:
			return SChart.ChartLegendPosition.TOP;
		case TOP_RIGHT:
			return SChart.ChartLegendPosition.TOP_RIGHT;
		case RIGHT:
		default:
			return SChart.ChartLegendPosition.RIGHT;
		}

	}

	public static AbstractChartAxisAdv.Crosses toAxisCrosses(AxisCrosses crosses) {
		switch (crosses) {
			case MAX:
				return AbstractChartAxisAdv.Crosses.MAX;
			case MIN:
				return AbstractChartAxisAdv.Crosses.MIN;
			case AUTO_ZERO:
			default:
				return AbstractChartAxisAdv.Crosses.AUTO_ZERO;
		}
	}

	public static AxisCrosses toPoiAxisCrosses(AbstractChartAxisAdv.Crosses crosses) {
		switch (crosses) {
			case MAX:
				return AxisCrosses.MAX;
			case MIN:
				return AxisCrosses.MIN;
			case AUTO_ZERO:
			default:
				return AxisCrosses.AUTO_ZERO;
		}
	}

	public static AxisOrientation toPoiAxisOrientation(AbstractChartAxisAdv.Orientation orientation) {
		switch (orientation) {
			case MAX_MIN:
				return AxisOrientation.MAX_MIN;
			case MIN_MAX:
			default:
				return AxisOrientation.MIN_MAX;
		}
	}

	public static AbstractChartAxisAdv.Orientation toAxisOrientation(AxisOrientation orientation) {
		switch (orientation) {
			case MAX_MIN:
				return AbstractChartAxisAdv.Orientation.MAX_MIN;
			case MIN_MAX:
			default:
				return AbstractChartAxisAdv.Orientation.MIN_MAX;
		}
	}

	public static AxisPosition toPoiAxisPosition(AbstractChartAxisAdv.Position position) {
		switch (position) {
			case LEFT:
				return AxisPosition.LEFT;
			case TOP:
				return AxisPosition.TOP;
			case RIGHT:
				return AxisPosition.RIGHT;
			case BOTTOM:
			default:
				return AxisPosition.BOTTOM;
		}
	}

	public static AbstractChartAxisAdv.Position toAxisPosition(AxisPosition position) {
		switch (position) {
			case LEFT:
				return AbstractChartAxisAdv.Position.LEFT;
			case TOP:
				return AbstractChartAxisAdv.Position.TOP;
			case RIGHT:
				return AbstractChartAxisAdv.Position.RIGHT;
			case BOTTOM:
			default:
				return AbstractChartAxisAdv.Position.BOTTOM;
		}
	}

	public static AxisTickLabelPosition toPoiAxisTickLabelPosition(AbstractChartAxisAdv.TickLabelPosition position) {
		switch (position) {
			case LOW:
				return AxisTickLabelPosition.LOW;
			case NONE:
				return AxisTickLabelPosition.NONE;
			case HIGH:
				return AxisTickLabelPosition.HIGH;
			case NEXT_TO:
			default:
				return AxisTickLabelPosition.NEXT_TO;
		}
	}

	public static AbstractChartAxisAdv.TickLabelPosition toAxisTickLabelPosition(
			AxisTickLabelPosition tickLabelPosition) {
		switch (tickLabelPosition) {
			case LOW:
				return AbstractChartAxisAdv.TickLabelPosition.LOW;
			case NONE:
				return AbstractChartAxisAdv.TickLabelPosition.NONE;
			case HIGH:
				return AbstractChartAxisAdv.TickLabelPosition.HIGH;
			case NEXT_TO:
			default:
				return AbstractChartAxisAdv.TickLabelPosition.NEXT_TO;
		}
	}

	public static AxisLabelAlignment toPoiAxisLabelAlign(CategoryAxisImpl.LabelAlign labelAlign) {
		switch (labelAlign) {
			case LEFT:
				return AxisLabelAlignment.LEFT;
			case RIGHT:
				return AxisLabelAlignment.RIGHT;
			case CENTER:
			default:
				return AxisLabelAlignment.CENTER;
		}
	}

	public static CategoryAxisImpl.LabelAlign toAxisLabelAlign(AxisLabelAlignment labelAlign) {
		switch (labelAlign) {
			case LEFT:
				return CategoryAxisImpl.LabelAlign.LEFT;
			case RIGHT:
				return CategoryAxisImpl.LabelAlign.RIGHT;
			case CENTER:
			default:
				return CategoryAxisImpl.LabelAlign.CENTER;
		}
	}

	// KEIKAI-405
	public static TimeUnit toPoiTimeUnit(DateAxisImpl.TimeUnit timeUnit) {
		switch (timeUnit) {
			case MONTHS:
				return TimeUnit.MONTHS;
			case YEARS:
				return TimeUnit.YEARS;
			case DAYS:
			default:
				return TimeUnit.DAYS;
		}
	}

	public static AxisCrossBetween toPoiAxisBetween(ValueAxisImpl.CrossBetween crossBetween) {
		switch (crossBetween) {
			case MIDPOINT_CATEGORY:
				return AxisCrossBetween.MIDPOINT_CATEGORY;
			case BETWEEN:
			default:
				return AxisCrossBetween.BETWEEN;
		}
	}

	public static ValueAxisImpl.CrossBetween toAxisBetween(AxisCrossBetween crossBetween) {
		switch (crossBetween) {
			case MIDPOINT_CATEGORY:
				return ValueAxisImpl.CrossBetween.MIDPOINT_CATEGORY;
			case BETWEEN:
			default:
				return ValueAxisImpl.CrossBetween.BETWEEN;
		}
	}

	public static int toPoiOperatorType(OperatorType type){
		switch (type) {
			case NOT_EQUAL:
				return DataValidationConstraint.OperatorType.NOT_EQUAL;
			case NOT_BETWEEN:	
				return DataValidationConstraint.OperatorType.NOT_BETWEEN;
			case LESS_THAN:
				return DataValidationConstraint.OperatorType.LESS_THAN;
			case LESS_OR_EQUAL:
				return DataValidationConstraint.OperatorType.LESS_OR_EQUAL;
			case GREATER_THAN:
				return DataValidationConstraint.OperatorType.GREATER_THAN;
			case GREATER_OR_EQUAL:
				return DataValidationConstraint.OperatorType.GREATER_OR_EQUAL;
			case EQUAL:
				return DataValidationConstraint.OperatorType.EQUAL;
			case BETWEEN:
			default:
				return DataValidationConstraint.OperatorType.BETWEEN;
		}
	}

	public static OperatorType toOperatorType(int poiOperator){
		switch(poiOperator){
			case DataValidationConstraint.OperatorType.EQUAL:
				return OperatorType.EQUAL;
			case DataValidationConstraint.OperatorType.GREATER_OR_EQUAL:
				return OperatorType.GREATER_OR_EQUAL;
			case DataValidationConstraint.OperatorType.GREATER_THAN:
				return OperatorType.GREATER_THAN;
			case DataValidationConstraint.OperatorType.LESS_OR_EQUAL:
				return OperatorType.LESS_OR_EQUAL;
			case DataValidationConstraint.OperatorType.LESS_THAN:
				return OperatorType.LESS_THAN;
			case DataValidationConstraint.OperatorType.NOT_BETWEEN:
				return OperatorType.NOT_BETWEEN;
			case DataValidationConstraint.OperatorType.NOT_EQUAL:
				return OperatorType.NOT_EQUAL;
			case DataValidationConstraint.OperatorType.BETWEEN:
			default:
				return OperatorType.BETWEEN;
		}
	}

	public static int toPoiErrorStyle(AlertStyle style){
		switch(style){
		case INFO:
			return DataValidation.ErrorStyle.INFO;
		case WARNING:
			return DataValidation.ErrorStyle.WARNING;
		case STOP:
		default:
			return DataValidation.ErrorStyle.STOP;
		}
	}
	
	public static AlertStyle toErrorStyle(int errorStyle){
		switch(errorStyle){
			case DataValidation.ErrorStyle.INFO:
				return AlertStyle.INFO;
			case DataValidation.ErrorStyle.WARNING:
				return AlertStyle.WARNING;
			case DataValidation.ErrorStyle.STOP:
			default:
					return AlertStyle.STOP;
		}
	}


	public static ValidationType toValidationType(int validationType){
		switch(validationType){
			case DataValidationConstraint.ValidationType.TIME:
				return ValidationType.TIME;
			case DataValidationConstraint.ValidationType.TEXT_LENGTH:
				return ValidationType.TEXT_LENGTH;
			case DataValidationConstraint.ValidationType.LIST:
				return ValidationType.LIST;
			case DataValidationConstraint.ValidationType.INTEGER:
				return ValidationType.INTEGER;
			case DataValidationConstraint.ValidationType.FORMULA:
				return ValidationType.CUSTOM;
			case DataValidationConstraint.ValidationType.DECIMAL:
				return ValidationType.DECIMAL;
			case DataValidationConstraint.ValidationType.DATE:
				return ValidationType.DATE;
			case DataValidationConstraint.ValidationType.ANY:
			default:
				return ValidationType.ANY;
		}
	}
	//POI has no way to set validation type

	public static ErrorValue toErrorCode(byte errorCellValue) {
		//ZSS-672
		return ErrorValue.valueOf(errorCellValue);
	}
}
