/* Converter.java

	Purpose:
		
	Description:
		
	History:
		3:00 PM 10/04/2017, Created by jumperchen

Copyright (C) 2017 Potix Corporation. All Rights Reserved.
*/
package io.keikai.util;

import java.util.Arrays;
import java.util.Collections;
import java.util.Deque;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * @author jumperchen
 */
public final class Converter {

	/**
	 * Converts 0-based index to abc.
	 * @param index 0-based.
	 * @return
	 */
	public static String numToAbc(int index) {
		final Deque<String> sb = new LinkedList<>();
		while (index >= 0) {
			final int j = (index % 26) + 65;
			sb.addFirst(String.valueOf((char) j));
			index = index / 26;
			index -= 1;
		}
		return String.join("", sb);
	}

	/**
	 * Converts a range to A1 notation.
	 */
	public static String rangeToA1(int row, int column) {
		return numToAbc(column) + (row + 1);
	}

	public static Ref rangeToRef(int row, int column) {
		return new Ref(column, row);
	}
	public static Ref areaToRef(int row, int column, int lastRow, int lastColumn) {
		return new Ref(column, row, lastColumn, lastRow);
	}
	public static Ref a1ToRef(String a1notation) {
		final String[] a1s = a1notation.split(":");
		if (a1s.length == 0) {
			return null;
		}
		final int len = a1s.length;
		for (int j = 0; j < len; j += 1) {
			// check if current ref is a cell
			// is a cell -> check if next ref is also a cell
			// >> next ref is not a cell -> parse current ref as a single cell
			// >> next ref is a cell -> parse current and next refs as a range
			// is not a cell -> must be a range
			// >> parse current and next refs as a range

			final List cr1 = splitA1(a1s[j]);
			if (cr1 == null) {
				return null;
			}
			final Integer left = (Integer) cr1.get(0);
			final Integer top = (Integer) cr1.get(1);

			if (left != null && top != null) { // current ref is a cell, check next ref
				if (j + 1 < len) {
					final List cr2 = splitA1(a1s[j + 1]);
					if (cr2 == null) {
						return null;
					}
					final Integer right = (Integer) cr2.get(0);
					final Integer bottom = (Integer) cr2.get(1);

					if (right != null && bottom != null) { // next is also a cell, current and next form a range
						return left == right && top == bottom ?
								new Ref(left, top) :
								new Ref(left, top, right, bottom);
					} else { // next is not a cell, current must be a single cell
						return new Ref(left, top);
					}
				} else { // current is the last ref, must be a single cell
					return new Ref(left, top);
				}
			} else { // current is not a cell, must form a col/row with next ref
				final List cr2 = len <= j+1 ? cr1 : splitA1(a1s[j + 1]);
				if (cr2 == null) {
					return null;
				}
				final Integer right = (Integer) cr2.get(0);
				final Integer bottom = (Integer) cr2.get(1);

				// cannot both be null, not a valid ref
				if (left == null && top == null)
					return null;
				// cannot both be null or not null, we're looking to build a col/row here
				if ((right == null && bottom == null) || (right != null
						&& bottom != null))
					return null;

				if (left == null) { // must be a row
					if (right == null) {
						return new Ref(0, top, Ref.MAX_COL - 1, bottom);
					} else {
						return null;
					}
				} else { // must be a col
					if (bottom == null) {
						return new Ref(left, 0, right, Ref.MAX_ROW - 1);
					} else {
						return null;
					}
				}
			}
		}
		return null;
	}

	private static boolean isAToZ(int code) {
		return 65 <= code && code <= 90;
	}

	private static boolean is0To9(int code) {
		return 48 <= code && code <= 57;
	}

	private static List splitA1(String a1) {
		a1 = a1.trim().toUpperCase();
		boolean absc = a1.startsWith("$");
		boolean absr = false;
		if (absc) {
			a1 = a1.substring(1);
  		}
		final int len = a1.length();
		int s = -1;
		int m = -1;
		for (int j = 0; j < len; ++j) {
			final int code = a1.charAt(j);
			if (isAToZ(code)) {
				if (m >= 0) {
					return null;
				}
				if (s < 0) {
					s = j;
				}
			} else if (is0To9(code) || (m < 0 && code == 36)) { // 0 ~ 9 or $
				if (m < 0) {
					m = j;
				}
			} else {
				return null;
			}
		}
		Integer n = null;
		Integer a = null;
		if (s < 0) {
			if (m >= 0) {
				// s < 0, m >=0; 1 only
				n = Integer.parseInt(a1) - 1;
				absr = absc;
				absc = false;
			}
		} else if (m < 0) {
			// s >= 0, m < 0; a only
			a = abcToNum(a1);
		} else {
			// s >= 0, m >=0; a1
			String r = a1.substring(m);
			absr = r.startsWith("$");
			if (absr) {
				r = r.substring(1);
			}
			if (r.isEmpty()) {
				return null;
			}
			n = Integer.parseInt(r) - 1;
			a = abcToNum(a1.substring(s, m));
		}
		if (n != null && (n < 0 || n >= Ref.MAX_ROW)) {
			return null;
		}
		if (a != null && a >= Ref.MAX_COL) {
			return null;
		}
		return Arrays.asList(a, n, absc, absr);
	}

	/**
	 * Converts abc to number. (like A to 0, B to 1)
	 * @param val
	 * @return
	 */
	public static int abcToNum(String val) {
		int index = 0;
		final int len = val.length();
		for (int j = 0; j < len; j += 1) {
			final int code = val.charAt(j);
			if (code < 65 || code > 90) {
				throw new IllegalArgumentException("The char code is ouf of range [" + code + "]");
			}
			final int digit = code - 64;
			index = index * 26 + digit;
		}
		return index - 1;
	}

	public static List<Ref> toRefs(String refs) {
		return toRefs(refs, " ");
	}

	public static List<Ref> toRefs(String refs, String separator) {
		return Stream.of(refs.split(separator)).map((s) -> a1ToRef(s)).collect(
				Collectors.toList());
	}

	public static String refToA1(Ref ref) {
		if (ref.isCell()) {
			// cell a1
			return numToAbc(ref.getLeft()) + (ref.getTop() + 1);
		} else {
			return numToAbc(ref.getLeft()) + (ref.getTop() + 1) + ":" +
				numToAbc(ref.getRight()) + (ref.getBottom() + 1);
		}
	}
	public static final <T> List<T> asList(T... ary) {
		if (ary == null || ary.length == 0)
			return Collections.EMPTY_LIST;
		return Stream.of(ary).collect(Collectors.toList());
	}

	public static Map asMap(Object... args) {
		Map<Object, Object> map = new HashMap<>(args.length / 2);
		for (int i = 0; i < args.length; i+= 2) {
			map.put(args[i], args[i+1]);
		}
		return map;
	}
	public static int ptToPx(Number pt) {
		return (int) (pt.doubleValue() * 4 / 3);
	}
	public static <T> T getFirst(T... args) {
		if (args == null || args.length == 0) return null;
		return args[0];
	}
	public static <T> T getLast(T... args) {
		if (args == null || args.length == 0) return null;
		return args[args.length - 1];
	}
}
