package io.keikai.util;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.regex.Pattern;

//Note: we use characters a-z, A-Z, 0-9 and _, s.t., the user can select all
//by double-clicking it (they also valid characters no need of escapes).
//So, it is 26 * 2 + 10 + _CC_EXTRA => 65 diff chars
//
//OID is 24 chars = 65^24 = 3.2e43
//so it is about 2.8e6 more than 128 bit UUID (where 122 is effective: 5.3e36)
//(note: Git revision is 36^40 about 1.79e62)

///**
// * The type of random generator
// * <p>
// * * [length] specifies the number of integers to generate.
// */
//typedef List<int> GetRandomInts(int length);

public interface Oid {
    ///Total number of characters per OID.
    int OID_LENGTH = 24;

    int[] _CC_EXTRA = {
            45/*-*/, 46/*.*/, 95/*_*/
    }; //33/*!*/, 40/*(*/, 41/*)*/ => not valid in email
//42/***/, => not safe
//126/*~*/, => conservative (https://www.cs.tut.fi/~jkorpela/tilde.html)


    ///The character range
    int _CC_RANGE = 65, //26*2+10+_CC_EXTRA
            _CC_0 = 48, //_CC_9 = _CC_0 + 9,
            _CC_A = 65, _CC_a = 97;
    int _INT_LEN = 5, //# of integers: _INT_LEN * _CHAR_PER_INT >= OID_LENGTH - 1 + 2
            _CHAR_PER_INT = 5; //65^5 < 2^31 (65^5: 1,160,290,625, 2^31: 2,147,483,648)
    Pattern _oidPattern = Pattern.compile("^[-0-9a-zA-Z._]*$");
    Random _random = new Random();

    /**
     * Returns the next unique object ID.
     */
    static String nextOid() {
        final int[] values = getRandomInts(_INT_LEN);
        assert (values.length == _INT_LEN);
        final List<Integer> bytes = new ArrayList<>();
        lGen:
        for (int i = values.length; --i >= 0; ) {
            int val = values[i];
            if (val < 0)
                val = -val;

            for (int j = _CHAR_PER_INT; ; ) {
                bytes.add(_escOid(val % _CC_RANGE));
                if (bytes.size() >= OID_LENGTH)
                    break lGen;

                if (--j == 0)
                    break;
                val = (int) Math.ceil((double) val / _CC_RANGE);
            }
        }
        final int[] arr = bytes.stream().mapToInt(i -> i).toArray();
        return new String(arr, 0, arr.length);
    }

    /**
     * Creates a new OID based two OIDs.
     * <p>
     * > To shorten the result OID, we retrieve the substring of [oid1] and [oid2]
     * and concatenate them together. Of course, there might be conflict but
     * the chance is so low that we can ignore (like OID generator),
     */
    static String mergeOid(String oid1, String oid2) {
        return oid1.substring(0, 12) + oid2.substring(0, 12);
    }

    ///Test if the given value is a valid OID.
    ///
    ///Note: for performance reason, it does only the basic check.
    static boolean isValidOid(String value) {
        return value.length() == OID_LENGTH && _oidPattern.matcher(value).find();
    }

    /**
     * The function used to generate a list of random integers to construct OID.
     * <p>
     * The default implementation uses [Random] to generate the random number.
     * When running at the browser, it is better to replace with
     * `Crypto.getRandomValues`.
     */
    ///Default implementation of [getRandomInts]
    static int[] getRandomInts(int length) {
        final int[] values = new int[length];
        for (int i = length; --i >= 0; )
            values[i] = _random.nextInt(0x7fffffff);
        return values;
    }

    static int _escOid(int v) {
        if (v < 10)
            return _CC_0 + v;
        if ((v -= 10) < 26)
            return _CC_A + v;
        if ((v -= 26) < 26)
            return _CC_a + v;
        return _CC_EXTRA[v - 26];
    }
}
