/* XlsxJsonHelper.java

	Purpose:
		
	Description:
		
	History:
		11:22 AM 2019/9/26, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.util;

import static io.keikai.util.Converter.asList;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * Xlsx Json Helper
 * @author jumperchen
 */
public class XlsxJsonHelper {
	private XlsxJsonHelper() {}


	// fix <code>$(foo, $(bar, baz))</code> wrong casting issue
	public static <T> T $1(Map map, String path) {
		return (T) map.get(path);
	}

	public static <T> T $(Map map, String... paths) {
		if (map == null) return null;
		Map nested = map;
		for (int i = 0, j = paths.length; i < j; i++) {
			Object r = nested.get(paths[i]);
			if (i + 1 == j) {
				return (T) r;
			} else if (r instanceof Map) {
				nested = (Map) r;
			} else {
				break;
			}
		}
		return null;
	}

	// query with the xpath-like, i.e. $x(map, "foo/bar/baz")
	public static <T> T $x(Map map, String xpath) {
		return $(map, xpath.split("/"));
	}

	public static Boolean toBool(Map jsonMap, String key) {
		return toBoolean(jsonMap.get(key), null);
	}

	public static Boolean toBool(Map jsonMap, String key, Boolean nullAs) {
		return toBoolean(jsonMap.get(key), nullAs);
	}

	public static Boolean toBoolean(Object val, Boolean nullAs) {
		if (val == null) {
			return nullAs;
		}
		if (val instanceof Boolean) {
			return (Boolean) val;
		}
		if (val instanceof String) {
			final String vall = val.toString().toLowerCase();
			return !vall.equals("0") && !vall.equals("false") && !vall
					.equals("no") && !vall.equals("off") && !vall.isEmpty();
		}
		if (val instanceof Number) {
			final double dv = ((Number) val).doubleValue();
			return Double.isNaN(dv) && dv != 0;
		}
		return false;
	}

	public static Double toDouble(Map jsonMap, String key) {
		return toDouble(jsonMap, key, null);
	}

	public static Double toDouble(Map jsonMap, String key, Double nullAs) {
		if (jsonMap == null)
			return null;
		Object v = getValue(jsonMap, key);
		if (v instanceof String)
			return Double.parseDouble((String) v);
		return nullAs;
	}

	public static Integer toInt(Map jsonMap, String key) {
		return toInt(jsonMap, key, null);
	}

	public static Integer toInt(Map jsonMap, String key, Integer nullAs) {
		if (jsonMap == null)
			return null;
		Object v = getValue(jsonMap, key);
		if (v instanceof String)
			return Integer.parseInt((String) v);
		return nullAs;
	}

	public static Long toLong(Map jsonMap, String key) {
		return toLong(jsonMap, key, null);
	}

	public static Long toLong(Map jsonMap, String key, Long nullAs) {
		if (jsonMap == null)
			return null;
		Object v = getValue(jsonMap, key);
		if (v instanceof String)
			return Long.parseLong((String) v);
		return nullAs;
	}

	public static Number toNum(Map jsonMap, String key) {
		if (jsonMap == null)
			return null;
		Object v = getValue(jsonMap, key);
		if (v instanceof String) {
			String val = (String) v;
			if (val.contains(".")) {
				return Double.parseDouble(val);
			} else {
				return Integer.parseInt(val);
			}
		} else if (v instanceof Number) {
			return (Number) v;
		}
		return null;
	}

	public static Float toFloat(Map jsonMap, String key) {
		if (jsonMap == null)
			return null;
		Object v = getValue(jsonMap, key);
		return v instanceof String ? Float.parseFloat((String) v) : null;
	}

	public static List toList(Object jsonMap, String key) {
		if (!(jsonMap instanceof Map))
			return null;
		Map jsonMap0 = (Map) jsonMap;
		Object v = jsonMap0.get(key);
		if (v instanceof List)
			return (List) v;
		else if (v == null) {
			return new ArrayList();
		} else {
			return asList(v);
		}
	}

	public static <T> T toValue(Map jsonMap, String key) {
		if (jsonMap == null)
			return null;
		return (T) getValue(jsonMap, key);
	}

	public static String toStringValue(Map jsonMap, String key) {
		return toStringValue(jsonMap, key, null);
	}

	public static String toStringValue(Map jsonMap, String key,
			String nullAs) {
		if (jsonMap == null)
			return null;
		Object r = getValue(jsonMap, key);
		if (r instanceof String)
			return (String) r;
		if (r != null)
			return r.toString();
		return nullAs;
	}
	public static Object getValue(Map jsonMap, String key) {
		if (jsonMap.containsKey(key)) {
			Object k = jsonMap.get(key);
			if (k instanceof Map) {
				Map kM = (Map) k;
				if (kM.containsKey("_")) { // text node
					return (String) kM.get("_");
				} else if (kM.containsKey("val")) {
					return (String) kM.get("val");
				} else {
					return kM;
				}
			} else {
				return k;
			}
		}
		return null;
	}

	public static <T> List<T> toList(Map jsonMap, String key, Class cls) {
		if (jsonMap == null || jsonMap.isEmpty()) return Collections.EMPTY_LIST;

		final Object value = jsonMap.get(key);
		if (value != null) {
			try {
				Constructor declaredConstructor = cls
						.getDeclaredConstructor(Map.class);
				if (value instanceof List) {
					return (List<T>) ((List<Map>) value).stream().map((c) -> {
						try {
							return declaredConstructor.newInstance(((c)));
						} catch (Exception e) {
							e.printStackTrace();
						}
						return null;
					}).collect(Collectors.toList());
				} else if (value instanceof Map) {
					return (List<T>) asList(
							declaredConstructor.newInstance((Map) (value)));
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return Collections.EMPTY_LIST;
	}
	//20141004, henrichen use hssf/usermodel/HeaderFooter.java solution, it is handles those case without &L, &C, &R
	public static String[] getParts(String text) {
		// default values
		String _left = "";
		String _center = "";
		String _right = "";

		if (text == null || text.length() == 0) {
			return new String[] { _left, _center, _right, };
		}

		outer:
		while (text.length() > 1) {
			//ZSS-1091
			char part = 'C';
			int start = 0;
			if (text.charAt(0) == '&') {
				part = text.charAt(1);
				start = 2;
			}
			int pos1 = 0;
			int pos = text.length();
			switch (part) {
			case 'L':
				pos1 = indexOf(text, "&L");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				pos1 = indexOf(text, "&C");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				pos1 = indexOf(text, "&R");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				_left = _left + text.substring(start, pos);
				text = text.substring(pos);
				break;
			case 'C':
				pos1 = indexOf(text, "&L");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				pos1 = indexOf(text, "&C");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				pos1 = indexOf(text, "&R");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				_center = _center + text.substring(start, pos);
				text = text.substring(pos);
				break;
			case 'R':
				pos1 = indexOf(text, "&L");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				pos1 = indexOf(text, "&C");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				pos1 = indexOf(text, "&R");
				if (pos1 >= 0) {
					pos = Math.min(pos, pos1);
				}
				_right = _right + text.substring(start, pos);
				text = text.substring(pos);
				break;
			default:
				// Mimics the behaviour of Excel, which would put it in the center.
				_center = _center + text.substring(start);
				break outer;
			}
		}
		return new String[] { _left, _center, _right, };
	}
	//20141003, henrichen: consider &&C which is not a valid &C
	private static int indexOf(String text, String token) {
		int start = 1;
		while (true) {
			start = text.indexOf(token, start);
			if (start >= 0) {
				int count = 1;
				for (int j = start; --j >= 0;) {
					if (text.charAt(j) != '&') break;
					++count;
				}
				if ((count & 1) != 0) { // ODD number of continuous '&', legal
					return start;
				} else {
					++start;
					continue;
				}
			}
			return start;
		}
	}
}
