/* AreaCtrl.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/

(function () {
	/**
	 * @param zss.AreaCtrl
	 * @param boolean
	 */
	function _showSelectionArea (obj) {
		var range = obj.lastRange;
		if (!range) return;

		const cell = obj.sheet.getCell(range.top, range.left);
		const l = range.left;
		const t = range.top;
		const r = range.right;
		const b = range.bottom;
		let isOneCell;
		// KEIKAI-279
		if (cell != null && cell.merid) {
			isOneCell = t === cell.mert && l === cell.merl &&
					b === cell.merb && r === cell.merr;
		} else {
			isOneCell = t === b && l === r;
		}
		jq(obj.comp).css('display', 'block');
		jq(obj.icomp)[isOneCell ? 'removeClass' : 'addClass']('zsselecti-r');
	}

/**
 * Area controller
 * 
 */
zss.AreaCtrl = zk.$extends(zk.Object, {
	$init: function (sheet, cmp, range, mode) {
		this.$supers('$init', arguments);
		this.id = cmp.id;
		cmp.ctrl = this;
		this.comp = cmp;
		this.icomp = cmp.firstChild;
		this.sheet = sheet;
		if (Array.isArray(range)) {
			range = range[0];
		}
		this.lastRange = range ? range : new zss.Range(0,0,0,0);
		this.mode = mode;//null,inner,outer
	},
	cleanup: function () {
		this.invalid = true;
		if (this.comp) this.comp.ctrl = null;
		this.comp = this.icomp = this.sheet = null;
	},
	/**
	 * Locate selection area
	 * @param zss.Range range of the selection
	 */
	relocate: function (range) {
		var sheet = this.sheet;
		if (!range) {
			range = this.lastRange;
			if (!range) return;
		} else
			this.lastRange = range;

		var custColWidth = sheet.custColWidth,
			custRowHeight = sheet.custRowHeight,
			l = custColWidth.getStartPixel(range.left),
			t = custRowHeight.getStartPixel(range.top),
			w = custColWidth.getStartPixel(range.right + 1) - l,
			h = custRowHeight.getStartPixel(range.bottom + 1) - t;
		
		this.relocate_(l, t, w, h);
	},
	relocate_: function(l, t, w, h) {
		const sheet = this.sheet;
		
		let l0 = l + sheet.leftWidth; //adjust to block position.
//		t += sheet.topHeight + 1;//adjust to block position. //ZSS-948
		let t0 = t + sheet.topHeight;//adjust to block position. // KEIKAI-249
		let w0 = w;
		let h0 = h;
		if (this.mode == "inner") {
			l0 -= 1;
			t0 -= 1;
			w0 += 1;
			h0 += 1;
		} else {
			l0 -= 2;
			t0 -= 2;
			w0 += 3;
			h0 += 3;
		}

		jq(this.comp).css({'width': jq.px0(w0), 'height': jq.px0(h0), 'left': jq.px(l0), 'top': jq.px(t0)});
	},
	isVisible: function () {
		return this._visible === true;
	},
	/**
	 * Display selection area
	 */
	showArea: function () {
		var range = this.lastRange;
		if (!range) return;
		if (this._visible) {
			return;
		}
		jq(this.comp).css('display', 'block');
		this._visible = true;
	},
	/**
	 * Hide selection area 
	 */
	hideArea: function () {
		if (!this._visible) {
			return;
		}
		jq(this.comp).css('display', 'none');
		this._visible = false;
	}
});

/**
 * Selection area controller
 * 
 * <P> This controls the spreadsheet's selection area. When user focus on
 * a cell or select multiple cells, shows the selection area.
 */
zss.SelAreaCtrl = zk.$extends(zss.AreaCtrl, {
	// KEIKAI-429: multiple selection range
	ranges: null, // [{range: zss.Range, comp: selection div}]
	//override
	extending: false,
	$init: function (sheet, cmp, range, mode) {
		this.$supers('$init', arguments);
		var wgt = this.sheet._wgt,
			n = this.comp;
		wgt.domListen_(n, 'onMouseMove', '_doSelAreaMouseMove');
		wgt.domListen_(n, 'onMouseOut', '_doSelAreaMouseOut');
		this.ranges = [];
	},
	cleanup: function () {
		this.ranges = [];
		var wgt = this.sheet._wgt,
			n = this.comp;
		wgt.domUnlisten_(n, 'onMouseMove', '_doSelAreaMouseMove');
		wgt.domUnlisten_(n, 'onMouseOut', '_doSelAreaMouseOut');
		this.$supers('cleanup', arguments);
	},
	/**
	 * Locate selection area
	 * @param zss.Range range of the selection
	 * @param boolean multiple range
	 */
	relocate: function (range, multiple) {
		var lastRange = this.lastRange,
			comp = this.comp,
			compStyle = comp.style,
			lastRangeDimension = {'width': compStyle.width, 'height': compStyle.height, 'left': compStyle.left, 'top': compStyle.top};
		this.$supers('relocate', arguments);
		if (!this.lastRange) {
			return;
		}
		var ranges = this.ranges;
		if (!multiple) { // clear first
			for (var i = 0, l = this.ranges.length; i < l; i++) {
				jq(ranges.pop().comp).detach();
			}
			jq(comp).find('.zsseldot').show();
		} else {
			if (!this.extending) {
				if (lastRange) {
					var rangeComp = jq('<div/>', {class: comp.className}); // copy last range
					lastRangeDimension['display'] = 'block';
					rangeComp.css(lastRangeDimension);
					jq(comp.parentNode).append(rangeComp);
					this.ranges.push({range: lastRange, comp: rangeComp[0]});
					jq(comp).find('.zsseldot').hide();
				}
			}
		}
	},
	showArea: function () {
		_showSelectionArea(this);
	},
	_doMouseMove: function (evt) {
		var sheet = this.sheet,
			mx = evt.pageX,
			my = evt.pageY,
			cellpos = zss.SSheetCtrl._calCellPos(sheet, mx, my, false),
			row = cellpos[0],
			col = cellpos[1],
			cell = sheet.getCell(row, col);
		//ZSS-454 Cannot click on hyperlink in the merge cell.
		//ZSS-1395 when moving a cursor, cell might be null for a sheet might not receive data yet
		if (cell) {
            cell = sheet.getMergerIfAny(cell);

			//ZSS-864
			if (sheet._showCellComment)
				sheet._showCellComment(cell);
			if (this._isStartHyperlink()) return; //the hyperlink followup is already started, so don't worry about the cursor
			var jqa = jq(cell.comp).find('a');
			if (jqa.length > 0) {
				var aelm = jqa[0];
				if (zkS.isOverlapByPoint(aelm, mx, my)) {
					this._setInnerCursor('pointer');
					this._setHyperlinkElment(aelm);
					return;
				}
			}
			this._resetHyperlink();
		}
	},
	_doMouseOut: function (evt) {
		this._resetHyperlink();
		
		//ZSS-864
		var sheet = this.sheet,
			mx = evt.pageX,
			my = evt.pageY,
			cellpos = zss.SSheetCtrl._calCellPos(sheet, mx, my, false),
			row = cellpos[0],
			col = cellpos[1],
			cell = sheet.getMergerIfAny(sheet.getCell(row, col));
		if (cell && sheet._showCellComment) {
			sheet._showCellComment(cell);
		}
	},
	_setInnerCursor: function (cursor) {
		// KEIKAI-200
		if (this.innerCursor === cursor) {
			return;
		}
		this.innerCursor = cursor;
		jq(this.icomp).css('cursor', cursor);
	},
	_setAnchorCursor: function (cursor) {
		// KEIKAI-200
		if (this.anchorCursor === cursor) {
			return;
		}
		this.anchorCursor = cursor;
		jq(this._aelm).css('cursor', cursor);
	},
	_setHyperlinkElment: function (elm) {
		this._hyperlinkElm = elm;
	},
	_startHyperlink: function(aelm) { //when mouse down
		this._stopHyperlink(); //just in case event is not properly fired ...
		//stupid ie8 will call _doMouseOut, and _doMouseMove between _startHyperlink and _tryAndEndHyperlink.
		//and #_hyperlinkElm will lost in #_resetHyperlink
		if(zk.ie8 && zk.ie < 11) this._start = true;
		var t = 1000;
		this._mdtimeout = (new Date()).getTime() + t;
		if((aelm && aelm.tagName.toLowerCase() == 'a') || this._hyperlinkElm) {
			var selArea = this, //delay 1 second
				anchor = aelm;
			setTimeout(function() {selArea._resetHyperlinkCursor(anchor)}, t);
		}
	},
	_tryAndEndHyperlink: function (row, col, evt) { //when mouse click
		var elm = this._hyperlinkElm;
		if (elm && elm.tagName.toLowerCase() == 'a') {
			if (!this._isHyperlinkTimeout(new Date().getTime())) {//not time out yet
				var href = elm.getAttribute('z.href'),
					type = elm.getAttribute('z.t');
				type = parseInt(type);//it is string in attribute
				this.sheet._wgt.linkTo(href, type, evt);
				
				//ZSS-807: "place in document" might link to other sheet; snapshot before leave
				if (type == 2) {
					var wgt = this.sheet._wgt,
						cacheCtrl = wgt._cacheCtrl,
						currSheetId = wgt.getSheetId();
					cacheCtrl.snap(currSheetId); //snapshot current sheet status
				}

				this.sheet._sendOnCellHyperlink(row, col, href, type, evt);
			}
		}
		//end of a hyperlink followup
		this._stopHyperlink();
	},
	_stopHyperlink: function () {
		//stupid ie8 will call _doMouseOut, and _doMouseMove between _startHyperlink and _tryAndEndHyperlink.
		//and #_hyperlinkElm will lost in #_resetHyperlink
		if(zk.ie8 && zk.ie < 11) delete this._start;
		delete this._mdtimeout;
		if (this._aelm) {
		    this._setAnchorCursor('');
			delete this._aelm;
		}
	},
	_isStartHyperlink: function() {
		return this._mdtimeout;
	},
	_isHyperlinkTimeout: function(t) {
		return t >= this._mdtimeout;
	},
	_resetHyperlinkCursor: function (aelm) {
		if (!aelm) {
			this._setInnerCursor('default');
		} else {
			this._setAnchorCursor('default');
		}
	},
	_resetHyperlink: function () {
		//stupid ie8 will call _doMouseOut, and _doMouseMove between _startHyperlink and _tryAndEndHyperlink.
		//and #_hyperlinkElm will lost in #_resetHyperlink. Must work around it
		if ((!zk.ie8 || zk.ie >= 11)|| !this._start) {
			delete this._hyperlinkElm;
			this._resetHyperlinkCursor();
		}
	}
});

/**
 * Selection changing area controller
 * 
 * <p> This controls the spreadsheet's selection change area. When user drag
 * current selection area to change selection area size, shows selection changing 
 * area. 
 */
zss.SelChgCtrl = zk.$extends(zss.AreaCtrl, {});

/**
 * Area controller at corner panel
 * 
 */
zss.AreaCtrlCorner = zk.$extends(zss.AreaCtrl, {});

/**
 * Selection area controller at corner panel
 */
zss.SelAreaCtrlCorner = zk.$extends(zss.AreaCtrlCorner, {
	//override
	showArea: function () {
		_showSelectionArea(this);
	}
});

/**
 * Selection changing area controller at corner panel
 */
zss.SelChgCtrlCorner = zk.$extends(zss.AreaCtrlCorner, {});

/**
 * Area controller at left panel
 */
zss.AreaCtrlLeft = zk.$extends(zss.AreaCtrl, {
	//override
  	relocate_: function (l, t, w, h) {
		var sheet = this.sheet;
		l += sheet.leftWidth - 1;
		
		if (sheet.lp.toppad) {
			t -= sheet.lp.toppad;
		}
		
		l = l - 2;
		t = t - 2;
		w = w + 3;
		h = h + 3;
		if (this.mode == "inner") {
			l = l + 1;
			t = t + 1;
			w = w - 2;
			h = h - 2;
		}

		jq(this.comp).css({'width': jq.px0(w), 'height': jq.px0(h), 'left': jq.px(l),'top' : jq.px(t)});
	}
});

/**
 * Selection area controller at left panel
 */
zss.SelAreaCtrlLeft = zk.$extends(zss.AreaCtrlLeft, {
	//override
	showArea: function () {
		_showSelectionArea(this);
	}
});

/**
 * Selection changing controller at left panel
 */
zss.SelChgCtrlLeft = zk.$extends(zss.AreaCtrlLeft, {});

/**
 * Area controller at top panel
 */
zss.AreaCtrlTop = zk.$extends(zss.AreaCtrl, {
	//override
  	relocate_: function(l , t, w, h) {
		var sheet = this.sheet;
		t += sheet.topHeight;//adjust to block position. //ZSS-948

		l = l - 2;
		t = t - 2;
		w = w + 3;
		h = h + 3;
		if (this.mode == "inner") {
			l = l + 1;
			t = t + 1;
			w = w - 2;
			h = h - 2;
		}

		jq(this.comp).css({'width': jq.px0(w), 'height': jq.px0(h), 'left': jq.px(l), 'top': jq.px(t)});
	}
});

/**
 * Selection area controller at top panel
 */
zss.SelAreaCtrlTop = zk.$extends(zss.AreaCtrlTop, {
	//override
	showArea: function () {
		_showSelectionArea(this);
	}
});

/**
 * Selection changing area controller at top panel
 */
zss.SelChgCtrlTop = zk.$extends(zss.AreaCtrlTop, {});
})();