/* CellBlockCtrl.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/

(function () {

/**
 * CellBlockCtrl is used for controlling cells, include load cell, creating cell, merge cell and cell position index
 */
zss.CellBlockCtrl = zk.$extends(zk.Widget, {
	widgetName: 'CellBlock',
	range: null, //zss.Range
	_lastDir: null,
	$o: zk.$void, //owner, fellows relationship no needed
	$init: function (sheet, tRow, lCol, bRow, rCol, data, type) {
		this.$supers(zss.CellBlockCtrl, '$init', []); //DO NOT pass "arguments" or all fields will be copied into this Object. 
		this.sheet = sheet;
		this.type = type;
		this.rows = []; //zss.Row
		this.range = new zss.Range(lCol, tRow, rCol, bRow);
		
		//ZSS-1149
		this._tmpMerges = {};
		this._toReset = {};
		this._toSet = {};
		
		var rows = data.rows,
			block = this;
		for (var r = tRow; r <= bRow; r++) {
			if (sheet.custRowHeight.isHidden(r)) continue;

			var row = new zss.Row(sheet, block, r, data);
			for (var c = lCol;  c <= rCol; c++) {
				if (sheet.custColWidth.isHidden(c)) {
					continue;
				}

				row.appendCell(new zss.Cell(sheet, block, r, c, data));
			}
			this.appendRow(row);
		}
	},
	//override
	setVisible: function (visible) {
		if (this._visible != visible) {
			this._visible = visible;
			var n = this.$n();
			if (n)
				n.style.visibility = visible ? 'visible' : 'hidden';
		}
	},
	redraw: function (out) {
		var rows = this.rows,
			vis = this.isVisible() ? 'visible' : 'hidden',
			size = rows.length;
		out.push('<div id="', this.uuid, '" class="', this.getZclass(), '" style="visibility:' + vis +';">');
		for (var i = 0; i < size; i++) {
			rows[i].redraw(out);
		}
		out.push('</div>');
	},
	//super//
	getZclass: function () {
		var t = this.type;
		return t ? 'zs' + t + 'block' : 'zsblock';
	},
	bind_: function () {
		//ZSS-1117: this._newrange used by Cell.js#bind_()
		this._newrange = this.range; 
		
		this.$supers(zss.CellBlockCtrl, 'bind_', arguments);
		
		var n = this.comp = this.$n();
//		zk(n).disableSelection(); //disableSelection()
		
		//ZSS-1117
		this._setTempMergeCells();
		delete this._newrange;
		this.sheet.addSSInitLater(this.sheet.batchShiftAlignedText.bind(this.sheet));
		this.sheet.listen({onProcessOverflow: this.proxy(this._onProcessOverflow)});
	},
	unbind_: function () {
		this.sheet.unlisten({onProcessOverflow: this.proxy(this._onProcessOverflow)});
		this.$supers(zss.CellBlockCtrl, 'unbind_', arguments);
		
		delete this.rows;
		this.sheet = this.comp = this.rows = this.range = null;
	},
	doMouseDown_: function (evt) {
		this.sheet._doMousedown(evt);
	},
	doMouseUp_: function (evt) {
		this.sheet._doMouseup(evt);
	},
	/**
	 * Returns the row of the spreadsheet
	 * @param int row 
	 */
	getRow: function (row) {
		var range = this.range;
		if(row < range.top || row > range.bottom)
			return null;
		return this.rows[this._getVisibleRowIndex(row - range.top)];
	},
	/**
	 * Returns the cell of the spreadsheet
	 * @param int row
	 * @param int col
	 */
	getCell: function (row, col) {
		var range = this.range;
		if(!range || row < range.top || row > range.bottom || col < range.left || col > range.right)
			return null;

		var sheet = this.sheet;

		if (sheet.custRowHeight.isHidden(row) || sheet.custColWidth.isHidden(col))
			return null;

        // this object doesn't contain cell data of left and top frozen cells
		let row0 = this.rows[this._getVisibleRowIndex(row - range.top)];
		if (!row0) {
			return null;
		}
		return row0.getCellAt(this._getVisibleColIndex(col - range.left));
	},
	/**
	 * Sets the merge range of the cell
	 * @param id
	 * @param int left column start index
	 * @param int top row start index
	 * @param int right column end index
	 * @param int bottom row end index
	 * @param int width
	 */
	addMergeRange: function (id, left, top, right, bottom, width) {
		var cell = ctrl = this.getCell(top, left);
		var comp;
		for (var r = top; r <= bottom; ++r) {
			for (var i = left; i <= right; i++) {
				if (i == left && r == top) {

					// KEIKAI-62: cell may be invisible.
					if (cell) {
						comp = cell.comp;
						jq(comp).addClass("zsmerge" + id);
						cell.mergeCls = "zsmerge" + id;
					}
				} else {
					cell = this.getCell(r, i);
					if (!cell) continue;//in top / left /corner , they might not have such cell
					comp = cell.comp,
					$comp = jq(comp);
					if (r == top) {
						cell.mergeCls = "zsmergee";
						$comp.removeClass("zsmergeeu").addClass("zsmergee");
					} else {
						cell.mergeCls = "zsmergeeu";
						$comp.removeClass("zsmergee").addClass("zsmergeeu");
					}
				}

				if (cell) {
					cell.merid = id;
					cell.merr = right;
					cell.merl = left;
					cell.mert = top;
					cell.merb = bottom;
				}
			}
		}
		//ZSS-944
		if (ctrl) {
			ctrl.redoRotate = true;
		}
		this.update_(top, left, top, left);
	},
	/**
	 * Remove the merge range of the cell
	 * @param id
	 * @param int left column start index
	 * @param int top row start index
	 * @param int right column end index
	 * @param int bottom row end index
	 * @param int width
	 */
	removeMergeRange: function (id, left, top, right, bottom, width) {
		var cell = ctrl = this.sheet.getCellData(top, left, true);
		if (!cell) return;
        var merid = cell.mergeId;
		if (!merid || id != merid) {
		    return;
		}
        var r = cell.merge,
            merl = r.left,
            merr = r.right,
            mert = r.top,
            merb = r.bottom;

		var ud,
			comp;
		for (var r = mert; r <= merb; ++r) {
			for (var i = merl; i <= merr; i++) {
				if (i == merl && r == mert) {
					let merger = this.getCell(r, i);
					if (merger) {
						comp = merger.comp;
						jq(comp).removeClass("zsmerge" + id);
						merger.mergeCls = merger.merid = merger.merr = merger.merl = merger.mert = merger.merb = ud;
					} else {
						let mergee = this.sheet._getVisibleClosestMergeeTargetByIndex(r, i);
						if (mergee) {
                            this.sheet.applyFakeMerger(mergee, id);
						}
					}
				} else {
					cell = this.getCell(r, i);
					if (cell) {
						comp = cell.comp;
						jq(comp).removeClass(r == mert ? "zsmergee" : "zsmergeeu");
						cell.mergeCls = cell.merid = cell.merr = cell.merl = cell.mert = cell.merb = ud;
					}
					if (i === merl && r === this.sheet.frozenRow + 1) {
						this.sheet.getCell(r, i)._resetMergeeuStyle();
					}
				}
			}
		}
		if (ctrl) {
			ctrl.redoOverflow = true;
			//ZSS-944
			ctrl.redoRotate = true;
		}
		this.update_(top, left, top, left);
	},

	/**
	 * Update cells
	 */
	update_: function (tRow, lCol, bRow, rCol, updateSrc) {
		var range = this.range,
			cctrl = this.sheet._wgt._cacheCtrl,
			csrc = cctrl ? cctrl.getSelectedSheet() : null;

		// turn range to index
		let tRow1 = this._getVisibleRowIndex(Math.max(tRow - range.top, 0));
		let bRow1 = this._getVisibleRowIndex(Math.max(bRow - range.top, 0));
		let lCol1 = this._getVisibleColIndex(Math.max(lCol - range.left, 0));
		let rCol1 = this._getVisibleColIndex(Math.max(rCol - range.left, 0));

		if (bRow1 >= this.rows.length) {
			bRow1 = this.rows.length - 1;
		}
		for (var r = tRow1; r <= bRow1; r++) {
			var row = this.rows[r]; // for visible row only.
			if (row) {
				// KEIKAI-62, the rows between tRow1 and bRow1 are not accurate with the given tRow and bRow range.
				// For example, see SSheetCtrl#_setRowHeight, it will create a temporary row
				// when the row is visible from invisible, but the rows between the upper bound of bRow
				// are not in the viewport which are not created yet, so we need to
				// escape if the row.r is greater than the bRow here.
				if (row.r > bRow || row.r < tRow) continue;

				if (updateSrc && csrc) { //ZSS-1181
					row._updateCacheSrc(csrc);
				}
				for (var c = lCol1; c <= rCol1; c++) {
					var cell = row.getCellAt(c);
					if (cell) {
						if (cell.c < lCol || cell.c > rCol) continue; // same as row.r
						cell.update_(updateSrc);
					}
				}
			}
		}
	},
	//ZSS-1181
	_updateRowCacheSrc: function (row) {
		var range = this.range;
		if(!range || row < range.top || row > range.bottom)
			return;
		
		var cctrl = this.sheet._wgt._cacheCtrl,
			csrc = cctrl ? cctrl.getSelectedSheet() : null;
		if (csrc) {
			this.rows[this._getVisibleRowIndex(row - range.top)]._updateCacheSrc(csrc);
		}
	},
	/**
	 * Dynamic create cells
	 * 
	 * @param dir the dir to create cells
	 * @param tRow
	 * @param lCol
	 * @param bRow
	 * @param rCol
	 */
	create_: function (dir, tRow, lCol, bRow, rCol, data) {
		var sheet = this.sheet,
			data = data || sheet._wgt._cacheCtrl.getSelectedSheet(),
			block = this,
			cr = this.range,
			rs = this.rows,
			isNewRow = false,
			isTop = 'north' == dir,
			isBtm = 'south' == dir,
			isJump = 'jump' == dir,
			isVer = isTop || isBtm || isJump,
			isLeft = 'west' == dir,
			isRight = 'east' == dir,
			nr = cr.clone(), //ZSS-1117
			width = rCol - lCol + 1, //ZSS-1117
			//ZSS-1132
			height = bRow - tRow + 1,
			finalNewrange = cr.clone();
		//ZSS-1132
		if (isLeft) {
			finalNewrange.left = lCol;
		} else if (isRight) {
			finalNewrange.right = rCol;
		} else if (isTop) {
			finalNewrange.top = tRow;
		} else if (isBtm) {
			finalNewrange.bottom = bRow;
		}
		this._finalNewrange = finalNewrange;
		
		//ZSS-1117
		if (isLeft)
			nr.left = lCol;
		else if (isRight)
			nr.right = rCol;
		this._newrange = nr; //used by Cell.js#bind_()

		for (var r = tRow, j = 0; r <= bRow; r++) {
			if (sheet.custRowHeight.isHidden(r)) {
				if (isVer) {
					if (isTop) {
						this.range.extendTop(1);
						if (this._newrange) this._newrange.extendTop(1);
					} else {
						this.range.extendBottom(1);
						if (this._newrange) this._newrange.extendBottom(1);
					}
				}
				continue;
			}

			let row = this.getRow(r);

			// force to create
			if (!isVer && !row) {
				isVer = true;
			}

			row = isVer ? new zss.Row(sheet, block, r, data) : row;
			let html = isVer ? row.getHtmlPrologHalf() : '';
			let cells = [];
			for (var c = lCol, i = 0;  c <= rCol; c++) {
				if (sheet.custColWidth.isHidden(c)) {
					// no need to extend the horizontal range, the code above has updated it already.
					continue;
				}

				var cell = new zss.Cell(sheet, block, r, c, data);
				if (isLeft) {
					// KEIKAI-225: batch insert child
					row.insertCell(i++, cell, true); //row has been attach to DOM tree
					html += cell.getHtml();
					cells.push(cell);
				} else if (isRight) {
					// KEIKAI-200: batch append child
					row.appendCell(cell, true); //row has been attach to DOM tree
					html += cell.getHtml();
					cells.push(cell);
				} else if (isVer) {
					row.appendCell(cell, true); //not attach to DOM tree yet
					html += cell.getHtml();
				}
			}
			if (isVer)
				html += row.getHtmlEpilogHalf();
			else if (row._prepareAutoFilterBtns)
				row._prepareAutoFilterBtns(); //ZSS-953
			
			if (isBtm)
				this.appendRow(row, html, true);
			else if (isTop) {
				this.insertRow(j++, row, html, true);
			} else if (isLeft || isRight) {
				const n = row.$n();
				if (n) {
				    isLeft ? jq(n).prepend(html) : jq(n).append(html);
					const dt = this.desktop;
					if (dt) {
						for (let i = 0; i < cells.length; i++) {
							cells[i].bind(dt);
						}
					}
				}
			}
		}
		//ZSS 125: wrap text processed on row.bind_, within appendRow / insertRow
		delete sheet._wrapRange;
		
		//ZSS-1117
		this.range = this._newrange;
		this._resetTempMergeCells(nr.left, nr.top, nr.right, nr.bottom);
		this._setTempMergeCells();
		delete this._newrange;
		
		// ZSS-1116: after Cell.js is bind_() in which it might register to 
		//    do _calcAutoHeight; so try to really render row height now
		if (sheet._cah) {
			sheet.fireProcessAutoHeight_();
			delete sheet._cah;
		}
		delete this._finalNewrange;	// ZSS-1332	
		this.sheet.addSSInitLater(this.sheet.batchShiftAlignedText.bind(this.sheet));
	},
	/**
	 * Sets rows's width position index
	 * @param int col the column
	 * @param int zsw the column's position index
	 */
	appendZSW: function (col, zsw) {
		if (col > this.range.right || col < this.range.left) return;
		rows = this.rows;
		var rowsize = rows.length;
		for (var i = 0; i < rowsize; i++)
			rows[i].appendZSW(this._getVisibleColIndex(col - this.range.left), zsw);
	},
	/**
	 * Sets rows's height position index
	 * @param int row the row
	 * @param int zsh the row's height position index
	 */
	appendZSH: function (row, zsh) {
		if (row > this.range.bottom || row < this.range.top) return;
		this.rows[this._getVisibleRowIndex(row - this.range.top)].appendZSH(zsh);
	},
	/**
	 * Insert new column
	 * @param int col the column to insert
	 * @param int size the size of the cell
	 * @param boolean forShow true if the operation is for show column to use
	 */
	insertNewColumn: function (col, size, forShow) {
		var index = Math.max(this._getVisibleColIndex(col - this.range.left), 0);
		
		rows = this.rows;
		var rowsize = rows.length;
		for (var i = 0; i < rowsize; i++)
			rows[i].insertNewCell(col, size, index, forShow);

		if (!forShow) {
			this.range.extendRight(size);
		} else {
			this.syncDataRange();
		}
	},
	/**
	 * Insert new row
	 * @param int row the row to insert
	 * @param int size the size of the row
	 * @param data
	 * @param boolean forShow true if the operation is for show row to use
	 */
	insertNewRow: function (row, size, data, forShow) {
		var index = Math.max(this._getVisibleRowIndex(row - this.range.top), 0),
			ctrl,
			sheet = this.sheet,
			data = data || sheet._wgt._cacheCtrl.getSelectedSheet(),
			block = this;
		for (var i = 0; i < size; i++) {
			var r = row + i,
				html = '';
			
			ctrl = new zss.Row(sheet, block, r, data);

			//ZSS-120: use default row height first, row's height will update by server if needed
			// ctrl.zsh = null;

			html += ctrl.getHtmlPrologHalf();
			for (var c = this.range.left, cEnd = this.range.right;  c <= cEnd; c++) {
				if (sheet.custColWidth.isHidden(c)) continue;

				var cell = new zss.Cell(sheet, block, r, c, data);
				ctrl.appendCell(cell, true); //not attach to DOM tree yet
				html += cell.getHtml();
			}

			this.insertRow(index + i, ctrl, html); // ZSS-450, ZSS-485: it should be row array index not row number, see also insertNewColumn()
		}

		// KEIKAI-62: don't update the range, if only for show and hide
		if (!forShow) {
			this.shiftRowInfo(index + size, row + size);
			this.range.extendBottom(size);
			if (this._newrange) this._newrange.extendBottom(size); //ZSS-1117
		} else {
			this.syncDataRange();
		}
	},
	syncDataRange: function () {
		let rows = this.rows;
		if (rows.length) {
			this.range.top = rows[0].r;
			this.range.bottom = rows[rows.length - 1].r;
			let cells = rows[0].cells;
			if (cells.length) {
				this.range.left = cells[0].c;
				this.range.right = cells[cells.length - 1].c;
			} else {
				this.range.right = this.range.left;
			}
		} else {
			this.range.bottom = this.range.top;
			this.range.right = this.range.left;
		}
		this.range.reCalcSize();

		if (this._newrange) {
			this._newrange.top = this.range.top;
			this._newrange.bottom = this.range.bottom;
			this._newrange.left = this.range.left;
			this._newrange.right = this.range.right;
			this._newrange.reCalcSize();
		} //ZSS-1117
	},
	/**
	 * Shift row info
	 */
	shiftRowInfo: function(index, newrow, ignoreHidden) {
		var rows = this.rows,
			size = this.rows.length;

		if (index >= size) return;

		let start = rows[index].r,
			hiddenCount = ignoreHidden ? 0 : this.sheet.custRowHeight.getHiddenCount(start > newrow ? newrow : start, start > newrow ? start : newrow),
			delta = newrow - start + hiddenCount;
		for (; index < size; index++) {
			let row = rows[index];
			row.resetRowIndex(row.r + delta);
		}
	},
	/**
	 * Remove column at the given column index
	 * @param int col the column to remove
	 * @param boolean forHidden true if the operation is for hidden column to use
	 */
	removeColumnAt: function (col, forHidden) {
		if (!this.rows.length || col > (this.range.right + 1) || col < this.range.left) return;

		// if forHidden is true, the visible col has already hidden, so the index will minus 1.
		let index = this._getVisibleColIndex(col - this.range.left) + (forHidden ? 1 : 0);
		
		let rows = this.rows,
			rowsize = rows.length;

		for (var i = 0; i < rowsize; i++)
			rows[i].removeCellAt(index, forHidden);

		if (!forHidden) {
			this.range.extendRight(-size);
		} else {
			this.syncDataRange();
		}
	},
	//KEIKAI-117
	/**
	 * Removes columns from lCol to rCol
	 * @param int lCol the left column index
	 * @param int rCol the right column index (inclusive)
	 */
	removeColumnFromRange: function (lCol, rCol) {
		if (!this.rows.length || lCol > (this.range.right + 1) || lCol < this.range.left) return;

		let custSize = this.sheet.custColWidth;

		// find first visible lCol
		for (let start = lCol, end = rCol+1; start < end; start++) {
			if (!custSize.isHidden(start)) {
				lCol = start;
				break;
			}
		}
		let index = this._getVisibleColIndex(lCol - this.range.left);

		if (index >= 0) {
			var rows = this.rows;
			var rowsize = rows.length;
			for (var i = 0; i < rowsize; i++)
				rows[i].removeCellFromRange(index, lCol, rCol);
		}

		this.range.extendRight(-(rCol - lCol + 1));
	},
	//KEIKAI-117
	/**
	 * Remove rows from tRow to bRow
	 * @param tRow the top index
	 * @param bRow the bottom index (inclusive)
	 */
	removeRowFromRange: function (tRow, bRow) {
		if (!this.rows.length || tRow > (this.range.bottom + 1) || tRow < this.range.top) return;

		let custSize = this.sheet.custRowHeight;

		// find first visible tRow
		for (let start = tRow, end = bRow+1; start < end; start++) {
			if (!custSize.isHidden(start)) {
				tRow = start;
				break;
			}
		}
		let index = this._getVisibleRowIndex(tRow - this.range.top),
			size = custSize.getVisibleCount(tRow, bRow);

		let rows = this.rows,
			rem = rows.splice(index, size);

		let ctrl = rem.pop();

		for (; ctrl; ctrl = rem.pop()) {
			ctrl.detach();
		}

		this.range.extendBottom(-size);
		if (this._newrange) this._newrange.extendBottom(-size); //ZSS-1117
		this.shiftRowInfo(index, tRow, true);
	},
	/**
	 * Remove row with given row index
	 * @param int row the row to remove
	 * @param boolean forHidden true if the operation is for hidden row to use
	 */
	removeRowAt: function (row, forHidden) {
		if (!this.rows.length || row > (this.range.bottom + 1) || row < this.range.top) return;

		// if forHidden true, the visible row has already hidden, so the index will minus 1.
		let index = this._getVisibleRowIndex(row - this.range.top) + (forHidden ? 1 : 0);
		let rows = this.rows;

		if (rows.length > index) {
			let rem = rows.splice(index, 1),
				ctrl;

			while(ctrl = rem.pop()) {
				ctrl.detach();
			}

			// KEIKAI-62: don't update the range, if only for show and hide
			if (!forHidden) {
				this.range.extendBottom(-1);
				if (this._newrange) this._newrange.extendBottom(-1); //ZSS-1117
				this.shiftRowInfo(index, row);
			}
		}

		if (forHidden) {
			this.syncDataRange();
		}
	},
	/**
	 * Insert zss.Row widget at start with html content [optional]
	 * 
	 * @param int index
	 * @param zss.Row rowwidget
	 * @param string htmlContent
	 * @param boolean extendRange whether shall extend block range or not
	 */
	insertRow: function (index, row, htmlContent, extendRange) {
		var ignoeChildDom = htmlContent === undefined || !!htmlContent,
			rows = this.rows,
			sibling = rows[index];
		if (!sibling) {
			this.appendRow(row, htmlContent, extendRange);
		} else {
			this.insertBefore(row, sibling, ignoeChildDom);
			rows.splice(index, 0, row);
			if (extendRange) {
				this.range.extendTop(1);
				if (this._newrange) this._newrange.extendTop(1); //ZSS-1117
			}
			if (htmlContent) {
				jq(htmlContent).insertBefore(sibling.$n());
				row.bind();
			}	
		}
	},
	/**
	 * Append zss.Row widget to the end with html content [optional]
	 * 
	 * @param row zss.Row widget
	 * @param string htmlContent
	 * @param boolean extendRange whether shall extend block range or not
	 */
	appendRow: function (row, htmlContent, extendRange) {
		var ignoeChildDom = htmlContent === undefined || !!htmlContent;
		this.appendChild(row, ignoeChildDom);
		this.rows.push(row);
		if (extendRange) {
			this.range.extendBottom(1);
			if (this._newrange) this._newrange.extendBottom(1); //ZSS-1117
		}
		if (htmlContent) {
			jq(this.comp).append(htmlContent);
			row.bind();
		}
	},
	/**
	 * Hide the cell
	 */
	hide: function () {
		jq(this.comp).css('display', 'none');
	},
	/**
	 * Show the cell
	 */
	show : function () {
		jq(this.comp).css('display', 'block');
	},
	removeColumnsFromStart_: function (size) {
		var rows = this.rows,
			rowSize = this.rows.length,
			i = 0;
		for(; i < rowSize; i++) {
			rows[i].removeLeftCell(size);
		}

		this.syncDataRange();
	},
	removeColumnsFromEnd_: function (size) {
		var rows = this.rows,
			rowSize = this.rows.length,
			i = 0;
		for(; i < rowSize; i++) {
			rows[i].removeRightCell(size);
		}

		this.syncDataRange();
	},
	removeRowsFromStart_: function (size) {
		zkS.removeWidgetFromStart(this.rows, size);
		this.syncDataRange();
	},
	removeRowsFromEnd_: function (size) {
		zkS.removeWidgetFromEnd(this.rows, size);
		this.syncDataRange();
	},
    _onProcessOverflow: function (evt) { //ZSS-1338
        const data = evt.data;
        const rows = this.rows;
        if (data) {
            const tRow = data.tRow;
            const bRow = data.bRow;
            const wholeRow = tRow === undefined && bRow === undefined;
            const rCol = data.col;
            for (let i = 0; i <= rows.length; i++) {
                const row = rows[i];// for visible row only.
                if (row) {
                    if (wholeRow || (row.r >= tRow && row.r <= bRow)) {
                        const cells = row.cells;
                        for (let j = 0; j < cells.length; j++) {
                          const cell = row.cells[j];
                          if (cell.overflow) {
                             // KEIKAI-153
                              if (cell.c === rCol - 1) {
                                cell._resetCache();
                                cell.computeTextWidth();
                              }
                              cell._processOverflow(true);
                              cell.visibleTextWidth = null; //enforce re-calculating
                              jq(cell).trigger('waitAlign', [cell]);
                          }
                        }
                    }
                }
            }
        }
    },
	//ZSS-1117, ZSS-1149
	_tmpMerges: null, //r_c -> [l,t,r,b]
	_toReset: null, //r_c -> [l,t,r,b]
	_toSet: null, //r_c -> [l,t,r,b]
	//ZSS-1117: called by Cell.js#bind_() 
	_addTempMerge: function (ml, mt, mr, mb, l, t, r, b, cutw, cuth) {
		var k = ""+mt+"_"+ml,
			m = this._tmpMerges[k];
		if (m) {
			this._toReset[k] = m;
			delete this._tmpMerges[k];
		}
		if (cutw || cuth) {
			this._toSet[k] = [l, t, r, b, cutw, cuth, ml, mt, mr, mb];
		} else { //ZSS-1178
			delete this._toSet[k];
		}
	},
	//ZSS-1117
	_setTempMergeCells: function () {
		var p = this._toSet;
		const keys = p != null ? Object.keys(p) : [];
		for (let i = 0; i < keys.length; ++i) {
			const key = keys[i];
			var m = p[key],
				l = m[0],
				t = m[1],
				r = m[2],
				b = m[3],
				cutw = m[4],
				cuth = m[5],
				ml = m[6],
				mt = m[7],
				mr = m[8],
				mb = m[9],
				cell = this.getCell(t, l);
			if (cell) {
				cell._setTempMergeCellStyle(l, t, r, b, cutw, cuth, ml, mt, mr, mb);
			}
			this._tmpMerges[key] = m;
		}
		this._toSet = {};
		this.sheet.resyncMergerTarget();
	},
	//ZSS-1117
	_resetTempMergeCells: function (rl, rt, rr, rb) {
		var re = this._toReset;
		let keys = re != null ? Object.keys(re) : [];
		for (var k = 0; k < keys.length; k++) {
			let key = keys[k];
			var m = re[key],
				l = m[0],
				t = m[1],
				cell = this.getCell(t, l);
			if (cell) cell._resetTempMergeCellStyle();
		}

		// remove those total out of range
		var rms = [];
		var p = this._tmpMerges;
		keys = p != null ?Object.keys(p) : [];
		for (var k = 0; k < keys.length; k++) {
			let key = keys[k];
			var m = p[key],
				l = m[0],
				t = m[1],
				r = m[2],
				b = m[3];
			if (r < rl || l > rr || b < rt || t > rb) {
				rms.push(key);
			}
		}
		for (var j = rms.length; --j >= 0;) {
			var key = rms[j];
			delete this._tmpMerges[key];
		}
		
		this._toReset = {};
	},
	//KEIKAI-36
	_getVisibleRowIndex: function (rowIndex) {
		return this.sheet.custRowHeight.getVisibleCount(this.range.top, this.range.top + rowIndex) - 1;// count to index
	},
	_getVisibleColIndex: function (colIndex) {
		return this.sheet.custColWidth.getVisibleCount(this.range.left, this.range.left + colIndex) - 1;// count to index
	}
});
})();