/* Group.js

	Purpose:
		
	Description:
		
	History:
		Mon Oct 30 14:13:54 CST 2023, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
(function () {
	// a tree node implementation
	let OutlineNode = zk.$extends(zk.Object, {
			start: -1,
			end: -1,
			parent: null,
			$init: function (level) {
				this.level = level;
			},
			appendChild: function (node) {
				if (!this.children) {
					this.children = [];
				}
				this.children.push(node);
				node.parent = this;
			},
			setIndex: function (index) {
				if (this.start < 0 || this.start > index) {
					this.start = index;
				}
				if (this.end < index) {
					this.end = index;
				}
				if (this.parent) {
					this.parent.setIndex(index);
				}
			},
			toJSON: function () {
				return {
					s: this.start,
					e: this.end,
					l: this.level,
					c: this.children
				};
			}
		}),
		OutlineTree = zk.$extends(zk.Object, {
			$init: function () {
				this.indexToNode = [];
				this.root = new OutlineNode(0);
			},
			getOrCreateNode: function (level, skipClear) {
				if (!level) {
					if (!skipClear) {
						this.indexToNode.length = 0;
					}
					return this.root;
				}
				var node = this.indexToNode[level];
				if (node) {
					// clear the rest.
					if (!skipClear) {
						for (var i = level + 1; i < 8; i++)
							this.indexToNode[i] = undefined;
					}
					return node;
				}
				node = new OutlineNode(level);
				var parent = this.getOrCreateNode(level - 1, true);
				parent.appendChild(node);
				this.indexToNode[level] = node;
				return node;
			},
			toJSON: function () {
				return this.root.children;
			}
		});

	zss.GroupRow = zk.$extends(zk.Widget, {
		_space: 20, // 20px height
		_outlineLevel: 0,
		$init: function (wgt) {
			this._wgt = wgt; //Spreadsheet
		},
		bind_: function () {
			this.$supers('bind_', arguments);
			this.syncOutline();
			this._wgt.listen({
				onHeaderUpdate: this.proxy(this._doHeaderUpdate),
				onActiveRangeUpdate: this.proxy(this._doActiveRangeUpdate)
			});
			this._wgt.sheetCtrl.listen({
				onScrolling: this.proxy(this._doScrolling)
			});
			zWatch.listen({afterSize: this});
		},
		unbind_: function () {
			this._wgt.unlisten({
				onHeaderUpdate: this.proxy(this._doHeaderUpdate),
				onActiveRangeUpdate: this.proxy(this._doActiveRangeUpdate)
			});
			this._wgt.sheetCtrl.unlisten({
				onScrolling: this.proxy(this._doScrolling)
			});
			zWatch.unlisten({afterSize: this});
			this.$supers('unbind_', arguments);
		},
		newButton: function (start, end, collapsed, summaryBelow) {
			let ss = this._wgt,
				pos = this.calculatePosition(start, end + 1),
				div = new zul.wgt.Div({
					top: jq.px0(pos[0]),
					width: this._space / 2 + 'px',
					height: jq.px0(pos[1]),
					sclass: (summaryBelow ? '' : 'k-group-reversed ') + 'k-group-outline' + (collapsed ? '-collapsed' : ''),
					startIndex: start,
					endIndex: end,
				});
			div.appendChild(new zul.wgt.Label({
				sclass: 'k-group-button z-icon-' + (collapsed ? 'plus' : 'minus'),
				listeners: {
					onClick: function () {
						ss.fire('onGroupOpen', {
							sheetId: ss.getSheetId(),
							start: start,
							end: end,
							open: !!collapsed,
							dir: 'row'
						});
					}
				}
			}));
			return div;
		},
		newSpot: function (index) {
			let rowHeight = this._wgt.sheetCtrl.custRowHeight,
				top = rowHeight.getStartPixel(index),
				height = rowHeight.getStartPixel(index + 1) - top,
				div = new zul.wgt.Div({
					top: jq.px0(top),
					width: this._space + 'px',
					height: jq.px0(height),
					sclass: 'k-group-spot',
					startIndex: index,
					endIndex: index + 1
				});
			div.appendChild(new zul.wgt.Label({sclass: 'z-icon-circle'}));
			return div;
		},
		_buildOutlineTree: function () {
			let ss = this._wgt,
				ctrl = ss._cacheCtrl.selected,
				headers = this.isRow_() ? ctrl.rowHeaders : ctrl.columnHeaders,
				keys = Object.keys(headers),
				tree = new OutlineTree();
			for (let i = 0, j = keys.length; i < j; i++) {
				let index = zk.parseInt(keys[i]),
					header = headers[index],
					outlineLevel = header.outlineLevel || 0,
					node = tree.getOrCreateNode(outlineLevel);
				node.setIndex(index);
			}
			return tree;
		},
		clearOutline: function () {
			if (this.isRow_()) {
				this.setWidth('0px');
			} else {
				this.setHeight('0px');
			}
			while (this.firstChild) {
				this.removeChild(this.firstChild);
			}
			this.tree = this.head = this.frozenBody = this.body = null;
			this._outlineLevel = 0;
		},
		initFrame: function () {
			if (this.head) return;
			this.setVflex(true);
			this.setWidth(this._outlineLevel * this._space + 'px');
			this.appendChild(new zul.box.Vlayout({
				hflex: true,
				vflex: true,
				spacing: '0px'
			}));
			this.lastChild.appendChild(this.head = new zul.box.Hlayout({
				height: this._space + 'px',
				spacing: '0px',
				sclass: 'k-group-head'
			}))
			if (this._wgt.getRowFreeze() >= 0) {
				this.lastChild.appendChild(this.frozenBody = new zul.box.Hlayout({
					spacing: '0px',
					sclass: 'k-group-body k-group-frozen-body'
				}));
			}
			this.lastChild.appendChild(this.body = new zul.box.Hlayout({
				vflex: true,
				spacing: '0px',
				sclass: 'k-group-body'
			}));
		},
		clearFrame: function () {
			[this.head, this.body, this.frozenBody].forEach(function (node) {
				if (node) {
					while (node.firstChild) {
						node.removeChild(node.firstChild);
					}
				}
			});
		},
		buildFrame: function () {
			let head = this.head, body = this.body,
				frozenBody = this.frozenBody,
				ss = this._wgt;
			for (let i = 0; i < this._outlineLevel; i++) {
				head.appendChild(new zul.wgt.Div({
					width: this._space + 'px',
					height: this._space + 'px',
					sclass: 'k-group-title', listeners: {
						onClick: (function (index) {
							return function () {
								if (ss.desktop) {
									ss.fire('onGroupHeaderClick', {
										sheetId: ss.getSheetId(),
										level: index,
										dir: 'row'
									});
								}
							};
						})(i)
					}
				}));
				head.lastChild.appendChild(new zul.wgt.Label({
					value: i + 1,
					sclass: 'k-group-text'
				}));

				if (frozenBody) {
					frozenBody.appendChild(new zul.wgt.Div({
						vflex: true,
						width: this._space + 'px',
						sclass: 'k-group-row-pane'
					}));
				}

				body.appendChild(new zul.wgt.Div({
					vflex: true,
					width: this._space + 'px',
					sclass: 'k-group-row-pane'
				}));
			}
		},
		syncFrame: function () {
			this.initFrame();
			this.clearFrame();
			this.buildFrame();
		},
		shouldShowButton: function (node, isSummaryBelow) {
			let min = 0, max = this._wgt.getMaxRows();
			if (isSummaryBelow) {
				if (node.end > max) return false;
				let buttonIndex = node.end + 1,
					parent = node.parent;
				return parent.end > buttonIndex || parent.end === buttonIndex;
			} else {
				if (node.start < min) return false;
				let buttonIndex = node.start - 1, parent = node.parent;
				return parent.start < buttonIndex || parent.start === buttonIndex;
			}
		},
		_markVisited: function(visited, node, summaryBelow) {
			if (summaryBelow) {
				for (let start = node.start, end = node.end + 1; start <= end; start++) {
					visited[start] = true;
				}
			} else {
				for (let start = node.start - 1, end = node.end; start <= end; start++) {
					visited[start] = true;
				}
			}
		},
		syncOutline: function () {
			let ss = this._wgt,
				ctrl = ss._cacheCtrl.selected,
				isRow = this.isRow_(),
				outlineLevel = isRow ? ctrl.outlineLevelRow : ctrl.outlineLevelCol;
			if (!outlineLevel) {
				this.clearOutline();
				return;
			}

			// i.e. 2 displays 3, 3 displays 4, ... between 1 to 8.
			this._outlineLevel = outlineLevel + 1;

			// build a tree to compare whether equals or not.
			let tree = this._buildOutlineTree();
			if (JSON.stringify(this.tree) == JSON.stringify(tree)) {
				return;
			}
			this.tree = tree;

			// re-build a frame
			this.syncFrame();

			let frozenBody = this.frozenBody,
				sheetCtrl = ss.sheetCtrl,
				headers = isRow ? ctrl.rowHeaders : ctrl.columnHeaders,
				sizes = isRow ? sheetCtrl.custRowHeight : sheetCtrl.custColWidth,
				freeze = isRow ? ss.getRowFreeze() : ss.getColumnFreeze(),
				summaryBelow = isRow ? ctrl.summaryBelow : ctrl.summaryRight,
				stack = [],
				fillStackFromBottomToTop = function (parent) {
					if (parent.children) {
						stack = parent.children.concat(stack);
						for (var i = 0, j = parent.children.length; i < j;) {
							fillStackFromBottomToTop(parent.children[--j]); // end to start
						}
					}
				};

			// build a stack to store each node for a whole tree from bottom to top.
			// For example,
			// |1--------8| --> a tree
			// |1-3| |5-7|
			//  |2|
			// -> [2, 1-3, 5-7, 1-8] -> a stack
			fillStackFromBottomToTop(this.tree.root);

			// build a visited 2d array to mark whether to draw a spot or not.
			let visited = [];
			for (let start = 0, end = this._outlineLevel; start <= end; start++) {
				visited.push([]);
			}

			// time complex: O(Nodes * Levels + 2N (the number of the range of a node to mark visited and draw spots))
			while (stack.length) {
				let node = stack.shift();
				if (this.shouldShowButton(node, summaryBelow)) {
					// show button
					let buttonIndex = summaryBelow ? node.end + 1 : node.start - 1;
					if (!sizes.isHidden(buttonIndex)) {
						let parent = node.parent,
							buttonLevel = node.level;
						while (parent) {
							if (summaryBelow ? parent.end != node.end : parent.start != node.start) {
								break;
							}
							buttonLevel = parent.level;
							parent = parent.parent;
						}
						let buttonHeader = headers[buttonIndex],
							collapsed = buttonHeader && buttonHeader.collapsed;
						if (frozenBody && freeze >= node.start - 1) {
							this.frozenBody.getChildAt(buttonLevel - 1).appendChild(this.newButton(node.start, node.end, collapsed, summaryBelow));
						}
						this.body.getChildAt(buttonLevel - 1).appendChild(this.newButton(node.start, node.end, collapsed, summaryBelow));

						this._markVisited(visited[buttonLevel - 1], node, summaryBelow);
					}
				}

				// draw a spot if the range of the node has not visited yet.
				for (let start = node.start, end = node.end; start <= end; start++) {
					if (!sizes.isHidden(start)) {
						if (!visited[node.level][start]) {
							if (frozenBody && freeze >= start) {
								this.frozenBody.getChildAt(node.level).appendChild(this.newSpot(start));
							}
							this.body.getChildAt(node.level).appendChild(this.newSpot(start));

							for (let level = node.level; level >= 0; level--)
								visited[level][start] = true;
						}
					}
				}
			}
		},
		afterSize: function () {
			if (this.frozenBody) {
				let rowHeight = this._wgt.sheetCtrl.custRowHeight,
					rowFrozen = this._wgt.getRowFreeze(),
					frozenHeight = rowHeight.getStartPixel(rowFrozen + 1) - rowHeight.getStartPixel(0),
					jbody = jq(this.body.$n());
				this.frozenBody.setHeight(jq.px0(frozenHeight + 1));
				jbody.css('transform', 'translate3d(0, ' + jq.px(-frozenHeight - 1) + ', 0)')
					.css('height', jq.px0(jbody.parent().height() + frozenHeight));
			}
			let sheetCtrl = this._wgt.sheetCtrl;
			if (sheetCtrl && sheetCtrl.sp) {
				// sync scrollbar position.
				this._doScrolling({data: {top: sheetCtrl.sp.scrollTop, left: sheetCtrl.sp.scrollLeft}});
			}
		},
		_doHeaderUpdate: function (evt) {
			if (!this._outlineLevel || !this.body) return;
			let data = evt.data;
			if (data.type === 'left') {
				for (let w = this.body.firstChild; w; w = w.nextSibling) {
					for (let c = w.firstChild; c; c = c.nextSibling) {
						if (c.startIndex >= data.index || c.endIndex >= data.index) {
							let pos = this.calculatePosition(c.startIndex, c.endIndex, c.getSclass().indexOf('k-group-spot') >= 0);
							c.setTop(jq.px0(pos[0]));
							c.setHeight(jq.px0(pos[1]));
						}
					}
				}
			}
		},
		calculatePosition: function (startIndex, endIndex, isSpot) {
			let ss = this._wgt,
				rowHeight = ss.sheetCtrl.custRowHeight,
				topPrev = rowHeight.getStartPixel(startIndex - 1),
				top = rowHeight.getStartPixel(startIndex),
				bottom = rowHeight.getStartPixel(endIndex),
				extraTop = isSpot ? 0 : Math.max(0, (top - topPrev - this._space) / 2),
				height = bottom - (top -= extraTop);
			return [top, height];
		},
		_doScrolling: function (evt) {
			if (!this._outlineLevel || !this.body) return;
			let data = evt.data,
				jbody = jq(this.body.$n());
			if (this.isRow_()) {
				let rowFrozen = this._wgt.getRowFreeze();
				let offsetTop = 0;
				if (rowFrozen >= 0) {
					let rowHeight = this._wgt.sheetCtrl.custRowHeight;
					offsetTop = rowHeight.getStartPixel(rowFrozen + 1) - rowHeight.getStartPixel(0) + 1;
				}

				jbody.css('transform', 'translate3d(0, ' + jq.px(-(data.top + offsetTop)) + ', 0)')
					.css('height', jq.px0(jbody.parent().height() + (data.top + offsetTop)));
			} else {
				let colFrozen = this._wgt.getColumnFreeze();
				let offsetLeft = 0;
				if (colFrozen >= 0) {
					var colWidth = this._wgt.sheetCtrl.custColWidth;
					offsetLeft = colWidth.getStartPixel(colFrozen + 1) - colWidth.getStartPixel(0) + 1;
				}
				jbody.css('transform', 'translate3d(' + jq.px(-(data.left + offsetLeft)) + ',0 , 0)')
					.css('width', jq.px0(jbody.parent().width() + data.left + offsetLeft));
			}
		},
		_doActiveRangeUpdate: function () {
			this.syncOutline();
		},
		isRow_: function () {
			return true;
		},
		getZclass: function () {
			return 'k-group-row';
		},
		redraw: zk.Page.$redraw,
	});
	zss.GroupCol = zk.$extends(zss.GroupRow, {
		isRow_: function () {
			return false;
		},
		initFrame: function () {
			if (this.head) return;
			this.setHflex(true);
			this.setHeight(this._outlineLevel * this._space + 'px');
			this.appendChild(new zul.box.Hlayout({
				hflex: true,
				vflex: true,
				spacing: '0px'
			}))
			this.lastChild.appendChild(this.head = new zul.box.Vlayout({
				hflex: 'min',
				spacing: '0px',
				sclass: 'k-group-head'
			}))
			if (this._wgt.getColumnFreeze() >= 0) {
				this.lastChild.appendChild(this.frozenBody = new zul.box.Vlayout({
					spacing: '0px',
					sclass: 'k-group-body k-group-frozen-body'
				}));
			}
			this.lastChild.appendChild(this.body = new zul.box.Vlayout({
				hflex: true,
				spacing: '0px',
				sclass: 'k-group-body'
			}))
		},
		buildFrame: function () {
			let head = this.head, body = this.body,
				frozenBody = this.frozenBody,
				ss = this._wgt,
				spaces = ss._cacheCtrl.selected.outlineLevelRow;
			for (let i = 0; i < this._outlineLevel; i++) {
				// add topLeft corner
				let headCave = head;
				headCave.appendChild(new zul.box.Hlayout({
					hflex: 'min',
					spacing: '0px'
				}))
				headCave = headCave.lastChild;
				headCave.appendChild(new zul.wgt.Div({
					width: this._space * (spaces ? spaces + 1 : 0) + 16 + 'px', // (36 - 20)px the leftPanel header width.
					height: this._space + 'px',
					sclass: 'k-group-corner'
				}));
				headCave.appendChild(new zul.wgt.Div({
					width: this._space + 'px',
					height: this._space + 'px',
					sclass: 'k-group-title', listeners: {
						onClick: (function (index) {
							return function () {
								ss.fire('onGroupHeaderClick', {
									sheetId: ss.getSheetId(),
									level: index,
									dir: 'column'
								});
							};
						})(i)
					}
				}));
				headCave.lastChild.appendChild(new zul.wgt.Label({
					value: i + 1,
					sclass: 'k-group-text'
				}));

				if (frozenBody) {
					frozenBody.appendChild(new zul.wgt.Div({
						hflex: true,
						height: this._space + 'px',
						sclass: 'k-group-col-pane'
					}));
				}
				body.appendChild(new zul.wgt.Div({
					hflex: true,
					height: this._space + 'px',
					sclass: 'k-group-col-pane'
				}));
			}
		},
		afterSize: function () {
			if (this.frozenBody) {
				let colWidth = this._wgt.sheetCtrl.custColWidth,
					colFrozen = this._wgt.getColumnFreeze(),
					frozenWidth = colWidth.getStartPixel(colFrozen + 1) - colWidth.getStartPixel(0),
					jbody = jq(this.body.$n());
				this.frozenBody.setWidth(jq.px0(frozenWidth + 1));
				jbody.css('transform', 'translate3d(' + jq.px(-frozenWidth - 1) + ',0, 0)')
					.css('width', jq.px0(jbody.parent().width() + frozenWidth));
			}
			let sheetCtrl = this._wgt.sheetCtrl;
			if (sheetCtrl && sheetCtrl.sp) {
				// sync scrollbar position.
				this._doScrolling({data: {top: sheetCtrl.sp.scrollTop, left: sheetCtrl.sp.scrollLeft}});
			}
		},
		_doHeaderUpdate: function (evt) {
			if (!this._outlineLevel || !this.body) return;
			let data = evt.data;
			if (data.type === 'top') {
				for (let w = this.body.firstChild; w; w = w.nextSibling) {
					for (let c = w.firstChild; c; c = c.nextSibling) {
						if (c.startIndex >= data.index || c.endIndex >= data.index) {
							let pos = this.calculatePosition(c.startIndex, c.endIndex, c.getSclass().indexOf('k-group-spot') >= 0);
							c.setLeft(jq.px0(pos[0]));
							c.setWidth(jq.px0(pos[1]));
						}
					}
				}
			}
		},
		getZclass: function () {
			return 'k-group-col';
		},
		calculatePosition: function (startIndex, endIndex, isSpot) {
			let ss = this._wgt,
				colWidth = ss.sheetCtrl.custColWidth,
				left = colWidth.getStartPixel(startIndex),
				right = colWidth.getStartPixel(endIndex),
				rightNext = colWidth.getStartPixel(endIndex + 1),
				extraRight = isSpot ? 0 : Math.max(0, (rightNext - right - this._space) / 2),
				width = right - left + extraRight;
			if (width === 0 && !isSpot) {
				left += (colWidth.getStartPixel(endIndex + 2) - rightNext - this._space) / 2;
			}
			return [left, width];
		},
		newButton: function (start, end, collapsed, summaryRight) {
			let ss = this._wgt,
				pos = this.calculatePosition(start, end + 1),
				div = new zul.wgt.Div({
					left: jq.px0(pos[0]), height: this._space / 2 + 'px',
					width: jq.px0(pos[1]),
					sclass: (summaryRight ? '' : 'k-group-reversed ') + 'k-group-outline' + (collapsed ? '-collapsed' : ''),
					startIndex: start,
					endIndex: end,
				});
			div.appendChild(new zul.wgt.Label({
				sclass: 'k-group-button z-icon-' + (collapsed ? 'plus' : 'minus'),
				listeners: {
					onClick: function () {
						ss.fire('onGroupOpen', {
							sheetId: ss.getSheetId(),
							start: start,
							end: end,
							open: !!collapsed,
							dir: 'column'
						});
					}
				}
			}));
			return div;
		},
		newSpot: function (index) {
			let colWidth = this._wgt.sheetCtrl.custColWidth,
				left = colWidth.getStartPixel(index),
				width = colWidth.getStartPixel(index + 1) - left,
				div = new zul.wgt.Div({
					left: jq.px0(left), width: this._space + 'px',
					width: jq.px0(width), sclass: 'k-group-spot',
					startIndex: index,
					endIndex: index + 1
				});
			div.appendChild(new zul.wgt.Label({sclass: 'z-icon-circle'}));
			return div;
		}
	});

})();