(function () {

	function element(name, props) {
		let el = document.createElement(name),
			classNames = props.classNames,
			id = props.id,
			textContent = props.textContent;
		if (classNames != undefined) {
			classNames = classNames instanceof Array ? classNames : [classNames];
			for (let i = 0; i < classNames.length; i++) {
				let className = classNames[i];
				if (className)
					el.classList.add(className);
			}
		}
		if (id != undefined) {
			el.id = id;
		}
		if (textContent != undefined) {
			el.innerText = textContent;
		}
		return el;
	}

	function div(classNames, id) {
		return element('div', {classNames: classNames, id: id});
	}

	function ul(classNames, id) {
		return element('ul', {classNames: classNames, id: id});
	}

	function li(classNames, id) {
		return element('li', {classNames: classNames, id: id});
	}

	function a(classNames, textContent) {
		return element('a', {classNames: classNames, textContent: textContent});
	}

	function input(classNames) {
		return element('input', {classNames: classNames});
	}

	/**
		<element>
			<content>
				<inner/>
			</content>
		</element>
	**/
	function btn(className, innerClassName, contentClassName) {
		let element = div(className);
		let inner = div(innerClassName);
		let content = div(contentClassName);
		element.appendChild(content);
		content.appendChild(inner);
		return {element: element, inner: inner, content: content};
	}

	function afterBind(comp) {
		let afterBind = comp.afterBind;
		if (afterBind) {
			afterBind();
		}
	}

	function listenCellSelect(comp) {
		let sheet = comp.wgt.sheetCtrl;
		if (sheet) {
			sheet.listen({onCellSelection: comp.onCellSelection, onCellUpdate: comp.onCellSelection});
		}
	}

	function fireToolbarAction(wgt, comp, action, disable) {
		if (disable || !wgt._toolbar.isVisible()) {
			return;
		}
		const sheet = wgt.sheetCtrl;
		comp.fireToolbarAction(action, triggerSelection(sheet)); //ZSS-717
	}

	function triggerSelection(sheet) {
		if (!sheet) {
			return {tRow: -1, lCol: -1, bRow: -1, rCol: -1, type: zss.SEL.ALL};
		}
		let s = sheet.getLastSelection(),
			tRow = s.top,
			lCol = s.left,
			bRow = s.bottom,
			rCol = s.right;
		sheet.triggerSelection(tRow, lCol, bRow, rCol);
		return {selection: sheet._getToolBarSelectionData()};
	}

	zss.KToolbar = function () {
		function KToolbar(wgt, btns, parent, showTooltip) {
			let _btns = btns;
			let _visible = true;
			let self = this;
			this.wgt = wgt;
			this.element = div(ktoolbarProtoClass.toolbar);
			this.setDisabled = function (v) {
				for (let i = 0; i < _btns.length; i++) {
					let btn = _btns[i];
					if (!(btn instanceof zss.KToolbarSeparator)) {
						if (v.$contains(btn.action)) {
							btn.setDisabled(true);
						} else {
							btn.setDisabled(false);
						}
					}
				}
			};
			this.bind = function () {
				for (let i = 0; i < _btns.length; i++) {
					_btns[i].bind();
				}
				afterBind(self);
			};
			this.render = function () {
				if (!_visible)
					return;
				let element = self.element;
				let btnElements = self.createBtnElements();
				for (let i = 0; i < btnElements.length; i++) {
					element.appendChild(btnElements[i]);
				}
				return element;
			};
			this.createBtnElements = function () {
				let btnElements = [];
				for (let i = 0; i < _btns.length; i++) {
					btnElements.push(_btns[i].render());
				}
				return btnElements;
			};
			this.setButtons = function (btns) {
				_btns = btns;
			}
			this.getButtons = function () {
				return _btns;
			}
			this.isVisible = function () {
				return _visible;
			}
			this.setVisible = function (visible) {
				if (_visible != visible) {
					_visible = visible;
					let el = this.element;
					if (el) {
						el.style.display = visible ? 'flex' : 'none';
					}
				}
			}
		}
		return KToolbar;
	}();

	zss.KToolbarSeparator = function () {
		function KToolbarSeparator() {
			this.bind = function () {
				afterBind(this);
			};
			this.render = function () {
				let element = this.element = div(ktoolbarProtoClass.separator);
				return element;
			};
		}
		return KToolbarSeparator;
	}();

	zss.KToolbarCustomButton = function () {
		function KToolbarCustomButton(builder, data) {
			let self = this;
			let _tooltip;
			let wgt = self.wgt = builder._wgt;

			data = data || {};

			this.bind = function() {
				let sheet = wgt.sheetCtrl;
				if (sheet) {
					sheet.listen({onCellSelection: _onCellSelection, onCellUpdate: _onCellSelection});
				}
				if (_tooltip) {
					let jqContent = jq(self.content);
					jqContent.on('mouseover', function (evt) {
						jqContent.tooltip('show');
						jqContent.tooltip('update');
					}).on('mouseout click', function (evt) {
						jqContent.tooltip('hide');
					});
				}
				afterBind(self);
			};
			this.unbind = function () {
				let afterUnbind = self.afterUnbind;
				if (afterUnbind) {
					afterUnbind();
				}
			}
			this.render = function () {
				let element = div([ktoolbarProtoClass.btn, ktoolbarProtoClass.customBtn]);
				let content = div(ktoolbarProtoClass.btnContent);
				self.element = element;
				self.content = content;
				_tooltip = data.tooltip;
				if (_tooltip) {
					jq(content).tooltip({title: _tooltip, placement: 'bottom', trigger: 'manual', animation: false,
						template: '<div class="tooltip k-toolbar-tooltip" role="tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>'});
				}
				element.appendChild(content);
				return element;
			};
			this.$n = function () {
				return self.content;
			};
			_onCellSelection = function (evt) {
				if (builder._skipEvent(evt, self)) {
					return;
				}
				let sheet = wgt.sheetCtrl;
				let cell;
				if (sheet) {
					let evtData = evt.data;
					cell = sheet.getCell(evtData.top, evtData.left);
				}
				let func = self.onCellSelection;
				if (func) {
					if (cell != undefined)
						evt.cell = cell;
					func(evt);
				}
			};
		}
		return KToolbarCustomButton;
	}();

	zss.KToolbarButton = function () {
		function KToolbarButton(builder, action, data) {
			let self = this;
			let wgt = self.wgt = builder._wgt;
			let _disable = false;
			let _popup;
			let _preSel;
			let _predropdownItem;
			let _preActionIcon;
			let _tooltip;

			data = data || {};

			this.action = action;
			this.setDisabled = function (v) {
				if (_disable != v) {
					let content = self.content;
					if (content) {
						jq(content).toggleClass(ktoolbarProtoClass.btnDisabled, v);
						if (_popup) {
							_popup.setDisabled(v);
						}
					}
					_disable = v;
				} else if (_popup) {
					_popup.setDisabled(v);
				}
			};
			this.isDisabled = function () {
				return _disable;
			};
			this.bind = function () {
				if (_popup) {
					_popup.bind();
				}
				if (!data.skipAction) {
					jq(self.inner).on('click', function (evt) {
						fireToolbarAction(wgt, self, action, _disable);
					});
				}

				if (_tooltip) {
					let jqContent = jq(self.content);
					jqContent.on('mouseover', function (evt) {
						jqContent.tooltip('show');
						jqContent.tooltip('update');
					}).on('mouseout click', function (evt) {
						jqContent.tooltip('hide');
					});
				}

				if (data.updateBtn) {
					listenCellSelect(self);
					if (_popup) {
						let items = _popup.items;
						for (let i = 0; i < items.length; i++) {
							let item = items[i];
							jq(item.element).on('click', function () {
								if (!_disable) {
									self.setSelectedEffect(true, item, true);
								}
							});
						}
					} else {
						jq(self.element).on('click', function () {
							if (!_disable) {
								self.setSelectedEffect(true);
							}
						});
					}
				}
				afterBind(self);
			};
			this.fireToolbarAction = function (action, props) {
				wgt.fireToolbarAction(action, props);
			};
			this.render = function () {
				_disable = wgt.getActionDisabled().$contains(action);
				const icon = data.iconClass || (data.custom ? '' : ktoolbarProtoClass.iconPrefix + action);
				let elements = btn([ktoolbarProtoClass.btn, ktoolbarProtoClass.btnPrefix + action],
					[ktoolbarProtoClass.btnInner, icon],
					ktoolbarProtoClass.btnContent);

				let element = self.element = elements.element;
				let inner = self.inner = elements.inner;
				let content = self.content = elements.content;

				let jqInner = jq(inner);
				if (data.custom) {
					jq(element).addClass('k-toolbar-btn-custom');
				}
				if (!icon && data.iconURI) {
					jqInner.append('<img src="' + data.iconURI + '">');
				}
				if (data.label) {
					jqInner.append('<span>' + data.label + '</span>');
				}

				let jqContent = jq(content);
				if (_disable) {
					jqContent.addClass(ktoolbarProtoClass.btnDisabled);
				}
				if (builder._showTooltip) {
					_tooltip = msgzss.action[action] || data.tooltip || '';
					if (_tooltip) {
						jqContent.tooltip({title: _tooltip, placement: 'bottom', container: 'body', trigger: 'manual',
							animation: false, template: '<div class="tooltip k-toolbar-tooltip" role="tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>'});
					}
				}
				if (_popup) {
					let popupElement = _popup.render(_disable);
					let cls = [ktoolbarProtoClass.dropdownBtn, ktoolbarProtoClass.dropdownBtnIcon];
					if (!_disable) {
						cls.push('dropdown');
					}
					cls.push('dropdown-' + action);
					const arrow = div(cls);
					content.appendChild(arrow);
					content.appendChild(popupElement);
					self.dropdownButton = arrow;
					arrow.setAttribute('data-toggle', 'dropdown');
					arrow.setAttribute('data-offset', -18);
				}
				return element;
			};
			this.onCellSelection = function (evt) {
				//ZSS-1171
				if (builder._skipEvent(evt, self)) {
					return;
				}
				let sheet = wgt.sheetCtrl;
				if (!sheet) {
					return;
				}
				const evtData = evt.data;
				const cell = sheet.getCell(evtData.top, evtData.left);
				if (!cell) {
					return;
				}
				let act = zss.Util.capitalize(action);
				let getterFunction = data.getterFunction;
				if (_popup) {
					let val = cell[getterFunction || 'get' + act]();
					let items = _popup.items;
					for (let i = 0; i < items.length; i++) {
						let item = items[i];
						if (item.action == val) { //ZSS-1276
							self.setSelectedEffect(true, item, true);
							return;
						}
					}
				} else {
					self.setSelectedEffect(cell[getterFunction || 'is' + act]());
				}
			};
			this.setSelectedEffect = function (seld, dropdownItem, updateIcon) {
				if (_preSel == seld && _predropdownItem == dropdownItem) {
					return;
				}
				_preSel = seld;
				_predropdownItem = dropdownItem;

				let jqElement = jq(self.element);
				if (seld && self.getPopup() && dropdownItem) {
					if (updateIcon && dropdownItem.actionIcon != _preActionIcon) {
						let jqInner = jq(self.inner);
						jqInner.addClass(ktoolbarProtoClass.iconPrefix + dropdownItem.actionIcon);
						if (_preActionIcon != undefined) {
							jqInner.removeClass(ktoolbarProtoClass.iconPrefix + _preActionIcon);
						}
						_preActionIcon = dropdownItem.actionIcon;
					}
				} else if (!dropdownItem) {
					jqElement.removeClass(ktoolbarProtoClass.btnSelected);
				}

				if (seld) {
					jqElement.addClass(ktoolbarProtoClass.btnSelected);
				} else {
					jqElement.removeClass(ktoolbarProtoClass.btnSelected);
				}
			};
			this.setPopup = function (popup) {
				_popup = popup;
				_popup.owner = self;
			};
			this.getPopup = function () {
				return _popup;
			};
		}
		return KToolbarButton;
	}();

	zss.KToolbarCombobox = function () {
		function KToolbarCombobox(builder, action, data) {
			let self = this;
			let wgt = self.wgt = builder._wgt;
			let _disable;
			let _popup;
			let _preRow;
			let _preCol;
			let _preValue;
			let _tooltip;

			data = data || {};

			this.action = action;

			this.setDisabled = function (v) {
				if (_disable != v) {
					let content = self.content;
					if (content) {
						// https://html.spec.whatwg.org/multipage/common-microsyntaxes.html#boolean-attributes
						if (v) {
							// If the attribute is present, its value must either be the empty string or a value that is
							// an ASCII case-insensitive match for the attribute's canonical name, with no leading or
							// trailing whitespace.
							self.inner.setAttribute('disabled', '');
						} else {
							// The values "true" and "false" are not allowed on boolean attributes. To represent a false
							// value, the attribute has to be omitted altogether.
							self.inner.removeAttribute('disabled');
						}

						let jqContent = jq(content);
						jqContent.toggleClass(ktoolbarProtoClass.btnDisabled, v);
						if (_popup) {
							jq(self.arrow).toggleClass(ktoolbarProtoClass.btnDisabled, v);
							if (!v) {
								self.arrow.setAttribute('data-toggle', 'dropdown');
								self.arrow.setAttribute('data-reference', 'parent');
							}
							_popup.setDisabled(v);
						}
					}
					_disable = v;
				}
			};
			this.isDisabled = function () {
				return _disable;
			};
			this.bind = function () {
				if (_popup) {
					_popup.bind();
				}
				listenCellSelect(self);
				let items = _popup.items;
				for (let i = 0; i < items.length; i++) {
					let item = items[i];
					const eventName = zk.mobile ? 'touchend' : 'click';
					jq(item.element).on(eventName, function (evt) {
						if (!_disable) {
							self.setSelectedEffect(item);
						}
					});
				}
				if (_tooltip) {
					let jqContent = jq(self.content);
					jqContent.on('mouseover', function (evt) {
						if (jq(evt.target).closest('.dropdown-menu').length == 0) {
							jqContent.tooltip('show');
							jqContent.tooltip('update');
						}
					}).on('mouseout click', function (evt) {
						jqContent.tooltip('hide');
					});
				}
				const $inner = jq(self.inner);
				// KEIKAI-182
				$inner.on('change', function (evt) {
					if (_disable) {
						return;
					}
					const val = $inner.val();
					for (let i = 0; i < items.length; ++i) {
						const item = items[i];
						if (val === item.action) {
							jq(item.element).trigger('click');
							return;
						}
					}
					$inner.val(_preValue);
				});
				if (zk.mobile) {
					$inner.on('touchstart mousedown', function (evt) {
						evt.preventDefault();
					});
				}
				afterBind(self);
			};
			this.render = function () {
				let element = self.element = div([ktoolbarProtoClass.btn, ktoolbarProtoClass.comboBtn, ktoolbarProtoClass.btnPrefix + action]);
				let inner = self.inner = input(ktoolbarProtoClass.btnInner);
				let content = self.content = div(ktoolbarProtoClass.btnContent);
				element.appendChild(content);
				content.appendChild(inner);

				_disable = wgt.getActionDisabled().$contains(action);
				if (_disable) {
					self.inner.setAttribute('disabled', '');
					jq(content).addClass(ktoolbarProtoClass.btnDisabled);
				}
				if (builder._showTooltip) {
					_tooltip = msgzss.action[action] || data.tooltip || '';
					if (_tooltip) {
						jq(content).tooltip({title: _tooltip, placement: 'bottom', trigger: 'manual', animation: false,
							template: '<div class="tooltip k-toolbar-tooltip" role="tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>'});
					}
				}
				if (_popup) {
					let popupElement = _popup.render();
					let cls = [ktoolbarProtoClass.dropdownBtn, ktoolbarProtoClass.dropdownBtnIcon];
					if (!_disable) {
						cls.push('dropdown');
					}
					cls.push('dropdown-' + action);
					let arrow = self.arrow = div(cls);
					content.appendChild(arrow);
					content.appendChild(popupElement);
					if (_disable) {
						jq(arrow).addClass(ktoolbarProtoClass.btnDisabled);
					} else {
						arrow.setAttribute('data-toggle', 'dropdown');
						arrow.setAttribute('data-reference', 'parent');
					}
				}
				return element;
			};
			this.onCellSelection = function (evt) {
				//ZSS-1171: no show context menu, skip
				if (builder._skipNoShow()) {
					return;
				}

				let evtData = evt.data;
				//ZSS-1171: skip if select on same cell
				if (evtData.top == _preRow && evtData.left == _preCol) {
					return;
				}
				_preRow = evtData.top;
				_preCol = evtData.left;
				const sheet = wgt.sheetCtrl;
				if (!sheet) {
					return;
				}
				let	cell = sheet.getCell(evtData.top, evtData.left);
				if (cell) {
					let act = cell[data.getterFunction || 'get' + zss.Util.capitalize(action)]();
					_updateFontStates(act);
				}
			};
			let _updateFontStates = function (act) {
				//ZSS-1171: skip if same state
				if (_preValue == act) {
					return;
				}
				_preValue = act;

				self.inner.value = act;
			};
			this.setSelectedEffect = function (item) {
				let act = item.action;
				_updateFontStates(act);
			};
			this.setPopup = function (popup) {
				_popup = popup;
				_popup.owner = self;
			};
			this.getPopup = function () {
				return _popup;
			};
		}
		return KToolbarCombobox;
	}();

	zss.KToolbarCheckbox = function () {
		function KToolbarCheckbox(builder, action, data) {
			let self = this;
			let wgt = self.wgt = builder._wgt;
			let _disable;
			let _preActionIcon;
			let _tooltip;

			data = data || {};

			this.action = action;
			this.setDisabled = function (v) {
				if (_disable != v) {
					let content = self.content;
					if (content) {
						jq(content).toggleClass(ktoolbarProtoClass.btnDisabled, v);
					}
					_disable = v;
				}
			};
			this.isDisabled = function () {
				return _disable;
			};
			this.bind = function () {
				let sheet = wgt.sheetCtrl;
				if (sheet) {
					const evt = data.onCheckEvt || 'on' + zss.Util.capitalize(action);
					let obj = {};
					obj[evt] = self.onCheck;
					sheet.listen(obj);
				}
				jq(self.element).on('click', function () {
					fireToolbarAction(wgt, self, action, _disable);
				});
				if (_tooltip) {
					let jqContent = jq(self.content);
					jqContent.on('mouseover', function (evt) {
						jqContent.tooltip('show');
						jqContent.tooltip('update');
					}).on('mouseout click', function (evt) {
						jqContent.tooltip('hide');
					});
				}
				afterBind(self);
			};
			this.fireToolbarAction = function (action, props) {
				wgt.fireToolbarAction(action, props);
			};
			this.render = function () {
				_disable = wgt.getActionDisabled().$contains(action);
				let elements = btn([ktoolbarProtoClass.btn, ktoolbarProtoClass.checkboxBtn, ktoolbarProtoClass.btnPrefix + action],
					[ktoolbarProtoClass.btnInner, data.iconClass || ktoolbarProtoClass.iconPrefix + action],
					ktoolbarProtoClass.btnContent);

				let element = self.element = elements.element;
				self.inner = elements.inner;
				let content = self.content = elements.content;

				let jqContent = jq(content);
				let cls = [ktoolbarProtoClass.checkboxBtnCheckbox, ktoolbarProtoClass.checkboxBtnCheckboxIcon];
				let chechbox = self.checkbox = div(cls);

				jqContent.prepend(chechbox);
				if (_disable) {
					jqContent.addClass(ktoolbarProtoClass.btnDisabled);
				}
				if (builder._showTooltip) {
					_tooltip = msgzss.action[action] || data.tooltip || '';
					if (_tooltip) {
						jqContent.tooltip({title: _tooltip, placement: 'bottom', trigger: 'manual', animation: false,
							template: '<div class="tooltip k-toolbar-tooltip" role="tooltip"><div class="tooltip-arrow"></div><div class="tooltip-inner"></div></div>'});
					}
				}
				return element;
			};
			this.onCheck = function (evt) {
				let prop = data.checkAction || 'checked';
				jq(self.checkbox).toggleClass(ktoolbarProtoClass.checkboxBtnCheckboxCheckedIcon, evt.data[prop]);
			};
		}
		return KToolbarCheckbox;
	}();

	zss.KMenupopup = function () {
		function KMenupopup(wgt, action, data, prefix) {
			let self = this;

			data = data || {};

			this.wgt = wgt;
			this.items = [];

			for (let i = 0; i < data.length; i++) {
				let item = new zss.KMenuitem(wgt, data[i], prefix, action);
				self.items.push(item);
				item.parent = self;
			}
			this.setDisabled = function (v) {
				let element = self.element;
				if (element) {
					jq(element).toggleClass('disable-show', v);
					let items = self.items;
					for (let i = 0; i < items.length; i++) {
						items[i].setDisabled(v);
					}
				}
			};
			this.bind = function () {
				let items = self.items;
				for (let i = 0; i < items.length; i++) {
					items[i].bind();
				}
				afterBind(self);
			};
			this.render = function (disable) {
				let element = self.element = ul('dropdown-menu');
				let items = self.items;
				for (let i = 0; i < items.length; i++) {
					let itemEl = items[i].render(disable);
					if (itemEl instanceof Node) {
						element.appendChild(itemEl);
					}
				}
				return element;
			};
			this.appendChild = function (child) {
				self.items.push(child);
				child.parent = self;
			};
		}
		return KMenupopup;
	}();

	zss.KMenuitem = function () {
		function KMenuitem(wgt, action, prefix, supAction, data) {
			let self = this;
			let _disable;
			let _popup;
			self.actionIcon = action.replace(/\s/g, '');

			data = data || {};

			this.action = action;

			this.setDisabled = function (v) {
				v = v || wgt.getActionDisabled().$contains(action);
				if (_disable != v) {
					let el = this.element;
					if (el) {
						jq(el).toggleClass(ktoolbarProtoClass.btnDisabled, v);
					}
					_disable = v;
				}
			};
			this.isDisabled = function () {
				return _disable;
			};
			this.bind = function () {
				if (_popup) {
					_popup.bind();
					let element = self.element;
					let jqElement = jq(element);
					element.addEventListener('mouseenter', function (evt) {
						let ppElement = _popup.element;
						let jqPpElement = jq(ppElement);
						let ppWidth = _popup.element.offsetWidth;
						if (ppWidth + element.offsetWidth + jqElement.offset().left > document.body.offsetWidth) {
							jqPpElement.css('left', -ppWidth);
						} else {
							jqPpElement.css('left', '');
						}
					})
				} else {
					let func = function () {
						fireToolbarAction(wgt, self, action, _disable);
						bootstrap.Dropdown.clearMenus();
					};
					this.element.addEventListener('click', func);
					this.element.addEventListener('touchend', func);
				}
				afterBind(self);
			};
			this.fireToolbarAction = function (action, props) {
				wgt.fireToolbarAction(action, props);
			};
			this.render = function (disable) {
				let element = self.element = li(_popup != undefined ? ['dropdown-submenu', ktoolbarProtoClass.dropdownSubBtnIcon] : undefined);
				let jqElement = jq(element);

				_disable = wgt.getActionDisabled().$contains(action) || disable;
				if (_disable) {
					jqElement.addClass(ktoolbarProtoClass.btnDisabled);
				}

				let cls = ['dropdown-item', ktoolbarProtoClass.iconPrefix + self.actionIcon];
				if (prefix != undefined) {
					let lowerAction = action.toLowerCase();
					cls = cls.concat([prefix + '-' + lowerAction.replace(/\s/g, '-')]);
				}
				let label = msgzss.action[action];
				let anchor = a(cls, label != undefined ? label : action);

				element.appendChild(anchor);
				if (_popup) {
					element.appendChild(_popup.render());
				}
				return element;
			};
			this.setPopup = function (popup) {
				_popup = popup;
				_popup.owner = self;
			};
			this.getPopup = function () {
				return _popup;
			};
		}
		return KMenuitem;
	}();

	zss.Colorpicker = function () {
		function Colorpicker(builder, action, owner, data) {
			let self = this;
			let wgt = self.wgt = builder._wgt;
			let _picker;

			data = data || {};

			this.setDisabled = function (v) {
				// do nothing
			};
			this.isDisabled = function () {
				// do nothing
			};

			this.bind = function () {
				if (self.onCellSelection) {
					listenCellSelect(self);
				}

				let jqPickerEl = _picker.element;
				jqPickerEl.on('colorpickerChange', function (evt) {
					updateHexRGBInputs(evt.color, evt.value);
				});
				/**
				 * @param {string} [value] - the argument to `setValue`
				 */
				function updateHexRGBInputs(color, value) {
					if (color) {
						const hexColor = color.toHexString();
						jqPickerEl.find('.current-color').css('backgroundColor', hexColor);
						jqPickerEl.find('.hex input').val(
							value && /^[0-9a-fA-F]{3}$/.test(value) ?
							value : // Keep the 3-digit hex value "in the middle of input", as `hexColor` is always 6-digit.
							hexColor.substring(1) // Remove the leading '#' character.
						);
						const [r, g, b] = color._color.rgb().color;
						jqPickerEl.find('.r input').val(Math.round(r));
						jqPickerEl.find('.g input').val(Math.round(g));
						jqPickerEl.find('.b input').val(Math.round(b));
					}
				}

				let jqPp = jq(self.parent.element);
				let func = function (evt) {
					let classList = evt.target.classList;
					if (!classList.contains('colorpicker-ok') &&
						!classList.contains('colorpicker-swatch--inner') &&
						!classList.contains('current-color') &&
						!classList.contains('colorpicker-customize')) {
						if (classList.contains('colorpicker-cancel')) {
							_picker.setValue(_picker.lastColor);
						} else {
							evt.stopPropagation();
						}
					} else {
						let color = _picker.color;
						let hexColor = color.toHexString();
						_picker.lastColor = color;
						if (classList.contains('colorpicker-ok')) {
							// Fill the Hex/RGB inputs with valid values in case they weren't.
							// `_picker.setValue(hexColor)` won't cut it because `setValue` won't update if the color stays the same.
							updateHexRGBInputs(color);

							let jqNewColor = jq(div('colorpicker-customize'));
							jqNewColor.on('click', function (evt) {
								if (evt.target.style.backgroundColor) {
									_picker.setValue(zss.Util.rgbToHex(evt.target.style.backgroundColor));
								}
							});
							jqNewColor.css('backgroundColor', hexColor);
							jqNewColor.attr('data-name', 'k-' + hexColor.substr(1).toLowerCase());
							jqPickerEl.find('.colorpicker-customize-wrapper').prepend(jqNewColor);
							jqPickerEl.find('.colorpicker-customize').last().remove();
						}
						const sht = wgt ? wgt.sheetCtrl : null;
						if (sht) {
							if (wgt._toolbar.isVisible()) {
								wgt.fireToolbarAction(action,
									{color: hexColor, selection: sht._getToolBarSelectionData()}); //ZSS-717
							}
							bootstrap.Dropdown.clearMenus();
						}
						jqPickerEl.find('.k-colorpicker-seleted-color').removeClass('k-colorpicker-seleted-color');
						jqPickerEl.find('[data-name="k-' + hexColor.substr(1).toLowerCase() + '"]').addClass('k-colorpicker-seleted-color');
						jq(owner.element).find('.color').css('backgroundColor', hexColor);
					}
				};
				jqPp.on('click', func);
				jqPp.on('touchend', func);

				jqPickerEl.find('.colorpicker-picker-toggle').on('click', function () {
					jqPickerEl.find('.colorpicker-saturation-wrapper').toggleClass('show');
				});
				afterBind(self);
			};
			this.render = function () {
				let ppEl = self.parent.element;
				jq(ppEl).on('afterColorpickerCreate', function () {
					let jqPickerEl = jq(this);
					let jqInlineEl = jqPickerEl.find('.colorpicker-inline');
					let swatches = jqPickerEl.find('.colorpicker-bar.colorpicker-swatches');
					jqInlineEl.prepend(swatches[2]);
					jqInlineEl.prepend(jq('<div class="colorpicker-title">' + colorpickerLabel.standardColorsTitle + '</div>'));
					jqInlineEl.prepend(swatches[1]);
					jqInlineEl.prepend(swatches[0]);
					jqInlineEl.prepend(jq('<div class="colorpicker-title">' + colorpickerLabel.popularColorsTitle + '</div>'));

					function okOnEnter(keyboardEvent) {
						// Allow user to submit Hex/RGB input while they are at it.
						// The "OK" button could be hidden (colorpicker-picker-toggle) which could confuse users from submitting the color.
						if (keyboardEvent.key === 'Enter') {
							keyboardEvent.preventDefault(); // Prevent the spreadsheet active cell from moving to the cell below.
							jqPickerEl.find('.colorpicker-ok').click();
						}
						keyboardEvent.stopPropagation(); // KEIKAI-616: backspace doesn't work in hex input
					}
					for (const [i, className] of ['r', 'g', 'b'].entries()) {
						const pos = i * 2 + 1;
						jqPickerEl.find(`.${className} input`)
							.on('keydown', okOnEnter)
							.on('input', function () {
								if (this.checkValidity()) {
									// `toHexString` always returns a 6-digit hex string prefixed with '#'.
									const hexString = _picker.color.toHexString();
									// For instance, `hexString === '#123456'`, `className === 'g'`. Then, `pos === 3`.
									//                            pos --^ ^-- pos + 2
									_picker.setValue(
										hexString.substring(0, pos) + // '#12'
										Number(this.value).toString(16).padStart(2, '0') + // E.g., 255 => 'ff', 128 => '80', 1 => '00'
										hexString.substring(pos + 2) // '56'
										// This should yield '#12gg56' where 'gg' is the two-digit hex value for the green channel.
									);
								}
							});
					}
					jqPickerEl.find('.hex input')
						.on('keydown', okOnEnter)
						.on('input', function () {
							if (this.checkValidity()) {
								_picker.setValue(this.value);
							}
						});
				});
				_picker = self.picker = owner.picker = new jq.colorpicker(ppEl, {
					inline: true,
					container: true,
					format: 'hsl',
					color: 'hsl(210, 0%, 100%)',
					horizontal: true,
					useAlpha: false,
					fallbackColor: true,
					template: zss.Colorpicker.prototype.template,
					extensions: [
						{
							name: 'swatches',
							options: {
								colors: zss.Colorpicker.prototype.swatches0,
								namesAsValues: false
							}
						},
						{
							name: 'swatches',
							options: {
								colors: zss.Colorpicker.prototype.swatches1,
								namesAsValues: false
							}
						},
						{
							name: 'swatches',
							options: {
								colors: zss.Colorpicker.prototype.swatches2,
								namesAsValues: false
							}
						}
					],
					slidersHorz: {
						saturation: {
							maxLeft: 295,
							maxTop: 140
						},
						hue: {
							maxLeft: 295
						}
					}
				});
			};
		}
		return Colorpicker;
	}();

	let ktoolbarProto = zss.KToolbar.prototype;
	let ktoolbarProtoClass = ktoolbarProto.class = {
		toolbar: 'k-toolbar',
		btn: 'k-toolbar-btn',
		btnPrefix: 'k-toolbar-btn-',
		iconPrefix: 'k-icon-',
		btnContent: 'k-toolbar-btn-content',
		btnInner: 'k-toolbar-btn-inner',
		dropdownBtn: 'k-toolbar-btn-dropdown',
		dropdownBtnIcon: 'k-icon-dropdown',
		dropdownSubBtnIcon: 'k-icon-dropdown-sub',
		separator: 'k-toolbar-separator',
		comboBtn: 'k-toolbar-btn-combo',
		checkboxBtn: 'k-toolbar-btn-checkbox',
		checkboxBtnCheckbox: 'k-toolbar-btn-check',
		checkboxBtnCheckboxIcon: 'k-icon-checkboxbtn',
		checkboxBtnCheckboxCheckedIcon: 'k-icon-checkboxbtn-checked',
		customBtn: 'k-toolbar-btn-custom',
		btnSelected: 'k-toolbar-btn-seld',
		btnDisabled: 'k-toolbar-btn-disabled',
	};

	let colorpickerProto = zss.Colorpicker.prototype;
	if (!msgzss.colorpicker) {
		msgzss.colorpicker = {
			standardColorsTitle: '',
			popularColorsTitle: '',
			recentColorTitle: '',
			customizeTitle: '',
		};
	}
	let colorpickerLabel = colorpickerProto.label = {
		standardColorsTitle: msgzss.colorpicker.standardColorsTitle,
		popularColorsTitle: msgzss.colorpicker.popularColorsTitle,
		recentColorTitle: msgzss.colorpicker.recentColorTitle,
		customizeTitle: msgzss.colorpicker.customizeTitle,
	};
	colorpickerProto.swatches0 = {
		'k-ffffff': '#ffffff',
		'k-000000': '#000000',
		'k-e7e6e6': '#e7e6e6',
		'k-eeece1': '#eeece1',
		'k-44546a': '#44546a',
		'k-4473c4': '#4473c4',
		'k-ec7d31': '#ec7d31',
		'k-a5a5a5': '#a5a5a5',
		'k-ffc00c': '#ffc00c',
		'k-5b9bd5': '#5b9bd5',
		'k-70ad47': '#70ad47',
		'k-c0504d': '#c0504d',
		'k-8064a2': '#8064a2',
		'k-4cacc6': '#4cacc6',
		'k-f79646': '#f79646'
	};
	colorpickerProto.swatches1 = {
		'k-f2f2f2': '#f2f2f2',
		'k-7f7f7f': '#7f7f7f',
		'k-d0cece': '#d0cece',
		'k-ddd9c3': '#ddd9c3',
		'k-d5dce4': '#d5dce4',
		'k-d9e2f3': '#d9e2f3',
		'k-fce4d5': '#fce4d5',
		'k-ededed': '#ededed',
		'k-fff2cc': '#fff2cc',
		'k-deebf6': '#deebf6',
		'k-e2efd9': '#e2efd9',
		'k-f2dcdb': '#f2dcdb',
		'k-e5e0ec': '#e5e0ec',
		'k-dbeef3': '#dbeef3',
		'k-fdeada': '#fdeada',
		'k-d8d8d8': '#d8d8d8',
		'k-595959': '#595959',
		'k-aeabab': '#aeabab',
		'k-c4bd97': '#c4bd97',
		'k-adb9ca': '#adb9ca',
		'k-b4c6e7': '#b4c6e7',
		'k-f7cbac': '#f7cbac',
		'k-dbdbdb': '#dbdbdb',
		'k-fee599': '#fee599',
		'k-bdd7ee': '#bdd7ee',
		'k-c6e0b3': '#c6e0b3',
		'k-e5b9b7': '#e5b9b7',
		'k-ccc1d9': '#ccc1d9',
		'k-b7dde8': '#b7dde8',
		'k-fbd5b5': '#fbd5b5',
		'k-bfbfbf': '#bfbfbf',
		'k-3f3f3f': '#3f3f3f',
		'k-757070': '#757070',
		'k-938953': '#938953',
		'k-8496b0': '#8496b0',
		'k-8eaadb': '#8eaadb',
		'k-f4b183': '#f4b183',
		'k-c9c9c9': '#c9c9c9',
		'k-ffd964': '#ffd964',
		'k-9cc3e5': '#9cc3e5',
		'k-a9d08d': '#a9d08d',
		'k-d99694': '#d99694',
		'k-b2a2c7': '#b2a2c7',
		'k-92cddc': '#92cddc',
		'k-fac08f': '#fac08f',
		'k-a5a5a5': '#a5a5a5',
		'k-262626': '#262626',
		'k-3a3838': '#3a3838',
		'k-494429': '#494429',
		'k-323f4f': '#323f4f',
		'k-2f5496': '#2f5496',
		'k-c55a12': '#c55a12',
		'k-7b7b7b': '#7b7b7b',
		'k-bf9006': '#bf9006',
		'k-2e75b5': '#2e75b5',
		'k-528135': '#528135',
		'k-953734': '#953734',
		'k-5f497a': '#5f497a',
		'k-31859b': '#31859b',
		'k-e36c09': '#e36c09',
		'k-828282': '#828282',
		'k-0c0c0c': '#0c0c0c',
		'k-171616': '#171616',
		'k-1d1b10': '#1d1b10',
		'k-222B35': '#222B35',
		'k-1f3864': '#1f3864',
		'k-833c0a': '#833c0a',
		'k-525252': '#525252',
		'k-7f6002': '#7f6002',
		'k-1d4e79': '#1d4e79',
		'k-375623': '#375623',
		'k-632423': '#632423',
		'k-3f3151': '#3f3151',
		'k-205867': '#205867',
		'k-974806': '#974806',
	};
	colorpickerProto.swatches2 = {
		'k-c00200': '#c00200',
		'k-ff0200': '#ff0200',
		'k-ff950e': '#ff950e',
		'k-fefb1a': '#fefb1a',
		'k-ffc00e': '#ffc00e',
		'k-91d051': '#91d051',
		'k-00b050': '#00b050',
		'k-00ccb5': '#00ccb5',
		'k-00afef': '#00afef',
		'k-016fbf': '#016fbf',
		'k-0137bf': '#0137bf',
		'k-001f5f': '#001f5f',
		'k-702f9f': '#702f9f',
		'k-9f2e7f': '#9f2e7f',
		'k-de39ab': '#de39ab',
	};
	colorpickerProto.template =
		'<div class="colorpicker">' +
			'<div class="colorpicker-title">' + colorpickerLabel.recentColorTitle + '</div>' +
			'<div class="colorpicker-bar colorpicker-customize-wrapper">' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
				'<div class="colorpicker-customize"></div>' +
			'</div>' +
			'<div class="colorpicker-bar colorpicker-input">' +
				'<div class="current-color"></div>' +
				'<div class="hex">' +
					'<input type="text" minlength="1" maxlength="6" pattern="([0-9a-fA-F]{6})|([0-9a-fA-F]{3})" required/>' +
					'<span>HEX</span>' +
				'</div>' +
				'<div class="r">' +
					'<input type="number" min="0" max="255" step="1" required/>' +
					'<span>R</span>' +
				'</div>' +
				'<div class="g">' +
					'<input type="number" min="0" max="255" step="1" required/>' +
					'<span>G</span>' +
				'</div>' +
				'<div class="b">' +
					'<input type="number" min="0" max="255" step="1" required/>' +
					'<span>B</span>' +
				'</div>' +
			'</div>' +
			'<div class="colorpicker-picker-toggle k-icon-dropdown-sub">' +
				'<div class="colorpicker-picker-toggle-btn"></div>' +
				'<span>' + colorpickerLabel.customizeTitle + '</span>' +
			'</div>' +
			'<div class="colorpicker-saturation-wrapper">' +
				'<div class="colorpicker-saturation"><i class="colorpicker-guide"></i></div>' +
				'<div class="colorpicker-hue"><i class="colorpicker-guide"></i></div>' +
				'<div class="colorpicker-alpha">' +
					'<div class="colorpicker-alpha-color"></div>' +
					'<i class="colorpicker-guide"></i>' +
				'</div>' +
				'<div class="colorpicker-btns">' +
					'<div class="colorpicker-ok">OK</div>' +
					'<div class="colorpicker-cancel">Cancel</div>' +
				'</div>' +
			'</div>' +
		'</div>';
})();
