/* LeftPanel.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/
(function () {

zss.Panel = zk.$extends(zk.Widget, {
	widgetName: 'Panel',
	$o: zk.$void, //owner, fellows relationship no needed
	$init: function (sheet, hide, start, end, data, isCorner) {
		this.$supers(zss.Panel, '$init', []);
		
		this.sheet = sheet;
		this.hidehead = hide;
		this.isCorner = isCorner;
		
		this.headers = this.initHeaders_(sheet, start, end, data, isCorner);
		this.block = this.initFrozenBlock_(sheet, start, end, data);
	},
	/**
	 * Returns {@link Header}
	 */
	getHeader: function (index) {
		var headers = this.headers;
		index = this._binarySearch(index);
		return headers[index];
	},
	getHeaderIndex: function (index) {
		return this._binarySearch(index);
	},
	_binarySearch: function (index) {
		let low = 0,
			headers = this.headers,
			high = this.headers.length - 1;
		while (low <= high) {
			let mid = (low + high) >> 1;
			let header = headers[mid];
			if (header.index < index) {
				low = mid + 1;
			} else if (header.index > index) {
				high = mid - 1;
			} else {
				return mid;
			}
		}
		return -(low + 1);
	},
	_getClosestHeaderIndex: function (index) {
		let headerIndex = this._binarySearch(index);
		if (headerIndex < 0) {
			headerIndex = -headerIndex - 1;
		}
		return headerIndex;
	},
	/**
	 * Create zss.Header child widgets, invoke on Panel widget initialization
	 */
	initHeaders_: function (sheet, start, end, data, isCorner) {
		//to be overridden
		return null;//return header widget array
	},
	isEmptyHeader: function() {
		return !(this.headers && this.headers.length > 0) || !this.getHeaderData_();
	},
	/**
	 * Create frozen block, invoke on Panel widget initialization 
	 */
	initFrozenBlock_: function (sheet, start, end, data) {
		//to be overridden
		return null;//return zss.CellBlockCtrl if there's frozen area
	},
	createHeaders_: function (dir, start, end) {
		var sheet = this.sheet,
			headers = this.getHeaderData_(),
			isInsert = this.isInsert_(dir),
			type = this.getHeaderType_(),
			html = '',
			hs = [];
		if ('jump' == dir)
			this._clearAllHeader();
		else {
			var sheetHeaders = this.headers,
				length = sheetHeaders.length;
			if (length > 0) {
				var headerStartIndex = sheetHeaders[0].index,
					headerEndIndex = sheetHeaders[length - 1].index;
				if (start >= headerStartIndex && end <= headerEndIndex)
					return null; //headers already exist
				else if (start <= headerEndIndex && end >= headerStartIndex) {
					if (start >= headerStartIndex)
						start = headerEndIndex + 1;
					else if (headerStartIndex != 0)
						end = headerStartIndex -1;
				}
			}
		}
		const sizes = type == zss.Header.HOR ? sheet.custColWidth : sheet.custRowHeight;
		for (let i = start; i <= end; i++) {
			if (!sizes.isHidden(i)) {
				hs.push(new zss.Header(sheet, type, headers[i]));
			}
		}
		if (isInsert) {
			this.insertHeaders_(0, hs);
		} else {
			this.appendHeaders_(hs);
		}
		this.updateSelectionCSS_();
		return [start, end];
	},
	updateSelectionCSS_: function () {
		//to be overridden
	},
	_clearAllHeader: function () {
		var hs = this.headers,
			size = hs ? hs.length : 0;
		this.removeChildFromStart_(size);
	},
	/**
	 * Returns the cached active range data source
	 */
	getHeaderData_: function () {
		//to be overridden
	},
	/**
	 * Returns frozen data
	 */
	getFrozenData_: function () {
		//to be overridden
	},
	/**
	 * Returns frozen header data
	 */
	getFrozenHeaderData_: function () {
		//to be overridden
	},
	getHeaderType_: function () {
		//to be overridden
	},
	isInsert_: function (dir) {
		//to be overridden
	},
	/**
	 * Append zss.Header widgets
	 *
	 * @param zss.Header[]
	 */
	appendHeaders_: function (hs) { // KEIKAI-200: batch append
		const headers = this.headers;
		const fragment = document.createDocumentFragment();
		for (let i = 0; i < hs.length; i++) {
			const h = hs[i];
			this.appendChild(h, true);
			fragment.appendChild(h.getHtmlDom());
		}
		this.hcomp.appendChild(fragment);
		for (let i = 0; i < hs.length; i++) {
			const h = hs[i];
			headers.push(h);
			h.bind();
		}
	},
	/**
	 * Insert zss.Header widgets
	 *
	 * @param int index
	 * @param zss.Header[]
	 */
	insertHeaders_: function (index, hs) { // KEIKAI-200: batch insert
		const headers = this.headers;
		const sibling = headers[index];
		if (!sibling) {
			this.appendHeaders_(hs);
		} else {
			const fragment = document.createDocumentFragment();
			for (let i = 0; i < hs.length; i++) {
				const h = hs[i];
				this.insertBefore(h, sibling, true);
				fragment.appendChild(h.getHtmlDom());
			}
			this.hcomp.insertBefore(fragment, sibling.$n());
			const dt = this.desktop;
			for (let i = 0; i < hs.length; ++i) {
				const h = hs[i];
				headers.splice(index + i, 0, h);
				h.bind();
			}
		}
	},
	/**
	 * Create cells and associated headers
	 */
	create_: function (dir, headerStart, headerEnd, frozenStart, frozenEnd, createFrozenOnly) {
		//to be overridden
	},
	/**
	 * Remove child from start
	 * 
	 * @param int size 
	 */
	removeChildFromStart_: function (size) {
		if (this.hidehead) {
			return;
		}
		// KEIKAI-225
		zkS.removeWidgetFromStart(this.headers, size);
	},
	/**
	 * Remove headers from end
	 * 
	 * @param int size 
	 */
	removeChildFromEnd_: function (size) {
		if (this.hidehead) {
			return;
		}
		// KEIKAI-225
		zkS.removeWidgetFromEnd(this.headers, size);
	},
	removeHeader_: function (index, size) {
		let headers = this.headers;
		
		if (index > headers.length) return;

		// ZSS-488: these code can cover every situation, we don't need to consider "first index" situation separately
		let rem = headers.splice(index, size);
		// KEIKAI-225
		zkS.removeWidgetFromStart(rem, size);
	},
	bind_: function () {
		this.$supers(zss.Panel, 'bind_', arguments);
		
		var n = this.comp = this.$n();
		n.ctrl = this;
		zk(n).disableSelection();//disable selectable

		this.padcomp = this.$n('pad'),
		this.icomp = this.$n('real'),
		this.hcomp = this.icomp.firstChild,
		this.wpcomp = this.$n(this.type+'wp');
		
		if (this.block) {
			this.bindFrozenCtrl_();
			
			//ZSS-1043
			var sheet = this.sheet;
			if(sheet.editorFocusMark) {
				var dummy = "";
				let _editorFocusMark = sheet.editorFocusMark;
				let keys = _editorFocusMark != null ? Object.keys(_editorFocusMark) : [];
				for (var k = 0; k < keys.length; k++) {
					let id = keys[k];
					var focusMark = _editorFocusMark[id];
					this.addEditorFocus(id, dummy);
				}
			}
		}
	},
	bindFrozenCtrl_: function () {
		//to be overridden
	},
	unbindFrozenCtrl_: function () {
		if (this.selArea) {
			this.selArea.cleanup();
			this.selArea = null;
		}
		if (this.selChgArea) {
			this.selChgArea.cleanup();
			this.selChgArea = null;
		}
		if (this.focusMark) {
			this.focusMark.cleanup();
			this.focusMark = null;
		}
		if (this.hlArea) {
			this.hlArea.cleanup();
			this.hlArea = null;
		}
	},
	unbind_: function () {
		
		if (this.headers) {
			this.headers.splice(0, this.headers.length);
		}
		
		//ZSS-1043
		if (this.editorFocusMark) {
			let _editorFocusMark = this.editorFocusMark;
			let keys = _editorFocusMark != null ? Object.keys(_editorFocusMark) : [];
			for (var k = 0; k < keys.length; k++) {
				let id = keys[k];
				var focusMark = _editorFocusMark[id];
				this._removeEditorFocusMark(focusMark);
			}
			this.editorFocusMark = null;
		}
		
		this.unbindFrozenCtrl_();
		this.comp = this.comp.ctrl = this.padcomp = this.wpcomp = this.icomp = 
		this.hcomp = this.headers = this.block = this.sheet = null;
		this.$supers(zss.Panel, 'unbind_', arguments);
	},
	getTypeAttr_: function () {
		//to be overridden
	},
	hasPad_: function () {
		return false;
	},
	doTooltipOver_: zk.$void,
	doTooltipOut_: zk.$void,
	doClick_: zk.$void,
	doMouseDown_: function (evt) {
		this.sheet._doMousedown(evt);
	},
	doMouseUp_: function (evt) {
		this.sheet._doMouseup(evt);
	},
	doRightClick_: function (evt) {
		this.sheet._doMouserightclick(evt);
	},
	doDoubleClick_: function (evt) {
		this.sheet._doMousedblclick(evt);
	},
	doMouseOver_: function (evt) {
		var n = evt.domTarget,
			header = zk.$(n);

		if (header.$instanceof(zss.Header) && !header.draging) {
			delete header._mouseLeaving;
			if (n.getAttribute('zs.t') == "SBoun")
				header._processDrag(true, false);
			if (n.getAttribute('zs.t') == "SBoun2")
				header._processDrag(true, true);
		}
	},
	doMouseOut_: function (evt) {
		var n = evt.domTarget,
			header = zk.$(n);
		if (header.$instanceof(zss.Header) && !header.draging) {
			if (n.getAttribute('zs.t') == "SBoun") {
				header._processDrag(false, false);
			} else if (n.getAttribute('zs.t') == "SBoun2") {
				header._processDrag(false, true);
			}
			if (evt.domEvent.relatedTarget) {
				if (header != zk.$(evt.domEvent.relatedTarget)) {
					header._doMouseLeave(evt);
				}
			}
		}
	},
	redraw: function (out) {
		var type = this.type,
			type = this.isCorner ? 'corner' + type : type,
			uid = this.uuid;
		out.push('<div id="', uid, '" class="zs', type, ' zsfz', type, '" zs.t="',
				this.getTypeAttr_(), '">', 
				(this.hasPad_() ? '<div id="' + uid + '-pad" class="zs' + type + 'pad"></div>' : ''),
				'<div id="', uid, '-real" class="zs', type, 
				'i"><div id="', uid, '-head" class="zs', type, 'head">');

		// ZSS-937 output DOM however. simply hiding headers for hidden state is enough. 
		var hs = this.headers,
			size = hs.length;
		for (var i = 0; i < size; i++) {
			out.push(hs[i].getHtml());
		}

		out.push('</div>'); //head div end
		if (this.block) { //frozen block
			this.block.redraw(out);
			out.push('<div id="', uid, '-select" class="zsselect" zs.t="SSelect">',
					'<div class="zsselecti" zs.t="SSelInner"></div><div class="zsseldot" zs.t="SSelDot"></div></div>',
					'<div id="', uid, '-selchg" class="zsselchg" zs.t="SSelChg"><div class="zsselchgi"></div></div>',
					'<div id="', uid, '-focmark" class="zsfocmark" zs.t="SFocus"><div class="zsfocmarki"></div></div>',
					'<div id="', uid, '-highlight" class="zshighlight" zs.t="SHighlight"><div class="zshighlighti"></div></div>');
					
		}
		out.push('</div>');
		out.push('<div id="', uid, '-',type,'wp" class="zswidgetpanel" zs.t="SWidgetpanel"></div>');
		out.push('</div>');
	},
	//ZSS-1043
	addEditorFocus : function(id, name){
		if (this.editorFocusMark && this.editorFocusMark[id])
			return;
		if (this.block) {
			var sheet = this.sheet,
				focusmarkcmp = this.$n('focmark'), 
				x = focusmarkcmp,
				div = x.cloneNode(true);
			
			div.id = div.id + '_' + id;
			div.style.borderWidth = "3px";
			x.parentNode.appendChild(div);
			this.addEditorFocus_(id, name, div);
		}
	},
	//ZSS-1043
	addEditorFocus_ : function(id, name, div) {
		//to be overridden
	},
	//ZSS-1043
	removeEditorFocus : function(id){
		if (this.block) {
			if (!this.editorFocusMark)
				return;
			var ctrl = this.editorFocusMark[id];
			this._removeEditorFocusMark(ctrl);
			this.editorFocusMark[id] = null;
		}
	},
	//ZSS-1043
	_removeEditorFocusMark : function(ctrl) {
		if (ctrl) {
			ctrl.comp.parentNode.removeChild(ctrl.comp);
			ctrl.cleanup();
		}
	},
	//ZSS-1043
	moveEditorFocus : function(id, name, color, row, col){
		if (this.block) {
			if(!this.editorFocusMark || !this.editorFocusMark[id]){
				this.addEditorFocus(id, name);
			}
			this.editorFocusMark[id].relocate(row, col);
			this.editorFocusMark[id].showMark(color, name); //new color/oldcolor with label
		}
	}
});
/**
 * LeftPanel represent the left area of the spreadsheet. It contains row headers of the spreadsheet
 */
zss.LeftPanel = zk.$extends(zss.Panel, {
	widgetName: 'LeftPanel',
	type: 'left',
	initHeaders_: function (sheet, start, end, data, isCorner) {		
		var ary = [],
			headers = data.rowHeaders,
			type = zss.Header.VER,
			leftHeaders = isCorner ? sheet.lp.headers : null,
			j = 0;
		for (var i = start; i <= end; i++) {
			if (!sheet.custRowHeight.isHidden(i)) {
				var h = new zss.Header(sheet, type, headers[i], leftHeaders ? leftHeaders[j++] : null);
				this.appendChild(h, true);
				ary.push(h);
			}
		}
		return ary;
	},
	/**
	 * Create cells and associated headers
	 */
	create_: function (dir, rowStart, rowEnd, frozenColStart, forzenColEnd, createFrozenOnly) {
		if (!createFrozenOnly)
			this.createHeaders_(dir, rowStart, rowEnd);
		
		var createFrozen = frozenColStart >= 0 && forzenColEnd >= 0;
		if ('jump' == dir && createFrozen) {
			var oldBlock = this.block;
			this.block = new zss.CellBlockCtrl(this.sheet, rowStart, frozenColStart, rowEnd, forzenColEnd, this.getFrozenData_(), 'left'); // ZSS-404: fixed missed sheet
			oldBlock ? oldBlock.replaceWidget(this.block) : this.appendChild(this.block);
		} else if (this.block && createFrozen) {
			this.block.create_(dir, rowStart, frozenColStart, rowEnd, forzenColEnd, this.getFrozenData_());
		}
	},
	initFrozenBlock_: function (sheet, tRow, bRow, data) {
		var c = sheet.frozenCol;
		if (c > -1) {
			var b = new zss.CellBlockCtrl(sheet, tRow, 0, bRow, c, data, 'left');
			this.appendChild(b, true);
			return b;
		}
		return null;
	},
	getHeaderData_: function () {
		// ZSS-404: if this is for corner panel's header, the header data must be fetched from top freeze panel
		if(this.isCorner) {
			return this.sheet._wgt._cacheCtrl.getSelectedSheet().topFrozen.rowHeaders;
		} else {
			return this.sheet._wgt._cacheCtrl.getSelectedSheet().rowHeaders;
		}
	},
	getFrozenData_: function () {
		var a = this.sheet._wgt._cacheCtrl.getSelectedSheet(),
			f = a.leftFrozen;
		return f ? f : a;
	},
	getFrozenHeaderData_: function () {
		var leftFrozen = this.getFrozenData_();
		return leftFrozen ? leftFrozen.columnHeaders : null;
	},
	getHeaderType_: function () {
		return zss.Header.VER;
	},
	isInsert_: function (dir) {
		return 'north' == dir; 
	},
	bindFrozenCtrl_: function () {
		var sheet = this.sheet,
			selArea = this.$n('select'),
			selChg = selArea.nextSibling,
			focus = selChg.nextSibling,
			highlight = focus.nextSibling;

		var selRanges = sheet.initparm.selrange,
			clonedSelRanges = [];
		for (var i = 0, l = selRanges.length; i < l; i++) {
			clonedSelRanges.push(selRanges[i].clone());
		}
		this.selArea = new zss.SelAreaCtrlLeft(sheet, selArea, clonedSelRanges);
		this.selChgArea = new zss.SelChgCtrlLeft(sheet, selChg);
		this.focusMark = new zss.FocusMarkCtrlLeft(sheet, focus, sheet.initparm.focus.clone());
		this.hlArea = new zss.HighlightLeft(sheet, highlight, sheet.initparm.hlrange.clone(), "inner");
	},
	//ZSS-1043
	addEditorFocus_ : function(id, name, div) {
		if(!this.editorFocusMark)
			this.editorFocusMark = new Object();
		this.editorFocusMark[id] = new zss.FocusMarkCtrlLeft(this.sheet, div, new zss.Pos(0, 0));
	},
	getTypeAttr_: function () {
		return 'SLeftPanel';
	},
	hasPad_: function () {
		return true;
	},
	removeChildFromStart_: function (size) {
		if (this.block) {
			this.block.removeRowsFromStart_(size);
		}
		this.$supers(zss.LeftPanel, 'removeChildFromStart_', arguments);
	},
	removeChildFromEnd_: function (size) {
		if (this.block) {
			this.block.removeRowsFromEnd_(size);
		}
		this.$supers(zss.LeftPanel, 'removeChildFromEnd_', arguments);
	},
	_updateHeight: function (height) {
		jq(this.comp).css('height', jq.px0(height));
		this.height = height;
		this._updateBlockHeight();
	},	
	_updatePos: function (toppos, leftpos) { // KEIKAI-422
		jq(this.icomp).css('transform',  'translate3d(0, ' + jq.px(toppos) + ', 0)');
		jq(this.sheet.dp.comp).css('transform', 'translate3d(' + jq.px(leftpos) + ', ' + jq.px(toppos)+ ', 0)'); // KEIKAI-7, KEIKAI-422
		jq(this.wpcomp).css('transform', 'translate3d(0, ' + jq.px(toppos) + ', 0)');
		this.toppos = toppos;
		this._updateBlockHeight();
	},
	_updateTopPadding: function (toppad) {
		if (this.toppad == toppad) return;
		jq(this.padcomp).css('height', jq.px0(toppad));
		this.toppad = toppad;	
		
		if (this.selArea)
			this.selArea.relocate();

		if (this.selChgArea)
			this.selChgArea.relocate();

		if (this.focusMark) {
			var pos = this.sheet.getLastFocus();
			this.focusMark.relocate(pos.row, pos.column);
		}
		this._updateBlockHeight();
	},
	_updateBlockHeight: function () {
		if (!this.block) return;
		var height = this.height,
			toppos = this.toppos,
			toppad = this.toppad;
		// ZSS-1281: grey area; side effect of fixing ZSS-1117; revert
		height = height - (toppos ? toppos : 0) - (toppad ? toppad : 0);
//			//ZSS-1117
//			range = this.block.range,
//			top = range.top,
//			bottom = range.bottom;
//		
//		height = this.sheet.custRowHeight.getDiffPixel(top, bottom);//ZSS-1117
		if (height < 0) height = 0;
		if (this.blockHeight === height) {
			return;
		}
		this.blockHeight = height;
		jq(this.block.comp).css('height', jq.px0(height));
	},
	/**
	 * Sets the selection range of the header
	 * @param int from row selection start index
	 * @param int to row selection end index
	 */
	updateSelectionCSS: function (from, to, remove) {
		const headers = this.headers;
		for (i = 0; i < headers.length; ++i) {
			const header = headers[i];
			const index = header.index;
			if (index >= from && index <= to) {
				const $header = jq(header.comp);
				if ($header.hasClass('zsleft-sel')) {
					if (remove) {
						$header.removeClass('zsleft-sel');
					}
				} else {
					if (!remove) {
						$header.addClass('zsleft-sel');
					}
				}
			}
		}
	},
	/**
	 * Sets the selection range of the header
	 * 
	 * Update CSS after header created (this.createHeaders_), ignore update selection when sheet CSS not ready
	 */
	updateSelectionCSS_: function () {
		var sheet = this.sheet,
			wgt = sheet._wgt,
			selRange = sheet.selArea.lastRange;
		if (!wgt.isSheetCSSReady())
			return;
		
		if (selRange) {
			var top = selRange.top,
				bottom = selRange.bottom;
			this.updateSelectionCSS(top, bottom, false || sheet.state == zss.SSheetCtrl.NOFOCUS);
		}
	},
	/**
	 * Sets the width position index
	 * @param int col column index
	 * @param int zsw the width position index
	 */
	appendZSW: function (col, zsw) {
		if (this.block)
			this.block.appendZSW(col, zsw);
	},
	/**
	 * Sets the height position index
	 * @param row row index
	 * @param int the height position index
	 */
	appendZSH: function (row, zsh) {
		if (this.block)
			this.block.appendZSH(row, zsh);
		
		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			return;
		}

		row = this.sheet.activeBlock._getVisibleRowIndex(row - this.headers[0].index);

		if (0 > row || this.headers.length - 1 < row) return;
		
		var header = this.headers[row];
		header.appendZSH(zsh);
	},
	/**
	 * Insert new column
	 * @param int col column
	 * @param int size the size of the column
	 * @param boolean forShow true if the operation is for show column to use
	 */
	insertNewColumn: function (col, size, forShow) {
		if (this.block) {
			if (forShow) {
				this._syncDataBlockRange();
			}
			this.block.insertNewColumn(col, size, forShow);
		}
	},
	_syncDataBlockRange: function () {
		if (this.block) {
			let range = this.block.range,
				dataRange = this.sheet.activeBlock.range,
				frozenRange = this.getFrozenData_().rect;
			range.top = dataRange.top;
			range.bottom = dataRange.bottom;
			range.left = frozenRange.left;
			range.right = frozenRange.right;
			range.reCalcSize();
		}
	},
	/**
	 * Insert new row
	 * @param int row row index
	 * @param int size the size of the row
	 * @param array extnm
	 * @param boolean forShow true if the operation is for show row to use
	 */
	insertNewRow: function (row, size, extnm, forShow) {
		if (this.block) {
			if (forShow) {
				this._syncDataBlockRange();
			}
			this.block.insertNewRow(row, size, this.getFrozenData_(), forShow);
		}

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if (this.isEmptyHeader()) {
			let rowHeaders;
			if (!forShow || !(rowHeaders = this.getHeaderData_())) {
				return;
			}

			// KEIKAI-62: we need to insert new row here
			var sheet = this.sheet,
				selRange = sheet.selArea.lastRange,
				type = zss.Header.VER;

			if (sheet.state != zss.SSheetCtrl.NOFOCUS && selRange) {
				var top = selRange.top,
					bottom = selRange.bottom;
				if (row <= bottom && (row+size-1) >= top)
					this.updateSelectionCSS(top, bottom, true);
			}
			let index = this._getClosestHeaderIndex(row);
			const hs = [];
			for (var i = 0; i < size; i++) {
				var r = row + i;
				const h = new zss.Header(sheet, type, rowHeaders[r] || {});
				hs.push(h)
			}
			this.insertHeaders_(index, hs);
			return;
		}

		var top = this.headers[0].index,
			bottom = this.headers[this.headers.length - 1].index;

		var index = Math.max(row - top, 0);
		// KEIKAI-62: some are invisible
		if (bottom - top + 1 != this.headers.length) {
			index = this._getClosestHeaderIndex(row);
		}

		var ctrl,
			rowHeaders = this.getHeaderData_(),
			type = zss.Header.VER;
		
		//insert row intersect with selection range, must remove selection CSS before insert cells
		var sheet = this.sheet,
			selRange = sheet.selArea.lastRange;
		if (sheet.state != zss.SSheetCtrl.NOFOCUS && selRange) {
			var top = selRange.top,
				bottom = selRange.bottom;
			if (row <= bottom && (row+size-1) >= top)
				this.updateSelectionCSS(top, bottom, true);
		}

		const hs = [];
		for (var i = 0; i < size; i++) {
			var r = row + i;
			const h = new zss.Header(sheet, type, rowHeaders[r] || {});
			hs.push(h);
		}
		this.insertHeaders_(index, hs);

		if (!forShow) {
			extnm = extnm.slice(size, extnm.length);
			this.shiftHeaderInfo(index + size, row + size, extnm);
		}
	},
	/**
	 * Shift header
	 * @param int index start shift index
	 * @param array extnm
	 */
	shiftHeaderInfo: function (index, newrow, extnm, ignoreHidden) {
		var size = this.headers.length,
			j = 0,
			sheet = this.sheet;

		if (index >= size) return;

		let start = this.headers[index].index,
			hiddenCount = ignoreHidden ? 0 : this.sheet.custRowHeight.getHiddenCount(start > newrow ? newrow : start, start > newrow ? start : newrow),
			delta = newrow - start + hiddenCount;
		for (; index < size; index++) {
			if (!extnm[j]) zk.log("undefined header to assing>>" + (newrow + j), "always");
			let header = this.headers[index];
			header.resetInfo(header.index + delta, extnm[header.index - start + hiddenCount]);
		}
	},
	/**
	 * Remove column at the given index
	 * @param int col column index
	 * @param boolean forHidden true if the operation is for hidden column to use
	 */
	removeColumnAt: function (col, forHidden) {
		if (this.block)
			this.block.removeColumnAt(col, forHidden);
	},
	//KEIKAI-117
	/**
	 * Removes columns from lCol to rCol
	 * @param int lCol the left column index
	 * @param int rCol the right column index (inclusive)
	 */
	removeColumnFromRange: function (lCol, rCol) {
		if (this.block)
			this.block.removeColumnFromRange(lCol, rCol);
	},
	//KEIKAI-117
	/**
	 * Remove rows from tRow to bRow
	 * @param tRow the top index
	 * @param bRow the bottom index (inclusive)
	 * @param array extnm
	 */
	removeRowFromRange: function (tRow, bRow, extnm) {
		if (this.block)
			this.block.removeRowFromRange(tRow, bRow);

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			return;
		}

		let custSize = this.sheet.custRowHeight;
		// find first visible lCol
		for (let start = tRow, end = bRow+1; start < end; start++) {
			if (!custSize.isHidden(start)) {
				tRow = start;
				break;
			}
		}

		let index = this.getHeaderIndex(tRow);

		if (index < 0) return;

		let size = custSize.getVisibleCount(tRow, bRow);
		this.removeHeader_(index, size);

		this.shiftHeaderInfo(index, tRow, extnm, true);
	},
	/**
	 * Remove row at the given index
	 * @param int row row index
	 * @param array extnm
	 * @param boolean forHidden true if the operation is for hidden row to use
	 */
	removeRowAt: function (row, extnm, forHidden) {
		if (this.block)
			this.block.removeRowAt(row, forHidden);

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			return;
		}

		let index = this.getHeaderIndex(row);

		if (index < 0) return;

		this.removeHeader_(index, 1);

		if (!forHidden) {
			this.shiftHeaderInfo(index, row, extnm);
		}
	},
	_fixSize: function() {
		var sheet = this.sheet,
			wgt = sheet._wgt,
			lw = sheet.leftWidth,
			leftw = lw-1,
			fzc = sheet.frozenCol,
			name = wgt.getSelectorPrefix(),
			sid = wgt.getSheetCSSId();
		
		if (fzc > -1)
			leftw = leftw + sheet.custColWidth.getStartPixel(fzc + 1);

		zcss.setRule(name + " .zsleft", ["width"], [(fzc > -1 ? leftw - 1 : leftw) + "px"], true, sid);
		zcss.setRule(name + " .zslefti", ["width"], [leftw + "px"], true, sid);
	}
});
})();