/* MergeMatrix.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/
(function () {

	// range to bbox
	RBush.prototype.toBBox = function (range) {
		if (range.minX === undefined) {
			return {minX: range.left,
				minY: range.top,
				maxX: range.right,
				maxY: range.bottom};
		}
		return range;
	};
/**
 * MergeMatrix handle merge on cells
 */
zss.MergeMatrix =  zk.$extends(zk.Object, {
	$init: function (matrix, sheet) {
		this.$supers('$init', arguments);
		this.rtree = new RBush().load(matrix); // array of zss.Range
		this.sheet = sheet;
	},
	getMergeMatrix: function (range) {
		return range ? this.rtree.search({
			minX: range.left,
			minY: range.top,
			maxX: range.right,
			maxY: range.bottom
		}) : this.rtree.all();
	},
	/**
	 * Returns row's merge range
	 * @return array result
	 */
	getRangesByRow: function (row) {
		return this.rtree.search({
			minX: 0,
			minY: row,
			maxX: this.sheet.maxCols,
			maxY: row
		});
	},
	/**
	 * Returns the top connected row
	 * @param int row row index
	 * @param int left column start index
	 * @param int right column end index
	 * @return int row
	 */
	getTopConnectedRow: function (row, left, right) {
		// With the fix of KEIKAI-115, we don't need the following code.

		// var result = this.rtree.search({
		// 	minX: Math.max(this.sheet.frozenCol, left),
		// 	minY: row,
		// 	maxX: right,
		// 	maxY: row
		// });
		//
		// var conti = true;
		// while (conti) {
		// 	conti = false;
		// 	size = result.length;
		// 	for(var i = 0; i < size; i++) {
		// 		range = result[i];
		// 		if (range.top < row && range.bottom >= row) {
		// 			row = range.top;
		// 			conti = true;
		// 			//TODO a fast way to remove range form result array
		// 			break;
		// 		}
		// 	}
		// }
		return row;
	},
	/**
	 * Returns the bottom connected row
	 * @param int row row index
	 * @param int left column start index
	 * @param int right column end index
	 * @return int row 
	 */
	getBottomConnectedRow: function (row, left, right) {
		// With the fix of KEIKAI-115, we don't need the following code.

		// var result = this.rtree.search({
		// 	minX: Math.max(this.sheet.frozenCol, left),
		// 	minY: row,
		// 	maxX: right,
		// 	maxY: row
		// });
		//
		// var conti = true;
		// while (conti) {
		// 	conti = false;
		// 	size = result.length;
		// 	for (var i = 0; i < size; i++) {
		// 		range = result[i];
		// 		if (range.bottom > row && range.top <= row) {
		// 			row = range.bottom;
		// 			conti = true;
		// 			//TODO a fast way to remove range from result array
		// 			break;
		// 		}
		// 	}
		// }
		return row;
	},
	/**
	 * Returns column's merge range
	 * @return array result
	 */
	getRangesByColumn: function (col) {
		return this.rtree.search({
			minX: col,
			minY: 0,
			maxX: col,
			maxY: this.sheet.maxRows
		});
	},
	/**
	 * Returns the left connected column
	 * @param int col column index
	 * @param int top row start index
	 * @param int bottom row end index
	 * @return int col
	 */
	getLeftConnectedColumn: function (col, top, bottom) {
		// With the fix of KEIKAI-115, we don't need the following code.

		// var result = this.rtree.search({
		// 	minX: col,
		// 	minY: Math.max(this.sheet.frozenRow, top),
		// 	maxX: col,
		// 	maxY: bottom
		// });
		//
		// var conti = true;
		// while (conti) {
		// 	conti = false;
		// 	size = result.length;
		// 	for(var i = 0; i < size; i++) {
		// 		range = result[i];
		// 		if (range.left < col && range.right >= col) {
		// 			col = range.left;
		// 			conti = true;
		// 			//TODO a fast way to remove range form result array
		// 			break;
		// 		}
		// 	}
		// }
		return col;
	},
	/**
	 * Returns the right connected column
	 * @param int col column index
	 * @param int top row start index
	 * @param int bottom row end index
	 * @return int col 
	 */
	getRightConnectedColumn: function (col, top, bottom) {
		// With the fix of KEIKAI-115, we don't need the following code.
		
		// var result = this.rtree.search({
		// 	minX: col,
		// 	minY: Math.max(this.sheet.frozenRow, top),
		// 	maxX: col,
		// 	maxY: bottom
		// });
		//
		// var conti = true;
		// while (conti) {
		// 	conti = false;
		// 	size = result.length;
		// 	for (var i = 0; i < size; i++) {
		// 		range = result[i];
		// 		if (range.right > col && range.left <= col) {
		// 			col = range.right;
		// 			conti = true;
		// 			//TODO a fast way to remove range form result array
		// 			break;
		// 		}
		// 	}
		// }
		return col;
	},
	/**
	 * Sets the merge range
	 * @param string id
	 * @param int left column start index
	 * @param int top row start index
	 * @param int right column end index
	 * @param int bottom row end index
	 */
	addMergeRange: function(id, left, top, right, bottom) {
		var range = new zss.Range(left, top, right, bottom);
		range.id = id;
		this.rtree.insert(range);
	},
	/**
	 * Remove merge range
	 * @param range
	 */
	removeMergeRange: function (range) {
		this.rtree.remove(range, function (a, b) {
			return a.id === b.id;
		});
	},
	//ZSS-927
	/**
	 * Returns the merged top row which overlaps the specified row. 
	 */
	getTopRow: function (row) {
		var mm = this.getRangesByRow(row),
			size = mm.length,
			top = row,
			range;
		for (var i = 0; i < size; i++) {
			range = mm[i];
			if (range.top <= row && range.bottom >= row && range.top < top)
				top = range.top;
		}
		return top;
	},
	//ZSS-927
	/**
	 * Returns the merged left row which overlaps the specified column.
	 */
	getLeftCol: function (col) {
		var mm = this.getRangesByColumn(col),
			size = mm.length,
			left = col,
			range;
		for (var i = 0; i < size; i++) {
			range = mm[i];
			if (range.left <= col && range.right >= col && range.left < left)
				left = range.left;
		}
		return left;
	},
	//ZSS-1168
	/**
	 * Remove All merge range and execute fn per merge range
	 */
	removeAllMergeRange: function (owner, fn) {
		if (fn) {
			var mm = this.rtree.all(),
				size = mm.length;
			for (var i = 0; i < size; i++) {
				var range = mm[i];
				fn.call(owner, range);
			}
		}
		this.rtree.clear();
	}
});

})();