/* PositionHelper.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/

(function () {
	
	function updateSize(customizedSizeInfo, size, custom) {
		customizedSizeInfo[1] = size;
		customizedSizeInfo[4] = custom;
	}

	const isInteger = Number.isInteger || (function (v) {return typeof v === 'number' && isFinite(v) && Math.floor(v) === v});

	function mostSignificantBit(value) {
		let result = value;
		result |= result >> 1;
		result |= result >> 2;
		result |= result >> 4;
		result |= result >> 8;
		result |= result >> 16;
		result |= result >> 32;
		result -= result >> 1;
		return result;
	}

	// The BIT implementation refers to https://github.com/microsoft/fast-binary-indexed-tree-js under MIT license
	/**
	 * Class BinaryIndexedTree
	 */
	zss.BinaryIndexedTree = zk.$extends(zk.Object, {
		$init: function (defaultFrequency, maxVal) {
			this.defaultFrequency = defaultFrequency || 0;
			this.maxVal = maxVal;
			this._tree = { 0: 0 };
			this._msb = mostSignificantBit(maxVal);
			this._countNeg = defaultFrequency < 0 ? maxVal : 0;
		},
		_getTree: function(index) {
			if (index in this._tree) {
				return this._tree[index];
			}
			return this.defaultFrequency * (index & -index);
		},
		_addTree: function(index, delta) {
			this._tree[index] = this._getTree(index) + delta;
		},
		_checkIndex: function(index) {
			if (!isInteger(index)) {
				throw new Error('Invalid index');
			}
			if (index < 0 || index >= this.maxVal) {
				throw new Error('Index out of range');
			}
		},
		_readSingle: function(idx) {
			// index out of range
			if (idx >= this.maxVal) return this.defaultFrequency;

			let index = idx + 1;
			let sum = this._getTree(index);
			let z = index - (index & -index);

			index--;

			while (index !== z) {
				sum -= this._getTree(index);
				index -= index & -index;
			}

			return sum;
		},
		_changed: function(freqCur, freqNew) {
			if (freqCur < 0 && freqNew >= 0) {
				this._countNeg --;
			} else if (freqCur >= 0 && freqNew < 0) {
				this._countNeg ++;
			}
		},
		_update: function(idx, delta) {
			let index = idx + 1;

			while (index <= this.maxVal) {
				this._addTree(index, delta);
				index += (index & -index);
			}
		},
		_writeSingle: function(idx, freq) {
			const freqCur = this._readSingle(idx);

			this._update(idx, freq - freqCur);
			this._changed(freqCur, freq);
		},
		_read: function(count) {
			let index = count;
			let sum = 0;
			while (index) {
				sum += this._getTree(index);
				index -= index & -index;
			}

			return sum;
		},
		_find: function(sum, before) {
			let left = 0;
			let right = this._msb << 1;
			let sumT = sum;

			while (right > left + 1) {
				let middle = (left + right) >> 1;
				const sumM = this._getTree(middle);

				if (middle <= this.maxVal && before(sumM, sumT)) {
					sumT -= sumM;
					left = middle;
				} else {
					right = middle;
				}
			}
			return left;
		},

		/**
		 * Read a single frequency
		 * @param {number} idx - The 0 based index for the frequency
		 * @return {number}
		 */
		readSingle: function(idx) {
			return this._readSingle(idx);
		},

		/**
		 * Update a single frequency with a delta value
		 * @param {number} idx - The 0 based index for the frequency
		 * @param {number} delta - The delta value of the freqency
		 * @return {undefined}
		 */
		update: function(idx, delta) {
			this._checkIndex(idx);
			const freqCur = this._readSingle(idx);

			this._update(idx, delta);
			this._changed(freqCur, freqCur + delta);
		},
		/**
		 * Update a single frequency with a given value
		 * @param {number} idx - The 0 based index for the frequency
		 * @param {number} freq - The new frequency
		 * @return {undefined}
		 */
		writeSingle: function(idx, freq) {
			this._checkIndex(idx);
			this._writeSingle(idx, freq);
		},
		/**
		 * Read the sum of the first `count` frequencies
		 * @param {number} count - The count of frequencies to accumulate
		 * @return {number}
		 */
		read: function(count) {
			if (!isInteger(count)) {
				throw new Error('Invalid count');
			}
			return this._read(Math.max(Math.min(count, this.maxVal), 0));
		},
		/**
		 * Read the lower-bound with the given cumulated frequency
		 * *REQUIRE ALL FREQUENCIES TO BE NON-NEGATIVE*
		 * @param {number} sum - The cumulated frequency
		 * @return {number}
		 */
		lowerBound: function(sum) {
			if (this._countNeg > 0) {
				throw new Error('Sequence is not non-descending');
			}
			return this._find(sum, function (x, y) { return  x < y;});
		},
		/**
		 * Read the upper-bound with the given cumulated frequency
		 * *REQUIRE ALL FREQUENCIES TO BE NON-NEGATIVE*
		 * @param {number} sum - The cumulated frequency
		 * @return {number}
		 */
		upperBound: function(sum) {
			if (this._countNeg > 0) {
				throw new Error('Sequence is not non-descending');
			}
			return this._find(sum, function (x, y) { return x <= y;});
		}
	});

/**
 * PositionHelper for calculate pixel to column/row index or vice versa
 */
zss.PositionHelper = zk.$extends(zk.Object, {
	$init: function (defaultSize, customizedSize, maxColRow, isRowDirection) {
		this.$supers('$init', arguments);
		this.size = defaultSize;
		this.custom = customizedSize;
		this.maxColRow = maxColRow;
		this.MIN_COL_WIDTH_ROW_HEIGHT = isRowDirection ?
			0 : // KEIKAI-600, KEIKAI-624: row height isn't constrained by the 5px that column width is
			5; // KEIKAI-456: 2 * 2 padding size + 1 border
		this._rebuildBIT();
	},
	cleanup: function () {
		this.size = this.custom = null;
	},
	updateDefaultSize: function (defaultSize) {
		this.size = defaultSize;
	},
	getIncUnhidden: function (idx, limit) {
		var customizedSize = this.custom,
			size = customizedSize.length,
			v0,
			v1;
		for(var j = 0; j < size && idx <= limit; ++j) {
			v0 = customizedSize[j][0];
			if (idx == v0) {
				v1 = customizedSize[j][3]; //hidden: true
				if (!v1) {//special size but not hidden, return
					break;
				} else {
					++idx; //a hidden one, try next unhidden
				}
			} else if (idx < v0) { //normal case, return
				break; 
			}
		}
		return idx > limit ? -1 : idx;
	},
	getDecUnhidden: function (idx, limit) {
		var customizedSize = this.custom,
			size = customizedSize.length,
			v0,
			v1;
		for(var j = size - 1; j >= 0 && idx >= limit; --j) {
			v0 = customizedSize[j][0];
			if (idx == v0) {
				v1 = customizedSize[j][3]; //hidden: true
				if (!v1) {//special size but not hidden, return
					break;
				} else {
					--idx; //a hidden one, try next unhidden
				}
			} else if (idx > v0) { //normal case, return
				break; 
			}
		}
		return idx < limit ? -1 : idx;
	},
	/**
	 * get cell index from a pixel
	 * @param int px
	 */
	getCellIndex: function (px, firstHidden) { // ZSS-1266
		if (px < 0) px = 0; //col1/row1 can be hidden...
		var sum = 0,
			sum2 = 0,
			index = 0,
			inc = 0,
			customizedSize = this.custom,
			defaultSize = this.size,
			size = customizedSize.length,
			i = 0,
			v0,
			v1;
		for (;i < size; i++) {
			v0 = customizedSize[i][0];
			v1 = customizedSize[i][3] ? 0 : customizedSize[i][1]; //hidden, then size is zero
			sum2 = sum2 + (v0 -index ) * defaultSize;
			if (sum2 > px) {
				inc = px - sum;
				var incx = Math.floor(inc / defaultSize),
					cpx = inc - incx * defaultSize;
				index = index + incx;
				return [firstHidden ? this.getFirstHiddenIndex(i, index) : index, cpx]; //ZSS-1266
			}
			sum2 = sum2 + v1;
			if (sum2 > px) {
				var cpx = px - sum2 + v1;
				return [firstHidden ? this.getFirstHiddenIndex(i, v0) : v0, cpx]; //ZSS-1266
			}
			sum = sum2;
			index = v0+1;
		}
		inc = px - sum; 
		var incx = Math.floor(inc / defaultSize),
			cpx = inc - incx * defaultSize;
		index = index + incx;
		return [firstHidden ? this.getFirstHiddenIndex(i, index) : index, cpx]; //ZSS-1266
	},
	//ZSS-1266
	// find backward the 1st consecutive hidden row/col index 
	getFirstHiddenIndex: function (i, index) {
		while (--i >= 0 && this.custom[i][3] && this.custom[i][0] == index - 1) {
			// locate index
			--index;
		}
		return index;
	},
	/**
	 * Returns start pixel(left or top) of a cell, 
	 * @param int cellIndex
	 * @return int
	 */
	getStartPixel: function (cellIndex) {
		if (cellIndex < 0) return 0;
		return this._getStartPixel0(cellIndex)[0];
	},
	//ZSS-1117: return [start-pixel-of-cellIndex, new start metaIndex, new start pixel for the new start metaIndex]
	// + cellIndex to find its startPixel
	// + sidx - starting metaIndex; optional; default 0
	// + spx - starting pixel for the starting metaIndex; optional; default 0
	_getStartPixel0: function (cellIndex, sidx, spx) {
		var count = sidx || 0,
			sum = spx || 0,
			customizedSize = this.custom,
			defaultSize = this.size,
			size = customizedSize.length;

		for (var i = count; i < size; i++) {
			if (cellIndex <= customizedSize[i][0])
				break;

			count++;
			sum += customizedSize[i][3] ? 0 : Math.max(customizedSize[i][1], this.MIN_COL_WIDTH_ROW_HEIGHT); // KEIKAI-456, KEIKAI-600, KEIKAI-624
		}
		var spx0 = sum + (count > 0 ? (customizedSize[count-1][0] + 1 - count) : 0) * defaultSize; 
		sum = sum + (cellIndex - count) * defaultSize;
		return [sum, count, spx0];
	},
	//ZSS-1117
	getDiffPixel: function (startIdx, endIdx) {
		var sinfo = this._getStartPixel0(startIdx),
			einfo = this._getStartPixel0(endIdx+1, sinfo[1], sinfo[2]);
		return einfo[0] - sinfo[0];
	},
	//ZSS-1117
	getMetaIndex: function (cellIdx, min, max) {
		var customizedSize = this.custom,
			defaultSize = this.size,
			size = customizedSize.length,
			imin = min || 0,
			imax = max || size-1;
		while (imin <= imax) {
			var imid = Math.floor((imax + imin) / 2),
				cidx = customizedSize[imid][0];
			if ( cidx == cellIdx) {
				return imid;
			} else if (cidx < cellIdx) {
				imin = imid + 1; 
			} else {
				imax = imid - 1;
			}
		}
		return -imid - 1; // not found
	},
	/**
	 * Shift Meta 
	 * @param int cellIndex
	 * @param int size
	 */
	shiftMeta: function (cellIndex, size) {
		let idx = this._binarySearch(cellIndex);

		if (idx < 0) {
			idx = -idx - 1;
		}

		var customizedSize = this.custom,
			len = customizedSize.length;
		for (var i = idx; i < len; i++) {
			customizedSize[i][0] += size;
		}
		this._rebuildBIT();
	},
	/**
	 * Unshift Meta
	 * @param int cellIndex
	 * @param int size
	 */
	unshiftMeta: function (cellIndex, size) {
		var customizedSize = this.custom,
			len = customizedSize.length,
			remove = [],
			end = cellIndex + size,
			index,
			i = this._binarySearch(cellIndex);
		if (i < 0) {
			i = -i - 1;
		}
		for (; i < len; i++) {
			index = customizedSize[i][0];
			if (index >= cellIndex && index < end) {
				remove.push(i);
			} else if (index >= end) {
				customizedSize[i][0] -= size;	
			}
		}
		end = remove.length;
		for (let i = end - 1; i >= 0; i--) {
			this._remove(remove[i]);
		}
		this._rebuildBIT();
	},
	/**
	 * Returns whether cell's size is default or not
	 * 
	 * @param int index
	 * @return boolean indicate whether is default size or not
	 */
	isDefaultSize: function (index) {
		var defaultSize = this.size,
			size = this.getSize(index),
			dif = Math.abs(size - defaultSize);
		if (Math.abs(size - defaultSize) > 1) { //default height tolerate range
			return false;
		}
		return true;
	},
	/**
	 * Returns the size of the cell
	 * @param int cellIndex
	 * @return int size
	 */
	getSize: function (cellIndex) {
		let idx = this._binarySearch(cellIndex);
		if (idx >= 0) {
			return this.custom[idx][1];
		}
		return this.size;
	},
	_binarySearch: function (index) {
		let low = 0,
			customizedSize = this.custom,
			high = customizedSize.length - 1;
		while (low <= high) {
			let mid = (low + high) >> 1;
			let custSize = customizedSize[mid][0];
			if (custSize < index) {
				low = mid + 1;
			} else if (custSize > index) {
				high = mid - 1;
			} else {
				return mid;
			}
		}
		return -(low + 1);
	},
	/**
	 * Returns Meta
	 * @param int cellIndex
	 * @return array meta
	 */
	getMeta: function (cellIndex) {
		let idx = this._binarySearch(cellIndex);
		if (idx >= 0) {
			return this.custom[idx];
		}
		return null;
	},
	/**
	 * Sets the customized size of the cell
	 * @param int cellIndex
	 * @param int size
	 * @param string id
	 * @param boolean hidden
	 * @param boolean custom
	 */
	setCustomizedSize: function (cellIndex, size, id, hidden, custom) {
		var customizedSize = this.custom,
			defaultSize = this.size,
			s = 0,
			e = customizedSize.length;
		if (e == 0) {
			this._insert(0, cellIndex, size, id, hidden, custom);
			this._updateBIT(cellIndex, hidden); // after insertion.
			return;
		}
		var i;
		while (s < e) {//binary search
			i = s + Math.floor((e - s) / 2);
			if (customizedSize[i][0] == cellIndex) {
				customizedSize[i][2] = id;
				customizedSize[i][3] = hidden;
				updateSize(customizedSize[i], size, custom);
				this._updateBIT(cellIndex, hidden);
				return;
			} else if (customizedSize[i][0] > cellIndex) {
				e = i - 1;
			} else {
				s = i + 1;
			}
			if (e == s) {
				if (e >= customizedSize.length || customizedSize[e][0] > cellIndex) {
					this._insert(e, cellIndex, size, id, hidden, custom);
					this._updateBIT(cellIndex, hidden); // after insertion.
				} else if (customizedSize[e][0] == cellIndex) {
					customizedSize[e][2] = id;
					customizedSize[e][3] = hidden;
					updateSize(customizedSize[e], size, custom);
					this._updateBIT(cellIndex, hidden);
				} else {
					this._insert(e + 1, cellIndex, size, id, hidden, custom);
					this._updateBIT(cellIndex, hidden); // after insertion.
				}
				break;
			} else if (e < s) {
				this._insert(s, cellIndex, size, id, hidden, custom);
				this._updateBIT(cellIndex, hidden); // after insertion.
			}
		}
	},
	/**
	 * Remove the customized size of the cell
	 * @param int cellIndex
	 */
	removeCustomizedSize: function (cellIndex) {
		var customizedSize = this.custom,
			s = 0,
			e = customizedSize.length;
		if (e == 0)
			return;

		var i;
		while (s < e){//binary search
			i = s + Math.floor((e - s) / 2);
			if (customizedSize[i][0] == cellIndex) {
				if (customizedSize[i][3]) {
					this._bit.writeSingle(cellIndex, 0);
				}
				this._remove(i);
				return;
			} else if (customizedSize[i][0] > cellIndex) {
				e = i - 1;
			} else {
				s = i + 1;
			}
			if (e == s) {
				if (e < customizedSize.length&&customizedSize[e][0] == cellIndex) {
					if (customizedSize[e][3]) {
						this._bit.writeSingle(cellIndex, 0);
					}
					this._remove(e);
				}
				break;
			} else if (e < s) {
				break;
			}
		}
	},
	_remove: function (index) {
		if (index > -1) {
			this.custom.splice(index, 1);
		}
	},
	_insert: function(index, cellIndex, size, id, hidden, custom) {
		let obj = [cellIndex, size, id, hidden, custom];
		this.custom.splice(index, 0, obj);
	},
	/**
	 * An custom size is set manually users. The size determined by Spreadsheet automatically for some features, i.e. wrap text, 
	 * is not custom.
	 * @return boolean indicate whether the size is determined automatically or not
	 */
	isCustomSize: function(rowIndex){
		var customSize = this.getMeta(rowIndex);
		if (customSize == null){ //no custom info, default size
			return false;
		}else{
			return customSize[4];
		}
	},
	getDefaultSize: function(){
		return this.size;
	},
	/**
	 * @param rowColumnIndex row or column index
	 * @return return "hidden" value in position helper of specified row (or column).
	 * If no position helper exists, return FALSE.  
	 */
	isHidden: function(rowColumnIndex){
		return this._hasCustom() ? this._isZeroSize(rowColumnIndex): false;
	},
    // KEIKAI-354
    _isZeroSize: function (rowColumnIndex) {
	    let custom = this.getMeta(rowColumnIndex);
	    return custom && (custom[1] === 0 || custom[3]);
    },
	_hasCustom: function () {
		return this.custom && this.custom.length;
	},
	//KEIKAI-62
	_initBIT: function () {
		if (this._hasCustom()) {
			var cs = this.custom;
			for (let i = 0, j = cs.length; i < j; i++) {
				let meta = cs[i];

				// only write for hidden rows
				if (meta && meta[3]) {
					this._bit.writeSingle(meta[0], 1);
				}
			}
		}
	},
	_rebuildBIT: function () {
		let size = 1000;
        const doubleSize = this._bit ? this._bit.maxVal * 2 : 0;
        if (this._hasCustom()) {
            const customMax = this.custom[this.custom.length - 1][0] + 1;
            if (customMax < this.maxColRow && this.maxColRow < doubleSize) {
                size = this.maxColRow;
            } else if (customMax > doubleSize) {
                size = customMax;
            } else {
                size = doubleSize;
            }
        }
		this._bit = new zss.BinaryIndexedTree(0, size); // re-create;
		this._initBIT();
	},
	_updateBIT: function (cellIndex, hidden) {
		if (this._bit.maxVal > cellIndex) {
			this._bit.writeSingle(cellIndex, hidden ? 1 : 0);
		} else {
			this._rebuildBIT();
		}
	},
	updateMaxCol: function(col) {
		if (this._bit.maxVal < col) {
			this.maxColRow = col;
			this._rebuildBIT();
		}
	},
	updateMaxRow: function(row) {
		if (this._bit.maxVal < row) {
			this.maxColRow = row;
			this._rebuildBIT();
		}
	},
	getHiddenCount: function (start, end) {
		return this._bit.read(end + 1) - this._bit.read(start);
	},
	getVisibleCount: function (start, end) {
		if (this._hasCustom()) {
			let total = end - start + 1;
			return total - this.getHiddenCount(start, end);
		}
		return end - start + 1;
	}
});
})();