/* ScrollInfo.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/

/**
 * ScrollInfo show current row/column information after scrolling
 */
zss.ScrollInfo = zk.$extends(zss.Info, {
	$init: function (sheet, cmp) {
		this.$supers('$init', arguments);
		this.horizontal = true;
		this.x = this.y = -1;
	},
	/**
	 * Show information by direction
	 * @param boolean horizontal
	 */
	showInfoOnDir: function (horizontal) {
		if (!this.enabled) return;
		const sheet = this.sheet;
		const sp = sheet.sp;
		const spcmp = sp.comp;
		
		if (horizontal) {
			const custColWidth = sheet.custColWidth;
			// KEIKAI-200: reduce reflow
			let col = sp.currentLeft || spcmp.scrollLeft;
			
			if (sheet.frozenCol > -1)
				col += custColWidth.getStartPixel(sheet.frozenCol + 1);
			
			col = custColWidth.getCellIndex(col)[0] + 1;
			this.setInfoText("Column : " + col);
			jq(this.comp).css('text-align', 'left');
		} else {
			const custRowHeight = sheet.custRowHeight;
			// KEIKAI-200: reduce reflow
			let row = sp.currentTop || spcmp.scrollTop;
			if (sheet.frozenRow > -1)
				row += custRowHeight.getStartPixel(sheet.frozenRow + 1);

			row = custRowHeight.getCellIndex(row)[0] + 1;
			this.setInfoText("Row : " + row);
			jq(this.comp).css('text-align', 'right');
		}
		this.showInfo(true);
	},
	/** 
	 * Pin xy point and enable
	 * @param int x
	 * @param int y
	 * @param boolean horizontal
	 */
	pinXY: function(x, y, horizontal) {
		this.x = x;
		this.y = y;
		this.horizontal = horizontal;
		this.enabled = true;

		if (this.visible) {
			this.pinLocation(horizontal);
			this.showInfo(horizontal);
		}
	},
	/**
	 * Pin location
	 * @param boolean horizontal
	 */
	pinLocation: function (horizontal) {
		if (!this.enabled) return ;
		var sheet = this.sheet,
			spcmp = sheet.sp.comp;
		if (this.horizontal != horizontal) {
			this.x = -1;
			this.y = -1;
		}
		this.horizontal = horizontal;

		if (!this.visible)
			jq(this.comp).css({'left': '-90000px', 'display': 'inline'});//show in non-visible for correct size

		if (horizontal) {
			const infox = (this.x == -1) ? sheet.leftWidth + 1 : this.x;
			// KEIKAI-200
			jq(this.comp).css({'left': jq.px(infox), 'right': 'auto', 'top': 'auto', 'bottom': jq.px(1)});
		} else {
			// KEIKAI-200
			const infoy = (this.y == -1) ? sheet.topHeight + 1 : this.y;
			jq(this.comp).css({'left': 'auto', 'right': jq.px(1), 'top': jq.px(infoy), 'bottom': 'auto'});
		}
	},
	showInfo: function (autohide) {
		if (!this.enabled) return ;
		this.lastautohide = autohide;
		this.visible = true;

		if (autohide) {
			var local = this;
			this.showcount++;
			setTimeout(function () {
				local.showcount -- ;
				if (local.showcount <= 0) {
					setTimeout(function () {
						if (local.showcount > 0) return;
						if (!local.lastautohide) return;
						local.hideInfo();
					}, local.displaytime);
				}
			},300);
		}
	},
});