/* ScrollPanel.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/

/**
 * ScrollPanel is used to handle spreadsheet scroll moving event,  
 */
zss.ScrollPanel = zk.$extends(zk.Object, {
	_scrollbar: null,
	/**
	 * Scroll direction
	 */
	dir: 'south',
	$init: function (sheet) {
		this.$supers('$init', []);
		var wgt = sheet._wgt,
			scrollPanel = sheet.$n('sp'),
			self = this;
		
		this.id = scrollPanel.id;
		this.sheet = sheet;
		this.comp = scrollPanel;
		this.currentTop = this.currentLeft = this.timerCount = 0;
		this.scrollTop = 0;
		this.scrollLeft = 0;
		this.leftWidth = 34;//left head width
		this.topHeight = 21;//top head height
		this.timerRunning = false;//is timer for scroll running.
		this.lastMove = "";//the last move of scrolling,
		scrollPanel.ctrl = this;
		
		var dtcmp = this.sheet.dp.comp,//zkSDatapanelCtrl._currcmp(self);
			sccmp = this.comp;
		this.minLeft = this._getMaxScrollLeft(dtcmp, sccmp);
		this.minTop = this._getMaxScrollTop(dtcmp, sccmp);
		this.minHeight = dtcmp.offsetHeight;
		this.minWidth = dtcmp.offsetWidth;

		var proxyMWFn = this.proxy(this._doMouseWheel);

		// KEIKAI-200: throttle onScroll event
		const proxyDoScrolling = this.proxy(this._doScrolling);
		this.ticking = false;
		const proxyRequestTick = this.proxy(function () {
            const sccmp = this.$scroll();
            this.scrollLeft = sccmp.scrollLeft;
            this.scrollTop = sccmp.scrollTop;
			if (!this.ticking) {
				window.requestAnimationFrame(proxyDoScrolling);
				this.ticking = true;
			}
		});
		this._onScroll = function (evt) {
			proxyRequestTick();
		};

		this.bindFrozenPanels();
		let jqScrollPanelFix = jq(sheet.$n('spfix'));
		jqScrollPanelFix.on('mousewheel', proxyMWFn);
		if (zk.mobile) {
			jqScrollPanelFix.on('touchstart', this.proxy(this._doTouchStart));
			jqScrollPanelFix.on('touchmove', proxyMWFn);
			jqScrollPanelFix.on('touchend', this.proxy(this._doTouchEnd));
		}

		if (zk.ie == undefined && !zk.mobile) {
			var self = this;
			zk.afterMount(function() {

				// add before init SimpleBar
				jq(sccmp).addClass('zsscroll-faker');

				SimpleBar.globalObserver.disconnect(); // KEIKAI-532: clear observations
				self._scrollbar = new SimpleBar(scrollPanel, {
					autoHide: false,
					scrollbarMinSize: 100
				});
				if (sheet._initiated) {
					self.sheet._fixSize();
					self._initScrollWidth();
				} else {
					//set cell focus after CSS ready
					sheet.addSSInitLater(function () {
						self.sheet._fixSize();
						self._initScrollWidth();
					});
				}
				var elem = self.$scroll();
				wgt.domListen_(elem, 'onScroll', self._onScroll);
			});
		} else {
			wgt.domListen_(scrollPanel, 'onScroll', this._onScroll);

		}

        // KEIKAI-358: width and height could be used in loadForVisible before onSize
		sheet.addSSInitLater(function() {
			self.width = scrollPanel.clientWidth;
			self.height = scrollPanel.clientHeight;
		});
	},
	_initScrollWidth: function () {
		if (!zss.Spreadsheet.scrollWidth) {
			var clientRects = this._scrollbar.axis.x.scrollbar.el.parentElement.getClientRects();
			if (clientRects.length > 0) { // is visible
				zss.Spreadsheet.scrollWidth = clientRects[0].height;
			}
		}
	},
	// returns the scroll element.
	$scroll: function () {
		return this._scrollbar ? this._scrollbar.getScrollElement() : this.comp;
	},
	hasHScroll: function () {
		if (this._scrollbar) {
			if (this._scrollbar.recalculate()) {
				return true;
			} else if (this._scrollbar.axis.x.isOverflowing) {
				return true;
			} else {
				const dp = this.sheet.dp;
				const dpWidth = dp.width;
				const width = this.width;
				if (dpWidth != undefined && width != undefined) {
					return dpWidth > width;
				} else {
					return dp.comp.clientWidth > this.comp.clientWidth;
				}
			}
		} else {
			return zk(this.comp).hasHScroll();
		}
	},
	hasVScroll: function () {
		if (this._scrollbar) {
			if (this._scrollbar.recalculate()) {
				return true;
			} else if (this._scrollbar.axis.y.isOverflowing) {
				return true;
			} else {
				const dp = this.sheet.dp;
				const dpHeight = dp.height;
				const height = this.height;
				if (dpHeight != undefined && height != undefined) {
					return dpHeight > height;
				} else {
					return dp.comp.clientHeight > this.comp.clientHeight;
				}
			}
		} else {
			return zk(this.comp).hasVScroll();
		}
	},
	/**
	 * Returns the direction that scroll to
	 * 
	 * <ul>
	 * 	<li>north</li>
	 * 	<li>west</li>
	 * 	<li>east</li>
	 * 	<li>south</li>
	 * </ul>
	 * 
	 * @return string direction
	 */
	getDirection: function () {
		return this.dir;
	},
	reset: function (top, left) {
		// doScrolling will update currentTop and currentLeft
		// this.currentTop = top;
		// this.currentLeft = left;
		var n = this.$scroll();
		n.scrollLeft = left;
		n.scrollTop = top;
		this.scrollLeft = left;
		this.scrollTop = top;
	},
	cleanup: function () {
		var sheet = this.sheet,
			wgt = sheet._wgt,
			n = this.$scroll(),
			proxyMWFn = this.proxy(this._doMouseWheel);
		this.unbindFrozenPanels();
		jq(sheet.$n('spfix')).off('mousewheel', proxyMWFn);

		wgt.domUnlisten_(n, 'onScroll', this._onScroll);

		this.invalid = true;
		if(this.comp) this.comp.ctrl = null;
		this._scrollbar = this.comp = this.sheet = null;

	},
	unbindFrozenPanels: function () {
		var jqFrozenPanels = [jq(this.sheet.cp.$n()), jq(this.sheet.lp.$n()), jq(this.sheet.tp.$n())],
			panelsLength = jqFrozenPanels.length,
			proxyMWFn = this.proxy(this._doMouseWheel);
		if (zk.mobile) {
			var proxyTSFn = this.proxy(this._doTouchStart),
				proxyTMFn = this.proxy(this._doTouchEnd);
			for (var i = 0; i < panelsLength; i++) {
				jqFrozenPanels[i].off('touchend', proxyTSFn).off('touchmove', proxyMWFn).off('touchstart', proxyTMFn);
			}
		}
		for (var i = 0; i < panelsLength; i++) {
			jqFrozenPanels[i].off('mousewheel', proxyMWFn);
		}
	},
	bindFrozenPanels: function () {
		var jqFrozenPanels = [jq(this.sheet.tp.$n()), jq(this.sheet.lp.$n()), jq(this.sheet.cp.$n())],
			panelsLength = jqFrozenPanels.length,
			proxyMWFn = this.proxy(this._doMouseWheel);
		for (var i = 0; i < panelsLength; i++) {
			jqFrozenPanels[i].on('mousewheel', proxyMWFn);
		}
		if (zk.mobile) {
			var proxyTSFn = this.proxy(this._doTouchStart),
				proxyTMFn = this.proxy(this._doTouchEnd);
			for (var i = 0; i < panelsLength; i++) {
				jqFrozenPanels[i].on('touchstart', proxyTSFn).on('touchmove', proxyMWFn).on('touchend', proxyTMFn);
			}
		}
	},
	//ZSS-996
	_doMouseWheel: function (evt) {
		var event = evt.originalEvent || evt /*fix test case simulate mouse wheel event*/,
			touches = event.touches,
			touch0,
			x,
			y,
			deltaX = event.deltaX,
			deltaY = event.deltaY,
			scrollPanel = this.$scroll();

		if (touches) {
			//disable swipe when target is zswidget
//			if (jq(event.target).parents('.zswidget').length > 0) {
//				return;
//			}
			touch0 = touches[0];
			x = touch0.pageX;
			y = touch0.pageY;
			let currentTouch = this._currentTouch;
			if (currentTouch == undefined) {
				return;
			}
			deltaX = event.deltaX || currentTouch.x - x;
			deltaY = event.deltaY || currentTouch.y - y;
		}
		this._currentTouch = {x: x, y: y};
		// IE doesn't have event.deltaY and deltaX
		if (deltaX === undefined) {
			deltaX = event.wheelDeltaX || 0;
			if (deltaX) {
				deltaY /= -1.2;
			}
		}
		if (deltaY === undefined) {
			deltaY = event.wheelDeltaY ? event.wheelDeltaY / 120 : event.wheelDelta ? event.wheelDelta / 120 : -event.detail / 3;
			if (deltaY) {
				deltaY *= -100;
			} else {
				deltaY = 0;
			}
		}
		if (event.deltaMode === 1) {
			deltaX *= 10;
			deltaY *= 10;
		}

		if(deltaY && deltaY != 0)
			scrollPanel.scrollTop += deltaY;
		
		// ff & IE doesn't support horizontal mouse wheel
		if(deltaX && deltaX != 0)
			scrollPanel.scrollLeft += deltaX;
	},
	_doMousedown: function (evt) {
		var cmp = this.comp;
		if(evt.domTarget != cmp) return;
		var data = zkS._getMouseData(evt, cmp);
		
		//calculate is xy in scroll bar
		var clickInHor = (data[1] < zk(cmp).offsetHeight() && data[1] > cmp.clientHeight);
			clickInVer = (data[0] < zk(cmp).offsetWidth() && data[0] > cmp.clientWidth);
		
		if( (clickInHor&&clickInVer) || (!clickInHor&&!clickInVer)) return;
		var sinfo = this.sheet.sinfo;
		sinfo.pinXY(data[0], data[1], clickInHor);
	},
	_doTouchStart: function (evt) {
		let touch0 = evt.originalEvent.touches[0],
			x = touch0.pageX,
			y = touch0.pageY;
		this._currentTouch = {x: x, y: y, t: new Date().getTime()};
	},
	_doTouchEnd: function (evt) {
		let sheet = this.sheet,
			prevTouch = this._prevTouch,
			touch0 = evt.originalEvent.changedTouches[0],
			x = touch0.pageX,
			y = touch0.pageY;
		if (prevTouch) {
			let diff = new Date().getTime() - prevTouch.t;
			if (diff < 300) {
				let prevCell = zss.SSheetCtrl._calCellPos(sheet, prevTouch.x, prevTouch.y);
				let currCell = zss.SSheetCtrl._calCellPos(sheet, x, y);
				if (prevCell[0] == currCell[0] && prevCell[1] == currCell[1]) {
					sheet._doMousedblclick(evt);
					this._prevTouch = null;
				}
			} else {
			    evt.preventDefault();
			}
		} else {
			evt.preventDefault(); // KEIKAI-494: Editing usability on tablet device
		}
		this._prevTouch = this._currentTouch;
		this._currentTouch = null;
	},
	_doScrolling: function () {
		this.ticking = false;
		var updateSimpleScrollbar = this._scrollbar != undefined;
		var sheet = this.sheet,
			scleft = this.scrollLeft,
			sctop = this.scrollTop,
			moveHorizontal = (scleft != this.currentLeft),
			moveVertical = (sctop != this.currentTop),
			moveLeft = scleft < this.currentLeft,
			moveTop = sctop < this.currentTop;

		// ignore same scroll position
		if (scleft == this.currentLeft && sctop == this.currentTop) {
			return;
		}

		if (moveHorizontal) {
			this.dir = moveLeft ? 'west' : 'east';
		} else {
			this.dir = moveTop ? 'north' : 'south';
		}
		this.currentLeft = scleft;
		this.currentTop = sctop;
		sheet.tp._updatePos(-this.currentTop, -this.currentLeft);
		sheet.lp._updatePos(-this.currentTop, -this.currentLeft);
		sheet.fire('onScrolling', {top: this.currentTop, left: this.currentLeft});
		if (this._visFlg) {
			//if fire by scroll to visible, don't show info 
			this._visFlg = false;
		} else {
			/*show scroll info*/
			var sinfo = sheet.sinfo;
			if (moveHorizontal) {
				if (!sinfo.visible || !sinfo.horizontal)
					sinfo.pinLocation(true);
				sinfo.showInfoOnDir(true);
			}
			
			if (moveVertical) {
				if (!sinfo.visible || sinfo.horizontal)
					sinfo.pinLocation(false);
				sinfo.showInfoOnDir(false);
			}
			if (moveVertical || moveHorizontal)
				this._trackScrolling(moveVertical);
		}
	},
	_trackScrolling: function (vertical) {
		// We need a timer here to trigger onVScroll and onHScroll,
		// because of the UX feeling that we give a browser a chance to paint layout,
		// otherwise, the browser will paint the layout after doing the heave task, which
		// feels the browser being hanging.

		if (vertical)
			this._fireOnVScroll();//fire scroll vertical
		else
			this._fireOnHScroll();//fire scroll horizontal
	},
	_vscrollTimeoutId: null,
	_hscrollTimeoutId: null,
	_fireOnVScroll: function (time) {
		clearTimeout(this._vscrollTimeoutId);

		//this._vscrollTimeoutId = setTimeout(this.proxy(this._doVScroll), time >= 0 ? time : zk.gecko ? 200 : 60);
		var self = this;
		this._vscrollTimeoutId = setTimeout(function () {
			self._doVScroll();
		}, time >= 0 ? time : 60);
	},
	_fireOnHScroll: function (time) {
		clearTimeout(this._hscrollTimeoutId);

		//this._hscrollTimeoutId = setTimeout(this.proxy(this._doHScroll), time >= 0 ? time : zk.gecko ? 200 : 60);
		var self = this;
		this._vscrollTimeoutId = setTimeout(function () {
			self._doHScroll();
		}, time >= 0 ? time : 60);
	},
	_doVScroll: function () {
		this._doScroll(true);
	},
	_doHScroll: function () {
		this._doScroll(false);
	},
	_doScroll: function (vertical) {
		this.sheet.activeBlock.doScroll(vertical);
		this.sheet.resyncMergerTarget(); // always check
	},
	_getMaxScrollLeft: function (dtcmp, sccmp) {
		return (dtcmp.offsetWidth - sccmp.offsetWidth) + zss.Spreadsheet.scrollWidth;
	},
	_getMaxScrollTop: function (dtcmp, sccmp) {
		return (dtcmp.offsetHeight - sccmp.offsetHeight) + zss.Spreadsheet.scrollWidth;
	},
	/**
	 * scroll a row, col or cell to visible
	 * @param {Object} cell cell ctrl
	 */
	scrollToVisible: function (row, col, cell, direction, pos) { // ZSS-475: add direction parameter
																// ZSS-664: add position parameter
		
		if(direction == undefined || direction == null) {
			direction = zss.SCROLL_DIR.BOTH; 
		}

		if (!pos) {
			pos = zss.SCROLL_POS.NONE;
		}
		
		//zss-219, scrollToVisible should able to not depends on cell dom instance
		
		var sheet = this.sheet;
		if (cell) {
			row = cell.r;
		    col = cell.c;
		}
		var cellLoc = this._getCellLocation(sheet,row,col);
		
		let block = sheet.activeBlock,
			w = cellLoc.width,//cellcmp.offsetWidth, // component width
			h = cellLoc.height,//cellcmp.offsetHeight, // component height
			l = cellLoc.left + sheet.leftWidth,//cellcmp.offsetLeft + block.comp.offsetLeft,//cell left in row + block left in datapanel = cell left in scroll panel
			t = cellLoc.top + sheet.topHeight,//cellcmp.parentNode.offsetTop + block.comp.offsetTop,//Row top in block + block top in datapanel = cell top in scroll panel.
			// KEIKAI-225
			sl = this.currentLeft,//current scroll left
			st = this.currentTop,//current scroll top
			sw = this.width, //scroll panel width(no scroll bar)
			sh = this.height, //scroll panel height(no scroll bar), 1 for the focus border
			th = sheet.topHeight, //sheet top height
			lw = sheet.leftWidth, //sheet left width
			lsl = sl, //final scroll left
			lst = st, //final scroll top
			dirty = false,
			fzr = sheet.frozenRow, //processing for frozen
			fzc = sheet.frozenCol,
			fh = 0,
			fw = 0;
		
		/**
		 * I wan to provide a feature, is moveFocus on frozenpanel then don't scroll datapanel 
		 * but if scroll to visible after a jump(for example, at the last column then click HOME)
		 * then the datapanel wil be recreate, if i dont scroll to the certain cell, 
		 * then i will see a blank (because the cell in this position is all cleared. 
		 */
		var moveonfr = moveonfc = false;
		 
		if (zkS.t(col) && fzc > -1) {
			if (fzc < col)
				fw = sheet.custColWidth.getStartPixel(fzc + 1);
			else
				moveonfc = true;// cell on frozen col, no need to scroll datapanel 
		}
	
		if (zkS.t(row) && fzr > -1) {
			if (fzr < row)
				fh = sheet.custRowHeight.getStartPixel(fzr + 1);
			else
				moveonfr = true;// cell on frozen col, no need to scroll datapanel
		}
		//end procesing for frozen
		
		/**
		 * if cell width large then scroll panel width
		 * or cell left small then scroll panel left
		 * then scroll left = cell-left
		 * 
		 * if scroll right < cell right
		 * then scroll left = cell right - scroll width (show cell right at the end of right side 
		 */
		if (zkS.t(col) && !moveonfc) {
			//cellcmp width large then scroll pane || cell-left  small then scroll=panel-left
			if (( (sw - lw) < (w - fw)) || (l-lw) < (sl+fw) ) {  
				lsl = l-lw - fw;
				dirty = true;
			 } else if(sl + sw < l + w) {
				lsl = l + w - sw;
				dirty = true;
			}
		}
		if (zkS.t(row) && !moveonfr) {
			if (((sh - th < h - fh || t - th < st + fh) && pos != zss.SCROLL_POS.BOTTOM) ||
				(pos == zss.SCROLL_POS.TOP)) {
				lst = t-th - fh;
				dirty = true;
			} else if (st + sh < t + h || pos == zss.SCROLL_POS.BOTTOM) {//top large then scroll panel bottom
				lst = t+h - sh;
				dirty = true;
			}
		}

		if (dirty) {
			this._visFlg = true;

			// ZSS-475: consider scrolling direction accroding to argument
			switch(direction) {
			case zss.SCROLL_DIR.HORIZONTAL :
				this.$scroll().scrollLeft = lsl;
				this.scrollLeft = lsl;
				break;
			case zss.SCROLL_DIR.VERTICAL :
				this.$scroll().scrollTop = lst;
				this.scrollTop = lst < 0 ? 0 : lst;
				break;
			case zss.SCROLL_DIR.BOTH :
				var $scroll = this.$scroll();
				$scroll.scrollLeft = lsl;
				$scroll.scrollTop = lst;
				this.scrollLeft = lsl;
				this.scrollTop = lst;
				break;
			}
			//after this , onScroll will be fired.
		}
	},
	_getCellLocation : function(sheet, row, column){
		var loc = {};
		loc.left = sheet.custColWidth.getStartPixel(column);
		loc.width = sheet.custColWidth.getSize(column);
		loc.top = sheet.custRowHeight.getStartPixel(row);
		loc.height = sheet.custRowHeight.getSize(row);
		return loc;
	}
});