/* TopPanel.js

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Mon Apr 23, 2007 17:29:18 AM , Created by sam
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/

/**
 * This widget renders the top frozen area including column headers.
 */
zss.TopPanel = zk.$extends(zss.Panel, {
	widgetName: 'TopPanel',
	type: 'top',
	$o: zk.$void, //owner, fellows relationship no needed
	initHeaders_: function (sheet, start, end, data, isCorner) {
		var ary = [],
			headers = data.columnHeaders,
			type = zss.Header.HOR,
			topHeaders = isCorner ? sheet.tp.headers : null,
			j = 0;
		for (var i = start; i <= end; i++) {
			if (!sheet.custColWidth.isHidden(i)) {
				var h = new zss.Header(sheet, type, headers[i], topHeaders ? topHeaders[j++] : null);
				this.appendChild(h, true);
				ary.push(h);
			}
		}
		return ary;
	},
	/**
	 * Create cells and associated headers
	 */
	create_: function (dir, colStart, colEnd, frozenRowStart, frozenRowEnd, createFrozenOnly) {
		var colsIndex = [colStart, colEnd];
		if (!createFrozenOnly)
			colsIndex = this.createHeaders_(dir, colStart, colEnd);

		var createFrozen = frozenRowStart >= 0 && frozenRowEnd >= 0;
		if ('jump' == dir && createFrozen) {
			var oldBlock = this.block;
			this.block = new zss.CellBlockCtrl(this.sheet, frozenRowStart, colStart, frozenRowEnd, colEnd, this.getFrozenData_(), 'top'); // ZSS-404: fixed missed sheet
			oldBlock ? oldBlock.replaceWidget(this.block) : this.appendChild(this.block);
		} else if (this.block && createFrozen && colsIndex) {
			this.block.create_(dir, frozenRowStart, colsIndex[0], frozenRowEnd, colsIndex[1], this.getFrozenData_());
		}
	},
	initFrozenBlock_: function (sheet, lCol, rCol, data) {
		var r = sheet.frozenRow;
		if (r > -1) {
			var b = new zss.CellBlockCtrl(sheet, 0, lCol, r, rCol, data, 'top');
			this.appendChild(b);
			return b;
		}
		return null;
	},
	getHeaderData_: function () {
		// ZSS-404: if this is for corner panel's header, the header data must be fetched from left freeze panel
		if(this.isCorner) {
			return this.sheet._wgt._cacheCtrl.getSelectedSheet().leftFrozen.columnHeaders;
		} else {
			return this.sheet._wgt._cacheCtrl.getSelectedSheet().columnHeaders;
		}
	},
	getFrozenData_: function () {
		var a = this.sheet._wgt._cacheCtrl.getSelectedSheet(),
			f = a.topFrozen;
		return f ? f : a;
	},
	getFrozenHeaderData_: function () {
		var topFrozen =  this.getFrozenData_();
		return topFrozen ? topFrozen.rowHeaders : null;
	},
	getHeaderType_: function () {
		return zss.Header.HOR;
	},
	isInsert_: function (dir) {
		return 'west' == dir;
	},
	bindFrozenCtrl_: function () {
		var sheet = this.sheet,
			selArea = this.$n('select'),
			selChg = selArea.nextSibling,
			focus = selChg.nextSibling,
			highlight = focus.nextSibling;
		var selRanges = sheet.initparm.selrange,
			clonedSelRanges = [];
		for (var i = 0, l = selRanges.length; i < l; i++) {
			clonedSelRanges.push(selRanges[i].clone());
		}
		this.selArea = new zss.SelAreaCtrlTop(sheet, selArea, clonedSelRanges);
		this.selChgArea = new zss.SelChgCtrlTop(sheet, selChg);
		this.focusMark = new zss.FocusMarkCtrlTop(sheet, focus, sheet.initparm.focus.clone());
		this.hlArea = new zss.HighlightTop(sheet, highlight, sheet.initparm.hlrange.clone(), "inner");
	},
	//ZSS-1043
	addEditorFocus_ : function(id, name, div) {
		if(!this.editorFocusMark)
			this.editorFocusMark = new Object();
		this.editorFocusMark[id] = new zss.FocusMarkCtrlTop(this.sheet, div, new zss.Pos(0, 0));
	},
	_getTopHeaderFontSize: function () {
		var head = this.$n('head'),
			col = head != null ? head.firstChild : null;
		if (col && col.getAttribute('zs.t') == 'STheader')
			return jq(col).css('font-size');
		return null;
	},
	bind_: function () {
		this.$supers(zss.TopPanel, 'bind_', arguments);
		var fontSize= this.isCorner && (zk.ie && zk.ie < 11) ? this._getTopHeaderFontSize() : null;
		if (fontSize)
			this.fontSize = fontSize;
	},
	unbind_: function () {
		this.$supers(zss.TopPanel, 'unbind_', arguments);
		this.fontSize = null;
	},
	getTypeAttr_: function () {
		return 'STopPanel';
	},
	removeChildFromStart_: function (size) {
		if (this.block) {
			this.block.removeColumnsFromStart_(size);
		}
		this.$supers(zss.TopPanel, 'removeChildFromStart_', arguments);
	},
	removeChildFromEnd_: function (size) {
		if (this.block) {
			this.block.removeColumnsFromEnd_(size);
		}
		this.$supers(zss.TopPanel, 'removeChildFromEnd_', arguments);
	},
	/**
	 *  IE need font size to display correctly in corner panel
	 *  @return string font size of the header , or null means this's not in corner panel
	 */
	_getCornerHeaderFontSize: function () {
		return this.fontSize;
	},
	_updateWidth: function (width) {
		jq(this.comp).css('width', jq.px0(width));
		this.width = width;	
		this._updateBlockWidth();
	},
	_updatePos: function (toppos, leftpos) { // KEIKAI-422
		jq(this.icomp).css('transform', 'translate3d(' + jq.px(leftpos) + ', 0, 0)');
		jq(this.sheet.dp.comp).css('transform', 'translate3d(' + jq.px(leftpos) + ', ' + jq.px(toppos)+ ', 0)'); // KEIKAI-7, KEIKAI-422
		jq(this.wpcomp).css('transform', 'translate3d(' + jq.px(leftpos) + ', 0, 0)');
		this.leftpos = leftpos;
		this._updateBlockWidth();
	},
	_updateLeftPadding: function (leftpad) {
		leftpad = leftpad - this.sheet.leftWidth;
		jq(this.icomp).css('padding-left', jq.px0(leftpad));
		this.leftpad = leftpad;
		this._updateBlockWidth();
	},
	_updateBlockWidth: function () {
		if (!this.block) return;
		var width = this.width,
			leftpos = this.leftpos,
			leftpad = this.leftpad;
		// ZSS-1281: grey area; side effect of fixing ZSS-1117; revert
		width = width - (leftpos ? leftpos : 0) - (leftpad ? leftpad : 0);		
//			//ZSS-1117
//			range = this.block.range,
//			left = range.left,
//			right = range.right;
//		
//		width = this.sheet.custColWidth.getDiffPixel(left, right);//ZSS-1117
		if (width < 0) width = 0;
		if (this.blockWidth === width) {
			return;
		}
		this.blockWidth = width;
		jq(this.block.comp).css('width', jq.px0(width));
	},
	/**
	 * Sets the selection range of the header
	 * @param int from row selection start index
	 * @param int to row selection end index
	 */
	updateSelectionCSS: function (from, to, remove) {
		const headers = this.headers;
		for (i = 0; i < headers.length; ++i) {
			const header = headers[i];
			const index = header.index;
			if (index >= from && index <= to) {
				const $header = jq(header.comp);
				if ($header.hasClass('zstop-sel')) {
					if (remove) {
						$header.removeClass('zstop-sel');
					}
				} else {
					if (!remove) {
						$header.addClass('zstop-sel');
					}
				}
			}
		}
	},
	/**
	 * Sets the width position index
	 * @param int col column index
	 * @param int zsw the width position index
	 */
	appendZSW: function (col, zsw) {
		if (this.block)
			this.block.appendZSW(col, zsw);

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			return;
		}
		col = this.sheet.activeBlock._getVisibleColIndex(col - this.headers[0].index);

		if (col < 0 || this.headers.length - 1 < col) return;
		
		var header = this.headers[col];
		header.appendZSW(zsw);
	},
	/**
	 * Sets the height position index
	 * @param row row index
	 * @param int the height position index
	 */
	appendZSH: function (row, zsh) {
		if (this.block)
			this.block.appendZSH(row, zsh);
	},
	/**
	 * Insert new column
	 * @param int col column
	 * @param int size the size of the column
	 * @param array extnm
	 * @param boolean forShow true if the operation is for show column to use
	 */
	insertNewColumn: function (col, size, extnm, forShow) {
		if (this.block)
			this.block.insertNewColumn(col, size);

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			let colHeaders;
			if (!forShow || !(colHeaders = this.getHeaderData_())) {
				return;
			}

			//insert columns intersect with selection range, must remove selection CSS before insert cells
			var sheet = this.sheet,
				selRange = sheet.selArea.lastRange,
				type = zss.Header.HOR;

			if (sheet.state != zss.SSheetCtrl.NOFOCUS && selRange) {
				var left = selRange.left,
					right = selRange.right;
				if (col <= right && (col+size-1) >= left)
					this.updateSelectionCSS(left, right, true);
			}

			// KEIKAI-62: we need to insert new column here
			const hs = [];
			let index = this._getClosestHeaderIndex(col);
			for (var i = 0; i < size; i++) {
				var c = col + i;
				const h = new zss.Header(sheet, type, colHeaders[c] || {});
                hs.push(h);
			}
			this.insertHeaders_(index, hs);
			return;
		}

		var left = this.headers[0].index,
			right = this.headers[this.headers.length - 1].index;

		var index = Math.max(col - left, 0);

		// KEIKAI-62: some are invisible
		if (right - left + 1 != this.headers.length) {
			index = this._getClosestHeaderIndex(col);
		}
		var ctrl,
			colHeaders = this.getHeaderData_(),
			type = zss.Header.HOR;
			
		//insert columns intersect with selection range, must remove selection CSS before insert cells
		var sheet = this.sheet,
			selRange = sheet.selArea.lastRange;
		if (sheet.state != zss.SSheetCtrl.NOFOCUS && selRange) {
			var left = selRange.left,
				right = selRange.right;
			if (col <= right && (col+size-1) >= left)
				this.updateSelectionCSS(left, right, true);
		}

		const hs = [];
		for (var i = 0; i < size; i++) {
			var c = col + i;
			const h = new zss.Header(sheet, type, colHeaders[c]);
			hs.push(h);
		}
		this.insertHeaders_(index, hs);
		if (!forShow) {
			extnm = extnm.slice(size, extnm.length);
			this.shiftHeaderInfo(index + size, col + size, extnm);
		}
	},
	/**
	 * Insert new row
	 * @param int row row index
	 * @param int size the size of the row
	 * @param boolean forShow true if the operation is for show row to use
	 */
	insertNewRow: function (row, size, forShow) {
		if (this.block)
			this.block.insertNewRow(row,size, this.getFrozenData_(), forShow);
	},
	/**
	 * Shift header
	 * @param int index start shift index
	 * @param int newcol new column index
	 * @param array extnm
	 */
	shiftHeaderInfo: function (index, newcol, extnm, ignoreHidden) {
		var size = this.headers.length,
			j = 0,
			sheet = this.sheet;

		if (index >= size) return;

		let start = this.headers[index].index,
			hiddenCount = ignoreHidden ? 0 : this.sheet.custColWidth.getHiddenCount(start > newcol ? newcol : start, start > newcol ? start : newcol),
			delta = newcol - start + hiddenCount;
		for (; index < size; index++) {
			if (!extnm[j]) zk.log("undefined header to assing>>" + (newrow + j), "always");
			let header = this.headers[index];
			header.resetInfo(header.index + delta, extnm[header.index - start + hiddenCount]);
		}
	},
	/**
	 * Remove column at the given index
	 * @param int col column index
	 * @param array extnm
	 * @param boolean forHidden true if the operation is for hidden column to use
	 */
	removeColumnAt: function (col, extnm, forHidden) {
		if (this.block)
			this.block.removeColumnAt(col, forHidden);

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			return;
		}

		let index = this.getHeaderIndex(col);

		if (index < 0) return;

		this.removeHeader_(index, 1);

		if (!forHidden) {
			this.shiftHeaderInfo(index, col, extnm);
		}
	},
	//KEIKAI-117
	/**
	 * Removes columns from lCol to rCol
	 * @param int lCol the left column index
	 * @param int rCol the right column index (inclusive)
	 * @param array extnm
	 */
	removeColumnFromRange: function (lCol, rCol, extnm) {
		if (this.block)
			this.block.removeColumnFromRange(lCol, rCol);

		// ZSS-515: header might be empty because of removing last row/column at freeze panels
		if(this.isEmptyHeader()) {
			return;
		}

		let custSize = this.sheet.custColWidth;
		// find first visible lCol
		for (let start = lCol, end = rCol+1; start < end; start++) {
			if (!custSize.isHidden(start)) {
				lCol = start;
				break;
			}
		}

		let index = this.getHeaderIndex(lCol);

		if (index < 0) return;


		let size = custSize.getVisibleCount(lCol, rCol);
		this.removeHeader_(index, size);

		this.shiftHeaderInfo(index, lCol, extnm, true);
	},
	//KEIKAI-117
	/**
	 * Remove rows from tRow to bRow
	 * @param tRow the top index
	 * @param bRow the bottom index (inclusive)
	 */
	removeRowFromRange: function (tRow, bRow) {
		if (this.block)
			this.block.removeRowFromRange(tRow, bRow);
	},
	/**
	 * Remove row at the given index
	 * @param int row row index
	 * @param boolean forHidden true if the operation is for hidden row to use
	 */
	removeRowAt: function (row, forHidden) {
		if (this.block)
			this.block.removeRowAt(row, forHidden);
	},
	/**
	 * Sets the selection range of the header
	 * 
	 * Update CSS after header created (this.createHeaders_), ignore update selection when sheet CSS not ready
	 */
	updateSelectionCSS_: function () {
		var sheet = this.sheet,
			wgt = sheet._wgt,
			selRange = sheet.selArea.lastRange;
		if (!wgt.isSheetCSSReady())
			return;
		
		if (selRange) {
			var left = selRange.left,
				right = selRange.right;
			this.updateSelectionCSS(left, right, false || sheet.state == zss.SSheetCtrl.NOFOCUS);
		}
	},
	_fixSize: function() {
		var sheet = this.sheet,
			wgt = sheet._wgt,
			th = sheet.topHeight,
			toph = th,
			fzr = sheet.frozenRow,
			name = wgt.getSelectorPrefix(),
			sid = wgt.getSheetCSSId();
		
		if (fzr > -1)
			toph = toph + sheet.custRowHeight.getStartPixel(fzr + 1);
		
		zcss.setRule(name + " .zstop", ["height"], [(fzr > -1 ? toph - 1: toph) + "px"], true, sid);
		zcss.setRule(name + " .zstopi", ["height"], [toph + "px"], true, sid);
	}
});