/* SheetProtection.java

	Purpose:
		
	Description:
		
	History:
		6:10 PM 2019/11/18, Created by jumperchen

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package io.keikai.api;

/**
 * SheetProtection values config.
 * <p>Using {@link SheetProtection.Builder} you can construct a SheetProtection object.</p>
 * @author jumperchen
 * @since 5.0.0
 */
public class SheetProtection {
	private String password;
	private boolean objectsEditable = true; // default to true
	private boolean scenarios = true; // default to true
	private boolean formatCellsAllowed;
	private boolean formatColumnsAllowed;
	private boolean formatRowsAllowed;
	private boolean insertColumnsAllowed;
	private boolean insertRowsAllowed;
	private boolean insertHyperlinksAllowed;
	private boolean deleteColumnsAllowed;
	private boolean deleteRowsAllowed;
	private boolean sortAllowed;
	private boolean autoFilterAllowed;
	private boolean pivotTablesAllowed;
	private boolean selectLockedCellsAllowed;
	private boolean selectUnlockedCellsAllowed;

	// use Builder to construct it.
	private SheetProtection() {}

	/**
	 * Returns the password.
	 * @return
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Returns whether to protect objects.
	 * <p>Default to true</p>
	 * @return
	 */
	public boolean isObjectsEditable() {
		return objectsEditable;
	}

	/**
	 * Returns whether to protect scenarios.
	 * <p>Default to true</p>
	 * @return
	 */
	public boolean isScenariosEditable() {
		return scenarios;
	}

	/**
	 * Returns whether to allow user to format any cell on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isFormatCellsAllowed() {
		return formatCellsAllowed;
	}

	/**
	 * Returns whether to allow user to format any columns on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isFormatColumnsAllowed() {
		return formatColumnsAllowed;
	}

	/**
	 * Returns whether to allow user to format any rows on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isFormatRowsAllowed() {
		return formatRowsAllowed;
	}

	/**
	 * Returns whether to allow user to insert columns on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isInsertColumnsAllowed() {
		return insertColumnsAllowed;
	}

	/**
	 * Returns whether to allow user to insert rows on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isInsertRowsAllowed() {
		return insertRowsAllowed;
	}

	/**
	 * Returns whether to allow user to insert hyperlinks on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isInsertHyperlinksAllowed() {
		return insertHyperlinksAllowed;
	}

	/**
	 * Returns whether to allow user to delete columns on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isDeleteColumnsAllowed() {
		return deleteColumnsAllowed;
	}

	/**
	 * Returns whether to allow user to delete rows on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isDeleteRowsAllowed() {
		return deleteRowsAllowed;
	}

	/**
	 * Returns whether to allow user to sort on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isSortAllowed() {
		return sortAllowed;
	}

	/**
	 * Returns whether to allow user to set filters on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isAutoFilterAllowed() {
		return autoFilterAllowed;
	}

	/**
	 * Returns whether to allow user to use pivot table reports on the protected sheet.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isPivotTablesAllowed() {
		return pivotTablesAllowed;
	}

	/**
	 * Returns whether to allow select locked cells.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isSelectLockedCellsAllowed() {
		return selectLockedCellsAllowed;
	}

	/**
	 * Returns whether to allow select unlocked cells.
	 * <p>Default to false</p>
	 * @return
	 */
	public boolean isSelectUnlockedCellsAllowed() {
		return selectUnlockedCellsAllowed;
	}

	/**
	 * Builder is used to build instances of SheetProtection from values configured by the setters
	 */
	public static final class Builder {
		private SheetProtection _protection;

		private Builder(SheetProtection protection) {
			_protection = protection;
		}

		/**
		 * Creates the builder instance.
		 * @return
		 */
		public static Builder create() {
			return new Builder(new SheetProtection());
		}

		/**
		 * Sets the case-sensitive password for the sheet; null or empty string
		 * means protect the sheet without password.
		 * <p>Default to null</p>
		 *
		 * @param password
		 */
		public Builder withPassword(String password) {
			_protection.password = password;
			return this;
		}

		/**
		 * Sets the objectsEditable to true to protect objects.
		 * <p>Default to true</p>
		 * @param objectsEditable
		 */
		public Builder withObjectsEditable(boolean objectsEditable) {
			_protection.objectsEditable = objectsEditable;
			return this;
		}

		/**
		 * Sets true to protect scenarios.
		 * <p>Default to true</p>
		 * @param scenarios
		 */
		public Builder withScenariosEditable(boolean scenarios) {
			_protection.scenarios = scenarios;
			return this;
		}

		/**
		 * Sets true to allow user to format any cell on the protected sheet.
		 * <p>Default to false</p>
		 * @param formatCellsAllowed
		 */
		public Builder withFormatCellsAllowed(boolean formatCellsAllowed) {
			_protection.formatCellsAllowed = formatCellsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to format any columns on the protected sheet.
		 * <p>Default to false</p>
		 * @param formatColumnsAllowed
		 */
		public Builder withFormatColumnsAllowed(boolean formatColumnsAllowed) {
			_protection.formatColumnsAllowed = formatColumnsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to format any rows on the protected sheet.
		 * <p>Default to false</p>
		 * @param formatRowsAllowed
		 */
		public Builder withFormatRowsAllowed(boolean formatRowsAllowed) {
			_protection.formatRowsAllowed = formatRowsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to insert columns on the protected sheet.
		 * <p>Default to false</p>
		 * @param insertColumnsAllowed
		 */
		public Builder withInsertColumnsAllowed(boolean insertColumnsAllowed) {
			_protection.insertColumnsAllowed = insertColumnsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to insert rows on the protected sheet.
		 * <p>Default to false</p>
		 * @param insertRowsAllowed
		 */
		public Builder withInsertRowsAllowed(boolean insertRowsAllowed) {
			_protection.insertRowsAllowed = insertRowsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to insert hyperlinks on the protected sheet.
		 * <p>Default to false</p>
		 * @param insertHyperlinksAllowed
		 */
		public Builder withInsertHyperlinksAllowed(boolean insertHyperlinksAllowed) {
			_protection.insertHyperlinksAllowed = insertHyperlinksAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to delete columns on the protected sheet.
		 * <p>Default to false</p>
		 * @param deleteColumnsAllowed
		 */
		public Builder withDeleteColumnsAllowed(boolean deleteColumnsAllowed) {
			_protection.deleteColumnsAllowed = deleteColumnsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to delete rows on the protected sheet.
		 * <p>Default to false</p>
		 * @param deleteRowsAllowed
		 */
		public Builder withDeleteRowsAllowed(boolean deleteRowsAllowed) {
			_protection.deleteRowsAllowed = deleteRowsAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to sort on the protected sheet.
		 * <p>Default to false</p>
		 * @param sortAllowed
		 */
		public Builder withSortAllowed(boolean sortAllowed) {
			_protection.sortAllowed = sortAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to set filters on the protected sheet
		 * <p>Default to false</p>
		 * @param autoFilterAllowed
		 */
		public Builder withAutoFilterAllowed(boolean autoFilterAllowed) {
			_protection.autoFilterAllowed = autoFilterAllowed;
			return this;
		}

		/**
		 * Sets true to allow user to use pivot table reports on the protected sheet.
		 * <p>Default to false</p>
		 * @param pivotTablesAllowed
		 */
		public Builder withPivotTablesAllowed(boolean pivotTablesAllowed) {
			_protection.pivotTablesAllowed = pivotTablesAllowed;
			return this;
		}

		/**
		 * Sets true to allow select locked cells.
		 * <p>Default to false</p>
		 * @param selectLockedCellsAllowed
		 */
		public Builder withSelectLockedCellsAllowed(boolean selectLockedCellsAllowed) {
			_protection.selectLockedCellsAllowed = selectLockedCellsAllowed;
			return this;
		}

		/**
		 * Sets true to allow select unlocked cells
		 * @param selectUnlockedCellsAllowed
		 */
		public Builder withSelectUnlockedCellsAllowed(boolean selectUnlockedCellsAllowed) {
			_protection.selectUnlockedCellsAllowed = selectUnlockedCellsAllowed;
			return this;
		}

		/**
		 * Returns an instance of SheetProtection created from the fields set on this builder.
		 * @return
		 */
		public SheetProtection build() {
			return _protection;
		}
	}
}
