/* InnerModelEventDispatcher.java

	Purpose:
		
	Description:
		
	History:
		2:44 PM 2020/1/17, Created by jumperchen

Copyright (C) 2020 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui;

import java.util.Objects;
import java.util.function.Consumer;

import io.keikai.api.AreaRef;
import io.keikai.api.model.Sheet;
import io.keikai.api.model.impl.SheetImpl;
import io.keikai.api.model.impl.SimpleRef;
import io.keikai.model.CellRegion;
import io.keikai.model.ModelEvent;
import io.keikai.model.ModelEvents;
import io.keikai.model.SBook;
import io.keikai.model.SChart;
import io.keikai.model.SPicture;
import io.keikai.model.SSheet;
import io.keikai.model.STable;
import io.keikai.ui.event.CellAreaEvent;
import io.keikai.ui.event.Events;
import io.keikai.ui.event.SheetDeleteEvent;
import io.keikai.ui.event.SheetEvent;
import io.keikai.ui.impl.Focus;
import io.keikai.ui.sys.SpreadsheetCtrl;

import org.zkoss.zk.au.out.AuInvoke;
import org.zkoss.zk.ui.Executions;
import org.zkoss.zk.ui.sys.WebAppCtrl;

/**
 * Internal use only
 * @author jumperchen
 * @since 5.1.0
 */ /* DataListener to handle sheet data event */
/*package*/ class InnerModelEventDispatcher extends ModelEventDispatcher {
	private static final long serialVersionUID = 20100330164021L;

	private final Spreadsheet _spreadsheet;

	/*package*/ InnerModelEventDispatcher(Spreadsheet spreadsheet) {
		this._spreadsheet = spreadsheet;
		addEventListener(ModelEvents.ON_SHEET_ORDER_CHANGE,
			evt -> wrapComponent(evt, this::onSheetOrderChange));
		addEventListener(ModelEvents.ON_SHEET_NAME_CHANGE,
			evt -> wrapComponent(evt, this::onSheetNameChange));
		addEventListener(ModelEvents.ON_SHEET_CREATE,
			evt -> wrapComponent(evt, this::onSheetCreate));
		addEventListener(ModelEvents.ON_SHEET_DELETE,
			evt -> wrapComponent(evt, this::onSheetDelete));
		//ZSS-832
		addEventListener(ModelEvents.ON_SHEET_VISIBLE_CHANGE,
			evt -> wrapComponent(evt, this::onSheetVisibleChange));
		//KEIKAI-668
		addEventListener(ModelEvents.ON_SHEET_SELECT,
			evt -> wrapComponent(evt, this::onSheetSelect));
		addEventListener(ModelEvents.ON_MODEL_FRIEND_FOCUS_MOVE,
			evt -> wrapComponent(evt, this::onFriendFocusChange));
		addEventListener(ModelEvents.ON_MODEL_FRIEND_FOCUS_DELETE,
			evt -> wrapComponent(evt, this::onFriendFocusChange));

		addEventListener(ModelEvents.ON_CELL_CONTENT_CHANGE,
			evt -> wrapComponent(evt, this::onCellContentChange));
		addEventListener(ModelEvents.ON_CHART_CONTENT_CHANGE,
			evt -> wrapComponent(evt, this::onChartContentChange));
		addEventListener(ModelEvents.ON_DATA_VALIDATION_CONTENT_CHANGE,
			evt -> wrapComponent(evt, this::onDataValidationContentChange));
		addEventListener(ModelEvents.ON_ROW_INSERT,
			evt -> wrapComponent(evt, this::onRowColumnInsertDelete));
		addEventListener(ModelEvents.ON_ROW_DELETE,
			evt -> wrapComponent(evt, this::onRowColumnInsertDelete));
		addEventListener(ModelEvents.ON_COLUMN_INSERT, evt -> wrapComponent(evt, this::onRowColumnInsertDelete));
		addEventListener(ModelEvents.ON_COLUMN_DELETE, evt -> wrapComponent(evt, this::onRowColumnInsertDelete));
		addEventListener(ModelEvents.ON_ROW_COLUMN_SIZE_CHANGE, evt -> wrapComponent(evt, this::onRowColumnSizeChange));
		addEventListener(ModelEvents.ON_AUTOFILTER_CHANGE, evt -> wrapComponent(evt, this::onAutoFilterChange));
		addEventListener(ModelEvents.ON_MERGE_ADD, evt -> wrapComponent(evt, this::onMergeAdd));

		addEventListener(ModelEvents.ON_MERGE_DELETE, evt -> wrapComponent(evt, this::onMergeDelete));
		//ZSS-1168
		addEventListener(ModelEvents.ON_MERGE_SYNC, evt -> wrapComponent(evt, this::onMergeSync));
		addEventListener(ModelEvents.ON_DISPLAY_GRIDLINES_CHANGE, evt -> wrapComponent(evt, this::onDisplayGridlines));
		addEventListener(ModelEvents.ON_PROTECT_SHEET_CHANGE, evt -> wrapComponent(evt, this::onProtectSheet));
		addEventListener(ModelEvents.ON_CHART_ADD, evt -> wrapComponent(evt, this::onChartAdd));
		addEventListener(ModelEvents.ON_CHART_DELETE, evt -> wrapComponent(evt, this::onChartDelete));
		addEventListener(ModelEvents.ON_CHART_UPDATE, evt -> wrapComponent(evt, this::onChartUpdate));

		addEventListener(ModelEvents.ON_PICTURE_ADD, evt -> wrapComponent(evt, this::onPictureAdd));
		addEventListener(ModelEvents.ON_PICTURE_DELETE, evt -> wrapComponent(evt, this::onPictureDelete));
		addEventListener(ModelEvents.ON_PICTURE_UPDATE, evt -> wrapComponent(evt, this::onPictureUpdate));
		addEventListener(ModelEvents.ON_FREEZE_CHANGE, evt -> wrapComponent(evt, this::onSheetFreeze));
		addEventListener(ModelEvents.ON_GROUP_CHANGE, evt -> wrapComponent(evt, this::onSheetGroupChange));
		// KEIKAI-607
		addEventListener(ModelEvents.ON_RELEASE_SELECTED_CACHE, evt -> wrapComponent(evt, this::onReleaseSelectedCache));
	}

	private void onSheetOrderChange(ModelEvent event) {
		_spreadsheet.smartUpdate(Spreadsheet.ClientCommands.SHEET_TABS, _spreadsheet.getSheetTabs());
		Sheet sheet = _spreadsheet.getBook().getSheet(event.getSheet().getSheetName());
		org.zkoss.zk.ui.event.Events.postEvent(
				new SheetEvent(Events.ON_AFTER_SHEET_ORDER_CHANGE, _spreadsheet, sheet));
	}

	private void onSheetNameChange(ModelEvent event) {
		_spreadsheet.smartUpdate(Spreadsheet.ClientCommands.SHEET_TABS, _spreadsheet.getSheetTabs());
		Sheet sheet = _spreadsheet.getBook().getSheet(event.getSheet().getSheetName());
		org.zkoss.zk.ui.event.Events.postEvent(
				new SheetEvent(Events.ON_AFTER_SHEET_NAME_CHANGE, _spreadsheet, sheet));
	}

	//ZSS-832
	private void onSheetVisibleChange(ModelEvent event) {
		_spreadsheet.smartUpdate(Spreadsheet.ClientCommands.SHEET_TABS, _spreadsheet.getSheetTabs());
		Sheet sheet = _spreadsheet.getBook().getSheet(event.getSheet().getSheetName());
		org.zkoss.zk.ui.event.Events.postEvent(new SheetEvent(Events.ON_AFTER_SHEET_VISIBLE_CHANGE,
				_spreadsheet, sheet));
	}

	// KEIKAI-668
	private void onSheetSelect(ModelEvent event) {
		_spreadsheet.setSelectedSheet(event.getSheet().getSheetName());
	}

	private void onSheetCreate(ModelEvent event) {
		_spreadsheet.smartUpdate(Spreadsheet.ClientCommands.SHEET_TABS, _spreadsheet.getSheetTabs());
		Sheet sheet = _spreadsheet.getBook().getSheet(event.getSheet().getSheetName());
		_spreadsheet.refreshToolbarDisabled();
		org.zkoss.zk.ui.event.Events.postEvent(
				new SheetEvent(Events.ON_AFTER_SHEET_CREATE, _spreadsheet, sheet));
	}

	private void onSheetDelete(ModelEvent event) {
		SBook book = _spreadsheet.getSBook();
		SSheet delSheet = event.getSheet();
		//TODO zss 3.5 clear active client cache and active range record

		//KEIKAI-430 delete Clipboard if deleting same sheet
		final UserActionContext.Clipboard cb = _spreadsheet.getClipboard();
		if (cb != null && cb.getSheet().getInternalSheet().equals(delSheet)) {
			_spreadsheet.clearClipboard();
		}

		if (delSheet == _spreadsheet.getSelectedSSheet()) {
			int delIndex = (Integer) event.getData(ModelEvents.PARAM_INDEX);
			//the sheet that selected is deleted, re select another
			if (delIndex >= book.getNumOfSheet() - 1) {
				delIndex = book.getNumOfSheet() - 1;
			}
			//if current select sheet name, equals the delete sheet, we should select to suggest new sheet
			_spreadsheet.setSelectedSheet(book.getSheet(delIndex).getSheetName());//this will also update sheet label
		} else {
			//just update sheet label
			_spreadsheet.smartUpdate(Spreadsheet.ClientCommands.SHEET_TABS, _spreadsheet.getSheetTabs());
		}
		org.zkoss.zk.ui.event.Events.postEvent(
				new SheetDeleteEvent(Events.ON_AFTER_SHEET_DELETE, _spreadsheet, delSheet.getSheetName()));
	}

	private void onFriendFocusChange(ModelEvent event) {
		if (!Objects.equals(_spreadsheet.getSelectedSSheet(), event.getSheet())) {
			_spreadsheet.syncFriendFocus(); //ZSS-998
			return;
		}
		final Focus focus = (Focus) event.getCustomData(); //other's spreadsheet's focus
		final String id = focus.getId();
		final Focus selfEditorFocus = _spreadsheet.getSelfEditorFocus();
		if (selfEditorFocus != null && !id
				.equals(selfEditorFocus.getId())) {
			if (ModelEvents.ON_MODEL_FRIEND_FOCUS_MOVE.equals(event.getName())) {
				_spreadsheet.addOrMoveFriendFocus(id, focus.getName(), focus.getColor(),
						focus.getSheetId(), focus.getRow(), focus.getColumn());
			} else {
				//NOTE should remove friend color, the friend is possible back (sheet switch back)
				//TODO current we dont' have a way to remove friend color cache
				_spreadsheet.removeFriendFocus(focus.getId());
			}
			_spreadsheet.syncFriendFocus();
		}
	}

	private void onChartAdd(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		final SChart chart = sheet.getChart(event.getObjectId());
		if (chart != null) {
			addChartWidget(sheet, chart);
		}
	}

	private void onChartDelete(ModelEvent event) {
		deleteChartWidget(event.getSheet(), event.getObjectId());
	}

	private void onChartUpdate(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		final SChart chart = sheet.getChart(event.getObjectId());
		if (chart != null) {
			updateChartWidget(sheet, chart);
		}
	}

	private void onPictureAdd(ModelEvent event) {
		final SSheet sheet = event.getSheet();

		SPicture picture = sheet.getPicture(event.getObjectId());
		if (picture != null) {
			addPictureWidget(event.getSheet(), picture);
		}
	}

	private void onPictureDelete(ModelEvent event) {
		deletePictureWidget(event.getSheet(), event.getObjectId());
	}

	private void onPictureUpdate(ModelEvent event) {
		final SSheet sheet = event.getSheet();

		SPicture picture = sheet.getPicture(event.getObjectId());
		if (picture != null) {
			updatePictureWidget(event.getSheet(), picture);
		}
	}

	private void onCellContentChange(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		final CellRegion region = event.getRegion();
		// ZSS-393: update cell could be not the current selected sheet.
		// e.g. co-editing, cut from other sheet... etc.
		// but has to update the client (might has cache)

		final int left = region.getColumn();
		final int top = region.getRow();
		final int right = region.getLastColumn();
		final int bottom = region.getLastRow();

		//ZSS-939
		final Integer cellAttrVal = (Integer) event.getData("cellAttr");
		final SpreadsheetCtrl.CellAttribute cellAttr = cellAttrVal == null ?
				SpreadsheetCtrl.CellAttribute.ALL :
				SpreadsheetCtrl.CellAttribute.values()[cellAttrVal - 1];
		_spreadsheet.updateCell(sheet, left, top, right, bottom, cellAttr);

		_spreadsheet.updateUnlockInfo();
		org.zkoss.zk.ui.event.Events.postEvent(
				new CellAreaEvent(Events.ON_AFTER_CELL_CHANGE, _spreadsheet,
						new SheetImpl(new SimpleRef<>(sheet.getBook()),
								new SimpleRef<>(sheet)),
						top, left, bottom, right));

		//ZSS-1168: if notify the whole sheet; invalidate the component
		final SBook book = event.getBook();
		if (left <= 0 && right >= book.getMaxColumnIndex() && top <= 0
				&& bottom >= book.getMaxRowIndex()) {
			_spreadsheet.invalidate();
		}
	}

	private void wrapComponent(ModelEvent e, Consumer<ModelEvent> f) {
		// Fix KEIKAI-41
		Executions.getCurrent().setAttribute(Spreadsheet.INSTANCE_KEY, _spreadsheet);
		f.accept(e);
	}

	private void onChartContentChange(ModelEvent event) {
		final SSheet sheet = event.getSheet();

		final String objid = event.getObjectId();
		SChart chart = sheet.getChart(objid);
		if (chart != null) {
			_spreadsheet.updateWidget(sheet, objid);
		}
	}

	private void onDataValidationContentChange(ModelEvent event) {
		final SSheet sheet = event.getSheet();

		final String objid = event.getObjectId();
		// ZSS-648: the validation could have been deleted, must notify client side
		//			SDataValidation validation = sheet.getDataValidation(objid);

		//			if(validation!=null){
		_spreadsheet.updateDataValidation(sheet, objid);
		//			}
	}

	private void onRowColumnInsertDelete(ModelEvent event) {
		// KEIKAI-210
		if ( _spreadsheet.isInvalidated()) {
			return;
		}

		// determine parameters
		boolean inserted = ModelEvents.ON_ROW_INSERT.equals(event.getName())
				|| ModelEvents.ON_COLUMN_INSERT.equals(event.getName());
		boolean isRow = ModelEvents.ON_ROW_INSERT.equals(event.getName())
				|| ModelEvents.ON_ROW_DELETE.equals(event.getName());

		// update client's row/column
		_spreadsheet.nextCellId();
		SSheet sheet = event.getSheet();
		CellRegion region = event.getRegion();
		if (isRow) {
			int row = region.getRow();
			int lastRow = region.getLastRow();
			int count = region.getRowCount();
			if (inserted) {
				((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl())
						.insertRows(sheet, row, count);
			} else {
				((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl())
						.removeRows(sheet, row, count);
			}

			// ZSS-306 a chart doesn't shrink its size when deleting rows or columns it overlaps
			int widgetBottom = inserted ? lastRow + count - 1 : lastRow;
			_spreadsheet.loadWidgetLoaders().forEach(
					widgetLoader -> widgetLoader
							.onRowChange(sheet, row, widgetBottom));
		} else { // column
			int col = region.getColumn();
			int lastCol = region.getLastColumn();
			int count = region.getColumnCount();
			if (inserted) {
				((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl())
						.insertColumns(sheet, col, count);
			} else {
				((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl())
						.removeColumns(sheet, col, count);
			}

			// ZSS-306 a chart doesn't shrink its size when deleting rows or columns it overlaps
			int widgetRight = inserted ? lastCol + count - 1 : lastCol;
			_spreadsheet.loadWidgetLoaders().forEach(
					widgetLoader -> widgetLoader
							.onColumnChange(sheet, col, widgetRight));
		}
		_spreadsheet.updateUnlockInfo();
	}

	private void onMergeAdd(ModelEvent event) {
		SSheet sheet = event.getSheet();
		//don't skip at here, let extraCtrl sync mergeMatrix helper and skip
		//			if (!getSelectedXSheet().equals(sheet)){
		//				releaseClientCache(sheet.getId());
		//				getMergeMatrixHelper(sheet);
		//				return;
		//			}
		CellRegion region = event.getRegion();
		((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl())
				.addMergeCell(sheet, region.getColumn(), region.getRow(), region.getLastColumn(), region.getLastRow());

		_spreadsheet.loadWidgetLoaders().forEach(widgetLoader -> widgetLoader.onRowColumnChange(sheet, region));

	}

	private void onMergeDelete(ModelEvent event) {
		SSheet sheet = event.getSheet();
		//don't skip at here, let extraCtrl sync mergeMatrix helper and skip
		//			if (!getSelectedXSheet().equals(sheet)){
		//				releaseClientCache(sheet.getId());
		//				return;
		//			}
		CellRegion region = event.getRegion();
		((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl())
				.deleteMergeCell(sheet, region.getColumn(), region.getRow(),
						region.getLastColumn(), region.getLastRow());

		_spreadsheet.loadWidgetLoaders().forEach(widgetLoader -> widgetLoader.onResetSparklineAnchor(sheet, region));

	}

	//ZSS-1168
	private void onMergeSync(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		((SpreadsheetCtrlImpl) _spreadsheet.getExtraCtrl()).syncMergeCells(sheet);
	}

	private void onRowColumnSizeChange(ModelEvent event) {
		//TODO shall pass the range over to the client side and let client side do it; rather than iterate each column and send multiple command
		final SSheet sheet = event.getSheet();
		final CellRegion region = event.getRegion();
		if (event.isWholeColumn()) {
			final int left = region.column;
			final int right = region.lastColumn;
			for (int c = left; c <= right; ++c) {
				_spreadsheet.updateColWidth(sheet, c);
			}

			//ZSS-455 Chart/Image doesn't move location after change column/row width/height
			_spreadsheet.loadWidgetLoaders().forEach(widgetLoader -> widgetLoader.onColumnChange(sheet, left, right));

			final AreaRef rect = ((SpreadsheetCtrl) _spreadsheet.getExtraCtrl()).getVisibleArea();
			_spreadsheet.syncFriendFocusPosition(left, rect.getRow(), rect.getLastColumn(), rect.getLastRow());
		} else if (event.isWholeRow()) {
			final int top = region.row;
			final int bottom = region.lastRow;

			for (int r = top; r <= bottom; ++r) {
				_spreadsheet.updateRowHeight(sheet, r);
			}

			//ZSS-455 Chart/Image doesn't move location after change column/row width/height
			_spreadsheet.loadWidgetLoaders().forEach(widgetLoader -> widgetLoader.onRowChange(sheet, top , bottom));

			final AreaRef rect = ((SpreadsheetCtrl) _spreadsheet.getExtraCtrl()).getVisibleArea();
			_spreadsheet.syncFriendFocusPosition(rect.getColumn(), top, rect.getLastColumn(), rect.getLastRow());
		}
	}

	private void onAutoFilterChange(ModelEvent event) {
		final SSheet sheet = event.getSheet();

		//ZSS-1083(refix ZSS-838): Retrieve affected row count caused by onZSSFilter
		// see AutoFilterHelper.java#enableAutoFilter0
		final STable table = (STable) event.getData("TABLE");
		final String key = (table == null ? sheet.getId() : table.getName())
				+ "_KK_AFFECTED_ROWS";
		final Integer affectedRowCount = (Integer) event.getData(key);
		_spreadsheet.updateAutoFilter(sheet, table, affectedRowCount); //ZSS-988
	}

	private void onDisplayGridlines(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.setDisplayGridlines((Boolean) event.getData(ModelEvents.PARAM_ENABLED));
		}
	}

	private void onProtectSheet(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.setProtectSheet((Boolean) event.getData(ModelEvents.PARAM_ENABLED));
		}
	}

	private void onSheetFreeze(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		//TODO
		_spreadsheet.invalidate();
		_spreadsheet.loadWidgetLoaders().forEach(widgetLoader -> widgetLoader.onSheetFreeze(sheet));
	}

	private void onSheetGroupChange(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		//TODO
		_spreadsheet.invalidate();
	}

	private void addPictureWidget(SSheet sheet, SPicture picture) {
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.loadWidgetLoaders().forEach(
					(widgetLoader -> widgetLoader
							.addPictureWidget(sheet, picture)));
		}
	}

	private void deletePictureWidget(SSheet sheet, String pictureId) {
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.loadWidgetLoaders().forEach(
					(widgetLoader -> widgetLoader
							.deletePictureWidget(sheet, pictureId)));
		}
	}

	private void updatePictureWidget(SSheet sheet, SPicture picture) {
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.loadWidgetLoaders().forEach(
					(widgetLoader -> widgetLoader
							.updatePictureWidget(sheet, picture)));
		}
	}

	private void addChartWidget(SSheet sheet, SChart chart) {
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.loadWidgetLoaders().forEach(
					(widgetLoader -> widgetLoader
							.addChartWidget(sheet, chart)));
		}
	}

	private void deleteChartWidget(SSheet sheet, String chartId) {
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.loadWidgetLoaders().forEach(
					(widgetLoader -> widgetLoader
							.deleteChartWidget(sheet, chartId)));
		}
	}

	private void updateChartWidget(SSheet sheet, SChart chart) {
		if (!_spreadsheet.tryReleaseCache(sheet)) {
			_spreadsheet.loadWidgetLoaders().forEach(
					(widgetLoader -> widgetLoader
							.updateChartWidget(sheet, chart)));
		}
	}
	// KEIKAI-607
	private void onReleaseSelectedCache(ModelEvent event) {
		final SSheet sheet = event.getSheet();
		((WebAppCtrl) _spreadsheet.getPage().getDesktop().getWebApp()).getUiEngine().addResponse(new AuInvoke(_spreadsheet, "_releaseSelectedCache", sheet.getId()));
	}
}
