/* Spreadsheet.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		Thu May 17 14:41:33     2007, Created by tomyeh
		Dec 19 10:10:10 2007, modify by Dennis Chen.
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under GPL Version 2.0 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
 */
package io.keikai.ui;

import static io.keikai.util.Converter.a1ToRef;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Serializable;
import java.net.URL;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import io.keikai.api.AreaRef;
import io.keikai.api.AreaRefWithType;
import io.keikai.api.CellRef;
import io.keikai.api.IllegalFormulaException;
import io.keikai.api.Importer;
import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.UnitUtil;
import io.keikai.api.impl.ImporterImpl;
import io.keikai.api.model.Book;
import io.keikai.api.model.Font;
import io.keikai.api.model.Hyperlink.HyperlinkType;
import io.keikai.api.model.Sheet;
import io.keikai.api.model.SheetProtection;
import io.keikai.api.model.impl.BookImpl;
import io.keikai.api.model.impl.SheetImpl;
import io.keikai.api.model.impl.SimpleRef;
import io.keikai.model.CellRegion;
import io.keikai.model.InvalidModelOpException;
import io.keikai.model.ModelEventListener;
import io.keikai.model.ModelEvents;
import io.keikai.model.SAutoFilter;
import io.keikai.model.SAutoFilter.NFilterColumn;
import io.keikai.model.SBook;
import io.keikai.model.SCell;
import io.keikai.model.SCellStyle;
import io.keikai.model.SColorFilter;
import io.keikai.model.SColumn;
import io.keikai.model.SColumnArray;
import io.keikai.model.SCustomFilter;
import io.keikai.model.SCustomFilters;
import io.keikai.model.SDependencyTracer;
import io.keikai.model.SDynamicFilter;
import io.keikai.model.SFill;
import io.keikai.model.SName;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.model.SSheet.SheetVisible;
import io.keikai.model.SSheetViewInfo;
import io.keikai.model.STable;
import io.keikai.model.STop10Filter;
import io.keikai.model.SheetRegion;
import io.keikai.model.impl.AbstractBookAdv;
import io.keikai.model.impl.TableImpl.DummyTable;
import io.keikai.model.sys.formula.EvaluationContributorContainer;
import io.keikai.range.SImporter;
import io.keikai.range.SImporters;
import io.keikai.range.SRange;
import io.keikai.range.SRanges;
import io.keikai.range.impl.NotifyChangeHelper;
import io.keikai.ui.au.in.Command;
import io.keikai.ui.au.out.AuCellAutoFitHeight;
import io.keikai.ui.au.out.AuCellFocus;
import io.keikai.ui.au.out.AuCellFocusTo;
import io.keikai.ui.au.out.AuDataUpdate;
import io.keikai.ui.au.out.AuHighlight;
import io.keikai.ui.au.out.AuRetrieveFocus;
import io.keikai.ui.au.out.AuSelection;
import io.keikai.ui.event.CellEvent;
import io.keikai.ui.event.CellHyperlinkEvent;
import io.keikai.ui.event.ClipboardPasteEvent;
import io.keikai.ui.event.Events;
import io.keikai.ui.event.StartEditingEvent;
import io.keikai.ui.event.StopEditingEvent;
import io.keikai.ui.event.SyncFriendFocusEvent;
import io.keikai.ui.impl.ActiveRangeHelper;
import io.keikai.ui.impl.BookCleaner;
import io.keikai.ui.impl.ComponentEvaluationContributor;
import io.keikai.ui.impl.DefaultUserActionManagerCtrl;
import io.keikai.ui.impl.DummyDataValidationHandler;
import io.keikai.ui.impl.DummyFreezeInfoLoader;
import io.keikai.ui.impl.DummyUndoableActionManager;
import io.keikai.ui.impl.Focus;
import io.keikai.ui.impl.HeaderPositionHelper;
import io.keikai.ui.impl.HeaderPositionHelper.HeaderPositionInfo;
import io.keikai.ui.impl.MergeMatrixHelper;
import io.keikai.ui.impl.MergedRect;
import io.keikai.ui.impl.SequenceId;
import io.keikai.ui.impl.SimpleCellDisplayLoader;
import io.keikai.ui.impl.VoidWidgetHandler;
import io.keikai.ui.impl.XUtils;
import io.keikai.ui.impl.undo.CellEditTextAction;
import io.keikai.ui.sys.CellDisplayLoader;
import io.keikai.ui.sys.DataValidationHandler;
import io.keikai.ui.sys.FormulaInfosLoader;
import io.keikai.ui.sys.FreezeInfoLoader;
import io.keikai.ui.sys.SpreadsheetCtrl;
import io.keikai.ui.sys.SpreadsheetCtrl.CellAttribute;
import io.keikai.ui.sys.UndoableActionManager;
import io.keikai.ui.sys.UserActionManagerCtrl;
import io.keikai.ui.sys.WidgetHandler;
import io.keikai.ui.sys.WidgetLoader;
import io.keikai.util.Ref;
import org.apache.poi.ss.SpreadsheetVersion;
import org.apache.poi.ss.formula.SheetNameFormatter;
import org.apache.poi.ss.util.AreaReference;
import org.apache.poi.ss.util.CellReference;

import org.zkoss.json.JSONArray;
import org.zkoss.json.JSONObject;
import org.zkoss.lang.Classes;
import org.zkoss.lang.Library;
import org.zkoss.lang.Objects;
import org.zkoss.lang.Strings;
import org.zkoss.util.resource.ClassLocator;
import org.zkoss.util.resource.Labels;
import org.zkoss.zk.au.AuRequest;
import org.zkoss.zk.au.AuResponse;
import org.zkoss.zk.au.DeferredValue;
import org.zkoss.zk.au.out.AuInvoke;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.Desktop;
import org.zkoss.zk.ui.Execution;
import org.zkoss.zk.ui.Executions;
import org.zkoss.zk.ui.Page;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zk.ui.WebApp;
import org.zkoss.zk.ui.event.Event;
import org.zkoss.zk.ui.event.EventListener;
import org.zkoss.zk.ui.event.SerializableEventListener;
import org.zkoss.zk.ui.ext.AfterCompose;
import org.zkoss.zk.ui.sys.ContentRenderer;
import org.zkoss.zk.ui.sys.PageCtrl;
import org.zkoss.zul.impl.XulElement;

/**
 * Spreadsheet is a rich ZK Component to handle EXCEL like behavior, it reads
 * the data from a data model({@link SBook}) and interact with this model by
 * event.<br/>
 * You can assign a Book by {@link #setBook(Book)} or just set the .xls file
 * location by {@link #setSrc(String)}. You can also set two attributes to
 * restrict max rows and columns to show on client side by
 * {@link #setMaxVisibleRows(int)} and {@link #setMaxVisibleColumns(int)}. <br/>
 * To use Spreadsheet in .zul file, just use <code>&lt;spreadsheet/&gt;</code>
 * tag like any other ZK Components.<br/>
 * An simplest example : <br/>
 * <span
 * style="font-family: courier new; font-size: 10pt;">&nbsp;&nbsp;&nbsp;&nbsp
 * ;<span style="color: rgb(0,128,128);">&lt;</span><span
 * style="color: rgb(63,127,127);">spreadsheet&nbsp;</span><span
 * style="color: rgb(127,0,127);">src</span>=<span
 * style="color: rgb(42,0,255);">
 * &quot;/WEB-INF/xls/my.xls&quot;&nbsp;</span><span
 * style="color: rgb(127,0,127);">maxrows</span>=<span
 * style="color: rgb(42,0,255);">&quot;300&quot;&nbsp;</span><span
 * style="color: rgb(127,0,127);">maxcolumns</span>=<span
 * style="color: rgb(42,0,255);">&quot;80&quot;&nbsp;<br />
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</span><span
 * style="color: rgb(127,0,127);">height</span>=<span
 * style="color: rgb(42,0,255);">&quot;400px&quot;&nbsp;</span><span
 * style="color: rgb(127,0,127);">width</span>=<span
 * style="color: rgb(42,0,255);">&quot;90%&quot;&nbsp;</span><span
 * style="color: rgb(0,128,128);">/&gt;</span></span>
 * 
 * 
 * @author dennischen
 * @author jumperchen
 */

public class Spreadsheet extends XulElement implements Serializable, AfterCompose {
	public static final String INSTANCE_KEY = "$KEIKAI_COMPONENT$";
	private static final long serialVersionUID = 1L;
	private static final String ROW_SIZE_HELPER_KEY = "_rowCellSize";
	private static final String COLUMN_SIZE_HELPER_KEY = "_colCellSize";
	private static final String MERGE_MATRIX_KEY = "_mergeRange";
	private static final String ACTIVE_RANGE_KEY = "_activeRange";

	private static final String WIDGET_HANDLER_CLS = "io.keikai.ui.sys.WidgetHandler.class";
	private static final String WIDGET_LOADERS = "io.keikai.ui.sys.WidgetLoader.class";
	
//	public static final String TOOLBAR_DISABLED_ACTION = "io.keikai.ui.ToolbarAction.disabled";
	private static final String USER_ACTION_MANAGER_CTRL_CLS = "io.keikai.ui.UserActionManagerCtrl.class";
	private static final String UNDOABLE_ACTION_MANAGER_CLS = "io.keikai.ui.UndoableActionManager.class";
	private static final String CELL_DISPLAY_LOADER_CLS = "io.keikai.ui.CellDisplayLoader.class";
	private static final String DATA_VALIDATION_HANDLER_CLS = "io.keikai.ui.DataValidationHandler.class";
	private static final String FREEZE_INFO_LOCADER_CLS = "io.keikai.ui.FreezeInfoLoader.class";
	private static final String FORMULA_INFOS_LOADER_CLS = "io.keikai.ui.FormulaInfosLoader.class";
	private static final String COLOR_PICKER_EX_USED_KEY= "io.keikai.colorPickerExUsed";

	// backward compatible with ZSS brand
	private static final String ZSS_WIDGET_HANDLER_CLS = "org.zkoss.zss.ui.sys.WidgetHandler.class";
	private static final String ZSS_WIDGET_LOADERS = "org.zkoss.zss.ui.sys.WidgetLoader.class";

	//	public static final String ZSS_TOOLBAR_DISABLED_ACTION = "org.zkoss.zss.ui.ToolbarAction.disabled";
	private static final String ZSS_USER_ACTION_MANAGER_CTRL_CLS = "org.zkoss.zss.ui.UserActionManagerCtrl.class";
	private static final String ZSS_UNDOABLE_ACTION_MANAGER_CLS = "org.zkoss.zss.ui.UndoableActionManager.class";
	private static final String ZSS_CELL_DISPLAY_LOADER_CLS = "org.zkoss.zss.ui.CellDisplayLoader.class";
	private static final String ZSS_DATA_VALIDATION_HANDLER_CLS = "org.zkoss.zss.ui.DataValidationHandler.class";
	private static final String ZSS_FREEZE_INFO_LOCADER_CLS = "org.zkoss.zss.ui.FreezeInfoLoader.class";
	private static final String ZSS_COLOR_PICKER_EX_USED_KEY= "org.zkoss.zss.colorPickerExUsed";

	private static final int DEFAULT_TOP_HEAD_HEIGHT = 20;
	private static final int DEFAULT_LEFT_HEAD_WIDTH = 36;
	/*package*/ static final int DEFAULT_CELL_PADDING = 2;
	private static final int DEFAULT_MAX_ROWS = 200;
	private static final int DEFAULT_MAX_COLUMNS = 40;
	
	private static final int DEFAULT_ROW_HEIGHT = 20;
	
	//For IE8: after 15000 cells, browser become unstable
	//For IE9: after 30000 cells, browser become slow
	//For Chrome and FF 10 : after 60000 cells, browser become slower but may acceptable
	private static final int DEFAULT_MAX_RENDERED_CELL_SIZE = 4000;
	// KEIKAI-430
	private static final String CLIPBOARD_KEY = "$keikai.clipboard$";

	transient private SBook _book; // the spreadsheet book

	private String _src; // the src to create an internal book
	transient private SImporter _importer; // the spreadsheet importer
	private int _maxRows = 0; //how many row of this spreadsheet; default to zero
	private int _maxColumns = 0; //how many column of this spreadsheet; default to zero
	//ZSS-1084
	private Map<String, int[]> _sheetMaxRowsCols; //the maximum visible rows/cols per SheetId
	
	private int _preloadRowSize = -1; //the number of row to load when receiving the rendering request
	private int _preloadColumnSize = -1; //the number of column to load when receiving the rendering request
	
	transient private SSheet _selectedSheet;
	
	private List<SSheet> _initializedSheets = new ArrayList<>();

	private boolean _hideRowhead; // hide row head
	private boolean _hideColhead; // hide column head*/
	
	private boolean _hideGridlines; //hide gridlines
	private boolean _protectSheet;
	private boolean _showFormulabar;
	private boolean _showToolbar;
	private boolean _showSheetbar;
	final private boolean _showZeroValues = getShowZeroValues();
	
	//ZSS-1082
	private boolean _showAddRow = true;
	//ZSS-1082
	private boolean _showAddColumn = true;

	// friend focus is a private api to control focus between share book;
	private Map<String, Focus> _friendFocuses = new HashMap<>(20); //id -> Focus
	private String _selfFocusId;

	private AreaRef _focusArea = new AreaRef(0, 0, 0, 0);
	private Set<AreaRefWithType> _selectionArea = new LinkedHashSet<>(1);
	private AreaRef _visibleArea = new AreaRef();
	private AreaRef _highlightArea = null;

	private WidgetHandler _widgetHandler;

	private List<WidgetLoader> _widgetLoaders;
	
	
	//a cleaner to clean book when detach or desktop cleanup
	private BookCleaner _bookCleaner;

	/**
	 * dynamic css version
	 */
	/*package*/ int _cssVersion = 0;

	/**
	 * width of left header
	 */
	private int _leftheadWidth = DEFAULT_LEFT_HEAD_WIDTH;

	/**
	 * height of top panel
	 */
	private int _topheadHeight = DEFAULT_TOP_HEAD_HEIGHT;

	/**
	 * customized row and column names.
	 */
	private Map _columnTitles;
	private Map _rowTitles;

	private ModelEventListener _modelEventListener = new InnerModelEventDispatcher(
			this);

	/**
	 * Server side customized column/row id, start with 1,3,5,7. If a
	 * column,which set from client side, the id will be 2,4,6 , check ss.js for
	 * detail
	 */
	private SequenceId _custColId = new SequenceId(-1, 2);
	private SequenceId _custRowId = new SequenceId(-1, 2);
	private SequenceId _updateCellId = new SequenceId(0, 1);// to handle batch
	private SequenceId _focusId = new SequenceId(0, 1);

	private String _userName;
	
	private boolean _clientCacheDisabled = isDefaultClientCacheDisabled();
	
	private static Boolean _defClientCache;
	
	private int _maxRenderedCellSize = getDefaultMaxRenderedCellSize();
	
	private static Integer _defMaxRenderedCellSize;
	
	private Set<AuxAction> _actionDisabled = new HashSet<>();
//	
//	private static Set<UserAction> _defToolbarActiobDisabled;
	
	private UndoableActionManager _undoableActionManager = null;
	
	private CellDisplayLoader _cellDisplayLoader = null;
	
	private DataValidationHandler _dataValidationHandler = null;
	
	private FreezeInfoLoader _freezeInfoLoader = null;
	
	transient private FormulaInfosLoader _formulaInfosLoader = null;
	
	//ZSS-1044: Whether keep the cell selection when this component lost focus;
	// default to true.
	private boolean _keepCellSelection = 
		!"false".equalsIgnoreCase(Optional.ofNullable(Library.getProperty("io.keikai.ui.keepCellSelection")).orElse(Library.getProperty("org.zkoss.zss.ui.keepCellSelection", "true")));
	
	//ZSS-1184
	// default to false
	private boolean _ignoreAutoHeight = false;

	//KEIKAI-10
	private List<ToolbarButton> _buttons;

	private List<String> _customFontFamily;

	//KEIKAI-163
	/*package*/ interface ClientCommands {
		String LINK_TO_NEW_TAB = "_linkToNewTab";
		String MAX_RENDERED_CELL_SIZE = "maxRenderedCellSize";
		String CLIENT_CACHE_DISABLED = "clientCacheDisabled";
		String ROW_FREEZE = "rowFreeze";
		String COLUMN_FREEZE = "columnFreeze";
		String AUTO_FILTER = "autoFilter";
		String TABLE_FILTERS = "tableFilters";
		String ROW_HEIGHT = "rowHeight";
		String COLUMN_WIDTH = "columnWidth";
		String DISPLAY_GRIDLINES = "displayGridlines";
		String PROTECT = "protect";
		String CUSTOM_SIZE_COLUMN = "csc";
		String CUSTOM_SIZE_ROW = "csr";
		String MERGE_RANGE = "mergeRange";
		String ACTIVE_RANGE = "activeRange";
		String DATA_VALIDATIONS = "dataValidations";
		String SHOW_TOOLBAR = "showToolbar";
		String SHOW_FORMULABAR = "showFormulabar";
		String SHOW_SHEETBAR = "showSheetbar";
		String SHOW_CONTEXT_MENU = "showContextMenu";
		String SHOW_SHEET_TAB_CONTEXT_MENU = "showSheetTabContextMenu";
		String SHOW_ZERO_VALUES = "showZeroValues";
		String MAX_ROWS = "maxRows";
		String MAX_COLUMNS = "maxColumns";
		String TOOLBAR_BUTTONS = "toolbarButtons";
		String CUSTOM_FONT_FAMILY = "customFontFamily";
		String DATA_UPDATE_STOP = "dataUpdateStop";
		String DATA_UPDATE_START = "dataUpdateStart";
		String DATA_UPDATE_CANCEL = "dataUpdateCancel";
		String DATA_UPDATE_RETRY = "dataUpdateRetry";
		String IGNORE_AUTO_HEIGHT = "ignoreAutoHeight";
		String SHOW_ADD_COLUMN = "showAddColumn";
		String SHOW_ADD_ROW = "showAddRow";
		String KEEP_CELL_SELECTION = "keepCellSelection";
		String UNLOCK_INFO = "unlockInfo";
		String OBJECT_EDITABLE = "objectEditable";
		String ACTION_DISABLED = "actionDisabled";
		String ALLOW_AUTO_FILTER = "allowAutoFilter";
		String ALLOW_FORMAT_ROWS = "allowFormatRows";
		String ALLOW_FORMAT_COLUMNS = "allowFormatColumns";
		String ALLOW_FORMAT_CELLS = "allowFormatCells";
		String ALLOW_SELECT_UNLOCKED_CELLS = "allowSelectUnlockedCells";
		String ALLOW_SELECT_LOCKED_CELLS = "allowSelectLockedCells";
		String SHEET_ID = "sheetId";
		String PRELOAD_ROW_SIZE = "preloadRowSize";
		String PRELOAD_COLUMN_SIZE = "preloadColumnSize";
		String COLOR_PICKER_EX_USED = "colorPickerExUsed";
		String SHEET_TABS = "sheetTabs";
		String TOP_PANEL_HEIGHT = "topPanelHeight";
		String LEFT_PANEL_WIDTH = "leftPanelWidth";
		String CELL_PADDING = "cellPadding";
		String FOCUS_RECT = "focusRect";
		String SELECTION_RECT = "selectionRect";
		String SEL_TYPE = "selType";
		String HIGH_LIGHT_RECT = "highLightRect";
		String COLUMN_HEAD_HIDDEN = "columnHeadHidden";
		String ROW_HEAD_HIDDEN = "rowHeadHidden";
		String INIT_ROW_SIZE = "initRowSize";
		String INIT_COLUMN_SIZE = "initColumnSize";
		String AFTER_SET = "afterSet";
		String FORMULA_INFOS = "formulaInfos";
	}

	//KEIKAI-378
	private boolean _readOnly;

	private enum SelectionMoveDirection {
		left,
		up,
		right,
		down
	}

	public Spreadsheet() {
		this._selectionArea.add(new AreaRefWithType(0, 0, 0, 0, CellSelectionType.CELL)); //ZSS-717
		this.addEventListener("onStartEditingImpl", new SerializableEventListener() {
			private static final long serialVersionUID = 2401696322103957589L;
			public void onEvent(Event event) throws Exception {
				Object[] data = (Object[]) event.getData();
				processStartEditing((String) data[0],
						(StartEditingEvent) data[1], (String) data[2]);
			}
		});
		this.addEventListener("onStopEditingImpl", new SerializableEventListener() {
			private static final long serialVersionUID = 2412586322103952998L;
			public void onEvent(Event event) throws Exception {
				Object[] data = (Object[]) event.getData();
				boolean isArrayFormula = data.length > 3 ? (boolean) data[3] : false;
				Map arrayFormulaSelection = data.length > 4 ? (Map) data[4] : Collections.EMPTY_MAP;
				processStopEditing((String) data[0], (StopEditingEvent) data[1], (String) data[2], isArrayFormula, arrayFormulaSelection);
			}
		});
		this.addEventListener("onClipboardPasteImpl", (SerializableEventListener) event -> {
			final Object[] data = (Object[]) event.getData();
			processPasteFromClipboard((ClipboardPasteEvent) data[0], (String[][]) data[1]);
		});
		//ZSS-816
		this.addEventListener(_ON_PROCESS_DEFER_OPERATIONS,  new SerializableEventListener() {
			private static final long serialVersionUID = 2401758232103952998L;
			public void onEvent(Event event) throws Exception {
				Map<String, DeferOperation> map = (Map<String, DeferOperation>) event.getData();
				if (map != null) {
					for (DeferOperation op : map.values()) {
						op.process();
					}
				}
			}
		});
		
		initComponentActionHandler();
		initToolbarButtons();
		initCustomFontFamily();
	}

	/**
	 * Trace the dependents of the reference.
	 * @param sheet
	 * @param cellRef
	 * @since 5.6.0
	 */
	public void traceDependents(Sheet sheet, CellRef cellRef) {
		Ranges.range(sheet, cellRef.getRow(), cellRef.getColumn()).traceDependents();
	}

	/**
	 * Trace the precedent of the reference.
	 * @param sheet
	 * @param cellRef
	 * @since 5.6.0
	 */
	public void tracePrecedents(Sheet sheet, CellRef cellRef) {
		Ranges.range(sheet, cellRef.getRow(), cellRef.getColumn()).tracePrecedents();
	}

	/**
	 * Clears the tracing.
	 * @since 5.6.0
	 */
	public void clearTrace() {
		SDependencyTracer sDependencyTracer = getSBook().getDependencyTracer();
		SheetRegion sheetRegion = sDependencyTracer.getSheetRegion();
		if (sheetRegion != null) {
			Ranges.range(getBook().getSheet(sheetRegion.getSheet().getSheetName()), sheetRegion.getRow(), sheetRegion.getColumn()).clearTrace();
		}
	}

	/**
	 * Gets the user action manager, then you can register/override your custom action by call {@link UserActionManager#registerHandler(String, String, UserActionHandler)}
	 * @return {@link UserActionManager} or null if doesn't support to override 
	 * @since 3.0.0
	 */
	public UserActionManager getUserActionManager(){
		UserActionManagerCtrl uamc = getUserActionManagerCtrl();
		if(uamc instanceof UserActionManager){
			return (UserActionManager)uamc;
		}
		return null;
	}
	
	private Set<String> _lastUAEvents;
	private boolean _ctrlKeysSet;
	private EventListener _uAEventDispatcher;
	
	private void initComponentActionHandler() {
		if(_uAEventDispatcher!=null && _lastUAEvents!=null){
			for(String evt:_lastUAEvents){
				this.removeEventListener(evt, _uAEventDispatcher);
			}
		}
		UserActionManagerCtrl ua = this.getUserActionManagerCtrl();
		_lastUAEvents = ua.getInterestedEvents();
		if(_lastUAEvents!=null && _lastUAEvents.size()>0){
			_uAEventDispatcher = new SerializableEventListener() {
				private static final long serialVersionUID = 2401696159873652998L;
				public void onEvent(Event event) throws Exception {
					UserActionManagerCtrl ua = getUserActionManagerCtrl();
					if(ua instanceof EventListener){
						((EventListener)ua).onEvent(event);
					}
				}
			};
			for(String evt:_lastUAEvents){
				this.addEventListener(evt, _uAEventDispatcher);
			}
		}else{
			_lastUAEvents = null;
			_uAEventDispatcher = null;
		}
		
		if(!_ctrlKeysSet){
			String ctrlKeys = ua.getCtrlKeys();
			if(ctrlKeys!=null){//null, don't set, keep the original
				super.setCtrlKeys(ctrlKeys);
			}
		}
	}
	
	/*package*/ void setUserActionManagerCtrl(UserActionManagerCtrl actionManagerCtrl) {
		if(!Objects.equals(_actionManagerCtrl,actionManagerCtrl)){
			_actionManagerCtrl = actionManagerCtrl;
			_actionManagerCtrl.bind(this);
			initComponentActionHandler();
		}
	}

	/*package*/ UserActionManagerCtrl getUserActionManagerCtrl() {
		if (_actionManagerCtrl == null) {
			String cls = finds(() -> getAttribute(USER_ACTION_MANAGER_CTRL_CLS,true),
					() -> getAttribute(ZSS_USER_ACTION_MANAGER_CTRL_CLS,true),
					() -> Library.getProperty(USER_ACTION_MANAGER_CTRL_CLS),
					() -> Library.getProperty(ZSS_USER_ACTION_MANAGER_CTRL_CLS));
			if (cls != null) {
				_actionManagerCtrl = (UserActionManagerCtrl) Spreadsheet.newInstance(cls); //ZSS-1319
			} else {
				_actionManagerCtrl = new DefaultUserActionManagerCtrl();
			}
			_actionManagerCtrl.bind(this);
		}
		return _actionManagerCtrl;
	}
	
	public void setCtrlKeys(String ctrlKeys){
		if(!_ctrlKeysSet && !Objects.equals(getCtrlKeys(),ctrlKeys)){
			_ctrlKeysSet = true;	
		}
		super.setCtrlKeys(ctrlKeys);
	}

	private static boolean isDefaultClientCacheDisabled() {
		if (_defClientCache == null)
			_defClientCache = Boolean.valueOf(Optional.ofNullable(Library.getProperty("io.keikai.spreadsheet.clientcache.disabed")).orElse(Library.getProperty("org.zkoss.zss.spreadsheet.clientcache.disabed", "false")));
		return _defClientCache;
	}

	private static boolean getShowZeroValues() {
		return Boolean.parseBoolean(Optional.ofNullable(
				Library.getProperty("io.keikai.spreadsheet.showZeroValues")).orElse("true"));
	}

	private static int getDefaultMaxRenderedCellSize() {
		if (_defMaxRenderedCellSize == null)
			_defMaxRenderedCellSize = Integer.valueOf(Optional.ofNullable(Library.getProperty("io.keikai.spreadsheet.maxRenderedCellSize")).orElse(Library.getProperty("org.zkoss.zss.spreadsheet.maxRenderedCellSize", "" + DEFAULT_MAX_RENDERED_CELL_SIZE)));
		return _defMaxRenderedCellSize;
	}
	
	/**
	 * Sets the max rendered cell size. When rendered cell size greater then this limit, 
	 * client side will prune extra cells (DOM Element). <br/>
	 * 
	 * @param maxRenderedCellSize
	 */
	public void setMaxRenderedCellSize(int maxRenderedCellSize) {
		if (_maxRenderedCellSize != maxRenderedCellSize) {
			_maxRenderedCellSize = maxRenderedCellSize;
			smartUpdate(ClientCommands.MAX_RENDERED_CELL_SIZE, maxRenderedCellSize);
		}
	}
	
	/**
	 * Returns the max rendered cell size
	 * 
	 * @return int maxRenderedCellSize
	 */
	public int getMaxRenderedCellSize() {
		return _maxRenderedCellSize;
	}
	
	/**
	 * Returns whether client cache disabled or not 
	 * 
	 * @return
	 */
	public boolean isClientCacheDisabled() {
		return _clientCacheDisabled;
	}

	/**
	 * Returns whether to show zero values or not
	 */
	public boolean isShowZeroValues() {
		return _showZeroValues;
	}
	
	/**
	 * Sets to disable client cache. Default is false
	 * @param clientCacheDisabled
	 */
	public void setClientCacheDisabled(boolean clientCacheDisabled) {
		if (_clientCacheDisabled != clientCacheDisabled) {
			_clientCacheDisabled = clientCacheDisabled;
			smartUpdate(ClientCommands.CLIENT_CACHE_DISABLED, _clientCacheDisabled);
		}
	}

	/**
	 * Don't call this, the spreadsheet is not draggable.
	 * 
	 * @exception UnsupportedOperationException if this method is called.
	 */
	public void setDraggable(String draggable) {
		throw new UnsupportedOperationException("doesn't support to be draggable");
	}

	/**
	 * Returns the book model of this Spreadsheet. If you call this method at
	 * first time and the book has not assigned by {@link #setSBook(SBook)}, this
	 * will create a new model depends on src;
	 * 
	 * @return the book model of this spreadsheet.
	 */
	public SBook getSBook() {
		if (_book == null) {
			if (_src == null) {
				return null;
			}
			try {
				SImporter importer = _importer;
				if (importer == null) {
					importer = SImporters.getImporter();
				}

				SBook book;
				{
					URL url = null;
					if (_src.startsWith("/")) {// try to load by application
						// context.
						WebApp wapp = Executions.getCurrent().getDesktop().getWebApp();
						String path = wapp.getRealPath(_src);
						if (path != null) {
							File file = new File(path);
//							if (file.isDirectory())
//								throw new IllegalArgumentException("Your input source is a directory, not a vaild file");
							if (file.exists())
								url = file.toURI().toURL();
						} else
							url = wapp.getResource(_src); 
					}
					if (url == null) {// try to load from class loader
						url = new ClassLocator().getResource(_src);
					}
					if (url == null) {// try to load from file
						File f = new File(_src);
						if (f.exists()) {
							url = f.toURI().toURL();
						}
					}

					if (url == null) {
						throw new UiException("resource for " + _src + " not found.");
					}

					String bookName = url.getFile();
					int i = bookName.lastIndexOf('/');
					if(i >=0){
						bookName = bookName.substring(i+1);
					}
					book = importer.imports(url,bookName);
				}
				initBook(book); //will set _book inside this method
			} catch (Exception ex) {
				throw UiException.Aide.wrap(ex);
			}
		}
		return _book;
	}

	/**
	 * Sets the book data model of this spread sheet.
	 * @param book the book data model.
	 */
	public void setSBook(SBook book) {
		if (!Objects.equals(book, _book)) {
			initBook0(book);
			invalidate();
		}
	}
	
	private void initBook(SBook book) {
		if (!Objects.equals(book, _book)) {
			initBook0(book);
		}
	}
	private Focus _selfEditorFocus;

	/*package*/ Focus getSelfEditorFocus() {
		return _selfEditorFocus;
	}

	private static final String FRIEND_FOCUS_KEY = "keikai.FirendFocus";
	
	private FriendFocusHelper getFriendFocusHelper(){
		return guardWithWriteLock(() -> {
			FriendFocusHelper helper = (FriendFocusHelper) _book.getAttribute(FRIEND_FOCUS_KEY);
			if(helper==null){
				_book.setAttribute(FRIEND_FOCUS_KEY, helper = new FriendFocusHelper());
			}
			return helper;
		});
	}

	private <T> T guardWithWriteLock(Supplier<T> run) {
		if (_book != null) {
			_book.getBookSeries().getLock().writeLock().lock();
			try {
				return run.get();
			} finally {
				_book.getBookSeries().getLock().writeLock().unlock();
			}
		}
		return null;
	}
	
	private void deleteSelfEditorFocus() {
		if (_selectedSheet != null && getSBook().getSheetIndex(_selectedSheet) != -1  && _selfEditorFocus != null) {
			final SRange rng = SRanges.range(_selectedSheet);
			getFriendFocusHelper().removeFocus(_selfEditorFocus);
			rng.notifyCustomEvent(ModelEvents.ON_MODEL_FRIEND_FOCUS_DELETE,_selfEditorFocus,true);
			_selfEditorFocus = null;
		}
	}
	private void moveSelfEditorFocus(String sheetId,int row, int column) {
		if (_selectedSheet != null) {
			if (_selfEditorFocus == null) {
				_selfEditorFocus = newSelfFocus(sheetId,row,column);
				getFriendFocusHelper().addFocus(_selfEditorFocus);
			}else{
				_selfEditorFocus.setSheetId(sheetId);
				_selfEditorFocus.setPosition(row, column);
			}
			final SRange rng = SRanges.range(_selectedSheet);
			rng.notifyCustomEvent(ModelEvents.ON_MODEL_FRIEND_FOCUS_MOVE,_selfEditorFocus,true);
		}
		syncFriendFocus();
	}
	private EventListener<Event> _focusListener = null;

	private UserActionManagerCtrl _actionManagerCtrl;

	private boolean _showContextMenu;
	private boolean _showSheetTabContextMenu;
	private void doMoveSelfFocus(CellEvent event){
		moveSelfEditorFocus(getSelectedSheetId(),event.getRow(),event.getColumn());
	}
	
	//ZSS-406 Spreadsheet doesn't be release when use application share scope
	/*package*/ void releaseBook() {
		_book = guardWithWriteLock(() -> {
			_book.removeEventListener(_modelEventListener);
			if(_book instanceof EvaluationContributorContainer
					&& ((EvaluationContributorContainer)_book).getEvaluationContributor() instanceof ComponentEvaluationContributor){
				((EvaluationContributorContainer)_book).setEvaluationContributor(null);
			}
			//delete self focus that stores in the book.
			deleteSelfEditorFocus();
			return null; // reset to null
		});
	}
	
	private void initBook0(SBook book) {
		guardWithWriteLock(() -> {
			_book.removeEventListener(_modelEventListener);
			if(_book instanceof EvaluationContributorContainer
					&& ((EvaluationContributorContainer)_book).getEvaluationContributor() instanceof ComponentEvaluationContributor){
				((EvaluationContributorContainer)_book).setEvaluationContributor(null);
			}
			if (_focusListener != null){
				removeEventListener(Events.ON_CELL_FOCUS, _focusListener);
				_focusListener = null;
			}
			deleteSelfEditorFocus();
			return null;
		});
		
		 //Shall clean selected sheet before set new book (ZSS-75: set book null, cause NPE)
		cleanSelectedSheet();
		
		removeAttribute(MERGE_MATRIX_KEY);
		clearHeaderSizeHelper(true, true);
		//ZSS-343 ActiveRangeHelper caches sheet object and doesn't release after reload a book
		removeAttribute(ACTIVE_RANGE_KEY);
		
		_custColId = new SequenceId(-1, 2);
		_custRowId = new SequenceId(-1, 2);
		
		//clear undo history
		clearUndoableActionManager();
		
		//ZSS-1084: initialize sheetMaxRowsCols
		_sheetMaxRowsCols = new HashMap<>();
		
		_book = book;
		guardWithWriteLock(() -> {
			_book.addEventListener(_modelEventListener);
			if(_book instanceof EvaluationContributorContainer
					&& ((EvaluationContributorContainer)_book).getEvaluationContributor()==null){
				((EvaluationContributorContainer)_book).setEvaluationContributor(new ComponentEvaluationContributor(this));
			}
			return null;
		});
		if (_book != null) {
			//20130523, dennis, if share-scope is not empty, then should always sync the  focus, not only application and session
			//TODO use a configuration to config this.
			if (!Strings.isEmpty(_book.getShareScope())) { //have to sync focus
				this.addEventListener(Events.ON_CELL_FOCUS, _focusListener = new SerializableEventListener() {
					private static final long serialVersionUID = 2716358947569822998L;
					@Override
					public void onEvent(Event event) throws Exception {
						doMoveSelfFocus((CellEvent) event);
					}
				});
			}
			updateBookCircularIteration();
		}
		_selfFocusId = null;//clean
		refreshToolbarDisabled();
		getUserActionManagerCtrl().doAfterLoadBook(getBook());
	}
	
	private Focus newSelfFocus(String sheetId, int row, int column) {
		//show id/uuid is useless for co-edit and ugly
		final String focusName = _userName == null ? ""/*+ (getId() == null ? getUuid() : getId()) */: _userName;
		_selfFocusId = _selfFocusId==null?getFriendFocusHelper().nextFocusId():_selfFocusId;
		return new Focus(_selfFocusId, focusName, "#000", sheetId,row,column,
				this);
	}

	/**
	 * Gets the selected sheet, the default selected sheet is first visible sheet.
	 * @return #{@link SSheet}
	 */	
	public SSheet getSelectedSSheet() {
		final SBook book = getSBook();
		if (book == null) {
			return null;
		}
		if (_selectedSheet == null) {
			if (book.getNumOfSheet() == 0)
				throw new UiException("sheet size of given book is zero");
			SSheet firstVisibleSheet = null;
			for (SSheet sheet : book.getSheets()) {
				if (sheet.getSheetVisible() == SheetVisible.VISIBLE) {
					if (firstVisibleSheet == null) {
						firstVisibleSheet = sheet;
					}
					if (!sheet.getViewInfo().isTabSelected()) {
						continue;
					}

					_selectedSheet = sheet;

					// KEIKAI-572
					SSheetViewInfo viewInfo = sheet.getViewInfo();
					Ref areaRef = a1ToRef(viewInfo.getActiveCell());
					_focusArea.setArea(areaRef.getTop(), areaRef.getLeft(), areaRef.getBottom(), areaRef.getRight());
					_selectionArea.clear();
					Set<AreaRef> areaRefs = new HashSet<>();
					String[] areaArray = viewInfo.getSelectionAreas().split(" ");
					for (String area : areaArray) {
						areaRef = a1ToRef(area);
						_selectionArea.add(new AreaRefWithType(areaRef.getTop(), areaRef.getLeft(), areaRef.getBottom(), areaRef.getRight(), CellSelectionType.CELL));
					}

					_initializedSheets.add(sheet);
					break;
				}
			}
			// if selectedSheet is null, the firstVisibleSheet is assumed.
			if (_selectedSheet == null) {
				_selectedSheet = firstVisibleSheet;
			}

			afterSheetSelected();
		}
		return _selectedSheet;
	}

	/*package*/ String getSelectedSheetId() {
		SSheet sheet = getSelectedSSheet();
		return sheet==null?null:sheet.getId();
	}

	/**
	 * Returns the src location of book model. This src is used by the specified
	 * importer to create the book data model of this spread sheet.
	 * 
	 * @return the src location
	 */
	public String getSrc() {
		return _src;
	}


	/**
	 * Sets the src location of the book data model to be imported into
	 * spreadsheet. A specified importer ({@link #getImporter}) will use this
	 * src to create the book data model.
	 * 
	 * @param src  the book src location
	 */
	public void setSrc(String src) {
		if (!Objects.equals(_src, src)) {
			_src = src;
			setBook(null);
			invalidate();
		}
	}

	/**
	 * Sets the selected sheet by a name
	 * @param name	the name of spreadsheet to be selected.
	 */
	public void setSelectedSheet(String name) {
		boolean update = setSelectedSheet0(name);
		
		//TODO: think if this is correct or not
		// the call of onSheetSelected must after invalidate,
		// because i must let invalidate clean lastcellblock first
		if(update){
			afterSheetSelected();
			invalidate();
		}
	}
	
	/**
	 * @return true if selected another sheet
	 */
	private boolean setSelectedSheet0(String name) {
		final SBook book = getSBook();
		if (book == null) {
			return false;
		}
		boolean update = false;
		//Note. check whether if the sheet has remove or not
		if (_selectedSheet != null && book.getSheetIndex(_selectedSheet) == -1) {
			cleanSelectedSheet();
			//_selectedSheet become null after clean
			_initializedSheets.remove(_selectedSheet);
			update = true;
		}

		if (_selectedSheet == null || !_selectedSheet.getSheetName().equals(name)) {
			SSheet sheet = book.getSheetByName(name);
			if (sheet == null) {
				throw new UiException("No such sheet : " + name);
			}
			if (_selectedSheet != null) {
				_selectedSheet.getViewInfo().setTabSelected(false);
			}

			cleanSelectedSheet();


			_selectedSheet = sheet;

			// KEIKAI-572
			SSheetViewInfo viewInfo = sheet.getViewInfo();
			viewInfo.setTabSelected(true);

			Ref areaRef = a1ToRef(viewInfo.getActiveCell());
			_focusArea.setArea(areaRef.getTop(), areaRef.getLeft(), areaRef.getBottom(), areaRef.getRight());
			_selectionArea.clear();
			Set<AreaRef> areaRefs = new HashSet<>();
			String[] areaArray = viewInfo.getSelectionAreas().split(" ");
			for (String area : areaArray) {
				areaRef = a1ToRef(area);
				_selectionArea.add(new AreaRefWithType(areaRef.getTop(), areaRef.getLeft(), areaRef.getBottom(), areaRef.getRight(), CellSelectionType.CELL));
			}

			if (!_initializedSheets.contains(_selectedSheet)) {
				_initializedSheets.add(_selectedSheet);
			}
			update = true;
		}
		
		return update;
	}
	
	/*
	 * DefaultUserActionManagerCtrl will update highlight area when selecting a sheet.
	 */
	/*package*/ void setSelectedSheetDirectly(String name, boolean cacheInClient, int row, int col,
			int left, int top, int right, int bottom
			/*, int highlightLeft, int highlightTop, int highlightRight, int highlightBottom,
			int rowfreeze, int colfreeze*/) {
		// KEIKAI-620, KEIKAI-622: when switching sheets, clear the visible area.
		// Since `_visibleArea` is shared among all sheets, `_visibleArea` is considered stale upon switching sheets.
		// `setSelectedSheetDirectly` is called upon `onCellHyperlink` (KEIKAI-620) and `onSheetSelected` (KEIKAI-622).
		_visibleArea = new AreaRef();

		boolean shouldInvalidate = true;
		for (SSheet s : _initializedSheets) {
			if (s.getSheetName() == name) {
				shouldInvalidate = false;
			}
		}
		boolean update = setSelectedSheet0(name);
		if (row >= 0 && col >= 0) {
			this.setCellFocusDirectly(new CellRef(row, col));
		}

		if (top >= 0 && right >= 0 && bottom >= 0 && left >=0) {
			this.setSelectionDirectly(new AreaRef(top, left, bottom, right));
		}

		if (update) {
			afterSheetSelected();
			if (shouldInvalidate) {
				invalidate();
			}
		}
		
		updateSheetAttributes(cacheInClient/*, rowfreeze, colfreeze*/);
		
		//ZSS-1040: must sync here because customRow/customHeight not ready yet
		syncFriendFocus(true);
	}
	
	private void updateSheetAttributes(boolean cacheInClient/*, int rowfreeze, int colfreeze*/) {
		
		SSheet sheet = _selectedSheet;

		if (!cacheInClient)	{
			
			smartUpdate(ClientCommands.ROW_FREEZE, getSelectedSheetRowfreeze());
			smartUpdate(ClientCommands.COLUMN_FREEZE, getSelectedSheetColumnfreeze());
			
			//handle AutoFilter
			Map afmap = convertAutoFilterToJSON(sheet.getAutoFilter());
			if (afmap != null) {
				smartUpdate(ClientCommands.AUTO_FILTER, afmap);
			} else {
				smartUpdate(ClientCommands.AUTO_FILTER, null);
			}
			// KEIKAI-224: should check auto filter bottom region when first loaded
			((SpreadsheetCtrlImpl) getExtraCtrl()).calcAutoFilterBottomRegion(sheet);
			
			//ZSS-988
			//handle Table's AutoFilter
			Map<String, Map> tbafsmap = convertTableFiltersToJSON(sheet);
			smartUpdate(ClientCommands.TABLE_FILTERS, tbafsmap);

			smartUpdate(ClientCommands.ROW_HEIGHT, getRowheight());
			smartUpdate(ClientCommands.COLUMN_WIDTH, getColumnwidth());

			smartUpdate(ClientCommands.DISPLAY_GRIDLINES, !_hideGridlines);
			refreshAllowedOptions();
			updateUnlockInfo();
			smartUpdate(ClientCommands.PROTECT, isProtectSheetOrReadOnly());

			// generate customized row & column information
			HeaderPositionHelper colHelper = getColumnPosition(sheet);
			HeaderPositionHelper rowHelper = getRowPosition(sheet);
			smartUpdate(ClientCommands.CUSTOM_SIZE_COLUMN, getSizeHelperStr(colHelper));
			smartUpdate(ClientCommands.CUSTOM_SIZE_ROW, getSizeHelperStr(rowHelper));
			
			// generate merge range information
			MergeMatrixHelper mmhelper = getMergeMatrixHelper(sheet);
			Iterator iter = mmhelper.getRanges().iterator();
			StringBuilder merr = new StringBuilder();
			while (iter.hasNext()) {
				MergedRect block = (MergedRect) iter.next();
				int left = block.getColumn();
				int top = block.getRow();
				int right = block.getLastColumn();
				int bottom = block.getLastRow();
				int id = block.getId();
				merr.append(left).append(",").append(top).append(",").append(right).append(",").append(bottom).append(",").append(id);
				if (iter.hasNext()) {
					merr.append(";");
				}
			}
			smartUpdate(ClientCommands.MERGE_RANGE, merr.toString());
			
			// ZSS-423: create multiple "active range" data for every panel
			JSONObject activeRange = createActiveRange(SpreadsheetCtrlImpl.newInstance(this), sheet, getInitColumnSize(), getInitRowSize());
			smartUpdate(ClientCommands.ACTIVE_RANGE, activeRange);
			
			//handle Validation, must after render("activeRange"
			List<Map<String, Object>> dvs = getDataValidationHandler().loadDataValidtionJASON(getSelectedSheet());
			smartUpdate(ClientCommands.DATA_VALIDATIONS, dvs);
		}

		// KEIKAI-183: selecting a sheet doesn't work for the first time.
		// Keep the SHEET_ID command after the ACTIVE_RANGE command.
		smartUpdate(ClientCommands.SHEET_ID, getSelectedSheetId());

	}

	public void setPreloadRowSize(int size) {
		if (_preloadRowSize != size) {
			_preloadRowSize = Math.max(size, 0);
			smartUpdate(ClientCommands.PRELOAD_ROW_SIZE, _preloadRowSize);
		}
	}
	
	public int getPreloadRowSize() {
		return _preloadRowSize;
	}
	
	/**
	 * Returns the number of rows rendered when Spreadsheet first render
	 * @return int
	 */
	private int getInitRowSize() {
		//ZSS-1215: invalidate() would call into this via renderProperties0() to setup the activeRange
		//Thus we have to use _visibleArea information when doing invalidate()
		final AreaRef vArea = _visibleArea;
		int rowSize = vArea.getLastRow() < 0 ?  // _visibleArea is not sync yet
				SpreadsheetCtrl.DEFAULT_LOAD_ROW_SIZE : vArea.getLastRow();
		int preloadRowSize = getPreloadRowSize();
		if (preloadRowSize < 0) { // preloadRowSize is not set 
			rowSize = Math.min(rowSize, getCurrentMaxVisibleRows() - 1); //ZSS-1084
		} else { // preloadRowSize is set; use it
			rowSize = Math.min(rowSize + preloadRowSize, getCurrentMaxVisibleRows() - 1); //ZSS-1084
		}
		return rowSize;
	}
	
	public void setPreloadColumnSize(int size) {
		if (_preloadColumnSize != size) {
			_preloadColumnSize = Math.max(size, 0);
			smartUpdate(ClientCommands.PRELOAD_COLUMN_SIZE, _preloadColumnSize); //ZSS-1215
		}
	}
	
	public int getPreloadColumnSize() {
		return _preloadColumnSize;
	}
	
	/**
	 * Returns the number of columns rendered when Spreadsheet first render
	 * @return int
	 */
	private int getInitColumnSize() {
		//ZSS-1215: invalidate() would call into this via renderProperties0() to setup the activeRange
		//Thus we have to use _visibleArea information when doing invalidate()
		final AreaRef vArea = _visibleArea;
		int colSize = vArea.getLastColumn() < 0 ? // _visibleArea is not sync yet 
			SpreadsheetCtrl.DEFAULT_LOAD_COLUMN_SIZE : vArea.getLastColumn();
		int preloadColSize = getPreloadColumnSize();
		if (preloadColSize < 0) { // preloadColSize is not set
			colSize = Math.min(colSize, getCurrentMaxVisibleColumns() - 1); //ZSS-1084
		} else { // preloadColSize is set; use it
			colSize = Math.min(colSize + preloadColSize, getCurrentMaxVisibleColumns() - 1); //ZSS-1084
		}
		return colSize;
	}

	final /*package*/ int getSelectedSheetRowfreeze(){
		return getFreezeInfoLoader().getRowFreeze(getSelectedSheet());
	}
	
	final /*package*/ int getSelectedSheetColumnfreeze(){
		return getFreezeInfoLoader().getColumnFreeze(getSelectedSheet());
	}

	/**
	 * Returns true if hide row head of this spread sheet; default to false.
	 * @return true if hide row head of this spread sheet; default to false.
	 */
	public boolean isHiderowhead() {
		return _hideRowhead;
	}

	/**
	 * Sets true to hide the row head of this spread sheet.
	 * @param hide true to hide the row head of this spread sheet.
	 */
	public void setHiderowhead(boolean hide) {
		if (_hideRowhead != hide) {
			_hideRowhead = hide;
			invalidate();
		}
	}

	/**
	 * Returns true if hide column head of this spread sheet; default to false.
	 * @return true if hide column head of this spread sheet; default to false.
	 */
	public boolean isHidecolumnhead() {
		return _hideColhead;
	}

	/**
	 * Sets true to hide the column head of this spread sheet.
	 * @param hide true to hide the row head of this spread sheet.
	 */
	public void setHidecolumnhead(boolean hide) {
		if (_hideColhead != hide) {
			_hideColhead = hide;
			invalidate();
		}
	}

	/**
	 * Sets the customized titles of column header.
	 * 
	 * @param titles a map for customized column titles, the key of column must be Integer object.
	 */
	public void setColumntitles(Map titles) {
		if (!Objects.equals(titles, _columnTitles)) {
			_columnTitles = titles;
			invalidate();
		}
	}

	/**
	 * Get the column titles Map object, modification of return object doesn't
	 * cause any update.
	 * 
	 * @return Map object of customized column names
	 */
	public Map getColumntitles() {
		return _columnTitles;
	}

	/**
	 * Sets the customized titles which split by ','. For example:
	 * "name,title,age" or "name,,age" , an empty string means use default
	 * column title. This method will split the input string to a Map with
	 * sequence index from 0, then call {@link #setColumntitles(Map)}<br/>
	 * 
	 * <p/>
	 * Note: this method will always invoke invalidate()
	 * 
	 * @param titles  the column titles
	 */
	public void setColumntitles(String titles) {
		String[] names = titles.split(",");
		Map map = new HashMap();
		for (int i = 0; i < names.length; i++) {
			if (names[i].length() > 0)
				map.put(Integer.valueOf(i), names[i]);
		}
		if (map.size() > 0)
			setColumntitles(map);

	}

	/**
	 * Sets the customized titles of row header.
	 * 
	 * @param titles map for customized column titles, the key of column must be Integer object.
	 */
	public void setRowtitles(Map titles) {
		if (!Objects.equals(titles, _rowTitles)) {
			_rowTitles = titles;
			invalidate();
		}
	}

	/**
	 * Get the row titles Map object, modification of the return object doesn't
	 * cause any update.
	 * 
	 * @return Map object of customized row names
	 */
	public Map getRowtitles() {
		return _rowTitles;
	}

	/**
	 * Sets the customized titles which split by ','. For example:
	 * "name,title,age" or "name,,age" , an empty string means use default row
	 * title. This method will split the input string to a Map with sequence
	 * index from 0, then call {@link #setRowtitles(Map)}<br/>
	 * 
	 * <p/>
	 * Note: this method will always invoke invalidate()
	 * 
	 * @param titles the row titles
	 */
	public void setRowtitles(String titles) {
		String[] names = titles.split(",");
		Map map = new HashMap();
		for (int i = 0; i < names.length; i++) {
			if (names[i].length() > 0)
				map.put(Integer.valueOf(i), names[i]);
		}
		if (map.size() > 0)
			setRowtitles(map);
	}

	/**
	 * Gets the default row height of the selected sheet
	 * @return default value depends on selected sheet
	 */
	public int getRowheight() {
		SSheet sheet = getSelectedSSheet();
		int rowHeight = sheet != null ? sheet.getDefaultRowHeight() : -1;

		return (rowHeight <= 0) ? DEFAULT_ROW_HEIGHT : rowHeight;
	}

	
	/**
	 * Sets the default row height of the selected sheet
	 * @param rowHeight the row height
	 */
	public void setRowheight(int rowHeight) {
		SSheet sheet = getSelectedSSheet();
		
		int dh = sheet.getDefaultRowHeight();

		if (dh != rowHeight) {
			sheet.setDefaultRowHeight(rowHeight);
			invalidate();
		}
	}
	
	/**
	 * Gets the default column width of the selected sheet
	 * @return default value depends on selected sheet
	 */
	public int getColumnwidth() {
		final SSheet sheet = getSelectedSSheet();
		return sheet.getDefaultColumnWidth();//XUtils.getDefaultColumnWidthInPx(sheet);
	}

	/**
	 * Sets the default column width of the selected sheet
	 * @param columnWidth the default column width
	 */
	public void setColumnwidth(int columnWidth) {
		final SSheet sheet = getSelectedSSheet();
		int dw = sheet.getDefaultColumnWidth();
		if (dw != columnWidth) {
			sheet.setDefaultColumnWidth(columnWidth);
			
			invalidate();
		}
	}

	/**
	 * Gets the left head panel width
	 * @return default value is 36
	 */
	public int getLeftheadwidth() {
		return _leftheadWidth;
	}
	
	/**
	 * rename setLeftheadwidth -> setLeftheadWidth
	 */
	/**
	 * Sets the left head panel width, must large then 0.
	 * @param leftWidth leaf header width
	 */
	public void setLeftheadwidth(int leftWidth) {
		if (_leftheadWidth != leftWidth) {
			_leftheadWidth = leftWidth;
			invalidate();
		}
	}

	/**
	 * Gets the top head panel height
	 * @return default value is 20
	 */
	public int getTopheadheight() {
		return _topheadHeight;
	}

	/**
	 * Sets the top head panel height, must large then 0.
	 * @param topHeight top header height
	 */
	public void setTopheadheight(int topHeight) {
		if (_topheadHeight != topHeight) {
			_topheadHeight = topHeight;
			invalidate();
		}
	}
	
	/**
	 * Sets whether show toolbar or not
	 * 
	 * Default: false
	 * @param showToolbar true to show toolbar
	 */
	public void setShowToolbar(boolean showToolbar) {
		if (_showToolbar != showToolbar) {
			_showToolbar = showToolbar;
			smartUpdate(ClientCommands.SHOW_TOOLBAR, _showToolbar);
		}
	}
	
	/**
	 * Returns whether shows toolbar
	 * @return boolean
	 */
	public boolean isShowToolbar() {
		return _showToolbar;
	}
	
	/**
	 * Sets whether show formula bar or not
	 * @param showFormulabar true if want to show formula bar
	 */
	public void setShowFormulabar(boolean showFormulabar) {
		if (_showFormulabar != showFormulabar) {
			_showFormulabar = showFormulabar;
			smartUpdate(ClientCommands.SHOW_FORMULABAR, _showFormulabar);
		}
	}
	
	/**
	 * Returns whether show formula bar
	 */
	public boolean isShowFormulabar() {
		return _showFormulabar;
	}
	
	/**
	 * Sets whether show sheetbar or not
	 * @param showSheetbar true if want to show sheet tab panel
	 */
	public void setShowSheetbar(boolean showSheetbar) {
		if (_showSheetbar != showSheetbar) {
			_showSheetbar = showSheetbar;
			smartUpdate(ClientCommands.SHOW_SHEETBAR, _showSheetbar);
		}
	}
	
	/**
	 * Returns whether show sheetbar
	 */
	public boolean isShowSheetbar() {
		return _showSheetbar;
	}
	
	/**
	 * Sets whether show ContextMenu or not
	 * @param showContextMenu
	 */
	public void setShowContextMenu(boolean showContextMenu) {
		if (_showContextMenu != showContextMenu) {
			_showContextMenu = showContextMenu;
			smartUpdate(ClientCommands.SHOW_CONTEXT_MENU, _showContextMenu);
		}
	}

	/**
	 * Returns whether to show ContextMenu or not
	 */
	public boolean isShowContextMenu() {
		return _showContextMenu;
	}
	
	/**
	 * Returns whether show ContextMenu in sheet tab
	 * @return
	 */
	public boolean isShowSheetTabContextMenu() {
		return _showSheetTabContextMenu;
	}

	/**
	 * Sets whether show ContextMenu or not
	 * @param showSheetTabContextMenu
	 */
	public void setShowSheetTabContextMenu(boolean showSheetTabContextMenu) {
		if (_showSheetTabContextMenu != showSheetTabContextMenu) {
			_showSheetTabContextMenu = showSheetTabContextMenu;
			smartUpdate(ClientCommands.SHOW_SHEET_TAB_CONTEXT_MENU, showSheetTabContextMenu);
		}
	}
	
	//See autofilter.js#setAutoFilter()
	private Map convertAutoFilterToJSON(SAutoFilter af) {
		if (af != null) {
			final CellRegion addr = af.getRegion();
			if (addr == null) {
				return null;
			}
			final Map addrmap = new HashMap();
			final int left = addr.getColumn();
			final int right = addr.getLastColumn();
			final int top = addr.getRow();
			final SSheet sheet = this.getSelectedSSheet();
			addrmap.put("left", left);
			addrmap.put("top", top);
			addrmap.put("right", right);
			addrmap.put("bottom", addr.getLastRow());
			
			final Collection<NFilterColumn> fcs = af.getFilterColumns();
			final List<Map> fcsary = fcs != null ? new ArrayList<Map>(fcs.size()) : null;
			if (fcsary != null) {
				List<String> filters = null;
				
				boolean on = true;
				int field = 0;
				for(int col = left; col <= right; ++col) {
					final NFilterColumn fc = af.getFilterColumn(col - left,false);
					if (fc == null) {
						on = true;
						continue; //ZSS-705: no filterColumn; default on and skip
					}
					
					//ZSS-1191
					boolean byFontColor = false;
					String pattern = null;
					String fgColor = null;
					String bgColor = null;

					//ZSS-1192
					boolean isAnd = false;
					SCustomFilter f1 = null;
					SCustomFilter f2 = null;
					
					//ZSS-1234
					boolean dynaf = false;
					
					//ZSS-1193
					boolean isAbove = false;
					Double avgVal = null;
					
					//ZSS-1193
					boolean isTop = false;
					boolean isPercent = false;
					Double filterVal = null;
					int countVal = 0;
					
					if (on) { // ZSS-705: only when previous showButton is on
						filters = fc.getFilters();
						on = fc.isShowButton();
						field = col - left + 1;
						
						//ZSS-1191
						SColorFilter cfilter = fc.getColorFilter();
						if (cfilter != null) {
							final SFill fill = cfilter.getExtraStyle().getFill();
							pattern = fill.getFillPattern().name();
							fgColor = fill.getFillColor().getHtmlColor();
							bgColor = fill.getBackColor().getHtmlColor();
							byFontColor = cfilter.isByFontColor();
						}
						
						//ZSS-1192
						SCustomFilters cusFilters = fc.getCustomFilters();
						if (cusFilters != null) {
							isAnd = cusFilters.isAnd();
							f1 = cusFilters.getCustomFilter1();
							f2 = cusFilters.getCustomFilter2();
						}
						
						//ZSS-1193
						SDynamicFilter dynaFilter = fc.getDynamicFilter();
						if (dynaFilter != null) {
							dynaf = true; //ZSS-1234
							isAbove = "aboveAverage".equals(dynaFilter.getType());
							avgVal = dynaFilter.getValue();
						}
						
						//ZSS-1193
						STop10Filter top10Filter = fc.getTop10Filter();
						if (top10Filter != null) {
							isTop = top10Filter.isTop();
							isPercent = top10Filter.isPercent();
							filterVal = top10Filter.getFilterValue();
							countVal = (int) top10Filter.getValue();
						}
					} // ZSS-705: if previous showButton is off; use previous field and filters(in merged cell case)
					
					Map fcmap = new HashMap();
					fcmap.put("col", Integer.valueOf(col));
					fcmap.put("filter", filters);
					fcmap.put("on", on);
					fcmap.put("field", field);
					
					//ZSS-1911: in filter; fg/bg should reversed to follow SCellStyle in CELL_COLOR operation
					if ("SOLID".equals(pattern) && !byFontColor) {
						final String tmp = fgColor;
						fgColor = bgColor;
						bgColor = tmp;
					}
					fcmap.put("colorpat", pattern);
					fcmap.put("colorfg", fgColor); 
					fcmap.put("colorbg", bgColor);
					fcmap.put("fontColor", byFontColor);
					
					//ZSS-1192
					if (f1 != null) {
						fcmap.put("and", isAnd);
						Map f1map = new HashMap();
						fcmap.put("f1", f1map);
						
						f1map.put("val", f1.getValue());
						f1map.put("op", f1.getOperator().name());
						
						if (f2 != null) {
							Map f2map = new HashMap();
							fcmap.put("f2", f2map);
							
							f2map.put("val", f2.getValue());
							f2map.put("op", f2.getOperator().name());
						}
					}
					
					//ZSS-1234
					if (dynaf) {
						fcmap.put("dynaf", Boolean.valueOf(dynaf));
						//ZSS-1193: DynamicFilter (aboveAverage/belowAverage)
						if (avgVal != null) {
							fcmap.put("avgVal", avgVal);
							fcmap.put("isAbove", isAbove);
						}
					}
					
					//ZSS-1193: Top10Filter
					if (filterVal != null) {
						fcmap.put("filterVal", filterVal);
						fcmap.put("isTop", isTop);
						fcmap.put("isPercent", isPercent);
						fcmap.put("countVal", countVal);
					}
					
					fcsary.add(fcmap);
				}
			}
			
			final Map afmap = new HashMap();
			afmap.put("range", addrmap);
			afmap.put("filterColumns", fcsary);
			return afmap;
		}
		return null;
	}

	//ZSS-988
	private Map<String, Map> convertATableFilterToJSON(STable table) {
		final AbstractBookAdv book = (AbstractBookAdv) table.getAllRegion().getSheet().getBook();
		Map<String, Map> tbmap = new HashMap<String, Map>();
		Map tafmap = table instanceof DummyTable || book.getTable(table.getName()) == null ? null : //ZSS-988: handle table deleted
			convertAutoFilterToJSON(table.getAutoFilter());
		tbmap.put(table.getName(), tafmap);
		return tbmap;
	}
	
	//ZSS-988
	private Map<String, Map> convertTableFiltersToJSON(SSheet sheet) {
		final AbstractBookAdv book = (AbstractBookAdv) sheet.getBook();
		Map<String, Map> tbmap = new HashMap<String, Map>();
		for (STable table : sheet.getTables()) {
			Map tafmap = book.getTable(table.getName()) == null ? null :  //ZSS-988: handle table deleted
					convertAutoFilterToJSON(table.getAutoFilter());
			tbmap.put(table.getName(), tafmap);
		}
		return tbmap.isEmpty() ? null : tbmap;
	}

	//ZSS-13: Support Open hyperlink in a separate browser tab window
	private boolean getLinkToNewTab() {
		final String linkToNewTab = Optional.ofNullable(Library.getProperty("io.keikai.ui.Spreadsheet.linkToNewTab")).orElse(Library.getProperty("org.zkoss.zss.ui.Spreadsheet.linkToNewTab", "true"));
		return Boolean.valueOf(linkToNewTab);
	}
	
	/**
	 * Returns each sheet's name and sheet uuid
	 * @return
	 */
	/*package*/ List<LinkedHashMap<String, String>> getSheetTabs() {
		int len = _book.getNumOfSheet();
		List<LinkedHashMap<String, String>> ary = new ArrayList<LinkedHashMap<String, String>>(len);

		for (int i = 0; i < len; i++) {
			//key: sheet names, value: sheet uuid
			LinkedHashMap<String, String> sheetTabs = new LinkedHashMap<String, String>();
			
			SSheet sheet = _book.getSheet(i);
			
			if(sheet.getSheetVisible() == SheetVisible.VISIBLE) {
				sheetTabs.put("id", sheet.getId());
				sheetTabs.put("name", sheet.getSheetName());
				sheetTabs.put("tabColor", sheet.getTabColor());
				if (sheet == _selectedSheet)
					sheetTabs.put("sel", "t");//stand for true, use for set selected tab only
					
				ary.add(sheetTabs);
			}
		}
		return ary.size() == 0 ? null : ary;
	}

	protected void renderProperties(ContentRenderer renderer) throws
			IOException {
		super.renderProperties(renderer);
		//I18N labels, must set first
		//TODO review this part 
//		Map<String, String> labels = getLabels();
//		if (labels != null) {
//			renderer.render("labels", labels);
//		}

		String moveDir = "io.keikai.ui.selectionMoveDirection";//up, down, left, right
		Object dir = getAttribute(moveDir, true);
		if (dir == null) {
			dir = Library.getProperty(moveDir);
		}
		if (dir != null) {
			dir = dir.toString().toLowerCase();
			for (SelectionMoveDirection s : SelectionMoveDirection.values()) {
				if (s.name().equals(dir)) {
					renderer.render("_selectionMoveDirection", dir);
					break;
				}
			}
		}

		String focusPosition = Library.getProperty("io.keikai.focus.cell.position"); // top, bottom(default)
		if (focusPosition != null && "top".equals(focusPosition.toLowerCase())) {
			renderer.render("_focusOnTop", true);
		}

		// KEIKAI-380
//		if (_selectedSheet != null) {
//			String propName = "io.keikai.chart.emptyAsGaps";
//			boolean emptyAsGaps = Utils.testAttribute(this, propName, false, true);
//			for (SChart chart : _selectedSheet.getCharts()) {
//				chart.setEmptyAsGaps(emptyAsGaps);
//			}
//			getWidgetHandler().invaliate();
//		}

		//KEIKAI-10
		renderer.render(ClientCommands.TOOLBAR_BUTTONS, _buttons); // before showToolbar
		renderer.render(ClientCommands.CUSTOM_FONT_FAMILY, _customFontFamily); // before showToolbar

		renderer.render(ClientCommands.COLOR_PICKER_EX_USED, isColorPickerExUsed()); //must before rendering showToolbar for the property used in creating toolbar buttons

		// KEIKAI-391: there are some actions that is able to be triggered by keyboard, so disabled actions should be sent to client under any circumstances
		renderer.render(ClientCommands.ACTION_DISABLED,
				convertToDisabledActionJSON(getUserActionManagerCtrl()
						.getSupportedUserAction(getSelectedSheet())));

		if (_showToolbar || _showContextMenu || _showSheetbar || _showSheetTabContextMenu) { // ZSS-252, _showContextMenu and _showSheetbar need actionDsiabled information in client-side.
			//20130507,Dennis,add commnet check, no actionDisabled json will cause client error when show context menu.
//			if (_actionDisabled.size() > 0) {
			// KEIKAI-391: there are some actions that is able to be triggered by keyboard, so disabled actions should be sent to client under any circumstances
//			renderer.render(ClientCommands.ACTION_DISABLED,
//					convertToDisabledActionJSON(getUserActionManagerCtrl()
//							.getSupportedUserAction(getSelectedSheet())));
//			}
			renderer.render(ClientCommands.SHOW_TOOLBAR, _showToolbar);
		}
			
		renderer.render(ClientCommands.SHOW_FORMULABAR, _showFormulabar);
		SSheet sheet = this.getSelectedSSheet();
		if (sheet == null) {
			return;
		}
		
		if (_showContextMenu) {
			renderer.render(ClientCommands.SHOW_CONTEXT_MENU, _showContextMenu);
		}

		if (_showSheetTabContextMenu) {
			renderer.render(ClientCommands.SHOW_SHEET_TAB_CONTEXT_MENU, _showSheetTabContextMenu);
		}
		
		if (_clientCacheDisabled) //default: use client cache
			renderer.render(ClientCommands.CLIENT_CACHE_DISABLED, _clientCacheDisabled);
		
		if (_maxRenderedCellSize != DEFAULT_MAX_RENDERED_CELL_SIZE)
			renderer.render(ClientCommands.MAX_RENDERED_CELL_SIZE, _maxRenderedCellSize);

		// KEIKAI-357, KEIKAI-166
		renderer.render("_topheadHeight", _topheadHeight);
		renderer.render("_leftheadWidth", _leftheadWidth);

		//Note: sheetTabs (sheet name, sheet uuid) must before showSheetTabpanel
		List<LinkedHashMap<String, String>> sheetTabs = getSheetTabs();
		if (sheetTabs != null) {
			renderer.render("sheetTabs", sheetTabs);
		}

		if (_showSheetbar) {
			renderer.render(ClientCommands.SHOW_SHEETBAR, _showSheetbar);
		}
		if (_showAddRow) { //ZSS-1082
			renderer.render(ClientCommands.SHOW_ADD_ROW, _showAddRow);
		}
		if (_showAddColumn) { //ZSS-1082
			renderer.render(ClientCommands.SHOW_ADD_COLUMN, _showAddColumn);
		}
		
		//handle link to new browser tab window; default to link to new tab
		if (!getLinkToNewTab()) {
			renderer.render(ClientCommands.LINK_TO_NEW_TAB, false);
		}
		
		//handle AutoFilter
		Map afmap = convertAutoFilterToJSON(sheet.getAutoFilter());
		if (afmap != null) {
			renderer.render(ClientCommands.AUTO_FILTER, afmap);
		} else {
			renderer.render(ClientCommands.AUTO_FILTER, (String) null);
		}
		// KEIKAI-224: should check auto filter bottom region when first loaded
		((SpreadsheetCtrlImpl) getExtraCtrl()).calcAutoFilterBottomRegion(sheet);
		
		//ZSS-988
		//handle Table's AutoFilter
		Map<String, Map> tbafsmap = convertTableFiltersToJSON(sheet);
		renderer.render(ClientCommands.TABLE_FILTERS, tbafsmap);

		int rowHeight = getRowheight();
		if (rowHeight != DEFAULT_ROW_HEIGHT) {
			renderer.render(ClientCommands.ROW_HEIGHT, getRowheight());
		}
		renderer.render(ClientCommands.COLUMN_WIDTH, getColumnwidth());
		
		if (_hideGridlines) {
			renderer.render(ClientCommands.DISPLAY_GRIDLINES, !_hideGridlines);
		}

		final boolean isProtected = isProtectSheetOrReadOnly();
		if (isProtected) {
			final SheetProtection sheetProtection = Ranges.range(getSelectedSheet()).getSheetProtection();

			// KEIKAI-378: sheetProtection may be null in read-only mode, if not
			// in read-only mode, we need to reset sheet's protection setting if any.
			if (_readOnly) {
				renderer.render(ClientCommands.ALLOW_SELECT_LOCKED_CELLS, false);
				renderer.render(ClientCommands.ALLOW_SELECT_UNLOCKED_CELLS, false);
				renderer.render(ClientCommands.ALLOW_FORMAT_CELLS, false);
				renderer.render(ClientCommands.ALLOW_FORMAT_COLUMNS, false);
				renderer.render(ClientCommands.ALLOW_FORMAT_ROWS, false);
				renderer.render(ClientCommands.ALLOW_AUTO_FILTER, false);
				renderer.render(ClientCommands.OBJECT_EDITABLE, false);
				renderer.render(ClientCommands.UNLOCK_INFO, new JSONObject());
			} else  if (sheetProtection != null) {
				renderer.render(ClientCommands.ALLOW_SELECT_LOCKED_CELLS,
						sheetProtection.isSelectLockedCellsAllowed());
				renderer.render(ClientCommands.ALLOW_SELECT_UNLOCKED_CELLS,
						sheetProtection.isSelectUnlockedCellsAllowed());
				renderer.render(ClientCommands.ALLOW_FORMAT_CELLS,
						sheetProtection.isFormatCellsAllowed());
				renderer.render(ClientCommands.ALLOW_FORMAT_COLUMNS,
						sheetProtection.isFormatColumnsAllowed());
				renderer.render(ClientCommands.ALLOW_FORMAT_ROWS,
						sheetProtection.isFormatRowsAllowed());
				renderer.render(ClientCommands.ALLOW_AUTO_FILTER,
						sheetProtection.isAutoFilterAllowed());
				renderer.render(ClientCommands.OBJECT_EDITABLE, sheetProtection.isObjectsEditable());
				if (!sheetProtection.isSelectLockedCellsAllowed()
						&& sheetProtection.isSelectUnlockedCellsAllowed()) {
					JSONObject unlockInfo = createUnlockInfo(getSelectedSheet());
					renderer.render(ClientCommands.UNLOCK_INFO, unlockInfo);
				}
			}
			renderer.render(ClientCommands.PROTECT, true);
		}
		
		// ZSS-938 set th as 0 is for hiding 1-pixel-height part of column header
		// in case, i didn't find any inpropriate look for using 1px to row header however.
		// so i decide to only adjust column header.
		renderer.render(ClientCommands.TOP_PANEL_HEIGHT, isHidecolumnhead() ? 0 : this.getTopheadheight());
		renderer.render(ClientCommands.LEFT_PANEL_WIDTH, isHiderowhead() ? 0 : this.getLeftheadwidth());

		renderer.render(ClientCommands.CELL_PADDING, DEFAULT_CELL_PADDING);

		int maxRows = getCurrentMaxVisibleRows(); //ZSS-1084
		renderer.render(ClientCommands.MAX_ROWS, maxRows); //ZSS-1084
		int maxCols = getCurrentMaxVisibleColumns(); //ZSS-1084
		renderer.render(ClientCommands.MAX_COLUMNS, maxCols); //ZSS-1084
		int rowFreeze = getSelectedSheetRowfreeze();
		if (rowFreeze > -1) {
			renderer.render(ClientCommands.ROW_FREEZE, rowFreeze);
		}
		int colFreeze = getSelectedSheetColumnfreeze();
		if (colFreeze > -1) {
			renderer.render(ClientCommands.COLUMN_FREEZE, colFreeze);
		}
		
		renderer.render(ClientCommands.SHEET_ID, getSelectedSheetId());
		renderer.render(ClientCommands.FOCUS_RECT, getRectStr(_focusArea));
		renderer.render(ClientCommands.SELECTION_RECT, getRectStr(_selectionArea));
		renderer.render(ClientCommands.SEL_TYPE, getSelTypeStr(_selectionArea)); //ZSS-717
		if (_highlightArea != null) {
			renderer.render(ClientCommands.HIGH_LIGHT_RECT, getRectStr(_highlightArea));
		}

		// generate customized row & column information
		HeaderPositionHelper colHelper = getColumnPosition(sheet);
		HeaderPositionHelper rowHelper = getRowPosition(sheet);
		renderer.render(ClientCommands.CUSTOM_SIZE_COLUMN, getSizeHelperStr(colHelper));
		renderer.render(ClientCommands.CUSTOM_SIZE_ROW, getSizeHelperStr(rowHelper));

		// generate merge range information

		MergeMatrixHelper mmhelper = getMergeMatrixHelper(sheet);
		Iterator iter = mmhelper.getRanges().iterator();
		StringBuilder merr = new StringBuilder();
		while (iter.hasNext()) {
			MergedRect block = (MergedRect) iter.next();
			int left = block.getColumn();
			int top = block.getRow();
			int right = block.getLastColumn();
			int bottom = block.getLastRow();
			int id = block.getId();
			merr.append(left).append(",").append(top).append(",").append(right).append(",").append(bottom).append(",").append(id);
			if (iter.hasNext()) {
				merr.append(";");
			}
		}
		renderer.render(ClientCommands.MERGE_RANGE, merr.toString());

		/**
		 * Add attr for UI renderer
		 */
		final SpreadsheetCtrl spreadsheetCtrl = ((SpreadsheetCtrl) this.getExtraCtrl());
		
		int initColSize = getInitColumnSize();
		int initRowSize = getInitRowSize();
		
		JSONObject activeRange = createActiveRange(spreadsheetCtrl, sheet, initColSize, initRowSize); // ZSS-423: create multiple "active range" data for every panel
		renderer.render(ClientCommands.ACTIVE_RANGE, activeRange);
		
		renderer.render(ClientCommands.PRELOAD_ROW_SIZE, getPreloadRowSize());
		renderer.render(ClientCommands.PRELOAD_COLUMN_SIZE, getPreloadColumnSize());
		
		renderer.render(ClientCommands.INIT_ROW_SIZE, initRowSize);
		renderer.render(ClientCommands.INIT_COLUMN_SIZE, initColSize);
		
		renderer.render(ClientCommands.COLUMN_HEAD_HIDDEN, _hideColhead);
		renderer.render(ClientCommands.ROW_HEAD_HIDDEN, _hideRowhead);
		
		//ZSS-1044
		renderer.render(ClientCommands.KEEP_CELL_SELECTION, _keepCellSelection);
		
		//handle Validation, must after render("activeRange" ...)
		List<Map<String, Object>> dvs = getDataValidationHandler().loadDataValidtionJASON(getSelectedSheet());
		if (dvs != null) {
			renderer.render(ClientCommands.DATA_VALIDATIONS, dvs);
		} else {
			renderer.render(ClientCommands.DATA_VALIDATIONS, (String) null);
		}

		//ZSS-1184
		if (_ignoreAutoHeight) {
			renderer.render(ClientCommands.IGNORE_AUTO_HEIGHT, _ignoreAutoHeight);
		}

		//KEIKAI-688
		renderer.render(ClientCommands.FORMULA_INFOS, getFormulaInfoLoader().getFormulaInfos());
	}
	
	private Boolean isColorPickerExUsed() {
		
		Object value = Optional.ofNullable(getAttribute(COLOR_PICKER_EX_USED_KEY, true)).orElse(getAttribute(ZSS_COLOR_PICKER_EX_USED_KEY, true));
		if(value == null){
			value = Optional.ofNullable(Library.getProperty(COLOR_PICKER_EX_USED_KEY)).orElse(Library.getProperty(ZSS_COLOR_PICKER_EX_USED_KEY, "false"));
		}
		if(value != null){
			return Boolean.parseBoolean(value.toString());
		}
		return false;
	}
	
	/**
	 * create active range from coordination (0, 0) to specific size.
	 */
	private JSONObject createActiveRange(SpreadsheetCtrl spreadsheetCtrl, SSheet sheet, int columnCount, int rowCount) {

		// KEIKAI-522
		int left = 0;
		int top = 0;
		int preloadCol = Math.max(getPreloadColumnSize(), SpreadsheetCtrl.DEFAULT_LOAD_COLUMN_SIZE);
		int preloadRow = Math.max(getPreloadRowSize(), SpreadsheetCtrl.DEFAULT_LOAD_ROW_SIZE);

		if (columnCount - left > preloadCol) {
			left = Math.max(0, Math.min(_visibleArea.getColumn(), columnCount - preloadCol));
		}
		if (rowCount - top > preloadRow) {
			top = Math.max(0, Math.min(_visibleArea.getRow(), rowCount - preloadRow));
		}

		// ZSS-423: create multiple "active range" data for every panel
		JSONObject activeRange = spreadsheetCtrl.getRangeAttrs(sheet, SpreadsheetCtrl.Header.BOTH,
				SpreadsheetCtrl.CellAttribute.ALL, left, top, columnCount, rowCount);
		
		int rowFreeze = getFreezeInfoLoader().getRowFreeze(sheet);
		int colFreeze = getFreezeInfoLoader().getColumnFreeze(sheet);
		
		if(rowFreeze >= 0) {
			activeRange.put("topFrozen", spreadsheetCtrl.getRangeAttrs(sheet, SpreadsheetCtrl.Header.BOTH,
					SpreadsheetCtrl.CellAttribute.ALL, 0, 0, columnCount, rowFreeze));
		}
		
		if(colFreeze >= 0) {
			activeRange.put("leftFrozen", spreadsheetCtrl.getRangeAttrs(sheet, SpreadsheetCtrl.Header.BOTH,
					SpreadsheetCtrl.CellAttribute.ALL, 0, 0, colFreeze, rowCount));
		}
		
		if(rowFreeze >= 0 && colFreeze >= 0) {
			activeRange.put("cornerFrozen", spreadsheetCtrl.getRangeAttrs(sheet, SpreadsheetCtrl.Header.BOTH,
					SpreadsheetCtrl.CellAttribute.ALL, 0, 0, colFreeze, rowFreeze));
		}

		return activeRange;
	}

	/**
	 * Get Column title of given index, it returns column name depends on
	 * following condition<br/>
	 * 1.if there is a custom column title assign on this spreadsheet on index,
	 * then return custom value <br/>
	 * 2.if there is a custom column title assign on selected sheet on index,
	 * the return custom value <br/>
	 * 3.return default column String, for example index 0 become "A", index 9
	 * become "J"
	 * 
	 * @return column name
	 */
	public String getColumntitle(int index) {

		if (XUtils.isTitleIndexMode(this)) {
			return Integer.toString(index);
		}
		String cname;
		if (_columnTitles != null
				&& (cname = (String) _columnTitles.get(Integer.valueOf(index))) != null) {
			return cname;
		}
		return org.apache.poi.ss.util.CellReference.convertNumToColString(index);
	}

	/**
	 * Get Row title of given index, it returns row name depends on following
	 * condition<br/>
	 * 1.if there is a custom row title assign on this spreadsheet on index,
	 * then return custom value <br/>
	 * 2.if there is a custom row title assign on selected sheet on index, the
	 * return custom value <br/>
	 * 3.return default row index+1 String, for example index 0 become "1",
	 * index 9 become "10"
	 * 
	 * @param index row index
	 * @return row name
	 */
	public String getRowtitle(int index) {
		String rname = null;

		if (XUtils.isTitleIndexMode(this)) {
			return Integer.toString(index);
		}

		if (_rowTitles != null && (rname = (String) _rowTitles.get(Integer.valueOf(index))) != null) {
			return rname;
		}

		return ""+(index+1);
	}

	/**
	 * Return current selection rectangle only if onCellSelection event listener is registered. 
	 * The returned value is a clone copy of current selection status. 
	 * Default Value:(0,0,0,0)
	 * @return current selection  (the last one if multiple)
	 */
	public AreaRef getSelection() {
		AreaRef lastSelection = null;
		for (AreaRefWithType sel : _selectionArea) {
			lastSelection = sel;
		}
		return lastSelection.cloneSelf();
	}

	/**
	 * Return all selection rectangle only if onCellSelection event listener is registered.
	 * The returned value is a clone copy of current selection status.
	 * @return all selection
	 * @since 6.0.0
	 */
	public Set<AreaRefWithType> getAllSelection() {
		Set<AreaRefWithType> selection = new LinkedHashSet<>();
		for (AreaRefWithType sel : _selectionArea) {
			selection.add((AreaRefWithType) sel.cloneSelf());
		}
		return selection;
	}

	/**
	 * Sets the selection rectangle. In general, if you set a selection, you must
	 * also set the focus by {@link #setCellFocus(CellRef)};. And, if you want
	 * to get the focus back to spreadsheet, call {@link #focus()} after set
	 * selection.
	 * 
	 * @param sel the selection rect
	 */
	public void setSelection(AreaRef sel) {
		if (!Objects.equals(_selectionArea, sel)) {
			checkSelectionArea(sel);
			setSelectionDirectly(sel);
		}
	}

	private void checkSelectionArea(AreaRef sel) {
		if (sel.getColumn() < 0 || sel.getRow() < 0
				|| sel.getLastColumn() > _book.getMaxColumnIndex()
				|| sel.getLastRow() > _book.getMaxRowIndex()
				|| sel.getColumn() > sel.getLastColumn()
				|| sel.getRow() > sel.getLastRow()) {
			throw new UiException("illegal selection : " + sel);
		}
	}

	private void setSelectionDirectly(AreaRef sel) {
		_selectionArea.clear();
		_selectionArea.add(new AreaRefWithType(sel.getRow(), sel.getColumn(), sel.getLastRow(), sel.getLastColumn(), CellSelectionType.CELL));

		HashMap args = new HashMap();
		args.put("type", "move");
		args.put("left", sel.getColumn());
		args.put("top", sel.getRow());
		args.put("right", sel.getLastColumn());
		args.put("bottom", sel.getLastRow());

		response("selection" + this.getUuid(), new AuSelection(this, args));
	}

	/**
	 * Sets the selection rectangle. In general, if you set a selection, you must
	 * also set the focus by {@link #setCellFocus(CellRef)};. And, if you want
	 * to get the focus back to spreadsheet, call {@link #focus()} after set
	 * selection.
	 *
	 * @param selection the multiple selection rect
	 * @since 6.0.0
	 */
	public void setSelection(Set<AreaRef> selection) {
		if (!Objects.equals(_selectionArea, selection)) {
			_selectionArea.clear();
			List selList = new LinkedList();
			for (AreaRef sel: selection) {
				checkSelectionArea(sel);
				_selectionArea.add(new AreaRefWithType(sel.getRow(), sel.getColumn(), sel.getLastRow(), sel.getLastColumn(), CellSelectionType.CELL));
				HashMap args = new HashMap();
				args.put("type", "move");
				args.put("left", sel.getColumn());
				args.put("top", sel.getRow());
				args.put("right", sel.getLastColumn());
				args.put("bottom", sel.getLastRow());
				selList.add(args);
			}
			response("selection" + this.getUuid(), new AuSelection(this, selList));
		}
	}

	/**
	 * Return current highlight rectangle. the returned value is a clone copy of
	 * current highlight status. Default Value: null
	 * 
	 * @return current highlight
	 */
	public AreaRef getHighlight() {
		if (_highlightArea == null)
			return null;
		return (AreaRef) _highlightArea.cloneSelf();
	}

	/**
	 * Sets the highlight rectangle or sets a null value to hide it.
	 * 
	 * @param highlight the highlight rect
	 */
	public void setHighlight(AreaRef highlight) {
		if (!Objects.equals(_highlightArea, highlight)) {
			setHighlightDirectly(highlight);
		}
	}
	
	private void setHighlightDirectly(AreaRef highlight) {
		HashMap args = new HashMap();
		
		if (highlight == null) {
			_highlightArea = null;
			args.put("type", "hide");
		} else {
			final int left = Math.max(highlight.getColumn(), 0);
			final int right = Math.min(highlight.getLastColumn(), this.getCurrentMaxVisibleColumns()-1); //ZSS-1084
			final int top = Math.max(highlight.getRow(), 0);
			final int bottom = Math.min(highlight.getLastRow(), this.getCurrentMaxVisibleRows()-1); //ZSS-1084
			if (left > right || top > bottom) {
				_highlightArea = null;
				args.put("type", "hide");
			} else {
				_highlightArea = new AreaRef(top, left, bottom, right);
				args.put("type", "show");
				args.put("left", left);
				args.put("top", top);
				args.put("right", right);
				args.put("bottom", bottom);
			}
		}
		response("selectionHighlight", new AuHighlight(this, args));
	}
	
	/**
	 * Sets whether display the gridlines.
	 * @param show true to show the gridlines.
	 */
	/*package*/ void setDisplayGridlines(boolean show) {
		if (_hideGridlines == show) {
			_hideGridlines = !show;
			smartUpdate(ClientCommands.DISPLAY_GRIDLINES, show);
		}
	}
	
	/**
	 * Update autofilter buttons.
	 */
	/*package*/ void updateAutoFilter(SSheet sheet, STable table, Integer affectedRowCount) { //ZSS-988
		if (tryReleaseCache(sheet)) {
			return;
		}
		if (this.isInvalidated())
			return;// since it is invalidate, we don't need to do anymore
		
		final SAutoFilter filter = table != null ? table.getAutoFilter() : sheet.getAutoFilter();
		if (affectedRowCount != null) {
			if (affectedRowCount.intValue() <= 0) { // wait last affected row
				return;
			}
		}
		if (table == null) {
			smartUpdate(ClientCommands.AUTO_FILTER, (DeferredValue) () -> convertAutoFilterToJSON(filter));
		} else {
			//ZSS-988: use response because there might be operations for different tables
			response(new AuInvoke(this, "setATableFilter", convertATableFilterToJSON(table)));
		}
	}

	/**
	 * Sets whether to enable this component as read-only mode or not.
	 * If true, the priority is higher than sheet's protection setting.
	 * @param readOnly true to be read only mode.
	 * @since 5.6.0
	 */
	public void setReadOnly(boolean readOnly) {
		if (_readOnly != readOnly) {
			_readOnly = readOnly;
			updateProtectionInfo();
		}
	}

	/**
	 * Returns whether the component is in read-ony mode or not.
	 * Default: false
	 * @since 5.6.0
	 */
	public boolean isReadOnly() {
		return _readOnly;
	}

	/**
	 * Returns whether this component is protected or in read-only mode.
	 * @since 5.6.0
	 */
	/*package*/ boolean isProtectSheetOrReadOnly() {
		return _readOnly || _protectSheet;
	}

    /**
     * Sets the sheet protection
     * @param protect
     */
	/*package*/ void setProtectSheet(boolean protect) {
		if (_protectSheet != protect) {
			_protectSheet = protect;
			updateProtectionInfo();
		}
	}

	//KEIKAI-378
	private void updateProtectionInfo() {
		boolean isProtected = isProtectSheetOrReadOnly();
		if (isProtected) {
			refreshAllowedOptions();
			updateUnlockInfo();
		}
		smartUpdate(ClientCommands.PROTECT, isProtected);

		refreshToolbarDisabled();
	}

	/**
	 * Return current cell(row,column) focus position. you can get the row by
	 * {@link CellRef#getRow()}, get the column by {@link CellRef#getColumn()}
	 * . The returned value is a copy of current focus status. Default
	 * Value:(0,0)
	 * 
	 * @return current focus
	 */
	public CellRef getCellFocus() {
		return new CellRef(_focusArea.getRow(), _focusArea.getColumn());
	}

	/*package*/ AreaRef getFocusArea() {
		return _focusArea;
	}

	/*package*/ Set<AreaRefWithType> getSelectionArea() {
		return _selectionArea;
	}

	/*package*/ AreaRef getVisibleArea() {
		return _visibleArea;
	}

	/**
	 * Sets the cell focus position.(this method doesn't focus the spreadsheet.)
	 * In general, if you set a cell focus, you also set the selection by
	 * {@link #setSelection(AreaRef)}; And if you want to get the focus back to
	 * spreadsheet, call {@link #focus()} to retrieve focus.
	 * 
	 * @param focus the cell focus position
	 */
	public void setCellFocus(CellRef focus) {
		if (_focusArea.getColumn() != focus.getColumn()
				|| _focusArea.getRow() != focus.getRow()) {
			if (focus.getColumn() < 0 || focus.getRow() < 0
					|| focus.getColumn() >= this.getCurrentMaxVisibleColumns() //ZSS-1084
					|| focus.getRow() >= this.getCurrentMaxVisibleRows()) { //ZSS-1084
				throw new UiException("illegal position : " + focus.toString());
			}
			setCellFocusDirectly(focus);
		}
	}
	
	private void setCellFocusDirectly(CellRef focus) {
		_focusArea.setArea(focus.getRow(), focus.getColumn(),
				focus.getRow(), focus.getColumn());
		Map args = new HashMap();
		args.put("type", "move");
		args.put("row", focus.getRow());
		args.put("column", focus.getColumn());

		response("cellFocus" + this.getUuid(), new AuCellFocus(this, args));
	}

	//ZSS-452, release client cache when it is about to out of sync.
	private void releaseClientCache(String sheetUuid){
		if(getSelectedSheetId().equals(sheetUuid)){
			return;
		}
		response(new AuInvoke(this, "_releaseClientCache",sheetUuid));
	}

	/*package*/ boolean tryReleaseCache(SSheet sheet) {
		if (!_selectedSheet.equals(sheet)){
			releaseClientCache(sheet.getId());
			_initializedSheets.remove(sheet);
			return true;
		}
		return false;
	}
	
	/*package*/ void updateColWidth(SSheet sheet, int col) {
		final int width = sheet.getColumn(col).getWidth();
		final boolean newHidden = sheet.getColumn(col).isHidden();
		HeaderPositionHelper posHelper = getColumnPosition(sheet);
		HeaderPositionInfo info = posHelper.getInfo(col);
		if ((info == null && (width != posHelper.getDefaultSize() || newHidden)) || (info != null && (info.size != width || info.hidden != newHidden))) {
			int id = info == null ? _custColId.next() : info.id;
			posHelper.setInfoValues(col, width, id, newHidden, true, true);
			((SpreadsheetCtrlImpl) getExtraCtrl()).setColumnWidth(sheet, col, width, id, newHidden);
		}
	}

	/*package*/ void updateRowHeight(SSheet sheet, int row) {
		final SRow rowobj = sheet.getRow(row);
		final int height = rowobj.getHeight();
		final boolean newHidden = rowobj.isHidden();
		HeaderPositionHelper posHelper = getRowPosition(sheet);
		HeaderPositionInfo info = posHelper.getInfo(row);
		if ((info == null && (height != posHelper.getDefaultSize() || newHidden)) || (info != null && (info.size != height || info.hidden != newHidden))) {
			int id = info == null ? _custRowId.next() : info.id;
			posHelper.setInfoValues(row, height, id, newHidden, rowobj.isCustomHeight(), rowobj.getHeightDirectly() != null);
			((SpreadsheetCtrlImpl) getExtraCtrl()).setRowHeight(sheet, row, height, id, newHidden, rowobj.isCustomHeight());
		}
	}
	
	/*package*/ ActiveRangeHelper getActiveRangeHelper() {
		ActiveRangeHelper activeRangeHelper = (ActiveRangeHelper) getAttribute(ACTIVE_RANGE_KEY);
		if (activeRangeHelper == null) {
			setAttribute(ACTIVE_RANGE_KEY, activeRangeHelper = new ActiveRangeHelper());
			return activeRangeHelper;
		}
		return activeRangeHelper;
	}
	
	/*package*/ MergeMatrixHelper getMergeMatrixHelper(SSheet sheet) {
		Map<String, MergeMatrixHelper> helpers = (Map<String, MergeMatrixHelper>) getAttribute(MERGE_MATRIX_KEY);
		if (helpers == null) {
			helpers = new HashMap<>();
			setAttribute(MERGE_MATRIX_KEY, helpers);
		}
		
		final String sheetId = sheet.getId();
		MergeMatrixHelper mmhelper = helpers.get(sheetId);
		int fzr = getSelectedSheetRowfreeze();
		int fzc = getSelectedSheetColumnfreeze();
		if (mmhelper == null) {
			final int sz = sheet.getNumOfMergedRegion();
			final List<CellRegion> mergeRanges = new ArrayList<>(sz);
			for(int j = sz - 1; j >= 0; --j) {
				final CellRegion addr = sheet.getMergedRegion(j);
				mergeRanges.add(addr);
			}
			helpers.put(sheetId, mmhelper = new MergeMatrixHelper(mergeRanges, fzr, fzc));
		} else {
			mmhelper.update(fzr, fzc);
		}
		return mmhelper;
	}
	
	//ZSS-1168: clear cached MergeMatrixHelper
	/*package*/ void removeMergeMatrixHelper(SSheet sheet) {
		Map<String, MergeMatrixHelper> helpers = (Map<String, MergeMatrixHelper>) getAttribute(MERGE_MATRIX_KEY);
		if (helpers == null) return;
		final String sheetId = sheet.getId();
		helpers.remove(sheetId);
		releaseClientCache(sheetId);
	}

	public HeaderPositionHelper getRowPosition(SSheet sheet) {
		if (sheet == null) {
			return null;
		}

		Map<String, HeaderPositionHelper> helpers = (Map<String, HeaderPositionHelper>) getAttribute(ROW_SIZE_HELPER_KEY);
		if (helpers == null) {
			setAttribute(ROW_SIZE_HELPER_KEY, helpers = new HashMap<>());
		}
		final String sheetId = sheet.getId();
		HeaderPositionHelper helper = helpers.get(sheetId);
				
		if (helper == null) {
			int defaultSize = this.getRowheight();
			
			List<HeaderPositionInfo> infos = new ArrayList<>();

			Iterator<SRow> iter = sheet.getRowIterator(); 
			while(iter.hasNext()) {
				SRow row = iter.next();
				final boolean hidden = row.isHidden();
				final int height = row.getHeight();
				final boolean customHeight = row.isCustomHeight();
				final boolean hasHeightInfo = row.getHeightDirectly() != null;
				if (height != defaultSize || hidden || customHeight || hasHeightInfo) { //special height, hidden, custom height, or height info (row attr - ht)
					infos.add(new HeaderPositionInfo(row.getIndex(), height, _custRowId.next(), hidden, customHeight, hasHeightInfo));
				}
			}
			
			helpers.put(sheetId, helper = new HeaderPositionHelper(defaultSize, infos));
		}
		return helper;
	}

	/*package*/ void updateWidget(SSheet sheet,String objId) {
		if (tryReleaseCache(sheet)) {
			return;
		}
		// KEIKAI-55: mark chart widgets to reinit even in invalidate
//		if (this.isInvalidated())
//			return;// since it is invalidate, we don't need to do anymore
		
		//by our implement, object id equals to widget id
		getWidgetHandler().updateWidget(sheet, objId);
	}
	
	/*package*/ void updateDataValidation(SSheet sheet,String objId) {
		// since it is invalidate, we don't need to do anymore
		if (tryReleaseCache(sheet) || this.isInvalidated()) {
			return;
		}

		//currently, we just update all validation (no suitable client api for now)
		//handle Validation, must after render("activeRange" ...)
		smartUpdate(ClientCommands.DATA_VALIDATIONS,
				(DeferredValue) () -> getDataValidationHandler().loadDataValidtionJASON(getSelectedSheet()));
	}
	
	/*package*/ void updateCell(SSheet sheet, int left, int top, int right, int bottom, CellAttribute cellAttr) { //ZSS-939
		if (this.isInvalidated())
			return;// since it is invalidate, we don't need to do anymore

		String sheetId = sheet.getId();
		if (!getActiveRangeHelper().containsSheet(sheet))
			return;
		
		//ZSS-939: optimize for border case
		if (cellAttr == CellAttribute.ALL || cellAttr == CellAttribute.STYLE) {
			left = left > 0 ? left - 1 : 0;// for border, when update a range, we
			// should also update the left - 1, top - 1 part
			top = top > 0 ? top - 1 : 0;
			
			//ZSS-568: for double border, when we update a range, we should also
			// update the right + 1, bottom + 1 part
			right = right + 1;
			bottom = bottom + 1;
		}
		
		//ZSS-701, ZSS-700
		final AreaRef rect = getActiveRangeHelper().getArea(sheet); 
		
		final int loadLeft = rect.getColumn();
		final int loadTop = rect.getRow();
		final int loadRight = rect.getLastColumn();
		final int loadBottom = rect.getLastRow();
		
		// ZSS-639: use corresponding sheet, not current selected sheet
		FreezeInfoLoader fil = getFreezeInfoLoader();
		final int frRow = fil.getRowFreeze(sheet);
		final int frCol = fil.getColumnFreeze(sheet);
		
		final int frTop = top <= frRow ? top : -1;
		final int frBottom = frRow;
		final int frLeft = left <= frCol ? left : -1;
		final int frRight = frCol;
		
		if (loadLeft > left) {
			left = loadLeft;
		}
		if (loadRight < right) {
			right = loadRight;
		}
		if (loadTop > top) {
			top = loadTop;
		}
		if (loadBottom < bottom) {
			bottom = loadBottom; 
		}
		
		// ZSS-393: update every panel separately
		JSONObject activeRange = null;
		//	data panel 
		if (top >= 0 && top <= bottom && left >= 0 && left <= right) {
			activeRange = responseUpdateCell(sheet, sheetId, left, top, right, bottom, cellAttr); //ZSS-939
		}
		// top freeze panel
		if (frTop >= 0 && frTop <= frBottom && left >= 0 && left <= right) { 
			if(activeRange == null) { // freeze panel needs payloader
				activeRange = responseUpdateCell(sheet, sheetId, loadLeft, loadTop, loadLeft, loadTop, cellAttr); //ZSS-939 
			}
			activeRange.put("topFrozen", responseUpdateCell(sheet, sheetId, left, frTop, right, frBottom, cellAttr)); //ZSS-939
		}
		// left freeze panel
		if (frLeft >= 0 && frLeft <= frRight && top >= 0 && top <= bottom) {
			if(activeRange == null) { // freeze panel needs payloader
				activeRange = responseUpdateCell(sheet, sheetId, loadLeft, loadTop, loadLeft, loadTop, cellAttr);  //ZSS-939
			}
			activeRange.put("leftFrozen", responseUpdateCell(sheet, sheetId, frLeft, top, frRight, bottom, cellAttr)); //ZSS-939
		}
		// corner freeze panel
		if (frTop >= 0 && frTop <= frBottom && frLeft >= 0 && frLeft <= frRight) {
			if(activeRange == null) { // freeze panel needs payloader
				activeRange = responseUpdateCell(sheet, sheetId, loadLeft, loadTop, loadLeft, loadTop, cellAttr); //ZSS-939
			}
			activeRange.put("cornerFrozen", responseUpdateCell(sheet, sheetId, frLeft, frTop, frRight, frBottom, cellAttr)); //ZSS-939
		}
		
		if(activeRange != null) {
			response(top + "_" + left + "_" + nextCellId(), new AuDataUpdate(this, "", sheetId, activeRange));
		}
	}

	/*package*/ int nextCellId() {
		return _updateCellId.next();
	}
	
	private JSONObject responseUpdateCell(SSheet sheet, String sheetId, int left, int top, int right, int bottom, CellAttribute cellAttr) { //ZSS-939
		SpreadsheetCtrl spreadsheetCtrl = ((SpreadsheetCtrl) this.getExtraCtrl());
		JSONObject result = spreadsheetCtrl.getRangeAttrs(sheet, SpreadsheetCtrl.Header.NONE, cellAttr, left, top, right, bottom); //ZSS-939
		result.put("type", "udcell");
		return result;
	}

	/*package*/
	public HeaderPositionHelper getColumnPosition(SSheet sheet) {
		Map<String, HeaderPositionHelper> helpers = (Map<String, HeaderPositionHelper>) getAttribute(COLUMN_SIZE_HELPER_KEY);
		if (helpers == null) {
			setAttribute(COLUMN_SIZE_HELPER_KEY, helpers = new HashMap<>());
		}
		final String sheetId = sheet.getId();
		HeaderPositionHelper helper = helpers.get(sheetId);
		
		if (helper == null) {
			final int defaultColSize = sheet.getDefaultColumnWidth();
			List<HeaderPositionInfo> infos = new ArrayList<HeaderPositionInfo>();
			Iterator<SColumnArray> iter = sheet.getColumnArrayIterator(); 
			while(iter.hasNext()) {
				SColumnArray columnArray = iter.next();
				final boolean hidden = columnArray.isHidden(); //whether this column is hidden
				final int columnWidth = columnArray.getWidth();//column width
				if (columnArray.isBestFit() && !hidden) {
					for (int i = columnArray.getIndex(); i <= columnArray.getLastIndex(); i++) {
						int maxWidth = calculateAutofitColumnWidth(sheet, i, false);
						infos.add(new HeaderPositionInfo(i, maxWidth == 0 ? columnWidth : maxWidth, _custColId.next(), false, true, true));
					}
				} else if (columnArray.isCustomWidth() || hidden) {
					for(int i = columnArray.getIndex(); i <= columnArray.getLastIndex();i++){
						infos.add(new HeaderPositionInfo(i, columnWidth, _custColId.next(), hidden, true, true));
					}
				}
			}

			helpers.put(sheetId, helper = new HeaderPositionHelper(defaultColSize, infos));
		}
		return helper;
	}
	
	@Override
	public Object getExtraCtrl() {
		return newExtraCtrl();
	}

	protected SpreadsheetCtrl newExtraCtrl() {
		return SpreadsheetCtrlImpl.newInstance(this);
	}

	public void invalidate() {
		super.invalidate();
		doInvalidate();
	}

	/**
	 * Retrieve client side spreadsheet focus.The cell focus and selection will
	 * keep at last status. It is useful if you want get focus back to
	 * spreadsheet after do some outside processing, for example after user
	 * click a outside button or menu item.
	 */
	public void focus() {
		// retrieve focus should work when spreadsheet init or after invalidate.
		// so I use response to implement it.
		JSONObject result = new JSONObject();
		result.put("type", "retrieve");
		
		/**
		 * rename zssfocus -> doRetrieveFocusCmd
		 */
		response("retrieveFocus" + this.getUuid(), new AuRetrieveFocus(this, result));
	}

	/**
	 * Retrieve client side spreadhsheet focus, move cell focus to position
	 * (row,column) and also scroll the cell to into visible view.
	 * 
	 * @param row row of cell to move
	 * @param column column of cell to move
	 */
	public void focusTo(int row, int column) {
		Map args = new HashMap();
		args.put("row", row);
		args.put("column", column);
		args.put("type", "moveto");
		
		response("cellFocusTo" + this.getUuid(), new AuCellFocusTo(this, args));

		_focusArea.setColumn(column);
		_focusArea.setLastColumn(column);
		_focusArea.setRow(row);
		_focusArea.setLastRow(row);
		_selectionArea.clear();
		_selectionArea.add(new AreaRefWithType(row, column, row, column, CellSelectionType.CELL));
	}

	private void cleanSelectedSheet() {
		if(_selectedSheet==null){
			return;
		}
		deleteSelfEditorFocus();
		
		List list = loadWidgetLoaders();
		int size = list.size();
		for (int i = 0; i < size; i++) {
			((WidgetLoader) list.get(i)).onSheetClean(_selectedSheet);
		}
		
//		_loadedRect.set(-1, -1, -1, -1);
		_selectionArea.clear();
		_selectionArea.add(new AreaRefWithType(0, 0, 0, 0, CellSelectionType.CELL));
		_focusArea.setArea(0, 0, 0, 0);
		
		_selectedSheet = null;
	}

	private void afterSheetSelected() {
		//Dennis, shouldn't post event in component by a server side operation call
//		org.zkoss.zk.ui.event.Events.postEvent(new Event(Events.ON_SHEET_SELECT, this));
		
		//load widgets
		List list = loadWidgetLoaders();
		int size = list.size();
		for (int i = 0; i < size; i++) {
			((WidgetLoader) list.get(i)).onSheetSelected(_selectedSheet);
		}
		//setup gridline
		setDisplayGridlines(_selectedSheet.getViewInfo().isDisplayGridlines());
		setProtectSheet(_selectedSheet.isProtected());
		
		//register collaborated focus
		CellRef cf = getCellFocus();
		moveSelfEditorFocus(getSelectedSheetId(),cf.getRow(),cf.getColumn());
		
		refreshToolbarDisabled();
		refreshAllowedOptions();
		updateUnlockInfo();
		
		//ZSS-1084
		smartUpdate(ClientCommands.MAX_ROWS, getCurrentMaxVisibleRows());
		smartUpdate(ClientCommands.MAX_COLUMNS, getCurrentMaxVisibleColumns());

		// KEIKAI-537
		smartUpdate(ClientCommands.FOCUS_RECT, getRectStr(_focusArea));
		smartUpdate(ClientCommands.SELECTION_RECT, getRectStr(_selectionArea));
	}
	
	public String getSelectedSheetName() {
		SSheet sheet = getSelectedSSheet();
		return sheet==null?null:sheet.getSheetName();
	}

	private void clearHeaderSizeHelper(boolean row, boolean col) {
		if (row)
			removeAttribute(ROW_SIZE_HELPER_KEY);
		if (col)
			removeAttribute(COLUMN_SIZE_HELPER_KEY);
	}

	private static String getSizeHelperStr(HeaderPositionHelper helper) {
		List<HeaderPositionInfo> infos = helper.getInfos();
		StringBuilder csc = new StringBuilder();
		for(HeaderPositionInfo info : infos) {
			if (csc.length() > 0)
				csc.append(",");
			csc.append(info.index).append(",")
				.append(info.size).append(",")
				.append(info.id).append(",")
				.append(info.hidden).append(",")
				.append(info.isCustom() || info.hasSizeInfo());
		}
		return csc.toString();
	}

	static private String getRectStr(AreaRef rect) {
		StringBuilder sb = new StringBuilder();
		sb.append(rect.getColumn()).append(",").append(rect.getRow()).append(",")
				.append(rect.getLastColumn()).append(",").append(rect.getLastRow());
		return sb.toString();
	}

	static private String getRectStr(Set<AreaRefWithType> rect) {
		StringBuilder sb = new StringBuilder();
		sb.append("[");
		int i = 0;
		for (AreaRef areaRef: rect) {
			if (i != 0) {
				sb.append(",");
			}
			i++;
			sb.append("[").append(areaRef.getColumn()).append(",").append(areaRef.getRow()).append(",")
					.append(areaRef.getLastColumn()).append(",").append(areaRef.getLastRow()).append("]");
		}
		sb.append("]");
		return sb.toString();
	}

	//ZSS-717
	static private String getSelTypeStr(Set<AreaRefWithType> rect) {
		StringBuilder sb = new StringBuilder();
		sb.append("[");
		int i = 0;
		for (AreaRefWithType areaRef: rect) {
			if (i != 0) {
				sb.append(",");
			}
			i++;
			String typeString = "cell";
			CellSelectionType type = areaRef.getSelType();
			switch (type) {
				case ALL:
					typeString = "all";
					break;
				case COLUMN:
					typeString = "col";
					break;
				case ROW:
					typeString = "row";
					break;
			}
			sb.append("'").append(typeString).append("'");
		}
		sb.append("]");
		return sb.toString();
	}

	private void doInvalidate() {
		//TODO: reset here ?
//		_loadedRect.set(-1, -1, -1, -1);
		
		SSheet sheet = getSelectedSSheet();

		clearHeaderSizeHelper(true, true);
		// remove this, beacuse invalidate will cause client side rebuild,
		// i must reinitial size helper since there are maybe some customized is
		// from client.
		// System.out.println(">>>>>>>>>>>remove this");
		// removeAttribute(MERGE_MATRIX_KEY);//TODO remove this, for insert
		// column test only

		_custColId = new SequenceId(-1, 2);
		_custRowId = new SequenceId(-1, 2);

		this.getWidgetHandler().invaliate();

		loadWidgetLoaders().forEach(WidgetLoader::invalidate);

		// KEIKAI-680
		if (sheet != null && !isVisible()) {
			getActiveRangeHelper().setActiveRange(sheet, 0, 0, Math.max(50, getMaxVisibleRows() - 1),
					Math.max(30, getMaxVisibleColumns() - 1));
		}
	}

	/**
	 * load widget loader
	 */
	/*package*/ List<WidgetLoader> loadWidgetLoaders() {
		if (_widgetLoaders != null)
			return _widgetLoaders;
		_widgetLoaders = new ArrayList<>();
		final String loaderclzs = finds(() -> Library.getProperty(WIDGET_LOADERS), () -> Library.getProperty(ZSS_WIDGET_LOADERS));
		if (loaderclzs != null) {
			String[] clzs = loaderclzs.split(",");
			WidgetLoader wl;
			for (int i = 0; i < clzs.length; i++) {
				clzs[i] = clzs[i].trim();
				if ("".equals(clzs[i]))
					continue;
				wl = (WidgetLoader) Spreadsheet.newInstance(clzs[i]);  //ZSS-1319
				wl.init(this);
				_widgetLoaders.add(wl);
			}
		}
		return _widgetLoaders;
	}

	/**
	 * Sets the {@link AuxAction} disabled
	 *
	 * @param action
	 * @param disabled
	 */
	public void disableUserAction(AuxAction action, boolean disabled) {
		boolean changed = false;
		if (disabled && !_actionDisabled.contains(action)) {
			_actionDisabled.add(action);
			changed = true;
		} else if (!disabled && _actionDisabled.contains(action)) {
			_actionDisabled.remove(action);
			changed = true;
		}
		if (changed) {
			refreshToolbarDisabled();
		}
	}
	
	private List<String> convertToDisabledActionJSON(Set<String> supported) {
		ArrayList<String> disd = new ArrayList<>();
		String act;
		for(AuxAction ua:AuxAction.values()){
			act = ua.toString();
			if(_actionDisabled.contains(ua) 
					|| supported==null || !supported.contains(act)){
				disd.add(act);
			}
		}
		return disd;
	}

	/**
	 * Adds toolbar button to the end of the buttons.
	 * @param btn
	 * @return true if success.
	 * @since 5.0.0
	 */
	public boolean addToolbarButton(ToolbarButton btn) {
		if (_buttons.add(btn)) {
			smartUpdate(ClientCommands.TOOLBAR_BUTTONS, _buttons);
			return true;
		}
		return false;
	}

	/**
	 * Inserts the specified button at the specified position in the buttons list.
	 * Shifts the button currently at that position (if any) and any subsequent
	 * buttons to the right (adds one to their indices).
	 *
	 * @param index index at which the specified element is to be inserted
	 * @param btn
	 * @return
	 */
	public void addToolbarButton(int index, ToolbarButton btn) {
		_buttons.add(index, btn);
		smartUpdate(ClientCommands.TOOLBAR_BUTTONS, _buttons);
	}

	/**
	 * Removes the first occurrence of the given button instance in the buttons list.
	 * @param btn
	 * @return
	 * @since 5.0.0
	 */
	public boolean removeToolbarButton(ToolbarButton btn) {
		if (_buttons.remove(btn)) {
			smartUpdate(ClientCommands.TOOLBAR_BUTTONS, _buttons);
			return true;
		}
		return false;
	}

	/**
	 * Removes the first occurrence of the given key in the buttons list.
	 * @param key the button's key
	 * @return
	 * @since 5.0.0
	 */
	public boolean removeToolbarButton(String key) {
		boolean result = false;
		for (ToolbarButton btn : _buttons) {
			if (key.equals(btn.getKey())) {
				_buttons.remove(btn);
				result = true;
				break;
			}
		}
		if (result) {
			smartUpdate(ClientCommands.TOOLBAR_BUTTONS, _buttons);
		}
		return result;
	}

	/**
	 * Removes the first occurrence of the given button instance in the buttons list.
	 * @param btn
	 * @return
	 * @since 5.1.0
	 */
	public boolean removeToolbarButton(AuxAction btn) {
		return removeToolbarButton(btn.getAction());
	}

	/**
	 * Removes the button at the specified position in this buttons list.
	 * @param index
	 * @return
	 */
	public boolean removeToolbarButton(int index) {
		boolean result = _buttons.remove(index) != null;
		if (result) {
			smartUpdate(ClientCommands.TOOLBAR_BUTTONS, _buttons);
		}
		return result;
	}

	/**
	 * Returns all toolbar buttons (read only).
	 * @return
	 * @since 5.0.0
	 */
	public List<ToolbarButton> getToolbarButtons() {
		return Collections.unmodifiableList(_buttons);
	}

	private void initToolbarButtons() {
		_buttons = Stream.of(
				AuxAction.NEW_BOOK, AuxAction.SAVE_BOOK, AuxAction.EXPORT_PDF, AuxAction.SEPARATOR,
				AuxAction.PASTE, AuxAction.CUT, AuxAction.COPY, AuxAction.SEPARATOR,
				AuxAction.FONT_FAMILY, AuxAction.FONT_SIZE, AuxAction.FONT_BOLD, AuxAction.FONT_ITALIC, AuxAction.FONT_UNDERLINE,
				AuxAction.FONT_STRIKE, AuxAction.BORDER, AuxAction.FONT_COLOR, AuxAction.BACK_COLOR, AuxAction.SEPARATOR,
				AuxAction.VERTICAL_ALIGN, AuxAction.HORIZONTAL_ALIGN, AuxAction.TEXT_INDENT_DECREASE, AuxAction.TEXT_INDENT_INCREASE,
				AuxAction.WRAP_TEXT, AuxAction.MERGE_AND_CENTER, AuxAction.SEPARATOR, AuxAction.INSERT, AuxAction.DELETE,
				AuxAction.SEPARATOR, AuxAction.CLEAR, AuxAction.SORT_AND_FILTER, AuxAction.SEPARATOR,
				AuxAction.PROTECT_SHEET, AuxAction.GRIDLINES, AuxAction.SEPARATOR,
				AuxAction.INSERT_PICTURE, AuxAction.INSERT_CHART, AuxAction.HYPERLINK
		).map(action -> ToolbarButton.Builder.create(action.getAction()).build()).collect(Collectors.toList());
	}

	//KEIKAI-208
	private void initCustomFontFamily() {
		final String str = Library.getProperty("io.keikai.ui.Spreadsheet.customFontFamily");
		if (str != null) {
			_customFontFamily = Stream.of(str.split(",")).map(String::trim).collect(Collectors.toList());
		}
	}

	/**
	 * get widget handler
	 */
	/*package*/ WidgetHandler getWidgetHandler() {
		if (_widgetHandler == null) {
			_widgetHandler = newWidgetHandler();
		}
		return _widgetHandler;
	}

	/**
	 * new widget handler
	 * 
	 * @return
	 */
	private WidgetHandler newWidgetHandler() {
		final String handlerclz = finds(() -> Library.getProperty(WIDGET_HANDLER_CLS), () -> Library.getProperty(ZSS_WIDGET_HANDLER_CLS));
		if (handlerclz != null) {
			_widgetHandler = (WidgetHandler) Spreadsheet.newInstance(handlerclz);  //ZSS-1319
		} else {
			_widgetHandler = new VoidWidgetHandler();
		}
		_widgetHandler.init(this);
		return _widgetHandler;
	}

	/**
	 * Add widget to the {@link WidgetHandler} of this spreadsheet,
	 * 
	 * @param widget a widget
	 * @return true if success to add a widget
	 */
	/*package*/ boolean addWidget(Widget widget) {
		return getWidgetHandler().addWidget(widget);
	}

	/**
	 * Remove widget from the {@link WidgetHandler} of this spreadsheet,
	 * 
	 * @param widget
	 * @return true if success to remove a widget
	 */
	/*package*/ boolean removeWidget(Widget widget) {
		return getWidgetHandler().removeWidget(widget);
	}

	private void processStartEditing(String token, StartEditingEvent event, String editingType) {
		if (!event.isCancel()) {
			Object val;
			// ZSS-536 Don't need to care about condition : clienttxt == null
			// only use server-side value when server-side modify it, otherwise, use client-side
			final boolean useEditValue = event.isEditingSet();
			if (useEditValue) {
				val = event.getEditingValue();
			} else {
				val = event.getClientValue();
			}
			processStartEditing0(token,event.getSheet(), event.getRow(), event
					.getColumn(), val, useEditValue, editingType);
		} else {
			processCancelEditing0(token, event.getSheet(), event.getRow(),
					event.getColumn(), false, editingType);
		}
	}

	private void processStopEditing(String token, StopEditingEvent event, String editingType, boolean isArrayFormula, Map selection) {
		if (!event.isCancel()) {
			processStopEditing0(token, event.getSheet(), event.getRow(), event.getColumn(), event.getEditingValue(), editingType, isArrayFormula, selection);
		} else
			processCancelEditing0(token, event.getSheet(), event.getRow(), event.getColumn(), false, editingType);
	}

	private void showFormulaErrorThenRetry(IllegalFormulaException ex, final String token, final Sheet sheet, final int rowIdx,final int colIdx, final Object value, final String editingType) {
		String title = Labels.getLabel("zss.msg.warn_title");
		String msg = Labels.getLabel("zss.msg.formula_error",new Object[]{ex.getMessage()});
		ZSMessagebox.show(msg, title, ZSMessagebox.OK, ZSMessagebox.EXCLAMATION, new EventListener() {
			public void onEvent(Event evt) {
				Spreadsheet.this.processRetryEditing0(token, sheet, rowIdx, colIdx, value, editingType);
			}
		});
	}
	
	private void showInvalidateModelOpErrorThenRetry(InvalidModelOpException ex, final String token, final Sheet sheet, final int rowIdx,final int colIdx, final Object value, final String editingType) {
		String title = Labels.getLabel("zss.msg.warn_title");
		String msg = Labels.getLabel("zss.msg.invalidate_model_op_error",new Object[]{ex.getMessage()});
		ZSMessagebox.show(msg, title, ZSMessagebox.OK, ZSMessagebox.EXCLAMATION, new EventListener() {
			public void onEvent(Event evt) {
				Spreadsheet.this.processRetryEditing0(token, sheet, rowIdx, colIdx, value, editingType);
			}
		});
	}
	

	// a local flag indicates that skip the validation and force this editing (ZSS-351) 
	private boolean forceStopEditing0 = false;

	private void processStopEditing0(final String token, final Sheet sheet,
			final int rowIdx, final int colIdx, final Object value,
			final String editingType, final boolean isArrayFormula, final Map selection) {
		try {
			
			String editText = value == null ? "" : value.toString();
			// ZSS-351: use force flag to skip the validation when user want force this editing
			if (!forceStopEditing0 && !getDataValidationHandler().validate(sheet, rowIdx, colIdx, editText,
				//callback
				new EventListener() {
					@Override
					public void onEvent(Event event) throws Exception {
						final String eventname = event.getName();
						if (ZSMessagebox.ON_CANCEL.equals(eventname)) { //cancel
							Spreadsheet.this.processCancelEditing0(token, sheet, rowIdx, colIdx, true, editingType); //skipMove
						} else if (ZSMessagebox.ON_OK.equals(eventname)) { //ok
							try {
								forceStopEditing0 = true;
								Spreadsheet.this.processStopEditing0(token, sheet, rowIdx, colIdx, value, editingType, isArrayFormula, selection);
							} finally {
								forceStopEditing0 = false;
							}
						} else { //retry
							Spreadsheet.this.processRetryEditing0(token, sheet, rowIdx, colIdx, value, editingType);
						}
					}
				}
			)) {
				return;
			}else{
				CellRegion cellRegionInArrayFormula = sheet.getInternalSheet()
						.getCellRegionInArrayFormula(sheet.getInternalSheet()
								.getCell(rowIdx, colIdx));
				if (!(cellRegionInArrayFormula == null || (isArrayFormula &&
						new CellRegion((int) selection.get("top"), (int) selection.get("left"),
								(int) selection.get("bottom"), (int) selection.get("right")).contains(cellRegionInArrayFormula)))) {
					String title = Labels.getLabel("zss.msg.warn_title");
					String msg = Labels.getLabel("zss.msg.operation_not_supported_with_array_formula");
					ZSMessagebox.show(msg, title, ZSMessagebox.OK | ZSMessagebox.CANCEL, ZSMessagebox.EXCLAMATION, new SerializableEventListener<Event>() {
						public void onEvent(Event evt) {
							Spreadsheet.this.processCancelEditing0(token, sheet, rowIdx, colIdx, false, editingType);
						}
					});
					return;
				}
				UndoableActionManager uam = getUndoableActionManager();
				if (isArrayFormula) {
					uam.doAction(new CellEditTextAction(Labels.getLabel("zss.undo.editText"),
							sheet, (int) selection.get("top"), (int) selection.get("left"), (int) selection.get("bottom"),
							(int) selection.get("right"), editText, true));
				} else {
					uam.doAction(new CellEditTextAction(Labels.getLabel("zss.undo.editText"),
							sheet, rowIdx, colIdx, rowIdx, colIdx, editText, false));
				}
			}

			//JSONObj result = new JSONObj();
			JSONObject result = new JSONObject();
			result.put("r", rowIdx);
			result.put("c", colIdx);
			result.put("type", "stopedit");
			result.put("val", "");
			result.put("et", editingType);

			smartUpdate(ClientCommands.DATA_UPDATE_STOP, new Object[] { token,	getSheetUuid(sheet), result});
		} catch (RuntimeException x) {
			if (x instanceof IllegalFormulaException) {
				showFormulaErrorThenRetry((IllegalFormulaException)x, token, sheet, rowIdx, colIdx, value, editingType);
			} else if (x instanceof InvalidModelOpException){
				showInvalidateModelOpErrorThenRetry((InvalidModelOpException)x, token, sheet, rowIdx, colIdx, value, editingType);
			} else {
				processCancelEditing0(token, sheet, rowIdx, colIdx, false, editingType);
				throw x;
			}
		}
	}
	
	private static String getSheetUuid(Sheet sheet){
		return sheet.getInternalSheet().getId();
	}

	private void processStartEditing0(String token, Sheet sheet, int row, int col, Object value, boolean useEditValue, String editingType) {
		try {
			JSONObject result = new JSONObject();
			result.put("r", row);
			result.put("c", col);
			result.put("type", "startedit");
			result.put("val", value == null ? "" : value.toString());
			result.put("et", editingType);
			if (useEditValue) { //shall use edit value from server
				result.put("server", true); 
			}
			smartUpdate(ClientCommands.DATA_UPDATE_START, new Object[] { token, getSheetUuid(sheet), result});
		} catch (RuntimeException x) {
			processCancelEditing0(token, sheet, row, col, false, editingType);
			throw x;
		}
	}

	private void processCancelEditing0(String token, Sheet sheet, int row, int col, boolean skipMove, String editingType) {
		JSONObject result = new JSONObject();
		result.put("r", row);
		result.put("c", col);
		result.put("type", "canceledit");
		result.put("val", "");
		result.put("sk", skipMove);
		result.put("et", editingType);
		smartUpdate(ClientCommands.DATA_UPDATE_CANCEL, new Object[] { token, getSheetUuid(sheet), result});
	}

	private void processRetryEditing0(String token, Sheet sheet, int row, int col, Object value, String editingType) {
		try {
			processCancelEditing0(token, sheet, row, col, true, editingType);
			JSONObject result = new JSONObject();
			result.put("r", row);
			result.put("c", col);
			result.put("type", "retryedit");
			result.put("val", value);
			result.put("et", editingType);
			smartUpdate(ClientCommands.DATA_UPDATE_RETRY, new Object[] { "", getSheetUuid(sheet), result});
		} catch (RuntimeException x) {
			processCancelEditing0(token, sheet, row, col, false, editingType);
			throw x;
		}
	}

	private void processPasteFromClipboard(ClipboardPasteEvent event, String[][] values) {
		final UndoableActionManager uam = getUndoableActionManager();
		uam.doAction(new CellEditTextAction(Labels.getLabel("zss.undo.editText"), event.getSheet(), event.getRow(),
				event.getColumn(), event.getLastRow(), event.getLastColumn(), values, true));
	}

	public boolean insertBefore(Component newChild, Component refChild) {
		// not all child can insert into spreadsheet a child want to insert to spreadsheet must provide some speaciall
		// attribute
		if (newChild.getAttribute(SpreadsheetCtrl.CHILD_PASSING_KEY) != null) {
			return super.insertBefore(newChild, refChild);
		} else {
			throw new UiException("Unsupported child for Spreadsheet: " + newChild);
		}
	}

	/**
	 * Remove editor's focus on specified name
	 */
	public void removeEditorFocus(String id){
		response("removeEditorFocus" + _focusId.next(), new AuInvoke((Component)this, "removeEditorFocus", id));
	}
	
	/*package*/ void removeFriendFocus(String id){
		//use same au response. in client side there is always editor focus
		if(_friendFocuses.remove(id)!=null){
			response("removeEditorFocus" + _focusId.next(), new AuInvoke((Component)this, "removeEditorFocus", id));
			//don't remove firendFocuses color cache, it might be back when switch sheet, only remove when sync firend focus 
		}
	}
	
	/**
	 *  Add and move other editor's focus
	 */
	public void moveEditorFocus(String id, String name, String color, int row ,int col){
		if (_selfEditorFocus != null && !_selfEditorFocus.getId().equals(id)) {
			response("moveEditorFocus" + _focusId.next(), new AuInvoke((Component)this, "moveEditorFocus", new String[]{id, name, color,""+row,""+col}));
		}
	}
	
	/*package*/ void addOrMoveFriendFocus(String id, String name, String color, String sheetId, int row ,int col){
		if (_selfEditorFocus != null && !_selfEditorFocus.getId().equals(id) && getSelectedSSheet()!=null) {
			if(sheetId!=null && sheetId.equals(getSelectedSheetId())){
				response("moveEditorFocus" + _focusId.next(), new AuInvoke((Component)this, "moveEditorFocus", new String[]{id, name, color,""+row,""+col}));
				_friendFocuses.put(id, new FriendFocus(id, name, color, sheetId, row, col));
			}else{
				removeFriendFocus(id);
			}
		}
	}	
	
	private static class FriendFocus extends Focus{
		public FriendFocus(String id, String name, String color,
				String sheetId, int row, int col) {
			super(id, name, color, sheetId, row, col,null);
		}
	}
	
	/*package*/ void syncFriendFocus() {
		syncFriendFocus(false);
	}
	//ZSS-1040
	private void syncFriendFocus(boolean always) {
		if (_book != null) {
			final Set<Object> bookFocuses;
			final Set<String> keep = new HashSet<String>(); //friend focus id in sheet
			final Set<String> inbook = new HashSet<String>(); //friend focus id in book
			
			//ZSS-998
			final List<Focus> inSheetFocus = new ArrayList<Focus>(); //friend focus in sheet
			final List<Focus> inBookFocus = new ArrayList<Focus>(); //friend focus in book
			
			bookFocuses = getFriendFocusHelper().getAllFocus();
			
			String sheetid = getSelectedSheetId();
			for(Object f:bookFocuses){
				if(!(f instanceof Focus) || f.equals(_selfEditorFocus)){
					continue;
				}
				Focus focus = (Focus)f;
				String id = focus.getId();
				inbook.add(id);
				inBookFocus.add(focus); //ZSS-998
				if(focus.getSheetId().equals(sheetid)){
					if(!_friendFocuses.containsKey(id) || always){ //ZSS-1040
						//same sheet, but not in friend focus, add back
						addOrMoveFriendFocus(id, focus.getName(), focus.getColor(), focus.getSheetId(), focus.getRow(), focus.getColumn());
					}
					keep.add(id);//same sheet, keep it in friend focus
					inSheetFocus.add(focus); //ZSS-998
				}else if(_friendFocuses.containsKey(id)){
					//different sheet and in firend, remove it
					removeFriendFocus(id);
				}
			}
			//remove other friend focus that not in book focus
			for(String fid:new HashSet<String>(_friendFocuses.keySet())){
				if(keep.contains(fid)) continue;
				removeFriendFocus(fid);				
			}
			
			//ZSS-998
			final Sheet sheet = getSelectedSheet();
			org.zkoss.zk.ui.event.Events.postEvent(
				new SyncFriendFocusEvent(Events.ON_SYNC_FRIEND_FOCUS, Spreadsheet.this, sheet, inBookFocus, inSheetFocus));
		}
	}
	
	/**
	 * Under collaboration edit mode, Spreadsheet renders multiple selection boxes in a sheet.
	 * Each selection box shows its user name when you hover on it except your owned one.
	 * @param name a name that shows with a selection box on other Spreadsheet
	 */
	public void setUserName(String name) {
		if(!Objects.equals(_userName, name)){
			_userName = name;
			if (_selfEditorFocus != null) {
				_selfEditorFocus.setName(name);
				//TODO UPDATE self and relative
				moveSelfEditorFocus(_selfEditorFocus.getSheetId(), _selfEditorFocus.getRow(), _selfEditorFocus.getColumn());
			}	
		}
	}

	//sync friend focus position after some size change
	/*package*/ void syncFriendFocusPosition(int left, int top, int right, int bottom) {
		int row = -1, col = -1;
		for(Focus focus : new ArrayList<Focus>(_friendFocuses.values())) {//avoid co-modify-exception
			row=focus.getRow();
			col=focus.getColumn();
			if(col>=left && col<=right && row>=top  && row<=bottom) {
				this.addOrMoveFriendFocus(focus.getId(), focus.getName(), focus.getColor(), focus.getSheetId(),row, col);
			}
		}
	}

	static {
		//ZSS-220 Can't get correct selection if I didn't listen to onCellSelection
		//onCellSelection should be a important event.
		addClientEvent(Spreadsheet.class, Events.ON_CELL_SELECTION,	CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_CELL_SELECTION_UPDATE, CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_CELL_FOCUS, CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		//can't ignore duplicate, for different header resize
		addClientEvent(Spreadsheet.class, Events.ON_HEADER_UPDATE, CE_IMPORTANT | CE_NON_DEFERRABLE);
		addClientEvent(Spreadsheet.class, Events.ON_SHEET_SELECT, CE_IMPORTANT | CE_DUPLICATE_IGNORE | CE_NON_DEFERRABLE);
		
		addClientEvent(Spreadsheet.class, Events.ON_CELL_CLICK, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_CLIPBOARD_PASTE, CE_IMPORTANT | CE_NON_DEFERRABLE);
		addClientEvent(Spreadsheet.class, Events.ON_CELL_RIGHT_CLICK, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_CELL_DOUBLE_CLICK, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_HEADER_CLICK, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_HEADER_RIGHT_CLICK,	CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_HEADER_DOUBLE_CLICK, CE_DUPLICATE_IGNORE);

		addClientEvent(Spreadsheet.class, Events.ON_EDITBOX_EDITING, 0);
		
		//ZSS-325 Cannot copy an area of cells from Excel to Spreadsheet
		addClientEvent(Spreadsheet.class, Events.ON_START_EDITING, CE_IMPORTANT | CE_NON_DEFERRABLE);
		addClientEvent(Spreadsheet.class, Events.ON_STOP_EDITING, CE_IMPORTANT | CE_NON_DEFERRABLE);
		
		addClientEvent(Spreadsheet.class, Events.ON_CELL_HYPERLINK, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_CELL_FILTER, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_CELL_VALIDATOR, CE_DUPLICATE_IGNORE);

		addClientEvent(Spreadsheet.class, Events.ON_CTRL_KEY, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_AUX_ACTION, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_WIDGET_CTRL_KEY, CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, Events.ON_WIDGET_UPDATE, CE_DUPLICATE_IGNORE);

		addClientEvent(Spreadsheet.class, Events.ON_GROUP_OPEN, CE_IMPORTANT | CE_DUPLICATE_IGNORE | CE_NON_DEFERRABLE);
		addClientEvent(Spreadsheet.class, Events.ON_GROUP_HEADER_CLICK, CE_IMPORTANT | CE_DUPLICATE_IGNORE | CE_NON_DEFERRABLE);

		//Event dispatcher
		addClientEvent(Spreadsheet.class, InnerEvts.ON_ZSS_CELL_MOUSE, CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, InnerEvts.ON_ZSS_HEADER_MOUSE, CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		
		//Inner
		addClientEvent(Spreadsheet.class, InnerEvts.ON_ZSS_CELL_FETCH, CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, InnerEvts.ON_ZSS_FETCH_ACTIVE_RANGE, CE_IMPORTANT | CE_DUPLICATE_IGNORE);
		addClientEvent(Spreadsheet.class, InnerEvts.ON_ZSS_SYNC_BLOCK, CE_IMPORTANT | CE_DUPLICATE_IGNORE);

		//TODO Dennis, why need this and is importnat?Review
//		addClientEvent(Spreadsheet.class, org.zkoss.zk.ui.event.Events.ON_BLUR,	CE_IMPORTANT | CE_DUPLICATE_IGNORE);//
	}



	// super//
	/**
	 * Processes an AU request. It is invoked internally.
	 * 
	 * <p>
	 * Default: in addition to what are handled by {@link XulElement#service(AuRequest, boolean)},
	 * it also handles onChange.
	 */
	public void service(AuRequest request, boolean everError) {
		final String cmd = request.getCommand();

		if (Events.ON_CELL_HYPERLINK.equals(cmd)) {
			final CellHyperlinkEvent evt = CellHyperlinkEvent.getHyperlinkEvent(request);
			if (evt != null) {
				//ZSS-807
				if (evt.getType() == HyperlinkType.DOCUMENT) {
					// might be a name range
					final String addr = evt.getAddress();
					final Object[] objs = parseLinkAddress(getBook(), addr);					
					final SName name = (SName) objs[0];
					final String sheetScopeName = (String) objs[1];
					if (name != null) {
						final CellRegion rgn = name.getRefersToCellRegion();
						final String sheetname = name.getRefersToSheetName();
						final Sheet sheet = getBook().getSheet(sheetname);
						if (sheet == null) {
							ZSMessagebox.show(Labels.getLabel("zssex.dlg.hyperlink.ref_not_valid"));
							return;
						} else if (!sheet.isHidden()) {
							final int left = rgn.getColumn();
							final int top = rgn.getRow();
							final int row = top;
							final int col = left;
							final int right = rgn.getLastColumn();
							final int bottom = rgn.getLastRow();
							final boolean cacheInClient = false; // assume no cache
							setSelectedSheetDirectly(sheetname, cacheInClient, row, col, 
									left, top, right, bottom);
						}
					} else if (!Strings.isBlank(addr)) { // ZSS-1211: should skip empty address
						try {
							final AreaReference area = new AreaReference(addr, SpreadsheetVersion.EXCEL2007);
							final CellReference cell1 = area.getFirstCell();
							final String sheetname = cell1.getSheetName();
							final Sheet sheet = getBook().getSheet(sheetname);
							if (sheet == null) {
								ZSMessagebox.show(Labels.getLabel("zssex.dlg.hyperlink.ref_not_valid"));
								return;
							} else if (!sheet.isHidden()) {
								final int left = cell1.getCol();
								final int top = cell1.getRow();
								final int row = top;
								final int col = left;
								final CellReference cell2 =  area.isSingleCell() ? null : area.getLastCell();
								final int right = cell2 == null ? left : cell2.getCol();
								final int bottom = cell2 == null ? top : cell2.getRow();
								final boolean cacheInClient = false; // assume no cache
								setSelectedSheetDirectly(sheetname, cacheInClient, row, col, 
										left, top, right, bottom);
							}
						} catch (Exception ex) { //ZSS-1211: illegal address
							ZSMessagebox.show(Labels.getLabel("zssex.dlg.hyperlink.ref_not_valid"));
							return;
						}
					}
				}
				org.zkoss.zk.ui.event.Events.postEvent(evt);
			}
			return;
		}

		// #KEIKAI-416
		if ("onAutoFitColumnWidth".equals(cmd)) {
			Map<String, Object> data = request.getData();
			int leftColumnIndex = (int)data.get("left");

			// multiple columns
			if (data.containsKey("right")) {
				int rightColumnIndex = (int)data.get("right");
				SSheet sheet = _book
						.getSheetById((String) data.get("sheetId"));
				setAutofitColumnWidth(sheet, leftColumnIndex, rightColumnIndex);
			} else {
				// single column
				setAutofitColumnWidth(_book.getSheetById((String) data.get("sheetId")),
						leftColumnIndex);
			}
			return;
		}
		Command command = InnerEvts.getCommand(cmd);
		if (command != null) {
			command.process(request);
			return;
		}
		
		super.service(request, everError);
	}

	public void smartUpdate(String attr, Object value) {
		super.smartUpdate(attr, value);
	}

	public void smartUpdate(String attr, Object value, boolean append) {
		super.smartUpdate(attr, value, append);
	}

	public void response(String key, AuResponse response) {
		super.response(key, response);
	}

	@Override
	public void afterCompose() {
		//KEIKAI-13 add zk touch related scripts for ipad ios13 support
		//should remove after zk fix ios13 ipad issue
		Execution exec = Executions.getCurrent();
		if (exec.getAttribute("$$IPAD_PATCH$$") == null) {
			exec.setAttribute("$$IPAD_PATCH$$", true);
			final StringBuilder sb = new StringBuilder();
			sb.append("if (navigator.platform == 'MacIntel' && navigator.maxTouchPoints > 0) {\n");
			sb.append("zk.mobile = true;\n");
			sb.append("zk.ios = true;\n");
			sb.append("zk.ipad = true;\n");
			final InputStream is = this.getClass().getResourceAsStream("/web/js/zss/ext/domtouch.js");
			final InputStream is0 = this.getClass().getResourceAsStream("/web/js/zss/ext/zswipe.js");
			try {
				final Reader reader = new BufferedReader(new InputStreamReader
						(is, Charset.forName(StandardCharsets.UTF_8.name())));
				int c;
				while ((c = reader.read()) != -1) {
					sb.append((char) c);
				}
				final Reader reader0 = new BufferedReader(new InputStreamReader
						(is0, Charset.forName(StandardCharsets.UTF_8.name())));
				while ((c = reader0.read()) != -1) {
					sb.append((char) c);
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
			sb.append("}");
			((PageCtrl) this.getPage()).addAfterHeadTags("<script>" + sb.toString() + "</script>");
		}
		updateBookCircularIteration();
	}
	
	@Override
	public void onPageAttached(Page newpage, Page oldpage){
		super.onPageAttached(newpage, oldpage);
		String uuid = getUuid();
		Desktop desktop = Executions.getCurrent().getDesktop();
		if(_bookCleaner==null){
			_bookCleaner = new BookCleaner(uuid);
			desktop.addListener(_bookCleaner);
		}
	}
	
	@Override
	public void onPageDetached(Page page){
		super.onPageDetached(page);
		Book book = getBook();
		if(book!=null){
			String scope = book.getShareScope();
			if(!Strings.isEmpty(scope)){
				setBook(null);
			}
		}
		Desktop desktop = Executions.getCurrent().getDesktop();
		if(_bookCleaner!=null){
			desktop.removeListener(_bookCleaner);
			_bookCleaner = null;
		}

		// #KEIKAI-415
		releaseBook();
	}

	//new wrapped API since 3.0.0
	
	/**
	 * Returns the book model of this Spreadsheet. If you call this method at
	 * first time and the book has not assigned by {@link #setBook(Book)}, this
	 * will create a new model depends on src;
	 * 
	 * @return the book model of this spread sheet.
	 */
	public Book getBook(){
		SBook book = getSBook();
		return book==null?null:new BookImpl(new SimpleRef<SBook>(book));
	}
	
	/**
	 * Sets the book data model of this spread sheet.
	 * 
	 * @param book the book data model.
	 */
	public void setBook(Book book) {
		setSBook((SBook)(book==null?null:((BookImpl)book).getNative()));
	}


	// #KEIKAI-416: 2 padding-left and padding-right + border-right 1px + 4px buffer.
	private final static int PADDING_SPACE = 2 + 2 + 1 + 4;

	/**
	 * AutoFit a single column width according to its data size.
	 * No resize when the whole column without any content.
	 * @see #setAutofitColumnWidth(SSheet, int, int) Check the note.
	 * @since 5.7.0
	 * @param sheet
	 * @param column
	 */
	public void setAutofitColumnWidth(SSheet sheet, int column) {
		setAutofitColumnWidth(sheet, column, column);
	}

	/**
	 * AutoFit multiple columns width according to each one's data size from
	 * the <t>fromColumn</t> to the <t>toColumn</t>.
	 * No resize when the whole column has no content.<br/>
	 * <b>Note:</b> be aware of that in different JDK vendor, version, or OS, the {@link FontMetrics} calculates a different width value of the same text, so the result on UI may look different when the server is running on these differences which are mentioned early.
	 * @since 5.7.0
	 * @param sheet
	 * @param fromColumn the start column to autofit width
	 * @param toColumn the end column to autofit width
	 */
	public void setAutofitColumnWidth(SSheet sheet, int fromColumn, int toColumn) {
		CellRegion dataRegion = sheet.getDataRegion();
		if (dataRegion != null) {
			for (int column = fromColumn; column <= toColumn; column++) {
				int maxWidth = calculateAutofitColumnWidth(sheet, column, true);
				if (maxWidth > 0) {
					SColumn sheetColumn = sheet.getColumn(column);
					sheetColumn.setCustomWidth(true);
					sheetColumn.setWidth(maxWidth);
					updateColWidth(sheet, column);
					sheetColumn.setBestFit(true);
				}
			}
		}
	}
	private int calculateAutofitColumnWidth(SSheet sheet, int column, boolean skipAllIfWrapped) {
		int maxWidth = 0;
		Graphics2D g2d = new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB).createGraphics();
		for (int row = 0, maxRows = sheet.getDataRegion().getLastRow(); row <= maxRows; row++) {
			Range range = Ranges.range(new SheetImpl(new SimpleRef<SBook>(sheet.getBook()),
					new SimpleRef<SSheet>(sheet)), row, column);
			// KEIKAI-513
			if (range.getCellStyle().isWrapText()) {
				if (skipAllIfWrapped) {
					maxWidth = 0;
					break;
				} else {
					continue;
				}
			}

			if (sheet.getRow(row).isHidden() || range.hasMergedCell()) {
				continue;
			}
			Font sFont = range.getCellStyle().getFont();
			Font.Boldweight boldWeight = sFont.getBoldweight();

			int jFontStyle = java.awt.Font.PLAIN;
			if (boldWeight != null && boldWeight == Font.Boldweight.BOLD) {
				jFontStyle += java.awt.Font.BOLD;
			}
			if (sFont.isItalic()) {
				jFontStyle += java.awt.Font.ITALIC;
			}
			java.awt.Font jFont = new java.awt.Font(sFont.getFontName(),
					jFontStyle, sFont.getFontHeightInPoint());
			FontMetrics fontMetrics = g2d.getFontMetrics(jFont);

			String formatText = range.getCellData().getFormatText();
			String[] splitStrings = formatText.split("\\n");
			for (String splitString : splitStrings) {
				if (!Strings.isEmpty(splitString)) {
					int max = fontMetrics.stringWidth(splitString);
					if (maxWidth < max) {
						maxWidth = max;
					}
				}
			}
		}
		if (maxWidth > 0) {
			// point to pixel
			maxWidth = UnitUtil.pointToPx((int) Math.ceil(maxWidth)) + PADDING_SPACE;
		}
		return maxWidth;
	}
	/**
	 * Adjust specified row's height of the currently-selected sheet according to the specified cell's content. Since it calculates the height at client side, it takes effect only when the specified cell is rendered in a browser, but the resulting height is more accurate than {@link io.keikai.api.CellOperationUtil#getAutoFitHeight(Sheet, int, int)}. If the target cell is a merged cell, please specify its left-top corner cell.
	 * @param row index of row
	 * @param col index of column
	 * @since 5.11.0
	 */
	public void autoFitRowHeight(int row, int col) {
		Map args = new HashMap();
		args.put("row", row);
		args.put("column", col);
		response("cellAutoFitHeight" + XUtils.nextUpdateId(), new AuCellAutoFitHeight(this, args));
	}
	/**
	 * Gets the selected sheet, the default selected sheet is first sheet.
	 * @return #{@link Sheet}
	 */
	public Sheet getSelectedSheet(){
		SSheet sheet = getSelectedSSheet();
		return sheet==null?null:new SheetImpl(new SimpleRef<SBook>(sheet.getBook()),new SimpleRef<SSheet>(sheet));
	}
	
	/**
	 * Returns the maximum visible number of columns of this spreadsheet. You can assign
	 * new numbers by calling {@link #setMaxVisibleColumns(int)}.
	 * 
	 * @return the maximum visible number of columns 
	 * @since 3.0.0
	 */
	public int getMaxVisibleColumns() {
		return _maxColumns; //ZSS-1084 
	}
	
	/**
	 * Returns the maximum visible number of rows of this spreadsheet. 
	 * You can assign new number by calling {@link #setMaxVisibleRows(int)}.
	 * 
	 * @return the maximum visible number of rows of the currently selected sheet.
	 * @since 3.0.0
	 */
	public int getMaxVisibleRows() {
		return _maxRows; //ZSS-1084
	}
	
	//ZSS-1084
	/**
	 * Returns the maximum visible number of columns of the currently selected
	 * sheet.
	 * @return
	 * @since 3.8.1
	 */
	public int getCurrentMaxVisibleColumns() {
		return getSheetMaxVisibleColumns(getSelectedSSheet());
	}

	//ZSS-1084
	/**
	 * Returns the maximum visible number of rows of the currently selected
	 * sheet. 
	 * @return
	 * @since 3.8.1
	 */
	public int getCurrentMaxVisibleRows() {
		return getSheetMaxVisibleRows(getSelectedSSheet());
	}
	
	/**
	 * Sets the maximum visible number of columns of this spreadsheet. For example, if you
	 * set this parameter to 40, it will allow showing only column 0 to column 39. the minimal value of
	 * max number of columns must large than 0. Since 3.8.1, if you set maximum visible number of columns
	 * to 0, it means you allow end user to adjust the visible number of columns. 
	 * <br/>
	 * Default : 0. (since 3.8.1)
	 * 
	 * @param maxcols  the maximum visible number of columns 
	 * @since 3.0.0
	 */
	public void setMaxVisibleColumns(int maxcols){
		if (maxcols < 1) {
			throw new UiException("maxcolumn must be greater than 0: " + maxcols);
		}

		if (_maxColumns != maxcols) {
			_maxColumns = maxcols;
			//ZSS-1082
			refreshMaxVisibleColumns();
		}
	}
	
	/**
	 * Sets the maximum visible number of rows of this spreadsheet. For example, if you set
	 * this parameter to 40, it will allow showing only row 0 to 39. The minimal value of max number of rows
	 * must large than 0. Since 3.8.1, if you set maximum visible number of rows
	 * to 0, it means you allow end user to adjust the visible number of rows. 
	 * <br/>
	 * Default : 0. (since 3.8.1)
	 * 
	 * @param maxrows  the maximum visible number of rows
	 * @since 3.0.0
	 */
	public void setMaxVisibleRows(int maxrows){
		if (maxrows < 1) {
			throw new UiException("maxrow must be greater than 0: " + maxrows);
		}

		if (_maxRows != maxrows) {
			_maxRows = maxrows;
			//ZSS-1082
			refreshMaxVisibleRows();
		}
	}

	//ZSS-1082
	private void refreshMaxVisibleColumns() {
		smartUpdate(ClientCommands.MAX_COLUMNS, getCurrentMaxVisibleColumns()); //ZSS-1084
		// 20140514, RaymondChao: unlock info records until max visible column,
		// so it needs to update when max visible column changed.
		if (getSelectedSheet() != null) {
			updateUnlockInfo();
		}
	}
	
	//ZSS-1082
	private void refreshMaxVisibleRows() {
		smartUpdate(ClientCommands.MAX_ROWS, getCurrentMaxVisibleRows()); //ZSS-1084
		// 20141104, henrichen: unlock info records until max visible row,
		// so it needs to update when max visible row changed.
		if (getSelectedSheet() != null) {
			updateUnlockInfo();
		}
	}
	
	/**
	 * Gets the importer that import the file in the specified src (
	 * {@link #getSrc}) to {@link Book} data model. The default importer is
	 * excel importer.
	 * 
	 * @return the importer
	 */
	public Importer getImporter(){
		return _importer == null ? null : new ImporterImpl(_importer);
	}
	
	/**
	 *
	 * @param importer the importer to import a spreadsheet file from a document
	 * format (e.g. an Excel file) by {@link Spreadsheet#setSrc(String)}.
	 * The default importer is {@link io.keikai.importer.XlsxImporter}
	 */
	public void setImporter(Importer importer){
		SImporter importer0 = importer == null ? null : ((ImporterImpl) importer).getNative();
		if (!Objects.equals(importer0, _importer)) {
			_importer = importer0;
			setBook(null);
		}
	}
	
	//ZSS-846
	/**
	 * Call this method to clear client side cache for a Sheet and fine 
	 * tune data loading speed. Note if the given sheet is the currently 
	 * selected sheet, it will be ignored.
	 * @param sheet the sheet
	 * @since 3.7.0
	 */
	public void clearClientCache(SSheet sheet) {
		if (sheet == null) return; 
		if (!getSBook().equals(sheet.getBook())) return;
		
		if (!sheet.equals(getSelectedSSheet())) {
			final ActiveRangeHelper helper = getActiveRangeHelper();
			if (helper.removeActiveRange(sheet) != null) {
				releaseClientCache(sheet.getId());
			}
		}
	}
	
	/*package*/ void refreshToolbarDisabled(){
		if(!isInvalidated()){
			smartUpdate(ClientCommands.ACTION_DISABLED,
					convertToDisabledActionJSON(getUserActionManagerCtrl()
							.getSupportedUserAction(getSelectedSheet())));
		}
	}
	
	private void refreshAllowedOptions() {

		// KEIKAI-378: sheetProtection may be null in read-only mode, if not
		// in read-only mode, we need to reset sheet's protection setting if any.
		if (_readOnly) {
			smartUpdate(ClientCommands.ALLOW_SELECT_LOCKED_CELLS, false);
			smartUpdate(ClientCommands.ALLOW_SELECT_UNLOCKED_CELLS, false);
			smartUpdate(ClientCommands.ALLOW_FORMAT_CELLS, false);
			smartUpdate(ClientCommands.ALLOW_FORMAT_COLUMNS, false);
			smartUpdate(ClientCommands.ALLOW_FORMAT_ROWS, false);
			smartUpdate(ClientCommands.ALLOW_AUTO_FILTER, false);
			smartUpdate(ClientCommands.OBJECT_EDITABLE, false);
		} else {
			final SheetProtection sheetProtection = Ranges.range(getSelectedSheet()).getSheetProtection();
			if (sheetProtection != null) {
				smartUpdate(ClientCommands.ALLOW_SELECT_LOCKED_CELLS,
						sheetProtection.isSelectLockedCellsAllowed());
				smartUpdate(ClientCommands.ALLOW_SELECT_UNLOCKED_CELLS,
						sheetProtection.isSelectUnlockedCellsAllowed());
				smartUpdate(ClientCommands.ALLOW_FORMAT_CELLS, sheetProtection.isFormatCellsAllowed());
				smartUpdate(ClientCommands.ALLOW_FORMAT_COLUMNS, sheetProtection.isFormatColumnsAllowed());
				smartUpdate(ClientCommands.ALLOW_FORMAT_ROWS, sheetProtection.isFormatRowsAllowed());
				smartUpdate(ClientCommands.ALLOW_AUTO_FILTER, sheetProtection.isAutoFilterAllowed());
				smartUpdate(ClientCommands.OBJECT_EDITABLE, sheetProtection.isObjectsEditable());
			}
		}
	}
	
	/*
	 *  Update unlock info including rows, cols and cells. 
	 */
	/*package*/ void updateUnlockInfo() {
		//ZSS-816: defer the unlockInfo operation or onCellContentChange might generate a storm
		if (!hasDeferOperation(ClientCommands.UNLOCK_INFO)) {
			addDeferOperation(ClientCommands.UNLOCK_INFO,
					(DeferOperation) this::updateUnlockInfo0);
		}
	}
	
	private void updateUnlockInfo0() {
		Sheet sht = getSelectedSheet();
		if (sht != null) {
			if (_readOnly) {
				smartUpdate(ClientCommands.UNLOCK_INFO, new JSONObject());
			} else {
				final SheetProtection sheetProtection = Ranges.range(sht).getSheetProtection();
				if (!sheetProtection.isSelectLockedCellsAllowed()
						&& sheetProtection.isSelectUnlockedCellsAllowed()) {
					JSONObject unlockInfo = createUnlockInfo(sht);
					smartUpdate(ClientCommands.UNLOCK_INFO, unlockInfo);
				}
			}
		}
	}
	
	/*
	 *  Unlock info:
	 *  chs: unlock columns
	 *  rhs: unlock rows
	 *  cs: unlock cells
	 */
	private JSONObject createUnlockInfo(Sheet sheet) {
		final SSheet ssheet = sheet.getInternalSheet();
		final int startColumn = ssheet.getStartColumnIndex(),
				// 20140513, RaymondChao: use getMaxVisibleColumns() instead of ssheet.getEndColumnIndex()
				// for better performance, because getEndColumnIndex() could be 16384.
				endColumn = getCurrentMaxVisibleColumns(), //ZSS-1084
				startRow = ssheet.getStartRowIndex(),
				endRow = getCurrentMaxVisibleRows(); //ZSS-1084
				
		JSONObject attrs = new JSONObject();
		JSONArray rows = new JSONArray(),
				columns = new JSONArray(),
				cells = new JSONArray();

		//ZSS-816
		SSheet sht = sheet.getInternalSheet();
		if (startColumn != -1) {
			for (final Iterator<SColumnArray> it = sht.getColumnArrayIterator(); it.hasNext();) {
				SColumnArray ca = it.next();
				if (ca.getLastIndex() < startColumn)
					continue;
				if (ca.getIndex() > endColumn) //impossible
					break;
				// overlaps
				SCellStyle style = ca.getCellStyle(true);
				// looking for unlocked columns
				if (style != null && !style.isLocked()) {
					final int start = Math.max(startColumn, ca.getIndex());
					final int end = Math.min(endColumn, ca.getLastIndex());
					columns.add(createUnlockGroup(start, end));
				}
			}
		}
		
		if (startRow != -1) {
			for (final Iterator<SRow> it = sht.getRowIterator(); it.hasNext();) {
				SRow srow = it.next();
				SCellStyle style = srow.getCellStyle(true);
				if (srow.getIndex() > endRow)
					break;
				final int rowIdx = srow.getIndex();
				// looking for unlocked row
				if (style != null && !style.isLocked()) {
					rows.add(rowIdx);
				}

				final Iterator<SCell> itc = srow.getCellIterator();
				if (itc.hasNext()) {
					int lockStart = -1;
					int lockPrev = -1;
					JSONArray lockData = new JSONArray();
					int start = -1;
					int prev = -1;
					JSONObject row = new JSONObject();
					JSONArray data = new JSONArray();
					row.put("i", rowIdx);
					
					while (itc.hasNext()) {
						SCell cell = itc.next();
						final int cellIdx = cell.getColumnIndex();
						SCellStyle stylec = cell.getCellStyle(true);
						if (stylec == null || stylec.isLocked()) { // KEIKAI-76
							//process lock group
							if (lockStart == -1) {
								lockStart = cellIdx;
							} else if (lockPrev != cellIdx - 1) {
								lockData.add(createLockGroup(lockStart, lockPrev));
								lockStart = cellIdx;
							}
							lockPrev = cellIdx;
							
							//process unlock group
							if (start != -1) {
								data.add(createUnlockGroup(start, prev));
								start = -1;
								prev = -1;
							} 
						} else {
							//process unlock group 
							if (start == -1) {
								start = cellIdx;
							} else if (prev != cellIdx - 1) {
								data.add(createUnlockGroup(start, prev));
								start = cellIdx;
							}
							prev = cellIdx;
							
							//process lock group
							if (lockStart != -1) {
								lockData.add(createLockGroup(lockStart, lockPrev));
								lockStart = -1;
								lockPrev = -1;
							}
						}
					}
					if (start != -1) {
						data.add(createUnlockGroup(start, prev));
					}
					if (lockStart != -1) {
						lockData.add(createLockGroup(lockStart, lockPrev));
					}
					row.put("data", data);
					row.put("lockData", lockData);
					cells.add(row);
				}
			}
		}
		attrs.put("chs", columns);
		attrs.put("rhs", rows);
		attrs.put("cs", cells);
		return attrs;
	}

	//ZSS-816
	/** Returns whether the named deferred operation is already exists.
	 * 
	 * @param name
	 */
	private boolean hasDeferOperation(String name) {
		Execution exec = Executions.getCurrent();
		Map<String, DeferOperation> map = 
			(Map<String, DeferOperation>) exec.getAttribute(_ZSS_DEFER_OP_MAP, false);
		return map != null && map.containsKey(name);
	}
	
	//ZSS-816
	/* Add a named deferred operation to be operated when execution is going
	 * to be cleanup. Operation with same name will override previous operation.
	 * 
	 * @param name operation name
	 * @param op the DeferOperation
	 */
	private void addDeferOperation(String name, DeferOperation op) {
		Execution exec = Executions.getCurrent();
		Map<String, DeferOperation> map = 
			(Map<String, DeferOperation>) exec.getAttribute(_ZSS_DEFER_OP_MAP, false);
		if (map == null) {
			map = new LinkedHashMap<String, DeferOperation>();
			exec.setAttribute(_ZSS_DEFER_OP_MAP, map, false);
			//post an low priority event to defer its operation!
			org.zkoss.zk.ui.event.Events.postEvent(_DEFER_OPERATOR_PRIORITY, _ON_PROCESS_DEFER_OPERATIONS, this, map);
		}
		map.put(name,  op);
	}
	
	/*
	 *  Unlocked group info:
	 *  start: start index of unlocked group
	 *  end: end index of unlocked group
	 */
	private JSONObject createUnlockGroup(int start, int end) {
		JSONObject group = new JSONObject();
		group.put("start", start);
		group.put("end", end);
		return group;
	}
	private JSONObject createLockGroup(int start, int end) {
		return createUnlockGroup(start, end);
	}

	/*package*/ CellDisplayLoader getCellDisplayLoader() {
		if(_cellDisplayLoader==null){
			String cls = finds(() -> Library.getProperty(CELL_DISPLAY_LOADER_CLS), () -> Library.getProperty(ZSS_CELL_DISPLAY_LOADER_CLS));
			if (cls != null) {
				_cellDisplayLoader = (CellDisplayLoader) Spreadsheet.newInstance(cls);  //ZSS-1319
			} else {
				_cellDisplayLoader = new SimpleCellDisplayLoader();
			}
		}
		return _cellDisplayLoader;
	}
	
	private DataValidationHandler getDataValidationHandler() {
		if(_dataValidationHandler==null){
			String cls = finds(() -> Library.getProperty(DATA_VALIDATION_HANDLER_CLS), () -> Library.getProperty(ZSS_DATA_VALIDATION_HANDLER_CLS));
			if (cls != null) {
				_dataValidationHandler = (DataValidationHandler) Spreadsheet.newInstance(cls);  //ZSS-1319
			} else {
				_dataValidationHandler = new DummyDataValidationHandler();
			}
		}
		return _dataValidationHandler;
	}
	
	/*package*/ FreezeInfoLoader getFreezeInfoLoader() {
		if(_freezeInfoLoader == null){
			String cls = finds(() -> Library.getProperty(FREEZE_INFO_LOCADER_CLS), () -> Library.getProperty(ZSS_FREEZE_INFO_LOCADER_CLS));
			if (cls != null) {
				_freezeInfoLoader = (FreezeInfoLoader) Spreadsheet.newInstance(cls);  //ZSS-1319
			} else {
				_freezeInfoLoader = new DummyFreezeInfoLoader();
			}
		}
		return _freezeInfoLoader;
	}
	
	private FormulaInfosLoader getFormulaInfoLoader() {
		if (_formulaInfosLoader == null) {
			String cls = Library.getProperty(FORMULA_INFOS_LOADER_CLS);
			if (cls != null) _formulaInfosLoader = (FormulaInfosLoader) Spreadsheet.newInstance(cls);
		}
		return _formulaInfosLoader;
	}
	
	public UndoableActionManager getUndoableActionManager(){
		if(_undoableActionManager==null){
			String cls = finds(() -> getAttribute(UNDOABLE_ACTION_MANAGER_CLS,true),
					() -> getAttribute(ZSS_UNDOABLE_ACTION_MANAGER_CLS,true),
					() -> Library.getProperty(UNDOABLE_ACTION_MANAGER_CLS),
					() -> Library.getProperty(ZSS_UNDOABLE_ACTION_MANAGER_CLS));
			if (cls != null) {
				_undoableActionManager = (UndoableActionManager) Spreadsheet.newInstance(cls);  //ZSS-1319
			} else {
				_undoableActionManager = new DummyUndoableActionManager();
			}
			_undoableActionManager.bind(this);
		}
		return _undoableActionManager;
	}
	
	private void clearUndoableActionManager(){
		if(_undoableActionManager!=null){
			_undoableActionManager.clear();
		}
	}

	//ZSS-816
	private interface DeferOperation extends Serializable {
		void process();
	}
	
	//ZSS-816
	private static final String _ZSS_DEFER_OP_MAP = "_KEIKAI_DEFER_OP_MAP";
	//ZSS-816
	private static final int _DEFER_OPERATOR_PRIORITY = -20000;
	//ZSS-816
	private static final String _ON_PROCESS_DEFER_OPERATIONS = "onProcessDeferOperations";


	/**
	 * Returns true if keep the cell selection box when lost focus; default to
	 * false.
	 * @return true if keep the cell selection box when lost focus; default to false.
	 * @since 3.8.1
	 */
	//ZSS-1044
	public boolean isKeepCellSelection() {
		return _keepCellSelection;
	}

	/**
	 * Sets true to keep the cell selection box when lost focus.
	 * @param keep true to keep the cell selection box when lost focus.
	 * @since 3.8.1
	 */
	//ZSS-1044
	public void setKeepCellSelection(boolean keep) {
		if (_keepCellSelection != keep) {
			_keepCellSelection = keep;
			smartUpdate(ClientCommands.KEEP_CELL_SELECTION, _keepCellSelection);
		}
	}
		
	//ZSS-1084
	/**
	 * Returns the max visible rows of the specified sheet
	 * @param sheet
	 * @since 3.8.1
	 */
	public int getSheetMaxVisibleRows(SSheet sheet) {
		if (sheet == null) {
			return _maxRows > 0 ? _maxRows : DEFAULT_MAX_ROWS;
		}
		String sheetId = sheet.getId();
		int[] maxRowsCols = _sheetMaxRowsCols.get(sheetId);
		if (_maxRows <= 0 && maxRowsCols == null) { //initialize
			maxRowsCols = new int[2];
			initSheetMaxRowsCols(sheet, maxRowsCols);
		}
		return _maxRows > 0 ? _maxRows : maxRowsCols[0];
	}
	
	//ZSS-1084
	/**
	 * Returns the max visible columns of the specified sheet
	 * @param sheet
	 * @since 3.8.1
	 */
	public int getSheetMaxVisibleColumns(SSheet sheet) {
		if (sheet == null) {
			return _maxColumns > 0 ? _maxColumns : DEFAULT_MAX_COLUMNS;
		}
		String sheetId = sheet.getId();
		int[] maxRowsCols = _sheetMaxRowsCols.get(sheetId);
		if (_maxColumns <= 0 && maxRowsCols == null) { //initialize
			maxRowsCols = new int[2];
			initSheetMaxRowsCols(sheet, maxRowsCols);
		}
		return _maxColumns > 0 ? _maxColumns : maxRowsCols[1];
	}
	
	//ZSS-1084
	private void initSheetMaxRowsCols(SSheet sheet, int[] maxRowsCols) {
		CellRegion region = sheet.getDataRegion(); //ZSS-1124
		if (region != null) {
			int maxRows = region.lastRow + 2;
			int maxCols = region.lastColumn + 2;
			if (maxRows > SpreadsheetVersion.EXCEL2007.getMaxRows()) {
				maxRows = SpreadsheetVersion.EXCEL2007.getMaxRows();
			} else if (maxRows < DEFAULT_MAX_ROWS) {
				maxRows = DEFAULT_MAX_ROWS;
			}
			if (maxCols > SpreadsheetVersion.EXCEL2007.getMaxColumns()) {
				maxCols = SpreadsheetVersion.EXCEL2007.getMaxColumns();
			} else if (maxCols < DEFAULT_MAX_COLUMNS) {
				maxCols = DEFAULT_MAX_COLUMNS;
			}
			maxRowsCols[0] = maxRows;
			maxRowsCols[1] = maxCols;
		} else {
			maxRowsCols[0] = DEFAULT_MAX_ROWS;
			maxRowsCols[1] = DEFAULT_MAX_COLUMNS;
		}
		_sheetMaxRowsCols.put(sheet.getId(), maxRowsCols);
	}
	
	//ZSS-1082
	/**
	 * Sets the max visible rows of the specified sheet
	 * @param sheet
	 * @since 3.8.1
	 */
	public void setSheetMaxVisibleRows(SSheet sheet, int maxRows) {
		if (sheet == null) {
			return;
		}
		if (_maxRows > 0) {
			if (maxRows <= SpreadsheetVersion.EXCEL2007.getLastRowIndex()) {
				if (_maxRows != maxRows) {
					_maxRows = maxRows;
					refreshMaxVisibleRows();
				}
			}
		} else {
			String sheetId = sheet.getId();
			int[] maxRowsCols = _sheetMaxRowsCols.get(sheetId);
			if (maxRowsCols[0] < SpreadsheetVersion.EXCEL2007.getLastRowIndex()) {
				if (maxRowsCols[0] != maxRows) {
					maxRowsCols[0] = maxRows;
					refreshMaxVisibleRows();
				}
			}
		}
	}
	
	//ZSS-1082
	/**
	 * Sets the max visibles columns of the specified sheet.
	 * @param sheet
	 * @since 3.8.1
	 */
	public void setSheetMaxVisibleColumns(SSheet sheet, int maxColumns) {
		if (sheet == null) {
			return;
		}
		if (_maxColumns > 0) {
			if (maxColumns <= SpreadsheetVersion.EXCEL2007.getLastRowIndex()) {
				if (_maxColumns != maxColumns) {
					_maxColumns = maxColumns;
					refreshMaxVisibleColumns();
				}
			}
		} else {
			String sheetId = sheet.getId();
			int[] maxRowsCols = _sheetMaxRowsCols.get(sheetId);
			if (maxRowsCols[1] < SpreadsheetVersion.EXCEL2007.getLastColumnIndex()) {
				if (maxRowsCols[1] != maxColumns) {
					maxRowsCols[1] = maxColumns;
					refreshMaxVisibleColumns();
				}
			}
		}
	}
	
	//ZSS-1082
	/**
	 * Sets whether show the add row button in sheetbar.
	 * @param showAddRow true if want to show add row button in sheetbar
	 * @since 3.8.1
	 */
	public void setShowAddRow(boolean showAddRow) {
		if (_showAddRow != showAddRow) {
			_showAddRow = showAddRow;
			smartUpdate(ClientCommands.SHOW_ADD_ROW, _showAddRow);
		}
	}
	
	//ZSS-1082
	/**
	 * Returns whether show the add row button in sheetbar.
	 * @since 3.8.1
	 */
	public boolean isShowAddRow() {
		return _showAddRow;
	}

	//ZSS-1082
	/**
	 * Sets whether show the add column button in sheetbar.
	 * @param showAddColumn true if want to show add column button in sheetbar
	 * @since 3.8.1
	 */
	public void setShowAddColumn(boolean showAddColumn) {
		if (_showAddColumn != showAddColumn) {
			_showAddColumn = showAddColumn;
			smartUpdate(ClientCommands.SHOW_ADD_COLUMN, _showAddColumn);
		}
	}
	
	//ZSS-1082
	/**
	 * Returns whether show the add column button in sheetbar.
	 * @since 3.8.1
	 */
	public boolean isShowAddColumn() {
		return _showAddColumn;
	}
	
	// parse Hyperlink linkAddress and return tokens
	// [0] SName or null if not a valid SName
	// [1] sheetScope in escaped form
	// [2] area reference
	public static Object[] parseLinkAddress(Book book, String linkAddress) {
		final String[] split = linkAddress.split("!");
		final String sheetScope = split.length == 1 ? null : split[0];
		final String ref = split.length > 1 ? split[1] : split[0];
		//ref could be a Name range
		final String rawSheetName = sheetScope == null ? null : SheetNameFormatter.unformat(sheetScope);
		final SName docSelectedName = book.getInternalBook().getNameByName(ref, rawSheetName);
		
		return new Object[] {docSelectedName, sheetScope, ref};
	}
	
	//ZSS-1181
	/**
	 * NotifyChange the currently loaded area and freeze area if exists.
	 * @since 3.8.3
	 */
	public void notifyLoadedAreaChange() {
		notifyAreaChange(newExtraCtrl().getLoadedArea());
	}
	
	//ZSS-1181
	/**
	 * NotifyChange the currently visible area and freeze area if exists.
	 * @since 3.8.3
	 */
	public void notifyVisibleAreaChange() {
		new NotifyChangeHelper().notifyReleaseSelectedCache(getSelectedSSheet());
		notifyAreaChange(newExtraCtrl().getVisibleArea());
	}
	
	//ZSS-1181
	private void notifyAreaChange(AreaRef area) {
		final int l = area.getColumn();
		final int t = area.getRow();
		final int r = area.getLastColumn();
		final int b = area.getLastRow();
		
		FreezeInfoLoader freezeInfo = getFreezeInfoLoader();
		Sheet sheet = this.getSelectedSheet();
		Ranges.range(sheet, area).notifyChange();
		final int col = freezeInfo.getColumnFreeze(sheet);
		final int row = freezeInfo.getRowFreeze(sheet);
		if (row >= 0 && col >= 0 && t != 0 && l != 0) { // corner
			Ranges.range(sheet, 0, 0, row, col).notifyChange();
		} 
		if (row >= 0 && t != 0) { // top panel
			Ranges.range(sheet, 0, l, row, r).notifyChange();
		}
		if (col >= 0 && l != 0) { // left panel
			Ranges.range(sheet, t, 0, b, col).notifyChange();
		}
	}
	
	//ZSS-1184
	/**
	 * Whether ignore row's autoHeight calculation. Default to false. 
	 * Calculating text height and doing row's autoHeight adjustment is an 
	 * expensive operation. In application, if your cell height is fixed then 
	 * you can set this flag to true to avoid Keikai Spreadsheet to calculate
	 * this information and thus speed up your application a bit.
	 * 
	 * However, if you need Keikai Spreadsheet to calculate row's autoHeight, you
	 * have to keep it at false.
	 * 
	 * Though not encouraged, note that whenever you set it from true to false, 
	 * Spreadsheet will invalidate itself to gurantee the autocalculation works.
	 * 
	 * @param b
	 * @since 3.8.3
	 */
	public void setIgnoreAutoHeight(boolean b) {
		if (_ignoreAutoHeight != b) {
			_ignoreAutoHeight = b;
			if (!b) {
				invalidate();
			} else {
				smartUpdate(ClientCommands.IGNORE_AUTO_HEIGHT, b);
			}
		}
	}

	//ZSS-1319
	/**
	 * Instantiates a new instance of the specified class name
	 * with the specified arguments.
	 *
	 * <p>It uses Class.forName to get the class(i.e. use defining classLoader of this Spreadsheet class).
	 *
	 * @param clsName the class name of the instance to create
	 * @return the new instance
	 */
	private static final Object newInstance(String clsName) {
		try {
			return Classes.newInstance(Class.forName(clsName), null, null);
		} catch (Throwable t) {
			throw new UiException(t);
		}
	}

	/**
	 * Finds with each supplier until its returned value is not null, or return null if not found.
	 * @param suppliers
	 * @since 5.1.0
	 */
	private static <T, C> T finds(Supplier<C>... suppliers) {
		return (T) Stream.of(suppliers)
				.map(Supplier::get)
				.map(Optional::ofNullable)
				.filter(Optional::isPresent).map(Optional::get).findFirst().orElse(null);
	}

	// KEIKAI-430
	/**
	 * Return current Server Clipboard
	 * @return current Server Clipboard; null if no copy source.
	 * @since 5.7.0
	 */
	public UserActionContext.Clipboard getClipboard() {
		return (UserActionContext.Clipboard)this.getAttribute(CLIPBOARD_KEY);
	}

	// KEIKAI-430
	/**
	 * Clear current Server Clipboard
	 * @since 5.7.0
	 */
	public void clearClipboard() {
		UserActionContext.Clipboard cp = (UserActionContext.Clipboard)this.removeAttribute(CLIPBOARD_KEY);
		if(cp!=null){
			this.setHighlight(null);
		}
	}

	// KEIKAI-430
	/**
	 * Set current Server Clipboard
	 * @param clipboard
	 * @since 5.7.0
	 */
	public void setClipboard(UserActionContext.Clipboard clipboard) {
		if (clipboard == null) {
			clearClipboard();
			return;
		}
		this.setAttribute(CLIPBOARD_KEY, clipboard);
		if(clipboard.getSheet().equals(this.getSelectedSheet())){
			this.setHighlight(clipboard.getSelection());
		}
	}

	// KEIKAI-509
	private void updateBookCircularIteration() {
		if (_book != null) {
			Object maxIterations = getAttribute("io.keikai.formula.maxIterations", true);
			_book.setFormulaMaxIterations(maxIterations != null ? maxIterations instanceof Integer ? ((Integer) maxIterations).intValue() : Integer.parseInt((String) maxIterations) : -1);
			Object maxChange = getAttribute("io.keikai.formula.maxChange", true);
			_book.setFormulaMaxChange(maxChange != null ? maxChange instanceof Double ? ((Double) maxChange).doubleValue() : Double.parseDouble((String) maxChange) : -1d);
		}
	}
}
