/* SpreadsheetCtrlImpl.java

	Purpose:
		
	Description:
		
	History:
		11:38 AM 2020/1/17, Created by jumperchen

Copyright (C) 2020 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import io.keikai.api.AreaRef;
import io.keikai.api.AreaRefWithType;
import io.keikai.api.CellOperationUtil;
import io.keikai.api.Ranges;
import io.keikai.api.model.Sheet;
import io.keikai.api.model.impl.SheetImpl;
import io.keikai.api.model.impl.SimpleRef;
import io.keikai.model.CellRegion;
import io.keikai.model.SAutoFilter;
import io.keikai.model.SBook;
import io.keikai.model.SCell;
import io.keikai.model.SCellStyle;
import io.keikai.model.SColumnArray;
import io.keikai.model.SComment;
import io.keikai.model.SFont;
import io.keikai.model.SRichText;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.model.STable;
import io.keikai.model.impl.AbstractCellAdv;
import io.keikai.model.impl.AbstractRowAdv;
import io.keikai.model.impl.AbstractSheetAdv;
import io.keikai.model.impl.AbstractTableAdv;
import io.keikai.model.impl.ConditionalStyleImpl;
import io.keikai.model.sys.format.FormatResult;
import io.keikai.model.util.RichTextHelper;
import io.keikai.range.impl.StyleUtil;
import io.keikai.ui.au.out.AuInsertRowColumn;
import io.keikai.ui.au.out.AuMergeCell;
import io.keikai.ui.au.out.AuRemoveRowColumn;
import io.keikai.ui.impl.CellFormatHelper;
import io.keikai.ui.impl.HeaderPositionHelper;
import io.keikai.ui.impl.MergeAggregation;
import io.keikai.ui.impl.MergeMatrixHelper;
import io.keikai.ui.impl.MergedRect;
import io.keikai.ui.impl.StringAggregation;
import io.keikai.ui.impl.XUtils;
import io.keikai.ui.impl.undo.AggregatedAction;
import io.keikai.ui.impl.undo.HideHeaderAction;
import io.keikai.ui.impl.undo.ResizeHeaderAction;
import io.keikai.ui.sys.FreezeInfoLoader;
import io.keikai.ui.sys.SpreadsheetCtrl;
import io.keikai.ui.sys.SpreadsheetInCtrl;
import io.keikai.ui.sys.SpreadsheetOutCtrl;
import io.keikai.ui.sys.UndoableAction;
import io.keikai.ui.sys.UndoableActionManager;
import io.keikai.ui.sys.UserActionManagerCtrl;
import io.keikai.ui.sys.WidgetHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.zkoss.json.JSONArray;
import org.zkoss.json.JSONObject;
import org.zkoss.lang.Objects;
import org.zkoss.lang.Strings;
import org.zkoss.util.resource.Labels;
import org.zkoss.zk.ui.UiException;

/**
 * Internal use only.
 * @since 5.1.0
 */
/*package*/ class SpreadsheetCtrlImpl implements SpreadsheetCtrl,
		SpreadsheetInCtrl, SpreadsheetOutCtrl {
	private static final Logger log = LoggerFactory.getLogger(SpreadsheetCtrlImpl.class);
	private Spreadsheet _spreadsheet;
	private SpreadsheetCtrlImpl(Spreadsheet spreadsheet) {
		_spreadsheet = spreadsheet;
	}

	/*package*/ static SpreadsheetCtrlImpl newInstance(Spreadsheet spreadsheet) {
		return new SpreadsheetCtrlImpl(spreadsheet);
	}

	public void setUserActionManagerCtrl(UserActionManagerCtrl actionHandler) {
		_spreadsheet.setUserActionManagerCtrl(actionHandler);
	}

	public UserActionManagerCtrl getUserActionManagerCtrl() {
		return _spreadsheet.getUserActionManagerCtrl();
	}

	public void setColumnSize(String sheetId, int column, int newsize, int id, boolean hidden) {
		SSheet xsheet = getSSheet(sheetId);

		// update helper size first before sheet.setColumnWidth, or it will fire a SSDataEvent
		HeaderPositionHelper helper = _spreadsheet.getColumnPosition(xsheet);
		helper.setInfoValues(column, newsize, id, hidden, true, true);

		Sheet sheet = new SheetImpl(new SimpleRef<SBook>(xsheet.getBook()),new SimpleRef<SSheet>(xsheet));
		if(sheet.isProtected() && !xsheet.getSheetProtection().isFormatColumns()){ //ZSS-1375
			return;
		}
		UndoableActionManager uam = _spreadsheet.getUndoableActionManager();

		if(hidden){
			uam.doAction(new HideHeaderAction(Labels.getLabel("zss.undo.hideColumn"),
					sheet, 0, column, 0, column, HideHeaderAction.Type.COLUMN, hidden));
		}else{
			uam.doAction(
					new AggregatedAction(Labels.getLabel("zss.undo.columnSize"),
							new UndoableAction[]{
									new HideHeaderAction(null,sheet, 0, column, 0, column, HideHeaderAction.Type.COLUMN, hidden),
									new ResizeHeaderAction(null,sheet, 0, column, 0, column, ResizeHeaderAction.Type.COLUMN, newsize, true)}
					));
		}
	}

	public void setRowSize(String sheetId, int row, int newsize, int id, boolean hidden, boolean isCustom) {
		SSheet xsheet = getSSheet(sheetId);

		HeaderPositionHelper helper = _spreadsheet.getRowPosition(xsheet);
		helper.setInfoValues(row, newsize, id, hidden, isCustom, true);

		Sheet sheet = new SheetImpl(new SimpleRef<SBook>(xsheet.getBook()),new SimpleRef<SSheet>(xsheet));
		if(sheet.isProtected()){
			return;
		}

		UndoableActionManager uam = _spreadsheet.getUndoableActionManager();

		if(hidden){
			uam.doAction(new HideHeaderAction(Labels.getLabel("zss.undo.hideRow"),
					sheet, row,0, row, 0, HideHeaderAction.Type.ROW, hidden));
		}else{
			if (isCustom){
				uam.doAction(
						new AggregatedAction(Labels.getLabel("zss.undo.rowSize"),
								new UndoableAction[]{
										new HideHeaderAction(null,sheet,  row,0, row, 0, HideHeaderAction.Type.ROW, hidden),
										new ResizeHeaderAction(null,sheet,  row,0, row, 0, ResizeHeaderAction.Type.ROW, newsize, isCustom)}
						));
			}else{ //ZSS-552 avoid auto-adjusting row height being put into undo history
				CellOperationUtil
						.setRowHeight(Ranges.range(sheet,row,0,row,0).toRowRange(),newsize, isCustom);
			}
		}
	}

	private SSheet getSSheet(String sheetId) {
		if (_spreadsheet.getSelectedSheetId().equals(sheetId)) {
			return _spreadsheet.getSelectedSSheet();
		} else {
			return _spreadsheet.getSBook().getSheetById(sheetId);
		}
	}
	public HeaderPositionHelper getColumnPositionHelper(String sheetId) {
		return _spreadsheet.getColumnPosition(getSSheet(sheetId));
	}

	public HeaderPositionHelper getRowPositionHelper(String sheetId) {
		return _spreadsheet.getRowPosition(getSSheet(sheetId));
	}

	public MergeMatrixHelper getMergeMatrixHelper(SSheet sheet) {
		return _spreadsheet.getMergeMatrixHelper(sheet);
	}

	public AreaRef getFocusArea() {
		return _spreadsheet.getFocusArea().cloneSelf();
	}

	public AreaRef getSelectionArea() {
		return _spreadsheet.getSelection();
	}

	public void setSelectionRect(int left, int top, int right, int bottom) {
		Set<AreaRefWithType> selectionArea = _spreadsheet.getSelectionArea();
		selectionArea.clear();
		selectionArea.add(new AreaRefWithType(top, left, bottom, right, CellSelectionType.CELL));
		SSheet selectedSSheet = _spreadsheet.getSelectedSSheet();
		if (selectedSSheet != null) {
			selectedSSheet.getViewInfo().setSelectionAreas(_spreadsheet.getSelectionArea().iterator().next().asString());
		}
	}

	public void setFocusRect(int left, int top, int right, int bottom) {
		_spreadsheet.getFocusArea().setArea(top, left, bottom, right);
		SSheet selectedSSheet = _spreadsheet.getSelectedSSheet();
		if (selectedSSheet != null) {
			selectedSSheet.getViewInfo().setActiveCell(_spreadsheet.getFocusArea().asString());
		}
	}

	public AreaRef getLoadedArea() {
		AreaRef rect = _spreadsheet.getActiveRangeHelper().getArea(_spreadsheet.getSelectedSSheet());
		if (rect == null)
			return null;
		return rect.cloneSelf();
	}

	public void setLoadedRect(int left, int top, int right, int bottom) {
		SSheet selectedSSheet = _spreadsheet.getSelectedSSheet();
		_spreadsheet.getActiveRangeHelper().setActiveRange(selectedSSheet, top, left, bottom, right);
		getWidgetHandler().onLoadOnDemand(selectedSSheet, left, top, right, bottom);
	}

	public void setVisibleRect(int left, int top, int right, int bottom) {
		_spreadsheet.getVisibleArea().setArea(top, left, bottom, right);
		getWidgetHandler().onLoadOnDemand(_spreadsheet.getSelectedSSheet(), left, top, right, bottom);
	}

	public AreaRef getVisibleArea() {
		return _spreadsheet.getVisibleArea().cloneSelf();
	}

	public boolean addWidget(Widget widget) {
		return _spreadsheet.addWidget(widget);
	}

	public boolean removeWidget(Widget widget) {
		return _spreadsheet.removeWidget(widget);
	}

	public WidgetHandler getWidgetHandler() {
		return _spreadsheet.getWidgetHandler();
	}

	public JSONObject getRowHeaderAttrs(SSheet sheet, int rowStart, int rowEnd) {
		return getHeaderAttrs(sheet, true, rowStart, rowEnd);
	}

	public JSONObject getColumnHeaderAttrs(SSheet sheet, int colStart, int colEnd) {
		return getHeaderAttrs(sheet, false, colStart, colEnd);
	}

	/**
	 * Header attributes
	 *
	 * <ul>
	 * 	<li>t: header type</li>
	 *  <li>s: index start</li>
	 *  <li>e: index end</li>
	 *  <li>hs: headers, a JSONArray object</li>
	 * </ul>
	 *
	 * @param isRow
	 * @param start
	 * @param end
	 * @return
	 */
	private JSONObject getHeaderAttrs(SSheet sheet, boolean isRow, int start, int end) {
		JSONObject attrs = new JSONObject();
		attrs.put("s", start);
		attrs.put("e", end);

		JSONArray headers = new JSONArray();
		attrs.put("hs", headers);
		if (isRow) {
			attrs.put("t", "r"); //type: row
			for (int row = start; row <= end; row++) {
				headers.add(getRowHeaderAttrs(sheet, row));
			}
		} else { //column header
			attrs.put("t", "c"); //type: column
			for (int col = start; col <= end; col++) {
				headers.add(getColumnHeaderAttrs(sheet, col));
			}
		}
		return attrs;
	}

	/**
	 * Column header attributes
	 *
	 * <ul>
	 * 	<li>i: column index</li>
	 *  <li>t: title</li>
	 *  <li>p: position info id</li>
	 * </ul>
	 *
	 * Ignore attribute if it's default
	 * Default attributes
	 * <ul>
	 * 	<li>hidden: false</li>
	 * </ul>
	 *
	 * @return
	 */
	private JSONObject getColumnHeaderAttrs(SSheet sheet, int col) {
		JSONObject attrs = new JSONObject();
		//			attrs.put("i", col);//getHeaderAttrs method has provide index info
		attrs.put("t", _spreadsheet.getColumntitle(col));


		SColumnArray columnArray = sheet.getColumnArray(col);
		if (columnArray != null) {
			if (columnArray.isCollapsed()) {
				attrs.put("collapsed", columnArray.isCollapsed());
			}
			if (columnArray.getOutlineLevel() > 0) {
				attrs.put("outlineLevel", columnArray.getOutlineLevel());
			}
		}

		HeaderPositionHelper colHelper = _spreadsheet.getColumnPosition(sheet);
		HeaderPositionHelper.HeaderPositionInfo info = colHelper.getInfo(col);
		if (info != null) {
			attrs.put("p", info.id);
		}
		return attrs;
	}

	/**
	 * Row header attributes
	 *
	 * <ul>
	 * 	<li>i: row index</li>
	 *  <li>t: title</li>
	 *  <li>p: position info id</li>
	 * </ul>
	 *
	 * Ignore attribute if it's default
	 * Default attributes
	 * <ul>
	 * 	<li>hidden: false</li>
	 * </ul>
	 *
	 * @return
	 */
	private JSONObject getRowHeaderAttrs(SSheet sheet, int row) {
		JSONObject attrs = new JSONObject();
		//			attrs.put("i", row);//getHeaderAttrs method has provide index info
		attrs.put("t", _spreadsheet.getRowtitle(row));


		SRow sheetRow = sheet.getRow(row);
		if (sheetRow != null) {
			if (sheetRow.isCollapsed()) {
				attrs.put("collapsed", sheetRow.isCollapsed());
			}
			if (sheetRow.getOutlineLevel() > 0) {
				attrs.put("outlineLevel", sheetRow.getOutlineLevel());
			}
		}

		HeaderPositionHelper rowHelper = _spreadsheet.getRowPosition(sheet);
		HeaderPositionHelper.HeaderPositionInfo info = rowHelper.getInfo(row);
		if (info != null) {
			attrs.put("p", info.id);
			//				if (info.hidden)
			//					attrs.put("h", 1);
		}
		return attrs;
	}

	/**
	 * Range attributes
	 *
	 * <ul>
	 * 	<li>id: sheet uuid</li>
	 * 	<li>l: range top</li>
	 *  <li>t: range top</li>
	 *  <li>r: range right</li>
	 *  <li>b: range bottom</li>
	 *  <li>at: range update Attribute Type</li>
	 *  <li>rs: rows, a JSONArray object</li>
	 *  <li>cs: cells, a JSONArray object</li>
	 * 	<li>s: strings, a JSONArray object</li>
	 *  <li>st: styles, a JSONArray object</li>
	 *  <li>m: merge attributes</li>
	 *  <li>rhs: row headers, a JSONArray object</li>
	 *  <li>chs: column headers, a JSONArray object</li>
	 * </ul>
	 *
	 * @param left
	 * @param top
	 * @param right
	 * @param bottom
	 * @param containsHeader
	 * @return
	 */
	//ZSS-939
	//@since 3.8.0
	public JSONObject getRangeAttrs(SSheet sheet, Header containsHeader, CellAttribute type, int left, int top, int right, int bottom) {
		JSONObject attrs = new JSONObject();

		attrs.put("id", sheet.getId());
		attrs.put("olr", sheet.getOutlineLevelRow());
		attrs.put("olc", sheet.getOutlineLevelCol());

		// default is true
		if (!sheet.isSummaryBelow()) {
			attrs.put("olsb", sheet.isSummaryBelow());
		}
		// default is true
		if (!sheet.isSummaryRight()) {
			attrs.put("olsr", sheet.isSummaryRight());
		}

		attrs.put("l", left);
		attrs.put("t", top);
		attrs.put("r", right);
		attrs.put("b", bottom);
		attrs.put("at", type);

		JSONArray rows = new JSONArray();
		attrs.put("rs", rows);

		StringAggregation styleAggregation = new StringAggregation();
		StringAggregation textAggregation = new StringAggregation();
		MergeAggregation mergeAggregation = new MergeAggregation(getMergeMatrixHelper(sheet));
		for (int row = top; row <= bottom; row++) {
			JSONObject r = getRowAttrs(row);
			rows.add(r);

			JSONArray cells = new JSONArray();
			r.put("cs", cells);
			for (int col = left; col <= right; col++) {
				cells.add(getCellAttr(sheet, type, row, col, styleAggregation, textAggregation, mergeAggregation));
			}
		}

		attrs.put("s", textAggregation);
		attrs.put("st", styleAggregation);
		attrs.put("m", mergeAggregation);

		//ZSS-1067: always send row and column header for client side caching
		attrs.put("rhs", getRowHeaderAttrs(sheet, top, bottom));
		attrs.put("chs", getColumnHeaderAttrs(sheet, left, right));

		List<CellRegion> allArrayFormulas = sheet.getAllArrayFormulas();
		List<CellRegion> including = new ArrayList<>();
		for (CellRegion cellRegion: allArrayFormulas) {
			if (cellRegion.getRow() >= top && cellRegion.getColumn() >= left && cellRegion.getLastRow() <= bottom
			&& cellRegion.getLastColumn() <= right) {
				including.add(cellRegion);
			}
		}
		if (!including.isEmpty()) {
			attrs.put("aflrs", including.stream().map(region -> {
				JSONObject cellAttr = new JSONObject();
				cellAttr.put("t", region.getRow());
				cellAttr.put("l", region.getColumn());
				cellAttr.put("b", region.getLastRow());
				cellAttr.put("r", region.getLastColumn());
				return cellAttr;
			}).collect(Collectors.toList()));
		}
		return attrs;
	}

	/**
	 * Row attributes
	 * <ul>
	 * 	<li>r: row number</li>
	 *  <li>h: height index</li>
	 *  <li>hd: hidden</li>
	 * </ul>
	 *
	 * Ignore if attribute is default
	 * <ul>
	 * 	<li>hidden: default is false</li>
	 * </ul>
	 */
	public JSONObject getRowAttrs(int row) {
		SSheet sheet = _spreadsheet.getSelectedSSheet();
		HeaderPositionHelper helper = _spreadsheet.getRowPosition(sheet);
		JSONObject attrs = new JSONObject();
		//row num
		attrs.put("r", row);

		HeaderPositionHelper.HeaderPositionInfo info = helper.getInfo(row);
		if (info != null) {
			attrs.put("h", info.id);
			if (info.hidden) {
				attrs.put("hd", "t"); //t stand for true
			}
		}
		return attrs;
	}

	public JSONObject getCellAttr(SSheet sheet, CellAttribute type, int row, int col, StringAggregation styleAggregation, StringAggregation textAggregation, MergeAggregation mergeAggregation) {
		boolean updateAll = type == CellAttribute.ALL,
				updateText = (updateAll || type == CellAttribute.TEXT),
				updateStyle = (updateAll || type == CellAttribute.STYLE),
				updateSize = (updateAll || type == CellAttribute.SIZE),
				updateMerge = (updateAll || type == CellAttribute.MERGE),
				updateComment = (updateAll || type == CellAttribute.COMMENT);

		SCell cell = sheet.getCell(row, col);
		JSONObject attrs = new JSONObject();

		//row num, cell num attr
		//			if (cell != null) {
		//				attrs.put("r", row);
		//				attrs.put("c", col);
		//			}

		//merge
		MergeAggregation.MergeIndex mergeIndex = mergeAggregation.add(row, col);
		if (updateMerge && mergeIndex != null) {
			attrs.put("mi", mergeIndex.getMergeId());
			attrs.put("mc", mergeIndex.getMergeCSSId());
		}

		//width, height id
		if (updateSize) {
			if (cell != null) {
				//process overflow when cell type is string, no wrap, no merge
				//ZSS-1338: apply overflow on both left and right alignment
				SCellStyle cellStyle = cell.getCellStyle();
				final boolean isString = cell.getType() == SCell.CellType.STRING || (
						cell.getType() == SCell.CellType.FORMULA && cell.getFormulaResultType() == SCell.CellType.STRING);
				if (isString && mergeIndex == null && !cellStyle.isWrapText()) {

					// ZSS-224: modify overflow flag spec. to carry more status in bitswise format
					// 1: needs overflow
					// 2: skip overflow when initializing
					// 4: [undefined now]
					// 8: ...
					int overflowOptions = 1; // needs overflow

					// ZSS-224: pre-check sibling cell's status and give current cell a hint
					// to process overflow or not when initializing.
					// ZSS-1338 check the right/left adjacent cell for left/right alignment
					SCell sibling = null;
					if (CellFormatHelper.getRealAlignment(cell) == SCellStyle.Alignment.LEFT) {
						sibling = sheet.getCell(row, col + 1);
					}else if (CellFormatHelper.getRealAlignment(cell) == SCellStyle.Alignment.RIGHT) {
						sibling = sheet.getCell(row, col - 1);
					}
					if(sibling!= null && sibling.getType() != SCell.CellType.BLANK) {
						overflowOptions |= 2; // skip overflow when initializing
					}
					attrs.put("ovf", overflowOptions);
				} else {
					((AbstractCellAdv)cell).setTextWidth(-1); // reset width for overflow
				}
			}
		}
		//ZSS-1142
		ConditionalStyleImpl cdCellStyle =
				((AbstractSheetAdv)sheet).getConditionalFormattingStyle(row, col);

		SCellStyle cellStyle = sheet.getCell(row, col).getCellStyle();
		CellFormatHelper cfh = new CellFormatHelper(sheet, row, col, getMergeMatrixHelper(sheet), cdCellStyle);
		StringBuffer doubleBorder = new StringBuffer(8);
		//ZSS-945: optimize calling CellFormatHelper#getFormatResult()
		//This implementation is super dirty! However, works.
		//@see cfh.getFontHtmlStyle(ft);
		//@see getCellDisplayLoader().getCellHtmlText(sheet, row, col, ft);
		//@see cfh.getCellFormattedText(ft);
		final FormatResult ft = updateStyle || updateText ? cfh.getFormatResult() : null;

		//ZSS-977
		STable table = ((AbstractSheetAdv)sheet).getTableByRowCol(row, col);
		SCellStyle tbCellStyle = table != null ? ((AbstractTableAdv)table).getCellStyle(row, col) : null;

		//style attr
		if (updateStyle) {
			Map<String, String> mergedBorder = new HashMap<>(4);
			String style = cfh.getHtmlStyle(doubleBorder, table, tbCellStyle, mergedBorder); //ZSS-1142

			final boolean hasMergedBorder = !mergedBorder.isEmpty();
			if (hasMergedBorder) {
				attrs.put("mgbd", mergedBorder);
			}

			if (!Strings.isEmpty(style)) {
				int idx = styleAggregation.add(style);
				attrs.put("s", idx);
			}
			String innerStyle = cfh.getInnerHtmlStyle(hasMergedBorder);
			boolean isTextWrapped = false;
			if (cell != null) {
				final SCellStyle cs = cell.getCellStyle();
				//ZSS-944: when rotate 90 degree, wrap must be false
				final int rotate90 = cs.getRotation();
				isTextWrapped = cs.isWrapText() && rotate90 != 90 && rotate90 != -90; //ZSS-1020
			}
			if (!Strings.isEmpty(innerStyle)) {
				int idx = styleAggregation.add(innerStyle);
				attrs.put("is", idx);
			}
			// ZSS-915
			String fontStyle = cfh.getRealHtmlStyle(ft, tbCellStyle, hasMergedBorder, isTextWrapped); //ZSS-945, ZSS-977, ZSS-1142
			if (!Strings.isEmpty(fontStyle)) {
				int idx = styleAggregation.add(fontStyle);
				attrs.put("os", idx);
			}
			if (cfh.hasRightBorder(table, tbCellStyle)) { //ZSS-977, ZSS-1142
				attrs.put("rb", 1);
			}

			//ZSS-509, handling lock info even cell is null ( lock in row,column style)
			boolean locked = cellStyle.isLocked();
			if (!locked)
				attrs.put("l", "f"); //f stand for "false"

			//ZSS-568, handling double border style
			final String db = doubleBorder.toString();
			if (!"____".equals(db)) {
				attrs.put("db", db);
			}

			//ZSS-901, handling auto filter border style
			final String af = cfh.getAutoFilterBorder();
			if (!"____".equals(af)) {
				attrs.put("af", "af"+af);
			}

			final String dp = cfh.getDependencyBorder();
			if (!"____".equals(dp)) {
				attrs.put("dp", "dp" + dp);
			}

			//ZSS-1142, handle dataBar border
			final boolean dbar = cfh.withDataBarBorder();
			if (dbar) {
				attrs.put("dbar", true);
			}
		}

		//ZSS-849
		//comment
		if (updateComment) {
			SComment comment = cell.getComment();
			if (comment != null) {
				SRichText rstr = comment.getRichText();
				String html;
				if (rstr == null) {
					html = RichTextHelper.escapeText(comment.getText(), true,
							true);
				} else {
					html = RichTextHelper.getCellRichTextHtml(cell, rstr,
							true); //ZSS-1138
				}
				boolean visible = comment.isVisible();
				Map map = new HashMap();
				map.put("t", html);
				map.put("v", visible);
				attrs.put("cmt", map);
			}
		}

		if (!cell.isNull()) {
			SCell.CellType cellType = cell.getType();
			if (cellType != SCell.CellType.BLANK)
				attrs.put("ct", cellType.value());

			if (cellType == SCell.CellType.FORMULA) {
				if (cell.isPartOfArrayFormulaGroup()) {
					CellRegion arrayFormulaRegion = cell.getArrayFormulaRegion();
					JSONObject region = new JSONObject();
					region.put("t", arrayFormulaRegion.getRow());
					region.put("l", arrayFormulaRegion.getColumn());
					region.put("b", arrayFormulaRegion.getLastRow());
					region.put("r", arrayFormulaRegion.getLastColumn());
					attrs.put("afl", "1");
					attrs.put("aflr", region);
				}
			}
			if (updateText) {
				if (cellType != SCell.CellType.BLANK || cell.getHyperlink() != null) {
					final String cellText = shouldShowValue(cell) ?
							_spreadsheet.getCellDisplayLoader().getCellHtmlText(sheet, row, col, ft, tbCellStyle, cdCellStyle) : //ZSS-945, ZSS-1018, ZSS-1142
							"";
					final String editText = cfh.getCellEditText();
					final String formatText = cfh.getCellFormattedText(ft); //ZSS-945

					if (Objects.equals(cellText, editText) && Objects.equals(editText, formatText)) {
						attrs.put("meft", textAggregation.add(cellText));
					} else {
						attrs.put("t", textAggregation.add(cellText));
						attrs.put("et", textAggregation.add(editText));
						attrs.put("ft", textAggregation.add(formatText));
					}
				}
			}

			if (updateStyle) {
				final boolean wrap = cellStyle.isWrapText();
				if (wrap)
					attrs.put("wp", 1);

				final int indention = cellStyle.getIndention();
				if(indention > 0)
					attrs.put("ind", indention);

				SCellStyle.Alignment horizontalAlignment = CellFormatHelper.getRealAlignment(cell);
				switch(horizontalAlignment) {
				case CENTER:
				case CENTER_SELECTION:
					attrs.put("ha", "c");
					break;
				case RIGHT:
					attrs.put("ha", "r");
					break;
				case LEFT:
				default:
					break;
				}

				SCellStyle.VerticalAlignment verticalAlignment = cellStyle.getVerticalAlignment();
				switch(verticalAlignment) {
				case TOP:
					attrs.put("va", "t");
					break;
				case CENTER:
					attrs.put("va", "c");
					break;
				//case CellStyle.VERTICAL_BOTTOM: //default
				//	break;
				}

				//ZSS-1171
				SFont font = StyleUtil
						.getFontStyle(sheet.getBook(), cellStyle, tbCellStyle, cdCellStyle); //ZSS-1142
				int fontSize = font.getHeightPoints();
				attrs.put("fs", fontSize); // fontsize

				final String fontName = font.getName();
				attrs.put("fn", fontName); // fontName

				//ZSS-1171: font format: bold(0x01)/italic(0x02)/underline(0x04)/strikout(0x08)
				int ff = 0;
				if (font.isItalic()) {
					ff |= 0x02;
				}
				if (font.isStrikeout()) {
					ff |= 0x08;
				}
				final SFont.Underline fontUnderline = font.getUnderline();
				boolean isUnderline = fontUnderline == SFont.Underline.SINGLE || fontUnderline == SFont.Underline.SINGLE_ACCOUNTING;
				if (isUnderline) {
					ff |= 0x04;
				}
				if (font.getBoldweight() == SFont.Boldweight.BOLD) {
					ff |= 0x01;
				}
				attrs.put("ff", ff);

				//ZSS-944: pass rotate info to browser
				final int rotate = cellStyle.getRotation();
				attrs.put("rot", rotate);
			}
		}

		//ZSS-1116: pass the is rich text flag to client
		if (cell.isRichTextValue()) {
			attrs.put("irt", true);
		}

		//ZSS-1116: pass the _calcAutoHeight flag to client
		if (((AbstractCellAdv)cell).isCalcAutoHeight())
			attrs.put("_cah", true);

		//ZSS-1171: pass the calculated text width to client
		final int textwidth = ((AbstractCellAdv)cell).getTextWidth();
		if (textwidth >= 0) {
			attrs.put("wd", textwidth);
		}
		return attrs;
	}

	public void insertColumns(SSheet sheet, int col, int size) {
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		if (size <= 0) {
			throw new UiException("size must > 0 : " + size);
		}
		//ZSS-1084
		int maxCols = _spreadsheet.getSheetMaxVisibleColumns(sheet);
		if (col > maxCols) {// not in scope, do nothing,
			return;
		}

		final AreaRef rect = _spreadsheet.getActiveRangeHelper().getArea(_spreadsheet.getSelectedSSheet());
		//ZSS-1284
		if (rect == null) { // Spreadsheet not loaded with data yet
			return;
		}

		// because of rowfreeze and colfreeze,
		// don't avoid insert behavior here, always send required data to
		// client,
		// let client handle it

		HashMap result = new HashMap();
		result.put("type", "column");
		result.put("col", col);
		result.put("size", size);

		int right = size + rect.getLastColumn();

		HeaderPositionHelper colHelper = _spreadsheet.getColumnPosition(sheet);
		colHelper.shiftMeta(col, size);
		result.put("hs", getColumnHeaderAttrs(_spreadsheet.getSelectedSSheet(), col, right));

		//_maxColumns += size;
		//			int cf = getColumnfreeze();
		//			if (cf >= col) {
		//				_colFreeze += size;
		//			}

		result.put("maxcol", maxCols); //ZSS-1084
		result.put("colfreeze", _spreadsheet.getSelectedSheetColumnfreeze());

		_spreadsheet.response("insertRowColumn" + XUtils
				.nextUpdateId(), new AuInsertRowColumn(_spreadsheet, "", sheet.getId(), result));

		rect.setLastColumn(right);

		// update surround cell
		int left = col;
		right = left + size - 1;
		right = right >= maxCols - 1 ? maxCols - 1 : right; //ZSS-1084
		int top = rect.getRow();
		int bottom = rect.getLastRow();
		if(log.isDebugEnabled()){
			log.debug("update cells when insert column " + col + ",size:" + size + ":" + left + "," + top + "," + right + "," + bottom);
		}
		_spreadsheet.updateCell(sheet, left, top, right, bottom, CellAttribute.ALL); //ZSS-939

		// ZSS-404: must update cell in freeze panels (previous range is only for data block)
		int rowFreeze = _spreadsheet.getSelectedSheetRowfreeze();
		if(rowFreeze >= 0) {
			_spreadsheet.updateCell(sheet, left, 0, right, rowFreeze, CellAttribute.ALL); // top freeze panel; ZSS-939
		}

		//update inserted column widths
		updateColWidths(sheet, col, size);

	}

	private boolean shouldShowValue(SCell cell) {
		if (_spreadsheet.isShowZeroValues()) {
			return true;
		}
		final SCell.CellType cellType = cell.getType();
		if (cellType == SCell.CellType.NUMBER || (cellType == SCell.CellType.FORMULA &&
						cell.getFormulaResultType() == SCell.CellType.NUMBER)) {
			final Object value = cell.getValue();
			if (value instanceof Number) {
				return ((Number) value).doubleValue() != 0;
			}
		}
		return true;
	}

	private void updateRowHeights(SSheet sheet, int row, int n) {
		for(int r = 0; r < n; ++r) {
			_spreadsheet.updateRowHeight(sheet, r+row);
		}
	}

	private void updateColWidths(SSheet sheet, int col, int n) {
		for(int r = 0; r < n; ++r) {
			_spreadsheet.updateColWidth(sheet, r+col);
		}
	}
	public void insertRows(SSheet sheet, int row, int size) {
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		if (size <= 0) {
			throw new UiException("size must > 0 : " + size);
		}
		//ZSS-1084
		int maxRows = _spreadsheet.getSheetMaxVisibleRows(sheet);
		if (row > maxRows) {// not in scrope, do nothing,
			return;
		}

		final AreaRef rect = _spreadsheet.getActiveRangeHelper().getArea(_spreadsheet.getSelectedSSheet());
		//ZSS-1284
		if (rect == null) { // Spreadsheet not loaded with data yet
			return;
		}

		// because of rowfreeze and colfreeze,
		// don't avoid insert behavior here, always send required data to
		// client,
		// let client handle it

		HashMap result = new HashMap();
		result.put("type", "row");
		result.put("row", row);
		result.put("size", size);

		int bottom = size + rect.getLastRow();

		HeaderPositionHelper rowHelper = _spreadsheet.getRowPosition(sheet);
		rowHelper.shiftMeta(row, size);

		result.put("hs", getRowHeaderAttrs(_spreadsheet.getSelectedSSheet(), row, bottom));

		//_maxRows += size;

		result.put("maxrow", maxRows); //ZSS-1084
		result.put("rowfreeze", _spreadsheet.getSelectedSheetRowfreeze());

		_spreadsheet.response("insertRowColumn" + XUtils.nextUpdateId(), new AuInsertRowColumn(_spreadsheet, "", sheet.getId(), result));

		rect.setLastRow(bottom);

		// update surround cell
		int top = row;
		bottom = top + size - 1;
		bottom = bottom >= maxRows - 1 ? maxRows - 1 : bottom; //ZSS-1084
		int left = rect.getColumn();
		int right = rect.getLastColumn();

		if (log.isDebugEnabled())
			log.debug("update cells when insert row " + row + ",size:" + size + ":" + left + "," + top + "," + right + "," + bottom);
		_spreadsheet.updateCell(sheet, left, top, right, bottom, CellAttribute.ALL);

		// ZSS-404: must update cell in freeze panels (previous range is only for data block)
		int colFreeze = _spreadsheet.getSelectedSheetColumnfreeze();
		if(colFreeze >= 0) {
			_spreadsheet.updateCell(sheet, 0, top, colFreeze, bottom, CellAttribute.ALL); // left freeze panel
		}

		// update the inserted row height
		updateRowHeights(sheet, row, size); //update row height
	}

	public void removeColumns(SSheet sheet, int col, int size) {
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		if (size <= 0) {
			throw new UiException("size must > 0 : " + size);
		} else if (col < 0) {
			throw new UiException("column must >= 0 : " + col);
		}
		//ZSS-1084
		int maxCols = _spreadsheet.getSheetMaxVisibleColumns(sheet);
		if (col >= maxCols) {
			return;
		}

		final AreaRef rect = _spreadsheet.getActiveRangeHelper().getArea(_spreadsheet.getSelectedSSheet());
		//ZSS-1284
		if (rect == null) { // Spreadsheet not loaded with data yet
			return;
		}

		if (col + size > maxCols) {
			size = maxCols - col;
		}

		// because of rowfreeze and colfreeze,
		// don't avoid insert behavior here, always send required data to
		// client,
		// let client handle it


		HashMap result = new HashMap();
		result.put("type", "column");
		result.put("col", col);
		result.put("size", size);

		int right = rect.getLastColumn() - size;
		if (right < col) {
			right = col - 1;
		}

		HeaderPositionHelper colHelper = _spreadsheet.getColumnPosition(sheet);
		colHelper.unshiftMeta(col, size);

		result.put("hs", getColumnHeaderAttrs(_spreadsheet.getSelectedSSheet(), col, right));


		//_maxColumns -= size;

		result.put("maxcol", maxCols); //ZSS-1084
		result.put("colfreeze", _spreadsheet.getSelectedSheetColumnfreeze());

		_spreadsheet.response("removeRowColumn" + XUtils.nextUpdateId(), new AuRemoveRowColumn(_spreadsheet, "", sheet.getId(), result));
		rect.setLastColumn(right);

		// update surround cell
		int left = col;
		right = left;

		_spreadsheet.updateCell(sheet, left, rect.getRow(), right, rect.getLastRow(), CellAttribute.ALL); //ZSS-939
	}

	public void removeRows(SSheet sheet, int row, int size) {
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		if (size <= 0) {
			throw new UiException("size must > 0 : " + size);
		} else if (row < 0) {
			throw new UiException("row must >= 0 : " + row);
		}
		//ZSS-1084
		int maxRows = _spreadsheet.getSheetMaxVisibleRows(sheet);
		if (row >= maxRows) {
			return;
		}

		final AreaRef rect = _spreadsheet.getActiveRangeHelper().getArea(_spreadsheet.getSelectedSSheet());
		//ZSS-1284
		if (rect == null) { // Spreadsheet not loaded with data yet
			return;
		}

		if (row + size > maxRows) {
			size = maxRows - row;
		}

		// because of rowfreeze and colfreeze,
		// don't avoid insert behavior here, always send required data to
		// client,
		// let client handle it

		HashMap result = new HashMap();
		result.put("type", "row");
		result.put("row", row);
		result.put("size", size);

		int bottom = rect.getLastRow() - size;
		if (bottom < row) {
			bottom = row - 1;
		}

		HeaderPositionHelper rowHelper = _spreadsheet.getRowPosition(sheet);
		rowHelper.unshiftMeta(row, size);

		result.put("hs", getRowHeaderAttrs(_spreadsheet.getSelectedSSheet(), row, bottom));

		//			_maxRows -= size;

		result.put("maxrow", maxRows); //ZSS-1084
		result.put("rowfreeze", _spreadsheet.getSelectedSheetRowfreeze());

		_spreadsheet.response("removeRowColumn" + XUtils.nextUpdateId(), new AuRemoveRowColumn(_spreadsheet, "", sheet.getId(), result));
		rect.setLastRow(bottom);

		// update surround cell
		int top = row;
		bottom = top;

		_spreadsheet.updateCell(sheet, rect.getColumn(), top, rect.getLastColumn(), bottom, CellAttribute.ALL); //ZSS-939
	}

	public void updateMergeCell(SSheet sheet, int left, int top, int right,
			int bottom, int oleft, int otop, int oright, int obottom) {
		deleteMergeCell(sheet, oleft, otop, oright, obottom);
		addMergeCell(sheet, left, top, right, bottom);
	}

	public void deleteMergeCell(SSheet sheet, int left, int top, int right, int bottom) {
		MergeMatrixHelper mmhelper = this.getMergeMatrixHelper(sheet);
		Set torem = new HashSet();
		mmhelper.deleteMergeRange(left, top, right, bottom, torem);
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		for (Iterator iter = torem.iterator(); iter.hasNext();) {
			MergedRect rect = (MergedRect) iter.next();

			updateMergeCell0(sheet, rect, "remove");
		}
		//updateCell(sheet, left > 0 ? left - 1 : 0, top > 1 ? top - 1 : 0, right + 1, bottom + 1);
		_spreadsheet.updateCell(sheet, left, top, right, bottom, CellAttribute.ALL); //ZSS-939
	}
	//ZSS-1168
	public void syncMergeCells(SSheet sheet) {
		_spreadsheet.removeMergeMatrixHelper(sheet);  //ZSS-1204

		//ZSS-1204
		if (_spreadsheet.isInvalidated())
			return;// since it is invalidate, we don't need to do anymore

		//ZSS-1168: so will NOT sync merge if NOT current selected sheet
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}

		MergeMatrixHelper mmhelper = this.getMergeMatrixHelper(sheet);
		JSONObject results = new JSONObject();
		results.put("type", "sync");
		JSONArray jarray = new JSONArray();
		results.put("rects", jarray);
		for (MergedRect block : mmhelper.getRanges()) {
			JSONObject result = new JSONObject();
			prepareMergeCell0(sheet, block, result);
			jarray.add(result);
		}
		_spreadsheet.response("mergeCell" + XUtils.nextUpdateId(), new AuMergeCell(_spreadsheet, "", sheet.getId(), results));
	}

	private void updateMergeCell0(SSheet sheet, MergedRect block, String type) {
		JSONObject result = new JSONObject();
		result.put("type", type);

		//ZSS-1168
		prepareMergeCell0(sheet, block, result);

		/**
		 * merge_ -> mergeCell
		 */
		_spreadsheet.response("mergeCell" + XUtils.nextUpdateId(), new AuMergeCell(_spreadsheet, "", sheet.getId(), result));
	}

	//ZSS-1168
	private void prepareMergeCell0(SSheet sheet, MergedRect block, JSONObject result) {
		result.put("id", block.getId());
		int left = block.getColumn();
		int top = block.getRow();
		int right = block.getLastColumn();
		int bottom = block.getLastRow();

		// don't check range to ignore update case,
		// because I still need to sync merge cell data to client side

		result.put("left", left);
		result.put("top", top);
		result.put("right", right);
		result.put("bottom", bottom);

		HeaderPositionHelper helper = _spreadsheet
				.getColumnPosition(sheet);
		final int w = helper.getStartPixel(block.getLastColumn() + 1) - helper.getStartPixel(block.getColumn());
		result.put("width", w);

		HeaderPositionHelper rhelper = _spreadsheet
				.getRowPosition(sheet);
		final int h = rhelper.getStartPixel(block.getLastRow() + 1) - rhelper.getStartPixel(block.getRow());
		result.put("height", h);
	}

	public void addMergeCell(SSheet sheet, int left, int top, int right, int bottom) {
		MergeMatrixHelper mmhelper = this.getMergeMatrixHelper(sheet);

		Set toadd = new HashSet();
		Set torem = new HashSet();
		mmhelper.addMergeRange(left, top, right, bottom, toadd, torem);

		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}

		for (Iterator iter = torem.iterator(); iter.hasNext();) {
			MergedRect rect = (MergedRect) iter.next();
			if (log.isDebugEnabled()) {
				log.debug("(A)remove merge:" + rect);
			}
			updateMergeCell0(sheet, rect, "remove");
		}
		for (Iterator iter = toadd.iterator(); iter.hasNext();) {
			MergedRect rect = (MergedRect) iter.next();
			if (log.isDebugEnabled()) {
				log.debug("add merge:" + rect);
			}
			updateMergeCell0(sheet, rect, "add");
		}
		//			updateCell(sheet, left > 0 ? left - 1 : 0, top > 1 ? top - 1 : 0, right + 1, bottom + 1);
		_spreadsheet.updateCell(sheet, left, top, right, bottom, CellAttribute.ALL); //ZSS-939
	}

	//in pixel
	public void setColumnWidth(SSheet sheet, int col, int width, int id, boolean hidden) {
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		JSONObject result = new JSONObject();
		result.put("type", "column");
		result.put("column", col);
		result.put("width", width);
		result.put("id", id);
		result.put("hidden", hidden);
		_spreadsheet.smartUpdate("columnSize", (Object) new Object[] { "", sheet.getId(), result}, true);
	}

	//in pixels
	public void setRowHeight(SSheet sheet, int row, int height, int id, boolean hidden, boolean isCustom) {
		if (_spreadsheet.tryReleaseCache(sheet)) {
			return;
		}
		JSONObject result = new JSONObject();
		result.put("type", "row");
		result.put("row", row);
		result.put("height", height);
		result.put("id", id);
		result.put("hidden", hidden);
		result.put("custom", isCustom);
		_spreadsheet.smartUpdate("rowSize", (Object) new Object[] { "", sheet.getId(), result}, true);
	}

	@Override
	public Boolean getLeftHeaderHiddens(int row) {
		SSheet sheet = _spreadsheet.getSelectedSSheet();
		HeaderPositionHelper rowHelper = _spreadsheet
				.getRowPosition(sheet);
		HeaderPositionHelper.HeaderPositionInfo info = rowHelper.getInfo(row);
		return info == null ? Boolean.FALSE : Boolean.valueOf(info.hidden);
	}

	@Override
	public Boolean getTopHeaderHiddens(int col) {
		SSheet sheet = _spreadsheet.getSelectedSSheet();
		HeaderPositionHelper colHelper = _spreadsheet
				.getColumnPosition(sheet);
		HeaderPositionHelper.HeaderPositionInfo info = colHelper.getInfo(col);
		return info == null ? Boolean.FALSE : Boolean.valueOf(info.hidden);
	}

	@Override
	public void setSelectedSheetDirectly(String name,
			boolean cacheInClient, int row, int col, int left, int top,
			int right, int bottom, int highlightLeft, int highlightTop,
			int highlightRight, int highlightBottom, int rowfreeze,
			int colfreeze) {
		_spreadsheet.setSelectedSheetDirectly(name, cacheInClient, row, col, left,
				top, right, bottom/*, highlightLeft, highlightTop,
					highlightRight, highlightBottom, rowfreeze, colfreeze*/);
	}

	@Override
	public FreezeInfoLoader getFreezeInfoLoader() {
		return _spreadsheet.getFreezeInfoLoader();
	}

	//ZSS-717
	@Override
	public void setSelType(CellSelectionType type) {
		getSpreadSheetSelectionArea().setSelType(type);
	}

	//ZSS-717
	@Override
	public CellSelectionType getSelType() {
		return getSpreadSheetSelectionArea().getSelType();
	}

	public void releaseBook() {
		_spreadsheet.releaseBook();
	}

	// KEIKAI-224
	void calcAutoFilterBottomRegion(SSheet sheet) {
		final SAutoFilter autoFilter = sheet.getAutoFilter();
		if (autoFilter == null) {
			return;
		}
		int lastVisible = autoFilter.getRegion().getLastRow();
		final Iterator<AbstractRowAdv> iter = ((AbstractSheetAdv) sheet).getRowIterator(true,
				autoFilter.getRegion().getRow(), autoFilter.getRegion().getLastRow());
		while (iter.hasNext()) {
			final AbstractRowAdv row = iter.next();
			final int index = row.getIndex();
			if (row.isHidden()) {
				if (lastVisible == index) {
					lastVisible = index - 1;
				} else {
					break;
				}
			} else {
				lastVisible = index;
				break;
			}
		}
		autoFilter.setLastVisibleRow(lastVisible);
	}

	// KEIKAI-429
	private AreaRefWithType getSpreadSheetSelectionArea() {
		Set<AreaRefWithType> selectionArea = _spreadsheet.getSelectionArea();
		if (selectionArea.isEmpty()) {
			selectionArea.add(new AreaRefWithType(0, 0, 0, 0, CellSelectionType.CELL));
		}
		return selectionArea.iterator().next();
	}

	@Override
	public void addSelectionRect(int left, int top, int right, int bottom) {
		addSelectionRect(left, top, right, bottom, CellSelectionType.CELL);
	}

	@Override
	public void addSelectionRect(int left, int top, int right, int bottom, CellSelectionType type) {
		AreaRefWithType newRect = new AreaRefWithType(top, left, bottom, right, type);
		_spreadsheet.getSelectionArea().add(newRect);
		SSheet selectedSSheet = _spreadsheet.getSelectedSSheet();
		if (selectedSSheet != null) {
			selectedSSheet.getViewInfo().addSelectionAreas(newRect.asString());
		}
	}
}
