package io.keikai.ui;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zkoss.mesg.Messages;
import org.zkoss.xel.VariableResolver;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.Desktop;
import org.zkoss.zk.ui.Executions;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zk.ui.event.Event;
import org.zkoss.zk.ui.event.EventListener;
import org.zkoss.zk.ui.event.SerializableEventListener;
import org.zkoss.zul.Messagebox;
import org.zkoss.zul.impl.MessageboxDlg;
import org.zkoss.zul.mesg.MZul;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class ZSMessagebox extends Messagebox {
	private static final Logger log = LoggerFactory.getLogger(ZSMessagebox.class);

	private static String _templ = "~./zssex/dlg/zsmessagebox.zul";

	private static final Button[] DEFAULT_BUTTONS;

	static {
		DEFAULT_BUTTONS = new Button[]{Button.OK};
	}

	public static Button show(String message, String title, Button[] buttons, String[] btnLabels, String icon, Button focus, EventListener<ClickEvent> listener, Map<String, String> params) {
		Map<String, Object> arg = new HashMap();
		Desktop desktop = Executions.getCurrent().getDesktop();
		arg.put("message", message);
		arg.put("title", title != null ? title : desktop.getWebApp().getAppName());
		arg.put("icon", icon);
		if (buttons == null) {
			buttons = DEFAULT_BUTTONS;
		}

		int btnmask = 0;

		for(int j = 0; j < buttons.length; ++j) {
			if (buttons[j] == null) {
				throw new IllegalArgumentException("The " + j + "-th button is null");
			}

			btnmask += buttons[j].id;
			arg.put(buttons[j].stringId, buttons[j].id);
		}

		arg.put("buttons", btnmask);
		if (params != null) {
			arg.putAll(params);
		}

		ZSMessageboxDlg dlg = (ZSMessageboxDlg)Executions.createComponents(_templ, desktop.getFirstPage(), (VariableResolver)null, arg)[0];
		dlg.setEventListener(listener);
		dlg.setButtons(buttons, btnLabels);
		if (focus != null) {
			dlg.setFocus(focus);
		}

		if (dlg.getDesktop().getWebApp().getConfiguration().isEventThreadEnabled()) {
			try {
				dlg.doModal();
			} catch (Throwable var15) {
				try {
					dlg.detach();
				} catch (Throwable var14) {
					log.warn("Failed to detach when recovering from an error", var14);
				}

				throw UiException.Aide.wrap(var15);
			}

			return dlg.getResult();
		} else {
			dlg.doHighlighted();
			return Button.OK;
		}
	}

	public static Button show(String message, String title, Button[] buttons, String[] btnLabels, String icon, Button focus, EventListener<ClickEvent> listener) {
		return show(message, title, buttons, btnLabels, icon, focus, listener, (Map)null);
	}

	public static Button show(String message, String title, Button[] buttons, String icon, Button focus, EventListener<ClickEvent> listener) {
		return show(message, title, buttons, (String[])null, icon, focus, listener, (Map)null);
	}

	public static Button show(String message, String title, Button[] buttons, String icon, EventListener<ClickEvent> listener) {
		return show(message, title, buttons, (String[])null, icon, (Button)null, listener, (Map)null);
	}

	public static Messagebox.Button show(String message, Messagebox.Button[] buttons, EventListener<Messagebox.ClickEvent> listener) {
		return show(message, (String)null, buttons, (String[])null, "z-messagebox-icon z-messagebox-information", (Messagebox.Button)null, listener, (Map)null);
	}

	public static int show(String message, String title, int buttons, String icon) {
		return show(message, title, buttons, icon, 0, (EventListener)null);
	}

	public static int show(String message, String title, int buttons, String icon, EventListener<Event> listener) {
		return show(message, title, buttons, icon, 0, listener);
	}

	public static int show(String message, String title, int buttons, String icon, int focus) {
		return show(message, title, buttons, icon, focus, (EventListener)null);
	}

	public static int show(String message, String title, int buttons, String icon, int focus, EventListener<Event> listener) {
//		String temp = Messagebox.getTemplate();
//		Messagebox.setTemplate(_templ);
		Button res = show(message, title, toButtonTypes(buttons), (String[])null, icon, focus != 0 ? toButtonType(focus) : null, toButtonListener(listener), (Map)null);
		return res != null ? res.id : 0;
//		int id = Messagebox.show(message, title, buttons, icon, focus, listener);
//		Messagebox.setTemplate(temp);
//		return id;
	}

	public static int show(String message) {
		return show(message, (String)null, 1, "z-messagebox-icon z-messagebox-information", 0, (EventListener)null);
	}

	public static int show(int messageCode, Object[] args, int titleCode, int buttons, String icon) {
		return show(messageCode, (Object[])args, titleCode, buttons, icon, 0, (EventListener)null);
	}

	public static int show(int messageCode, Object[] args, int titleCode, int buttons, String icon, int focus) {
		return show(messageCode, (Object[])args, titleCode, buttons, icon, focus, (EventListener)null);
	}

	public static int show(int messageCode, Object[] args, int titleCode, int buttons, String icon, int focus, EventListener<Event> listener) {
		return show(Messages.get(messageCode, args), titleCode > 0 ? Messages.get(titleCode) : null, buttons, icon, focus, listener);
	}

	public static int show(int messageCode, Object arg, int titleCode, int buttons, String icon) {
		return show(messageCode, (Object)arg, titleCode, buttons, icon, 0, (EventListener)null);
	}

	public static int show(int messageCode, Object arg, int titleCode, int buttons, String icon, int focus) {
		return show(messageCode, (Object)arg, titleCode, buttons, icon, focus, (EventListener)null);
	}

	public static int show(int messageCode, Object arg, int titleCode, int buttons, String icon, int focus, EventListener<Event> listener) {
		return show(Messages.get(messageCode, arg), titleCode > 0 ? Messages.get(titleCode) : null, buttons, icon, focus, listener);
	}

	public static int show(int messageCode, int titleCode, int buttons, String icon) {
		return show(messageCode, titleCode, buttons, icon, 0);
	}

	public static int show(int messageCode, int titleCode, int buttons, String icon, int focus) {
		return show(messageCode, titleCode, buttons, icon, focus, (EventListener)null);
	}

	public static int show(int messageCode, int titleCode, int buttons, String icon, int focus, EventListener<Event> listener) {
		return show(Messages.get(messageCode), titleCode > 0 ? Messages.get(titleCode) : null, buttons, icon, focus, listener);
	}

	private static Button toButtonType(int btn) {
		switch(btn) {
			case 2:
				return Button.CANCEL;
			case 16:
				return Button.YES;
			case 32:
				return Button.NO;
			case 256:
				return Button.ABORT;
			case 512:
				return Button.RETRY;
			case 1024:
				return Button.IGNORE;
			default:
				return Button.OK;
		}
	}

	private static Button[] toButtonTypes(int buttons) {
		List<Button> btntypes = new ArrayList();
		if ((buttons & 1) != 0) {
			btntypes.add(toButtonType(1));
		}

		if ((buttons & 2) != 0) {
			btntypes.add(toButtonType(2));
		}

		if ((buttons & 16) != 0) {
			btntypes.add(toButtonType(16));
		}

		if ((buttons & 32) != 0) {
			btntypes.add(toButtonType(32));
		}

		if ((buttons & 512) != 0) {
			btntypes.add(toButtonType(512));
		}

		if ((buttons & 256) != 0) {
			btntypes.add(toButtonType(256));
		}

		if ((buttons & 1024) != 0) {
			btntypes.add(toButtonType(1024));
		}

		return (Button[])btntypes.toArray(new Button[btntypes.size()]);
	}


	private static EventListener<ClickEvent> toButtonListener(EventListener<Event> listener) {
		return listener != null ? new ButtonListener(listener) : null;
	}

	public static class ClickEvent extends Event {
		public ClickEvent(String name, Component target, Button button) {
			super(name, target, button);
		}

		public Button getButton() {
			return (Button)this.getData();
		}
	}

	private static class ButtonListener implements SerializableEventListener<ClickEvent> {
		private final EventListener<Event> _listener;

		private ButtonListener(EventListener<Event> listener) {
			this._listener = listener;
		}

		public void onEvent(ClickEvent event) throws Exception {
			Button btn = event.getButton();
			this._listener.onEvent(new Event(event.getName(), event.getTarget(), btn != null ? btn.id : -1));
		}

		public String toString() {
			return this._listener.toString();
		}
	}

	public static enum Button {
		OK(1, "onOK", MZul.OK, "OK"),
		CANCEL(2, "onCancel", MZul.CANCEL, "CANCEL"),
		YES(16, "onYes", MZul.YES, "YES"),
		NO(32, "onNo", MZul.NO, "NO"),
		ABORT(256, "onAbort", MZul.ABORT, "ABORT"),
		RETRY(512, "onRetry", MZul.RETRY, "RETRY"),
		IGNORE(1024, "onIgnore", MZul.IGNORE, "IGNORE");

		public final int id;
		public final String event;
		public final int label;
		private final String stringId;

		private Button(int id, String event, int label, String stringId) {
			this.id = id;
			this.event = event;
			this.label = label;
			this.stringId = stringId;
		}
	}

}
