package io.keikai.ui;

import org.zkoss.mesg.Messages;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zk.ui.event.EventListener;
import io.keikai.ui.ZSMessagebox;
import org.zkoss.zul.Window;

public class ZSMessageboxDlg extends Window {
	private ZSMessagebox.Button[] _buttons;
	private ZSMessagebox.Button _result;
	private EventListener<ZSMessagebox.ClickEvent> _listener;

	public ZSMessageboxDlg() {
	}

	public void onOK() throws Exception {
		if (this.contains(ZSMessagebox.Button.OK)) {
			this.endModal(ZSMessagebox.Button.OK);
		} else if (this.contains(ZSMessagebox.Button.YES)) {
			this.endModal(ZSMessagebox.Button.YES);
		} else if (this.contains(ZSMessagebox.Button.RETRY)) {
			this.endModal(ZSMessagebox.Button.RETRY);
		}

	}

	public void onCancel() throws Exception {
		if (this._buttons.length == 1 && this._buttons[0] == ZSMessagebox.Button.OK) {
			this.endModal(ZSMessagebox.Button.OK);
		} else if (this.contains(ZSMessagebox.Button.CANCEL)) {
			this.endModal(ZSMessagebox.Button.CANCEL);
		} else if (this.contains(ZSMessagebox.Button.NO)) {
			this.endModal(ZSMessagebox.Button.NO);
		} else if (this.contains(ZSMessagebox.Button.ABORT)) {
			this.endModal(ZSMessagebox.Button.ABORT);
		}

	}

	private boolean contains(ZSMessagebox.Button button) {
		for(int j = 0; j < this._buttons.length; ++j) {
			if (this._buttons[j] == button) {
				return true;
			}
		}

		return false;
	}

	public void setButtons(ZSMessagebox.Button[] buttons, String[] btnLabels) {
		this._buttons = buttons;
		Component parent = this.getFellowIfAny("buttons");
		if (parent != null && parent.getFirstChild() == null) {
			String sclass = (String)parent.getAttribute("button.sclass");
			String zclass = (String)parent.getAttribute("button.zclass");

			for(int j = 0; j < this._buttons.length; ++j) {
				ZSMessageboxDlg.Button mbtn = new ZSMessageboxDlg.Button();
				mbtn.setButton(this._buttons[j], btnLabels != null && j < btnLabels.length ? btnLabels[j] : null);
				mbtn.setSclass(sclass);
				mbtn.setZclass(zclass);
				mbtn.setAutodisable("self");
				parent.appendChild(mbtn);
			}
		}

	}

	public void setEventListener(EventListener<ZSMessagebox.ClickEvent> listener) {
		this._listener = listener;
	}

	public void setFocus(ZSMessagebox.Button button) {
		if (button != null) {
			ZSMessageboxDlg.Button btn = (ZSMessageboxDlg.Button)this.getFellowIfAny("btn" + button.id);
			if (btn != null) {
				btn.focus();
			}
		}

	}

	public void endModal(ZSMessagebox.Button button) throws Exception {
		this._result = button;
		if (this._listener != null) {
			ZSMessagebox.ClickEvent evt = new ZSMessagebox.ClickEvent(button.event, this, button);
			this._listener.onEvent(evt);
			if (!evt.isPropagatable()) {
				return;
			}
		}

		this.detach();
	}

	public ZSMessagebox.Button getResult() {
		return this._result;
	}

	public void onClose() {
		if (this._listener != null) {
			ZSMessagebox.ClickEvent evt = new ZSMessagebox.ClickEvent("onClose", this, (ZSMessagebox.Button)null);

			try {
				this._listener.onEvent(evt);
				if (!evt.isPropagatable()) {
					return;
				}
			} catch (Exception var3) {
				throw UiException.Aide.wrap(var3);
			}
		}

		super.onClose();
	}

	public static class Button extends org.zkoss.zul.Button {
		private ZSMessagebox.Button _button;

		public Button() {
		}

		public void setButton(ZSMessagebox.Button button, String label) {
			this._button = button;
			this.setLabel(label != null ? label : Messages.get(this._button.label));
			this.setId("btn" + this._button.id);
			if (label != null && label.length() > 7) {
				this.setWidth("auto");
			}

		}

		public void setButton(ZSMessagebox.Button button) {
			this.setButton(button, (String)null);
		}

		public void onClick() throws Exception {
			((ZSMessageboxDlg)this.getSpaceOwner()).endModal(this._button);
		}

		protected String getDefaultMold(Class klass) {
			return super.getDefaultMold(org.zkoss.zul.Button.class);
		}

	}
}
