/* CellSelectionCommand.java
 * 
{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		January 10, 2008 03:10:40 PM , Created by Dennis.Chen
}}IS_NOTE

Copyright (C) 2007 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
	This program is distributed under Lesser GPL Version 2.1 in the hope that
	it will be useful, but WITHOUT ANY WARRANTY.
}}IS_RIGHT
*/
package io.keikai.ui.au.in;


import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import io.keikai.api.AreaRefWithType;
import io.keikai.api.model.Sheet;
import io.keikai.api.model.impl.SheetImpl;
import io.keikai.model.SBook;
import io.keikai.ui.CellSelectionType;
import io.keikai.ui.Spreadsheet;
import io.keikai.ui.event.CellMultipleSelectionEvent;
import io.keikai.ui.event.CellSelectionEvent;
import io.keikai.ui.sys.SpreadsheetInCtrl;

import org.zkoss.json.JSONArray;
import org.zkoss.json.JSONObject;
import org.zkoss.lang.Objects;
import org.zkoss.zk.au.AuRequest;
import org.zkoss.zk.mesg.MZk;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zk.ui.event.Events;


/**
 * A Command (client to server) for handling cell selection
 * @author Dennis.Chen
 *
 */
public class CellSelectionCommand extends AbstractCommand implements Command {

	public void process(AuRequest request) {
		final Component comp = request.getComponent();
		if (comp == null)
			throw new UiException(MZk.ILLEGAL_REQUEST_COMPONENT_REQUIRED, CellSelectionCommand.class.getCanonicalName());

		final Map data = (Map) request.getData();
		if (data == null || (data.size() != 2 && data.size() != 3))
			throw new UiException(MZk.ILLEGAL_REQUEST_WRONG_DATA, new Object[] {Objects.toString(data), this});
			
		String sheetId= (String) data.get("sheetId");
		
		Sheet sheet = ((Spreadsheet)comp).getSelectedSheet();
		
		if(!getSheetUuid(sheet).equals(sheetId))
			return;
		
		//TODO request shall send back maxcol/maxrow (do it in client side)
		final SBook book = ((SheetImpl)sheet).getNative().getBook();
		final int maxcol = book.getMaxColumnIndex();
		final int maxrow = book.getMaxRowIndex();
		
		JSONArray areas = (JSONArray) data.get("areas");
		int areasSize = areas.size();
		SpreadsheetInCtrl ctrl = ((SpreadsheetInCtrl)((Spreadsheet)comp).getExtraCtrl());
		Set<AreaRefWithType> areasWithTypes = new LinkedHashSet<>();
		for (int i = 0; i < areasSize; i++) {
			JSONObject areaData = (JSONObject) areas.get(i);
			int left = (Integer) areaData.get("left");
			int top = (Integer) areaData.get("top");
			String t = (String) areaData.get("type");
			CellSelectionType type = null;
			if ("all".equals(t)) {
				type = CellSelectionType.ALL;
			} else if ("col".equals(t)) {
				type = CellSelectionType.COLUMN;
			} else if ("row".equals(t)) {
				type = CellSelectionType.ROW;
			} else if ("cell".equals(t)) {
				type = CellSelectionType.CELL;
			} else {
				throw new UiException(MZk.ILLEGAL_REQUEST_WRONG_DATA, new Object[]{Objects.toString(data), this});
			}
			int right = (type == CellSelectionType.ROW || type == CellSelectionType.ALL) ? maxcol : (Integer) areaData.get("right");
			int bottom = (type == CellSelectionType.COLUMN || type == CellSelectionType.ALL) ? maxrow : (Integer) areaData.get("bottom");
			AreaRefWithType areaRefWithType = new AreaRefWithType(left, top, right, bottom, type);
			areasWithTypes.add(areaRefWithType);
			if (i == 0) {
				ctrl.setSelectionRect(left, top, right, bottom);
				ctrl.setSelType(type); //ZSS-717
				Events.postEvent(new CellSelectionEvent(
						io.keikai.ui.event.Events.ON_CELL_SELECTION, comp, sheet,
						top, left, bottom, right,type));
			} else {
				ctrl.addSelectionRect(left, top, right, bottom, type);
			}
		}
		if (areasSize > 1) { // multiple selection
			Events.postEvent(new CellMultipleSelectionEvent(
				io.keikai.ui.event.Events.ON_CELL_MULTIPLE_SELECTION, comp, sheet, areasWithTypes, data));
		}
	}
}