package io.keikai.ui.au.in;

import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.model.Sheet;
import io.keikai.ui.Spreadsheet;
import io.keikai.ui.event.ClipboardPasteEvent;
import org.zkoss.json.JSONArray;
import org.zkoss.util.resource.Labels;
import org.zkoss.zk.au.AuRequest;
import org.zkoss.zk.au.out.AuInvoke;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.event.Event;
import org.zkoss.zk.ui.event.Events;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * A command for client-side ctrl+v paste.
 * KEIKAI-40: Fire an event when ZSS completed client-side paste
 * @author wenninghsu
 */
public class ClipboardPasteCommand extends AbstractCommand implements Command {

	@Override
	public void process(AuRequest request) {
		Map data = request.getData();
		final Component comp = request.getComponent();
		Sheet sheet = ((Spreadsheet) comp).getSelectedSheet();
		final String sheetId = (String) data.get("sheetId");
		if (!getSheetUuid(sheet).equals(sheetId)) {
			return;
		}
		final int left = AuDataUtil.getInt(data, "left");
		final int top = AuDataUtil.getInt(data, "top");
		final int right = AuDataUtil.getInt(data, "right");
		final int bottom = AuDataUtil.getInt(data, "bottom");

		if (AuDataUtil.getBoolean(data, "recheckProtected") && isRangeProtected(sheet, top, bottom, left, right)) {
			((Spreadsheet) comp).response("cancelClipboardPaste", new AuInvoke((Component) comp, "cancelClipboardPaste"));
		} else {
			final ClipboardPasteEvent event = new ClipboardPasteEvent(comp, sheet, top, left, bottom, right);
			final JSONArray json = (JSONArray) data.get("values");
			final List<String[]> values = new ArrayList<>();
			for (final Object json0: json) {
				final List<String> arr = new ArrayList<>();
				for (final Object json1: (JSONArray) json0) {
					arr.add(json1.toString());
				}
				values.add(arr.toArray(new String[0]));
			}
			final String[][] values0 = values.toArray(new String[0][0]);
			Events.postEvent(new Event("onClipboardPasteImpl", comp, new Object[] {event, values0}));
			Events.postEvent(event);
		}
	}

	private boolean isRangeProtected(Sheet sheet, int row, int lastRow, int col, int lastCol){
		boolean protect = sheet.isProtected();
		if (!protect) {
			return false;
		}
		for (int i = row; i <= lastRow; i++) {
			for (int j = col; j <= lastCol; j++) {
				Range r = Ranges.range(sheet, i, j);
				boolean lock = r.getCellStyle().isLocked();
				if (lock) {
					return true;
				}
			}
		}
		return false;
	}
}
