/* GroupHeaderClickCommand.java

        Purpose:
                
        Description:
                
        History:
                2023/11/2 AM 09:42, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui.au.in;

import java.util.Map;

import io.keikai.api.model.Sheet;
import io.keikai.model.CellRegion;
import io.keikai.model.SColumn;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.range.impl.NotifyChangeHelper;
import io.keikai.ui.Spreadsheet;
import io.keikai.ui.event.GroupHeaderClickEvent;

import org.zkoss.lang.Objects;
import org.zkoss.zk.au.AuRequest;
import org.zkoss.zk.mesg.MZk;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.UiException;

public class GroupHeaderClickCommand extends AbstractCommand implements Command {
	@Override
	public void process(AuRequest request) {
		final Component comp = request.getComponent();
		if (comp == null)
			throw new UiException(MZk.ILLEGAL_REQUEST_COMPONENT_REQUIRED, this);
		Map<String, Object> data = request.getData();
		if (data == null || data.size() != 3)
			throw new UiException(MZk.ILLEGAL_REQUEST_WRONG_DATA, new Object[]{Objects.toString(data), this});

		String sheetId = (String) data.get("sheetId");
		int level = (int) data.get("level");
		String dir = (String) data.get("dir");
		Sheet sheet = ((Spreadsheet) comp).getSelectedSheet();
		if (!getSheetUuid(sheet).equals(sheetId))
			return;
		SSheet internalSheet = sheet.getInternalSheet();
		int previousOutlineLevel = -1;
		if ("row".equals(dir)) {
			if (internalSheet.isSummaryBelow()) {
				for (int i = sheet.getFirstRow(), j = sheet.getLastRow(); i <= j; j--) {
					SRow prev = internalSheet.getRow(j);
					int outlineLevel = prev.getOutlineLevel();
					if (outlineLevel == 0) {
						previousOutlineLevel = outlineLevel;
						continue;
					}
					boolean collapsed = level < outlineLevel;
					if (previousOutlineLevel != outlineLevel) {
						internalSheet.getRow(prev.getIndex() + 1).setCollapsed(collapsed);
					}
					previousOutlineLevel = outlineLevel;
					prev.setHidden(collapsed);
				}
			} else {
				for (int i = sheet.getFirstRow(), j = sheet.getLastRow(); i <= j; i++) {
					SRow next = internalSheet.getRow(i);
					int outlineLevel = next.getOutlineLevel();
					if (outlineLevel == 0) {
						previousOutlineLevel = outlineLevel;
						continue;
					}
					boolean collapsed = level < outlineLevel;
					if (previousOutlineLevel != outlineLevel) {
						internalSheet.getRow(next.getIndex() - 1).setCollapsed(collapsed);
					}
					previousOutlineLevel = outlineLevel;
					next.setHidden(collapsed);
				}
			}
		} else {
			if (internalSheet.isSummaryRight()) {
				CellRegion dataRegion = internalSheet.getDataRegion();
				for (int i = dataRegion.column, j = dataRegion.lastColumn; i <= j; j--) {
					SColumn prev = internalSheet.getColumn(j);
					int outlineLevel = prev.getOutlineLevel();
					if (outlineLevel == 0) {
						previousOutlineLevel = outlineLevel;
						continue;
					}
					boolean collapsed = level < outlineLevel;
					if (previousOutlineLevel != outlineLevel) {
						SColumn column = internalSheet.getColumn(
								prev.getIndex() + 1);
						column.setCollapsed(collapsed);
					}
					previousOutlineLevel = outlineLevel;
					prev.setHidden(collapsed);
				}
			} else {
				CellRegion dataRegion = internalSheet.getDataRegion();
				for (int i = dataRegion.column, j = dataRegion.lastColumn; i <= j; i++) {
					SColumn next = internalSheet.getColumn(i);
					int outlineLevel = next.getOutlineLevel();
					if (outlineLevel == 0) {
						previousOutlineLevel = outlineLevel;
						continue;
					}
					boolean collapsed = level < outlineLevel;
					if (previousOutlineLevel != outlineLevel) {
						SColumn column = internalSheet.getColumn(
								next.getIndex() - 1);
						column.setCollapsed(collapsed);
					}
					previousOutlineLevel = outlineLevel;
					next.setHidden(collapsed);
				}
			}
		}
		org.zkoss.zk.ui.event.Events.postEvent(new GroupHeaderClickEvent(request.getCommand(), comp, sheet,
				level, dir));
		new NotifyChangeHelper().notifySheetGroupChange(sheet.getInternalSheet());
	}
}