/* GroupOpenCommand.java

        Purpose:
                
        Description:
                
        History:
                2023/11/2 AM 09:20, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui.au.in;

import java.util.Map;

import io.keikai.api.model.Sheet;
import io.keikai.model.SColumn;
import io.keikai.model.SRow;
import io.keikai.model.SSheet;
import io.keikai.range.impl.NotifyChangeHelper;
import io.keikai.ui.Spreadsheet;
import io.keikai.ui.event.GroupOpenEvent;
import org.apache.poi.ss.SpreadsheetVersion;

import org.zkoss.lang.Objects;
import org.zkoss.zk.au.AuRequest;
import org.zkoss.zk.mesg.MZk;
import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.UiException;

/**
 * The group open command to update group open info.
 *
 * @author jumperchen
 * @since 6.0.0
 */
public class GroupOpenCommand extends AbstractCommand implements Command {
	@Override
	public void process(AuRequest request) {
		final Component comp = request.getComponent();
		if (comp == null)
			throw new UiException(MZk.ILLEGAL_REQUEST_COMPONENT_REQUIRED, this);
		Map<String, Object> data = request.getData();
		if (data == null || data.size() != 5)
			throw new UiException(MZk.ILLEGAL_REQUEST_WRONG_DATA, new Object[]{Objects.toString(data), this});
		String sheetId = (String) data.get("sheetId");
		int start = (int) data.get("start");
		int end = (int) data.get("end");
		String dir = (String) data.get("dir");
		boolean open = (boolean) data.get("open");

		Sheet sheet = ((Spreadsheet) comp).getSelectedSheet();
		if (!getSheetUuid(sheet).equals(sheetId))
			return;
		SSheet internalSheet = sheet.getInternalSheet();
		if ("row".equals(dir)) {
			int base = -1;
			for (int i = start; i < SpreadsheetVersion.EXCEL2007.getLastRowIndex(); i++) {
				SRow next = internalSheet.getRow(i);
				if (base < 0) {
					base = next.getOutlineLevel();
				}
				if (next.getOutlineLevel() == 0 || next.getOutlineLevel() < base) break;
				next.setHidden(!open);
			}
			if (internalSheet.isSummaryBelow()) {
				internalSheet.getRow(end + 1).setCollapsed(!open);
			} else {
				internalSheet.getRow(start - 1).setCollapsed(!open);
			}
		} else {
			int base = -1;
			for (int i = start; i < SpreadsheetVersion.EXCEL2007.getLastColumnIndex(); i++) {
				SColumn next = internalSheet.getColumn(i);
				if (base < 0) {
					base = next.getOutlineLevel();
				}
				if (next.getOutlineLevel() == 0 || next.getOutlineLevel() < base) break;
				next.setHidden(!open);
			}
			if (internalSheet.isSummaryRight()) {
				SColumn next = internalSheet.getColumn(end + 1);
				next.setCollapsed(!open);
			} else {
				SColumn prev = internalSheet.getColumn(start - 1);
				prev.setCollapsed(!open);
			}
		}
		org.zkoss.zk.ui.event.Events.postEvent(new GroupOpenEvent(request.getCommand(), comp, sheet,
				start, end, dir, open));
		new NotifyChangeHelper().notifySheetGroupChange(sheet.getInternalSheet());
	}
}
