/* GroupHeaderClickEvent.java

        Purpose:
                
        Description:
                
        History:
                2023/11/2 AM 09:42, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui.event;

import java.util.List;

import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.model.Sheet;

import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.event.Event;

/**
 * Represents an event fired by a user clicking on a button of group head to collapse or expand all groups.
 * @author jumperchen
 * @since 6.0.0
 */
public class GroupHeaderClickEvent extends Event {
	private final Sheet sheet;
	private final int level;
	private final String dir;
	public GroupHeaderClickEvent(String name, Component target, Sheet sheet, int level, String dir) {
		super(name, target);
		this.sheet = sheet;
		this.level = level;
		this.dir = dir;
	}

	/**
	 * Returns the current sheet.
	 */
	public Sheet getSheet() {
		return sheet;
	}

	/**
	 * Returns whether it's a row-based outline.
	 * @return true a row base, unless a column base.
	 */
	public boolean isRow() {
		return "row".equals(dir);
	}

	/**
	 * Returns the level triggered by the click event.
	 */
	public int getOutlineLevel() {
		return level;
	}

	/**
	 * Returns all the collapsed groups in rows or columns.
	 */
	public List<Range> getCollapsedGroups() {
		if (isRow()) {
			return Ranges.range(getSheet()).getRowGroups(level + 1, 8);
		} else {
			return Ranges.range(getSheet()).getColumnGroups(level + 1, 8);
		}
	}

	/**
	 * Returns all the expanded groups in rows or columns.
	 */
	public List<Range> getExpandedGroups() {
		if (isRow()) {
			return Ranges.range(getSheet()).getRowGroups(1, level + 1);
		} else {
			return Ranges.range(getSheet()).getColumnGroups(1, level + 1);
		}
	}
}
