/* GroupOpenEvent.java

        Purpose:
                
        Description:
                
        History:
                2023/11/2 AM 09:03, Created by jumperchen

Copyright (C) 2023 Potix Corporation. All Rights Reserved.
*/
package io.keikai.ui.event;

import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.model.Sheet;

import org.zkoss.zk.ui.Component;
import org.zkoss.zk.ui.event.Event;

/**
 * Represents an event fired by a user clicking on a button of group row or column to collapse or expand a group.
 * @author jumperchen
 * @since 6.0.0
 */
public class GroupOpenEvent extends Event {
	private final Sheet sheet;
	private final int startIndex;
	private final int endIndex;
	private final String dir;
	private final boolean open;

	public GroupOpenEvent(String name, Component target, Sheet sheet, int startIndex, int endIndex, String dir, boolean open) {
		super(name, target);
		this.sheet = sheet;
		this.startIndex = startIndex;
		this.dir = dir;
		this.open = open;
		this.endIndex = endIndex;
	}

	/**
	 * Returns whether is opened or not for the group.
	 */
	public boolean isOpen() {
		return this.open;
	}

	/**
	 * Returns the group range in a whole rows or columns based on {@link #isRow()}.
	 */
	public Range getRange() {
		if (isRow()) {
			return Ranges.range(sheet, startIndex, 0, endIndex, sheet.getBook().getMaxColumns());
		} else {
			return Ranges.range(sheet, 0, startIndex, sheet.getBook().getMaxRows(), endIndex);
		}
	}

	/**
	 * Returns whether it's a row-based outline.
	 * @return true a row base, unless a column base.
	 */
	public boolean isRow() {
		return "row".equals(dir);
	}
}
