package io.keikai.ui.impl.ua;

import io.keikai.api.model.Book;
import io.keikai.api.model.Sheet;
import io.keikai.model.SAutoFilter;
import io.keikai.model.STable;
import io.keikai.model.impl.AbstractSheetAdv;
import io.keikai.range.SRange;
import io.keikai.range.SRanges;
import io.keikai.ui.UserActionContext;

public abstract class AbstractFilterHandler extends AbstractHandler {
    @Override
    public boolean isEnabled(Book book, Sheet sheet) {
        return book != null && sheet != null;
    }

    /**
     * get auto filter of a table at the selection first, then a sheet's auto filter
     */
    protected SAutoFilter getAutoFilter(UserActionContext context) {
        AbstractSheetAdv internalSheet = (AbstractSheetAdv) context.getSheet().getInternalSheet();
        final String rangeStr = (String) context.getData("range");
        final SRange range = SRanges.range(internalSheet, rangeStr);
        STable table = internalSheet.getTableByRowCol(range.getRow(), range.getColumn());

        if (table != null && table.getAutoFilter() != null){
            return table.getAutoFilter();
        }
        // KEIKAI-230
        if (internalSheet.getAutoFilter() == null) {
            throw new NullPointerException("Cannot find auto filter in range " + rangeStr);
        }
        return internalSheet.getAutoFilter();
    }
}

