/* FontFamilyAction.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2013/8/5 , Created by dennis
}}IS_NOTE

Copyright (C) 2013 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.ui.impl.ua;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import io.keikai.api.AreaRefWithType;
import io.keikai.api.CellOperationUtil;
import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.model.Font.Boldweight;
import io.keikai.api.model.Sheet;
import io.keikai.ui.CellSelectionType;
import io.keikai.ui.UserActionContext;
import io.keikai.ui.impl.ActionHelper;
import io.keikai.ui.impl.undo.AggregatedAction;
import io.keikai.ui.impl.undo.FontStyleAction;
import io.keikai.ui.sys.UndoableAction;
import io.keikai.ui.sys.UndoableActionManager;

import org.zkoss.util.resource.Labels;

/**
 * @author dennis
 *
 */
public class FontBoldHandler extends AbstractCellHandler {
	private static final long serialVersionUID = 4224308822648778702L;

	@Override
	protected boolean processAction(UserActionContext ctx) {
		Sheet sheet = ctx.getSheet();
		Set<AreaRefWithType> allSelection = ctx.getAllSelection();
		List<UndoableAction> actions = new ArrayList<>();
		for (AreaRefWithType selection : allSelection) {
			CellSelectionType type = selection.getSelType();
			Range range = Ranges.range(sheet, selection);
			//ZSS-576
			if (range.isProtected() && !range.getSheetProtection().isFormatCellsAllowed()) {
				showProtectMessage();
				return true;
			}
			//zss-623, extends to row,column area
			switch (type) {
				case ROW:
					range = range.toRowRange();
					break;
				case COLUMN:
					range = range.toColumnRange();
					break;
				case ALL:
					//we don't allow to set whole sheet style, use column range instead
					range = range.toColumnRange();
			}
			selection = new AreaRefWithType(range.getRow(), range.getColumn(), range.getLastRow(), range.getLastColumn(), type);

			Boldweight bw = range.getCellStyle().getFont().getBoldweight();
			if (Boldweight.BOLD.equals(bw)) {
				bw = Boldweight.NORMAL;
			} else {
				bw = Boldweight.BOLD;
			}
			actions.add(new FontStyleAction("", sheet, selection.getRow(), selection.getColumn(), selection.getLastRow(), selection.getLastColumn(), CellOperationUtil.getFontBoldweightApplier(bw)));
			ActionHelper.collectRichTextStyleActions(range, CellOperationUtil.getRichTextFontBoldweightApplier(bw), actions);
		}
		AggregatedAction action = new AggregatedAction(Labels.getLabel("zss.undo.fontStyle"), actions.toArray(new UndoableAction[actions.size()]));
		UndoableActionManager uam = ctx.getSpreadsheet().getUndoableActionManager();
		uam.doAction(action);
		return true;
	}

}
