/* FontFamilyAction.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2014/8/18 , Created by henrichen
}}IS_NOTE

Copyright (C) 2014 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.ui.impl.ua;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import io.keikai.api.AreaRef;
import io.keikai.api.AreaRefWithType;
import io.keikai.api.CellOperationUtil;
import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.model.Font.TypeOffset;
import io.keikai.api.model.Sheet;
import io.keikai.ui.CellSelectionType;
import io.keikai.ui.UserActionContext;
import io.keikai.ui.impl.ActionHelper;
import io.keikai.ui.impl.undo.AggregatedAction;
import io.keikai.ui.impl.undo.FontStyleAction;
import io.keikai.ui.sys.UndoableAction;
import io.keikai.ui.sys.UndoableActionManager;

import org.zkoss.util.resource.Labels;

/**
 * Handle font super/sub operation.
 * @author henrichen
 * @since 3.6.0
 */
public class FontTypeOffsetHandler extends AbstractCellHandler {
	private static final long serialVersionUID = -8745551581702240409L;

	@Override
	protected boolean processAction(UserActionContext ctx) {
		final String offstr = (String) ctx.getData("typeOffset"); 
		TypeOffset offset = 
			"SUPER".equals(offstr) ? TypeOffset.SUPER :
			"SUB".equals(offstr) ? TypeOffset.SUB : TypeOffset.NONE;

		Sheet sheet = ctx.getSheet();
		Set<AreaRefWithType> allSelection = ctx.getAllSelection();
		List<UndoableAction> actions = new ArrayList<UndoableAction>();
		for (AreaRefWithType selection : allSelection) {
			CellSelectionType type = selection.getSelType();
			Range range = Ranges.range(sheet, selection);
			//ZSS-576
			if (range.isProtected() && !range.getSheetProtection().isFormatCellsAllowed()) {
				showProtectMessage();
				return true;
			}
			//zss-623, extends to row,column area
			switch (type) {
				case ROW:
					range = range.toRowRange();
					break;
				case COLUMN:
					range = range.toColumnRange();
					break;
				case ALL:
					//we don't allow to set whole sheet style, use column range instead
					range = range.toColumnRange();
			}
			selection = new AreaRefWithType(range.getRow(), range.getColumn(), range.getLastRow(), range.getLastColumn(), type);
			actions.add(new FontStyleAction("", sheet, selection.getRow(), selection.getColumn(), selection.getLastRow(), selection.getLastColumn(), CellOperationUtil.getFontTypeOffsetApplier(offset)));
			ActionHelper.collectRichTextStyleActions(range, CellOperationUtil.getRichTextFontTypeOffsetApplier(offset), actions);
		}
		UndoableActionManager uam = ctx.getSpreadsheet().getUndoableActionManager();
		AggregatedAction action = new AggregatedAction(Labels.getLabel("zss.undo.fontStyle"), actions.toArray(new UndoableAction[actions.size()]));
		uam.doAction(action);
		return true;
	}
}
