/* TextIndentHandler.java

{{IS_NOTE
	Purpose:
		
	Description:
		
	History:
		2014/3/23 , Created by JerryChen
}}IS_NOTE

Copyright (C) 2015 Potix Corporation. All Rights Reserved.

{{IS_RIGHT
}}IS_RIGHT
*/
package io.keikai.ui.impl.ua;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import io.keikai.api.AreaRefWithType;
import io.keikai.api.CellOperationUtil;
import io.keikai.api.Range;
import io.keikai.api.Ranges;
import io.keikai.api.model.Sheet;
import io.keikai.ui.CellSelectionType;
import io.keikai.ui.UserActionContext;
import io.keikai.ui.impl.undo.AggregatedAction;
import io.keikai.ui.impl.undo.CellStyleAction;
import io.keikai.ui.sys.UndoableAction;
import io.keikai.ui.sys.UndoableActionManager;

import org.zkoss.util.resource.Labels;

/**
 * @author JerryChen
 */
public class TextIndentHandler extends AbstractCellHandler {
	private static final long serialVersionUID = -2346819389887059442L;
	private int offset;
	
	
	public TextIndentHandler(int offset) {
		this.offset = offset;
	}


	/* (non-Javadoc)
	 * @see io.keikai.ui.sys.ua.impl.AbstractHandler#processAction(io.keikai.ui.UserActionContext)
	 */
	@Override
	protected boolean processAction(UserActionContext ctx) {
		Sheet sheet = ctx.getSheet();
		Set<AreaRefWithType> allSelection = ctx.getAllSelection();
		List<UndoableAction> actions = new ArrayList<>();
		for (AreaRefWithType selection : allSelection) {
			CellSelectionType type = selection.getSelType();
			Range range = Ranges.range(sheet, selection);
			//ZSS-576
			if (range.isProtected() && !range.getSheetProtection().isFormatCellsAllowed()) {
				showProtectMessage();
				return true;
			}
			//zss-623, extends to row,column area
			switch (type) {
				case ROW:
					range = range.toRowRange();
					break;
				case COLUMN:
					range = range.toColumnRange();
					break;
				case ALL:
					//we don't allow to set whole sheet style, use column range instead
					range = range.toColumnRange();
			}
			selection = new AreaRefWithType(range.getRow(), range.getColumn(), range.getLastRow(), range.getLastColumn(), type);
			actions.add(new CellStyleAction(Labels.getLabel("zss.undo.cellStyle"), sheet, selection.getRow(), selection.getColumn(), selection.getLastRow(), selection.getLastColumn(), CellOperationUtil.getIndentionApplier(offset)));
		}
		AggregatedAction action = new AggregatedAction(Labels.getLabel("zss.undo.cellStyle"), actions.toArray(new UndoableAction[actions.size()]));
		UndoableActionManager uam = ctx.getSpreadsheet().getUndoableActionManager();
		uam.doAction(action);
		return true;
	}

}
